/*
  Kalman.c : SiGN-SSM - Gene Network Estimation with State Sapce Model

  Copyright (C) 2010
    Yoshinori Tamada, Rui Yamaguchi, Seiya Imoto, Osamu Hirose, Ryo Yoshida
	Masao Nagasaki, and Satoru Miyano
      Laboratory of DNA Information Analysis &
	  Laboratory of Sequence Analysis, Human Genome Center,
      The Institute of Medical Science, The Unversity of Tokyo

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  
  Visit <http://sign.hgc.jp/signssm/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
  Originally developed by Osamu Hirose, Ryo Yoshida, Seiya Imoto, Rui Yamaguchi,
                          Tomoyuki Higuchi, and Satoru Miyano.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>

#include <math/SGN_Matrix.h>

#include "signcom.h"
#include "SSM.h"
#include "SSMData.h"
#include "Kalman.h"

#ifdef SSM_DEBUG
#include "debug.h"
#endif

#ifdef HGC
#include "hgc.h"
#endif
#include "pvalues.h"

#define Math_PI 3.14159265358979323846264

/* SSM
  x_n = F x_(n-1) + v_n,   v_n ~ N(0, Q)
  y_n = H x_n     + w_n,   w_n ~ N(0, R)
*/

/*!\defgroup Kalman Kalman
 *\brief Kalman filter and smoother.
 *
 *\sa SSM, SSMState, SSMSuffStat, SSMData, SSM_EM, SSMInit, SSMWork */

/* Proto-types for internal functions */

void Kalman_ddgimm(const double * A, double * B, const int p, const int k);
void Kalman_predict(double * F, double * Q, double * x, double * V,
					double * FV, double * px, double * pV, double * pVinv,
					int * pivot, double * work, int worksize, int k);
void Kalman_filter_observed(double * H, double * R, double * x, double * V, double * Vinv,
							double * tHRinvH, double * RinvH,
							double * e, double * tHRinve, double * Ke, 
							double * fx, double * fV, double * I_KH,
							const double * Y, int * missing, int p, int k,
							double * w_tHRinvH, double * M,	double * LU, int * pivot);
double Kalman_log_likelihood(double * R, double * e, double * Ke, double * tHRinve,
							 double * I_KH, int * missing,  
							 int p, int k, 
							 double * LU, int * pivot);
void Kalman_print_predict(int n, int k, 
						  double * exp, double * var, double * FV, double * Vinv);

void Kalman_print_filter_observed(int n, int p, int k, double * exp, double * var,
								  double * e, double * tHRinve, double * Ke, double * I_KH);

/*!\brief Kalman filter
 *
 * This routine performs the Kalman filter for a single replicate and
 * returns the log likelihood of the current models.
 *
 * Called by SSM_EM_estimate().
 *
 * A value in thd column position index table \a table[i] represents the
 * true column position of the matrix \a Y for the \a i th time point.
 *
 *\param ssm     [INPUT] SSM model parameters.
 *\param X       [OUTPUT] Not need to be zero-cleared.
 *\param p       Number of objects.
 *\param k       Size of dimension.
 *\param N       Number of time points.
 *\param Y       [INPUT] (\a p, \a s)-input data set matrix where \a s is the number of observed samples.
 *\param table   [INPUT] Column position index table for \a Y. 
 *\param missing [INPUT] Missing value table where \a missing[n] represents the list of rows of
 *                 missing values, terminating by -1.
 *\param work    [WORK] Working memory allocated by SSMWork_new().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup Kalman */
double Kalman_filter(SSM * ssm, SSMState * X, int p, int k, int N,
					 double * Y, int * table, int ** missing,  SSMWork * work){

#ifdef SSM_DEBUG
  //fprintf(SIGN_ERR, "Calculation of H' Rinv H\n");
#endif
  /* Calculates H' Rinv H */
  /* X->RinvH := (X->R)^inv H */
  /*   (p,k)  := (p,p)  (p,k) */
  memcpy(work->RinvH, ssm->H, sizeof(double) * p * k);
  /*            (p)      (p,k)        */
  Kalman_ddgimm(ssm->R, work->RinvH, p, k);
  /* X->tHRinvH := H' Rinv H */
  /*   (k,k)    := (k,p) (p,k) */
  SGN_Matrix_dgemm(SGN_Matrix_TRANS, SGN_Matrix_NO_TRANS, &k, &k, &p,
				   &SGN_Matrix_DOUBLE_ONE, ssm->H, &p,
				   work->RinvH, &p,
				   &SGN_Matrix_DOUBLE_ZERO, work->tHRinvH, &k);

  /* Set initial values */
  memcpy(X->filter_exp[0], ssm->exp, sizeof(double) * k);
  memcpy(X->filter_var[0], ssm->var, sizeof(double) * k * k);

  double llh = 0.0; /* Log likelihood */

  /* Visit all the time points. */
  for(int n = 0; n < N; n++){
	/* Prediction */
	Kalman_predict(ssm->F, ssm->Q, X->filter_exp[n], X->filter_var[n],         /* input */
				   X->FV[n], X->predict_exp[n], X->predict_var[n], X->Vinv[n], /* output */
				   work->pivot, work->dgetri_work, work->dgetri_work_size, k);

#ifdef SSM_DEBUG
	Kalman_print_predict(n, k, X->predict_exp[n], X->predict_var[n], X->FV[n], X->Vinv[n]);
	if(n == 0){
	  char buff[SSM_BUFF_SIZE];
	  snprintf(buff, SSM_BUFF_SIZE, "Kalman:Filter:RinvH");
	  debug_check(buff, work->RinvH, p, k, SSM_DEBUG, 0);
	  snprintf(buff, SSM_BUFF_SIZE, "Kalman:Filter:tHRinvH");
	  debug_check(buff, work->tHRinvH, k, k, SSM_DEBUG, 0);
	}
	//fprintf(SIGN_ERR, "table[n=%d] = %d\n", n, table[n]);
#endif

	/* Filtering */
	if(table[n] != -1){
	  /* Observed time point */
	  double * Yn = Y + table[n] * p;

	  Kalman_filter_observed(ssm->H, ssm->R,
							 X->predict_exp[n], X->predict_var[n], X->Vinv[n],
							 work->tHRinvH, work->RinvH, 
							 work->e, work->tHRinve, work->Ke, 
							 X->filter_exp[n+1], X->filter_var[n+1], work->I_KH,
							 Yn, missing[table[n]], p, k,
							 work->w_tHRinvH, work->M, work->LU, work->pivot);
#ifdef SSM_DEBUG
	  Kalman_print_filter_observed(n, p, k, X->filter_exp[n + 1], X->filter_var[n + 1],
								   work->e, work->tHRinve, work->Ke, work->I_KH);
#endif
	  llh += Kalman_log_likelihood(ssm->R, work->e, work->Ke, work->tHRinve,
								   work->I_KH, missing[table[n]], p, k, 
								   work->LU, work->pivot);
#ifdef SSM_DEBUG
	  char buff[SSM_BUFF_SIZE];
	  snprintf(buff, SSM_BUFF_SIZE, "Kalman:Filter:llh");
	  debug_check(buff, &llh, 1, 1, SSM_DEBUG, 0);
#endif
	} else {
	  /* Unobserved (missing) time point
		   X->filter_exp[n + 1] := X->predict_exp[n]
		   X->filter_var[n + 1] := X->predict_var[n]  */
	  memcpy(X->filter_exp[n + 1], X->predict_exp[n], sizeof(double) * k);
	  memcpy(X->filter_var[n + 1], X->predict_var[n], sizeof(double) * k * k);
#ifdef SSM_DEBUG
	  char buff[SSM_BUFF_SIZE];
	  snprintf(buff, SSM_BUFF_SIZE, "Kalman:Filter:filter_exp[%d]", n + 1);
	  debug_check(buff, X->filter_exp[n + 1], k, 1, SSM_DEBUG, 0);
	  snprintf(buff, SSM_BUFF_SIZE, "Kalman:Filter:filter_var[%d]", n + 1);
	  debug_check(buff, X->filter_var[n + 1], k, k, SSM_DEBUG, 0);
#endif
	}
  }/* for n < N */
  return llh;
}
/* Kalman_filter() ========================================================= */

/*!\brief Kalman Prediction.
 *
 * This calculates the followings:<br>
 * \a FV := \a F \a V <br>
 * \a px := \a F \a x <br>
 * \a pV := \a F \a V \a F ' + diag(\a Q) <br>
 * \a pVinv := \a pV <sup>-1</sup> = (\a F \a V \a F ' + diag(\a Q)) <sup>-1</sup> <br>
 * where \a x is assumed to be SSMState::filter_exp[<i>n</i>], \a px, SSMState::predict_exp[<i>n</i>]. 
 *
 * Called by Kalman_filter().
 *
 *\param F [INPUT] (\a k, \a k)-matrix. System matrix. SSM::F.
 *\param Q [INPUT] (\a k)-vector. System noise. SSM::Q.
 *\param x [INPUT] (\a k)-vector. SSMState::filter_exp.
 *\param V [INPUT] (\a k, \a k)-matrix. SSMState::filter_var.
 *\param FV [OUTPUT] (\a k, \a k)-matrix. SSMState::FV.
 *\param px [OUTPUT] (\a k)-vector. SSMState::predict_exp.
 *\param pV [OUTPUT] (\a k, \a k)-vector. SSMState::predict_var.
 *\param pVinv [OUTPUT] (\a k, \a k)-vector. Inverse of \a pV.
 *\param pivot [WORK] (\a k)-integer vector.
 *\param work [WORK] Work space for LU decomposition (dgetri) of \a pV (\a k by \a k matrix).
 *\param worksize  size of \a work.
 *\param k  Size of dimension.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup Kalman */
void Kalman_predict(double * F, double * Q, double * x, double * V,
					double * FV, double * px, double * pV, double * pVinv,
					int * pivot, double * work, int worksize, int k){
  int info;

  /* F %*% V -> FV */
  /* (k,k) %*% (k,k) -> (k,k) */
  SGN_Matrix_dgemm(SGN_Matrix_NO_TRANS, SGN_Matrix_NO_TRANS, &k, &k, &k,
				   &SGN_Matrix_DOUBLE_ONE, F, &k, V, &k,
				   &SGN_Matrix_DOUBLE_ZERO, FV, &k);

  /* Prediction (mean) */
  /* F %*% x -> px */
  /* (k,k) %*% (k,1) -> (k,1) */
  SGN_Matrix_dgemv(SGN_Matrix_NO_TRANS, &k, &k, 
				   &SGN_Matrix_DOUBLE_ONE, F, &k, 
				   x, &SGN_Matrix_INT_ONE,
				   &SGN_Matrix_DOUBLE_ZERO, px, &SGN_Matrix_INT_ONE);

  /* Prediction (variance) */
  /* FV %*% (F)' + Q -> pV */
  /* (k,k) %*% (k,k) + (k,k) -> (k,k) */
  SGN_Matrix_diag(k, SGN_Matrix_DOUBLE_ONE, Q, pV, k);
  SGN_Matrix_dgemm(SGN_Matrix_NO_TRANS, SGN_Matrix_TRANS, &k, &k, &k,
				   &SGN_Matrix_DOUBLE_ONE, FV, &k, F, &k,
				   &SGN_Matrix_DOUBLE_ONE, pV, &k);

  /* pV^-1 -> pVinv */
  /* (k,k)^-1 -> (k,k) */
  /*    LU factorization of pV */
  memcpy(pVinv, pV, sizeof(double) * k * k);
  SGN_Matrix_dgetrf(&k, &k, pVinv, &k, pivot, &info);
  /*    Inverse the LU-factorized matrix */
  SGN_Matrix_dgetri(&k, pVinv, &k, pivot, work, &worksize, &info);
}
/* Kalman_predict() ======================================================== */

/*!\brief inverse of diagonal matrix x general matrix.
 *
 * Calculates (diag(\a A)<sup>-1</sup>) x \a B and set it to \a B.
 *
 *\param A (\a p)-vector.
 *\param B (\a p, \a k)-matrix. This will be over-witten.
 *\param p length of vector \a A.
 *\param k rows of matrix \a B.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup Kalman */
void Kalman_ddgimm(const double * A, double * B, const int p, const int k){
  for(int i = 0; i < p; i++){
	for(int j = 0; j < k; j++){
	  B[i + j * p] = B[i + j * p] / A[i];
	}
  }
}
/* Kalman_ddgimm() ========================================================= */

/*!\brief Subroutine for Kalman_filter().
 *
 * This is called by Kalman_filter() for the observed time points.
 *
 *\param H  [INPUT] (\a p,\a k)-matrix. SSM::H.
 *\param R  [INPUT] (\a) p-vector. SSM::R.
 *\param x [INPUT] (\a k)-vector. Assumed SSMState::predict_exp[n - 1].
 *\param V [INPUT] (\a k, \a k)-matrix. Assumed SSMState::predict_var[n - 1].
 *\param Vinv [INPUT]  (\a k, \a k)-matrix. SSMState::Vinv.
 *\param tHRinvH [INPUT] (\a k, \a k)-matrix. SSMState::tHRinvH.
 *\param RinvH [INPUT] (\a p, \a k)-matrix. SSMState::RinvH.
 *
 *\param e [OUTPUT] (\a p)-vector containing residues.
 *\param tHRinve [OUTPUT] (\a k)-vector.
 *\param Ke [OUTPUT] (\a k)-vector.
 *\param fx [OUTPUT] (\a k)-vector. SSMState::filter_exp[n].
 *\param fV [OUTPUT] (\a k, \a k)-matrix. SSMState::filter_var[n].
 *\param I_KH [OUTPUT] (\a k, \a k)-matrix.
 *
 *\param Y  [INPUT] (\a p, 1)-observed data vector.
 *\param missing [INPUT] Missing rows in \a Y. -1 represents the end of the array.
 *\param p [INPUT] Number of objects (nodes, or genes).
 *\param k [INPUT] Size of dimention.
 *
 *\param w_tHRinvH [WORK] (\a k, \a k)-matrix.
 *\param M [WORK] (\a k, \a k)-matrix.
 *\param LU [WORK] (\a k, \a k)-matrix.
 *\param pivot [WORK] (\a k)-integer vector.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup Kalman */
void Kalman_filter_observed(double * H, double * R, double * x, double * V, double * Vinv,
							double * tHRinvH, double * RinvH,
							double * e, double * tHRinve, double * Ke, 
							double * fx, double * fV, double * I_KH,
							const double * Y, int * missing, int p, int k,
							double * w_tHRinvH, double * M,	double * LU, int * pivot){

  int k_k = k * k;

  /* e := Y - H x (residue) */  /* x = px[n-1] */
  /*  [ (p,1) := (p,1) - (p,k) %*% (k,1) ] */
  memcpy(e, Y, sizeof(double) * p);                      /* e := Y */
  SGN_Matrix_dgemv(SGN_Matrix_NO_TRANS, &p, &k,
				   &SGN_Matrix_DOUBLE_MINUS_ONE, H, &p, x, &SGN_Matrix_INT_ONE, 
				   &SGN_Matrix_DOUBLE_ONE, e, &SGN_Matrix_INT_ONE); /* e := - H x  + e
																	      = Y - H x    */
  /*
  fprintf(SIGN_ERR, "Kalman_filter_observed: Y = \n");
  SGN_Matrix_fprintf(SIGN_ERR, 15, Y, p, 1);
  fprintf(SIGN_ERR, "\n");
  fprintf(SIGN_ERR, "Kalman_filter_observed: H = \n");
  SGN_Matrix_fprintf(SIGN_ERR, 15, H, p, k);
  fprintf(SIGN_ERR, "\n");
  fprintf(SIGN_ERR, "Kalman_filter_observed: x = \n");
  SGN_Matrix_fprintf(SIGN_ERR, 15, x, k, 1);
  fprintf(SIGN_ERR, "\n");
  fprintf(SIGN_ERR, "Kalman_filter_observed: e = \n");
  SGN_Matrix_fprintf(SIGN_ERR, 15, e, p, 1);
  fprintf(SIGN_ERR, "\n");*/

  memcpy(w_tHRinvH, tHRinvH, sizeof(double) * k_k);    /* w_tHRinvH := tHRinvH */
  for(int ii = 0; missing[ii] != -1; ii++){
	int i = missing[ii];
	double v = -1.0 / R[i];
	e[i] = 0.0;
	SGN_Matrix_dger(&k, &k, &v, H + i, &p,
					H + i, &p, w_tHRinvH, &k);  /* w_tHRinvH := w_tHRinvH - H_i H_i' / R_i */
  }

  /* tHRinve := (Rinv H)' e = H' Rinv' e */
  /*  (k,1)  :=  (k,p) %*% (p,1) */
  SGN_Matrix_dgemv(SGN_Matrix_TRANS, &p, &k,
				   &SGN_Matrix_DOUBLE_ONE, RinvH, &p,
				   e, &SGN_Matrix_INT_ONE,
				   &SGN_Matrix_DOUBLE_ZERO, tHRinve, &SGN_Matrix_INT_ONE);

  /* M := w_tHRinvH + Vinv = H' Rinv H + Vinv */
  /* (k,k) := (k,k) + (k,k) */
  memcpy(M, Vinv, sizeof(double) * k_k);
  SGN_Matrix_daxpy(&k_k, &SGN_Matrix_DOUBLE_ONE, w_tHRinvH, &SGN_Matrix_INT_ONE,
				   M, &SGN_Matrix_INT_ONE);

  /* Ke :=  Minv (H' Rinv e) */ /* by solving M Ke = H' Rinv e  for Ke */
  /* (k,1) := (k,k) (k,1) */   

  int info;
  memcpy(LU, M, sizeof(double) * k_k);      /* LU := M = H' Rinv H + Vinv */
  memcpy(Ke, tHRinve, sizeof(double) * k);
  SGN_Matrix_dgesv(&k, &SGN_Matrix_INT_ONE, LU, &k, pivot, Ke, &k, &info);
                                                      /* Solve  M Ke = H' Rinv e  for  Ke */
                                                      /*    i.e   Ke = (M)^-1 H' Rinv e */
                                                      /* Here, LU is the factorized M */  

  /* fx := x + Ke */
  /* (k,1) := (k,1) + (k,1) */
  memcpy(fx, x, sizeof(double) * k);
  SGN_Matrix_daxpy(&k, &SGN_Matrix_DOUBLE_ONE, Ke, &SGN_Matrix_INT_ONE,
				   fx, &SGN_Matrix_INT_ONE);

  /* I_KH := I - (Minv H' Rinv H) */   /* = I - ((H' Rinv H + Vinv)inv H' Rinv H) */
  /*   At first, solve  M I_KH = H' Rinv H  for  I_KH. */
  /*                 i.e. I_KH = Minv H' Rinv H */
  /*   Recycling M.  LU contains Minv. */
  memcpy(M, w_tHRinvH, sizeof(double) * k_k); /* M := H' Rinv H */ 
  SGN_Matrix_dgetrs(SGN_Matrix_L_NO_TRANS, &k, &k, LU, &k, pivot, M, &k, &info); 
                                                         /* M := Minv H' Rinv H */
  SGN_Matrix_I(k, I_KH); /* M := I */
  SGN_Matrix_daxpy(&k_k, &SGN_Matrix_DOUBLE_MINUS_ONE, M, &SGN_Matrix_INT_ONE,
				   I_KH, &SGN_Matrix_INT_ONE);           /* I_KH := I - (Minv H' Rinv H) */
			   
  /* fV := I_KH V */
  SGN_Matrix_dgemm(SGN_Matrix_NO_TRANS, SGN_Matrix_NO_TRANS, &k, &k, &k,
				   &SGN_Matrix_DOUBLE_ONE, I_KH, &k, V, &k,
				   &SGN_Matrix_DOUBLE_ZERO, fV, &k);
}
/* Kalman_filter_observed() ================================================ */

/*!\brief Calculates log likelihood.
 *
 * This is assumed to be called after Kalman_filter_observed() from Kalman_filter().
 *
 *\param R [INPUT] (\a p)-vector.
 *\param e [INPUT] (\a p)-vector.
 *\param Ke [INPUT] (\a k)-vector.
 *\param tHRinve [INPUT] (\a k)-vector.
 *\param I_KH [INPUT] (\a k, \a k)-matrix.
 *\param missing
 *\param p  Number of objects.
 *\param k  Size of dimension.
 *\param LU [WORK] (\a k, \a k)-matrix.
 *\param pivot [WORK] \a k integer vector.
 *
 *\return Log likelihood.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup Kalman */
double Kalman_log_likelihood(double * R, double * e, double * Ke, double * tHRinve,
							 double * I_KH, int * missing,  
							 int p, int k, 
							 double * LU, int * pivot){
  double llh = 0.0;
  double v = 0.0;

#ifdef SSM_DEBUG
  //fprintf(SIGN_ERR, "Kalman_log_likelihood: e = \n");
  //SGN_Matrix_fprintf(SIGN_ERR, 15, e, p, 1);
  //fprintf(SIGN_ERR, "\n");
  //fprintf(SIGN_ERR, "Kalman_log_likelihood: R = \n");
  //SGN_Matrix_fprintf(SIGN_ERR, 15, R, p, 1);
  //fprintf(SIGN_ERR, "\n");
#endif

  /* Quadratic part */
  for(int i = 0; i < p; i++){
	v += e[i] * e[i] / R[i];
	//fprintf(SIGN_ERR, "v=%.15f\n", v);
  }
  double v2 = SGN_Matrix_ddot(&k, tHRinve, &SGN_Matrix_INT_ONE, Ke, &SGN_Matrix_INT_ONE);
  llh += v - v2;
#ifdef SSM_DEBUG
  //fprintf(SIGN_ERR, "Kalman_log_likelihood: quadratic part llh = %f (= %f - %f)\n", 
  //llh, v, v2);
#endif
    
  /* Determinant of R */
  v = 0.0;
  for(int i = 0; i < p; i++){
	v += log(R[i]);
  }
  int nna = 0;
  for(int ii = 0; missing[ii] != -1; ii++){
	nna++;
	int i = missing[ii];
	v -= log(R[i]);
  }
  llh += v;

#ifdef SSM_DEBUG
  //fprintf(SIGN_ERR, "Kalman_log_likelihood: det R llh = %f\n", v);
#endif

  /* constant */
  v = (double) (p - nna) * log(2.0 * Math_PI);
  llh += v;
#ifdef SSM_DEBUG
  //fprintf(SIGN_ERR, "Kalman_log_likelihood: constant llh = %f\n", v);
#endif

  /* Determinant of I - KH */
  int info;
  memcpy(LU, I_KH, sizeof(double) * k * k);
  SGN_Matrix_dgetrf(&k, &k, LU, &k, pivot, &info);
  v = 0.0;
  int pos = 0;
  for(int i = 0; i < k; i++){
	v += log(fabs(LU[pos]));
	pos += (k + 1);
  }
  llh -= v;
  llh /= -2.0;

  return llh;
}
/* Kalman_log_likelihood() ================================================== */

/*!\brief Kalman smoother.
 *
 *
 * Updating SSMState::smooth_exp, SSMState::smooth_var, and Sg.
 *
 * Called by SSM_EM_estimate().
 *
 * Following fields in \a X are used for the calculation as inputs:
 * - SSMState::filter_exp[n]
 * - SSMState::filter_var[n]
 * - SSMState::predict_exp[n]
 * - SSMState::predict_var[n]
 * - SSMState::Vinv[n]
 *
 * Then updates followings:
 * - SSMState::smooth_exp[n]
 * - SSMState::smooth_var[n]
 * - SSMState::smooth_gain[n]
 *
 *\param X  [INPUT/OUTPUT] SSM state variables.
 *\param k  Size of dimension.
 *\param N  Number of time points.
 *\param w   [WORK] (\a k)-vector
 *\param W1  [WORK] (\a k, \a k)-matrix
 *\param W2  [WORK] (\a k, \a k)-matrix
 * 
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup Kalman */
void Kalman_smoother(SSMState * X, int k, int N, double * w, double * W1, double * W2){

  int k_k = k * k;

  /* Set Initial Values */
  memcpy(X->smooth_exp[N], X->filter_exp[N], sizeof(double) * k);
  memcpy(X->smooth_var[N], X->filter_var[N], sizeof(double) * k_k);
  for(int n = N - 1; n >= 0; n--){
	/* Smooth Gain */
    /* SG := FV' pVinv = (F fV)' pVinv */  /* fV: filter exp, pV: predict exp */
	/* (k,k) := (k,k) (k,k) */
	SGN_Matrix_dgemm(SGN_Matrix_TRANS, SGN_Matrix_NO_TRANS, &k, &k, &k,
					 &SGN_Matrix_DOUBLE_ONE, X->FV[n], &k,
					 X->Vinv[n], &k,
					 &SGN_Matrix_DOUBLE_ZERO, X->smooth_gain[n], &k);

	/* Update smooth_exp (x) */
	memcpy(w, X->smooth_exp[n + 1], sizeof(double) * k);  /* w := sx; (k)-vector */
	SGN_Matrix_daxpy(&k, &SGN_Matrix_DOUBLE_MINUS_ONE, X->predict_exp[n], &SGN_Matrix_INT_ONE,
					 w, &SGN_Matrix_INT_ONE);             /* w := sx - px */
	memcpy(X->smooth_exp[n], X->filter_exp[n], sizeof(double) * k);
	SGN_Matrix_dgemv(SGN_Matrix_NO_TRANS, &k, &k, 
					 &SGN_Matrix_DOUBLE_ONE, X->smooth_gain[n], &k,
					 w, &SGN_Matrix_INT_ONE, 
					 &SGN_Matrix_DOUBLE_ONE, X->smooth_exp[n], &SGN_Matrix_INT_ONE);       
	                                         /* x[n] := SG (sx - pv) + fx */
	/* Update smooth_var (V) */
	memcpy(W1, X->smooth_var[n + 1], sizeof(double) * k_k); /* W1 := sV_n1; (k,k)-matrix */
	SGN_Matrix_daxpy(&k_k, &SGN_Matrix_DOUBLE_MINUS_ONE, X->predict_var[n], &SGN_Matrix_INT_ONE,
					 W1, &SGN_Matrix_INT_ONE);              /* W1 := sV_n1 - pV */
	SGN_Matrix_dgemm(SGN_Matrix_NO_TRANS, SGN_Matrix_NO_TRANS, &k, &k, &k,
					 &SGN_Matrix_DOUBLE_ONE, X->smooth_gain[n], &k,
					 W1, &k,
					 &SGN_Matrix_DOUBLE_ZERO, W2, &k);    /* W2 := SG (sV_n1 - pV); (k,k)-matrix */
	SGN_Matrix_dgemm(SGN_Matrix_NO_TRANS, SGN_Matrix_TRANS, &k, &k, &k,
					 &SGN_Matrix_DOUBLE_ONE, W2, &k,
					 X->smooth_gain[n], &k,
					 &SGN_Matrix_DOUBLE_ZERO, W1, &k);    /* W1 := SG (sV - pV) SG' */
	memcpy(X->smooth_var[n], X->filter_var[n], sizeof(double) * k_k); /* sV_n := fV     */
	SGN_Matrix_daxpy(&k_k, &SGN_Matrix_DOUBLE_ONE, W1, &SGN_Matrix_INT_ONE,
					 X->smooth_var[n], &SGN_Matrix_INT_ONE);
                                                        /* sV_n := fV + SG (sV_n1 - pV) SG' */

#ifdef SSM_DEBUG
	char buff[SSM_BUFF_SIZE];
	snprintf(buff, SSM_BUFF_SIZE, "Kalman:Smoother:smooth_exp[%d]", n);
	debug_check(buff, X->smooth_exp[n], k, 1, SSM_DEBUG, 0);
	snprintf(buff, SSM_BUFF_SIZE, "Kalman:Smoother:smooth_var[%d]", n);
	debug_check(buff, X->smooth_var[n], k, k, SSM_DEBUG, 0);
	snprintf(buff, SSM_BUFF_SIZE, "Kalman:Smoother:smooth_gain[%d]", n);
	debug_check(buff, X->smooth_gain[n], k, k, SSM_DEBUG, 0);
#endif
  }/* for n >= 0 */
}
/* Kalman_smoother() ======================================================== */

/*!\brief Calculates SSMState::lagone_cov in \a X.
 *
 * This calculates SSMState::lagone_cov[\a n] (\a k by \a k matrix) from
 *  SSMState::smooth_gain[\a n] and SSMState::smooth_var[\a n] for \a n = 0, ... \a N - 1.
 *
 * This is called by SSM_EM_estimate().
 *
 *\param X  [INPUT/OUTPUT] SSM state variables.
 *\param k  Size of dimension.
 *\param N  Number of time points.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup Kalman */
void Kalman_lagone_cov(SSMState * X, int k, int N){

  for(int n = 0; n < N; n++){
	/* X->lagone_cov[n] := X->smooth_gain[n] X->smooth_var[n+1] */
	/*  ( smooth_gain := F V' pVinv ) */
	SGN_Matrix_dgemm(SGN_Matrix_NO_TRANS, SGN_Matrix_NO_TRANS, &k, &k, &k,
					 &SGN_Matrix_DOUBLE_ONE, X->smooth_gain[n], &k,
					 X->smooth_var[n+1], &k,
					 &SGN_Matrix_DOUBLE_ZERO, X->lagone_cov[n], &k);

#ifdef SSM_DEBUG
	char buff[SSM_BUFF_SIZE];
	snprintf(buff, SSM_BUFF_SIZE, "Kalman:lagone_cov:lagone_cov[%d]", n);
	debug_check(buff, X->lagone_cov[n], k, k, SSM_DEBUG, 0);
#endif
				 
  }/* for n < N */
}
/* Kalman_lagone_cov() ====================================================== */

/*!\brief [DEBUG] Prints the results of Kalman prediction.
 *
 *\param n    Time point.
 *\param k    Size of dimension.
 *\param exp  (\a k)-expectation vector.
 *\param var  (\a k, \a k)-variance matrix.
 *\param FV   (\a k, \a k)-matrix.
 *\param Vinv (\a k, \a k)-matrix.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup Kalman */
void Kalman_print_predict(int n, int k, 
						  double * exp, double * var, double * FV, double * Vinv){
#ifdef SSM_DEBUG
  char buff[SSM_BUFF_SIZE];
  snprintf(buff, SSM_BUFF_SIZE, "Kalman:Filter:predict_exp[%d]", n);
  debug_check(buff, exp, k, 1, SSM_DEBUG, 0);
  snprintf(buff, SSM_BUFF_SIZE, "Kalman:Filter:predict_var[%d]", n);
  debug_check(buff, var, k, k, SSM_DEBUG, 0);
  snprintf(buff, SSM_BUFF_SIZE, "Kalman:Filter:FV[%d]", n);
  debug_check(buff, FV, k, k, SSM_DEBUG, 0);
  snprintf(buff, SSM_BUFF_SIZE, "Kalman:Filter:Vinv[%d]", n);
  debug_check(buff, Vinv, k, k, SSM_DEBUG, 0);
#endif
}
/* Kalman_print_predict() ================================================== */

/*!\brief [DEBUG] Prints the results Kalman filter for the observed time point.
 *
 *\param n  Time point.
 *\param p  Number of objects.
 *\param k  Size of dimension.
 *\param exp  (\a k)-expectation vector.
 *\param var  (\a k, \a k)-variance matrix.
 *\param e    (\a p)-vector.
 *\param tHRinve (\a k)-vector.
 *\param Ke   (\a k)-vector.
 *\param I_KH  (\a k, \a k)-matrix.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup Kalman */
void Kalman_print_filter_observed(int n, int p, int k, double * exp, double * var,
								  double * e, double * tHRinve, double * Ke, double * I_KH){
#ifdef SSM_DEBUG
  char buff[SSM_BUFF_SIZE];
  snprintf(buff, SSM_BUFF_SIZE, "Kalman:Filter:filter_exp[%d]", n + 1);
  debug_check(buff, exp, k, 1, SSM_DEBUG, 0);
  snprintf(buff, SSM_BUFF_SIZE, "Kalman:Filter:filter_var[%d]", n + 1);
  debug_check(buff, var, k, k, SSM_DEBUG, 0);
  snprintf(buff, SSM_BUFF_SIZE, "Kalman:Filter:e[%d]", n + 1);
  debug_check(buff, e, p, 1, SSM_DEBUG, 0);
  snprintf(buff, SSM_BUFF_SIZE, "Kalman:Filter:tHRinve[%d]", n + 1);
  debug_check(buff, tHRinve, k, 1, SSM_DEBUG, 0);
  snprintf(buff, SSM_BUFF_SIZE, "Kalman:Filter:Ke[%d]", n + 1);
  debug_check(buff, Ke, k, 1, SSM_DEBUG, 0);
  snprintf(buff, SSM_BUFF_SIZE, "Kalman:Filter:I_KH[%d]", n + 1);
  debug_check(buff, I_KH, k, k, SSM_DEBUG, 0);
#endif
}
/* Kalman_print_filter_observed() ========================================== */

/*!\brief Applies the Kalman filter and smoother to generate filtering,
 *   prediction and smoothing data, and then outputs them.
 *
 * This calculates the estimated state vectors by
 * simply calling Kalman_filter() and Kalman_smoother() from
 * the estimated SSM model parameters, then
 * outputs the state vectors <i>x<sub>n</sub></i> and predicted data <i>y<sub>n</sub></i>.
 *
 * The state vectors and the predicted, filtering, smoothing data are generated for each
 * replicate data.
 * The total number of files to be generated is (# of replicates) x 7.
 *
 * The file names are formed:
 * &quot;<i>prefix</i><tt>.D</tt><i>000</i><tt>.S</tt><i>000</i><tt>.<i>X</i>.<i>Y</i>.</tt><i>postfix</i>&quot;
 * where <i>X</i> is one of
 *  - <tt>Xp</tt> : 
 *  - <tt>Xf</tt> : 
 *  - <tt>Xs</tt> :
 *  - <tt>Yp</tt> :
 *  - <tt>Yf</tt> :
 *  - <tt>Ys</tt> :
 *  - <tt>Y</tt> : the input time series data.
 *
 * and <i>Y</i> is a replicate number given as \a rep.
 *
 * If you want to calculate state vectors and predicted data for different
 * input data from one used for estimate the parameters, then
 * it is recommended to estimated the initial values from the new input data
 * by Kalman_infer_mu0().
 *
 *\param ssm  [INPUT]  SSM model parameters.  Unchanged on exit.
 *\param k    [INPUT]  Size of dimension.
 *\param Y    [INPUT]  Input time series data.
 *\param X    [OUTPUT] Calculated filter, prediction, and smooth distributions.
 *\param work [WORK]
 *\param prefix  Output file prefix.
 *\param postfix Output file postfix (extension).
 *\param id      ID inserted to the output file.
 *\param pv      Enables calculation and output of p-values. 
 *
 *\sa Kalman_infer_mu0()
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup Kalman */
void Kalman_apply_output_each(SSM * ssm, int k, SSMData * Y, SSMState * X, SSMWork * work,
							  const char * prefix, const char * postfix, int id, int pv){

  double * pvpos = work->PV;

  for(int i = 0; i < Y->replicates; i++){
	/* Calculate the state vectors */
	Kalman_filter(ssm, X, Y->p, k, Y->N,
				  Y->Y, Y->table[i], Y->missing, work);
	Kalman_smoother(X, k, Y->N, work->w, work->M, work->LU);

	/* Output the calculated "X" into a file. */
	SSMState_output_stats(ssm->H, X, Y->p, k, Y->N,
						  prefix, postfix, id, i + 1, work->e);

	/* Calculate the p values */
	if(pv) pvpos = calc_pvalues(ssm, X, Y, k, i, pvpos, work->Eyx, work->e);

	/* Output the input data Y. */
	SSMData_output2(Y, prefix, postfix, i);
  }/* for i < # of replicates */

  if(pv) output_pvalues(prefix, k, id, postfix, Y->p, Y->samples, work->PV);
}
/* Kalman_apply_output_each() ============================================== */

/*!\brief Applies the Kalman filter and smoother to generate filter,
 *    prediction and smooth distributions, and then outputs them into a single file.
 *
 * This is the same as Kalman_apply_output_each() except that this outputs
 * to a single file named
 * &quot;<i>prefix</i><tt>.D</tt><i>000</i><tt>.S</tt><i>000</i><tt>.K.</tt><i>postfix</i>&quot;.
 *
 * Each matrices are separated by an empty line, and a column representing the
 * time points is added to the matrices.
 * This file is useful when you use Gnuplot.
 *
 *\param pv      Enables calculation and output of p-values. 
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup Kalman */
void Kalman_apply_output(SSM * ssm, int k, SSMData * Y, SSMState * X, SSMWork * work,
						 const char * prefix, const char * postfix, int id, int pv){
  char buff[SSM_BUFF_SIZE];
  int len = snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.K.%s",
					 prefix, k, id, postfix);
  if(SSM_BUFF_SIZE <= len){
	fprintf(SIGN_ERR, "ERROR: Output prefix and/or postfix too long.\n");
	exit(1);
  }
  fprintf(SIGN_ERR, "Outputting the calculated state variables: %s\n",
		  buff);
  fflush(SIGN_ERR);
  FILE * fp = fopen(buff, "w");
  if(fp == NULL){
	fprintf(SIGN_ERR, "WARNING: Cannot open file: %s\n", buff);
	return;
  }

  double * pvpos = work->PV;

  for(int i = 0; i < Y->replicates; i++){
	fprintf(fp, "# REPLICATE %d\n", i + 1);
	/* Calculate the state vectors */
	Kalman_filter(ssm, X, Y->p, k, Y->N,
				  Y->Y, Y->table[i], Y->missing, work);
	Kalman_smoother(X, k, Y->N, work->w, work->M, work->LU);

	/* Output the calculated "X" into a file. */
	SSMState_output_stats_fp(ssm->H, X, Y->p, k, Y->N, fp, work->e, 1);

	if(pv) pvpos = calc_pvalues(ssm, X, Y, k, i, pvpos, work->Eyx, work->e);
	
	/* Output the input data Y. */
	fprintf(fp, "# Y\n");
	SSMData_output_fp(Y, fp, Y->table[i], 1);
	fprintf(fp, "\n\n");
  }/* for i < */
  fclose(fp);

  if(pv) output_pvalues(prefix, k, id, postfix, Y->p, Y->samples, work->PV);
}
/* Kalman_apply_output() =================================================== */

/*!\brief Infers the initial value <i>x</i><sub>0</sub> (or <i>mu</i><sub>0</sub>)
 *   by applying Kalman filter, prediction and smoother.
 *
 * This is used to estimate the initial value for the different data
 * set from one used for estimating the SSM model parameters,
 * in order to perform Kalman_apply() for the new data set.
 *
 *\param ssm   [INPUT/OUTPUT] SSM model parameters to be updated.  Only SSM::exp 
 *   is overwitten.
 *\param X     [WORK]  SSM state variables used to estimate the initial value.
 *\param k     Size of dimension
 *\param Y     [INPUT]  Input time series data.
 *\param work  [WORK] Only SSMWork::w is used.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup Kalman */
void Kalman_infer_mu0(SSM * ssm, SSMState * X, int k, SSMData * Y, SSMWork * work){
  /* Use SSMWORK::w as the working memory ((k)-vector). */
  memset(work->w, 0, sizeof(double) * k);
  double scale = 1.0 / (double) Y->replicates;
  for(int i = 0; i < Y->replicates; i++){
	Kalman_filter(ssm, X, Y->p, k, Y->N,
				  Y->Y, Y->table[i], Y->missing, work);
	Kalman_smoother(X, k, Y->N, work->w, work->M, work->LU);
	
	/* work->w += X->smooth_exp[0] / m   (m: # of replicates) */
	/*   Note: This is not efficient.  A much better way is to divide by m after
		 all the computation is finished. */
	SGN_Matrix_daxpy(&k, &scale, X->smooth_exp[0], &SGN_Matrix_INT_ONE, 
					 work->w, &SGN_Matrix_INT_ONE);
  }/* for i < # of replicates */
  
  /* Update x0 (mu0) as x0 := work->w */
  memcpy(ssm->exp, work->w, sizeof(double) * k);
}
/* Kalman_infer_mu0() ====================================================== */

/* End of file.  Copyright (C) 2010  SiGN Project Members */
