/*
  SSM.c : SiGN-SSM - Gene Network Estimation with State Sapce Model

  Copyright (C) 2010 - 2011, 2014
    Yoshinori Tamada, Rui Yamaguchi, Seiya Imoto, Osamu Hirose, Ryo Yoshida,
	Masao Nagasaki and Satoru Miyano
      Laboratory of DNA Information Analysis &
	  Laboratory of Sequence Analysis, Human Genome Center,
      The Institute of Medical Science, The Unversity of Tokyo

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  
  Visit <http://sign.hgc.jp/signssm/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
  Originally developed by Osamu Hirose, Ryo Yoshida, Seiya Imoto, Rui Yamaguchi,
                          Tomoyuki Higuchi, and Satoru Miyano.
*/

#include "signcom.h"
#include "SSM.h"
#include "SSM_EM.h"
#ifdef SSM_DEBUG
#include "debug.h"
#endif

#include <math/SGN_Matrix.h>
#include <math/SGN_Stat.h>
#include <util/SGN_Str.h>

#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>

/* Internal Prototype */
void SSMSuffStat_print(SSMSuffStat * T, int p, int k, char * m);

/*!\brief Generates an SSM instance.
 *
 * The generated instance can be used for
 * any <i>p'</i> &lt;= \a p and <i>k'</i> &lt;= \a k.
 *
 *\param p  Number of objects.
 *\param k  Size of dimension of the system model.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM */
SSM * SSM_new(int p, int k){
  SSM * ssm = malloc(sizeof(SSM));
  ssm->exp = malloc(sizeof(double) * k);
  ssm->var = malloc(sizeof(double) * k * k);
  ssm->H = malloc(sizeof(double) * p * k);
  ssm->R = malloc(sizeof(double) * p);
  ssm->F = malloc(sizeof(double) * k * k);
  ssm->Q = malloc(sizeof(double) * k);

  return ssm;
}
/* SSM_new() =============================================================== */

/*!\brief Returns the required size in bytes to allocate an SSM instance.
 *
 *\param p  Number of objects.
 *\param k  Size of dimension of the system model.
 *
 *\return memory size in bytes to allocate an SSM instance by \ref SSM_new(\a p, \a k).
 *\sa SSM_new()
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM */
size_t SSM_new_size(int p, int k){
  size_t mem = 0;
  /* SSM */
  mem += sizeof(SSM);
  /* SSM::exp */
  mem += sizeof(double) * k;
  /* SSM::var */
  mem += sizeof(double) * k * k;
  /* SSM::H */
  mem += sizeof(double) * p * k;
  /* SSM::R */
  mem += sizeof(double) * p;
  /* SSM::F */
  mem += sizeof(double) * k * k;
  /* SSM::Q */
  mem += sizeof(double) * k;

  return mem;
}
/* SSM_new_size() ========================================================== */

/*!\brief Copies the SSM instance to another.
 *
 *\param dst    Destination of copying the SSM instance. All the values are overwritten by \a src.
 *\param src    Source of the SSM instance to be copied.
 *\param p      Number of objects
 *\param k      Size of dimension
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM */
void SSM_copy(SSM * dst, const SSM * src, const int p, const int k){
  memcpy(dst->exp, src->exp, sizeof(double) * k);
  memcpy(dst->var, src->var, sizeof(double) * k * k);
  memcpy(dst->H, src->H, sizeof(double) * p * k);
  memcpy(dst->R, src->R, sizeof(double) * p);
  memcpy(dst->F, src->F, sizeof(double) * k * k);
  memcpy(dst->Q, src->Q, sizeof(double) * k);
}
/* SSM_copy() ============================================================== */

/*!\brief Generates the initial model parameters at random according to the
 *        initializing parameters.
 *
 * The current implementation does not use SSMInit::H_SD.
 * Instead of it, dimention \a k is used.
 *
 *\param ssm  [OUTPUT] SSM model parameters to be filled with the generate initial values.
 *\param init [INPUT] Parameters for generating the initial model parameters.
 *\param p   Number of objects.
 *\param k   Size of Dimension.
 *\param Y   (\a p, \a n)-input time series data matrix.
 *\param n   Number of observed samples in \a Y.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM */
void SSM_init(SSM * ssm, SSMInit * init, const int p, const int k,
			  const double * Y, const int n){
  
  /* ssm->exp (k x 1) := init->mu0 (1 x 1) */
  double v = init->mu0; 
  int kk = k;
  SGN_Matrix_dcopy(&kk, &v, &SGN_Matrix_INT_ZERO, ssm->exp, &SGN_Matrix_INT_ONE);

  /* ssm->var (k x k) := diag(init->SD0^2,...) */
  v = init->SD0 * init->SD0;
  memset(ssm->var, 0, sizeof(double) * k * k);
  int k_1 = k + 1;
  SGN_Matrix_dcopy(&kk, &v, &SGN_Matrix_INT_ZERO, ssm->var, &k_1);

  /* H:  ssm->H (p x k) := gaussian random value */
  SGN_RNG_normal(init->rng, ssm->H, p * k, init->H_mean, (double) k);

  /* R:  ssm->R (p) := variance in data */
  for(int i = 0; i < p; i++){
	double mean = 0.0, var = 0.0;
	for(int j = 0; j < n; j++){
	  double v = Y[i + j * p];
	  mean += v;
	  var += v * v;
	}
	mean /= (double) n;
	var /= (double) n;
	var -= mean * mean;
	
	ssm->R[i] = var;
  }/* for i < p */

  /* F: ssm->F (k x k) := uniform random value  */
  SGN_RNG_uniform(init->rng, ssm->F, k * k, init->F_min, init->F_max);
  
  /* Q:  ssm->Q (k x 1) := 1.0 */
  v = 1.0;
  SGN_Matrix_dcopy(&kk, &v, &SGN_Matrix_INT_ZERO, ssm->Q, &SGN_Matrix_INT_ONE);

}
/* SSM_init() ============================================================== */

/*!\brief [DEBUG] Prints and checks the contents of the SSM instance.
 *
 *\param ssm SSM model parameters to be checked.
 *\param p    Number of objects.
 *\param k    Size of dimension.
 *\param abs  Checks by the absolute values if non-zero.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM */
void SSM_check(SSM * ssm, int p, int k, int abs){
#ifdef SSM_DEBUG
  double v[2] = {p, k};

  debug_check("SSM_p_k", v, 1, 2, SSM_DEBUG, abs);
  debug_check("init.exp", ssm->exp, k, 1, SSM_DEBUG, abs);
  debug_check("init.var", ssm->var, k, k, SSM_DEBUG, abs);
  debug_check("init.H", ssm->H, p, k, SSM_DEBUG, abs);
  debug_check("init.R", ssm->R, p, 1, SSM_DEBUG, abs);
  debug_check("init.F", ssm->F, k, k, SSM_DEBUG, abs);
  debug_check("init.Q", ssm->Q, k, 1, SSM_DEBUG, abs);
#endif
}
/* SSM_check() ============================================================= */

/*!\brief Sets the constant initial model parameters.
 *
 * This is prepared for debugging and currently not used anywhere.
 *
 *\param ssm  SSM model parameters to be filled with constant initial values.
 *\param p  Number of objects.
 *\param k  Size of dimension.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM */
void SSM_init_const(SSM * ssm, int p, int k){

  /* ssm->exp (k,1) := 0.0 */
  double v = 0.0;
  SGN_Matrix_dcopy(&k, &v, &SGN_Matrix_INT_ZERO, ssm->exp, &SGN_Matrix_INT_ONE);
  
  /* ssm->var (k,k) := 1.0 */
  v = 1.0;
  memset(ssm->var, 0, sizeof(double) * k * k);
  int k_1 = k + 1;
  SGN_Matrix_dcopy(&k, &v, &SGN_Matrix_INT_ZERO, ssm->var, &k_1);

  /* R:  ssm->R (p) := variance in data */
  for(int i = 0; i < p; i++) ssm->R[i] = 1.0;

  /* H:  ssm->H (p,k) := 1.0 */
  for(int i = 0; i < p; i++){
	for(int j = 0; j < k; j++){
	  ssm->H[i + j * p] = 1.0;
	}
  }/* for i < p */
  
  /* F: ssm->F (k,k) := 0.0  */
  for(int i = 0; i < k; i++){
	for(int j = 0; j < k; j++){
	  ssm->F[i + j * k] = 1.0;
	}
  }/* for i < k */
  
  /* Q:  ssm->Q (k,1) := 1.0 */
  v = 1.0;
  SGN_Matrix_dcopy(&k, &v, &SGN_Matrix_INT_ZERO, ssm->Q, &SGN_Matrix_INT_ONE);
}
/* SSM_init() ============================================================== */

/*!\brief Applies the constraint on the model parameters.
 *
 * Use this before outputting the estimated SSM model parameters.
 *
 *\param ssm    [INPUT/OUTPUT]  SSM model parameters.
 *\param p      Number of objects.
 *\param k      Size of dimension.
 *\param L      [OUTPUT] (\a k)-vector. (Diagonal elements of (\a k, \a k)-matrix).
 *\param sqR    [OUTPUT] (\a p)-vector.
 *\param D      [WORK] (\a k)-vector.
 *\param U      [WORK] (\a k, \a k)-matrix.
 *\param W      [WORK] max(\a p, \a k, \a W_size, \a k * \a k)-vector.
 *\param W_size [WORK] size of \a W.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM */
void SSM_constrain(SSM * ssm, int p, int k, double * L, double * sqR, double * D, 
				   double * U, double * W, int W_size){

  /* sqR := sqrt(R) */
  for(int i = 0; i < p; i++) sqR[i] = sqrt(ssm->R[i]);
  
#ifdef SSM_DEBUG
  debug_check("Constrain:sqR", sqR, p, 1, SSM_DEBUG, 0);
#endif

  /* H := SqR^-1 H  */
  SGN_Matrix_dgeidm2(p, k, sqR, ssm->H, p);

  int info;

  /* L diag(D) U  :=  sqR^-1 H  (Singular value decomposition)  [real] */
  /* H := L [code] */
  /*    H : (p,k)-matrix */
  /*    D : (k)-vector. Singular values of H */
  /*    U : (k,k)-matrix */
  SGN_Matrix_dgesvd(SGN_Matrix_LO, SGN_Matrix_LA, &p, &k, ssm->H, &p,
					D, NULL, &p, U, &k, W, &W_size, &info);

#ifdef SSM_DEBUG
  //fprintf(SIGN_ERR, "Constraint:H = \n");
  //SGN_Matrix_printf(SIGN_ERR, ssm->H, p, k, "%.15f");
  //fprintf(SIGN_ERR, "Constraint:U = \n");
  debug_check("Constrain:H", ssm->H, p, k, SSM_DEBUG, 0);
  debug_check("Constrain:U", U, k, k, SSM_DEBUG, 0);
#endif

  /* H     := R^(1/2)  L     diag(D)   [real] */
  /*             ^ Here, there is an error in Hirose et al. (2008).  */
  /* H     := diag(sqR) H     diag(D)  [code] */
  /* (p,k)    (p,p)     (p,k)  (k,k) */
  /*   (1)   H := sqR   H */
  /*      (p,k)   (p,p) (p,k) */
  SGN_Matrix_dgedm2(p, k, sqR, ssm->H, p);
  /*   (2)   H := (sqR L) diag(D)  */
  /*       (p,k)   (p,k)   (k,k)   */
  SGN_Matrix_dgemd2(p, k, ssm->H, p, D);

#ifdef SSM_DEBUG
  debug_check("Constrain:H", ssm->H, p, k, SSM_DEBUG, 1);
#endif

  /* F    := U    F     U' */
  /* (k,k)   (k,k) (k,k) (k,k) */
  /*   (1) W :=  U   F  */
  SGN_Matrix_dgemm(SGN_Matrix_NO_TRANS, SGN_Matrix_NO_TRANS, &k, &k, &k,
				   &SGN_Matrix_DOUBLE_ONE, U, &k, ssm->F, &k,
				   &SGN_Matrix_DOUBLE_ZERO, W, &k);

#ifdef SSM_DEBUG
  //fprintf(SIGN_ERR, "Constraint:VF = \n");
  //SGN_Matrix_printf(SIGN_ERR, W, k, k, "%.15f");
  debug_check("Constrain:VF", W, k, k, SSM_DEBUG, 0);
#endif
  
  /*   (2) F :=  W       U' */
  /*       F :=  (U  F)  U' */
  SGN_Matrix_dgemm(SGN_Matrix_NO_TRANS, SGN_Matrix_TRANS, &k, &k, &k,
				   &SGN_Matrix_DOUBLE_ONE, W, &k, U, &k,
				   &SGN_Matrix_DOUBLE_ZERO, ssm->F, &k);

#ifdef SSM_DEBUG
  //fprintf(SIGN_ERR, "Constraint:F = \n");
  //SGN_Matrix_printf(SIGN_ERR, ssm->F, k, k, "%.15f");
  debug_check("Constrain:F", ssm->F, k, k, SSM_DEBUG, 1);
#endif

  /* exp   := U      exp */
  /* (k,1)    (k,k)  (k,1) */
  SGN_Matrix_dgemv(SGN_Matrix_NO_TRANS, &k, &k, &SGN_Matrix_DOUBLE_ONE, U, &k,
				   ssm->exp, &SGN_Matrix_INT_ONE, 
				   &SGN_Matrix_DOUBLE_ZERO, W, &SGN_Matrix_INT_ONE);
  memcpy(ssm->exp, W, sizeof(double) * k);

#ifdef SSM_DEBUG
  fprintf(SIGN_ERR, "Constraint:exp = \n");
  SGN_Matrix_printf(SIGN_ERR, ssm->exp, k, 1, "%.15f");
  debug_check("Constrain:exp", ssm->exp, k, 1, SSM_DEBUG, 1);
#endif

  /* L <- D^2 */
  for(int i = 0; i < k; i++) L[i] = D[i] * D[i];

  /* L <- H' R^-1 H [slow] */
  //for(int j = 0; j < k; j++){
	/* l_j = \sum_{i=1}^{p} h_{ij}^{2} / r_{ii} */
  //L[j] = 0.0;
  //for(int i = 0; i < p; i++){
  //  double h = ssm->H[i + j * p];
	//  L[j] += h * h / ssm->R[i];
  //	}
  //}
  
}
/* SSM_constrain() ========================================================= */

/*!\brief Calculates the module projection matrix.
 *
 * This calculates a projection matrix D := L<sup>-1</sup> H' R<sup>-1/2</sup> that
 * maps from the data space to the state space,
 * where L = H' R<sup>-1</sup> H.
 *
 *\param p  Number of objects.
 *\param k  Size of dimension.
 *\param H     [INPUT] (\a p, \a k)-observation matrix SSM::H.
 *\param L     [INPUT] (\a k)-vector calculated by SSM_constrain().
 *\param sqR   [INPUT] (\a p)-vector calculated by SSM_constrain().
 *\param D     [OUTPUT] (\a k,\a p)-matrix.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM */
void SSM_calc_module(int p, int k, double * H, double * L, double * sqR, double * D){

  /* D := H' */
  /* (k,p)  (p,k) */
  SGN_Matrix_trans(p, k, H, D);
  
  /* D := diag(L)^-1 D */ 
  /*    = diag(L)^-1 (H') */
  /* (k,p)   (k,k)         (k,p) */
  SGN_Matrix_dgeidm2(k, p, L, D, k);

  /* D := D diag(sqR)^-1 */
  /*      = (diag(L)^-1 H') diag(sqR)^-1 */
  /* (k,p)    (k,p)              (p,p)  */
  SGN_Matrix_dgemid2(k, p, D, k, sqR);
}
/* SSM_calc_module() ======================================================= */

/*!\brief Writes the SSM model parameters to the file stream.
 *
 * This outputs the contents of an SSM instance to the given file stream.
 *
 * Basically, this outputs in tab-separated style format where
 * empty lines are used to distinguish particular matrices or vectors.
 * Before each matrix/vector, a one-line header is put.  The header includes
 * the label, the number of rows and columns.
 *
 * The output file can be read by SSM_read_fp() to reconstruct the SSM
 * instance.
 *
 * The order (the format) of matrices/vectors is as follows:
 *
 * \code
 * "SSM_p_k"  1  2
 * p   k
 *
 * "SSM::exp"  k  1
 * SSM::exp (k x 1)
 *
 * "SSM::var"  k  k
 * SSM::var (k x k)
 *
 * "SSM::H"  p  k
 * SSM::H (p x k)
 *
 * "SSM::R"  1  p
 * SSM::R (1 x p)
 *
 * "SSM::F"  k  k
 * SSM::F (k x k)
 *
 * "SSM::Q"  k  1
 * SSM::Q (k, 1)
 *
 * \endcode
 *
 *\param ssm  SSM model parameters to output.
 *\param p    Number of objects.
 *\param k    Size of dimension.
 *\param fp   Output file stream.
 *
 *\sa SSM_read_fp()
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM */
void SSM_write_fp(SSM * ssm, int p, int k, FILE * fp){
  int d = 18;
  fprintf(fp, "SSM_p_k\t1\t2\n");
  fprintf(fp, "%d\t%d\n", p, k);
  fprintf(fp, "\n");
  fprintf(fp, "SSM::exp\t%d\t1\n", k);
  SGN_Matrix_fprintf(fp, d, ssm->exp, k, 1);
  fprintf(fp, "\n");
  fprintf(fp, "SSM::var\t%d\t%d\n", k, k);
  SGN_Matrix_fprintf(fp, d, ssm->var, k, k);
  fprintf(fp, "\n");
  fprintf(fp, "SSM::H\t%d\t%d\n", p, k);
  SGN_Matrix_fprintf(fp, d, ssm->H, p, k);
  fprintf(fp, "\n");
  fprintf(fp, "SSM::R\t%d\t1\n", p);
  SGN_Matrix_fprintf(fp, d, ssm->R, p, 1);
  fprintf(fp, "\n");
  fprintf(fp, "SSM::F\t%d\t%d\n", k, k);
  SGN_Matrix_fprintf(fp, d, ssm->F, k, k);
  fprintf(fp, "\n");
  fprintf(fp, "SSM::Q\t%d\t1\n", k);
  SGN_Matrix_fprintf(fp, d, ssm->Q, k, 1);
  fprintf(fp, "\n");
}
/* SSM_write_fp() ========================================================== */

/*!\brief Reads the SSM model parameters from the file stream.
 *
 * See SSM_write_fp() for the format of the SSM parameter file.
 *
 *\param fp   File stream.
 *\param p    [OUTPUT] Number of objects of the read model.
 *\param k    [OUTPUT] Size of dimension of the read model.
 *\return  Newly generated SSM instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM */
SSM * SSM_read_fp(FILE * fp, int * p, int * k){
  char * buff = NULL;
  size_t buff_size = 0;
  char * pos;
  /* p, k */
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the first label line. */
  SGN_Str_readline(fp, &buff, &buff_size);
  if(strlen(buff) == 0){
	fprintf(SIGN_ERR, "ERROR: Failed to read the SSM parameters.\n");
	exit(1);
  }
  pos = buff;
  *p = atoi(pos);
  pos = strchr(pos, '\t');
  if(pos == NULL){
	fprintf(SIGN_ERR, "ERROR: Failed to read the SSM parameters.\n");
	exit(1);
  }
  *k = atoi(pos + 1);
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the empty line. */

  SSM * ssm = SSM_new(*p, *k);

  /* ssm->exp: (k, 1) */
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the first label line. */
  pos = buff;
  for(int i = 0; i < *k; i++){
	SGN_Str_readline(fp, &buff, &buff_size);
	ssm->exp[i] = atof(buff);
  }
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the empty line. */

  /* ssm->var: (k,k) */
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the first label line. */
  for(int i = 0; i < *k; i++){
	SGN_Str_readline(fp, &buff, &buff_size);
	pos = buff;
	for(int j = 0; j < *k; j++){
	  ssm->var[i + j * (*k)] = atof(pos);
	  pos = strchr(pos, '\t') + 1;
	}
  }
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the empty line. */

  /* ssm->H: (p,k) */
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the first label line. */
  for(int i = 0; i < *p; i++){
	SGN_Str_readline(fp, &buff, &buff_size);
	pos = buff;
	for(int j = 0; j < *k; j++){
	  ssm->H[i + j * *p] = atof(pos);
	  pos = strchr(pos, '\t') + 1;
	}
  }
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the empty line. */
  
  /* ssm->R: (p,1) */
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the first label line. */
  pos = buff;
  for(int i = 0; i < *p; i++){
	SGN_Str_readline(fp, &buff, &buff_size);
	ssm->R[i] = atof(buff);
  }
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the empty line. */
  
  /* ssm->F: (k,k) */
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the first label line. */
  for(int i = 0; i < *k; i++){
	SGN_Str_readline(fp, &buff, &buff_size);
	pos = buff;
	for(int j = 0; j < *k; j++){
	  ssm->F[i + j * *k] = atof(pos);
	  pos = strchr(pos, '\t') + 1;
	}
  }
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the empty line. */
  
  /* ssm->Q: (k,1) */
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the first label line. */
  for(int i = 0; i < *k; i++){
	SGN_Str_readline(fp, &buff, &buff_size);
	ssm->Q[i] = atof(buff);
  }
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the empty line. */

  free(buff);
  return ssm;
}
/* SSM_read_fp() =========================================================== */

/*!\brief Allocates a new SSMState instance.
 *
 * This is used to store filter, prediction, and smooth distributions
 * calculated by Kalman filter and smoother.
 *
 *\param k  Size of dimension.
 *
 *\param N  Number of time points in a single replicate.  This does not
 *          the number of the observed samples in a replicate.  This represents
 *          the total number of points split by the same time inteval.  Thus,
 *          for example, if the observed time points are 1, 2, and 4, then
 *          \a N equals to 4, and not 3.
 *
 *\sa Kalman_filter(), Kalman_smoother(), Kalman
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMState */
SSMState * SSMState_new(int k, int N){
  
  SSMState * state = malloc(sizeof(SSMState));

  state->filter_exp = SGN_Matrix_new_ar(k, 1, N + 1); /* N+1 (k)-vectors */
  state->filter_var = SGN_Matrix_new_ar(k, k, N + 1); /* N+1 (k,k)-matrices */

  state->predict_exp = SGN_Matrix_new_ar(k, 1, N); /* N (k)-vectors */
  state->predict_var = SGN_Matrix_new_ar(k, k, N); /* N (k,k)-matrices */

  state->smooth_exp = SGN_Matrix_new_ar(k, 1, N + 1); /* N+1 (k)-vectors */
  state->smooth_var = SGN_Matrix_new_ar(k, k, N + 1); /* N+1 (k,k)-matrices */
  state->smooth_gain = SGN_Matrix_new_ar(k, k, N);    /* N (k,k)-matrices */

  state->FV = SGN_Matrix_new_ar(k, k, N); /* N (k,k)-matrices */
  state->Vinv = SGN_Matrix_new_ar(k, k, N); /* N (k,k)-matrices */
  
  state->lagone_cov = SGN_Matrix_new_ar(k, k, N); /* N (k,k)-matrices */

  return state;
}
/* SSMState_new() ========================================================== */

/*!\brief Outputs the array of vectors to the file stream.
 *
 *\param fp   Output file stream.
 *\param X     Array of vectors.  \a X[\a i][\a j] represents the
 *   \a j -th element of the \a i -th vector.
 *\param k     size of each vector.
 *\param N     Number of vectors.
 *\param t     Whether or not add the time column.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMState */
void SSMState_output_fp(FILE * fp, double ** X, int k, int N, int t){
  for(int j = 0; j < N; j++){
	if(t) fprintf(fp, "%d\t", j + 1);
	for(int i = 0; i < k; i++){
	  if(i != 0) fprintf(fp, "\t");
	  fprintf(fp, "%.17f", X[j][i]);
	}
	fprintf(fp, "\n");
  }/* j < N */
}
/* SSMState_output_fp() ==================================================== */

/*!\brief Outputs the array of vectors to the file.
 *
 *\param file  Output file name.
 *\param X     Array of vectors.  \a X[\a i][\a j] represents the
 *   \a j -th element of the \a i -th vector.
 *\param k     size of each vector.
 *\param N     Number of vectors.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMState */
void SSMState_output(const char * file, double ** X, int k, int N){
  FILE * fp = fopen(file, "w");
  if(fp == NULL){
	fprintf(SIGN_ERR, "ERROR: Cannot open file: %s\n", file);
	exit(1);
  }

  SSMState_output_fp(fp, X, k, N, 0);

  fclose(fp);
}
/* SSMState_output() ======================================================= */

/*!\brief Outputs <i>Y</i> = <i>H</i> <i>X</i> into a file.
 *
 * This outputs <i>Y</i> = <i>H</i> <i>X</i> where <i>Y</i> is a
 * (\a p, \a N) matrix, <i>H</i>, (\a p, \a k)-observation matrix,
 * and <i>X</i>, (\a k, \a N)-system state vectors for <i>N</i> time points.
 * The system state vectors are given as an array of (pointer to) (\a k)-order
 * vectors, instead of a (\a k, \a N)-matrix.  SSMState::filter_exp, SSMState::predict_exp,
 * and SSMState::smooth_exp calculated by Kalman_filter() and Kalman_smoother()
 * are assumed as argument \a x.
 *
 * Note that the output matrix is the transpose of the above <i>Y</i>, that is,
 * (\a N, \a p)-matrix <i>Y</i> '.
 *
 * This is called by SSMState_output_stats_fp().
 *
 *\param fp Output file stream.
 *\param H  [INPUT] (\a p, \a k)-observation matrix (SSM::H).
 *\param x  [INPUT] Array of (\a k)-order state variables.  The size of array is
 *                    \a N.
 *\param p  Number of objects.
 *\param k  Size of dimension.
 *\param N  Number of time points.
 *\param y  [WORK] (\a p)-vector.
 *\param t  Whehter or not add the time column.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM */
void SSM_output_Hx_fp(FILE * fp, double * H, double ** x, 
					  int p, int k, int N, double * y, int t){
  
  /* Output Y' where Y   :=  H     X     */
  /*        (N,p)    (p,N)   (p,k) (k,N) */
  for(int i = 0; i < N; i++){
	/* Calculate  y   :=  H     x     */
	/*            (p,1)   (p,k) (k,1) */
	SGN_Matrix_dgemv(SGN_Matrix_NO_TRANS, &p, &k, &SGN_Matrix_DOUBLE_ONE, H, &p,
					 x[i], &SGN_Matrix_INT_ONE, 
					 &SGN_Matrix_DOUBLE_ZERO, y, &SGN_Matrix_INT_ONE);
	if(t) fprintf(fp, "%d\t", i + 1);
	for(int j = 0; j < p; j++){
	  if(j != 0) fprintf(fp, "\t");
	  fprintf(fp, "%.17f", y[j]);
	  fflush(fp);
	}
	fprintf(fp, "\n");
  }

}
/* SSM_output_Hx_fp() ====================================================== */

/*!\brief Outputs <i>Y</i> = <i>H</i> <i>X</i> into a file.
 *
 * This outputs <i>Y</i> = <i>H</i> <i>X</i> where <i>Y</i> is a
 * (\a p, \a N) matrix, <i>H</i>, (\a p, \a k)-observation matrix,
 * and <i>X</i>, (\a k, \a N)-system state vectors for <i>N</i> time points.
 * The system state vectors are given as an array of (pointer to) (\a k)-order
 * vectors, instead of a (\a k, \a N)-matrix.  SSMState::filter_exp, SSMState::predict_exp,
 * and SSMState::smooth_exp calculated by Kalman_filter() and Kalman_smoother()
 * are assumed as argument \a x.
 *
 * Note that the output matrix is the transpose of the above <i>Y</i>, that is,
 * (\a N, \a p)-matrix <i>Y</i> '.
 *
 * This is called by SSM_output_stats().
 *
 *\param file  Output file name.
 *\param H  [INPUT] (\a p, \a k)-observation matrix (SSM::H).
 *\param x  [INPUT] Array of (\a k)-order state variables.  The size of array is
 *                    \a N.
 *\param p  Number of objects.
 *\param k  Size of dimension.
 *\param N  Number of time points.
 *\param w  [WORK] (\a p)-vector.
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM */
void SSM_output_Hx(const char * file,
				   double * H, double ** x, int p, int k, int N, double * w){
  FILE * fp = fopen(file, "w");
  if(fp == NULL){
	fprintf(SIGN_ERR, "ERROR: Cannot open file: %s\n", file);
	exit(1);
  }

  SSM_output_Hx_fp(fp, H, x, p, k, N, w, 0);

  fclose(fp);
}
/* SSM_output_Hx() ========================================================= */

/*!\brief Outputs the estimated state variables <i>x</i><sub><i>n</i></sub> and the estimated
 *   observation variables <i>y</i><sub><i>n</i></sub>.
 *
 * This makes six (6) files named
 * &quot;<i>prefix</i><tt>.</tt><i>X</i><tt>.D</tt><i>000</i><tt>.S</tt><i>000</i><tt>.<i>Y</i>.</tt><i>postfix</i>&quot;
 * where <i>X</i> is one of
 *  - <tt>Xp</tt> : prediction state variables.
 *  - <tt>Xf</tt> : filter state variables.
 *  - <tt>Xs</tt> : smooth state variables.
 *  - <tt>Yp</tt> : prediction observation variables.
 *  - <tt>Yf</tt> : filter observation variables.
 *  - <tt>Ys</tt> : sommth observation variables.
 *
 * and <i>Y</i> is a replicate number given as \a rep.
 *
 *\param H  [INPUT]  (\a p, \a k)-obseration matrix. (SSM::H)
 *\param X  [INPUT]  Estimated state variables.
 *\param p  Number of objects.
 *\param k  Size of dimension.
 *\param N  Number of time points.
 *\param prefix   Output file name prefix.
 *\param postfix  Output file name postfix (extension).
 *\param id       ID label inserted to the output file.
 *\param rep      Replicate ID inserted to the output file.
 *\param w  [WORK] (\a p)-working vector.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMState */
void SSMState_output_stats(double * H, SSMState * X, int p, int k, int N,
						   const char * prefix, const char * postfix, int id, int rep,
						   double * w){
  char buff[SSM_BUFF_SIZE];
  
  /* Output state vectors */
  snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.%s.%d.%s", prefix, k, id, "Xp", rep, postfix);
  fprintf(SIGN_ERR, "%s ...\n", buff); fflush(SIGN_ERR);
  SSMState_output(buff, X->predict_exp, k, N);

  snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.%s.%d.%s", prefix, k, id, "Xf", rep, postfix);
  fprintf(SIGN_ERR, "%s ...\n", buff); fflush(SIGN_ERR);
  SSMState_output(buff, X->filter_exp + 1, k, N); /* Does not print the initial point */

  snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.%s.%d.%s", prefix, k, id, "Xs", rep, postfix);
  fprintf(SIGN_ERR, "%s ...\n", buff); fflush(SIGN_ERR);
  SSMState_output(buff, X->smooth_exp + 1, k, N); 

  /* Output y     = H     x     */
  /*        (p,1)   (p,k) (k,1) */
  /*        Y     = H     X     (for N time points) */
  /*        (p,N) = (p,k) (k,N) */
  snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.%s.%d.%s", prefix, k, id, "Yp", rep, postfix);
  fprintf(SIGN_ERR, "%s ...\n", buff); fflush(SIGN_ERR);
  SSM_output_Hx(buff, H, X->predict_exp, p, k, N, w);

  snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.%s.%d.%s", prefix, k, id, "Yf", rep, postfix);
  fprintf(SIGN_ERR, "%s ...\n", buff); fflush(SIGN_ERR);
  SSM_output_Hx(buff, H, X->filter_exp + 1, p, k, N, w);

  snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.%s.%d.%s", prefix, k, id, "Ys", rep, postfix);
  fprintf(SIGN_ERR, "%s ...\n", buff); fflush(SIGN_ERR);
  SSM_output_Hx(buff, H, X->smooth_exp + 1, p, k, N, w);
}
/* SSMState_output_stats() ================================================= */

/*!\brief Outputs the estimated state variables <i>x</i><sub><i>n</i></sub> and the estimated
 *   observation variables <i>y</i><sub><i>n</i></sub>.
 *
 * This outputs six (6) matrices into a file stream.
 * The order of the matrices are:
 *  - <tt>Xp</tt> : prediction state variables.
 *  - <tt>Xf</tt> : filter state variables.
 *  - <tt>Xs</tt> : smooth state variables.
 *  - <tt>Yp</tt> : prediction observation variables.
 *  - <tt>Yf</tt> : filter observation variables.
 *  - <tt>Ys</tt> : sommth observation variables.
 *
 *\param H  [INPUT]  (\a p, \a k)-obseration matrix. (SSM::H)
 *\param X  [INPUT]  Estimated state variables.
 *\param p  Number of objects.
 *\param k  Size of dimension.
 *\param N  Number of time points.
 *\param fp Output file stream
 *\param w  [WORK] (\a p)-working vector.
 *\param t  Whether or not adds the time column.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMState */
void SSMState_output_stats_fp(double * H, SSMState * X, int p, int k, int N,
							  FILE * fp, double * w, int t){
  
  /* See SSMState_output_stats() for details. */
  fprintf(fp, "# Xp\n");
  SSMState_output_fp(fp, X->predict_exp, k, N, t);
  fprintf(fp, "\n\n");
  fprintf(fp, "# Xf\n");
  SSMState_output_fp(fp, X->filter_exp + 1, k, N, t);
  fprintf(fp, "\n\n");
  fprintf(fp, "# Xs\n");
  SSMState_output_fp(fp, X->smooth_exp + 1, k, N, t); 
  fprintf(fp, "\n\n");

  fprintf(fp, "# Yp\n");
  SSM_output_Hx_fp(fp, H, X->predict_exp, p, k, N, w, t);
  fprintf(fp, "\n\n");
  fprintf(fp, "# Yf\n");
  SSM_output_Hx_fp(fp, H, X->filter_exp + 1, p, k, N, w, t);
  fprintf(fp, "\n\n");
  fprintf(fp, "# Ys\n");
  SSM_output_Hx_fp(fp, H, X->smooth_exp + 1, p, k, N, w, t);
  fprintf(fp, "\n\n");
}
/* SSMState_output_stats_fp() ============================================== */

/*!\brief Allcates work space for the SSM calculation.
 *
 *\param p  Number of objects.
 *\param k  Size of dimension.
 *\param n  Number of observed samples.
 *\param constrain_F  If non-zero, contrain \a F when updating.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMWork */
SSMWork * SSMWork_new(int p, int k, int n, int constrain_F){
  SSMWork * work = malloc(sizeof(SSMWork));
  int k_k = k * k;
  int p_k = p * k;

  work->dgetri_work_size = SGN_Matrix_dgetri_work_size(k);
#ifdef SSM_DEBUG
  fprintf(SIGN_ERR, "dgetri_work_size = %d\n", work->dgetri_work_size);
#endif
  work->dgetri_work = malloc(sizeof(double) * work->dgetri_work_size);

  work->pivot = malloc(sizeof(int) * k);
  work->RinvH = malloc(sizeof(double) * p_k);
  work->tHRinve = malloc(sizeof(double) * k);
  work->tHRinvH = malloc(sizeof(double) * k_k);
  work->w_tHRinvH = malloc(sizeof(double) * k_k);
  work->e = malloc(sizeof(double) * p);
  work->Ke = malloc(sizeof(double) * k);
  work->I_KH = malloc(sizeof(double) * k_k);
  work->w = malloc(sizeof(double) * k);
  work->M = malloc(sizeof(double) * k_k);
  work->LU = malloc(sizeof(double) * k_k);
  work->Eyy = malloc(sizeof(double) * p);
  work->Eyx = malloc(sizeof(double) * p_k);
  work->Exx = malloc(sizeof(double) * k_k);

  double v, w;
  int lw = -1, info;

  SGN_Matrix_dgesvd(SGN_Matrix_LO, SGN_Matrix_LA, &p, &k, &v, &p, &v, NULL, 
				&SGN_Matrix_INT_ONE, &v, &k, &w, &lw, &info);
  work->dgesvd_work_size = (int) w;
#ifdef SSM_DEBUG
  fprintf(SIGN_ERR, "dgesvd_work_size = %d\n", work->dgesvd_work_size);
#endif

  lw = p;
  if(lw < k) lw = k;
  if(lw < work->dgesvd_work_size) lw = work->dgesvd_work_size;
  if(lw < k_k) lw = k_k;
  work->SVD = malloc(sizeof(double) * lw);

  if(constrain_F){
	work->C = malloc(sizeof(double) * k_k * k);
	work->C1 = malloc(sizeof(double) * k_k * k);
	work->C2 = malloc(sizeof(double) * k_k * k);
	work->C3 = malloc(sizeof(double) * k_k * k_k);
  } else {
	work->C = NULL;
	work->C1 = NULL;
	work->C2 = NULL;
	work->C3 = NULL;
  }

  work->PV = malloc(sizeof(double) * p * n);
  
  return work;
}
/* SSMWork_new() =========================================================== */

/*!\brief Returns the size of memory required to allocate an SSMWork instance.
 *
 *\param p  Number of objects.
 *\param k  Size of dimension.
 *\param constrain_F  If non-zero, includes the working memory for 
 *          constraining F.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMWork */
size_t SSMWork_new_size(int p, int k, int n, int constrain_F){
  size_t mem = 0;
  int k_k = k * k; 
  int p_k = p * k;
  
  /* SSMWork */
  mem += sizeof(SSMWork);
  /* SSMWork::dgetri_work */
  mem += sizeof(double) * SGN_Matrix_dgetri_work_size(k);
  /* SSMWork::pivot */
  mem += sizeof(int) * k;
  /* SSMWork::RinvH */
  mem += sizeof(double) * p_k;
  /* SSMWork::tHRinve */
  mem += sizeof(double) * k;
  /* SSMWork::tHRinvH */
  mem += sizeof(double) * k_k;
  /* SSMWork::w_tHRinvH */
  mem += sizeof(double) * k_k;
  /* SSMWork::e */
  mem += sizeof(double) * p;
  /* SSMWork::Ke */
  mem += sizeof(double) * k;
  /* SSMWork::I_KH */
  mem += sizeof(double) * k_k;
  /* SSMWork::w */
  mem += sizeof(double) * k;
  /* SSMWork::M */
  mem += sizeof(double) * k_k;
  /* SSMWork::LU */
  mem += sizeof(double) * k_k;
  /* SSMWork::Eyy */
  mem += sizeof(double) * p;
  /* SSMWork::Eyx */
  mem += sizeof(double) * p_k;
  /* SSMWork::Exx */
  mem += sizeof(double) * k_k;
  
  double v, w;
  int lw = -1, info;
  SGN_Matrix_dgesvd(SGN_Matrix_LO, SGN_Matrix_LA, &p, &k, &v, &p, &v, NULL, 
				&SGN_Matrix_INT_ONE, &v, &k, &w, &lw, &info);
  lw = p;
  if(lw < k) lw = k;
  if(lw < (int) w) lw = (int) w;
  if(lw < k_k) lw = k_k;
  /* SSMWork::SVD */
  mem += sizeof(double) * lw;

  if(constrain_F){
	/* SSMWork::C */
	mem += sizeof(double) * k_k * k;
	/* SSMWork::C1 */
	mem += sizeof(double) * k_k * k;
	/* SSMWork::C2 */
	mem += sizeof(double) * k_k * k;
	/* SSMWork::C3 */
	mem += sizeof(double) * k_k * k_k;
  }

  /* SSMWork::PV */
  mem += sizeof(double) * p * n;

  return mem;
}
/* SSMWork_new_size() ====================================================== */

/*!\brief Generates a new SSMSuffStat instance.
 *
 * The generated instance can be used for the estimation with
 * dimention <i>k</i>' &lt;= \a k.
 *
 *\param p  Number of objects.
 *\param k  Size of dimension.
 *\return  pointer to a newly allocated SSMSuffStat instance.
 *\sa SSMSuffStat_set_zero()
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMSuffStat */
SSMSuffStat * SSMSuffStat_new(int p, int k){
  SSMSuffStat * T = malloc(sizeof(SSMSuffStat));

  /* Tmu0: (k)-vector. */
  T->Tmu0 = malloc(sizeof(double) * k);
  
  /* Tyy: (p)-vector. */
  T->Tyy = malloc(sizeof(double) * p);

  /* Tyx: (p,k)-matrix */
  T->Tyx = malloc(sizeof(double) * p * k);

  /* Txx: (k,k)-matrix */
  T->Txx = malloc(sizeof(double) * k * k);
  
  /* Txx1: (k,k)-matrix */
  T->Txx1 = malloc(sizeof(double) * k * k);

  /* Tx1x1: (k,k)-matrix */
  T->Tx1x1 = malloc(sizeof(double) * k * k);

  return T;
}
/* SSMSuffStat_new() ======================================================= */

/*!\brief Sets zeros to all the fields in SSMSuffStat.
 *
 * Memo: Do not need to call this to set zeros for the SSMSuffStat instance,
 *       if it is passed to and calculated in SSMSuffStat_calc().
 *       This is because all the fileds are set in SSMSuffStat_calc().
 *
 *\param T  pointer to SSMSuffStat instance to be filled with zero values.
 *\param p  Number of objects.
 *\param k  Size of dimension.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMSuffStat */
void SSMSuffStat_set_zero(SSMSuffStat * T, int p, int k){
  /* Tmu0: (k)-vector. */
  memset(T->Tmu0, 0, sizeof(double) * k);
  
  /* Tyy: (p)-vector. */
  memset(T->Tyy, 0, sizeof(double) * p);

  /* Tyx: (p,k)-matrix */
  memset(T->Tyx, 0, sizeof(double) * p * k);

  /* Txx: (k,k)-matrix */
  memset(T->Txx, 0, sizeof(double) * k * k);
  
  /* Txx1: (k,k)-matrix */
  memset(T->Txx1, 0, sizeof(double) * k * k);

  /* Tx1x1: (k,k)-matrix */
  memset(T->Tx1x1, 0, sizeof(double) * k * k);
}
/* SSMSuffStat_set_zero() ================================================== */

/*!\brief Calculates the sufficient statistics.
 *
 * This is called by SSM_EM_estimate().
 *
 *\param T    [OUTPUT]  Sufficient statistics.
 *\param X    [INPUT]   State variables.
 *\param ssm  [INPUT]   SSM model parameters.
 *\param p       Number of objects.
 *\param k       Size of dimension.
 *\param N       Number of observed time points.
 *\param table   Column position table for \a Y.
 *\param Y       Input time series data matrix.
 *\param missing Missing row position arrays for \a Y.
 *\param Eyy  [WORK] (\a p)-vector
 *\param Eyx  [WORK] (\a p, \a k)-matrix
 *\param Exx  [WORK] (\a k, \a k)-matrix
 *\param w    [WORK] (\a k)-vector
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMSuffStat */
void SSMSuffStat_calc(SSMSuffStat * T, SSMState * X, SSM * ssm, int p, int k, int N, 
					  int * table, double * Y, int ** missing,
					  double * Eyy, double * Eyx, double * Exx, double * w){

  int k_k = k * k;
  int p_k = p * k;

  SSMSuffStat_set_zero(T, p, k);
  
  /* T->Tmu0 := X->smooth_exp[0] */
  memcpy(T->Tmu0, X->smooth_exp[0], sizeof(double) * k);

  for(int n = 0; n < N; n++){

	/* T->Tx1x1 := sx[n] sx[n]' + sV[n] */
	/*   (k,k) := (k,1) (1,k) + (k,k) */
	SGN_Matrix_dger(&k, &k, &SGN_Matrix_DOUBLE_ONE, X->smooth_exp[n], &SGN_Matrix_INT_ONE,
					X->smooth_exp[n], &SGN_Matrix_INT_ONE,
					T->Tx1x1, &k);
	SGN_Matrix_daxpy(&k_k, &SGN_Matrix_DOUBLE_ONE, X->smooth_var[n], &SGN_Matrix_INT_ONE,
					 T->Tx1x1, &SGN_Matrix_INT_ONE);
	
	/* T->Txx1 := sx[n+1] sx[n]' + cov[n] */
	SGN_Matrix_dger(&k, &k, &SGN_Matrix_DOUBLE_ONE, X->smooth_exp[n+1], &SGN_Matrix_INT_ONE,
					X->smooth_exp[n], &SGN_Matrix_INT_ONE,
					T->Txx1, &k);
	SGN_Matrix_daxpy(&k_k, &SGN_Matrix_DOUBLE_ONE, X->lagone_cov[n], &SGN_Matrix_INT_ONE,
					 T->Txx1, &SGN_Matrix_INT_ONE);

	if(table[n] != -1){
	  double * Yn = Y + table[n] * p;

	  /* Eyy_i := Yn_i * Yn_i */
	  /* (p,1) :=  (p,1) x (p,1) */
	  for(int i = 0; i < p; i++) Eyy[i] = Yn[i] * Yn[i];

	  /* Eyx  :=  Yn sx[n+1]' */
	  /* (p,k) := (p,1) (1,k) */
	  memset(Eyx, 0, sizeof(double) * p_k);
	  SGN_Matrix_dger(&p, &k, &SGN_Matrix_DOUBLE_ONE, Yn, &SGN_Matrix_INT_ONE,
					  X->smooth_exp[n+1], &SGN_Matrix_INT_ONE, Eyx, &p);

	  /* Exx := sV[n+1] + sx[n+1] sx[n+1]' */
	  /* (k,k) := (k,k) + (k,1) (1,k) */
	  memcpy(Exx, X->smooth_var[n+1], sizeof(double) * k_k);
	  SGN_Matrix_dger(&k, &k, &SGN_Matrix_DOUBLE_ONE, X->smooth_exp[n + 1], &SGN_Matrix_INT_ONE,
					  X->smooth_exp[n + 1], &SGN_Matrix_INT_ONE, Exx, &k);

	  for(int ii = 0; missing[table[n]][ii] != -1; ii++){
		int i = missing[table[n]][ii];

		SGN_Matrix_dgemv(SGN_Matrix_NO_TRANS, &k, &k,
						 &SGN_Matrix_DOUBLE_ONE, Exx, &k,
						 ssm->H + i, &p,
						 &SGN_Matrix_DOUBLE_ZERO, w, &SGN_Matrix_INT_ONE);         /* w := Exx H_i */
		double v = SGN_Matrix_ddot(&k, ssm->H + i, &p, w, &SGN_Matrix_INT_ONE); /* v := H_i' w */
		Eyy[i] = v + ssm->R[i];                                      /*      (1,k) (k,1) */
		
		SGN_Matrix_dgemv(SGN_Matrix_NO_TRANS, &k, &k,
						 &SGN_Matrix_DOUBLE_ONE, Exx, &k,
						 ssm->H + i, &p,
						 &SGN_Matrix_DOUBLE_ZERO, Eyx + i, &p); /* Eyx_i := Exx H_i */
	  }/* for ii */

	  /* Tyy (p) := Tyy + Eyy */
	  SGN_Matrix_daxpy(&p, &SGN_Matrix_DOUBLE_ONE, Eyy, &SGN_Matrix_INT_ONE, T->Tyy, &SGN_Matrix_INT_ONE);
	  /* Tyx (p,k) := Tyx + Eyx */
	  SGN_Matrix_daxpy(&p_k, &SGN_Matrix_DOUBLE_ONE, Eyx, &SGN_Matrix_INT_ONE, T->Tyx, &SGN_Matrix_INT_ONE);
	  /* Txx (k,k) := Txx + Exx */
	  SGN_Matrix_daxpy(&k_k, &SGN_Matrix_DOUBLE_ONE, Exx, &SGN_Matrix_INT_ONE, T->Txx, &SGN_Matrix_INT_ONE);
	  
	}/* if table[n] */
  }/* for n < N */

#ifdef SSM_DEBUG
  SSMSuffStat_print(T, p, k, "calc");
#endif
}
/* SSMSuffStat_calc() ===================================================== */

/*!\brief Adds values in sufficient statistics \a S to \a T.
 *
 *\param S  [INPUT] Unchanged on exit.
 *\param T  [INPUT/OUTPUT] Fields in \a S are added to those of \a T.
 *\param p  [INPUT] Number of objects (nodes, genes).
 *\param k  [INPUT] Size of dimension.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMSuffStat */
void SSMSuffStat_add(SSMSuffStat * S, SSMSuffStat * T, int p, int k){
  /* Tmu0: (K)-vector */
  SGN_Matrix_daxpy(&k, &SGN_Matrix_DOUBLE_ONE, S->Tmu0, &SGN_Matrix_INT_ONE, T->Tmu0, &SGN_Matrix_INT_ONE);
  
  /* Tyy: (p)-vector */
  SGN_Matrix_daxpy(&p, &SGN_Matrix_DOUBLE_ONE, S->Tyy, &SGN_Matrix_INT_ONE, T->Tyy, &SGN_Matrix_INT_ONE);

  /* Tyx: (p,k)-matrix */
  int p_k = p * k;
  SGN_Matrix_daxpy(&p_k, &SGN_Matrix_DOUBLE_ONE, S->Tyx, &SGN_Matrix_INT_ONE, T->Tyx, &SGN_Matrix_INT_ONE);

  /* Txx: (k,k)-matrix */
  int k_k = k * k;
  SGN_Matrix_daxpy(&k_k, &SGN_Matrix_DOUBLE_ONE, S->Txx, &SGN_Matrix_INT_ONE, T->Txx, &SGN_Matrix_INT_ONE);

  /* Txx1: (k,k)-matrix */
  SGN_Matrix_daxpy(&k_k, &SGN_Matrix_DOUBLE_ONE, S->Txx1, &SGN_Matrix_INT_ONE, T->Txx1, &SGN_Matrix_INT_ONE);

  /* Tx1x1: */
  SGN_Matrix_daxpy(&k_k, &SGN_Matrix_DOUBLE_ONE, S->Tx1x1, &SGN_Matrix_INT_ONE, T->Tx1x1, &SGN_Matrix_INT_ONE);

#ifdef SSM_DEBUG
  SSMSuffStat_print(T, p, k, "add");
#endif

}
/* SSMSuffStat_add() ======================================================= */

/*!\brief [DEBUG] Prints matrices in an SSMSuffStat instance into the
 *    debug log file.
 *
 * This is enabled only when compiled with SSM_DEBUG flag.
 * Otherwise, do nothing.
 *
 *\param T  SSMSuffStat instance. Unchanged on exit.
 *\param p  Number of objects.
 *\param k  Size of dimension.
 *\param m  Label inserted in the output label.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMSuffStat */
void SSMSuffStat_print(SSMSuffStat * T, int p, int k, char * m){
#ifdef SSM_DEBUG
  char buff[SSM_BUFF_SIZE];
  snprintf(buff, SSM_BUFF_SIZE, "Sufficient:%s:Tmu0", m);
  debug_check(buff, T->Tmu0, k, 1, SSM_DEBUG, 0);
  snprintf(buff, SSM_BUFF_SIZE, "Sufficient:%s:Tyy", m);
  debug_check(buff, T->Tyy, p, 1, SSM_DEBUG, 0);
  snprintf(buff, SSM_BUFF_SIZE, "Sufficient:%s:Tyx", m);
  debug_check(buff, T->Tyx, p, k, SSM_DEBUG, 0);
  snprintf(buff, SSM_BUFF_SIZE, "Sufficient:%s:Txx", m);
  debug_check(buff, T->Txx, k, k, SSM_DEBUG, 0);
  snprintf(buff, SSM_BUFF_SIZE, "Sufficient:%s:Txx1", m);
  debug_check(buff, T->Txx1, k, k, SSM_DEBUG, 0);
  snprintf(buff, SSM_BUFF_SIZE, "Sufficient:%s:Tx1x1", m);
  debug_check(buff, T->Tx1x1, k, k, SSM_DEBUG, 0);
#endif
}
/* SSMSuffStat_print() ===================================================== */

/* End of file.  Copyright (C) 2010-2011, 2014  SiGN Project Members */
