/*
  SSM.h : SiGN-SSM - Gene Network Estimation with State Sapce Model

  Copyright (C) 2010
    Yoshinori Tamada, Rui Yamaguchi, Seiya Imoto, Osamu Hirose, Ryo Yoshida,
	Masao Nagasaki and Satoru Miyano
      Laboratory of DNA Information Analysis &
	  Laboratory of Sequence Analysis, Human Genome Center,
      The Institute of Medical Science, The Unversity of Tokyo

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  
  Visit <http://sign.hgc.jp/signssm/> for more information.

  Written by Yoshinori Tamada <tamada at ims.u-tokyo.ac.jp>
  Originally developed by Osamu Hirose, Ryo Yoshida, Seiya Imoto, Rui Yamaguchi,
                          Tomoyuki Higuchi, and Satoru Miyano.
*/

#ifndef __SSM_H
#define __SSM_H

#include <stdlib.h>
#include <stdio.h>
#include "SSMInit.h"

/*!\brief Common buffer size for strings.
 *\memberof SSM */
#define SSM_BUFF_SIZE 1024

/*!\brief Structure for SSM model parameters.
 *
 * <i>p</i> and <i>k</i> in the documentation represent
 * the number of objects (genes) and the size of dimension, respectively.
 *
 *\sa SSM_EM, SSMState, SSMSuffStat, SSMInit, SSMData, SSMResult, SSMWork, \ref Kalman
 */
typedef struct {
  /*!\brief Expectation: (\a k)-vector */
  double * exp;     /* alias: inie, ie */

  /*!\brief Variance: (\a k x \a k)-matrix */
  double * var; 

  /*!\brief (\a p x \a k)-observation matrix */
  double * H; 

  /*!\brief Observation noise: (\a p)-diagonal matrix  */
  double * R; 

  /*!\brief (\a k x \a k)-system matrix */
  double * F;

  /*!\brief System noise: (\a k)-diagonal matrix */ 
  double * Q;
} SSM;
/* typedef SSM ============================================================= */

/*!\brief Structur for SSM state variables.
 *
 *\sa SSM, SSMSuffStat, \ref Kalman, Kalman_predict()
 */
typedef struct { /* k: dimension */
  /*!\brief \a N of (\a k)-vectors */
  double ** predict_exp;

  /*!\brief \a N of (\a k,\a k)-matrices */
  double ** predict_var;

  /*!\brief N+1 * (k)-vectors */
  double ** filter_exp;  

  /*!\brief N+1 * (k,k)-matrices */
  double ** filter_var;

  /*!\brief N+1 * (k)-vectors */
  double ** smooth_exp;

  /*!\brief N+1 of (k,k)-matrices */  
  double ** smooth_var;

  /*!\brief \a N of (\a k,\a k)-matrices */
  double ** smooth_gain;

  /*!\brief \a N of (\a k, \a k) matrix to store <i>F V</i>. */
  double ** FV;

  /*!\brief \a N of (\a k, \a k) matrix to store the inverse of \ref predict_var. */
  double ** Vinv;

  /*!\brief \a N of (\a k, \a k)-matrices. */
  double ** lagone_cov;
} SSMState;
/* typedef SSMState ======================================================== */

/*!\brief Structure for sufficient statistics.
 *
 * \a k and \a p in this documentation represent the size of dimension and
 * the number of objects (genes), respectively.
 *
 *\sa SSM, SSMState, SSMData, SSMWork, Kalman
 */
typedef struct {
  /*!\brief  (\a k)-vector. */
  double * Tmu0;

  /*!\brief (\a p)-vector. */
  double * Tyy;

  /*!\brief (\a p, \a k)-matrix. */
  double * Tyx;

  /*!\brief (\a k, \a k)-matrix. */
  double * Txx;

  /*!\brief (\a k, \a k)-matrix. */
  double * Txx1;

  /*!\brief (\a k, \a k)-matrix. */
  double * Tx1x1;
} SSMSuffStat;
/* typedef SSMSuffStat ===================================================== */

/*!\brief Structure for working memory spaces.
 *
 *\sa Kalman_filter_observed()
 *
 *\sa SSM, SSMState, SSMSuffStat, SSM_EM, \ref Kalman
 */
typedef struct {

  /*!\brief Work space for dgetri routine. */
  double * dgetri_work;

  /*!\brief Size of SSMWork::dgetri_work. */
  int dgetri_work_size;

  /*!\brief Pivot table for LU factorization. (\a k)-integer vector.*/
  int * pivot;

  /*!\brief (\a p, \a k) matrix. */
  double * RinvH;

  /*!\brief (\a k) vector */
  double * tHRinve;

  /*!\brief (\a k, \a k) matrix */
  double * tHRinvH;

  /*!\brief (\a k, \a k) matrix */
  double * w_tHRinvH;

  /*!\brief (\a p) vector */
  double * e;

  /*!\brief (\a k) vector */
  double * Ke;

  /*!\brief (\a k, \a k) matrix */
  double * I_KH;

  /*!\brief (\a k)-vector. */
  double * w;

  /*!\brief (\a k, \a k) matrix. */
  double * M;

  /*!\brief (\a k, \a k) matrix. */
  double * LU;

  /*!\brief (\a p)-vector. */
  double * Eyy;
  
  /*!\brief (\a p, \a k)-matrix. */
  double * Eyx;

  /*!\brief (\a k, \a k)-matrix. */
  double * Exx;

  /*!\brief max{\a p, \a k, \a dgesvd_size, \a k * \a k} vector. */
  double * SVD;
  
  /*!\brief working are size for dgesvd routine. */
  int dgesvd_work_size;

  /*!\brief (\a k, \a k <sup>2</sup>)-constraint matrix */
  double * C;

  /*!\brief (\a k, \a k <sup>2</sup>)-matrix. */
  double * C1;
  
  /*!\brief (\a k, \a k <sup>2</sup>)-matrix. */
  double * C2;

  /*!\brief (\a k <sup>2</sup>, \a k <sup>2</sup>)-matrix. */
  double * C3;

  /*!\brief (\a p, \a n)-matrix where \a n represents the number of observed samples. */
  double * PV;
  
} SSMWork;

SSM *  SSM_new(int p, int k);
size_t SSM_new_size(int p, int k);
void   SSM_copy(SSM * dst, const SSM * src, const int p, const int k);
void   SSM_init(SSM * ssm, SSMInit * init, const int p, const int k,
				const double * Y, const int n);
void   SSM_init_const(SSM * ssm, int p, int k);
void   SSM_write_fp(SSM * ssm, int p, int k, FILE * fp);
SSM *  SSM_read_fp(FILE * fp, int * p, int * k);
void   SSM_check(SSM * ssm, int p, int k, int abs);
void   SSM_calc_module(int p, int k, double * H, double * eigen, double * sqR, double * D);
void   SSM_constrain(SSM * ssm, int p, int k, double * eigen, double * sqR, double * S, 
					 double * tV, double * W, int W_size);

SSMState * SSMState_new(int k, int N);
void SSMState_output_stats(double * H, SSMState * X, int p, int k, int N,
						   const char * prefix, const char * postfix, int id, int rep,
						   double * w);
void SSMState_output_stats_fp(double * H, SSMState * X, int p, int k, int N,
							  FILE * fp, double * w, int t);


SSMWork * SSMWork_new(int p, int k, int n, int constrain_F);
size_t    SSMWork_new_size(int p, int k, int n, int constrain_F);


SSMSuffStat * SSMSuffStat_new(int p, int k);
void SSMSuffStat_set_zero(SSMSuffStat * T, int p, int k);
void SSMSuffStat_calc(SSMSuffStat * T, SSMState * X, SSM * ssm, int p, int k, int N, 
					  int * table, double * Y, int ** missing,
					  double * Eyy, double * Eyx, double * Exx, double * w);

void SSMSuffStat_add(SSMSuffStat * S, SSMSuffStat * T, int p, int k);

#endif /* __SSM_H */

/* End of file.  Copyright (C) 2010  HGC, IMS, The University of Tokyo       */
