/*
  SSMData.c : SiGN-SSM - Gene Network Estimation with State Sapce Model

  Copyright (C) 2010
    Yoshinori Tamada, Rui Yamaguchi, Seiya Imoto, Osamu Hirose, Ryo Yoshida,
	Masao Nagasaki and Satoru Miyano
      Laboratory of DNA Information Analysis &
	  Laboratory of Sequence Analysis, Human Genome Center,
      The Institute of Medical Science, The Unversity of Tokyo

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  
  Visit <http://sign.hgc.jp/signssm/> for more information.

  Written by Yoshinori Tamada <tamada at ims.u-tokyo.ac.jp>
  Originally developed by Osamu Hirose, Ryo Yoshida, Seiya Imoto, Rui Yamaguchi,
                          Tomoyuki Higuchi, and Satoru Miyano.
*/

#include "signcom.h"
#include "SSMData.h"

#include "util/SGN_IntArray.h"
#include "util/SGN_DoubleArray.h"
#include "util/SGN_StrArray.h"
#include "util/SGN_Str.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <strings.h>

#ifdef USE_SIGN
/*!\brief Converts a Dataset into SSMData.
 *
 * The current implementation assumes that the primary ID represents
 * a time point and secondary ID, a replicate ID.
 *
 * NaN (Not a Number) values are regarded as missing values and
 * handled properly by the program.
 *
 * Primary IDs and Secondary IDs need to be set in the given Dataset.
 * If not so, then an error occurs and the program quits with error code 1.
 *
 *\param dataset   SiGN Dataset instance.
 *\return Pointer to the newly generated SSMData instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMData */
SSMData * SSMData_new_Data(const SGN_Data * data){

#ifdef SSM_DEBUG
  fprintf(SIGN_ERR, "SSMData: Converting data structure.\n"); fflush(SIGN_ERR);
#endif

  /* Primary ID = Time Point */
  /* Secondary ID = Replicate ID */
  
  /* Check if the primary and secondary IDs are set */
  if(data->primaryID == NULL || data->secondaryID == NULL){
	fprintf(SIGN_ERR, "ERROR: Primary and/or secondary IDs are missing in the input data.\n");
	exit(1);
  }

  SGN_IntArray * replicateIDs = SGN_Data_unique_secondaryID(data);

  SSMData * Y = malloc(sizeof(SSMData));
  /* Set basic files of SSMData */
  Y->p = data->p;  /* Number of objects (genes) */
  Y->samples = data->n;  /* Number of total samples */
  Y->replicates = (int) SGN_IntArray_size(replicateIDs); /* Number of replicates */
  Y->N = SGN_Data_max_primaryID(data); /* Number of time points */

  /* Create Replicate ID -> index table */
  int max_replicateID = SGN_Data_max_secondaryID(data);
  int * repl_table = malloc(sizeof(int) * (max_replicateID + 1));
  /* repl_table[i] represents position of exiting replicateIDs. */
  for(int i = 0; i < max_replicateID; i++) repl_table[i] = -1;
  for(int i = 0; i < Y->replicates; i++){
	repl_table[SGN_IntArray_get(replicateIDs, i)] = i;
  }

  /* Allocation of memory for tables and matrices */
  Y->Y = malloc(sizeof(double) * Y->p * Y->samples);
  Y->table = malloc(sizeof(int *) * Y->replicates);
  int * ptr = malloc(sizeof(int) * Y->replicates * Y->N);
  for(int i = 0; i < Y->replicates; i++){
	Y->table[i] = ptr;
	ptr += Y->N;
	for(int j = 0; j < Y->N; j++) Y->table[i][j] = -1;
  }
  Y->missing = malloc(sizeof(int *) * Y->samples);

  /* Set values into Y->Y, Y->missing and Y->table */
  SGN_IntArray * missing = SGN_IntArray_new();
  for(int k = 0; k < Y->samples; k++){
	int n = SGN_IntArray_get(data->primaryID, k) - 1;
	int l = repl_table[SGN_IntArray_get(data->secondaryID, k)];

	Y->table[l][n] = k;
	SGN_IntArray_clear(missing);
	for(int i = 0; i < Y->p; i++){
	  double v = data->X[k + i * data->n];
	  if(isnan(v)){
		SGN_IntArray_add(missing, i);
		v = 0.0;
	  }
	  Y->Y[i + k * Y->p] = v;
	}
	size_t s = SGN_IntArray_size(missing);
	Y->missing[k] = malloc(sizeof(int) * (s + 1));
	memcpy(Y->missing[k], SGN_IntArray_ptr(missing), sizeof(int) * s);
	Y->missing[k][s] = -1; /* End marker */
  }

  free(repl_table);
  SGN_IntArray_delete(missing);
  SGN_IntArray_delete(replicateIDs);

#ifdef SSM_DEBUG
  fprintf(SIGN_ERR, "  objects = %d\n", Y->p);
  fprintf(SIGN_ERR, "  samples = %d\n", Y->samples);
  fprintf(SIGN_ERR, "  time points = %d\n", Y->N);
  fprintf(SIGN_ERR, "  replicates = %d\n", Y->replicates);

  for(int i = 0; i < Y->replicates; i++){
	fprintf(SIGN_ERR, "Rep #%d;", i + 1);
	for(int j = 0; j < Y->N; j++){
	  fprintf(SIGN_ERR, "  %d:%d", j, Y->table[i][j]);
	}
	fprintf(SIGN_ERR, "\n");
  }
  fflush(SIGN_ERR);

  fprintf(SIGN_ERR, "missing:\n");
  for(int i = 0; i < Y->samples; i++){
	fprintf(SIGN_ERR, "[%d] ", i);
	int j = 0;
	do {
	  fprintf(SIGN_ERR, "%d ", Y->missing[i][j]);
	} while(Y->missing[i][j++] != -1);
	fprintf(SIGN_ERR, "\n");
  }
  fflush(SIGN_ERR);
#endif

  return Y;
}
/* SSMData_new_Data() =================================================== */
#endif /* USE_SIGN */

/*!\brief Output a single replicate data into a file stream.
 *
 *\param Y [INPUT]
 *\param fp   Output file stream.
 *\param table column index table for the replicate.
 *\param t  Whether adds the time column or not.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMData */
void SSMData_output_fp(SSMData * Y, FILE * fp, int * table, int t){
  for(int i = 0; i < Y->N; i++){

	if(t) fprintf(fp, "%d\t", i + 1);
	
	int index = table[i];
	if(index == -1){
	  /* Missing time point */
	  for(int j = 0; j < Y->p; j++){
		if(j != 0) fprintf(fp, "\t");
		fprintf(fp, "NAN");
	  }
	  fprintf(fp, "\n");
	} else {
	  int midx = 0;
	  for(int j = 0; j < Y->p; j++){
		if(j != 0) fprintf(fp, "\t");
		if(Y->missing[index][midx] == j){
		  fprintf(fp, "NAN");
		  midx++;
		} else {
		  fprintf(fp, "%.15f", Y->Y[j + index * Y->p]);
		}
	  }
	  fprintf(fp, "\n");
	}
  }/* for i < N */
}
/* SSMData_output_fp() ===================================================== */

/*!\brief Output a single replicate data into a file.
 *
 *\param Y [INPUT]
 *\param file  Output file name
 *\param rep  Replicate ID.  The ID of the first replicate is 0 (0-origin).
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMData */
void SSMData_output(SSMData * Y, const char * file, int rep){

  FILE * fp = fopen(file, "w");
  if(fp == NULL){
	fprintf(SIGN_ERR, "Cannot open file: %s\n", file);
	exit(1);
  }

  SSMData_output_fp(Y, fp, Y->table[rep], 0);

  fclose(fp);
}
/* SSMData_output() ======================================================== */

/*!\brief Output a single replicate data into a file with a prefix and postfix.
 *
 *\param Y        Pointer to the SSMData instance to be printed.
 *\param prefix   Prefix of the output file name.
 *\param postfix  Postfix of the output file name.
 *\param rep  Replicate ID.  The ID starts from 0 (0-origin).  The ID in the output file
 *              name becomes (\a rep + 1).
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMData */
void SSMData_output2(SSMData * Y, const char * prefix, const char * postfix, int rep){
  int size = 256;
  char buff[size];
  
  /* Output state vectors */
  snprintf(buff, size, "%s.%s.%d.%s", prefix, "Y", rep + 1, postfix);
  fprintf(SIGN_ERR, "%s ...\n", buff); fflush(SIGN_ERR);
  SSMData_output(Y, buff, rep);
}
/* SSMData_output2() ======================================================= */

/*!\brief Performs mean shift for each object.
 *
 *\param Y  [INPUT/OUTPUT] SSMData instance.
 *\param replicate  If 1, mean shift for each replicate, otherwise
 *                   mean shift for all the replicates at once.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMData */
void SSMData_node_mean_shift(SSMData * Y, int replicate){
  double v[Y->p];
  int c[Y->p];
  
  if(replicate == 1){
	/* Mean shift for each replicate */
	for(int n = 0; n < Y->replicates; n++){
	  memset(v, 0, sizeof(double) * Y->p);
	  memset(c, 0, sizeof(int) * Y->p);
	  
	  for(int i = 0; i < Y->N; i++){
		int pos = Y->table[n][i];
		if(pos == -1) continue;
		int m = 0;
		for(int j = 0; j < Y->p; j++){
		  if(j == Y->missing[pos][m]){
			m++;
			continue;
		  }
		  v[j] += Y->Y[j + pos * Y->p];
		  c[j]++;
		}
	  }/* for i < Y->N */
	  
	  /* Calculates mean */
	  for(int j = 0; j < Y->p; j++){
		if(c[j] == 0){
		  v[j] = 0;
		} else {
		  v[j] /= (double) c[j];
		}
	  }
	  
	  /* Mean shifting */
	  for(int i = 0; i < Y->N; i++){
		int pos = Y->table[n][i];
		if(pos == -1) continue;
		int m = 0;
		for(int j = 0; j < Y->p; j++){
		  if(j == Y->missing[pos][m]){
			m++;
			continue;
		  }
		  Y->Y[j + pos * Y->p] -= v[j];
		}
	  }/* for i < N */
	}
  } else {
	memset(v, 0, sizeof(double) * Y->p);
	memset(c, 0, sizeof(int) * Y->p);

	for(int n = 0; n < Y->samples; n++){
		int m = 0;
		for(int j = 0; j < Y->p; j++){
		  if(j == Y->missing[n][m]){
			m++;
			continue;
		  }
		  v[j] += Y->Y[j + n * Y->p];
		  c[j]++;
		}
	}/* for i < Y->N */
	
	/* Calculates mean */
	for(int j = 0; j < Y->p; j++){
	  if(c[j] == 0){
		v[j] = 0;
	  } else {
		v[j] /= (double) c[j];
	  }
	}

	for(int n = 0; n < Y->samples; n++){
	  int m = 0;
	  for(int j = 0; j < Y->p; j++){
		if(j == Y->missing[n][m]){
		  m++;
		  continue;
		}
		Y->Y[j + n * Y->p] -= v[j];
	  }
	}/* for i < samples */
  }/* if ... else */
}
/* SSMData_node_mean_shift() =============================================== */

/*!\brief Reads the time series data from a file stream.
 *
 * The file format is a subset version of EDF (Expression Data Format)
 * developed in the authors' laboratory.  See other documents for details.
 *
 *\param fp  Pointer to the file stream to read.
 *\return  Newly generated SSMData instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMData */
SSMData * SSMData_read_fp(FILE * fp){
  char * primary_key = "time";
  char * secondary_key = "replicate";

  char * buff = NULL;
  size_t buff_size = 0;
  int ln = 0;
  int section = 0; /* 0: meta data, 1: attribute, 2: data */
  SGN_IntArray * uniq_repIDs = SGN_IntArray_new();
  SGN_IntArray * repIDs = SGN_IntArray_new();
  SGN_IntArray * times = SGN_IntArray_new();
  SGN_DoubleArray * values = SGN_DoubleArray_new();
  SGN_StrArray * names = SGN_StrArray_new();
  int p = 0;
  while(SGN_Str_readline(fp, &buff, &buff_size) == 0){
	ln++;
	if(buff[0] == '#') continue; /* Skip a comment line */
	char * line = SGN_Str_trim(buff);

	if(section == 0 && line[0] == '@'){
	  section = 1;
	}
	if(section <= 1 && (line[0] != '@' && line[0] != '$')){
	  section = 2;

	  /* Case if the replicate ID is not specified */
	  if(SGN_IntArray_size(repIDs) == 0){
		for(int i = 0; i < SGN_IntArray_size(times); i++) SGN_IntArray_add(repIDs, 1);
		SGN_IntArray_add(uniq_repIDs, 1);
	  }
	}

	/* --------------------------- */
	/* Section Specific Processing */
	/* --------------------------- */
	if(section == 0){
	  /* Meta data section */
	  char * delim = strchr(line, '\t');
	  if(delim == NULL){
		fprintf(SIGN_ERR, "WARNING: Broken meta data line found. Skip...\n");
		continue;
	  }
	  *delim = '\0';
	  char * key = SGN_Str_trim(line);
	  char * value = SGN_Str_trim(delim + 1);
	  if(strcasecmp(key, "$primarykeygroupid") == 0){
		primary_key = SGN_Str_copy(value);
		fprintf(SIGN_ERR, "  $PrimaryKeyGroupID: \"%s\"\n", primary_key);
	  } else if(strcasecmp(key, "$secondarykeygroupid") == 0){
		secondary_key = SGN_Str_copy(value);
		fprintf(SIGN_ERR, "  $SecondaryPrimaryKeyGroupID: \"%s\"\n", value);
	  } else {
		fprintf(SIGN_ERR, "WARNING: Unsupported meta data key: %s\n", key);
	  }
	} else if(section == 1){
	  /* Attribute section  */
	  char * delim = strchr(line, '\t');
	  *delim = '\0';
	  char * key = SGN_Str_trim(line);
	  if(key[0] == '@' && strcasecmp(key + 1, primary_key) == 0){
		/* Primary Key (Time points) */
		char * prev = delim + 1;
		while((delim = strchr(prev, '\t')) != NULL){
		  SGN_IntArray_add(times, atoi(prev));
		  prev = delim + 1;
		}
		SGN_IntArray_add(times, atoi(prev));

		if(SGN_IntArray_size(repIDs) != 0 && 
		   (SGN_IntArray_size(repIDs) != SGN_IntArray_size(times))){
		  fprintf(SIGN_ERR, "ERROR: Inconsistent number of columns: line %d\n", ln);
		  exit(1);
		}
	  } else if(key[0] == '@' && strcasecmp(key + 1, secondary_key) == 0){
		/* Secondary Key (Replicate IDs) */
		char * prev = delim + 1;
		while((delim = strchr(prev, '\t')) != NULL){
		  int v = atoi(prev);
		  SGN_IntArray_add_nr(uniq_repIDs, v);
		  SGN_IntArray_add(repIDs, v);
		  prev = delim + 1;
		}
		int v = atoi(prev);
		SGN_IntArray_add_nr(uniq_repIDs, v);
		SGN_IntArray_add(repIDs, v);
		if(SGN_IntArray_size(times) != 0 &&
		   (SGN_IntArray_size(times) != SGN_IntArray_size(repIDs))){
			 fprintf(SIGN_ERR, "ERROR: Inconsistent number of columns: line %d\n", ln);
			 exit(1);
		}
	  }
	} else if(section == 2){
	  /* Data section */
	  p++; /* Number of objects */
	  char * pos = strchr(line, '\t');
	  int c = 0;
	  char * endptr;

	  char * name = malloc(sizeof(char) * (pos - line + 1));
	  memcpy(name, line, pos - line);
	  name[pos - line] = '\0';
	  SGN_StrArray_add(names, name);

	  while(pos != NULL){
		double v = strtod(pos + 1, &endptr);
		if(pos + 1 == endptr){
		  fprintf(SIGN_ERR, "ERROR: Unrecognizable character found at line %d\n", ln);
		  exit(1);
		}
		SGN_DoubleArray_add(values, v);
		c++;
		pos = strchr(pos + 1, '\t');
	  }
	  
	  if(c != SGN_IntArray_size(times)){
		fprintf(SIGN_ERR, "ERROR: Inconsistent number of columns: line %d\n", ln);
		fprintf(SIGN_ERR, "     :   %d (previous) vs %d (now).\n", 
				(int) SGN_IntArray_size(times), c);
		exit(1);
	  }
	}
  }/* while() */

  SSMData * Y = malloc(sizeof(SSMData));
  Y->p = p;
  Y->replicates = SGN_IntArray_size(uniq_repIDs);
  Y->samples = SGN_IntArray_size(times);
  Y->N = SGN_IntArray_max(times);

  /* Checks the size of values */
  if(SGN_DoubleArray_size(values) != Y->p * Y->samples){
	fprintf(SIGN_ERR, "ERROR: Input file broken.\n");
	exit(1);
  }
  
  /* Create Replicate ID -> index table */
  int * repl_table = malloc(sizeof(int) * SGN_IntArray_max(uniq_repIDs));
  for(int i = 0; i < Y->replicates; i++){
	repl_table[SGN_IntArray_get(uniq_repIDs, i)] = i;
  }

  /* Allocation of memory for tables and matrices */
  Y->Y = malloc(sizeof(double) * Y->p * Y->samples);
  Y->table = malloc(sizeof(int *) * Y->replicates);
  int * ptr = malloc(sizeof(int) * Y->replicates * Y->N);
  for(int i = 0; i < Y->replicates; i++){
	Y->table[i] = ptr;
	ptr += Y->N;
	for(int j = 0; j < Y->N; j++) Y->table[i][j] = -1;
  }
  Y->missing = malloc(sizeof(int *) * Y->samples);

  /* Set values into Y->Y, Y->missing and Y->table */
  SGN_IntArray * missing = SGN_IntArray_new();
  SGN_IntArray * missing_pos = SGN_IntArray_new();
  for(int k = 0; k < Y->samples; k++){
	int n = SGN_IntArray_get(times, k) - 1;
	int l = repl_table[SGN_IntArray_get(repIDs, k)];

	Y->table[l][n] = k;
	for(int i = 0; i < Y->p; i++){
	  double v = SGN_DoubleArray_ptr(values)[k + i * Y->samples];
	  if(isnan(v)){
		SGN_IntArray_add(missing, i);
		v = 0.0;
	  }
	  Y->Y[i + k * Y->p] = v;
	}
	SGN_IntArray_add(missing, -1); /* End marker */
	SGN_IntArray_add(missing_pos, SGN_IntArray_size(missing));
  }
  Y->missing[0] = SGN_IntArray_gen_array(missing);
  for(int i = 1; i < Y->samples; i++){
	Y->missing[i] = Y->missing[0] + SGN_IntArray_get(missing_pos, i - 1);
  }

  Y->names = (char **) names->ptr;

  SGN_IntArray_delete(missing_pos);
  SGN_IntArray_delete(missing);
  SGN_IntArray_delete(uniq_repIDs);
  SGN_IntArray_delete(repIDs);
  SGN_IntArray_delete(times);
  SGN_DoubleArray_delete(values);
  free(repl_table);
  free(buff);
  free(names); /* Do not need to delete the instance itself. */

  return Y;
}
/* SSMData_read_fp() ======================================================= */

/*!\brief [DEBUG] Prints all the values in the SSMData instance.
 *
 *\param Y  [INPUT] SSMData instance to be printed.
 *\param fp  Output file stream.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMData */
void SSMData_dump(SSMData * Y, FILE * fp){
  for(int i = 0; i < Y->replicates; i++){
	fprintf(SIGN_ERR, "Rep #%d;", i + 1);
	for(int j = 0; j < Y->N; j++){
	  fprintf(SIGN_ERR, "  %d:%d", j, Y->table[i][j]);
	}
	fprintf(SIGN_ERR, "\n");
  }
  fflush(SIGN_ERR);

  fprintf(SIGN_ERR, "Missing:\n");
  for(int i = 0; i < Y->samples; i++){
	fprintf(SIGN_ERR, "[%d] ", i);
	int j = 0;
	do {
	  fprintf(SIGN_ERR, "%d ", Y->missing[i][j]);
	} while(Y->missing[i][j++] != -1);
	fprintf(SIGN_ERR, "\n");
  }
  fflush(SIGN_ERR);

  fprintf(SIGN_ERR, "Data:\n");
  for(int i = 0; i < Y->p; i++){
	fprintf(SIGN_ERR, "[%d]", i);
	for(int j = 0; j < Y->samples; j++){
	  fprintf(SIGN_ERR, "  %f", Y->Y[i + j * Y->p]);
	}
	fprintf(SIGN_ERR, "\n");
  }
  fflush(SIGN_ERR);
}
/* SSMData_dump() ========================================================== */

#ifdef USE_MPI
#include <mpi.h>

/*!\brief [MPI] Broadcast the dataset to all the MPI processes.
 *
 * The current implementation only supports SSMData instances
 * generated by SSM_read_fp().
 *
 *\param pY  Pointer to the pointer to the SSMData instance.  If the
 *   root process calls this, then it broadcasts this SSMData instance
 *   to all the processes associated with \a comm MPI communicator.
 *   Otherwise, the broadcasted instance is set to this pointer.
 *\param root  Root MPI rank ID to broadcast.
 *\param comm MPI communicator to which the SSMData instance is 
 *  broadcasted.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMData */
void SSMData_MPI_Bcast(SSMData ** pY, int root, MPI_Comm comm){
  int rank;
  MPI_Comm_rank(comm, &rank);
  SSMData * Y;
  if(rank != root){
	*pY = malloc(sizeof(SSMData));
  }
  Y = *pY;

  /* At first, broadcast 5 int values: p, N, samples, replicates and missings */
  int iar[5];
  if(rank == root){
	iar[0] = Y->p;
	iar[1] = Y->N;
	iar[2] = Y->samples;
	iar[3] = Y->replicates;

	/* Count the total size of missing */
	iar[4] = 0;
	for(int i = 0; i < Y->samples; i++){
	  int j = 0;
	  while(Y->missing[i][j++] != -1);
	  iar[4] += j;
	}/* for i < Y->samples */
  }
  MPI_Bcast(iar, 5, MPI_INT, root, comm);
  if(rank != root){
	Y->p = iar[0];
	Y->N = iar[1];
	Y->samples = iar[2];
	Y->replicates = iar[3];
	Y->Y = malloc(sizeof(double) * Y->p * Y->samples);
	Y->table = malloc(sizeof(int *) * Y->replicates);
	Y->table[0] = malloc(sizeof(int) * Y->replicates * Y->N);
	for(int i = 1; i < Y->replicates; i++) Y->table[i] = Y->table[0] + Y->N * i;
	Y->missing = malloc(sizeof(int *) * Y->samples);
	Y->missing[0] = malloc(sizeof(int) * iar[4]);
  }
  MPI_Bcast(Y->Y, Y->p * Y->samples, MPI_DOUBLE, root, comm);
  MPI_Bcast(Y->table[0], Y->replicates * Y->N, MPI_INT, root, comm);
  MPI_Bcast(Y->missing[0], iar[4], MPI_INT, root, comm);

  /* Calculate the position of missing values */
  if(rank != root){
	int i = 1;
	for(int j = 0; j < iar[4] - 1; j++){
	  if(Y->missing[0][j] == -1){
		Y->missing[i++] = Y->missing[0] + j + 1;
	  }
	}/* for j < iar[4] */
	Y->names = NULL; /* names field is not shared among processes. */
  }
}
/* SSMData_MPI_Bcast() ===================================================== */
#endif

/* End of file.  Copyright (C) 2010  HGC, IMS, The University of Tokyo  */
