/*
  SSMData.h : SiGN-SSM - Gene Network Estimation with State Sapce Model

  Copyright (C) 2010
    Yoshinori Tamada, Rui Yamaguchi, Seiya Imoto, Osamu Hirose, Ryo Yoshida,
	Masao Nagasaki and Satoru Miyano
      Laboratory of DNA Information Analysis &
	  Laboratory of Sequence Analysis, Human Genome Center,
      The Institute of Medical Science, The Unversity of Tokyo

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  
  Visit <http://sign.hgc.jp/signssm/> for more information.

  Written by Yoshinori Tamada <tamada at ims.u-tokyo.ac.jp>
  Originally developed by Osamu Hirose, Ryo Yoshida, Seiya Imoto, Rui Yamaguchi,
                          Tomoyuki Higuchi, and Satoru Miyano.
*/

#ifndef __SSM_DATA_H
#define __SSM_DATA_H

#include <stdio.h>
#ifdef USE_SIGN
#include "util/SGN_Data.h"
#endif

/*!\brief Structure for the time series data with replicates.
 *
 * To generate an \ref SSMData instance, use SSMData_new_Dataset().
 *
 *\sa SSM, SSM_EM, SSMState, SSMSuffStat, SSMInit, SSMResult, SSMWork, \ref Kalman
 */
typedef struct {
  /*!\brief \ref p by \ref samples fortran-style (column-first) matrix.

	This contains the input data values as a single matrix.
	The value of the i-th object at the j-th sample can be
	accessible as Y[i + j * p].

	Field \ref table specifies which column represents the replicate and 
	the time point.
  */
  double * Y;  

  /*!\brief Number of observed samples. */
  int samples; 

  /*!\brief Number of objects. */
  int p;       

  /*!\brief Length of time points. */
  int N;

  /*!\brief Number of replicates. */
  int replicates;

  /*!\brief Column position index table for replicated data.
   *
   * (\ref replicates)-by-(\ref N) C-style 2-dimensional array where
   * \ref table<code>[l][n]</code> represents the column position of the <code>n</code>-th 
   * time point in the <code>l</code>-th replicate for \ref Y.
   * If the time point is not observed (or does not exist in the input dataset),
   * then <code>-1</code> is set.
   */
  int ** table;

  /*!\brief Arrays containing the row positions of missing (NA, NaN) values.
   *
   * This points to a C-style 2-dimensional array where the length of the
   * first dimension is \ref samples and each first dimension element
   * corresponds to the list of row positions of the missing values for
   * columns in \ref Y.
   * Thus the length of the second dimension is dependant on the number of
   * missing values at that column position of \ref Y.  -1 is set at the
   * end of arrays for each column position.
   */
  int ** missing;

  /*!\brief Gene names.
   *
   * Gene names are not sent by SSMData_MPI_Bcast().
   */
  char ** names;
} SSMData;

/* Proto-types */

#ifdef USE_SIGN
SSMData * SSMData_new_Data(const SGN_Data * data);
#endif
void      SSMData_output_fp(SSMData * Y, FILE * fp, int * table, int t);
void      SSMData_output(SSMData * Y, const char * file, int rep);
void      SSMData_output2(SSMData * Y, const char * prefix, const char * postfix, int rep);
void      SSMData_node_mean_shift(SSMData * Y, int replicate);
SSMData * SSMData_read_fp(FILE * fp);
void      SSMData_dump(SSMData * Y, FILE * fp);

#ifdef USE_MPI
#include <mpi.h>
void SSMData_MPI_Bcast(SSMData ** pY, int root, MPI_Comm comm);
#endif

#endif /* __SSM_DATA_H */

/* End of file.  Copyright (C) 2010  HGC, IMS, The University of Tokyo       */
