/*
  SSMResult.c : SiGN-SSM - Gene Network Estimation with State Sapce Model

  Copyright (C) 2010-2012
    Yoshinori Tamada, Rui Yamaguchi, Seiya Imoto, Osamu Hirose, Ryo Yoshida,
	Masao Nagasaki and Satoru Miyano
      Laboratory of DNA Information Analysis &
	  Laboratory of Sequence Analysis, Human Genome Center,
      The Institute of Medical Science, The Unversity of Tokyo

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  
  Visit <http://sign.hgc.jp/signssm/> for more information.

  Written by Yoshinori Tamada <tamada at ims.u-tokyo.ac.jp>
  Originally developed by Osamu Hirose, Ryo Yoshida, Seiya Imoto, Rui Yamaguchi,
                          Tomoyuki Higuchi, and Satoru Miyano.
*/

#include "signcom.h"
#include "SSMResult.h"
#include <util/SGN_Str.h>
#include <math/SGN_Matrix.h>
#ifdef SSM_DEBUG
#include "debug.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <strings.h>

void SSMResult_alloc(SSMResult * result, int p, int dim){
  result->D = malloc(sizeof(double) * dim * p);
  result->L = malloc(sizeof(double) * dim);
}
/* SSMResult_alloc() ======================================================= */

void SSMResult_free(SSMResult * result){
  free(result->D); 
  free(result->L);
}
/* SSMResult_alloc() ======================================================= */

/*!\brief Outputs a matrix to a single file.
 *
 * This is the internal function to output matrices into a file,
 * used in SSMResult_output().
 *
 *\param file  file name.
 *\param X     pointer to a matrix to be printed.
 *\param m     Rows of the matrix \a X.
 *\param n     Columns of the matrix \a X.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMResult  */
void SSMResult_output_matrix(const char * file, double * X, int m, int n){
  FILE * fp = fopen(file, "w");
  if(fp == NULL){
	fprintf(SIGN_ERR, "ERROR: Failed to open file: %s\n", file);
	exit(1);
  }
  SGN_Matrix_printf(fp, X, m, n, "%.15f");
  fclose(fp);
}
/* SSMResult_output_matrix() =============================================== */

/*!\brief  Applies the constraint and calculates SSMResult::M and SSMResult::E.
 *
 *\param result  Estimation result by SSM_EM_algo() or SSM_EM_iterate().
 *\param work    Working memory.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMResult */
void SSMResult_calc_D_L(SSMResult * result, SSMWork * work){

  /* Recycle working memory */
  //double * L = work->Ke;     /* (k)-vector */
  double * sqR = work->e;    /* (p)-vector */
  double * S = work->w;      /* (k)-vector */
  double * tV = work->M;     /* (k,k)-matrix */
  //double * D = work->RinvH;  /* (k,p) = (p,k)-matrix */
  
  SSM_constrain(result->ssm, result->p, result->dim, result->L, sqR, S, tV,
				work->SVD, work->dgesvd_work_size);
  SSM_calc_module(result->p, result->dim, result->ssm->H, result->L, sqR, 
				  result->D);

  //result->D = D;
  //result->L = L;
}
/* SSMResult_calc_D_L() ==================================================== */

/*!\brief Outputs the estimated results of SSM model parameters.
 *
 * This outputs eight matrices (vectors) in a single file.
 * The file name would be
 * &quot;<i>prefix</i><tt>.D</tt><i>000</i><tt>.S</tt><i>000</i><tt>.A.</tt><i>postfix</i>&quot;
 * where <tt>D</tt><i>000</i>: Size of the dimension, and <tt>S</tt><i>000</i>: Set ID specified by
 * \a id. The matrices to be output are:
 *  - SSM::exp      : (\a k)-vector of the initial state variables (x<sub>0</sub>).
 *  - SSM::var      : (\a k, \a k)-matrix of the variance matrix of the state variables.
 *  - SSM::H        : (\a p, \a k)-observation matrix <i>H</i>.
 *  - SSM::R        : (\a p)-observation noise vector (diagonal elements of (\a p, \a p)-matrix) <i>R</i>.
 *  - SSM::F        : (\a k, \a k)-system matrix <i>F</i>.
 *  - SSM::Q        : (\a k)-system noise vector (diagonal elements of (\a k, \a k)-matrix) <i>R</i>.
 *  - SSMResult::D  : (\a k, \a p)-object-to-module projection matrix.
 *  - SSMResult::L  : (\a k)-diagonal elements of the diagonal matrix <i>L</i> = <i>H</i> ' <i>R</i><sup>-1</sup> <i>H</i>.
 *
 * In addition, the result summary file
 * &quot;<i>prefix</i><tt>.D</tt><i>000</i><tt>.S</tt><i>000</i><tt>.B.</tt><i>postfix</i>&quot;
 * is also output.
 *
 * To output each matrix in different files, use SSMResult_output_each().
 * The output file can be read by SSMResult_read_fp().
 *
 *\memberof SSMResult */
void SSMResult_output(SSMResult * result, const char * prefix, const char * postfix, int id){
  int d = 18; /* 18 digits */
  char buff[SSM_BUFF_SIZE];
  int len = snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.A.%s", prefix, result->dim, id, postfix);
  if(SSM_BUFF_SIZE <= len){
	fprintf(SIGN_ERR, "ERROR: Output prefix and/or postfix too long.\n");
	exit(1);
  }
  fprintf(SIGN_ERR, "Outputting the result into a single file: %s\n", buff);
  fflush(SIGN_ERR);
  FILE * fp = fopen(buff, "w");
  SSM_write_fp(result->ssm, result->p, result->dim, fp);
  fprintf(fp, "%s\t%d\t%d\n", "SSMResult::D", result->dim, result->p);
  SGN_Matrix_fprintf(fp, d, result->D, result->dim, result->p);
  fprintf(fp, "\n");
  fprintf(fp, "%s\t%d\t%d\n", "SSMResult::L", result->dim, 1);
  SGN_Matrix_fprintf(fp, d, result->L, result->dim, 1);
  fprintf(fp, "\n");
  fclose(fp);

  snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.B.%s", prefix, result->dim, id, postfix);
  fprintf(SIGN_ERR, "Outputting the summary file: %s\n", buff);
  fp = fopen(buff, "w");
  SSMResult_output_summary_fp(result, id, fp);
  fclose(fp);
}
/* SSMResult_output() ====================================================== */

/*!\brief Outputs the estimated results of SSM model parameters.
 *
 * This makes nine (8) files named 
 * &quot;<i>prefix</i><tt>.D</tt><i>000</i><tt>.S</tt><i>000</i><tt>.</tt><i>X</i><tt>.</tt><i>postfix</i>&quot;
 * where <i>X</i> is one of
 *  - <tt>H</tt> : (\a p, \a k)-observation matrix <i>H</i>.
 *  - <tt>R</tt> : (\a p)-observation noise vector (diagonal elements of (\a p, \a p)-matrix) <i>R</i>.
 *  - <tt>F</tt> : (\a k, \a k)-system matrix <i>F</i>.
 *  - <tt>D</tt> : (\a k, \a p)-object-to-module projection matrix.
 *  - <tt>L</tt> : (\a k)-diagonal elements of the diagonal matrix <i>L</i> = <i>H</i> ' <i>R</i><sup>-1</sup> <i>H</i>.
 *  - <tt>x</tt> : (\a k)-vector of the initial state variables (x<sub>0</sub>).
 *  - <tt>B</tt> : Estimation Summary includeing the BIC value,
 *
 * and <tt>D</tt><i>000</i>: Size of the dimension and <tt>S</tt><i>000</i>: Set ID specified by
 * \a id.
 *
 * After the estimation of the model parameters, use SSMResult_calc_M_E() to
 * apply contraint on the model parameters.
 *
 *\param result   Pointer of the EM algorithm result after calling SSMResult_calc_M_E().
 *\param prefix   Output file prefix.
 *\param postfix  Output file postfix (extension).
 *\param id       ID inserted to the output file name.
 *
 *\sa SSMResult_calc_D_E(), SSM_EM_iterate(), SSM_EM_algo()
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMResult */
void SSMResult_output_each(SSMResult * result, const char * prefix, const char * postfix, int id){
  char buff[SSM_BUFF_SIZE];

  int len = snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.%s.%s", prefix, result->dim, id, "*", postfix);
  if(SSM_BUFF_SIZE <= len){
	fprintf(SIGN_ERR, "ERROR: Output prefix and/or postfix too long.\n");
	exit(1);
  }
  fprintf(SIGN_ERR, "Outputting files: %s\n", buff);
  fflush(SIGN_ERR);

  snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.%s.%s", prefix, result->dim, id, "H", postfix);
  fprintf(SIGN_ERR, "%s ...\n", buff);
  SSMResult_output_matrix(buff, result->ssm->H, result->p, result->dim);

  snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.%s.%s", prefix, result->dim, id, "R", postfix);
  fprintf(SIGN_ERR, "%s ...\n", buff);
  SSMResult_output_matrix(buff, result->ssm->R, result->p, 1);

  snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.%s.%s", prefix, result->dim, id, "F", postfix);
  fprintf(SIGN_ERR, "%s ...\n", buff);
  SSMResult_output_matrix(buff, result->ssm->F, result->dim, result->dim);

  snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.%s.%s", prefix, result->dim, id, "D", postfix);
  fprintf(SIGN_ERR, "%s ...\n", buff);
  SSMResult_output_matrix(buff, result->D, result->dim, result->p);

  snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.%s.%s", prefix, result->dim, id, "L", postfix);
  fprintf(SIGN_ERR, "%s ...\n", buff);
  SSMResult_output_matrix(buff, result->L, result->dim, 1);

  snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.%s.%s", prefix, result->dim, id, "x", postfix);
  fprintf(SIGN_ERR, "%s ...\n", buff);
  SSMResult_output_matrix(buff, result->ssm->exp, result->dim, 1);

  snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.%s.%s", prefix, result->dim, id, "B", postfix);
  fprintf(SIGN_ERR, "%s ...\n", buff);
  FILE * fp = fopen(buff, "w");
  SSMResult_output_summary_fp(result, id, fp);
  fclose(fp);
#ifdef SSM_DEBUG
  SSM_check(result->ssm, result->p, result->dim, 1);
  debug_check("Output:D", result->D, result->dim, result->p, SSM_DEBUG, 1);
  debug_check("Output:L", result->L, result->dim, 1, SSM_DEBUG, 0);
#endif
}
/* SSMResult_output_each() ================================================= */

/*!\brief Outputs the estimation summary into a file stream.
 *
 *\param result   Pointer of the EM algorithm result after calling SSMResult_calc_D_L().
 *\param fp       Output file stream
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMResult */
void SSMResult_output_summary_fp(SSMResult * result, int id, FILE * fp){
  /* FORMAT:  dim  set_id  LLH  BIC LOOPS  Monotone converged */
  fprintf(fp, "%03d\t%03d\t%f\t%f\t%d\t%s\t%s\n", 
		  result->dim, id, result->likelihood, result->bic, result->loops, 
		  (result->monotone ? "Monotone" : "Not Monotone"),
		  (result->converged ? "Converged" : "Not Converged"));
}
/* SSMResult_output_summary_fp() =========================================== */

/*!\brief Reads the estimation result file output by SSMResult_output().
 *
 * New memory space is allocated for SSM, SSMResult::D and SSMResult::L.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMResult */
SSMResult SSMResult_read_fp(FILE * fp){
  SSMResult result;
  char * buff = NULL;
  size_t buff_size = 0;
  char * pos;
  
  result.likelihood = 0;
  result.bic = 0;
  result.monotone = 0;
  result.loops = 0;

  /* Read SSM model parameters. */
  fprintf(SIGN_ERR, "  Reading model parameters.\n");
  fflush(SIGN_ERR);
  result.ssm = SSM_read_fp(fp, &(result.p), &(result.dim));
  fprintf(SIGN_ERR, "    p=%d, dim=%d\n", result.p, result.dim);
  fflush(SIGN_ERR);

  result.D = malloc(sizeof(double) * result.dim * result.p);
  result.L = malloc(sizeof(double) * result.dim);

  fprintf(SIGN_ERR, "  Reading aditional matrices.\n");
  fflush(SIGN_ERR);
  /* Read D: (k, p)-projection matrix. */
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the first label line. */
  fflush(SIGN_ERR);
  for(int i = 0; i < result.dim; i++){
    SGN_Str_readline(fp, &buff, &buff_size);
	pos = buff;
	for(int j = 0; j < result.p; j++){
	  if(pos == (char *) 1){
	    fprintf(SIGN_ERR, "ERROR: Number of columns inconsistent.\n");
	    exit(1);
	  }
	  result.D[i + j * result.dim] = atof(pos);
	  pos = strchr(pos, '\t') + 1;
	}
  }
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the empty line. */
  
  /* Read L: (k)-diagonal matrix. */
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the first label line. */
  pos = buff;
  for(int i = 0; i < result.dim; i++){
	SGN_Str_readline(fp, &buff, &buff_size);
	result.L[i] = atof(buff);
  }
  SGN_Str_readline(fp, &buff, &buff_size); /* Skip the empty line. */

  free(buff);
  return result;
}
/* SSMResult_read_fp() ===================================================== */

/*!\brief Calculates an object-to-object coefficient matrix.
 *
 * This calculates \a Psi = \a D' \a L \a F \a D where
 *  \a D is SSMResult::D, \a L diag( SSMResult::L ), and \a F SSMResult::SSM::F.
 *
 *\param result [INPUT] SSM estimation result.
 *\param Psi    [OUTPUT] (\a p, \a p)-coefficient matrix. The (i,j)-element represents
 *   the effect from the j-th object to the i-th object.
 *\param  W    [WORK]   (\a k, \a p)-working matrix.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSMResult */
void SSMResult_calc_Psi(SSMResult * result, double * Psi, double * W){
  /* Psi := D '   diag(L)     F     D     */
  /*        (p,k) (k,k) (k,k) (k,p) */
  /* Psi_ij represents j -> i edge effect. */

  int k = result->dim;
  int p = result->p;

  /* Psi     := diag(L)    F     */  /* Use Psi as working are */
  /* (k,k)   (k,k) (k,k) */
  memcpy(Psi, result->ssm->F, sizeof(double) * k * k);
  SGN_Matrix_dgedm2(k, k, result->L, Psi, k);

  /* W    := Psi    D    */
  /* (k,p)   (k,k) (k,p) */
  /*       =  diag(L)  F  D   */
  SGN_Matrix_dgemm(SGN_Matrix_NO_TRANS, SGN_Matrix_NO_TRANS, &k, &p, &k,
				   &SGN_Matrix_DOUBLE_ONE, Psi, &k,
				   result->D, &k,
				   &SGN_Matrix_DOUBLE_ZERO, W, &k);

  /* Psi  :=  D'   W   */
  /* (p,p)   (p,k) (k,p) */
  /*       =  D'  diag(L)  F  D  */
  SGN_Matrix_dgemm(SGN_Matrix_TRANS, SGN_Matrix_NO_TRANS, &p, &p, &k,
				   &SGN_Matrix_DOUBLE_ONE, result->D, &k,
				   W, &k,
				   &SGN_Matrix_DOUBLE_ZERO, Psi, &p);
}
/* SSMResult_calc_Psi() ==================================================== */

/* End of file.  Copyright (C) 2010-2012  HGC, IMS, The University of Tokyo  */
