/*
  SSM_EM.c : SiGN-SSM - Gene Network Estimation with State Sapce Model

  Copyright (C) 2010
    Yoshinori Tamada, Rui Yamaguchi, Seiya Imoto, Osamu Hirose, Ryo Yoshida,
	Masao Nagasaki and Satoru Miyano
      Laboratory of DNA Information Analysis &
	  Laboratory of Sequence Analysis, Human Genome Center,
      The Institute of Medical Science, The Unversity of Tokyo

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  
  Visit <http://sign.hgc.jp/signssm/> for more information.

  Written by Yoshinori Tamada <tamada at ims.u-tokyo.ac.jp>
  Originally developed by Osamu Hirose, Ryo Yoshida, Seiya Imoto, Rui Yamaguchi,
                          Tomoyuki Higuchi, and Satoru Miyano.
*/

#include "signcom.h"
#include "SSM.h"
#include "SSM_Defaults.h"
#include "SSM_EM.h"
#include "SSMData.h"
#include "SSMResult.h"
#include "Kalman.h"

#ifdef SSM_DEBUG
#include "debug.h"
#endif

#include <math/SGN_Matrix.h>

#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>

/*!\brief Generates a new SSM_EM instance initialized by default values.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM_EM */
SSM_EM * SSM_EM_new(){
  SSM_EM * em = malloc(sizeof(SSM_EM));
  
  em->converge     = 1.0E-5;
  em->loops        = SSM_DEFAULT_EM_LOOPS;
  em->R_rI         = 0;
  em->update_mu0   = 1;
  em->abs_converge = 1;
  em->ridge        = 0.0;
  em->constrain_F  = 1;
  em->g            = SSM_DEFAULT_CONSTRAIN_F_G;
  em->retry        = -1;
  em->verbose      = 0;

  return em;
}
/* SSM_EM_new() ============================================================ */

/*!\brief Iterates the EM algorithm for the specified times.
 *
 * This tries to estimate the SSM model paramters with different initial
 * model parameters for \a iterations times, and then returns the best result
 * in terms of the log likelihood.
 *
 *\param em         [INPUT] Parameters for the EM algorithm.
 *\param init       [INPUT] Used to generate the initial model parameters.
 *\param iterations [INPUT] Number of iterations of the EM algorithm.  The best result
 *                       in these \a iterations will return.
 *\param Y          [INPUT] Input time series data.
 *\param k          [INPUT] Size of dimension.
 *\param ssm1       [WORK]  Either \a ssm1 or \a ssm2 is used to return the best result
 *                             in the returning SSMResult structure.
 *\param ssm2       [WORK]  Either \a ssm1 or \a ssm2 is used to return the best result
 *                             in the returning SSMResult structure.
 *\param X          [WORK]  Working area for state variables.
 *\param T          [WORK]  Working area for sufficient statistics.
 *\param sumT       [WORK]  Working area for sufficient statistics.
 *\param work       [WORK]  Working area.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM_EM */
SSMResult SSM_EM_iterate(SSM_EM * em, SSMInit * init, int iterations,
						 SSMData * Y, int k,
						 SSM * ssm1, SSM * ssm2,
						 SSMState * X, SSMSuffStat * T, SSMSuffStat * sumT,
						 SSMWork * work){

  SSMResult best;
  best.likelihood = - DBL_MAX;
  /* Set values for suppressing warnings */
  best.bic = - DBL_MAX;
  best.ssm = NULL;
  best.loops = -1;
  best.converged = 0;
  best.dim = -1;
  best.p = -1;
  best.monotone = 0;
  best.D = NULL;
  best.L = NULL;

  for(int i = 0; i < iterations; i++){
	
	/* Perform the EM algorithm once. */
	SSMResult result;
	int retry = 0;
	while(1){
	  /* Generate the initial parameters. */
#if SSM_DEBUG == 2
	  /* Check-Out mode */
	  fprintf(SIGN_ERR, "SSM_EM_iterate: Reading SSM model parameters from file.\n");
	  fflush(SIGN_ERR);
	  ssm1 = SSM_read_fp(DEBUG_FP, &Y->p, &k);
	  //SSM_check(ssm1, Y->p, k, 0);
#elif SSM_DEBUG == 1
	  /* Writing-Out mode or Normal */
	  fprintf(SIGN_ERR, "SSM_EM_iterate: Initializing SSM model parameters at random.\n");
	  SSM_init(ssm1, init, Y->p, k, Y->Y, Y->samples);
	  SSM_write_fp(ssm1, Y->p, k, DEBUG_FP);
#else
	  SSM_init(ssm1, init, Y->p, k, Y->Y, Y->samples);
#endif
	  if(em->verbose > 0) SSM_write_fp(ssm1, Y->p, k, SIGN_ERR);
	  result = SSM_EM_algo(em, ssm1, X, T, sumT, Y, k, work);

	  if((em->retry < 0 || retry < em->retry)
		 && (!isnormal(result.likelihood) || result.monotone == 0)){
		retry++;
		fprintf(SIGN_ERR, "Estimation failed.  Retrying (%d)...\n", retry);
		fflush(SIGN_ERR);
		continue;
	  } else {
		if(em->retry != -1 && retry >= em->retry){
		  fprintf(SIGN_ERR, "WARNING: Retry count exceeded the limitation.\n");
		  fflush(SIGN_ERR);
		}
		break;
	  }
	}

	fprintf(SIGN_ERR, "%d: Loops=%d, Likelihood=%f\n", i + 1, result.loops, result.likelihood);
	fflush(SIGN_ERR);

	if(result.likelihood > best.likelihood){
	  /* Save the best result */
	  best = result; /* Shallow copy first */
	  /*     Next, simply swapping the pointers here. */
	  SSM * tmp = ssm2;
	  ssm2 = ssm1;
	  ssm1 = tmp;
	}
  }/* for i < iterations */

  best.ssm = ssm2;

  return best;
}
/* SSM_EM_iterate() ======================================================== */

/*!\brief Performs the EM algorithm to estimate SSM model parameters.
 *
 * This performs the EM algorithm to estimate the parameters of the SSM (State
 * Space Model).  The algorithm starts with the initial model parameters
 * specified in \a ssm, and then iterates updating of the model parameters
 * by EM algorithm until the algorithm is converged.
 *
 * The initial model parameters can be set by SSM_init().
 *
 * The results are returned as a value of SSMResult structure.
 * The estimated parameters are also set in \a ssm.
 *
 * This calls the following sub routines:
 *  - SSM_EM_estimate()
 *  - SSM_EM_update()
 *
 * This is called by SSM_EM_iterate() that iterates this routine to
 * estimate the best parameters in terms of the log likelihood with different
 * initial values of the model parameters.
 *
 *\param em   [INPUT] Parameters of the EM algorithm.  Unchanged on exit.
 *\param ssm  [INPUT/OUTPUT] Model parameters with initial values.
 *            Values of the fields are updated with the estimated results.
 *\param X    [WORK] Working area for state variables.
 *\param T    [WORK] Working area for sufficient statistics.
 *\param sumT [WORK] Working area for sufficient statistics.
 *\param Y  Input time series data.
 *\param k    Size of dimension.
 *\param work [WORK] Working area.
 *
 *\return  SSMResult structure containing the esimated model parameters and the information
 *  about the execution results of the EM algorithm.  SSMResult::ssm is set by the same
 *  value as \a ssm.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM_EM */
SSMResult SSM_EM_algo(SSM_EM * em, SSM * ssm, SSMState * X, 
				SSMSuffStat * T, SSMSuffStat * sumT, 
				SSMData * Y, int k, SSMWork * work){

  int i;
  SSMResult result;
  double prev_llh = - DBL_MAX;
  double llh = - DBL_MAX;

  /* Set the initial values to SSMResult. */
  result.dim = k;
  result.p = Y->p;
  result.converged = 0;
  result.monotone = 1;
  result.D = NULL;
  result.L = NULL;

  /* At most, try to update parameters "em->loops" times. */
  i = 0;
  while(1){

	/* Perform a single iteration of the EM algorithm. */
	llh = SSM_EM_estimate(ssm, X, T, sumT, Y, k, em->ridge, work);

	if(i > (em->loops / 10) && isnan(llh)) break;

	/* Check convergence */
	double diff = llh - prev_llh;
	if(em->abs_converge) diff = fabs(diff);
	if(diff < 0) result.monotone = 0;
	if(!isnan(llh) && diff < em->converge){
	  result.converged = 1;
	  break;
	}
	if(i == em->loops - 1) break;
	
	/* Update model parameters */
	SSM_EM_update(ssm, sumT, Y->p, k, Y->replicates, Y->samples,
				  em->update_mu0, em->R_rI, work->C, em->g, work);
	prev_llh = llh;

#ifndef SSM_DEBUG
	if(em->verbose >= 2){
#else
	{
#endif
	  fprintf(SIGN_ERR, "(%d) llh = %f\n", i + 1, llh);
	  fflush(SIGN_ERR);
	}
	i++;
  }/* while(1) */

  if(em->verbose >= 1){
	fprintf(SIGN_ERR, "SSM_EM_algo(): Estimation finished. llh=%f\n", llh);
	fflush(SIGN_ERR);
  }

  /* Set the returning results. */
  result.likelihood = llh;
  result.loops = i + 1;
  result.ssm = ssm;
  /* set BIC here */
  double df = Y->p * (k + 1) + k * (k + 1) / 2.0;
  if(em->update_mu0) df += k;
  result.bic = -2.0 * llh + log(Y->samples) * df;
#ifdef SSM_DEBUG
  fprintf(SIGN_ERR, "BIC: df=%f\n", df);
  fprintf(SIGN_ERR, "BIC: log(N)=%f\n", log(Y->samples));
  fprintf(SIGN_ERR, "BIC: llh=%f\n", llh);
  fprintf(SIGN_ERR, "BIC: BIC=%f\n", result.bic);
#endif

  return result;
}
/* SSM_EM_algo() =========================================================== */

/*!\brief Performs the EM algorithm estimation for one step.
 *
 * Called by SSM_EM_algo().
 *
 * This performs the EM algorithm for one step by calling mainly:
 *  - Kalman_filter()
 *  - Kalman_smoother()
 *  - Kalman_lagone_cov()
 *  - SSMSuffStat_calc()
 *
 *\param ssm  [INPUT] SSM model parameters.
 *\param X    [OUTPUT] SSM state variables.
 *\param T    [WORK]  Working are for sufficient statistics.
 *\param sumT [OUTPUT]  Final sufficient statistics used for updating model parameters.
 *\param Y    [INPUT]  Input time series data.
 *\param k     Size of dimension.
 *\param ridge Reidge parameter.
 *\param work  [WORK] Working area.
 *
 *\return Log likelihood of the updated model parameters.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM_EM */
double SSM_EM_estimate(SSM * ssm, SSMState * X, 
					   SSMSuffStat * T,SSMSuffStat * sumT,
					   SSMData * Y, int k, double ridge, SSMWork * work){

  /* Zero-clear sumT. Do not need to clear T because all the fileds in it
     will be over-written. */
  SSMSuffStat_set_zero(sumT, Y->p, k);

  double llh = 0.0; /* Log likelihood */

  /* Executes below for all the replicates */
  for(int i = 0; i < Y->replicates; i++){
	llh += Kalman_filter(ssm, X, Y->p, k, Y->N,
						 Y->Y, Y->table[i], Y->missing, work);
	Kalman_smoother(X, k, Y->N, work->w, work->M, work->LU);
	Kalman_lagone_cov(X, k, Y->N);
	SSMSuffStat_calc(T, X, ssm, Y->p, k, Y->N, Y->table[i], Y->Y,
					 Y->missing, work->Eyy, work->Eyx, work->Exx, work->w);

	SSMSuffStat_add(T, sumT, Y->p, k);
	
  }/* for i < # of replicates */

  if(ridge > 0.0) SSM_EM_ridge(k, ridge, T->Txx, T->Tx1x1);

  return llh;
}
/* SSM_EM_estimate() ======================================================= */

/*!\brief Generates the (\a k, \a k <sup>2</sup>)-constraint matrix \a C.
 *
 * This is for generating the constraint matrix used in
 * SSM_EM_update_constraint_F().
 *
 *\param k  Size of dimension.
 *\param C  [OUTPUT] (\a k, \a k <sup>2</sup>)-constraint matrix.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM_EM */
void SSM_EM_set_C(int k, double * C){
  int k_k = k * k;
  int k_k_k = k * k + k;
  memset(C, 0, sizeof(double) * k_k * k);

  for(int j = 0; j < k; j++){
	C[j] = 1.0;
	C += k_k_k;
  }
}
/* SSM_EM_set_C() ========================================================== */

/*!\brief Subroutine for SSM_EM_update_constraint_F(), that calculates
 *   the Kronecker product of \a Ainv and diag(\a Q).
 *
 *\param k    Size of dimension.
 *\param Ainv [INPUT]  (\a k, \a k)-matrix.
 *\param Q    [INPUT]  (\a k)-vector.
 *\param Z    [OUTPUT] (\a k <sup>2</sup>, \a k <sup>2</sup>)-matrix.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM_EM */
void SSM_EM_kron(int k, double * Ainv, double * Q, double * Z){
  int k_k = k * k;
  int k_k_km = k * k * (k - 1);
  for(int j = 0; j < k; j++){
	for(int i = 0; i < k; i++){
	  SGN_Matrix_diag(k, *(Ainv++), Q, Z, k_k);
	  Z += k;
	}
	Z += k_k_km;
  }
}
/* SSM_EM_kron() =========================================================== */

/*!\brief Updates the system matrix F with constriant.
 *
 * This updates the system matrix \a F by
 *
 * vec(\a F) := \a B + \a Z \a C ' {\a C \a Z \a C ' }<sup>-1</sup> {\a g - \a B }
 *
 * where
 * vec(\a X) = column vector created by stacking all the columns of \a X,<br/>
 * \a B = \a Txx1 \a Tx1x1 <sup>-1</sup>,<br/>
 * \a Z = \a Tx1x1 <sup>-1</sup> (X) diag(\a Q) &nbsp;&nbsp; ((X) represents Kronecker's product),<br/>
 * \a Q : (\a k)-vector containing diagonal elements of the system noise variances.<br/>
 * \a C : (\a k, \a k <sup>2</sup>)-constraint matrix,<br/>
 * \a g : (\a k, 1)-constant vector.
 *
 *\param Txx1  [INPUT] (\a k, \a k)-matrix (SSMSuffStat::Txx1).
 *\param Tx1x1 [INPUT] (\a k, \a k)-matrix (SSMSuffStat::Tx1x1).
 *\param Q     [INPUT] (\a k)-vector (SSM::Q).
 *\param C     [INPUT]   (a k, \a k <sup>2</sup>)-constraint matrix generated by SSM_EM_set_C().
 *\param g     Constant for the constraint of calculating \a F.
 *\param F     [OUTPUT]  Updated (\a k, \a k)-system matrix (SSM::F).
 *\param k     Size of dimension.
 *\param V     [WORK] At least (\a k, \a k <sup>2</sup>)-matrix.
 *\param W     [WORK] At least (\a k, \a k <sup>2</sup>)-matrix.
 *\param Z     [WORK] At least (\a k <sup>2</sup>, \a <sup>2</sup>)-matrix.
 *\param pivot [WORK] (\a k)-integer vector.
 *\param work  [WORK] Working mmemory for LAPACK dgetri routine for inversion of (\a k, \a k)-matrix.
 *\param worksize  Size of \a work.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM_EM */
void SSM_EM_update_constraint_F(double * Txx1, double * Tx1x1,
								double * Q, double * C, double g, double * F,
								int k, double * V, double * W, double * Z,
								int * pivot, double * work,	int worksize){
  int info;
  int k_k = k * k;
  /*-----------------------------------------------------
   * vec F := vec F + Z C' {C Z C'}^-1 {g - C vec F}
   *  where
   * C : Constraint matrix
   * F = Txx1  Tx1x1^-1  
   * Z = W (X)  diag(Q)   ( (X) is Kronecker's product )
   *---------------------------------------------------*/

  /* W  :=  Tx1x1^-1 */
  /* (k,k)  (k,k) */
  memcpy(W, Tx1x1, sizeof(double) * k * k);
  SGN_Matrix_dgetrf(&k, &k, W, &k, pivot, &info);
  SGN_Matrix_dgetri(&k, W, &k, pivot, work, &worksize, &info);
  /* F :=  Txx1  Tx1x1^-1 */
  /* (k,k) (k,k) (k,k)    */
  SGN_Matrix_dgemm(SGN_Matrix_NO_TRANS, SGN_Matrix_NO_TRANS, &k, &k, &k,
				   &SGN_Matrix_DOUBLE_ONE, Txx1, &k, W, &k,
				   &SGN_Matrix_DOUBLE_ZERO, F, &k);
  /* Z      := W (X) diag(Q)  */
  /* (k^2,k^2)                */
  SSM_EM_kron(k, W, Q, Z);
  /* Here Tx1x1^-1 in W is no longer needed. */
  /* W     := C        Z         */
  /* (k,k^2)  (k,k^2)  (k^2,k^2) */
  SGN_Matrix_dgemm(SGN_Matrix_NO_TRANS, SGN_Matrix_NO_TRANS, &k, &k_k, &k_k,
				   &SGN_Matrix_DOUBLE_ONE, C, &k, Z, &k_k,
				   &SGN_Matrix_DOUBLE_ZERO, W, &k);
  /* V   := W        C'       */
  /* (k,k)  (k,k^2)  (k^2,k)  */
  /*      = C Z C' */
  SGN_Matrix_dgemm(SGN_Matrix_NO_TRANS, SGN_Matrix_TRANS, &k, &k, &k_k,
				   &SGN_Matrix_DOUBLE_ONE, W, &k, C, &k_k,
				   &SGN_Matrix_DOUBLE_ZERO, V, &k);
  /* V  := V^-1   */
  /* (k,k)        */
  SGN_Matrix_dgetrf(&k, &k, V, &k, pivot, &info);
  SGN_Matrix_dgetri(&k, V, &k, pivot, work, &worksize, &info);
  /* W     := C'       V     */
  /* (k^2,k)  (k^2,k)  (k,k) */
  /*        = C' {C Z C'}^-1 */
  SGN_Matrix_dgemm(SGN_Matrix_TRANS, SGN_Matrix_NO_TRANS, &k_k, &k, &k,
				   &SGN_Matrix_DOUBLE_ONE, C, &k, V, &k,
				   &SGN_Matrix_DOUBLE_ZERO, W, &k_k);
  /* V    :=  Z         W         */
  /* (k^2,k)  (k^2,k^2) (k^2,k)   */
  /*       = Z C' {C Z C'}^-1    */
  SGN_Matrix_dgemm(SGN_Matrix_NO_TRANS, SGN_Matrix_NO_TRANS, &k_k, &k, &k_k,
				   &SGN_Matrix_DOUBLE_ONE, Z, &k_k, W, &k_k,
				   &SGN_Matrix_DOUBLE_ZERO, V, &k_k);
  /* W    :=  g   -    C       vec F.   :Simply extracting the diagonal elements */
  /* (k,1)   (k,1)   (k,k^2) (k^2,1) */  
  double * Fp = F;
  for(int i = 0; i < k; i++){
	W[i] = g - *Fp; /* (*Fp) corresponds to the i-th diagonal element. */
	Fp += (k + 1);
  }
  /* vec F     := vec F  + C       W */
  /*  (k^2,1)    (k^2,1) + (k^2,k) (k,1)  */
  SGN_Matrix_dgemv(SGN_Matrix_NO_TRANS, &k_k, &k, 
				   &SGN_Matrix_DOUBLE_ONE, C, &k_k,
				   W, &SGN_Matrix_INT_ONE,
				   &SGN_Matrix_DOUBLE_ONE, F, &SGN_Matrix_INT_ONE);
}
/* SSM_EM_update_constraint_F() ============================================ */

/*!\brief Calculates \a X = \a A \a B <sup>-1</sup>.
 *
 * called by SSM_EM_update().
 *
 *\param A  [INPUT] (\a p, \a k)-matrix. Unchanged on exit.
 *\param B  [INPUT] (\a k, \a k)-matrix, Unchanged on exit.
 *\param X    [OUTPUT] (\a p, \a k)-updated observation matrix.
 *\param p     Number of objects.
 *\param k     Size of dimension.
 *\param LU    [WORK] (\a k, \a k)-working matrix.
 *\param pivot [WORK] (\a k)-integer working vector.
 *\param W     [WORK] (\a p, \a k)-working matrix;
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM_EM */
void SSM_EM_calc_ABinv(double * A, double * B, double * X, int p, int k,
					   double * LU, int * pivot, double * W){
  /* C :=  A B^-1 */
  /* (p,k) :=  (p,k) (k,k) */ 
  /*  dgesv solves  A' C' = B' for C' i.e.  C' = A'^-1 B' */
  /*                                        C  = B A^-1   */
  int info;
  memcpy(LU, B, sizeof(double) * k * k);       /* LU := B */
  SGN_Matrix_dgetrf(&k, &k, LU, &k, pivot, &info); /* LU factorization of B */
  SGN_Matrix_trans(p, k, A, W);                    /* W := A'  (k, p) */
  SGN_Matrix_dgetrs(SGN_Matrix_L_TRANS, &k, &p, LU, &k, pivot, W, &k, &info); /* W := B'^-1 A' */
  SGN_Matrix_trans(k, p, W, X); /* X := W' =  A B^-1 */
}
/* SSM_EM_update_H() ======================================================= */

/*!\brief Updates SSM model parameters.
 *
 * Updates \a ssm->exp (SSM::exp), \a ssm->H (SSM::H), \a ssm->F (SSM::F), 
 * and \a ssm->R (SSM::R)
 * using the sufficient statisitics \a T.
 *
 * Called by SSM_EM_algo().
 *
 *\param ssm      SSM model parameters to be updated.
 *\param T        Sufficient statistics to update the SSM model parameters in \a ssm.
 *\param p        Number of objects (nodes).
 *\param k        Size of dimension.
 *\param m        Number of replicates.
 *\param samples  Number of total samples.
 *\param mu0_update   If non-zero, updates mu0 (initial state SSM::exp).
 *\param unique_r     If non-zero, assumes \a R = \a r \a I where R is SSM::R.
 *\param C     (\a k, \a k <sup>2</sup>-constraint matrix set by SSM_EM_set_C().
 *             If NULL, constraint on F (SSM::F) is not applied.
 *\param g     Constraint strength ranging from 0.0 to 1.0.
 *\param work  Working m.
 *
 *\sa SSM_EM_algo().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM_EM */
void SSM_EM_update(SSM * ssm, SSMSuffStat * T,
				   int p, int k, int m, int samples,
				   int mu0_update, int unique_r, 
				   double * C, double g, SSMWork * work){

  /* mu0 (ssm->exp): (k)-vector */
  /*  ssm->exp := Tmu0 / m */
  if(mu0_update){
	memcpy(ssm->exp, T->Tmu0, sizeof(double) * k);
	double v = 1.0 / (double) m;
	SGN_Matrix_dscal(&k, &v, ssm->exp, &SGN_Matrix_INT_ONE); 
  }

  /* H :=  Tyx Txx^-1 */
  /* (p,k) :=  (p,k) (k,k) */ 
  SSM_EM_calc_ABinv(T->Tyx, T->Txx, ssm->H, p, k,
					work->LU, work->pivot, work->Eyx);

  /* F := Txx1 Tx1x1^-1 */
  /* (k,k) := (k,k) (k,k)   How to calculate is the same as above. */
  if(C != NULL){
	SSM_EM_update_constraint_F(T->Txx1, T->Tx1x1,
							   ssm->Q, C, g, ssm->F, 
							   k, work->C1, work->C2, work->C3,
							   work->pivot, work->dgetri_work, work->dgetri_work_size);
  } else {
	SSM_EM_calc_ABinv(T->Txx1, T->Tx1x1, ssm->F, k, k, 
					  work->LU, work->pivot, work->Eyx);
  }
  
  /* R */
  for(int i = 0; i < p; i++){
	double c1 = T->Tyy[i]; /* Tyy: (p)-vector */
	double c2 = SGN_Matrix_ddot(&k, ssm->H + i, &p, T->Tyx + i, &p); /* H_i' Tyx_' */
	SGN_Matrix_dgemv(SGN_Matrix_NO_TRANS, &k, &k,
					 &SGN_Matrix_DOUBLE_ONE, T->Txx, &k, ssm->H + i, &p,     /* w := Txx H_i */
					 &SGN_Matrix_DOUBLE_ZERO, work->w, &SGN_Matrix_INT_ONE);    /* (k,1) := (k,k) (k,1) */
	double c3 = SGN_Matrix_ddot(&k, ssm->H + i, &p, work->w, &SGN_Matrix_INT_ONE); /* c3 := H_i'  Txx  H_i */
	                                                              /*      (1,k) (k,k) (k,1) */
	ssm->R[i] = (c1 - 2.0 * c2 + c3) / (double) samples;
  }/* for i < p */

  if(unique_r){
	double v = 0.0;
	for(int i = 0; i < p; i++) v += ssm->R[i];
	v /= (double) p;
	SGN_Matrix_dcopy(&p, &v, &SGN_Matrix_INT_ZERO, ssm->R, &SGN_Matrix_INT_ONE);
  }/* if(unique_r) */
}
/* SSM_EM_update() ========================================================= */

/*!\brief Applys ridge estimation.
 *
 * Adds \a ridge to the diagonal elements of \a Txx and \a Tx1x1.
 *
 * Called by SSM_EM_estimate().
 *
 *\param k      Size of dimension.
 *\param ridge  Value of ridge estimation.
 *\param Txx    [INPUT] (\a k, \a k)-matrix. SSMSuffStat::Txx.
 *\param Tx1x1  [INPUT] (\a k, \a k)-matrix. SSMSuffStat::Tx1x1.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SSM_EM */
void SSM_EM_ridge(int k, double ridge, double * Txx, double * Tx1x1){
  int pos = 0;
  for(int i = 0; i < k; i++){
	Txx[pos] += ridge;
	Tx1x1[pos] += ridge;

	pos += k + 1;
  }/* for i < k */
}
/* SSM_EM_ridge() ========================================================== */

/* End of file.  Copyright (C) 2010  HGC, IMS, The University of Tokyo       */
