/*
  csml/CSMLObject.c: General CSML Object

  Copyright (C) 2009 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.
  
  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
 */

#include "CSMLObject.h"
//#include "../signlib.h"
#include "../lang/SGN_Obj.h"
	   
/*!\brief Generates a new CSMLObject instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
CSMLObject * CSMLObject_new(){
  CSMLObject * this = malloc(sizeof(CSMLObject));
  this->type = CSMLObject_TYPE_GENERAL;
  this->attr = NULL;
  this->x = 0.0;
  this->y = 0.0;
  this->width = 100;
  this->height = 30;
  this->fill_color = NULL;
  this->line_color = NULL;
  this->line_width = 1.0;
  this->depth = 0;
  this->visible = 1;
  this->CSO_type = CSOType_UNDEF;
  return this;
}
/* CSMLObject_new() ======================================================== */

/*\brief Deletes a CSMLObject instance.
 *
 *
 * NULL acceptable.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_delete(void * this){
  if(this == NULL) return;
  if(((CSMLObject *) this)->attr != NULL) SGN_KeyObjects_delete(((CSMLObject *) this)->attr);
  SGN_Color_delete(((CSMLObject *) this)->fill_color);
  SGN_Color_delete(((CSMLObject *) this)->line_color);
  free(this);
}
/* CSMLObject_delete() ===================================================== */

/*!\brief Generates a complete copy of the \ref CSMLObject instance.
 *
 * NULL acceptable.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
CSMLObject * CSMLObject_clone(const void * this){
  if(this == NULL) return NULL;

  CSMLObject * obj = CSMLObject_new();

  /* Perform shallow copy at first. */
  *obj = *((const CSMLObject *) this);
  
  /* Perform required deep copy next. */
  //obj->attr = SGN_KeyValues_clone(((const CSMLObject *) this)->attr);
  obj->attr = SGN_KeyObjects_clone(((const CSMLObject *) this)->attr);
  obj->fill_color = SGN_Color_clone(((const CSMLObject *) this)->fill_color);
  obj->line_color = SGN_Color_clone(((const CSMLObject *) this)->line_color);

  return obj;
}
/* CSMLObject_clone() ====================================================== */

/*!\brief Sets the type of this \ref CSMLObject instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_set_type(CSMLObject * this, int type){
  this->type = type;
}
/* CSMLObject_set_type() =================================================== */

/*!\brief Returns the type of the \ref CSMLObject instnace.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
int CSMLObject_get_type(const CSMLObject * this){
  return this->type;
}
/* CSMLObject_get_type() =================================================== */

/*!\brief Sets the CSO type of this CSMLObject instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_set_CSO_type(CSMLObject * this, CSOType type){
  this->CSO_type = type;
}
/* CSMLObject_set_CSO_type() =============================================== */

/*!\brief Returns the CSO type of this CSMLObject instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
CSOType CSMLObject_get_CSO_type(const CSMLObject * this){
  return this->CSO_type;
}
/* CSMLObject_get_CSO_type() =============================================== */

/*!\brief Returns the string expression of this CSMLObject instance CSO type.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
const char * CSMLObject_get_CSO_type_name(const CSMLObject * this){
  return CSOType_name(this->CSO_type);
}
/* CSMLObject_get_CSO_type_name() ========================================== */

/*!\brief Sets the position of the \ref CSMLObject.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_set_position(CSMLObject * this, double x, double y){
  this->x = x;
  this->y = y;
}
/* CSMLObject_set_position() =============================================== */

/*!\brief Sets the relative position of the \ref CSMLObject to the
 *  current one.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_move_position(CSMLObject * this, double x, double y){
  this->x += x;
  this->y += y;
}
/* CSMLObject_move_position() ============================================== */

/*!\brief Gets the position of X.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
double CSMLObject_get_position_x(const CSMLObject * this){
  return this->x;
}
/* CSMLObject_get_position_x() ============================================= */

/*\brief Gets the position of Y.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
double CSMLObject_get_position_y(const CSMLObject * this){
  return this->y;
}
/* CSMLObject_get_position_y() ============================================= */

/*!\brief Sets the height of the \ref CSMLObject instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_set_height(CSMLObject * this, int height){
  this->height = height;
}
/* CSMLObject_set_height() ================================================= */

/*!\brief Returns the height of the \ref CSMLObject.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
double CSMLObject_get_height(const CSMLObject * this){
  return this->height;
}
/* CSMLObject_get_height() ================================================= */

void CSMLObject_set_size_move(CSMLObject * this, double width, double height){
  double x = (this->width - width) / 2.0;
  double y = (this->height - height) / 2.0;

  this->width = width;
  this->height = height;
  this->x += x;
  this->y += y;
}
/* CSMLObject_set_size_move() ============================================== */

/*!\brief Sets the width of the \ref CSMLObject instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_set_width(CSMLObject * this, int width){
  this->width = width;
}
/* CSMLObject_set_width() ================================================== */

/*!\brief Returns the width of the \ref CSMLObject.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
double CSMLObject_get_width(const CSMLObject * this){
  return this->width;
}
/* CSMLObject_get_width() ================================================== */

/*!\brief Sets the fill color by the \ref SGN_Color instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_set_fill_color(CSMLObject * this, const SGN_Color * color){
  if(this->fill_color == NULL) this->fill_color = SGN_Color_new();
  SGN_Color_set(this->fill_color, color);
}
/* CSMLObject_set_fill_color() ============================================= */

/*!\brief Sets the fill color in RGB.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_set_fill_color_RGB(CSMLObject * this, int r, int g, int b, int a){
  if(this->fill_color == NULL) this->fill_color = SGN_Color_new();
  SGN_Color_set_RGB(this->fill_color, r, g, b, a);
}
/* CSMLObject_set_fill_color() ============================================= */

/*!\brief Sets the line color by the \ref SGN_Color instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_set_line_color(CSMLObject * this, const SGN_Color * color){
  if(this->line_color == NULL) this->line_color = SGN_Color_new();
  SGN_Color_set(this->line_color, color);
}
/* CSMLObject_set_line_color() ============================================= */

/*!\brief Sets the line color in RGB.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_set_line_color_RGB(CSMLObject * this, int r, int g, int b, int a){
  if(this->line_color == NULL) this->line_color = SGN_Color_new();
  SGN_Color_set_RGB(this->line_color, r, g, b, a);
}
/* CSMLObject_set_line_color() ============================================= */

/*!\brief Sets the line width.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_set_line_width(CSMLObject * this, double width){
  this->line_width = width;
}
/* CSMLObject_set_line_width() ============================================= */

/*!\brief Returns the line width.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
double CSMLObject_get_line_width(const CSMLObject * this){
  return this->line_width;
}
/* CSMLObject_get_line_width() ============================================= */

void CSMLObject_set_depth(CSMLObject * this, int depth){
  if(depth > SIGN_MAX_DEPTH) depth = SIGN_MAX_DEPTH;
  this->depth = depth;
}
/* CSMLObject_set_depth() ================================================== */

int CSMLObject_get_depth(const CSMLObject * this){
  return this->depth;
}
/* CSMLObject_get_depth() ================================================== */

/*!\brief Sets the visibility of the object.
 *
 *\param visible zero or non-zero specifying this to be visible or
 *            invisible respectively.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_set_visible(CSMLObject * this, int visible){
  this->visible = visible;
}
/* CSMLObject_set_visible() ================================================ */

/*!\brief Returns the visibility.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
int CSMLObject_get_visible(const CSMLObject * this){
  return this->visible;
}
/* CSMLObject_get_visible() ================================================ */

/*****************************************************************************/
/*                                                                           */
/*  P r o p e r t y  R e l a t e d   F u n c t i o n s                       */
/*                                                                           */
/*****************************************************************************/

/*!\brief Returns the number of properties.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
size_t CSMLObject_num_properties(const CSMLObject * this){
  if(this->attr == NULL){
	return 0;
  } else {
	return SGN_KeyObjects_size(this->attr);
  }
}
/* CSMLObject_num_properties() ============================================= */

void CSMLObject_clear_properties(CSMLObject * this){
  if(this->attr == NULL) return;

  SGN_KeyObjects_clear(this->attr);
}
/* CSMLObject_clear_properties() =========================================== */

/*!\brief Adds a property.
 *
 * The specfied Object is managed by \a this CSMLObject 
 * (i.e. memory handling).
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_add_property(CSMLObject * this, const char * key, SGN_Object * obj){
  if(this->attr == NULL) this->attr = SGN_KeyObjects_new();
  SGN_KeyObjects_add(this->attr, key, obj);
}
/* CSMLObject_add_property() =============================================== */

/*!\brief Returns the property associated with the specified key.
 *
 *\param this
 *\param key
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
SGN_Object * CSMLObject_get_property(const CSMLObject * this, const char * key){
  if(this->attr == NULL) return NULL;
#ifdef SIGN_DEBUG
  fprintf(SIGN_ERR, "key=%s\n", key);
  //CSMLObject_check_validity(this);

  SGN_Object * obj = SGN_KeyObjects_get(this->attr, key);
  
  if((int) SGN_Object_type(obj) > (int) SGN_Type_ERROR){
	fprintf(SIGN_ERR, "WARNING: CSMLObject_get_property(): Illigal type number detected: %d\n",
			(int) this->type);
	abort();
  }
  return obj;
#else
  return SGN_KeyObjects_get(this->attr, key);
#endif
}
/* CSMLObject_get_property() =============================================== */

/*!\brief Returns the property key.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
const char * CSMLObject_get_property_key(const CSMLObject * this, 
										 size_t index){
  if(this->attr == NULL) return NULL;
  return SGN_KeyObjects_get_key(this->attr, index);
}
/* CSMLObject_get_property_key() =========================================== */

SGN_Object * CSMLObject_get_property_value(const CSMLObject * this,
										   size_t index){
  if(this->attr == NULL) return NULL;
  return SGN_KeyObjects_get_value(this->attr, index);
}
/* CSMLObject_get_property_value() ========================================= */

SGN_Type CSMLObject_get_property_type(const CSMLObject * this, size_t index){
  return SGN_KeyObjects_get_type(this->attr, index);
}

/*
void CSMLObject_add_property_typeX(CSMLObject * this, const char * key,
								  SGN_Type type, void * value){
  if(this->attr == NULL) this->attr = SGN_KeyObjects_new();
  SGN_KeyObjects_add_type(this->attr, value, type);
}
*/
/* CSMLObject_add_property_type() ========================================== */

void CSMLObject_add_property_int(CSMLObject * this, const char * key, int value){
  if(this->attr == NULL) this->attr = SGN_KeyObjects_new();
  SGN_KeyObjects_add_int(this->attr, key, value);
}
/* CSMLObject_add_property_int() =========================================== */
							 

void CSMLObject_set_property_int(CSMLObject * this, const char * key,
								 int value){
  if(this->attr == NULL) this->attr = SGN_KeyObjects_new();
  SGN_KeyObjects_set_int(this->attr, key, value);
}
/* CSMLObject_set_property_int() =========================================== */

int CSMLObject_get_property_int(const CSMLObject * this, const char * key, int def){
  if(this->attr == NULL){
	return def;
  } else {
	return SGN_KeyObjects_get_int(this->attr, key, def);
  }
}
/* CSMLObject_get_property_int() =========================================== */

int CSMLObject_get_property_int_at(const CSMLObject * this, size_t index, int def){
  return SGN_KeyObjects_get_int_at(this->attr, index, def);
}

void CSMLObject_add_property_double(CSMLObject * this, const char * key, double value){
  if(this->attr == NULL) this->attr = SGN_KeyObjects_new();
  SGN_KeyObjects_add_double(this->attr, key, value);
}

void CSMLObject_set_property_double(CSMLObject * this, const char * key,
									double value){
  if(this->attr == NULL) this->attr = SGN_KeyObjects_new();
  SGN_KeyObjects_set_double(this->attr, key, value);
}
/* CSMLObject_set_property_double() ======================================== */

void CSMLObject_add_property_string(CSMLObject * this, const char * key,
									const char * string){
  if(this->attr == NULL) this->attr = SGN_KeyObjects_new();
  SGN_KeyObjects_add_string(this->attr, key, string);
}
/* CSMLObject_add_property_string() ======================================== */

/*
char * CSMLObject_get_property_string_(const CSMLObject * this, const char * key){
  if(this->attr == NULL) return NULL;
  SGN_KeyObjects_get_string(this->attr, key);
}
*/
/* CSMLObject_get_property_string() ======================================== */

void CSMLObject_set_properties(CSMLObject * this, SGN_KeyObjects * ko){
  if(this->attr != NULL) SGN_KeyObjects_delete(this->attr);
  this->attr = ko;
}
/* CSMLObject_set_properties() ============================================= */

size_t CSMLObject_find_property(CSMLObject * this, const char * key){
  if(this->attr == NULL) return CSMLObject_ERROR_VALUE;
  return SGN_KeyObjects_find(this->attr, key);
}
/* CSMLObject_find_property() ============================================== */

void CSMLObject_append_properties_KeyValues(CSMLObject * this, const SGN_KeyValues * kv){
  SGN_KeyObjects_append_KeyValues(this->attr, kv);
}

void CSMLObject_remove_property_at(CSMLObject * this, size_t index){
  if(this->attr != NULL){
	SGN_KeyObjects_remove_at(this->attr, index);
  }
}
/* CSMLObject_remvoe_property_at() ========================================= */

/* End of property functions.                                                */
/*****************************************************************************/

/*
 */
void CSMLObject_get_line_color_expr(const CSMLObject * this, char * buff){
  if(this->line_color == NULL){
	switch(this->type){
	case CSMLObject_TYPE_EDGE:
	  sprintf(buff, "%s", CSMLObject_DEFAULT_EDGE_LINE_COLOR_EXPR);
	  break;
	case CSMLObject_TYPE_NODE:
	  sprintf(buff, "%s", SGN_Node_DEFAULT_COLOR_EXPR);
	  break;
	default:
	  sprintf(buff, "%s", CSMLObject_DEFAULT_LINE_COLOR_EXPR);
	  break;
	}
  } else {
	SGN_Color_sprint(this->line_color, buff);
  }
}
/* CSMLObject_get_line_color_expr() ======================================== */

/*
 */
void CSMLObject_get_fill_color_expr(const CSMLObject * this, char * buff){
  if(this->fill_color == NULL){
	switch(this->type){
	case CSMLObject_TYPE_NODE:
	  sprintf(buff, "%s", SGN_Node_DEFAULT_COLOR_EXPR);
	  break;
	default:
	  sprintf(buff, "%s", CSMLObject_DEFAULT_FILL_COLOR_EXPR);
	  break;
	}
  } else {
	SGN_Color_sprint(this->fill_color, buff);
  }
}
/* CSMLObject_get_fill_color_expr() ======================================== */

/*!\brief
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_copy_properties(CSMLObject * this, const CSMLObject * obj){
  //fprintf(SIGN_ERR, "CSMLObject_copy_properties(): called\n");
  if(this->attr != NULL){
	SGN_KeyObjects_clear(this->attr);
  }
  
  if(obj->attr == NULL){
	/* Do nothing. Nothing to copy */
	return;
  } else if(this->attr == NULL){
	this->attr = SGN_KeyObjects_new();
  }
  SGN_KeyObjects_append_KeyObjects(this->attr, obj->attr);
}
/* CSMLObject_copy_properties() ============================================ */

void CSMLObject_fprint_properties(const CSMLObject * this, FILE * fp){
  if(this->attr != NULL){
	SGN_KeyObjects_fprint(this->attr, fp);
  }
}
/* CSMLObject_fprint_properties() ========================================== */

/*!\brief Prints the attribute value bound to the given key to the stream.
 *
 *\param this   pointer to the \ref CSMLObject instance.
 *\param key    key of the value.
 *\param fp     output stream.
 *\return 1 if succeeded, or 0 if failed.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
int CSMLObject_fprint_property_value(const CSMLObject * this, const char * key, FILE * fp){
  size_t pos = SGN_KeyObjects_find(this->attr, key);
  if(pos != SGN_KeyObjects_ERROR_VALUE){
	const SGN_Object * obj = SGN_KeyObjects_get_const_value(this->attr, pos);
	SGN_Object_fprint(obj, fp);
	return 1;
	
  } else {
	return 0;
  }
}
/* CSMLObject_fprint_property_value() ====================================== */

/*!\brief Puts the \ref CSMLObject in CSML.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof CSMLObject */
void CSMLObject_put_CSML(const CSMLObject * this, FILE * fp){

  const char * text = "";

  if(this->attr != NULL) {
	//const void * obj = SGN_KeyValues_get_string(this->attr, "text");
	//if(obj != NULL) text = (const char *) obj;
	SGN_Object * obj = SGN_KeyObjects_get(this->attr, "text");
	if(obj != NULL && SGN_Object_type(obj) == SGN_Type_STRING){
	  text = (char *) SGN_Object_ptr(obj);
	} else {
	  fprintf(SIGN_ERR, "NULL!\n");
	}
  }
  
  /* <csml:fact id="XX" name="XX" type="cso30:c:FactViewElement"> */
  
  //<csml:logProperty/>
  fprintf(fp, "<csml:viewProperty>\n");
  fprintf(fp, "<csml:position position=\"rotation:0.0\" positionID=\"default\" x=\"%f\" y=\"%f\"/>\n",
		  this->x, this->y);
  fprintf(fp, "<csml:shape depth=\"0\" shapeID=\"default\" visible=\"true\">\n");
  fprintf(fp, "<csml:toolSpecificGraphics>\n");
  fprintf(fp, "<fig:figure xmlns:fig=\"http://www.csml.org/csml/cigraphics\">\n");
  fprintf(fp, "<fig:textArea arcRatio=\"5\" depth=\"0\"");
  if(this->fill_color == NULL){
	fprintf(fp, " fillColor=\"%s\"", CSMLObject_DEFAULT_FILL_COLOR_EXPR);
  } else {
	fprintf(fp, " fillColor=\"");
	SGN_Color_fprint(this->fill_color, fp);
	fprintf(fp, "\"");
  }
  fprintf(fp, " fontName=\"Lucida Grande\" fontSize=\"16\" fontStyle=\"plain\" "
		  "outlineColor=\"0 0 255 255\" outlineDash=\"\" outlineWidth=\"2.0\" ");
  fprintf(fp, "size=\"%f %f\" ", this->width, this->height);
  fprintf(fp, "spacing=\"0.0\" textColor=\"0 0 0 255\" "
		  "visible=\"true\">%s</fig:textArea>\n", text);
  fprintf(fp, "</fig:figure>\n");
  fprintf(fp, "</csml:toolSpecificGraphics>\n");
  fprintf(fp, "</csml:shape>\n");
  fprintf(fp, "</csml:viewProperty>\n");
  fprintf(fp, "<csml:animationProperty/>\n");
  fprintf(fp, "<csml:biologicalProperty/>\n");
  fprintf(fp, "<csml:comments/>\n");
  //</csml:fact>
}
/* CSMLObject_put_CSML() =================================================== */

void CSMLObject_check_validity(const CSMLObject * this){
  if(this->attr == NULL){
	fprintf(SIGN_ERR, "CSMLObject_check_validity(): NULL\n");
  } else {
	SGN_KeyObjects_check_validity(this->attr);
	fprintf(SIGN_ERR, "CSMLObject_check_validity(): Validated\n");
  }
}

/* End of file.  Copyright (C) 2009 SiGN Project Members. */
