/*
  csml/CSMLObject.h: General CSML Object

  Copyright (C) 2010 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.
  
  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
 */

#ifndef __CSML_OBJECT_H
#define __CSML_OBJECT_H

//#include "../util/SGN_KeyValues.h"
#include "../util/SGN_KeyObjects.h"
#include "../util/SGN_Color.h"
#include "CSOType.h"

#define CSMLObject_TYPE_GENERAL     0
#define CSMLObject_TYPE_TEXT_AREA   1
#define CSMLObject_TYPE_NODE        2
#define CSMLObject_TYPE_EDGE        3
#define CSMLObject_DEFAULT_LINE_COLOR_EXPR      "0 0 0 255"  /* black */
#define CSMLObject_DEFAULT_EDGE_LINE_COLOR_EXPR "0 0 0 255"  /* black */

#define CSMLObject_DEFAULT_FILL_COLOR_EXPR      "0 0 0 0"
#define CSMLObject_DEFAULT_NODE_FILL_COLOR_EXPR "0 0 255 255" /* blue */

#define CSMLObject_ERROR_VALUE SGN_KeyObjects_ERROR_VALUE


/*!\brief CSMLObject.
 *
 * \section textarea TextArea
 * Available properties.
 * <DL>
 * <DT><TT>text</TT> (String)</DT>
 * </DL>
 *
 */
typedef struct {
#ifdef DOXY
private:
#endif
  int type;
  SGN_KeyObjects * attr;
  double x;
  double y;
  double width;
  double height;
  SGN_Color * fill_color;
  SGN_Color * line_color;
  double line_width;
  int depth;
  int visible;
  CSOType CSO_type;
} CSMLObject;

CSMLObject * CSMLObject_new();
void         CSMLObject_delete(void * this);
CSMLObject * CSMLObject_clone(const void * this);
void         CSMLObject_copy(CSMLObject * this, const CSMLObject * obj);

void         CSMLObject_set_type(CSMLObject * this, int type);
int          CSMLObject_get_type(const CSMLObject * this);
void         CSMLObject_set_CSO_type(CSMLObject * this, CSOType type);
CSOType      CSMLObject_get_CSO_type(const CSMLObject * this);
const char * CSMLObject_get_CSO_type_name(const CSMLObject * this);

void         CSMLObject_set_position(CSMLObject * this, double x, double y);
void         CSMLObject_move_position(CSMLObject * this, double x, double y);
double       CSMLObject_get_position_x(const CSMLObject * this);
double       CSMLObject_get_position_y(const CSMLObject * this);
void         CSMLObject_set_height(CSMLObject * this, int height);
double       CSMLObject_get_height(const CSMLObject * this);
void         CSMLObject_set_width(CSMLObject * this, int width);
double       CSMLObject_get_width(const CSMLObject * this);
void         CSMLObject_set_size_move(CSMLObject * this, double width, double height);


void         CSMLObject_set_fill_color(CSMLObject * this, const SGN_Color * color);
void         CSMLObject_set_fill_color_RGB(CSMLObject * this, int r, int g, int b, int a);
void         CSMLObject_get_fill_color_expr(const CSMLObject * this, char * buff);
void         CSMLObject_set_line_color(CSMLObject * this, const SGN_Color * color);
void         CSMLObject_set_line_color_RGB(CSMLObject * this, int r, int g, int b, int a);
void         CSMLObject_get_line_color_expr(const CSMLObject * this, char * buff);

void         CSMLObject_set_line_width(CSMLObject * this, double width);
double       CSMLObject_get_line_width(const CSMLObject * this);
void         CSMLObject_set_depth(CSMLObject * this, int depth);
int          CSMLObject_get_depth(const CSMLObject * this);
void         CSMLObject_set_visible(CSMLObject * this, int visible);
int          CSMLObject_get_visible(const CSMLObject * this);


size_t       CSMLObject_num_properties(const CSMLObject * this);
const char * CSMLObject_get_property_key(const CSMLObject * this, size_t index);
SGN_Type     CSMLObject_get_property_type(const CSMLObject * this, size_t index);
void         CSMLObject_clear_properties(CSMLObject * this);
void         CSMLObject_add_property_type(CSMLObject * this, const char * key,
										  SGN_Type type, void * value);
void         CSMLObject_add_property(CSMLObject * this, const char * key, SGN_Object * obj);
SGN_Object * CSMLObject_get_property(const CSMLObject * this, const char * key);
SGN_Object * CSMLObject_get_property_value(const CSMLObject * this, size_t index);
void         CSMLObject_add_property_int(CSMLObject * this, const char * key, int value);
void         CSMLObject_set_property_int(CSMLObject * this, const char * key, int value);
int          CSMLObject_get_property_int(const CSMLObject * this, const char * key, int def);
int          CSMLObject_get_property_int_at(const CSMLObject * this, size_t index, int def);

void         CSMLObject_add_property_double(CSMLObject * this, const char * key, double value);
void         CSMLObject_set_property_double(CSMLObject * this, const char * key, double value);

void         CSMLObject_add_property_string(CSMLObject * this, const char * key, const char * string);
char *       CSMLObject_get_property_string(const CSMLObject * this, const char * key);
void         CSMLObject_set_properties(CSMLObject * this, SGN_KeyObjects * ko);
void         CSMLObject_copy_properties(CSMLObject * this, const CSMLObject * obj);
void         CSMLObject_fprint_properties(const CSMLObject * this, FILE * fp);
int          CSMLObject_fprint_property_value(const CSMLObject * this, const char * key, FILE * fp);
size_t       CSMLObject_find_property(CSMLObject * this, const char * key);

void         CSMLObject_append_properties_KeyValues(CSMLObject * this, const SGN_KeyValues * kv);

void         CSMLObject_remove_property_at(CSMLObject * this, size_t index);

void         CSMLObject_put_CSML(const CSMLObject * this, FILE * fp);
void         CSMLObject_check_validity(const CSMLObject * this);

#endif /* __CSML_OBJECT_H */

/* End of file.  Copyright (C) 2010 HGC, IMS, The University of Tokyo */
