/*
  dispatch.c : SiGN-SSM - Gene Network Estimation with State Sapce Model

  Copyright (C) 2010
    Yoshinori Tamada, Rui Yamaguchi, Seiya Imoto, Osamu Hirose, Ryo Yoshida,
	Masao Nagasaki and Satoru Miyano
      Laboratory of DNA Information Analysis &
	  Laboratory of Sequence Analysis, Human Genome Center,
      The Institute of Medical Science, The Unversity of Tokyo

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  
  Visit <http://sign.hgc.jp/signssm/> for more information.

  Written by Yoshinori Tamada <tamada at ims.u-tokyo.ac.jp>
  Originally developed by Osamu Hirose, Ryo Yoshida, Seiya Imoto, Rui Yamaguchi,
                          Tomoyuki Higuchi, and Satoru Miyano.
*/

#include "signcom.h"

#include "dispatch.h"
#include "util/SGN_IntQueue.h"
#include "util/SGN_IntArray.h"
#include "util/SGN_Array.h"

#include <string.h>
#include <strings.h>

/*
  Protocol Memo:

  SERVER              CLIENT
    <--------------------
    'score' (double)          : Request & the previous result
	
  SERVER              CLIENT
    -------------------->
    (all integers)
	
    '0',     setID2, dim  : Dispatch setID2
    setID1,  setID2, dim  : Output setID1 & dispatch setID2 (dim=-1, finished)
    -setID1, setID2, dim  : Dispose setID1 & dispatch setID2 (dim=-1, finished)
    '0',     '0',    '-1' : Request again
	'0',     '0',    '-2' : Wait for the final call.

 */

/*
  Returns the number of positive values in 'queue'.
  That is, the number of setIDs required to output for this rank's queue.
 */
int count_outputs(const SGN_IntQueue * queue){
  int num = 0;
  for(int i = 0; i < SGN_IntQueue_size(queue); i++){
	if(SGN_IntQueue_get(queue, i) > 0) num++;
  }
  return num;
}
/* count_outputs() ========================================================= */

/*
  If the assignment (dispatch) is over-flowed, then 0 is set to 'set'.
  'rank' starts from 1.
 */
void calc_first_dispatch(int in_iter, int in_set, int in_dim, int rank,
						 int * dim, int * set_id){
  /* Jobs: interation x set x dim */

  rank--;

  rank /= in_iter;
  *set_id = rank;

  *dim = rank / in_set;

  if(*dim >= in_dim){
	*set_id = 0; /* Assignment over flowed. */
  } else {
	*set_id += 1;
  }
}
/* calc_first_dispatch() =================================================== */

#ifdef USE_MPI
void dispatch_jobs(Args * args){
  fprintf(SIGN_ERR, "THIS IS THE ROOT PROCESS [SERVER]\n");
  fflush(SIGN_ERR);

  int total = args->sets * args->iteration * args->len_dims;

  fprintf(SIGN_ERR, 
		  "In total, %d (set) x %d (iteration) x %d (dimension) = %d jobs will be dispatched.\n",
		  args->sets, args->iteration, args->len_dims, total);

  int finished_procs = 0;
  double likelihood = 0;
  MPI_Status status;
  int total_sets = args->len_dims * args->sets;
  int rank_set_id[args->MPI_size]; /* set ID assigned to each rank */
  double best_scores[total_sets + 1]; /* Best score for each set ID */
  int best_ranks[total_sets + 1];
  int finished_iter[total_sets + 1];
  int iteration = 0;
  int set = 0;
  int dim_idx = 0;
  int set_id = 1; /* 1-origin */
  int msg[3];
  SGN_IntQueue * queue = SGN_IntQueue_new_ar(args->MPI_size);
  SGN_IntArray * set_ranks[total_sets + 1];
  SGN_Array * stack = SGN_Array_new(); /* for recycling IntArray of 'set_ranks' */
  int count = args->MPI_size - 1;
  if(count > total) count = total;

  fprintf(SIGN_ERR, "Total sets = %d\n", total_sets);
  fflush(SIGN_ERR);

  /* Note: set ID begins with 1.  Suppress subtracting 1 from the index,
	 arrays indexed by the set ID have 1-larger size and the first element
	 has no data. */
  
  /* Initialize variables */
  memset(finished_iter + 1, 0, sizeof(int) * total_sets);
  memset(set_ranks + 1, 0, sizeof(SGN_IntArray *) * total_sets);
  for(int i = 1; i <= total_sets; i++) best_scores[i] = - DBL_MAX;
  proc_first_dispatch(args->MPI_size, rank_set_id,
					  args->iteration, args->sets, args->len_dims,
					  &iteration, &set, &dim_idx, &set_id);
  for(int i = 1; i < args->MPI_size; i++){
	int id = rank_set_id[i];
	if(id == 0) continue;
	if(set_ranks[id] == NULL) set_ranks[id] = SGN_IntArray_new();
	SGN_IntArray_add_nr(set_ranks[id], i);
  }
  if(total < args->MPI_size - 1){
	finished_procs = args->MPI_size - 1 - total;
  }

  fprintf(SIGN_ERR, "After the first dispatch: iter=%d, set=%d, dim_idx=%d, set_id=%d.\n",
		  iteration, set, dim_idx, set_id);
  fflush(SIGN_ERR);
  
  while(finished_procs < args->MPI_size - 1){
	/* Receives a job request as the likelihood */
	fprintf(SIGN_ERR, "Waiting a job request...\n"); fflush(SIGN_ERR);
	MPI_Recv(&likelihood, 1, MPI_DOUBLE, MPI_ANY_SOURCE, 2, MPI_COMM_WORLD, &status);
	int recv_id = rank_set_id[status.MPI_SOURCE]; /* set_id assigned to the receiving proc. */

	fprintf(SIGN_ERR, "Request received: rank=%d, likelihood=%f (set ID=%d)\n",
			status.MPI_SOURCE, likelihood, recv_id);
	fflush(SIGN_ERR);

	if(recv_id != 0 && !isnan(likelihood) && best_scores[recv_id] < likelihood){
	  fprintf(SIGN_ERR, "  Best likelihood updated for set ID=%d: %f count: %d\n", recv_id, likelihood, count);
	  fflush(SIGN_ERR);
	  best_scores[recv_id] = likelihood;
	  best_ranks[recv_id] = status.MPI_SOURCE;
	}
	if(recv_id != 0) finished_iter[recv_id]++;

	if(recv_id != 0 && finished_iter[recv_id] == args->iteration){
	  fprintf(SIGN_ERR, "------------------------\n");
	  fprintf(SIGN_ERR, " Set finished. setID=%d.\n", recv_id);
	  fprintf(SIGN_ERR, "------------------------\n");
	  fflush(SIGN_ERR);
	  fprintf(SIGN_ERR, "  |set_ranks[recv_id]|=%zu\n", 
			  SGN_IntArray_size(set_ranks[recv_id]));
	  fflush(SIGN_ERR);
	  
	  /* Enqueue the actions for the finished set. */
	  for(int i = 0; i < SGN_IntArray_size(set_ranks[recv_id]); i++){
		int rank = SGN_IntArray_get(set_ranks[recv_id], i);
		if(rank == best_ranks[recv_id]) {
		  SGN_IntQueue_enqueue(queue + rank, recv_id); /* Output */
		} else {
		  SGN_IntQueue_enqueue(queue + rank, - recv_id); /* Dispose */
		}
	  }/* for i < |set_ranks[recv_id]| */
	  SGN_IntArray_clear(set_ranks[recv_id]);
	  SGN_Array_add(stack, set_ranks[recv_id]); /* Recycling the IntArray instance. */
	  set_ranks[recv_id] = NULL;

	  fprintf(SIGN_ERR, "  Best likelihood: %f @ rank: %d  count: %d\n", 
			  best_scores[recv_id], best_ranks[recv_id], count);
	  fflush(SIGN_ERR);
	}/* if(recv_id != 0 && finished_iter[recv_id] == args->iteration) */

	/* Action order */
	if(SGN_IntQueue_is_empty(queue + status.MPI_SOURCE)){
	  msg[0] = 0; /* Nothing to order */
	} else {
	  SGN_IntQueue_dequeue(queue + status.MPI_SOURCE, msg + 0);
	}

	/* Dispatched job */
	msg[1] = set_id;  /* set_id == 0 if already finished */

	/* dim */
	if(set_id == 0){ /* if finished */
	  if(SGN_IntQueue_is_empty(queue + status.MPI_SOURCE)){
		finished_procs++;
		msg[2] = -2; /* Wait for the final call */
	  } else {
		fprintf(SIGN_ERR, "Queue not empty for rank %d: ", status.MPI_SOURCE);
		SGN_IntQueue_fprint(queue + status.MPI_SOURCE, SIGN_ERR);
		fprintf(SIGN_ERR, "\n"); fflush(SIGN_ERR);
		msg[2] = -1; /* Request again */
	  }
	} else {
	  msg[2] = args->dims[dim_idx];
	}
	MPI_Send(&msg, 3, MPI_INT, status.MPI_SOURCE, 1, MPI_COMM_WORLD);
	fprintf(SIGN_ERR, "Job Dispatched: rank=%d, setID1=%d, setID2=%d, dim=%d\n",
			status.MPI_SOURCE, msg[0], msg[1], msg[2]);
	fprintf(SIGN_ERR, "  iteration=%d, set=%d\n", iteration, set);
	if(set_id != 0){
	  fprintf(SIGN_ERR, "  %d/%d\n", ++count, total);
	}
	fflush(SIGN_ERR);
	  
	rank_set_id[status.MPI_SOURCE] = msg[1]; /* Dispatched set ID */

	/* Update variables */
	if(set_id != 0){
	  if(set_ranks[set_id] == NULL){
		if(SGN_Array_size(stack) != 0){
		  set_ranks[set_id] = (SGN_IntArray *) SGN_Array_pop(stack);
		} else {
		  set_ranks[set_id] = SGN_IntArray_new();
		}
	  }
	  SGN_IntArray_add_nr(set_ranks[set_id], status.MPI_SOURCE);
	  
	  inc_variables(args->iteration, args->sets, args->len_dims,
					&iteration, &set, &dim_idx, &set_id);
	}
  }/* while() */
  fprintf(SIGN_ERR, "All jobs finished.  Sending the final order.\n"); fflush(SIGN_ERR);

  for(int i = 1; i < args->MPI_size; i++){
	int num = count_outputs(queue + i);
	fprintf(SIGN_ERR, "For Client:%d, num = %d\n", i, num); fflush(SIGN_ERR);
	MPI_Send(&num, 1, MPI_INT, i, 2, MPI_COMM_WORLD);
	if(num != 0){
	  int ids[num];
	  int j = 0, v;
	  while(SGN_IntQueue_dequeue(queue + i, &v) == 0) if(v > 0) ids[j++] = v;
	  MPI_Send(ids, num, MPI_INT, i, 3, MPI_COMM_WORLD);
	}
  }
}
/* dispatch_jobs() ========================================================= */
#endif /* USE_MPI */

/*
 *\param in_set  Number of sets for a single size of dimensions.
 */
void inc_variables(int in_iter, int in_set, int in_dim,
				   int * iteration, int * set, int * dim, int * set_id){

  if(*set_id != 0){
	(*iteration)++;

	if(*iteration == in_iter){
	  *iteration = 0;
	  (*set)++;
	  (*set_id)++;

	  if(*set == in_set){
		*set = 0;
		(*dim)++;

		if(*dim == in_dim){
		  *set_id = 0;
		}
	  }
	}
  }
}
/* inc_variables() ========================================================= */

void proc_first_dispatch(int MPI_size, int * rank_set_id,
						 int in_iter, int in_set, int in_dim,
						 int * iteration, int * set, int * dim, int * set_id){
  for(int i = 1; i < MPI_size; i++){
	fprintf(SIGN_ERR, "proc_first_dispatch: rank=%d, set_id=%d\n", i, *set_id);
	fflush(SIGN_ERR);
	rank_set_id[i] = *set_id;
	inc_variables(in_iter, in_set, in_dim, iteration, set, dim, set_id);
  }/* for i < MPI_size */
}
/* proc_first_dispatch() =================================================== */

/* End of file.  Copyright (C) 2010  HGC, IMS, The University of Tokyo       */
