/*
  execute.c : SiGN-SSM - Gene Network Estimation with State Sapce Model

  Copyright (C) 2010
    Yoshinori Tamada, Rui Yamaguchi, Seiya Imoto, Osamu Hirose, Ryo Yoshida,
	Masao Nagasaki and Satoru Miyano
      Laboratory of DNA Information Analysis &
	  Laboratory of Sequence Analysis, Human Genome Center,
      The Institute of Medical Science, The Unversity of Tokyo

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  
  Visit <http://sign.hgc.jp/signssm/> for more information.

  Written by Yoshinori Tamada <tamada at ims.u-tokyo.ac.jp>
  Originally developed by Osamu Hirose, Ryo Yoshida, Seiya Imoto, Rui Yamaguchi,
                          Tomoyuki Higuchi, and Satoru Miyano.
*/

#include "signcom.h"

#include "execute.h"
#include "dispatch.h"
#include "SSMResult.h"
#include "Kalman.h"
#include "util/SGN_Array.h"

#include <limits.h>

int int_ar_max(int * ar, int size){
  int max = - INT_MAX;
  for(int i = 0; i < size; i++){
	if(max < ar[i]) max = ar[i];
  }
  return max;
}
/* int_ar_max() ============================================================ */

void update(int set_id, double * best_scores, 
			SSMResult * result, SSMResult * best_results, 
			SGN_Array * stack, int max_k){
  fprintf(SIGN_ERR, "Best likelihood updated for set id = %d.\n", set_id); 
  fflush(SIGN_ERR);
  best_scores[set_id] = result->likelihood;

  SSM * ssm;
  if(best_results[set_id].ssm == NULL){
	if(SGN_Array_size(stack) == 0){
	  ssm = SSM_new(result->p, max_k);
	} else {
	  ssm = (SSM *) SGN_Array_pop(stack);
	  if(ssm == NULL){
		fprintf(SIGN_ERR, "FATAL ERROR: execute.c:update(): ssm NULL\n");
		fflush(SIGN_ERR);
		exit(1);
	  }
	}
  } else {
	ssm = best_results[set_id].ssm;
  }

  /* Copy SSM instance */
  if(ssm == NULL){
	fprintf(SIGN_ERR, "FATAL ERROR: execute.c:update(): ssm NULL\n");
	fflush(SIGN_ERR);
	exit(1);
  }
  SSM_copy(ssm, result->ssm, result->p, result->dim);

  /* Shallow copy first */
  best_results[set_id] = *result;
  /* Set the SSM instance */
  best_results[set_id].ssm = ssm;
}
/* update() ================================================================ */

#ifdef USE_MPI
void execute_jobs(Args * args, SSMData * Y){
  fprintf(SIGN_ERR, "****************************\n");
  fprintf(SIGN_ERR, "**   MPI Execution Mode   **\n");
  fprintf(SIGN_ERR, "****************************\n");
  fprintf(SIGN_ERR, "THIS IS THE EXECUTION PROCESS [CLIENT:%d]\n", args->MPI_rank);
  fflush(SIGN_ERR);

  /* Prepare the execution */
  fprintf(SIGN_ERR, "--- INPUT DATA ---------\n");
  fprintf(SIGN_ERR, " observed samples = %d\n", Y->samples);
  fprintf(SIGN_ERR, " objects = %d\n", Y->p);
  fprintf(SIGN_ERR, " time points = %d\n", Y->N);
  fprintf(SIGN_ERR, " replicates = %d\n", Y->replicates);
  fprintf(SIGN_ERR, "------------------------\n");
  //SSMData_dump(Y, SIGN_ERR);

  int max_dim = int_ar_max(args->dims, args->len_dims);
  fprintf(SIGN_ERR, "Max dimension = %d\n", max_dim); fflush(SIGN_ERR);
  int prev_dim = -1;

  /* Prepare working instances */
  SSM * ssm = SSM_new(Y->p, max_dim);
  SSMState * X = SSMState_new(max_dim, Y->N);
  SSMSuffStat * T = SSMSuffStat_new(Y->p, max_dim);
  SSMSuffStat * sumT = SSMSuffStat_new(Y->p, max_dim);
  SSMWork * work = SSMWork_new(Y->p, max_dim, Y->samples, args->em->constrain_F);

  int total_sets = args->len_dims * args->sets;
  double likelihood = 0.0;
  MPI_Status status;
  int count = 0;
  SGN_Timer * timer = SGN_Timer_new();
  int msg[3];
  double best_scores[total_sets + 1];
  SSMResult best_results[total_sets + 1];
  SGN_Array * stack = SGN_Array_new(); /* SSM instance stack */

  for(int i = 1; i <= total_sets; i++){
	best_scores[i] = - DBL_MAX;
	best_results[i].ssm = NULL;
  }
  
  while(1){
	int dim, set_id;
	if(count == 0){
	  /* For the first time, calculate the first job request in this process. */
	  int dim_idx;
	  calc_first_dispatch(args->iteration, args->sets, args->len_dims, 
						  args->MPI_rank, &dim_idx, &set_id);
	  if(set_id == 0){
		dim = -2; /* No jobs for this process. */
	  } else {
		dim = args->dims[dim_idx];
	  }
	  msg[0] = 0; /* do nothing */
	} else {
	  /* Send the job request to the root process, and receive the resoponse. */
	  fprintf(SIGN_ERR, "Requesting a job.\n");  fflush(SIGN_ERR);
	  MPI_Send(&likelihood, 1, MPI_DOUBLE, 0, 2, MPI_COMM_WORLD);
	  MPI_Recv(&msg, 3, MPI_INT, 0, 1, MPI_COMM_WORLD, &status);
	  set_id = msg[1];
	  dim = msg[2];
	}
	  
	fprintf(SIGN_ERR, "[JobID:%d] Job dispatched. dim=%d, set ID=%d\n",	count, dim, set_id);
	fflush(SIGN_ERR);

	if(msg[0] > 0){
	  /* Output */
	  fprintf(SIGN_ERR, "Output requested: set ID=%d\n", msg[0]); fflush(SIGN_ERR);
	  ssm_output(best_results + msg[0], work, X, Y,
				 args->output_file, args->output_ext, (msg[0] % args->sets) + 1,
				 args->output_each, args->output_state, args->output_pvalues);

	} else if(msg[0] < 0){
	  /* Dispose */
	  fprintf(SIGN_ERR, "Disposing requested: set ID=%d\n", - msg[0]); fflush(SIGN_ERR);
	  if(best_results[- msg[0]].ssm != NULL){
		SGN_Array_add(stack, best_results[- msg[0]].ssm);
		best_results[- msg[0]].ssm = NULL;
	  }
	  /* NOTE: NULL check is required because, when job "-msg[0]" processed NaN,
		       it is not stored in best_result[-msg[0]] even if it is the first
			   job assined to the set ID "-msg[0]".  The root process assumes
			   that at least one dispatched job for the same IDs is stored.  However,
			   the case that NaN produceed, no best results are stored.
	  */			   
	}

	if(dim == -2){
	  /* Job finished. */
	  fprintf(SIGN_ERR, "Finished and waiting for the final call.\n"); fflush(SIGN_ERR);
	  break;
	} else if(dim == -1){
	  fprintf(SIGN_ERR, "Requesting again.\n"); fflush(SIGN_ERR);
	  likelihood = 0.0;
	  continue;
	}

	fprintf(SIGN_ERR, "Performing the SSM estimation.\n"); fflush(SIGN_ERR);
	SGN_Timer_clear(timer); SGN_Timer_start(timer);
	//SSM_write_fp(ssm, Y->p, dim, SIGN_ERR);
	if(args->em->constrain_F && prev_dim != dim){ 
	  SSM_EM_set_C(dim, work->C);
	  prev_dim = dim;
	}
	SSMResult result;
	int retry = 0;
	while(1){
	  SSM_init(ssm, args->init, Y->p, dim, Y->Y, Y->samples);
	  result = SSM_EM_algo(args->em, ssm, X, T, sumT, Y, dim, work);
	  if((args->em->retry < 0 || retry < args->em->retry)
		 && (!isnormal(result.likelihood) || result.monotone == 0)){
		retry++;
		fprintf(SIGN_ERR, "Estimation failed.  Retrying (%d)...\n", retry);
		fflush(SIGN_ERR);
		continue;
	  } else {
		if(args->em->retry != -1 && retry >= args->em->retry){
		  fprintf(SIGN_ERR, "WARNING: Retry count exceeded the limitation.\n");
		  fflush(SIGN_ERR);
		}
		break;
	  }
	}/* while(1); */
	SGN_Timer_stop(timer);
	likelihood = result.likelihood;
	fprintf(SIGN_ERR, "Estimation Finished for dim=%d, set id=%d: Likelihood=%f, Loops=%d\n",
			dim, set_id, likelihood, result.loops);
	fprintf(SIGN_ERR, "   time = %f sec\n", SGN_Timer_get(timer));
	fprintf(SIGN_ERR, "X-SINGLE-TIME:\t%f\n", SGN_Timer_get(timer));
	fflush(SIGN_ERR);

	if(isnormal(likelihood) && likelihood > best_scores[set_id]){
	  /* Update the saved best result. */
	  update(set_id, best_scores, &result, best_results, stack, max_dim);
	}

	count++;
  }/* while(1) */

  int num;
  MPI_Recv(&num, 1, MPI_INT, 0, 2, MPI_COMM_WORLD, &status);
  if(num != 0){
	fprintf(SIGN_ERR, "%d outputs requested.\n", num); fflush(SIGN_ERR);
	int ids[num];
	MPI_Recv(ids, num, MPI_INT, 0, 3, MPI_COMM_WORLD, &status);
	for(int i = 0; i < num; i++){
	  if(i != 0) fprintf(SIGN_ERR, ", ");
	  fprintf(SIGN_ERR, "%d", ids[i]); fflush(SIGN_ERR);
	  ssm_output(best_results + ids[i], work, X, Y,
				 args->output_file, args->output_ext, (ids[i] % args->sets) + 1,
				 args->output_each, args->output_state, args->output_pvalues);
	}
	fprintf(SIGN_ERR, "\n");
	fflush(SIGN_ERR);
  } else {
	fprintf(SIGN_ERR, "No extra outputs required.\n"); fflush(SIGN_ERR);
  }

  fprintf(SIGN_ERR, "All jobs finished.\n"); fflush(SIGN_ERR);
  
}
/* execute_jobs() ========================================================== */
#endif /* USE_MPI */
/* End of file.  Copyright (C) 2010  HGC, IMS, The University of Tokyo       */
