/*
  lang/SGN_Object.c : Generic Object

  Copyright (C) 2009-2012 SiGN Project Members.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#include "SGN_Object.h"

/*!\struct SGN_Object
 *
 *\brief Generic object class.
 *
 * \ref SGN_Object module is to provide functions to handle generic
 * values and/or objects.
 * 
 * The main purpose of \ref SGN_Object module is to exclude chances
 * to handle void pointers directly.
 *
 * This module can hide void pointers that have no type information
 * from your programs.
 *
 * The acutual type specific operations are handled by the \ref SGN_Type
 * module, not by this module.  This simply provides the interfaces
 * to handle both type information and its value simultaneously.
 *
 *\see \ref SGN_Type, \ref SGN_ObjArray, \ref SGN_KeyObjects
 *\author Yoshinori Tamada &lt;tamada at ims.u-tokyo.ac.jp&gt;
 */
struct SGN_Object_t {
  void * ptr;
  SGN_Type type;
};

#include "../util/SGN_Memory.h"
#include "../signcom.h"

/*!\brief Generates an empty \ref SGN_Object instance.
 *
 * By default, the generated instance has \ref SGN_Type_NULL type at first.
 * and NULL pointer.  Use SGN_Object_set() to set a value in the generated instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Object */
SGN_Object * SGN_Object_new(){
  SGN_Object * this = malloc(sizeof(SGN_Object));

  this->type = SGN_Type_NULL;
  this->ptr = NULL;

  return this;
}
/* SGN_Object_new() ============================================================ */

SGN_Object * SGN_Object_new_value(void * ptr, SGN_Type type){
  SGN_Object * this = malloc(sizeof(SGN_Object));
  this->type = type;
  this->ptr = ptr;
  return this;
}
/* SGN_Object_new_value() ================================================== */

SGN_Object * SGN_Object_new_int(int value){
  SGN_Object * this = malloc(sizeof(SGN_Object));

  this->type = SGN_Type_INT;
  this->ptr = SGN_Type_int_clone(&value);

  return this;
}
/* SGN_Object_new_int() ======================================================== */

SGN_Object * SGN_Object_new_error(int value){
  SGN_Object * this = malloc(sizeof(SGN_Object));

  this->type = SGN_Type_ERROR;
  this->ptr = SGN_Type_int_clone(&value);

  return this;
}
/* SGN_Object_new_error() ====================================================== */

/*!\brief Generates an SGN_Object instance containing a double value.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Object */
SGN_Object * SGN_Object_new_double(double value){
  SGN_Object * this = malloc(sizeof(SGN_Object));

  this->type = SGN_Type_DOUBLE;
  this->ptr = SGN_Type_double_clone(&value);

  return this;
}
/* SGN_Object_new_double() ===================================================== */

SGN_Object * SGN_Object_new_string(const char * value){
  SGN_Object * this = malloc(sizeof(SGN_Object));

  this->type = SGN_Type_STRING;
  this->ptr = SGN_Type_string_clone(value);

  return this;
}
/* SGN_Object_new_string() ===================================================== */

/*!\brief Generates an \ref SGN_Object from string expression with its type.
 *
 *\param value  string expression of the value.
 *\param type   \ref Type ID of \a value.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Object */
SGN_Object * SGN_Object_new_by_string(const char * value, SGN_Type type){
  SGN_Object * this = malloc(sizeof(SGN_Object));

  this->type = type;
  this->ptr = SGN_Type_from_str(type, value, NULL);
  
  return this;
}
/* SGN_Object_new_by_string() ================================================== */

/*!\brief Deletes the \ref SGN_Object instance.
 *
 * This peforms deep deletion, i.e, deletes both the instance pointed by this
 * and this \ref SGN_Object instance.  The type specified deletion is
 * performed by SGN_Type_delete() in \ref Type module.
 *
 *\see SGN_Type_delete().
 *
 *\param  this  \ref SGN_Object intance to be deleted.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Object */
void SGN_Object_delete(SGN_Object * this){
  SGN_Type_delete(this->ptr, this->type);
  free(this);
}
/* SGN_Object_delete() ========================================================= */


void SGN_Object_delete_ref(void * this){
  free(this);
}

void SGN_Object_delete_v(void * this){ SGN_Object_delete((SGN_Object *) this); }

void SGN_Object_release(SGN_Object * this){
  SGN_Type_delete(this->ptr, this->type);
  this->type = SGN_Type_NULL;
  this->ptr = NULL;
}

/*!\brief Generates a clone of this \ref SGN_Object instance.
 *
 * This performs the deep copy of this \ref SGN_Object instance.
 * The type specific cloning is performed by SGN_Type_clone() in \ref Type module.
 *
 *\see \ref Type, SGN_Type_clone().
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Object */
SGN_Object * SGN_Object_clone(const SGN_Object * this){
  SGN_Object * object = malloc(sizeof(SGN_Object));
  object->type = this->type;
  object->ptr = SGN_Type_clone(this->ptr, this->type);

#ifdef SIGN_DEBUG
  if(this->type > (int) SGN_Type_ERROR){
	fprintf(SIGN_ERR, "WARNING: SGN_Object_clone(): Illigal type number detected: %d\n",
			(int) this->type);
	abort();
  }
#endif

  return object;
}
/* SGN_Object_clone() ========================================================== */


/*!\brief Sets the \ref SGN_Object to point to the specified pointer.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Object */
void SGN_Object_set(SGN_Object * this, void * pointer, SGN_Type type){
  SGN_Type_delete(this->ptr, this->type);
  this->ptr = pointer;
  this->type = type;
}
/* SGN_Object_set() ============================================================ */

/*!\brief Sets this \ref SGN_Object to point to the specified pointer.
 *
 *\return previously set pointer.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Object */
void * SGN_Object_set_ret(SGN_Object * this, void * pointer, SGN_Type type){
  void * ret = this->ptr;
  this->ptr = pointer;
  this->type = type;
  return ret;
}
/* SGN_Object_set_ret() ======================================================== */

void * SGN_Object_pointer(const SGN_Object * this){
  return this->ptr;
}

void * SGN_Object_ptr(const SGN_Object * this){
  return this->ptr;
}

void * SGN_Object_get(SGN_Object * this, SGN_Type type){
  if(this->type == type){
	return this->ptr;
  } else {
	return NULL;
  }
}
/* SGN_Object_get() ======================================================== */

/*!\brief Returns the type of this object.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Object */
SGN_Type SGN_Object_type(const SGN_Object * this){
  return this->type;
}

/*!\brief Returns the integer value.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Object */
int SGN_Object_int_value(const SGN_Object * this, int def_value){
  if(this->type == SGN_Type_INT){
	return *((int *) this->ptr);
  } else if(this->type == SGN_Type_STRING){
	return atoi((char *) this->ptr);
  } else {
	return def_value;
  }
}
/* SGN_Object_int_value() ================================================== */

double SGN_Object_double_value(const SGN_Object * this, double def_value){
  if(this->type == SGN_Type_DOUBLE){
	return *((double *) this->ptr);
  } else if(this->type == SGN_Type_STRING){
	return atof((char *) this->ptr);
  } else {
	return def_value;
  }
}
/* SGN_Object_double_value() =============================================== */

/*!\brief Retturns the string value.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Object */
const char * SGN_Object_string_value(const SGN_Object * this, const char * def_value){
  if(this->type == SGN_Type_STRING){
	return (const char *) this->ptr;
	//} else if(this->type == SGN_Type_INT){
	
  } else {
	return def_value;
  }
}
/* SGN_Object_string_value() =============================================== */

/*
CSMLObject * SGN_Object_CSMLObject_value(const SGN_Object * this, CSMLObject * def_value){
  if(this->type == SGN_Type_CSMLObject){
	return (CSMLObject *) this->ptr;
  } else {
	return def_value;
  }
}
*/
/* SGN_Object_CSMLObject_value() =========================================== */

/*!\brief Returns the name of the type.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Object */
const char * SGN_Object_type_name(const SGN_Object * this){
#ifdef SIGN_DEBUG
  if(this->type > (int) SGN_Type_ERROR){
	fprintf(SIGN_ERR, "WARNING: SGN_Object_type_name(): Illegal type number detected: %d\n",
			(int) this->type);
	abort();
  }
#endif
  return SGN_Type_name(this->type);
}
/* SGN_Object_type_name() ====================================================== */

/*!\brief Prints the string expression of this \ref SGN_Object into the file 
 *  pointer.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Object */
void SGN_Object_fprint(const SGN_Object * this, FILE * fp){
  SGN_Type_fprint_value(this->ptr, this->type, fp);
}
/* SGN_Object_fprintf() ======================================================== */


#ifdef SIGN_DEBUG
#include "../util/SGN_Node.h"
#include "../util/SGN_Edge.h"

int * Obj_int(const SGN_Object * this){
  if(this->type != SGN_Type_INT){
	return NULL;
  } else {
	return (int *) this->ptr;
  }
}

double * Obj_double(const SGN_Object * this){
  if(this->type != SGN_Type_DOUBLE){
	return NULL;
  } else {
	return (double *) this->ptr;
  }
}

char * Obj_str(const SGN_Object * this){
  if(this->type != SGN_Type_STRING){
	return NULL;
  } else {
	return (char *) this->ptr;
  }
}

SGN_Node * Obj_Node(const SGN_Object * this){
  if(this->type != SGN_Type_NODE){
	return NULL;
  } else {
	return (SGN_Node *) this->ptr;
  }
}

SGN_Edge * Obj_Edge(const SGN_Object * this){
  if(this->type != SGN_Type_EDGE){
	return NULL;
  } else {
	return (SGN_Edge *) this->ptr;
  }
}

CSMLObject * Obj_CSMLObject(const SGN_Object * this){
  if(this->type != SGN_Type_CSMLObject){
	return NULL;
  } else {
	return (CSMLObject *) this->ptr;
  }
}
#endif /* SIGN_DEBUG */

/* End of file.  Copyright (C) 2010-2012  SiGN Project Members */
