/*
  sign/lang/SGN_Type.c : Routines for dealing with types

  Copyright (C) 2009-2012 SiGN Project Members.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

/*!\defgroup SGN_Type
 *\brief SGN_Type module provides routines for handling types.
 *
 *\author Yoshinori Tamada &lt;tamada at is.s.u-tokyo.ac.jp&gt;
 */

#include "../signcom.h"
#include "SGN_Type.h"
#include "../util/SGN_Str.h"
#include "../util/SGN_KeyValues.h"
#include "SGN_Object.h"
#include "../util/SGN_DoubleArray.h"

#include <string.h>
#include <strings.h>

typedef void * (* SGN_Type_new_t)(void);
typedef void (* SGN_Type_delete_t)(void *);
typedef void * (* SGN_Type_clone_t)(const void *);
typedef void * (* SGN_Type_new_str_t)(const char *);
typedef int (* SGN_Type_cmp_t)(const void * x, const void * y);
typedef void (* SGN_Type_fprint_t)(const void *, FILE *);

typedef struct {
  char * name;
  SGN_Type_new_t new;
  SGN_Type_delete_t delete;
  SGN_Type_clone_t clone;
  SGN_Type_new_str_t new_str;
  SGN_Type_cmp_t cmp;
  SGN_Type_fprint_t fprint;
} SGN_Type_DB_t;

/* Proto-types */
//void SGN_Array_fprint_v(const void * v, FILE * fp);

static SGN_Type_DB_t SGN_Type_db[] = {
  /* SGN_Type_NULL */
  { "NULL", NULL, SGN_Type_NULL_delete, NULL, NULL, NULL,},

  /* SGN_Type_INT */
  {"int", SGN_Type_int_new, free, SGN_Type_int_clone, SGN_Type_int_new_str, 
   SGN_Type_int_cmp, SGN_Type_int_fprint,},

  /* SGN_Type_UINT */
  {"uint", NULL, free, NULL, NULL, NULL, NULL,},

  /* SGN_Type_FLOAT */
  {"float", NULL, free,	NULL, NULL,	NULL, NULL,},

  /* SGN_Type_DOUBLE */
  {"double", SGN_Type_double_new, free,	SGN_Type_double_clone, SGN_Type_double_new_str,	
   SGN_Type_double_cmp, SGN_Type_double_fprint, },

  /* SGN_Type_STRING */
  {"string", SGN_Type_string_new, free,	SGN_Type_string_clone, SGN_Type_string_new_str,
   SGN_Type_string_cmp, SGN_Type_string_fprint, },

  /* SGN_Type_CHAR */
  {"char", NULL, free, NULL, NULL, NULL, NULL, },

  /* SGN_Type_OBJECT */
  {"Object", (void *) SGN_Object_new, SGN_Object_delete_v, NULL, NULL, NULL, NULL,},

  /* SGN_Type_ARRAY */  { "Array", NULL, NULL, NULL, NULL, NULL, SGN_Array_fprint_v},
  /* SGN_Type_INT_ARRAY */ { "IntArray", NULL, NULL, NULL, NULL, NULL, NULL},
  /* SGN_Type_DOUBLE_ARRAY */ { "DoubleArray", NULL, SGN_DoubleArray_delete_v, NULL, NULL, NULL, SGN_DoubleArray_fprint_v},
  /* SGN_Type_STR_ARRAY */ { "StrArray", NULL, SGN_StrArray_delete_v, NULL, NULL, NULL, SGN_StrArray_fprint_v},
  /* SGN_Type_KEY_VALUES */ { "KeyValues",   NULL, SGN_KeyValues_delete_v, NULL, NULL, NULL, SGN_KeyValues_print_v},
  /* SGN_Type_Node */ { "Node", NULL, NULL, NULL, NULL, NULL, NULL,},
  /* SGN_Type_Edge */ { "Edge", NULL, NULL, NULL, NULL, NULL, NULL,},
  /* SGN_Type_Network */ { "Network", NULL, NULL, NULL, NULL, NULL, NULL,},
  /* SGN_Type_DataFrame */ { "DataFrame", NULL, NULL, NULL, NULL, NULL, NULL,},
  /* SGN_Type_CSMLObject */ {"CSMLObject", NULL, NULL, NULL, NULL, NULL, NULL},
  /* SGN_Type_Data */ {"Data", NULL, NULL, NULL, NULL, NULL, NULL},
  /* SGN_Type_KEY_VALUES_ADD */ {NULL},
  /* SGN_Type_INT_OPT */ {NULL},
  /* SGN_Type_INT_ADD */ {NULL},
  /* SGN_Type_STRING_OPT */ {NULL},
  /* SGN_Type_STRING_ADD */ {NULL},
  /* SGN_Type_ON_OFF */ {NULL},
  /* SGN_Type_ON */ {NULL},
  /* SGN_Type_OFF */ {NULL},
  /* SGN_Type_INT_ARRAY_ADD */ {NULL},
  /* SGN_Type_INT_ARRAY_ADD2 */ {NULL},
  /* SGN_Type_DOUBLE_ARRAY_ADD */ {NULL},
  /* SGN_Type_DOUBLE_ARRAY_ADD2 */ {NULL},
  /* SGN_Type_STR_ARRAY_ADD */ {NULL},
  /* SGN_Type_STR_ARRAY_ADD2 */ {NULL},
  /* SGN_Type_FUNC */ {NULL},
  /* SGN_Type_FUNC2 */ {NULL},
  /* SGN_Type_FUNC3 */ {NULL},
  /* SGN_Type_OTHER */ {NULL},
  /* SGN_Type_StrMatrix */ {NULL},
  /* SGN_Type_INT_REGION */ {NULL},
  /* SGN_Type_GRAPH */ {"Graph"},
  /* SGN_Type_ADJ_GRAPH */ {"AdjGraph"},
  /* SGN_Type_COMPLETE_GRAPH */ {"CompleteGraph"},
  /* [SGN_Type_NONE] */ {"NONE", NULL, NULL, NULL, NULL, NULL, NULL,},
  /* [SGN_Type_ERROR] */ {"ERROR", NULL, NULL, NULL, NULL, NULL, NULL},
  {NULL},
};

/*!\brief Returns the name of the given type.
 *
 * The following code always becomes true:
 * \code
 * type == SGN_Type_get(SGN_Type_name(type))
 * \endcode
 *
 *\param type  SGN_Type value.
 *\return string expression of \a type.  DO NOT modify or free the
 * returned pointer.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Type */
const char * SGN_Type_name(SGN_Type type){
  if(type > SGN_Type_ERROR || SGN_Type_db[type].name == NULL){
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_Type_name(): Unsupported type ID: %d\n",
			(int) type);
	exit(1);
  }
  return SGN_Type_db[type].name;
}
/* SGN_Type_name() ============================================================= */

/*!\brief Returns \ref SGN_Type value corresponding to a given string.
 *
 * - <TT>int</TT>
 * - <TT>integer</TT>
 * - <TT>float</TT>
 * - <TT>double</TT>
 * - <TT>string</TT> (\ref SGN_Type_STRING)
 * - <TT>str</TT>  (\ref SGN_Type_STRING)
 *
 *\return \ref SGN_Type value, or \ref SGN_Type_ERROR if there is no
 *   type specified by \a str.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Type */
SGN_Type SGN_Type_get(const char * str){

  if(strcasecmp(str, "int") == 0 || 
	 strcasecmp(str, "integer") == 0){
	return SGN_Type_INT;
  } else if(strcasecmp(str, "uint") == 0 ||
			strcasecmp(str, "unsigned int") == 0 ||
			strcasecmp(str, "unsigned integer") == 0){
	return SGN_Type_UINT;
  } else if(strcasecmp(str, "double") == 0){
	return SGN_Type_DOUBLE;
  } else if(strcasecmp(str, "float") == 0){
	return SGN_Type_FLOAT;
  } else if(strcasecmp(str, "string") == 0 ||
			strcasecmp(str, "str") == 0){
	return SGN_Type_STRING;
  } else if(strcasecmp(str, "doublearray") == 0){
	return SGN_Type_DOUBLE_ARRAY;
  } else {
	return SGN_Type_ERROR;
  }
}
/* SGN_Type_get() =============================================================== */

/*!\brief Geretates a clone of the given instance.
 *
 * This generates a new instance of type \a type that is idential to
 * the original instance pointed by \a obj.
 *
 *\param obj  pointer to the instance to be cloned.
 *\param type  SGN_Type value of \a obj.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Type */
void * SGN_Type_clone(const void * obj, SGN_Type type){
  if(SGN_Type_db[type].clone == NULL){
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_Type_clone(): Unsupported type: %s\n",
			SGN_Type_name(type));
	exit(1);
  }
  return SGN_Type_db[type].clone(obj);
}
/* SGN_Type_clone() ============================================================ */

/*!\breif Deletes the instance whose type is specified by the \ref SGN_Type value.
 *
 * This deletes the instance and releases resources allocated by it.
 *
 *\param obj pointer to the instance to be deleted.
 *\param type  SGN_Type value of \a obj.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Type */
void SGN_Type_delete(void * obj, SGN_Type type){
  if(SGN_Type_db[type].delete == NULL){
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_Type_delete():"
			" Unsupported type ID given: %s (%d)\n",
			SGN_Type_name(type), type);
	abort();
  }
#ifdef SIGN_DEBUG
  fprintf(SIGN_ERR, "SGN_Type: Deleting \'%s\'. addr=%p\n", 
		  SGN_Type_name(type), obj);
  if(SGN_Type_db[type].delete == free){
	free(obj);
  } else {
	SGN_Type_db[type].delete(obj);
  }
#else
  SGN_Type_db[type].delete(obj);
#endif 
}
/* SGN_Type_delete() ----------------------------------------------------------- */

/*!\brief
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Type */
void * SGN_Type_new_str(const char * str, SGN_Type type){
  if(SGN_Type_db[type].new_str == NULL){
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_Type_new_str(): Unsupported type ID: %d\n",
			(int) type);
	exit(1);
  }
  return SGN_Type_db[type].new_str(str);
}
/* SGN_Type_new_str() ========================================================== */

/*!\brief Prints a value into a file stream.
 *
 *\param type 
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Type */
void SGN_Type_fprint_value(const void * obj, SGN_Type type, FILE * fp){
  if(SGN_Type_db[type].fprint == NULL){
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_Type_fprint_value(): Unsupported type: %s\n",
			SGN_Type_name(type));
	exit(1);
  }
  SGN_Type_db[type].fprint(obj, fp);
}
/* SGN_Type_fprint_value()====================================================== */

void * SGN_Type_int_new(){
  return malloc(sizeof(int));
}
/* SGN_Type_int_new() ========================================================== */

void * SGN_Type_int_clone(const void * ptr){
  int * p = malloc(sizeof(int));
  *p = *((const int *) ptr);
  return p;
}
/* SGN_Type_int_clone() ======================================================== */

void SGN_Type_int_fprint(const void * ptr, FILE * fp){
  const int * p = (const int *) ptr;
  fprintf(fp, "%d", *p);
}
/* SGN_Type_int_fprint() ======================================================= */

void * SGN_Type_int_new_str(const char * str){
  void * ptr = SGN_Type_int_new();
  *((int *) ptr) = atoi(str);
  return ptr;
}
/* SGN_Type_int_new_str() ====================================================== */

int SGN_Type_int_cmp(const void * x, const void * y){
  if(*((const int *) x) < *((const int *) y)){
	return -1;
  } else if(*((const int *) x) > *((const int *) y)){
	return 1;
  } else {
	return 0;
  }
}
/* SGN_Type_int_cmp() ------------------------------------------------------ */

void * SGN_Type_float_new(){
  return malloc(sizeof(float));
}

void * SGN_Type_double_new(){
  return malloc(sizeof(double));
}

void * SGN_Type_double_clone(const void * ptr){
  double * p = malloc(sizeof(double));
  *p = *((const double *) ptr);
  return p;
}
/* SGN_Type_double_clone() ===================================================== */

void SGN_Type_double_fprint(const void * ptr, FILE * fp){
  const double * p = (const double *) ptr;
  fprintf(fp, "%f", *p);
}
/* SGN_Type_double_fprint() ==================================================== */

int SGN_Type_double_cmp(const void * x, const void * y){
  if(*((const double *) x) < *((const double *) y)){
	return -1;
  } else if(*((const double *) x) > *((const double *) y)){
	return 1;
  } else {
	return 0;
  }
}
/* SGN_Type_double_cmp() ------------------------------------------------------ */

void * SGN_Type_double_new_str(const char * str){
  void * ptr = SGN_Type_double_new();
  *((double *) ptr) = atof(str);
  return ptr;
}
/* SGN_Type_double_new_str() =================================================== */

/*!\brief Generates an instance from the string expression.
 *
 *
 * Currently supported types are:
 *  - \ref SGN_Type_INT
 *  - \ref SGN_Type_DOUBLE
 *  - \ref SGN_Type_STRING
 *  - \ref SGN_Type_DOUBLE_ARRAY (\ref SGN_DoubleArray)
 *
 *\param type \ref SGN_Type ID of the value given by \a str.
 *\param str  string expression of the value to be instanciated.
 *\param ptr  pointer to which the instanciated value is set.  If NULL is
 *            specified, new pointer (memory) is allocated.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Type */
void * SGN_Type_from_str(SGN_Type type, const char * str, void * ptr){
  switch(type){
  case SGN_Type_INT:
	{
	  if(ptr == NULL) ptr = malloc(sizeof(int));
	  *((int *) ptr) = atoi(str);
	}
	break;

  case SGN_Type_DOUBLE:
	{
	  if(ptr == NULL) ptr = malloc(sizeof(double));
	  *((double *) ptr) = atof(str);
	}
	break;

  case SGN_Type_STRING:
	{
	  if(ptr == NULL){
		ptr = SGN_Str_copy(str);
	  } else {
		memcpy(ptr, str, sizeof(char) * (strlen(str) + 1));
	  }
	}
	break;

  case SGN_Type_DOUBLE_ARRAY:
	{
	  if(ptr == NULL){
		ptr = SGN_DoubleArray_new_str(str);
	  } else {
		SGN_DoubleArray_set_str((SGN_DoubleArray *) ptr, str);
	  }
	}
	break;

  default:
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_Type_from_str():"
			" unsupported type id given: %s (%d)\n",
			SGN_Type_name(type), type);
	abort();
  }
  return ptr;
}
/* SGN_Type_from_str() -------------------------------------------------------- */

char * SGN_Type_to_str(SGN_Type type, const void * value, char * str){
  return NULL;
}

/*!\brief Duplicates the object whose type is specified by the
 *    \ref SGN_Type ID.  
 *
 *\param obj  pointer to the object to be duplicated.
 *\param type  \ref SGN_Type ID.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Type */
void * SGN_Type_duplicate(const void * obj, SGN_Type type){
  void * ret = NULL;

  switch(type){
  case SGN_Type_INT:
	ret = malloc(sizeof(int));
	*((int *) ret) = *((const int *) obj);
	break;
	
  case SGN_Type_DOUBLE:
	ret = malloc(sizeof(double));
	*((double *) ret) = *((const double *) obj);
	break;
	
  case SGN_Type_STRING:
	ret = SGN_Str_copy((const char *) obj);
	break;

  case SGN_Type_KEY_VALUES:
	ret = SGN_KeyValues_clone((SGN_KeyValues *) obj);
	break;

  case SGN_Type_DOUBLE_ARRAY:
	ret = SGN_DoubleArray_copy((const SGN_DoubleArray *) obj);
	break;
	
  default:
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_Type_duplicate(): "
			"unsupported value type: %d\n", type);
	abort();
  }
  return ret;
}
/* SGN_Type_duplicate() ------------------------------------------------------- */


/*!\brief Compares two values.
 *
 * This is simply an interface for the comparison functions listed below.
 *  - SGN_Type_double_cmp()
 *
 *\param type  type ID of the two values.
 *\param x   pointer to the value to be compared.
 *\param y   pointer to the value to be compared.
 *\return an integer less than, equal to, or greater than zero if
 *     \a x is less than, equal to, or greater than \a y.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Type */
int SGN_Type_compare(SGN_Type type, const void * x, const void * y){
  SGN_Type_cmp_t f = SGN_Type_db[type].cmp;
  if(f == NULL){
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_Type_compare():"
			" Unsupported type specified: %s (%d)\n",
			SGN_Type_name(type), type);
	abort();
  }
  return f(x, y);
  
  //switch(type){
  //case SGN_Type_DOUBLE: return SGN_Type_double_cmp(x, y);
  //  default:
  //fprintf(SIGN_ERR, "FATAL ERROR: SGN_Type_compare():"
  //			" Unsupported type specified: %s (%d)\n",
  //			SGN_Type_name(type), type);
  //abort();
  //}
  //#ifdef __FUJITSU
  //  return 0;
  //#endif /* __FUJITSU */
}
/* SGN_Type_compare() --------------------------------------------------------- */

/*!\brief Compares two values by the specified way.
 *
 * This compares the given two values by SGN_Type_compare().  Then
 * based on its result, this returns if the condition given as
 * an operator \a op is satisfied.
 *
 * <DL>
 * <DT><TT>OP_EQ</TT></DT><DD>Returns 1 if two values are equal.</DD>
 * <DT><TT>OP_GT</TT></DT><DD>Returns 1 if \a x is greater than \a y.</DD>
 * <DT><TT>OP_GE</TT></DT><DD>Returns 1 if \a x is greater than or equal to \a y.</DD>
 * <DT><TT>OP_LT</TT></DT><DD>Returns 1 if \a x is less than \a y.</DD>
 * <DT><TT>OP_LE</TT></DT><DD>Returns 1 if \a x is less than or equal to \a y.</DD>
 * </DL>
 *
 *\param type type ID of two values.
 *\param op Operator to be used for comparison.
 *\param x  pointer to the value to be compared.
 *\param y  pointer to the value to be compared.
 *
 *\return 1 if the specified operation or condition is satisfied, or 0
 *   if not.
 *
 *\sa SGN_Type_get_compare_op()
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Type */
int SGN_Type_compare_op(SGN_Type type, SGN_Type_Operator op, const void * x, const void * y){
  int r = SGN_Type_compare(type, x, y);
  switch(op){
  case SGN_OP_EQ: return r == 0 ? 1 : 0;
  case SGN_OP_GT: return r > 0 ? 1 : 0;
  case SGN_OP_GE: return r >= 0 ? 1 : 0;
  case SGN_OP_LT: return r < 0 ? 1 : 0;
  case SGN_OP_LE: return r <= 0 ? 1 : 0;
  default:
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_Type_compare_op():"
			" Unsupported operator IDt specified: %d\n", op);
	abort();
  }
#ifdef __FUJITSU
  return 0;
#endif
}
/* SGN_Type_compare_op() ------------------------------------------------------ */

/*!\brief Returns the operator from the string expression.
 *
 * This returns the comparison operator corresponding to its string expression.
 * Their corresponding pairs are:
 * - &quot;<tt>EQ</tt>&quot; or &quot;<tt>=</tt>&quot; : SGN_OP_EQ
 * - &quot;<tt>GT</tt>&quot; or &quot;<tt>&gt;</tt>&quot; : SGN_OP_GT
 * - &quot;<tt>GE</tt>&quot; or &quot;<tt>&gt;=</tt>&quot; : SGN_OP_GE
 * - &quot;<tt>LT</tt>&quot; or &quot;<tt>&lt;</tt>&quot; : SGN_OP_LT
 * - &quot;<tt>LE</tt>&quot; or &quot;<tt>&lt;=</tt>&quot; : SGN_OP_LE
 *
 * The string expression is evaluated case-insensitively.
 *
 *\Returns an SGN_Type_Operator of the given string expression or SGN_OP_ERROR if
 *  there is no such operator.
 *\ingroup SGN_Type  */
SGN_Type_Operator SGN_Type_get_compare_op(const char * str){
  if(str == NULL){
	return SGN_OP_ERROR;
  } else if(strcasecmp(str, "eq") == 0 || strcasecmp(str, "=") == 0){
	return SGN_OP_EQ;
  } else if(strcasecmp(str, "gt") == 0 || strcasecmp(str, ">") == 0){
	return SGN_OP_GT;
  } else if(strcasecmp(str, "ge") == 0 || strcasecmp(str, ">=") == 0){
	return SGN_OP_GE;
  } else if(strcasecmp(str, "lt") == 0 || strcasecmp(str, "<") == 0){
	return SGN_OP_LT;
  } else if(strcasecmp(str, "le") == 0 || strcasecmp(str, "<=") == 0){
	return SGN_OP_LE;
  } else {
	return SGN_OP_ERROR;
  }
}
/* SGN_Type_get_compare_op() =============================================== */

void SGN_Type_NULL_delete(void * ptr){
  /* do nothing */
}
/* SGN_Type_NULL_delete() ====================================================== */

/*****************************************************************************/
/* String                                                                    */
/*****************************************************************************/

void * SGN_Type_string_new(){
  return malloc(sizeof(char));
}

void * SGN_Type_string_clone(const void * ptr){
  char * p = malloc(sizeof(char) * strlen((const char *) ptr));
  strcpy(p, (const char *) ptr);
  return p;
}
/* SGN_Type_string_clone() ===================================================== */

void * SGN_Type_string_new_str(const char * str){
  void * ptr = malloc(sizeof(char) * (strlen(str) + 1));
  strcpy((char *) ptr, str);
  return ptr;
}
/* SGN_Type_string_new_str() =================================================== */

int SGN_Type_string_cmp(const void * x, const void * y){
  return strcmp((const char *) x, (const char *) y);
}
/* SGN_Type_string_cmp() --------------------------------------------------- */

void SGN_Type_string_fprint(const void * ptr, FILE * fp){
  const char * p = (const char *) ptr;
  fprintf(fp, "%s", p);
}
/* SGN_Type_string_fprint() ==================================================== */

/* End of file.  Copyright (C) 2009-2012 SiGN Project Members */
