/*
  sign/lang/SGN_Type.h : Routines for dealing with types

  Copyright (C) 2009-2012 SiGN Project Members.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#ifndef __SGN_TYPE_H
#define __SGN_TYPE_H

#include <stdio.h>

/*!\brief Types of values supported by SiGN library.
 *
 * Some value types are used only in \ref KeyValues module.
 *
 *\see KeyValues
 *\ingroup SGN_Type */
typedef enum {

  /* !!!!! IF INSERT A NEW TYPE, ALSO ADD TO SGN_Type_db[] in SGN_Type.c !!!!! */

  /*! This represents undefined type. */
  SGN_Type_NULL,

  /*! Integer (<TT>int</TT>) value. */
  SGN_Type_INT,

  /*! Unsigned integer (<TT>unsigned int</TT>) value. */
  SGN_Type_UINT,
  
  /*! Single precision float (<TT>float</TT>) value. */
  SGN_Type_FLOAT,

  /*! Double precision float (<TT>double</TT>) value. */
  SGN_Type_DOUBLE,

  /*! String (null-terminated <TT>char *</TT>) value.
	If used in \ref KeyValues module, this means that
	the string is copied (duplicated) and then stored
	to the variable.
   */
  SGN_Type_STRING,

  /*! Character (<TT>char</TT>) value. */
  SGN_Type_CHAR,

  /*! Generic \ref SGN_Object. */
  SGN_Type_OBJECT,

  /*! \ref Array instance. */
  SGN_Type_ARRAY,
  
  /*! \ref IntArray (integer array) instance. */
  SGN_Type_INT_ARRAY,

  /*! \ref SGN_DoubleArray (double array) instance. */
  SGN_Type_DOUBLE_ARRAY,

  /*! String array (\ref StrArray) instance.  */
  SGN_Type_STR_ARRAY,

  /*! \ref KeyValues instance. */
  SGN_Type_KEY_VALUES,

  /*! \ref Node instance. */
  SGN_Type_NODE,

  /* \ref Edge instance */
  SGN_Type_EDGE,

  /* \ref SGN_Network instance */
  SGN_Type_NETWORK,

  /* \ref SGN_DataFrame instance */
  SGN_Type_DATA_FRAME,

  /* \ref CSMLObject instance */
  SGN_Type_CSMLObject,

  /* \ref SGN_DATA instance */
  SGN_Type_DATA,

  /*! [\ref KeyValues type specifier] The parsed pairs of keys and their
	values are added to the \ref KeyValues instance specified by the
	variable. */
  SGN_Type_KEY_VALUES_ADD,

  /*! [\ref KeyValues type specifier] This specifies the optional 
	integer value.  If the value is not given,  the default
    value is set to the variable.  The fourth field of \ref KeyValueSet 
	is required to specify the pointer to the default integer value. */
  SGN_Type_INT_OPT,

  /*! [\ref KeyValues type specifier] The value is added to the the
	variable, instead of replacing it. */
  SGN_Type_INT_ADD,

  /*! [\ref SGN_KeyValues type specifier] This specifies the optional
	string value.  If the value is not given, the dfault value is
	set to the variable.  The fourth field of \ref SGN_KeyValueSet
	is required to specify the pointer to the dfault string (char poitner) value. */
  SGN_Type_STRING_OPT,
  
  /*! [\ref KeyValues type specifier] */
  SGN_Type_STRING_ADD,

  /*! [\ref KeyValues type specifier] */
  SGN_Type_ON_OFF,

  /*! [\ref KeyValues type specifier] If specified, integer value one
	(1) is set to the variable. */
  SGN_Type_ON,

  /*! [\ref KeyValues type specifier] If specified, integer value zero
	(0) is set to the variable. */
  SGN_Type_OFF,

  /*! [\ref KeyValues type specifier] The elements need to be
	split by ':' if these are given by a string.  The elements
	are added to the array specified by the variable.
	If the variable is NULL, then the new \ref IntArray instance
	is generated and is set to the variable with the elements. */
  SGN_Type_INT_ARRAY_ADD,

  /*! [\ref KeyValues type specifier] Same as
	\ref TYPE_INT_ARRAY_ADD except that this expects ',' (comma) as
	a delimitor. */
  SGN_Type_INT_ARRAY_ADD2,

  /*! [\ref KeyValues type specifier] Values for 
	\ref SGN_DoubleArray instance.
	The elements need to be
	split by ':' if these are given by a string.  The elements
	are added to the array specified by the variable.
	If the variable is NULL, then the new \ref SGN_DoubleArray instance
	is generated and is set to the variable with the elements. */
  SGN_Type_DOUBLE_ARRAY_ADD,

  /*! [\ref KeyValues type specifier] Same as
	\ref TYPE_DOUBLE_ARRAY_ADD except that this expects ',' (comma) as
	a delimitor. */
  /*! [\ref KeyValues type specifier] */
  SGN_Type_DOUBLE_ARRAY_ADD2,

  /*! [\ref KeyValues type specifier] The elements need to be
	split by ':' if these are given by a string.  The elements
	are added to the array specified by the variable.
	If the variable is NULL, then the new \ref StrArray instance
	is generated
	and is set to the variable with the elements. */
  SGN_Type_STR_ARRAY_ADD,

  /*! [\ref KeyValues type specifier] Same as
	\ref SGN_Type_STR_ARRAY_ADD except that this expects ',' (comma) as
	a delimitor. */
  SGN_Type_STR_ARRAY_ADD2,

  /*! [\ref KeyValues type specifier] Function pointer.  The type of
	the function needs to be <TT>void (*) ()</TT>, that is, a function
	that takes no arguments and return nothing. */
  SGN_Type_FUNC,

  /*! [\ref KeyValues type specifier] Function pointer.  The type of
	the function needs to be <TT>void (*) (const char *, const char *)</TT>.
	The function will receive the key and the value as a string. **/
  SGN_Type_FUNC2,

  /*! [\ref KeyValues type specifier] Function pointer.  The type of
	the function needs to be 
	<TT>void (*) (const char *, const char *, void *)</TT>.
	The first two arguments are the same as that of \ref TYPE_FUNC2.
	The value <TT>arg3</TT> of \ref KeyValueSet is passed to the 
	third argument.
  */
  SGN_Type_FUNC3,

  /*! */
  SGN_Type_OTHER,

  /*! */
  SGN_Type_StrMatrix,

  /*! [\ref KeyValues type specifier] This accepts a 'X-Y:Z' style
	format string and parse it into an IntArray objects. */
  SGN_Type_INT_REGION,

  /*! */
  SGN_Type_GRAPH,
  
  /*! */
  SGN_Type_ADJ_GRAPH,

  /*! */
  SGN_Type_COMPLETE_GRAPH,

  /*! [\ref KeyValues type specifier] This tells that there is
	nothing to do in parsing by \ref KeyValues functions.
   */
  SGN_Type_NONE,

  /*! Error value */
  SGN_Type_ERROR,
} SGN_Type;

typedef enum {
  SGN_OP_EQ,  
  SGN_OP_LT,
  SGN_OP_LE,
  SGN_OP_GT,
  SGN_OP_GE,
  SGN_OP_ERROR,
} SGN_Type_Operator;

typedef struct {
  int a;
  int b;
} SGN_IntPair;

typedef struct {
  double a;
  double b;
} SGN_DoublePair;

const char * SGN_Type_name(SGN_Type type);
SGN_Type     SGN_Type_get(const char * str);
void *       SGN_Type_clone(const void * obj, SGN_Type type);
void         SGN_Type_fprint_value(const void * obj, SGN_Type type, FILE * fp);

void *       SGN_Type_int_new();
void *       SGN_Type_int_clone(const void * ptr);
void *       SGN_Type_int_new_str(const char * str);
int          SGN_Type_int_cmp(const void * x, const void * y);
void         SGN_Type_int_fprint(const void * ptr, FILE * fp);

void *       SGN_Type_float_new();

void *       SGN_Type_double_new();
void *       SGN_Type_double_clone(const void * ptr);
void *       SGN_Type_double_new_str(const char * str);
int          SGN_Type_double_cmp(const void * x, const void * y);
void         SGN_Type_double_fprint(const void * ptr, FILE * fp);

void *       SGN_Type_string_new();
void *       SGN_Type_string_clone(const void * ptr);
void *       SGN_Type_string_new_str(const char * str);
int          SGN_Type_string_cmp(const void * x, const void * y);
void         SGN_Type_string_fprint(const void * ptr, FILE * fp);

void         SGN_Type_NULL_delete(void * ptr);


void *       SGN_Type_new_str(const char * str, SGN_Type type);
void *       SGN_Type_duplicate(const void * obj, SGN_Type type);
void         SGN_Type_delete(void * obj, SGN_Type type);
void *       SGN_Type_from_str(SGN_Type type, const char * str, void * ptr);
int          SGN_Type_compare_op(SGN_Type type, SGN_Type_Operator op, const void * x, const void * y);
SGN_Type_Operator SGN_Type_get_compare_op(const char * str);


#endif /* __SGN_TYPE_H */

/* End of file.  Copyright (C) 2009-2012 SiGN Project Members */
