/*
  math/OTO_PK.c : Online Topological Ordering by PK Algorithm.

  Copyright (C) 2009 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.
  
  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

/*!\defgroup OTO_PK
 *\brief Online Topological Ordering by PK Algorithm.
 *
 * The algorithm is written in the paper:
 *  D. J. Pearch and P. H. J. Kelly (2006). A dynamic topological
 *  sort algorithm for directed acyclic graphs.
 *  <I>ACM Journal of Experimental Algorithms</I>, <B>11</B>, 1-24.
 *
 *\sa PCGraph
 *\author Yoshinori Tamada &lt;tamada at is.s.u-tokyo.ac.jp&gt;
 */

#include "../signcom.h"
#include "SGN_PCGraph.h"
#include "OTO_PK.h"
#include <stdlib.h>
#include <string.h>
#include <strings.h>

/*!\brief Generates a new \ref OTO_PK instance for
 *   online topological ordering.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup OTO_PK */
OTO_PK * OTO_PK_new(int size){
  OTO_PK * this = malloc(sizeof(OTO_PK));
  int i;
  this->ord = malloc(sizeof(int) * size * 2);
  this->visited = this->ord + size;
  this->Rf = malloc(sizeof(OTO_PK_sort_t) * size * 2);
  this->Rb = this->Rf + size;
  
  for(i = 0; i < size; i++)	this->ord[i] = i;
  bzero(this->visited, sizeof(int) * size);

  this->size = size;

  return this;
}
/* OTO_PK_new() ------------------------------------------------------------ */

/*!\brief Generates a new \ref OTO_PK instance with a \ref SGN_PCGraph.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup OTO_PK */
OTO_PK * OTO_PK_new_pcgraph(SGN_PCGraph * g){
  OTO_PK * oto = OTO_PK_new(SGN_PCGraph_num_nodes(g));
  size_t i, id;

  id = SGN_PCGraph_first_edge(g);
  for(i = 0; i < SGN_PCGraph_num_edges(g); i++){
	int pa, ch;

	SGN_PCGraph_get_edge(g, id, &pa, &ch);
	if(OTO_PK_check_cyclic_reorder(oto, pa, ch, g) == 1){
	  OTO_PK_delete(oto);
	  return NULL;
	}
	id = SGN_PCGraph_next_edge(g, id);
  }

  return oto;
}
/* OTO_PK_new_pcgraph() ---------------------------------------------------- */

void OTO_PK_copy(const OTO_PK * this, OTO_PK * oto){
  if(this->size != oto->size){
	fprintf(SIGN_ERR, "FATAL ERROR: OTO_PK_copy(): size inconsisitent.\n");
	exit(1);
  }

  memcpy(oto->ord, this->ord, sizeof(int) * this->size);
}
/* OTO_PK_copy() ----------------------------------------------------------- */

/*!\brief Deletes the \ref OTO_PK instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup OTO_PK */
void OTO_PK_delete(void * this){
  free(((OTO_PK *) this)->ord);
  free(((OTO_PK *) this)->Rf);
  free(this);
}
/* OTO_PK_delete() --------------------------------------------------------- */

/*
  Internal function that implements dfs-f(n).
 */
int OTO_PK_dfs_f(OTO_PK * this, int n, SGN_PCGraph * g, int ub){
  size_t id;
  int w;
  //size_t prev_id;

  //fprintf(SIGN_ERR, "dfs_f(n=%d,ub=%d)\n", n, ub);
  //SGN_PCGraph_print_children(g, n, SIGN_ERR);

  this->visited[n] = 1;
  this->Rf[this->Rfp].id = n;
  this->Rf[this->Rfp++].ord = this->ord[n];

  id = SGN_PCGraph_first_child_id(g, n);
  while(id != SGN_PCGraph_NULL){
	//prev_id = id;
	id = SGN_PCGraph_child_next(g, id, &w);
	if(w == -1) continue;

	/*
	fprintf(SIGN_ERR, "[id=%zu]  n=%d -> w=%d (ord(w) = %d) [next_id=%zu]\n",
			prev_id, n, w, this->ord[w], id);
	*/

	if(this->ord[w] == ub) return 1;

	if(this->visited[w] == 0 && this->ord[w] < ub){
	  if(OTO_PK_dfs_f(this, w, g, ub) == 1) return 1;
	}
  }/* while */

  return 0;
}
/* OTO_PK_dfs_f() --------------------------------------------------------- */

/*
  Inernal funtion that implements dfs-b(n).
*/
int OTO_PK_dfs_b(OTO_PK * this, int n, SGN_PCGraph * g, int lb){
  size_t id;
  int w;

  //fprintf(SIGN_ERR, "dfs_b(n=%d)\n", n);

  this->visited[n] = 1;
  this->Rb[this->Rbp].id = n;
  this->Rb[this->Rbp++].ord = this->ord[n];
  
  id = SGN_PCGraph_first_parent_id(g, n);
  while(id != SGN_PCGraph_NULL){
	id = SGN_PCGraph_parent_next(g, id, &w);
	if(w == -1) continue;

	//fprintf(SIGN_ERR, "  n=%d <- w=%d (ord(w) = %d)\n", n, w, this->ord[w]);

	if(this->ord[w] == lb) return 1;
	if(this->visited[w] == 0 && lb < this->ord[w]){
	  if(OTO_PK_dfs_b(this, w, g, lb) == 1) return 1;
	}
  }
  return 0;
}
/* OTO_PK_df_b() ---------------------------------------------------------- */

int OTO_PK_cmp(const void * a, const void * b){
  return  ((const OTO_PK_sort_t *) a)->ord - ((const OTO_PK_sort_t *) b)->ord;
}

void OTO_PK_merge(OTO_PK_sort_t * A, int A_len, 
				  OTO_PK_sort_t * B, int B_len, int * C){
  int c = 0;
  int a = 0;
  int b = 0;
  while(a < A_len && b < B_len){
	if(A[a].id <= B[b].id){
	  C[c++] = A[a++].id;
	} else {
	  C[c++] = B[b++].id;
	}
  }
  while(a < A_len) C[c++] = A[a++].id;
  while(b < B_len) C[c++] = B[b++].id;
}
/* OTO_PK_merge() ---------------------------------------------------------- */

void OTO_PK_reorder(OTO_PK * this){
  int i;
  int Lp = 0;
  int L[this->size];
  int R[this->size];

  qsort(this->Rb, this->Rbp, sizeof(OTO_PK_sort_t), OTO_PK_cmp);
  qsort(this->Rf, this->Rfp, sizeof(OTO_PK_sort_t), OTO_PK_cmp);

  for(i = 0; i < this->Rbp; i++){
	int w = this->Rb[i].id;
	this->Rb[i].id = this->ord[w];
	//this->visited[w] = 0;
	L[Lp++] = w;
  }

  for(i = 0; i < this->Rfp; i++){
	int w = this->Rf[i].id;
	this->Rf[i].id = this->ord[w];
	//this->visited[w] = 0;
	L[Lp++] = w;
  }

  OTO_PK_merge(this->Rf, this->Rfp, this->Rb, this->Rbp, R);

  for(i = 0; i < Lp; i++) this->ord[L[i]] = R[i];
}
/* OTO_PK_reorder() ------------------------------------------------------- */

/*!\brief Adds an edge with checking if it makes the graph cyclic.
 *
 *\param this   pointer to the \ref OTO_PK instance.
 *\param x
 *\param y
 *\param g  pointer to the \ref SGN_PCGraph instance.
 *\return 0 if successful, or 1 if a cycle becomes with \a x - \a y edge.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup OTO_PK */
int OTO_PK_add_edge(OTO_PK * this, int x, int y, SGN_PCGraph * g){
  int lb = this->ord[y];
  int ub = this->ord[x];

  /*
  fprintf(SIGN_ERR, "add_edge(x,y): ord(x=%d)=%d (ub) vs ord(y=%d)=%d (lb)\n",
		  x, this->ord[x], y, this->ord[y]);
  */
  
  if(lb < ub){
	//fprintf(SIGN_ERR, "lb < ub: rquires reordering.\n");
	this->Rfp = 0; this->Rbp = 0;
	bzero(this->visited, sizeof(int) * this->size);

	if(OTO_PK_dfs_f(this, y, g, ub) == 1) return 1;
	if(OTO_PK_dfs_b(this, x, g, lb) == 1) return 1;
	OTO_PK_reorder(this);
	SGN_PCGraph_add_edge(g, x, y);
	return 0;
  } else {
	/* x -> y does not break the topological order */
	SGN_PCGraph_add_edge(g, x, y);
	return 0;
  }
}
/* OTO_PK_add_edge() ------------------------------------------------------- */

/*!\brief Checkes if the specified edge makes the graph cyclic.
 *
 * This does not affect the current topological order.
 *
 *\return 1 if the edges makes the graph cyclic, or 0 otherwise.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup OTO_PK */
int OTO_PK_check_cyclic(OTO_PK * this, int x, int y, SGN_PCGraph * g){
  int lb = this->ord[y];
  int ub = this->ord[x];

  if(lb < ub){
	this->Rfp = 0; this->Rbp = 0;
	bzero(this->visited, sizeof(int) * this->size);

	if(OTO_PK_dfs_f(this, y, g, ub) == 1) return 1;
	if(OTO_PK_dfs_b(this, x, g, lb) == 1) return 1;
	/* Do not re-order */
	return 0;
  } else {
	return 0;
  }
}
/* OTO_PK_check_cyclic() --------------------------------------------------- */

/*!\brief Checkes if the specified edge makes the graph cyclic.
 *
 * This affects (changes) the current topological order.
 *
 *\return 1 if the edges makes the graph cyclic, or 0 otherwise.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup OTO_PK */
int OTO_PK_check_cyclic_reorder(OTO_PK * this, int x, int y, SGN_PCGraph * g){
  int lb = this->ord[y];
  int ub = this->ord[x];

  if(lb < ub){
	this->Rfp = 0; this->Rbp = 0;
	bzero(this->visited, sizeof(int) * this->size);

	if(OTO_PK_dfs_f(this, y, g, ub) == 1) return 1;
	if(OTO_PK_dfs_b(this, x, g, lb) == 1) return 1;
	OTO_PK_reorder(this);
	return 0;
  } else {
	return 0;
  }
}
/* OTO_PK_check_cyclic() --------------------------------------------------- */

void OTO_PK_print_order(OTO_PK * this, FILE * fp){
  int ordinv[this->size];
  int i;
  for(i = 0; i < this->size; i++){
	ordinv[this->ord[i]] = i;
  }
  fprintf(fp, "nodes: ");
  for(i = 0; i < this->size; i++){
	fprintf(fp, " %d", ordinv[i]);
  }
  fprintf(fp, "\n");
}
/* OTO_PK_print_order() ---------------------------------------------------- */

/* End of file.  Copyright (C) 2009  SiGN Project Members */
