/*
  math/SGN_AdjGraph.c : Graph structure by an adjacency matrix.

  Copyright (C) 2011,2013 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.
  
  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#include "../signcom.h"
#include "SGN_AdjGraph.h"

#include <string.h>
#include <stdlib.h>

void SGN_AdjGraph_check_type(const SGN_AdjGraph * g){
  const SGN_ObjectI * obj = (const SGN_ObjectI *) g;
  if(obj->type != SGN_Type_ADJ_GRAPH){
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_AdjGraph: invalid type object received.\n");
	exit(1);
  }
}
/* SGN_AdjGraph_check_type() =================================================== */

/*!\brief Generates a new empty SGN_AdjGraph instance.
 *
 * The generated graph has no nodes nor edges.
 *
 *\return Pointer to the SGN_AdjGraph instance.
 *
 *<div align="right"><a href="#top">Go to top</a></div>
 *\memberof SGN_AdjGraph */
SGN_AdjGraph * SGN_AdjGraph_new(){
  SGN_AdjGraph * g = malloc(sizeof(SGN_AdjGraph));

  g->nodes = 0;
  g->matrix = NULL;

  SGN_ObjectI * obj = (SGN_ObjectI *) g;
  obj->type = SGN_Type_ADJ_GRAPH;
  obj->delete = SGN_AdjGraph_delete;
  obj->clone = SGN_AdjGraph_clone;

  /* Sets the implemented methods. */
  SGN_GraphI * gg = (SGN_GraphI *) g;
  gg->add_edge = SGN_AdjGraph_add_edge;
  gg->check_edge = SGN_AdjGraph_check_edge;
  gg->remove_edge = SGN_AdjGraph_remove_edge;
  gg->print_name = SGN_AdjGraph_print_name;
  gg->print_name2 = SGN_AdjGraph_print_name2;

  return g;
}
/* SGN_AdjGraph_new() ========================================================== */

/*!\brief Generates a new empty SGN_AdjGraph instance.
 *
 *
 *\param nodes Number of nodes in the graph.
 *\return Pointer to the SGN_AdjGraph instance.
 *
 *<div align="right"><a href="#top">Go to top</a></div>
 *\memberof SGN_AdjGraph */
SGN_AdjGraph * SGN_AdjGraph_news(size_t nodes){
  SGN_AdjGraph * g = SGN_AdjGraph_new();

  g->nodes = nodes;
  g->matrix = malloc(sizeof(char) * nodes * nodes);

  if(g->matrix == NULL){
	fprintf(SIGN_ERR, "WARNING: SGN_AdjGraph_news(): failed to allocate memory.\n");
	free(g);
	return NULL;
  } else {

	memset(g->matrix, 0, sizeof(char) * nodes * nodes);
	
	return g;
  }
}
/* SGN_AdjGraph_news() ========================================================= */

SGN_ObjectI * SGN_AdjGraph_clone(const SGN_ObjectI * obj){
  const SGN_AdjGraph * g = (const SGN_AdjGraph *) obj;
  SGN_AdjGraph_check_type(g);

  SGN_AdjGraph * new_g = SGN_AdjGraph_news(g->nodes);
  memcpy(new_g->matrix, g->matrix, sizeof(char) * g->nodes * g->nodes);

  return (SGN_ObjectI *) new_g;
}
/* SGN_AdjGraph_clone() ======================================================== */

/*!\brief Deletes the SGN_AdjGraph instance.
 *
 *\param obj Pointer to the SGN_AdjGraph insance to be deleted.
 *               Need to be casted to SGN_ObjectI.
 *
 *<div align="right"><a href="#top">Go to top</a></div>
 *\memberof SGN_AdjGraph */
void SGN_AdjGraph_delete(SGN_ObjectI * obj){
  SGN_AdjGraph * g = (SGN_AdjGraph *) obj;
  SGN_AdjGraph_check_type(g);
  
  free(g->matrix);
  free(g);
}
/* SGN_AdjGraph_delete() ======================================================= */

/*!\brief Adds an edge.
 *
 * Implementation of SGN_GraphI::add_edge().
 *
 *<div align="right"><a href="#top">Go to top</a></div>
 *\memberof SGN_AdjGraph */
void SGN_AdjGraph_add_edge(SGN_GraphI * gg, int u, int v){
  SGN_AdjGraph * g = (SGN_AdjGraph *) gg;
  SGN_AdjGraph_check_type(g);

  if(u >= g->nodes || v >= g->nodes) return;

  /* Set (u,v)-edge. */
  g->matrix[u + v * g->nodes] = 1;

}
/* SGN_AdjGraph_add_edge() ===================================================== */

/*!\brief Checks if the edge exists.
 *
 * Implementation of SGN_GraphI::check_edge().
 *
 *<div align="right"><a href="#top">Go to top</a></div>
 *\memberof SGN_AdjGraph */
int SGN_AdjGraph_check_edge(const SGN_GraphI * gg, int u, int v){
  const SGN_AdjGraph * g = (const SGN_AdjGraph *) gg;
  SGN_AdjGraph_check_type(g);

  if(u >= g->nodes || v >= g->nodes) return 0;
  
  return (g->matrix[u + v * g->nodes] != 0);
}
/* SGN_AdjGraph_check_edge() =================================================== */

/*!\brief Removes an edge.
 *
 * Implementation of SGN_GraphI::remove_edge().
 *
 *<div align="right"><a href="#top">Go to top</a></div>
 *\memberof SGN_AdjGraph */
void SGN_AdjGraph_remove_edge(SGN_GraphI * gg, int u, int v){
  SGN_AdjGraph * g = (SGN_AdjGraph *) gg;
  SGN_AdjGraph_check_type(g);

  if(u >= g->nodes || v >= g->nodes) return;

  /* Set (u,v)-edge. */
  g->matrix[u + v * g->nodes] = 0;
}
/* SGN_AdjGraph_remove_edge() ================================================== */

/*!\brief Makes the graph complete.
 *
 *
 *
 *<div align="right"><a href="#top">Go to top</a></div>
 *\memberof SGN_AdjGraph */
void SGN_AdjGraph_complete(SGN_AdjGraph * g){
  memset(g->matrix, 1, sizeof(char) * g->nodes * g->nodes);
}
/* SGN_AdjGraph_complete() ===================================================== */

/*!\brief
 *<div align="right"><a href="#top">Go to top</a></div>
 *\memberof SGN_AdjGraph */
void SGN_AdjGraph_print_name(const SGN_GraphI * gg, FILE * fp, const SGN_StrArray * names){
  const SGN_AdjGraph * g = (const SGN_AdjGraph *) gg;
  SGN_AdjGraph_check_type(g);

  int i, j;
  for(i = 0; i < g->nodes; i++){
	fprintf(fp, "%s(%d) :", SGN_StrArray_get(names, i), i);
	for(j = 0; j < g->nodes; j++){
	  if(SGN_AdjGraph_check_edge(gg, j, i)){
		fprintf(fp, " %s(%d)", SGN_StrArray_get(names, j), j);
	  }
	}
	fprintf(fp, "\n");
  }
}
/* SGN_AdjGraph_print_name() =================================================== */

/*!\brief
 *<div align="right"><a href="#top">Go to top</a></div>
 *\memberof SGN_AdjGraph */
void SGN_AdjGraph_print_name2(const SGN_GraphI * gg, FILE * fp, const SGN_StrArray * names){
  const SGN_AdjGraph * g = (const SGN_AdjGraph *) gg;
  SGN_AdjGraph_check_type(g);

  int i, j;
  for(i = 0; i < g->nodes; i++){
	fprintf(fp, "%s(%d) :", SGN_StrArray_get(names, i), i);
	for(j = 0; j < g->nodes; j++){
	  if(SGN_AdjGraph_check_edge(gg, i, j)){
		fprintf(fp, " %s(%d)", SGN_StrArray_get(names, j), j);
	  }
	}
	fprintf(fp, "\n");
  }
}
/* SGN_AdjGraph_print_name() =================================================== */

/*!\brief Casts the SGN_AdjGraph instance to SGN_GraphI.
 *
 *
 *\memberof SGN_AdjGraph */
SGN_GraphI * SGN_AdjGraph_to_GraphI(SGN_AdjGraph * g){
  return (SGN_GraphI *) g;
}
/* SGN_AdjGraph_to_GraphI() ==================================================== */

/* End of file. Copyright (C) 2011,2013  SiGN Project Members. */


  
