/*
  math/Graph.c : Graph module for efficient manipulation

  Copyright (C) 2008 - 2011 SiGN Project Members
  
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.
  
  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/


#include "SGN_Graph.h"

#include "../lang/SGN_Object.h"
#include "../util/SGN_Memory.h"
#include "../signcom.h"
#include "../util/SGN_StrArray.h"

#include <strings.h>

void SGN_Graph_check_type(const SGN_Graph * g){
  const SGN_ObjectI * obj = (const SGN_ObjectI *) g;
  if(obj->type != SGN_Type_GRAPH){
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_Graph: invalid type object received.\n");
	exit(1);
  }
}
/* SGN_Graph_check_type() ====================================================== */

/*!\brief Generates the new \ref SGN_Graph instance.
 *
 * <div align="right"><A href="#top">Go to top</A></div>
 *\memberof SGN_Graph */
SGN_Graph * SGN_Graph_new(size_t nodes){
  SGN_Graph * g = (SGN_Graph *) malloc(sizeof(SGN_Graph));
  if(g == NULL) return NULL;

  g->nodes = nodes;
  g->parents = SGN_IntArray_new_ar(nodes, 1);
  if(g->parents == NULL){
	free(g);
	return NULL;
  }

  SGN_ObjectI * obj = (SGN_ObjectI *) g;
  obj->type = SGN_Type_GRAPH;
  obj->delete = SGN_Graph_deleteI;

  SGN_GraphI * gg = (SGN_GraphI *) g;
  gg->num_nodes = SGN_Graph_num_nodesI;
  gg->check_edge = SGN_Graph_check_edgeI;
  gg->print_name = SGN_Graph_print_nameI;
  gg->print_name2 = SGN_Graph_print_name2I;
  
  return g;
}
/* SGN_Graph_new() ------------------------------------------------------------- */

/*!\brief Deletes the \ref SGN_Graph instance.
 *
 * <div align="right"><A href="#top">Go to top</A></div>
 *\memberof SGN_Graph */
void SGN_Graph_delete(void * this){
  if(this == NULL) return;
  SGN_IntArray_delete_ar(((SGN_Graph *) this)->parents, ((SGN_Graph *) this)->nodes);
  free(this);
}
/* SGN_Graph_delete() ---------------------------------------------------------- */

void SGN_Graph_deleteI(SGN_ObjectI * this){
  SGN_Graph_check_type((const SGN_Graph *) this);
  SGN_Graph_delete(this);
}
/* SGN_Graph_deleteI() ========================================================= */

/*!\brief Returns the number of nodes in a \ref SGN_Graph.
 *
 *\return the number of nodes in the \ref SGN_Graph instance.
 *
 * <div align="right"><A href="#top">Go to top</A></div>
 *\memberof SGN_Graph */
size_t SGN_Graph_num_nodes(const SGN_Graph * g){
  return g->nodes;
}

int SGN_Graph_num_nodesI(const SGN_GraphI * g){
  SGN_Graph_check_type((const SGN_Graph *) g);
  return (int) ((const SGN_Graph *)g)->nodes;
}

size_t SGN_Graph_num_edges(const SGN_Graph * g){
  size_t i, sum = 0;
  for(i = 0; i < g->nodes; i++){
	sum += SGN_IntArray_size(g->parents + i);
  }
  return sum;
}

/*!\brief Returns the set of parents of the specified node.
 *
 * Changing the contents of the array causes the changing of
 * the graph structure.
 *
 *\param this   pointer to the \ref SGN_Graph intance.
 *\param u   index of the node.
 *
 * <div align="right"><A href="#top">Go to top</A></div>
 *\memberof SGN_Graph */
SGN_IntArray * SGN_Graph_parents(const SGN_Graph * this, int u){
  return this->parents + u;
}
/* SGN_Graph_parents() --------------------------------------------------------- */

/*!\brief Returns the number of parents of the specified node.
 *
 *\param this  pointer to the \ref SGN_Graph instance.
 *\param u  index of the node.
 *
 * <div align="right"><A href="#top">Go to top</A></div>
 *\memberof SGN_Graph */
int SGN_Graph_num_parents(const SGN_Graph * this, int u){
  return SGN_IntArray_size(this->parents + u);
}
/* SGN_Graph_num_parents() ----------------------------------------------------- */

SGN_IntArray * SGN_Graph_gen_children(const SGN_Graph * this, int u){
  SGN_IntArray * ret = SGN_IntArray_new();
  size_t j;
  for(j = 0; j < this->nodes; j++){
	int r = SGN_IntArray_find(this->parents + j, u);
	if(r != SGN_IntArray_ERROR_VALUE){
	  SGN_IntArray_add(ret, j);
	}
  }
  return ret;
}
/* SGN_Graph_gen_children() ---------------------------------------------------- */

/*!\brief Sets children for all nodes into SGN_IntArrays.
 *
 *
 * Note that argument \a ar needs to be a pointer to the consecutive
 * \ref SGN_IntArray instances of size <I>n</I>, where <I>n</I> is the
 * number of nodes in the graph.  Such instances can be obtained by
 * SGN_IntArray_new_ar().
 * The given arrays do not need to be cleared.  This automatically
 * clears all elements in the arrays by SGN_IntArray_clear().
 *
 *\param this  pointer to the \ref SGN_Graph instance.
 *\param ar pointer to the <I>n</I> consecutive \ref SGN_IntArray instances
 *         where <I>n</I> is the number of nodes in the graph.
 *
 * <div align="right"><A href="#top">Go to top</A></div>
 *\memberof SGN_Graph */
void SGN_Graph_set_children_all(const SGN_Graph * this, SGN_IntArray * ar){
  size_t j, k;
  for(j = 0; j < this->nodes; j++) SGN_IntArray_clear(ar + j);
  for(j = 0; j < this->nodes; j++){
	for(k = 0; k < SGN_IntArray_size(this->parents + j); k++){
	  int jk = SGN_IntArray_get(this->parents + j, k);
	  SGN_IntArray_add(ar + jk, j);
	}
  }
}
/* SGN_Graph_set_children_all() ------------------------------------------------ */

/*!\brief Adds an edge to the graph.
 *
 * This does not check whether \a u - \a v edge exsits already
 * in the graph.
 *
 *\param this
 *\param u
 *\param v
 *
 * <div align="right"><A href="#top">Go to top</A></div>
 *\memberof SGN_Graph */
void SGN_Graph_add_edge(SGN_Graph * this, int u, int v){
  SGN_IntArray_add(this->parents + v, u);
}
/* SGN_Graph_add_edge() -------------------------------------------------------- */

void SGN_Graph_add_edge_nr(SGN_Graph * g, int u, int v){
  size_t i = SGN_IntArray_find(g->parents + v, u);
  if(i == SGN_IntArray_ERROR_VALUE){
	SGN_IntArray_add(g->parents + v, u);
  }
}

/*!\brief Collects all paths by depth-first search.
 *
 * <div align="right"><A href="#top">Go to top</A></div>
 *\memberof SGN_Graph */
SGN_Graph * SGN_Graph_collect_paths(const SGN_Graph * g,
							const SGN_IntArray * starts, const SGN_IntArray * ends,
							int length, int exact){
  int flag[g->nodes];
  int back[g->nodes];
  int targets[g->nodes];
  int i, uu;
  int depth = 0;
  SGN_Graph * rg = SGN_Graph_new(g->nodes);
  int c = 0;

  /* initizalize 'targets' */
  bzero(targets, sizeof(int) * g->nodes);
  for(i = 0; i < SGN_IntArray_size(starts); i++){
	targets[SGN_IntArray_get(starts, i)] = 1;
  }

  for(uu = 0; uu < SGN_IntArray_size(ends); uu++){
	/* start point */
	int u = SGN_IntArray_get(ends, uu);

	fprintf(stderr, "SGN_Graph: %d / %zu start.\n", uu + 1, SGN_IntArray_size(ends));
	depth = 0;

	/* initialize 'flag' */
	for(i = 0; i < g->nodes; i++) flag[i] = -1;

	/* ================= */
	/*  Start searching  */
	/* ================= */
	/*  u represents the current position in the network. */
	/*  back[u] : where from */
	/*  flag[u] : whether visited or parent index to visit */
	back[u] = -1;
	while(u != -1){
	  flag[u]++;

	  /* check u is involved in the targets */
	  if(targets[u] && flag[u] == 0){
		/* found ! */

		if(exact == 0 || (exact == 1 && length == depth)){
		  int x = back[u]; int y = u;

		  c++;
		  
		  /* add */
		  while(x != -1){
			SGN_Graph_add_edge_nr(rg, y, x);
			y = x;
			x = back[x];
		  }
		}
	  }/* if(targets[u]) */

	  if(depth == length){
		flag[u] = -1;
		u = back[u];
		depth--;
		continue;
	  }

	  if(flag[u] < SGN_IntArray_size(g->parents + u)){
		int nu = SGN_IntArray_get(g->parents + u, flag[u]);
		if(flag[nu] < 0){
		  /* not visited. go next */
		  back[nu] = u;
		  u = nu;
		  depth++;
		}
	  } else {
		/* visited all */
		/* go back */
		flag[u] = -1;
		u = back[u];
		depth--;
	  }
	}/* while(1) */
  }
  fprintf(stderr, "Found %d paths.\n", c);
  return rg;
}
/* SGN_Graph_collect_paths() --------------------------------------------------- */

/*!\brief Checks if the specified edge exists in the graph.
 *\param this  pointer to the \ref SGN_Graph instance. NULL acceptable.
 *                  If so, this always returns 0.
 *\param u   index of starting node of the edge.
 *\param v   index of ending node of the edge.
 *\return    1 if the edge exists, or 0 otherwise.
 *
 * <div align="right"><A href="#top">Go to top</A></div>
 *\memberof SGN_Graph */
int SGN_Graph_check_edge(const SGN_Graph * this, int u, int v){
  
  int r;

  if(this == NULL) {
	return 0;
  } else {

	r = SGN_IntArray_find(this->parents + v, u);
	if(r == SGN_IntArray_ERROR_VALUE){
	  return 0;
	} else {
	  return 1;
	}
  }
}
/* SGN_Graph_check_edge() ------------------------------------------------------ */

int SGN_Graph_check_edgeI(const SGN_GraphI * g, int u, int v){
  const SGN_Graph * this = (const SGN_Graph *) g;
  SGN_Graph_check_type(this);
  
  int r;

  if(this == NULL) {
	return 0;
  } else {

	r = SGN_IntArray_find(this->parents + v, u);
	if(r == SGN_IntArray_ERROR_VALUE){
	  return 0;
	} else {
	  return 1;
	}
  }
}
/* SGN_Graph_check_edgeI() ----------------------------------------------------- */

/*!\brief Checks if the graph becomes cyclic by adding the edge.
 *
 * This checks whether the graph becomes cyclic if \a u - \a v edge
 * is added.
 *
 *\param this   pointer to the \ref SGN_Graph instance.
 *\param u   index of the starting node of the edge
 *\param v   index of the ending node of the edge
 *\return 1 if the graph becomes cyclic with \a u - \a v edge, or 0
 *  otherwise.
 *
 * <div align="right"><A href="#top">Go to top</A></div>
 *\memberof SGN_Graph */
int SGN_Graph_check_cyclic(SGN_Graph * this, int u, int v){

  int list_a[this->nodes]; /* reached nodes added at the previous turn. */
  int list_a_size = 0;     /* size of 'list_a' */
  int flag_a[this->nodes]; /* flag of nodes in 'list_a'. */

  int flag_b[this->nodes]; /* flag of already reached nodes. */

  int list_c[this->nodes]; /* newly reached nodes in a turn. */
  int list_c_size = 0;     /* size of 'list_c' */
  int flag_c[this->nodes]; /* flag of nodes in 'list_c' */

  int i, k;

  bzero(flag_a, sizeof(int) * this->nodes);
  bzero(flag_b, sizeof(int) * this->nodes);
  bzero(flag_c, sizeof(int) * this->nodes);

  list_a[list_a_size++] = u; /* the first node */
  flag_a[u] = 1;

  while(1){
	/* Visit all nodes in 'list_a', to collect nodes reachable
	   from nodes in 'list_a'.  The collected nodes are stored in
	   'list_c'	*/
	for(i = 0; i < list_a_size; i++){
	  SGN_IntArray * parents = this->parents + list_a[i];

	  for(k = 0; k < SGN_IntArray_size(parents); k++){
		int n = SGN_IntArray_get(parents, k);
		
		/* check 'v' is in the newly reached nodes */
		if(n == v) return 1; /* The graph becomes cyclic. */

		if(!flag_a[n] && !flag_b[n] && !flag_c[n]){
		  list_c[list_c_size++] = n;
		  flag_c[n] = 1;
		}
	  }/* for k < |parents| */
	}/* for i < |list_a| */

	if(list_c_size == 0){
	  /* All the reachable nodes were visited. */
	  /* If 'v' is not found at this moment, the graph is acyclic with u->v. */
	  return 0; /* acyclic */
	}

	/* Add the ``front-line'' nodes ('list_a') to 'list_b', 
	   and clear 'list_a' & 'flag_a'. */
	for(i = 0; i < list_a_size; i++){
	  int n = list_a[i];
	  flag_b[n] = 1;
	  flag_a[n] = 0; /* clear flag_a */
	}
	list_a_size = 0;

	/* Copy nodes in 'list_c' into 'list_a',
	 and clear 'list_c' & 'flag_c'. */
	for(i = 0; i < list_c_size; i++){
	  int n = list_c[i];
	  list_a[list_a_size++] = n;
	  flag_a[n] = 1;
	  flag_c[n] = 0;
	}
	list_c_size = 0;
  }
  return 0;
}
/* SGN_Graph_check_cyclic() ---------------------------------------------------- */

int SGN_Graph_check_cyclic2(const SGN_Graph * this, int u, int v,
							 int j, const SGN_IntArray * j_parents){

  int list_a[this->nodes]; /* reached nodes added at the previous turn. */
  int list_a_size = 0;     /* size of 'list_a' */
  int flag_a[this->nodes]; /* flag of nodes in 'list_a'. */

  int flag_b[this->nodes]; /* flag of already reached nodes. */

  int list_c[this->nodes]; /* newly reached nodes in a turn. */
  int list_c_size = 0;     /* size of 'list_c' */
  int flag_c[this->nodes]; /* flag of nodes in 'list_c' */

  int i, k;

  memset(flag_a, 0, sizeof(int) * this->nodes);
  memset(flag_b, 0, sizeof(int) * this->nodes);
  memset(flag_c, 0, sizeof(int) * this->nodes);

  list_a[list_a_size++] = u; /* the first node */
  flag_a[u] = 1;

  while(1){
	/* Visit all nodes in 'list_a', to collect nodes reachable
	   from nodes in 'list_a'.  The collected nodes are stored in
	   'list_c'	*/
	for(i = 0; i < list_a_size; i++){
	  const SGN_IntArray * parents = list_a[i] != j ? this->parents + list_a[i] : j_parents;

	  for(k = 0; k < SGN_IntArray_size(parents); k++){
		int n = SGN_IntArray_get(parents, k);
		
		/* check 'v' is in the newly reached nodes */
		if(n == v) return 1; /* The graph becomes cyclic. */

		if(!flag_a[n] && !flag_b[n] && !flag_c[n]){
		  list_c[list_c_size++] = n;
		  flag_c[n] = 1;
		}
	  }/* for k < |parents| */
	}/* for i < |list_a| */

	if(list_c_size == 0){
	  /* All the reachable nodes were visited. */
	  /* If 'v' is not found at this moment, the graph is acyclic with u->v. */
	  return 0; /* acyclic */
	}

	/* Add the ``front-line'' nodes ('list_a') to 'list_b', 
	   and clear 'list_a' & 'flag_a'. */
	for(i = 0; i < list_a_size; i++){
	  int n = list_a[i];
	  flag_b[n] = 1;
	  flag_a[n] = 0; /* clear flag_a */
	}
	list_a_size = 0;

	/* Copy nodes in 'list_c' into 'list_a',
	 and clear 'list_c' & 'flag_c'. */
	for(i = 0; i < list_c_size; i++){
	  int n = list_c[i];
	  list_a[list_a_size++] = n;
	  flag_a[n] = 1;
	  flag_c[n] = 0;
	}
	list_c_size = 0;
  }
  return 0;
}
/* SGN_Graph_check_cyclic2() =============================================== */

void SGN_Graph_swap_parents(SGN_Graph * this, int u, SGN_IntArray * parents){
  SGN_IntArray_swap(this->parents + u, parents);
}
/* SGN_Graph_swap_parents() ---------------------------------------------------- */

void SGN_Graph_set_parents(SGN_Graph * this, int u, const SGN_IntArray * parents){
  SGN_IntArray_copy(parents, this->parents + u);
}
/* SGN_Graph_set_parents() ----------------------------------------------------- */

void SGN_Graph_remove_parent_index(SGN_Graph * this, int u, int index){
  SGN_IntArray_remove(this->parents + u, index);
}
/* SGN_Graph_delete_parent_index() --------------------------------------------- */

/*!\brief Removes all the edges.
 *
 * <div align="right"><A href="#top">Go to top</A></div>
 *\memberof SGN_Graph */
void SGN_Graph_clear(SGN_Graph * this){
  int i;
  for(i = 0; i < this->nodes; i++){
	SGN_IntArray_clear(this->parents + i);
  }
}
/* SGN_Graph_clear() ----------------------------------------------------------- */

/*!\brief Copies the \ref SGN_Graph structure into another \ref SGN_Graph.

 * <div align="right"><A href="#top">Go to top</A></div>
 *\memberof SGN_Graph */
void SGN_Graph_copy(const SGN_Graph * this, SGN_Graph * dst){
  int i;
  if(this->nodes != dst->nodes){
	fprintf(SIGN_ERR, 
			"FATAL ERROR: SGN_Graph_copy(): inconsistent numbers of nodes.\n");
	exit(1);
  }
  for(i = 0; i < this->nodes; i++){
	SGN_IntArray_copy(this->parents + i, dst->parents + i);
  }
}
/* SGN_Graph_copy() ------------------------------------------------------------ */

/*!\brief Copies the \ref SGN_Graph structure into another \ref SGN_Graph.

 * <div align="right"><A href="#top">Go to top</A></div>
 *\memberof SGN_Graph */
void SGN_Graph_copy_GraphI(const SGN_GraphI * g, SGN_Graph * dst){
  SGN_Graph * this = (SGN_Graph *) g;
  SGN_Graph_check_type(this);

  if(SGN_GraphI_num_nodes(g) != dst->nodes){
	fprintf(SIGN_ERR, 
			"FATAL ERROR: SGN_Graph_copy_GraphI(): inconsistent numbers of nodes.\n");
	exit(1);
  }

  if(SGN_Object_type((const SGN_Object *) g) == SGN_Type_GRAPH){
	SGN_Graph_copy(this, dst);
  } else {
	SGN_Graph_clear(dst);
	int n = SGN_GraphI_num_nodes(g);
	for(int i = 0; i < n; i++){
	  for(int j = 0; j < n; j++){
		if(SGN_GraphI_check_edge(g,i,j)){
		  SGN_Graph_add_edge(dst,i,j);
		}
	  }/* for j < n */
	}/* for i < n */
  }
}
/* SGN_Graph_copy_GraphI() ----------------------------------------------------- */

/*!\brief Prints the graph structure.
 *
 *<div align="right"><a href="#top">Go to top</a></div>
 *\memberof SGN_Graph */
void SGN_Graph_show(FILE * fp, const SGN_Graph * this){
  int i, j;
  for(i = 0; i < this->nodes; i++){
	fprintf(fp, "%d :", i);
	for(j = 0; j < SGN_IntArray_size(this->parents + i); j++){
	  fprintf(fp, " %d", SGN_IntArray_get(this->parents + i, j));
	}
	fprintf(fp, "\n");
  }
}
/* SGN_Graph_show() ------------------------------------------------------------ */

/*!\brief Prints the graph structure with the given node names.
 *
 *<div align="right"><a href="#top">Go to top</a></div>
 *\memberof SGN_Graph */
void SGN_Graph_print_name(const SGN_Graph * this, FILE * fp, const SGN_StrArray * names){
  int i, j;
  for(i = 0; i < this->nodes; i++){
	fprintf(fp, "%s(%d) :", SGN_StrArray_get(names, i), i);
	for(j = 0; j < SGN_IntArray_size(this->parents + i); j++){
	  int v = SGN_IntArray_get(this->parents + i, j);
	  fprintf(fp, " %s(%d)", SGN_StrArray_get(names, v), v);
	}
	fprintf(fp, "\n");
  }
}
/* SGN_Graph_print_name() ====================================================== */

/*!\brief Prints the graph structure with the given node names.
 *
 *<div align="right"><a href="#top">Go to top</a></div>
 *\memberof SGN_Graph */
void SGN_Graph_print_name2(const SGN_Graph * this, FILE * fp, const SGN_StrArray * names){
  int i, j;
  for(i = 0; i < this->nodes; i++){
	fprintf(fp, "%s(%d) :", SGN_StrArray_get(names, i), i);
	for(j = 0; j < this->nodes; j++){
	  if(SGN_Graph_check_edge(this, i, j)){
		fprintf(fp, " %s(%d)", SGN_StrArray_get(names, j), j);
	  }
	}
	fprintf(fp, "\n");
  }
}
/* SGN_Graph_print_name2() ===================================================== */

void SGN_Graph_print_nameI(const SGN_GraphI * g, FILE * fp, const SGN_StrArray * names){
  const SGN_Graph * this = (const SGN_Graph *) g;
  SGN_Graph_check_type(this);

  int i, j;
  for(i = 0; i < this->nodes; i++){
	fprintf(fp, "%s(%d) :", SGN_StrArray_get(names, i), i);
	for(j = 0; j < SGN_IntArray_size(this->parents + i); j++){
	  int v = SGN_IntArray_get(this->parents + i, j);
	  fprintf(fp, " %s(%d)", SGN_StrArray_get(names, v), v);
	}
	fprintf(fp, "\n");
  }
}
/* SGN_Graph_print_name() ====================================================== */

void SGN_Graph_print_name2I(const SGN_GraphI * g, FILE * fp, const SGN_StrArray * names){
  const SGN_Graph * this = (const SGN_Graph *) g;
  SGN_Graph_check_type(this);

  int i, j;
  for(i = 0; i < this->nodes; i++){
	fprintf(fp, "%s(%d) :", SGN_StrArray_get(names, i), i);
	for(j = 0; j < this->nodes; j++){
	  if(SGN_Graph_check_edge(this, i, j)){
		fprintf(fp, " %s(%d)", SGN_StrArray_get(names, j), j);
	  }
	}
	fprintf(fp, "\n");
  }
}
/* SGN_Graph_print_name() ====================================================== */

SGN_Graph * SGN_Graph_convert(const SGN_Graph * this, SGN_IntArray * table, size_t size){
  SGN_Graph * g = SGN_Graph_new(size);
  for(int j = 0; j < SGN_Graph_num_nodes(this); j++){
	int v = SGN_IntArray_get(table, j);

	SGN_IntArray * parents = SGN_Graph_parents(this, j);
	for(int i = 0; i < SGN_IntArray_size(parents); i++){
	  int u = SGN_IntArray_get(table, SGN_IntArray_get(parents, i));
	  SGN_Graph_add_edge(g, u, v);
	}
  }
  return g;
}
/* SGN_Graph_convert() ========================================================= */

/*!\brief Casts the SGN_Graph instance to SGN_GraphI.
 *
 * <div align="right"><A href="#top">Go to top</A></div>
 *\memberof SGN_Graph */
SGN_GraphI * SGN_Graph_to_GraphI(SGN_Graph * this){
  return  (SGN_GraphI *) this;
}
/* SGN_Graph_to_GraphI() ======================================================= */


#ifdef USE_MPI

/*!\brief Broadcasts the graph using MPI.
 *
 * This can be used only when \c USE_MPI is defined.
 *
 *
 * <div align="right"><A href="#top">Go to top</A></div>
 *\memberof SGN_Graph */
void SGN_Graph_Bcast(SGN_Graph ** this, int root, MPI_Comm comm){
  int rank;
  SGN_Graph * g = *this;

  MPI_Comm_rank(comm, &rank);
  if(rank != root){
	*this = (SGN_Graph *) malloc(sizeof(SGN_Graph));
	g = *this;
  }
  SGN_IntArray_MPI_Bcast_ar(&g->parents, &g->nodes, root, comm);
}
/* SGN_Graph_Bcast() ----------------------------------------------------------- */

#endif /* USE_MPI */

/* End of file.  Copyright (C) 2008  SiGN Project Members */
