/*
  sign/math/SGN_Math.c : Utility functions of mathematics.

  Copyright (C) 2008-2012 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

/*!\defgroup Math
 *\brief  Mathematical functions
 *
 *
 *\sa \ref Matrix
 *\author Yoshinori Tamada &lt;tamada at is.s.u-tokyo.ac.jp&gt;
  */
#include <stdlib.h>
#include "SGN_Math.h"

/*!\brief Returns the maximum value in a vector.
 *
 *\ingroup Math */
int SGN_Math_vmaxv(const size_t num, const int * vec){
  int ret = vec[0];
  size_t i;
  for(i = 1; i < num; i++){
	if(vec[i] > ret) ret = vec[i];
  }
  return ret;
}
/* SGN_Math_vmaxv() ------------------------------------------------------------ */

/*!\brief Searches the minimun and the maximum values in a vector.
 *
 * if 'num' is equal to 0, this does nothing.
 *
 *\ingroup Math */
void SGN_Math_minmax(const double * ar, const size_t num,
				 double * min, double * max){
  int lmin, lmax;
  if(num == 0) return;
  lmin = lmax = ar[0];

  for(size_t i = 1; i < num; i++){
	if(ar[i] < lmin) lmin = ar[i];
	if(lmax < ar[i]) lmax = ar[i];
  }/* for i < num */

  *min = lmin;
  *max = lmax;
}
/* SGN_Math_minmax() =========================================================== */

/*!\brief Calculates the sum of values in an array.
 *\ingroup Math */
double SGN_Math_sum(const double * ar, const int n){
  double ret = 0.0;
  int i;
  for(i = 0; i < n; i++) ret += ar[i];
  return ret;
}
/* SGN_Math_sum() -------------------------------------------------------------- */

/*!\brief Calculates the sum of values in an array.
 *\ingroup Math */
int SGN_Math_isum(const int * ar, const int n){
  int ret = 0;
  int i;
  for(i = 0; i < n; i++) ret += ar[i];
  return ret;
}
/* SGN_Math_isum() ============================================================= */

/*!\brief Calculates the sum of values in an array.
 *\ingroup Math */
int SGN_Math_isum2(const int * ar, const int n, const int inc){
  int ret = 0;
  int pos = 0;
  int i;
  for(i = 0; i < n; i++){
	ret += ar[pos];
	pos += inc;
  }
  return ret;
}
/* SGN_Math_isum2() ============================================================ */

/*!\brief Calculates the greatest common divisor.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup Math */
int SGN_Math_gcd(int x, int y){
  int t;
  while(y != 0){
	t = x % y; x = y; y = t;
  }
  return x;
}
/* SGN_Math_gcd() -------------------------------------------------------------- */


size_t SGN_Math_choose2(int n, int c){
  size_t ret = 1;
  int i;

  if(c > n - c) c = n - c;

  for(i = 1; i <= c; i++){
	int r, nn, mm;
	r = SGN_Math_gcd(n, i);
	nn = n / r;
	mm = i / r;
	ret /= mm;
	ret *= nn;
	n--;
  }
  return ret;
}
/* Match_choose() ---------------------------------------------------------- */

/*!\brief Calculates the number of combinations of \a n things taken \a k at
 *  a time (binomial coefficient).
 *
 * This is the C implementation of ACM Algorithm 160.
 *
 *\param n
 *\param k
 *\return  number of combinations of \a n things taken \a k at a time
 *            (binomical coefficient for \a n and \a k).
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup Math */
size_t SGN_Math_choose(int n, int k){
  int p = n - k, i;
  size_t r;
  if(k < p){
	p = k;
	k = n - p;
  }
  if(p == 0){
	return 1;
  }
  r = k + 1;
  for(i = 2; i <= p; i++){
	r = (r * (k + i)) / i;
  }
  return r;
}
/* SGN_Math_coose2() =========================================================== */

/*!\brief Generates the combination.
 *
 * To generate the first combination, set -1 for the first positions of \a ar.
 *
 *\param n
 *\param c
 *\param ar
 *\return 1 if the generated combination is the last one, 0 otherwise.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup Math */
int SGN_Math_combo(int n, int c, int * ar){
  int i, j;

  if(ar[0] == -1){
	for(i = 0; i < c; i++){
	  ar[i] = i;
	}
	return 0;
  }

  for(i = 0; i < c; i++){
	ar[c - i - 1]++; /* increment */

	if(ar[c - i - 1] != n - i){

	  /* update the lower digits */
	  for(j = 0; j < i; j++){
		ar[c - i + j] = ar[c - i - 1] + j + 1;
	  }
	  
	  return 0; /* OK */
	}/* if () */
  }/* for i */
  return 1; /* Finished */
}
/* SGN_Math_combo() ------------------------------------------------------------ */

/*!\brief Generates the index table for SGN_Math_combo_index().
 *
 * The table has to have size <TT>sizeof(size_t) * n * c</TT>.
 *
 *
 *\param n
 *\param c
 *\param table pointer to room for storing the index table.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup Math */
void SGN_Math_combo_index_table(int n, int c, size_t * table){
  int i, j;

  for(j = 0; j < n; j++) table[j] = j;
  
  for(i = 1; i < c; i++){
	size_t prev = 0;
	table[i * n] = prev;
	
	for(j = 1; j < n; j++){
	  prev = prev + table[(i - 1) * n + j];
	  table[i * n + j] = prev;
	}/* for j < n */
  }/* for i < c */
}
/* SGN_Math_combo_index_table() ------------------------------------------------ */

/*!\brief Calculates the index for the given combination.
 *
 * Note that this returns the reverse order (index) of the combination.
 * To obtain the normal index, calculate (SGN_Math_choose(n, c) - \t id) where
 * \t id is the reverse index returned by this function.
 *
 *\param ar sorted list (combination) of values.
 *\param table index table calculated by SGN_Math_combo_index_table().
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup Math */
size_t SGN_Math_combo_index(int n, int c, const int * ar, 
						const size_t * table, int tn){
  int i;
  size_t ret = 0;
  int n_c = n - c;
  
  for(i = 0; i < c; i++){
	int v = n_c - ar[i] + i; /* = n - ar[i] - 1 - (c - i - 1) */
	ret += table[(c - i - 1) * tn + v];
  }
  return ret;
}
/* SGN_Math_combo_index() ------------------------------------------------------ */

/*!\brief Calculates the index for the given combination after the conversion.
 *
 * This is the same as SGN_Math_combo_index() except that this converts values
 * in \a ar by the conversion table \a ar2.  So the i-th value of the combination
 * for indexing is <TT>ar2[ar[i]]</TT>.
 *
 *\param ar sorted list of values.
 *\param table index table calculated by SGN_Math_combo_index_table().
 *\param ar2  converstion table for \a ar.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup Math */
size_t SGN_Math_combo_index2(int n, int c, const int * ar,
						 const size_t * table, int tn, const int * ar2){
  int i;
  size_t ret = 0;
  int n_c = n - c;
  
  for(i = 0; i < c; i++){
	int v = n_c - ar2[ar[i]] + i; /* = n - ar[i] - 1 - (c - i - 1) */
	ret += table[(c - i - 1) * tn + v];
  }
  return ret;
}
/* SGN_Math_combo_index2() ----------------------------------------------------- */

/*!\brief Calculates the index for the given combination after the conversion.
 *
 * This is the same as SGN_Math_combo_index() except that this converts values
 * in \a ar by the conversion table \a ar2 and \a ar3.  So the i-th value of
 * the combination for indexing is <TT>ar2[ar3[ar[i]]]</TT>.
 *
 *\param ar sorted list of values.
 *\param table index table calculated by SGN_Math_combo_index_table().
 *\param ar2  converstion table for \a ar3.
 *\param ar3  converstion table for \a ar.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup Math */
size_t SGN_Math_combo_index3(int n, int c, const int * ar,
						 const size_t * table, int tn, const int * ar2, const int * ar3){
  int i;
  size_t ret = 0;
  int n_c = n - c;
  
  for(i = 0; i < c; i++){
	int v = n_c - ar2[ar3[ar[i]]] + i; /* = n - ar[i] - 1 - (c - i - 1) */
	ret += table[(c - i - 1) * tn + v];
  }
  return ret;
}
/* SGN_Math_combo_index2() ----------------------------------------------------- */

/*!\brief Obtain the combination corresponding to the given index.
 *
 * This is the inverse function of SGN_Math_combo_index().
 *
 *\param tn table size specified to SGN_Math_combo_index_table().
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup Math */
void SGN_Math_index_combo(int n, int c, size_t index, int * ar, 
					  const size_t * table, int tn){
  int i, j;
  int st = n - c;
  for(i = 0; i < c; i++){
	int pos = (c - i - 1) * tn;
	for(j = st; j >= 0; j--){
	  if(index >= table[pos + j]) {
		ar[i] = n - j - c + i;
		index -= table[pos + j];
		st = j;
		break;
	  }
	}/* for j */
  }/* for i */
}
/* SGN_Math_index_combo() ------------------------------------------------------ */

/*!\brief Obtain the combination corresponding to the given index.
 *
 * This is the inverse function of SGN_Math_combo_index().
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup Math */
void SGN_Math_index_combo2(int n, int c, size_t index, int * ar, 
					   const size_t * table, int tn, const int * ar2){
  int i, j;
  int st = n - c;
  for(i = 0; i < c; i++){
	int pos = (c - i - 1) * tn;
	for(j = st; j >= 0; j--){
	  if(index >= table[pos + j]) {
		ar[i] = ar2[n - j - c + i];
		index -= table[pos + j];
		st = j;
		break;
	  }
	}/* for j */
  }/* for i */
}
/* SGN_Math_index_combo2() ----------------------------------------------------- */

/* End of file.  Copyright (C) 2008-2012  SiGN Project Members */
