/*
  sign/math/SGN_Matrix.h : Matrix and LAPACK wrapper functions

  Copyright (C) 2008-2012 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

/*!\defgroup SGN_Matrix
 * \brief Matrix computation functions
 *
 * \code
 * #include <math/SGN_Matrix.h>
 * \endcode
 *
 * \section Macos
 *
 * The following macros are acceptable when compiling this module.
 *
 * - <TT>USE_MACOSX_VECLIB</TT>
 *  - Specifies to use Apple's Mac OS X vecLib in Accelerate Framework.
 *
 * - <TT>USE_SUNPERF64</TT>
 *  - Specifies to use Sun's performace library 64-bit version. 
 *      (Currently not supported)
 *
 * - <TT>USE_SUNPER32</TT>
 *  - Specifies to use Sun's performace library 32-bit version. 
 *      (Currently not supported)
 *
 * - <TT>USE_IBM_ESSL</TT>
 *  - Specifies to use IBM ESSL library. (Currently not supported)
 *
 * - <TT>USE_FUJITSU_FCC_SSL</TT>
 *  - Specifies to use Fujitsu C Compiler (fcc) and SSL.
 *
 *\author Yoshinori Tamada &lt;tamada at ims.u-tokyo.ac.jp&gt;
 */

#ifndef __SIGN_SSM
#include "../util/SGN_Memory.h"
#endif /* __SIGN_SSM */
#include "../signcom.h"
#include "SGN_Matrix.h"
#include "../util/SGN_Str.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <assert.h>
#include <math.h>

char * SGN_Matrix_LU = "U";
char * SGN_Matrix_LL = "L";
char * SGN_Matrix_LN = "N";
char * SGN_Matrix_LV = "V";
char * SGN_Matrix_LA = "A";
char * SGN_Matrix_LO = "O";

#ifndef DOXY
char * SGN_Matrix_L_TRANS = "T";
char * SGN_Matrix_L_NO_TRANS = "N";
#endif

int SGN_Matrix_INT_ONE = 1;
int SGN_Matrix_INT_ZERO = 0;
double SGN_Matrix_DOUBLE_ONE = 1.0;
double SGN_Matrix_DOUBLE_ZERO = 0.0;
double SGN_Matrix_DOUBLE_MINUS_ONE= -1.0;

/* For FORTRAN BLAS */
#ifndef USE_CBLAS
char * SGN_Matrix_U = "U";
char * SGN_Matrix_L = "L";
char * SGN_Matrix_T = "T";
char * SGN_Matrix_N = "N";
char * SGN_Matrix_TRANS = "T";
char * SGN_Matrix_NO_TRANS = "N";
char * SGN_Matrix_LEFT = "L";
char * SGN_Matrix_RIGHT = "R";
#endif /* #ifndef CBLAS */

#ifdef DOXY
/*!\brief This specifies transposing of a matrix in BLAS routines.
 *  Use \ref SGN_Matrix_L_Transpose for LAPACK routines.
 *
 *\ingroup SGN_Matrix */
typedef enum {
  /*!\brief No transpose */
  SGN_Matrix_NO_TRANS, 

  /*!\brief Transpose */
  SGN_Matrix_TRANS
} SGN_Matrix_Transpose;

/*!\brief This specifies transposing of a matrix in LAPACK routines.
 *   Use \ref SGN_Matrix_Transpose for BLAS routines.
 *
 *\ingroup SGN_Matrix */
typedef enum {
  /*!\brief No transpose. */
  SGN_Matrix_L_NO_TRANS,

  /*!\brief Transpose. */
  SGN_Matrix_L_TRANS
} SGN_Matrix_L_Transpose;

#endif /* DOXY */

/*!\brief Allocates the memory for a matrix or a vector.
 *
 * Element (i,j) is accessible by <TT>A[i + j * m]</TT> where
 * <TT>A</TT> is the pointer to the matrix (or the double array)
 * allocated by this function.
 *
 *\param m   number of rows.
 *\param n   number of columns.
 *\ingroup SGN_Matrix */
SGN_Matrix_double * SGN_Matrix_new(int m, int n){
  return (SGN_Matrix_double *) malloc(sizeof(SGN_Matrix_double) * m * n);
}
/* SGN_Matrix_new() ------------------------------------------------------------ */

/*!\brief Allocates the memory for an array of matrices or vectors.
 *
 * The i-th matrix (vector) can be accessed by <tt>r[i]</tt> where
 * <tt>r</tt> is the pointer returned by this function.
 *
 *\param m  rows of each matrix.
 *\param n  columns of each matrix.  If you want an array of vectors, then
 *          specify one (1).
 *\param k  number of matrices to allocate.
 *
 *\return pointer to the array containing \a k elements.  Each of them
 *          corresponds  \a m by \a n matrix.
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
SGN_Matrix_double ** SGN_Matrix_new_ar(const int m, const int n, const int k){
  double ** r = malloc(sizeof(SGN_Matrix_double *) * k);
  double * p = malloc(sizeof(SGN_Matrix_double) * m * n * k);
  for(int i = 0; i < k; i++){
	r[i] = p;
	p += m * n;
  }
  return r;
}
/* SGN_Matrix_new_ar() ========================================================= */

/*!\brief Deletes (releases) the memory allocated to the matrix.
 *\param M  pointer to the matrix returned by SGN_Matrix_new().
 *\ingroup SGN_Matrix */
void SGN_Matrix_delete(SGN_Matrix_double * M){
  free(M);
}
/* SGN_Matrix_delete() --------------------------------------------------------- */

/* ========================================================================= */
/*  FOR  DOXYGEN                                                             */
/* ========================================================================= */
#ifdef DOXY 

/*!\brief BLAS DSCAL; \a X := \a alpha \a X.
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_dscal(int * N, double * alpha, double * X, int * incX);

/*!\brief BLAS DCOPY; \a Y := \a X.
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_dcopy(int N, double * X, int incX, double * Y, int incY);

/*!\brief BLAS DAXPY; \a Y := \a alpha \a X + \a Y.
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_daxpy(int * N, double * alpha, double * X, int * incX, double * Y, int * incY);

/*!\brief BLAS DDOT; \a X' \a Y.
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
double SGN_Matrix_ddot(int * N, double * X, int * incX, double * Y, int * incY);

/*!\brief BLAS DNRM2;
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
double SGN_Matrix_nrm2(int * N, double * X, int * incX);

/*!\brief BLAS DASUM; Sum of |\a x<sub>i</sub>|.
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
double SGN_Matrix_dasum(int * N, double * X, int * incX);

#endif /* DOXY */
/* ======================================================================== */

void SGN_Matrix_dsyaxpy(const int n, double alpha, double * X, double * Y)
{
  int i;
  int pos = 0;

  for(i = 0; i < n; i++){
	int ii = i + 1;
	SGN_Matrix_daxpy(&ii, &alpha, X + pos, &SGN_Matrix_INT_ONE, Y + pos, &SGN_Matrix_INT_ONE);
	pos += n;
  }
}
/* SGN_Matrix_dsyaxpy() -------------------------------------------------------- */
 
void SGN_Matrix_dsycopy(const int n, const double * X, double * Y){
  int i;
  int pos = 0;
  for(i = 0; i < n; i++){
	memcpy(Y + pos, X + pos, sizeof(double) * (i + 1));
	//memcpy(Y + i * n, X + i * n, sizeof(double) * (i + 1));
#ifdef BNRC_DEBUG
	//fprintf(SIGN_ERR, "DEBUG: i=%2d: X=", i);
	//SGN_Matrix_print(SIGN_ERR, X + pos, 1, n);
	//fprintf(SIGN_ERR, "DEBUG:       Y=");
	//SGN_Matrix_print(SIGN_ERR, Y + pos, 1, n);
#endif
	pos += n;
  }
}
/* SGN_Matrix_dsycopy() -------------------------------------------------------- */

void SGN_Matrix_dsyge(const int n, double * X){
  int i, j;
  for(j = 0; j < n; j++){
	for(i = j + 1; i < n; i++){
	  X[i + j * n] = X[j + i * n];
	}
  }
}

/*!\brief Calculates the transpose of \a A and set it to \a B.
 *
 * \a B := \a A<sup>T</sup>.
 *
 *\param M  Number of rows of \a A.
 *\param N  Number of columns of \a A.
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_trans(int M, int N, double * A, double * B){

  for(int i = 0; i < N; i++){
	SGN_Matrix_dcopy(&M, A, &SGN_Matrix_INT_ONE, B + i, &N);
	A += M;
  }
}
/* SGN_Matrix_trans() ========================================================== */

#ifdef DOXY
/*!\brief BLAS DGEMV; double, general matrix, matrix-vector multiplication.
 *
 * Wrapper macro for BLAS DGEMV routine.
 *
 * This calculates \a alpha op(\a A) * \a X + \a beta \a Y and set it to \a Y
 * where op(\a A) is, depending on \a transA, one of 
 *  - op(\a A) = \a A (\a transA = \ref SGN_Matrix_NO_TRANS)
 *  - op(\a A) = \a A' (\a transA = \ref SGN_Matrix_TRANS)
 *
 *\param A  \a M by \a N matrix.
 *\param Y  \a M size vector
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_dgemv(SGN_Matrix_Transpose transA, int M, int N,
				  double alpha, double * A, int ldA,
				  double * X, int incX, double beta, double * Y, int incY);

/*!\brief BLAS DGER; \a A := \a alpha \a X \a Y' + \a A.
 *
 *\param A  \a M by \a N matrix.
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_dger(int M, int N, double alpha, double * X, int incX,
				 double * Y, int incY, double * A, int ldA);

/*!\brief BLAS DGEMM; double general matrix, matrix-matrix multiplication.
 *
 *
 *\param transA \ref SGN_Matrix_NO_TRANS or \ref SGN_Matrix_TRANS.
 *\param M
 *\param N
 *\param K
 *\param A  [INPUT] \a M by \a K matrix
 *\param B  [INPUT] \a K by \a M matrix
 *\param C  [INPUT/OUTPUT] \a M by \a N matrix
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_dgemm(SGN_Matrix_Transpose transA, SGN_Matrix_Transpose transB,
				  int M, int N, int K, double alpha, double * A, int ldA,
				  double * B, int ldB, double beta, double * C, int ldC);
#endif
 
/*!\brief Prints the matrix.
 *
 * The delimiter is a space.
 *
 *\ingroup SGN_Matrix */
void SGN_Matrix_print(FILE * fp, const double * X, int m, int n){

  for(int i = 0; i < m; i++){

	if(0 < n) fprintf(fp, "%8.4f", X[i]);
	for(int j = 1; j < n; j++){
	  fprintf(fp, "   %8.4f", X[i + j * m]);
	}
	fprintf(fp, "\n");
  }
}
/* SGN_Matrix_print() ---------------------------------------------------------- */
 
void SGN_Matrix_printf(FILE * fp, const double * X, const int m, const int n,
				   const char * fmt){
  int i, j;
  for(i = 0; i < m; i++){
	for(j = 0; j < n; j++){
	  if(j != 0) fprintf(fp, "  ");
	  fprintf(fp, fmt, X[i + j * m]);
	}
	fprintf(fp, "\n");
  }
}
/* SGN_Matrix_fprint() --------------------------------------------------------- */

/*!\brief Prints the matrix with specified precision.
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_fprintf(FILE * fp, const int digits, 
					const double * X, const int m, const int n){
  char buff[32];
  sprintf(buff, "%%.%df", digits);
  int i, j;
  for(i = 0; i < m; i++){
	for(j = 0; j < n; j++){
	  if(j != 0) fprintf(fp, "\t");
	  fprintf(fp, buff, X[i + j * m]);
	}
	fprintf(fp, "\n");
  }
}
/* SGN_Matrix_fprint() ========================================================= */

void SGN_Matrix_debug_write(FILE * fp, const char * label, const double * X,
						const int m, const int n){
  fprintf(fp, "%s\t%d\t%d\n", label, m, n);
  SGN_Matrix_fprintf(fp, 15, X, m, n);
  fprintf(fp, "\n");
}
/* SGN_Matrix_debug_write() ==================================================== */

/*
  mode = 0 : Do Nothing
  mode = 1 : Writing
  mode = 2 : Reading & check
 */
double SGN_Matrix_debug_check(FILE * fp, const char * label, const double * X, 
						  const int m, const int n, const int mode, const int abs){
  char * buff = NULL;
  size_t buff_size = 0;
  char * pos;
  char * next_pos;
  double max_diff = 0.0;

  if(mode == 1){
	SGN_Matrix_debug_write(fp, label, X, m, n);
	return 0.0;
  }

  assert(SGN_Str_readline(fp, &buff, &buff_size) == 0);

  /* Checks the label and the dimension */
  pos = buff;
  next_pos = strchr(pos, '\t');
  assert(next_pos != NULL);
  *next_pos = '\0';
  if(strcmp(label, pos) != 0){
	fprintf(SIGN_ERR, "SGN_Matrix_debug_check: Different label detected.\n");
	fprintf(SIGN_ERR, "                    Old vs New: \"%s\" vs \"%s\"\n",
			pos, label);
	exit(1);
  }
  pos = next_pos + 1; next_pos = strchr(pos, '\t');
  assert(pos != NULL && m == atoi(pos));
  pos = next_pos + 1;
  assert(pos != NULL && n == atoi(pos));

  for(int i = 0; i < m; i++){
	SGN_Str_readline(fp, &buff, &buff_size);
	pos = buff;
	for(int j = 0; j < n; j++){
	  double old_v = atof(pos);
	  double new_v = X[i + j * m];
	  double d;
	  if(abs){
		d = fabs(fabs(old_v) - fabs(new_v));
	  } else {
		d = fabs(old_v - new_v);
	  }
	  if(mode == 2 && d > 0.0001){
		fprintf(SIGN_ERR, "SGN_Matrix_debug_check: Failed for \"%s\" diff=%.15f\n",	label, d);
		fprintf(SIGN_ERR, "                  : X(%d,%d): Old vs New: %.15f vs %.15f\n", 
				i, j, old_v, new_v);
		exit(1);
	  }
	  if(mode == 2 && d > max_diff) max_diff = d;
	  pos = strchr(pos + 1, '\t') + 1;
	}
  }

  /* Cheks the last empty line */
  assert(SGN_Str_readline(fp, &buff, &buff_size) == 0);
  assert(strlen(buff) == 0);

  //fprintf(SIGN_ERR, "MAX DIFF = %.15f (%s)\n", max_diff, label);
  
  free(buff);
  return max_diff;
}
/* SGN_Matrix_debug_check() ==================================================== */

/*!\brief Returns the optimal working memory size for dsytrf routine.
 *
 * 
 *\return optimal working memory size used by SGN_Matrix_dsytrf(),
 *   or 0 if the native LAPACK routine does not need it.
 *\ingroup SGN_Matrix */
SGN_Matrix_int SGN_Matrix_dsytrf_work_size(SGN_Matrix_int n){
  SGN_Matrix_int size = 0;
  SGN_Matrix_double * p = NULL;
  SGN_Matrix_double work[1];
  SGN_Matrix_int lwork = -1;
  SGN_Matrix_int pv[n];
  SGN_Matrix_int info;
 
  SGN_Matrix_dsytrf(SGN_Matrix_LU, &n, p, &n, pv, work, &lwork, &info);
  size = (SGN_Matrix_int) work[0];

  return size;
}
/* SGN_Matrix_dsytrf_work_size() ----------------------------------------------- */

/*!\brief Returns the optimal working memory size for SGN_Matrix_dgetri().
 *
 *\return optimal working memory size used by SGN_Matrix_dgetri(),
 *   or 0 if the native LAPACK routine does not need it.
 *\ingroup SGN_Matrix */
SGN_Matrix_int SGN_Matrix_dgetri_work_size(SGN_Matrix_int n){
  SGN_Matrix_int size = 0;
  SGN_Matrix_double dummy;
  SGN_Matrix_int pv[n];
  SGN_Matrix_double work[1];
  SGN_Matrix_int lwork = -1;
  SGN_Matrix_int info;

  SGN_Matrix_dgetri(&n, &dummy, &n, pv, work, &lwork, &info);
  size = (SGN_Matrix_int) work[0];

  /*
#if (defined USE_IBM_ESSL) || (defined USE_INTEL_MKL) || (defined 
  SGN_Matrix_dgetri(n, dummy, n, pv, work, lwork, &info);
  size = (SGN_Matrix_int) work[0];
#else
  info = 0; lwork = 0; work[0] = 0.0; pv[0] = 1; dummy = NULL;
#endif  
  */

  return size;
}

SGN_Matrix_int SGN_Matrix_dsyev_work_size(SGN_Matrix_int n){
#ifdef USE_IBM_ESSL
  return 0;
#else
  SGN_Matrix_int size = 0;
  SGN_Matrix_double * A = NULL;
  SGN_Matrix_double * w = NULL;
  SGN_Matrix_double work[1];
  SGN_Matrix_int lwork = -1;
  SGN_Matrix_int info;

  SGN_Matrix_dsyev(SGN_Matrix_LN, SGN_Matrix_LU, &n, A, &n, w, work, &lwork, &info);
  size = (SGN_Matrix_int) work[0];
  return size;
#endif
}
/* SGN_Matrix_dsyev_work_size() ------------------------------------------------ */


void SGN_Matrix_dgemv_l(const SGN_Matrix_int M, const SGN_Matrix_int N,
					const double alpha, const double * A, const SGN_Matrix_int lda,
					const double * X, const SGN_Matrix_int incX, 
					const double beta, double * Y, const SGN_Matrix_int incY){
  int i, j;
  double * YY = Y;

  for(i = 0; i < M; i++){
	double sum = 0.0;
	const double * XX = X;
	for(j = 0; j < N; j++){
	  sum += A[i + j * lda] * *XX;
	  XX += incX;
	}
	*YY = alpha * sum + beta * *YY;
	YY += incY;
  }
}
/* SGN_Matrix_dgemv_l() -------------------------------------------------------- */

/*!\brief
 *\ingroup SGN_Matrix */
void SGN_Matrix_sy2ge(SGN_Matrix_double * X, const SGN_Matrix_int n){
  int i, j;
  for(j = 0; j < n; j++){
	for(i = 0; i < j; i++){
	  X[j + i * n] = X[i + j * n];
	}
  }
}
/* SGN_Matrix_sy2ge() ---------------------------------------------------------- */

/*!\brief Calculates the trace of the squre matrix.
 *\ingroup SGN_Matrix */
double SGN_Matrix_trace(const double * X, const SGN_Matrix_int n){
  double tr = 0.0;
  int pos = 0;
  for(int i = 0; i < n; i++){
	tr += X[pos];
	pos += n + 1;
  }
  return tr;
}
/* SGN_Matrix_trace() ========================================================== */

/*!\brief Converts a vector into the diagonal matrix.
 *
 * \a A := \a alpha diag(\a X).
 *
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_diag(const int N, double alpha, const double * X, double * A, int LDA){
  memset(A, 0, sizeof(double) * N * N);
  int pos = 0;
  for(int i = 0; i < N; i++){
	A[pos++] = alpha * X[i];
	pos += LDA;
  }
}
/* SGN_Matrix_diag() =========================================================== */

/*!\brief Sets the identity matrix.
 *
 *\param A \a N by \a N matrix.
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_I(const int N, double * A){
  memset(A, 0, sizeof(double) * N * N);
  int pos = 0;
  for(int i = 0; i < N; i++){
	A[pos++] = 1.0;
	pos += N;
  }
}
/* SGN_Matrix_I() ============================================================== */

/*!\brief Calculates the multiplication of a diagonal matrix and a matrix.
 *
 * This calculates \a A := diag(x) \a A where \a x is a \a M size vector and
 * \a A is an \a M by \a N matrix.
 *
 * Note that this overwrites the values in \a A.
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_dgedm2(int M, int N, double * x, double * A, int LDA){
  for(int j = 0; j < N; j++){
	for(int i = 0; i < M; i++){
	  *(A + i) *= x[i];
	}
	A += LDA;
  }
}
/* SGN_Matrix_dgedm2() ========================================================= */

/*!\brief Calculates the multiplication of a diagonal matrix and a matrix.
 *
 * This calculates \a A := \a A diag(x)  where
 * \a A is an \a M by \a N matrix and
 *  \a x is an \a N size vector.
 *
 * Note that this overwrites the values in \a A.
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_dgemd2(int M, int N, double * A, int LDA, double * x){
  for(int j = 0; j < N; j++){
	double v = x[j];
	for(int i = 0; i < M; i++){
	  *(A + i) *= v;
	}
	A += LDA;
  }
}
/* SGN_Matrix_dgemd2() ========================================================= */

/*!\brief inverse of diagonal matrix x general matrix.
 *
 * Calculates \a A := (diag(\a x)<sup>-1</sup>) \a A where
 * \a x is an \a M size vector and \a A is an \a M by \a N matrix.
 *
 *\param M
 *\param N
 *\param x
 *\param A
 *\param LDA   The first dimension of \a A.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_dgeidm2(int M, int N, double * x, double * A, int LDA){
  for(int j = 0; j < N; j++){
	for(int i = 0; i < M; i++){
	  *(A + i) /= x[i];
	}
	A += LDA;
  }
}
/* SGN_Matrix_dgeidm() ========================================================= */

/*!\brief Calculates \a A := \a A (diag(x)<sup>-1</sup>).
 *
 *\param A \a M by \a N matrix.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_dgemid2(int M, int N, double * A, int LDA, double * x){
  for(int j = 0; j < N; j++){
	double v = x[j];
	for(int i = 0; i < M; i++){
	  *(A + i) /= v;
	}
	A += LDA;
  }
}
/* SGN_Matrix_dgemid() ========================================================= */

#ifdef DOXY
/*!\brief LAPACK DGETRF; LU Factorization of a general matrix.
 *
 *
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_dgetrf(int * M, int * N, double * A, int * ldA, int * ipiv, int * info);

/*!\brief LAPACK DGETRS; Solves a system of linear equations using
 *            LU factorized matrix.
 *
 *
 *<DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup SGN_Matrix */
void SGN_Matrix_dgetrs(SGN_Matrix_L_Transpose trans, int * N, int * M,
				   double * A, int * ldA, int * ipiv, double * B, int * ldB,
				   int * info);
#endif

/*!\brief Returns the maximum value of the absolute elements in a vector.
 *
 *
 */
double SGN_Matrix_vector_abs_max(int n, const double * x){
  double v = fabs(x[0]);
  for(int i = 1; i < n; i++){
	double a = fabs(x[i]);
	if(v < a) v = a;
  }
  return v;
}
/* SGN_Matrix_vector_abs_max() ============================================= */

double SGN_Matrix_vector_abs_min(int n, const double * x){
  double v = fabs(x[0]);
  for(int i = 1; i < n; i++){
	double a = fabs(x[i]);
	if(a < v) v = a;
  }
  return v;
}
/* SGN_Matrix_vector_abs_max() ============================================= */

/* End of file.  Copyright (C) 2008-2014  SiGN Project Members  */
