/*
  math/SGN_PCGraph.c : SGN_PCGraph module for efficient manipulation
  
  Copyright (C) 2009 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.
  
  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#include "SGN_PCGraph.h"

#include "../util/SGN_Memory.h"
#include "../signcom.h"
#include "../util/SGN_IntArray.h"
#include "OTO_PK.h"

#include <string.h>
#include <strings.h>


/*!\brief Generates a new \ref SGN_PCGraph instance.
 *
 *\param nodes  the number of nodes in the graph.
 *\returns pointer to a new \ref SGN_PCGraph instance.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
SGN_PCGraph * SGN_PCGraph_new(size_t nodes){
  SGN_PCGraph * g = (SGN_PCGraph *) malloc(sizeof(SGN_PCGraph));
  int i;

  g->buff_size = 0;
  g->ptr = malloc(sizeof(SGN_PCGraph_Edge) * nodes);
  g->nodes = nodes;
  g->edges = 0;

  for(i = 0; i < nodes; i++){
	g->ptr[i].pa = -3;
	g->ptr[i].ch = -3;
	g->ptr[i].pa_next = SGN_PCGraph_NULL;
	g->ptr[i].pa_prev = SGN_PCGraph_NULL;
	g->ptr[i].ch_next = SGN_PCGraph_NULL;
	g->ptr[i].ch_prev = SGN_PCGraph_NULL;
  }

  g->pa_last = malloc(sizeof(size_t) * nodes * 2);
  g->ch_last = g->pa_last + nodes;

  for(i = 0; i < nodes; i++){
	g->pa_last[i] = i;
	g->ch_last[i] = i;
  }

  g->last = SGN_PCGraph_NULL;
  g->in_degree = malloc(sizeof(int) * nodes);
  g->out_degree = malloc(sizeof(int) * nodes);
  memset(g->in_degree, 0, sizeof(int) * nodes);
  memset(g->out_degree, 0, sizeof(int) * nodes);
  
  return g;
}
/* SGN_PCGraph_new() ----------------------------------------------------------- */

/*!\brief Deletes the \SGN_PCGraph instance.
 *
 *\param this  pointer to the \ref SGN_PCGraph instance to delete.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
void SGN_PCGraph_delete(void * this){
  SGN_PCGraph * g = (SGN_PCGraph *) this;

  free(g->ptr);
  free(g->pa_last);
  free(g);
}
/* SGN_PCGraph_delete() -------------------------------------------------------- */

/*!\brief Copies the \ref SGN_PCGraph instance to another instance.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
void SGN_PCGraph_copy(const SGN_PCGraph * this, SGN_PCGraph * g){
  if(g->nodes != this->nodes){
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_PCGraph_copy(): size inconsisitent.\n");
	exit(1);
  }

  g->edges = this->edges;
  SGN_PCGraph_resize_buff(g, this->buff_size);
  
  memcpy(g->ptr, this->ptr, (g->buff_size + g->nodes) * sizeof(SGN_PCGraph_Edge));
  
  memcpy(g->pa_last, this->pa_last, sizeof(size_t) * this->nodes * 2);
  g->last = this->last;
}
/* SGN_PCGraph_copy() ---------------------------------------------------------- */

/*!\brief Resizes the edge buffer size.
 *
 * If \a size is smaller than the current buffer size,
 * it is regarded as the current buffer size.
 *
 * The current implementation will abort if it fails
 * to resize the memory, especially when enlarging it.
 *
 *\param this  pointer to the \ref SGN_PCGraph instance.
 *\param size  new buffer size.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
void SGN_PCGraph_resize_buff(SGN_PCGraph * this, size_t size){
  void * new_ptr;

  if(size < this->edges) size = this->edges;

  new_ptr = realloc(this->ptr, sizeof(SGN_PCGraph_Edge) * (this->nodes + size));
  if(new_ptr == NULL){
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_PCGraph_resize_buff(): "
			"memory allocation failure.\n");
	abort();
  }

  this->ptr = new_ptr;
  this->buff_size = size;
}
/* SGN_PCGraph_resize_buff() --------------------------------------------------- */

/*!\brief Returns the number of nodes.
 *
 *\param this  pointer to the \ref SGN_PCGraph instance.
 *\return number of nodes in the graph.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_num_nodes(const SGN_PCGraph * this){
  return this->nodes;
}
/* SGN_PCGraph_num_nodes() ----------------------------------------------------- */

/*!\brief Returns the number of edges.
 *
 *\param this  pointer to the \ref SGN_PCGraph instance.
 *\return number of edges in the graph.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_num_edges(const SGN_PCGraph * this){
  return this->edges;
}
/* SGN_PCGraph_num_edges() ----------------------------------------------------- */

/*!\brief Adds a new edge from \a u to \a v.
 *
 * If the buffer size is not enough to add a new edge,
 * this automatically extends its size.
 *
 * This does not check if the edge already exists in the graph.
 *
 * This is a constant time operation.
 *
 *\param this  pointer to the \ref SGN_PCGraph instance.
 *\param u     start node of the edge to add.
 *\param v     end node of the edge to add.
 *\return edge ID of the newly added one.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_add_edge(SGN_PCGraph * this, int u, int v){
  size_t id;
  
  if(this->buff_size == this->edges){
	/* Extend the edge buffer */
	size_t new_size = this->buff_size * 1.5;
	if(new_size == this->buff_size) new_size = this->buff_size + 1;

	SGN_PCGraph_resize_buff(this, new_size);
  }

  if(this->last == SGN_PCGraph_NULL){
	id = this->edges + this->nodes;
  } else {
	/* Re-use an empty slot */
	id = this->last;
	this->last = this->ptr[this->last].pa_next;
  }

  this->ptr[id].pa = u;
  this->ptr[id].ch = v;
  this->ptr[id].pa_next = SGN_PCGraph_NULL;
  this->ptr[id].ch_next = SGN_PCGraph_NULL;

  this->ptr[this->pa_last[v]].pa_next = id;
  this->ptr[id].pa_prev = this->pa_last[v];
  this->pa_last[v] = id;

  this->ptr[this->ch_last[u]].ch_next = id;
  this->ptr[id].ch_prev = this->ch_last[u];
  this->ch_last[u] = id;

  this->edges++;
  this->out_degree[u]++;
  this->in_degree[v]++;

  return id;
}
/* SGN_PCGraph_add_edge() ------------------------------------------------------ */

/*!\brief Checks if the specified edge exists.
 *
 * This takes at most <I>O</I>(<I>n</I>) time where <I>n</I> represents
 * the number of parents or children.
 *
 *\param this  pointer to the \ref SGN_PCGraph instance.
 *\param u  start node of the edge to check
 *\param v  end node of the edge to check
 *\return 1 if the edge exsits, or 0 otherwise.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
int SGN_PCGraph_check_edge(SGN_PCGraph * this, int u, int v){

  size_t u_next, v_next;
  
  u_next = this->ptr[u].ch_next;
  v_next = this->ptr[v].pa_next;

  while(u_next != SGN_PCGraph_NULL && v_next != SGN_PCGraph_NULL){
	if(this->ptr[u_next].ch == v) return 1;
	if(this->ptr[v_next].pa == u) return 1;

	u_next = this->ptr[u_next].ch_next;
	v_next = this->ptr[v_next].pa_next;
  }

  /* If you reache here, it means you cannot find the edge. */
  return 0;
}
/* SGN_PCGraph_check_edge() ---------------------------------------------------- */

/*!\brief Checks if the specified edge exists.
 *
 * This takes at most <I>O</I>(<I>n</I>) time where <I>n</I> represents
 * the number of parents or children.
 *
 * This is identical to SGN_PCGraph_check_edge() except that this returns
 * an edge index (not an edge ID) if found.
 *
 *\param this  pointer to the \ref SGN_PCGraph instance.
 *\param u  start node of the edge to check
 *\param v  end node of the edge to check
 *\return index of the edge if it exsits, or \ref SGN_PCGraph_NULL otherwise.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_check_edge_index(SGN_PCGraph * this, int u, int v){
  size_t u_next, v_next;
  u_next = this->ptr[u].ch_next;
  v_next = this->ptr[v].pa_next;
  while(u_next != SGN_PCGraph_NULL && v_next != SGN_PCGraph_NULL){
	if(this->ptr[u_next].ch == v) return u_next - this->nodes;
	if(this->ptr[v_next].pa == u) return v_next - this->nodes;

	u_next = this->ptr[u_next].ch_next;
	v_next = this->ptr[v_next].pa_next;
  }

  return SGN_PCGraph_NULL;
}
/* SGN_PCGraph_check_edge_index() ---------------------------------------------- */

/*!\brief Checks if the specified edge exists.
 *
 * This takes at most <I>O</I>(<I>n</I> / 2) time where <I>n</I> represents
 * the number of parents or children.
 *
 * This is identical to SGN_PCGraph_check_edge() except that this returns
 * an edge ID (not an edge index) if found.
 *
 *\param this  pointer to the \ref SGN_PCGraph instance.
 *\param u  start node of the edge to check
 *\param v  end node of the edge to check
 *\return ID of the edge if it exsits, or \ref SGN_PCGraph_NULL otherwise.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_check_edge_id(SGN_PCGraph * this, int u, int v){
  size_t u_next, v_next, u_prev, v_prev;
  u_next = this->ptr[u].ch_next;
  v_next = this->ptr[v].pa_next;
  u_prev = this->ch_last[v];
  v_prev = this->pa_last[v];
  while(u_next != SGN_PCGraph_NULL && v_next != SGN_PCGraph_NULL &&
		u_prev != SGN_PCGraph_NULL && u_prev != SGN_PCGraph_NULL){
	if(this->ptr[u_next].ch == v) return u_next;
	if(this->ptr[v_next].pa == u) return v_next;
	if(this->ptr[u_prev].ch == v) return u_prev;
	if(this->ptr[v_prev].pa == u) return v_prev;

	u_next = this->ptr[u_next].ch_next;
	v_next = this->ptr[v_next].pa_next;
	u_prev = this->ptr[u_prev].ch_prev;
	v_prev = this->ptr[v_prev].pa_prev;
  }

  return SGN_PCGraph_NULL;
}
/* SGN_PCGraph_check_edge_index() ---------------------------------------------- */

/*!\brief Obtains the start and end nodes of the specified edge.
 *
 * This returns <I>u'</I> and <I>v'</I> of (<I>u', v'</I> )-edge,
 * specified by its ID \a id,  into the variables given by pointer
 * parameters \a u and \a v.
 *
 * Note that argument \a id is not an index of edges.
 * To visit all the edges in the graph, use SGN_PCGraph_get_edge_index().
 * The first edge does not have an ID == 0.
 *
 *\param this  pointer to the \ref SGN_PCGraph instance.
 *\param id   ID of the edge to obtain
 *\param u    initial node of the specified edge will be set.
 *\param v    terminal node of the specified edge will be set.
 *\sa SGN_PCGraph_get_edge_index().
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
void SGN_PCGraph_get_edge(SGN_PCGraph * this, size_t id, int * u, int * v){
  *u = this->ptr[id].pa;
  *v = this->ptr[id].ch;
}
/* SGN_PCGraph_get_edge() ------------------------------------------------------ */

/*!\brief Obtains the start and end nodes of the specified edge.
 *
 * This returns <I>u'</I> and <I>v'</I> of (<I>u', v'</I> )-edge,
 * specified by its index \a index, into the variables given by pointer
 * parameters \a u and \a v.
 *
 * The idexes of edges start from 0 and end at <I>n</I> - 1 where
 * <I>n</I> is the number of edges.
 *
 * Please do not confuse with SGN_PCGraph_get_edge() which returns
 * an edge specified by its ID.
 *
 *\param this  pointer to the \ref SGN_PCGraph instance.
 *\param index index of the edge to obtain.
 *\param u     initial node of the specified edge will be set.
 *\param v     terminal node of the specified edge will be set.
 *\sa SGN_PCGraph_get_edge().
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
void SGN_PCGraph_get_edge_index(const SGN_PCGraph * this, size_t index, int * u, int * v){
  *u = this->ptr[index + this->nodes].pa;
  *v = this->ptr[index + this->nodes].ch;
}
/* SGN_PCGraph_get_edge_index() ----------------------------------------------- */

/*!\brief Removes all the edges in the graph.
 *
 *\param this pointer to the \ref SGN_PCGraph instance.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
void SGN_PCGraph_clear(SGN_PCGraph * this){
  int i;

  this->edges = 0;
  for(i = 0; i < this->nodes; i++){
	this->ptr[i].pa = -4;
	this->ptr[i].ch = -4;
	this->ptr[i].pa_next = SGN_PCGraph_NULL;
	this->ptr[i].ch_next = SGN_PCGraph_NULL;
	this->ptr[i].pa_prev = SGN_PCGraph_NULL;
	this->ptr[i].ch_prev = SGN_PCGraph_NULL;
	this->pa_last[i] = i;
	this->ch_last[i] = i;
  }
  this->last = SGN_PCGraph_NULL;
  memset(this->in_degree, 0, sizeof(int) * this->nodes);
  memset(this->out_degree, 0, sizeof(int) * this->nodes);
}
/* SGN_PCGraph_clear() --------------------------------------------------------- */

/*!\brief Returns the edge ID of the first parent of node \a v.
 *
 * Note that the returning value is not a parent node, but
 * it is an edge ID.
 * To obtain a parent node of the edge ID, use SGN_PCGraph_get_parent().
 *
 * This is convenient if you want to visit all of parents of a certain
 * node.  To do this, after obtaining the first edge ID by this function,
 * use SGN_PCGraph_next_parent_id(), or SGN_PCGraph_parent_next(), which returns
 * both edge ID and the corresponding parent, 
 *iteratively until \ref SGN_PCGraph_NULL is
 * returned.
 *
 *\param this  pointer to the \ref SGN_PCGraph instance.
 *\param v    node
 *\return  edge ID for the first parent of node \a v, or
 *  \ref SGN_PCGraph_NULL if it has no parents.
 *
 *\sa SGN_PCGraph_get_parent(), SGN_PCGraph_next_parent_id(), SGN_PCGraph_parent_next(),
 *      SGN_PCGraph_first_child_id()
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_first_parent_id(SGN_PCGraph * this, int v){
  return this->ptr[v].pa_next;
}
/* SGN_PCGraph_first_parent(SGN_PCGraph * this, int v) ----------------------------- */

/*!\brief Return the edge ID of the first child of node \a v.
 *
 * Note that the returning value is not a child node, but
 * it is an edge ID.
 * To obtain a child node of the edge ID, use SGN_PCGraph_get_child().
 *
 * This is convenient if you want to visit all of children of a certain
 * node.  To do this, after obtaining the first edge ID by this function,
 * use SGN_PCGraph_next_child_id(), or SGN_PCGraph_child_next(), which returns
 * both edge ID and the corresponding parent, 
 *iteratively until \ref SGN_PCGraph_NULL is
 * returned.
 *
 *\param this  pointer to the \ref SGN_PCGraph instance.
 *\param v    node
 *\return  edge ID for the first child of node \a v, or
 *    \ref SGN_PCGraph_NULL if it has no children.
 *
 *\sa SGN_PCGraph_get_child(), SGN_PCGraph_next_child_id(), 
 *        SGN_PCGraph_child_next(), SGN_PCGraph_first_parent_id()
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_first_child_id(SGN_PCGraph * this, int v){
  return this->ptr[v].ch_next;
}
/* SGN_PCGraph_first_child_id() ------------------------------------------------ */

/*!\brief Returns the edge ID of the next parent.
 *
 * The second parameter \id needs to be a value returned by
 * SGN_PCGraph_first_parent_id() or SGN_PCGraph_parent_next().
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_next_parent_id(SGN_PCGraph * this, size_t id){
  return this->ptr[id].pa_next;
}
/* SGN_PCGraph_next_parent_id(SGN_PCGraph * this, size_t id) ----------------------- */

/*!\brief Returns the edge ID of the next child.
 *
 * The second parameter \id needs to be a value returned by
 * SGN_PCGraph_first_child_id() or SGN_PCGraph_child_next().
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_next_child_id(SGN_PCGraph * this, size_t id){
  return this->ptr[id].ch_next;
}
/* SGN_PCGraph_next_child_id(SGN_PCGraph * this, size_t id) ------------------------ */

/*!\brief Returns the parent node and the edge ID of the next parent.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_parent_next(SGN_PCGraph * this, size_t id, int * pa){
  *pa = this->ptr[id].pa;
  return this->ptr[id].pa_next;
}
/* SGN_PCGraph_parent_next(SGN_PCGraph * this, size_t * id, int * pa) -------------- */

/*!\brief Returns the child node and the edge ID of the next child.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_child_next(SGN_PCGraph * this, size_t id, int * ch){
  *ch = this->ptr[id].ch;
  return this->ptr[id].ch_next;
}
/* SGN_PCGraph_child_next(SGN_PCGraph * this, size_t * id, int * ch) --------------- */

/*!\brief Returns the parent (initial) node
 *    corresponding to the given edge ID.
 *
 * This returns the parent (initial) node of the edge specified by
 * its ID \a id.  If the invalid edge ID is given,
 * this returns a meaningless value.
 * To get both initial and terminate node of the edge, use
 * SGN_PCGraph_get_edge().
 *
 *\param this  pointer to the \ref SGN_PCGraph instance.
 *\param id    edge ID.
 *\return initial node of the specified edge.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
int SGN_PCGraph_get_parent(SGN_PCGraph * this, size_t id){
  return this->ptr[id].pa;
}
/* SGN_PCGraph_get_parent(SGN_PCGraph * this, size_t id) --------------------------- */

/*!\brief Returns the child (terminate) node
 *   corresponding to the given edge ID.
 *
 * This returns the child (terminate) node of the edge specified
 * by its ID \a id.  If the invalid edge ID is given,
 * this returns a meaningless value.
 * To get both initial and terminate node of the edge, use
 * SGN_PCGraph_get_edge().
 * 
 *\param this  pointer to the \ref SGN_PCGraph instance.
 *\param id    edge ID.
 *\return terminate node of the specified edge.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
int SGN_PCGraph_get_child(SGN_PCGraph * this, size_t id){
  return this->ptr[id].ch;
}
/* SGN_PCGraph_get_child(SGN_PCGraph * this, size_t id) ---------------------------- */


void SGN_PCGraph_status(SGN_PCGraph * this, FILE * fp){
  size_t i, id;
  fprintf(fp, "edges = %zu\n", this->edges);
  fprintf(fp, "=== All Edges ===\n");
  id = SGN_PCGraph_first_edge(this);
  for(i = 0; i < SGN_PCGraph_num_edges(this); i++){
	fprintf(fp, "--- Edge %zu (ID = %zu) ---\n", i, id);
	fprintf(fp, "  pa = %d  ch = %d\n", this->ptr[id].pa, this->ptr[id].ch);
	fprintf(fp, "  pa_prev = %zu <- -> pa_next = %zu\n", 
			this->ptr[id].pa_prev, this->ptr[id].pa_next);
	fprintf(fp, "  ch_prev = %zu <- -> ch_next = %zu\n",
			this->ptr[id].ch_prev, this->ptr[id].ch_next);
	fprintf(fp, "\n");
	id = SGN_PCGraph_next_edge(this, id);
  }
}
/* SGN_PCGraph_status() -------------------------------------------------------- */

/*!\brief Prints lists of parents for all nodes.
 *
 *\param this  pointer to the \ref SGN_PCGraph instance.
 *\param fp    FILE pointer to which lists of parents are printed.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
void SGN_PCGraph_print_all_parents(SGN_PCGraph * this, FILE * fp){
  int j;
  size_t id;
  for(j = 0; j < this->nodes; j++){
	id = SGN_PCGraph_first_parent_id(this, j);
	fprintf(fp, "%d:", j);
	while(id != SGN_PCGraph_NULL){
	  int i;
	  id = SGN_PCGraph_parent_next(this, id, &i);
	  fprintf(fp, " %d", i);
	}
	fprintf(fp, "\n");
  }/* for j < this->nodes */
}
/* SGN_PCGraph_print_all_parents(SGN_PCGraph * this, FILE * fp) --------------------- */

void SGN_PCGraph_print_parents(SGN_PCGraph * this, int j, FILE * fp){
  size_t id = SGN_PCGraph_first_parent_id(this, j);
  fprintf(fp, "PARENTS of %d:", j);
  while(id != SGN_PCGraph_NULL){
	int pa;
	id = SGN_PCGraph_parent_next(this, id, &pa);
	fprintf(fp, " %d", pa);
  }
  fprintf(fp, "\n");
}
/* SGN_PCGraph_print_parents() -------------------------------------------------- */

void SGN_PCGraph_print_children(SGN_PCGraph * this, int j, FILE * fp){
  size_t id = SGN_PCGraph_first_child_id(this, j);
  fprintf(fp, "CHILDREN of %d:", j);
  while(id != SGN_PCGraph_NULL){
	int ch;
	id = SGN_PCGraph_child_next(this, id, &ch);
	fprintf(fp, " %d", ch);
  }
  fprintf(fp, "\n");
}
/* SGN_PCGraph_print_parents() -------------------------------------------------- */

/*!\brief Returns the edge ID for the first edge.
 *
 *\param this  pointer to the \ref SGN_PCGraph instance.
 *\sa SGN_PCGraph_next_edge()
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_first_edge(SGN_PCGraph * this){
  return this->nodes;
}
/* SGN_PCGraph_first_edge() ---------------------------------------------------- */

/*!\brief Returns the next valid ID of the given edge ID
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_next_edge(SGN_PCGraph * this, size_t id){

  while(this->ptr[++id].pa == -1);

  return id;
}
/* SGN_PCGraph_next_edge() ----------------------------------------------------- */

/*!\brief Checks if the graph becomes cyclic by adding the edge.
 *
 * This check whether the graph becomes cyclic if \a u - \a v edge
 * is added.
 *
 * Note that this algorithm is very slow.  If you want to
 * add edges repeadedly with keeping the graph acyclic,  use \ref OTO_PK
 * with the \ref SGN_PCGraph instance.
 *
 *
 *\param this   pointer to the \ref Graph instance.
 *\param u   index of the starting node of the edge
 *\param v   index of the ending node of the edge
 *\return 1 if the graph becomes cyclic with \a u - \a v edge, or 0
 *  otherwise.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
int SGN_PCGraph_check_cyclic(SGN_PCGraph * this, int u, int v){
  int list_a[this->nodes]; /* reached nodes added at the previous turn. */
  int list_a_size = 0;     /* size of 'list_a' */
  int flag_a[this->nodes]; /* flag of nodes in 'list_a'. */

  int flag_b[this->nodes]; /* flag of already reached nodes. */

  int list_c[this->nodes]; /* newly reached nodes in a turn. */
  int list_c_size = 0;     /* size of 'list_c' */
  int flag_c[this->nodes]; /* flag of nodes in 'list_c' */

  int i;

  bzero(flag_a, sizeof(int) * this->nodes);
  bzero(flag_b, sizeof(int) * this->nodes);
  bzero(flag_c, sizeof(int) * this->nodes);

  list_a[list_a_size++] = u; /* the first node */
  flag_a[u] = 1;

  while(1){
	/* Visit all nodes in 'list_a', to collect nodes reachable
	   from nodes in 'list_a'.  The collected nodes are stored in
	   'list_c'	*/
	for(i = 0; i < list_a_size; i++){
	  size_t id = SGN_PCGraph_first_parent_id(this, list_a[i]);

	  while(id != SGN_PCGraph_NULL){
		int n;
		size_t new_id = SGN_PCGraph_parent_next(this, id, &n);

		/* check 'v' is in the newly reached nodes */
		if(n == v) return 1; /* The graph becomes cyclic. */
		
		if(!flag_a[n] && !flag_b[n] && !flag_c[n]){
		  list_c[list_c_size++] = n;
		  flag_c[n] = 1;
		}

		id = new_id;
	  }
	}/* for i < |list_a| */

	if(list_c_size == 0){
	  /* All the reachable nodes were visited. */
	  /* If 'v' is not found at this moment, the graph is acyclic with u->v. */
	  return 0; /* acyclic */
	}

	/* Add the ``front-line'' nodes ('list_a') to 'list_b', 
	   and clear 'list_a' & 'flag_a'. */
	for(i = 0; i < list_a_size; i++){
	  int n = list_a[i];
	  flag_b[n] = 1;
	  flag_a[n] = 0; /* clear flag_a */
	}
	list_a_size = 0;

	/* Copy nodes in 'list_c' into 'list_a',
	 and clear 'list_c' & 'flag_c'. */
	for(i = 0; i < list_c_size; i++){
	  int n = list_c[i];
	  list_a[list_a_size++] = n;
	  flag_a[n] = 1;
	  flag_c[n] = 0;
	}
	list_c_size = 0;
  }
  return 0;
}
/* SGN_PCGraph_check_cyclic() -------------------------------------------------- */

/*!\brief Return the number of parents of the node.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_num_parents(SGN_PCGraph * this, int v){
  /*
  size_t num = 0;
  size_t id = SGN_PCGraph_first_parent_id(this, v);
  while(id != SGN_PCGraph_NULL){
	id = SGN_PCGraph_next_parent_id(this, id);
	num++;
  }
  return num;
  */
  return this->in_degree[v];
}
/* SGN_PCGraph_num_parents() --------------------------------------------------- */

/*\brief Returns the number of children of the node.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_num_children(SGN_PCGraph * this, int v){
  /*
  size_t num = 0;
  size_t id = SGN_PCGraph_first_child_id(this, v);
  while(id != SGN_PCGraph_NULL){
	id = SGN_PCGraph_next_child_id(this, id);
	num++;
  }
  return num;
  */
  return this->out_degree[v];
}
/* SGN_PCGraph_num_child() ----------------------------------------------------- */

/*!\brief Returns the degree of the node.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_degree(SGN_PCGraph * this, int v){
  //return SGN_PCGraph_num_parents(this, v) + SGN_PCGraph_num_children(this, v);
  return this->in_degree[v] + this->out_degree[v];
}
/* SGN_PCGraph_degree() -------------------------------------------------------- */

/*!\brief Removes an edge specified by its ID.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
void SGN_PCGraph_remove_edge_id(SGN_PCGraph * this, size_t id){

  /* Reconnect the linked list */
  this->ptr[this->ptr[id].pa_prev].pa_next = this->ptr[id].pa_next;
  this->ptr[this->ptr[id].ch_prev].ch_next = this->ptr[id].ch_next;

  /* Re-point pa_last and ch_last if the removed edge is the
	 last edge for pa or ch */
  if(this->pa_last[this->ptr[id].ch] == id){
	this->pa_last[this->ptr[id].ch] = this->ptr[id].pa_prev;
  } else {
	this->ptr[this->ptr[id].pa_next].pa_prev = this->ptr[id].pa_prev;
  }

  if(this->ch_last[this->ptr[id].pa] == id){
	this->ch_last[this->ptr[id].pa] = this->ptr[id].ch_prev;
  } else {
	this->ptr[this->ptr[id].ch_next].ch_prev = this->ptr[id].ch_prev;
  }

  
  /* Push the edge onto the edge stack */
  this->ptr[id].pa_next = this->last;
  this->last = id;

  /* Decreases the degrees */
  this->out_degree[this->ptr[id].pa]--;
  this->in_degree[this->ptr[id].ch]--;

  this->ptr[id].pa = -2;
  this->ptr[id].ch = -2;

  /* Decrease the number of edges */
  this->edges--;

}
/* SGN_PCGraph_remove_edge_id() ------------------------------------------------ */

/*!\brief Removes an edge specified by its starting and ending node.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
void SGN_PCGraph_remove_edge(SGN_PCGraph * this, int u, int v){
  size_t id = SGN_PCGraph_check_edge_id(this, u, v);

  if(id != SGN_PCGraph_NULL){
	SGN_PCGraph_remove_edge_id(this, id);
  }
}
/* SGN_PCGraph_remove_edge() --------------------------------------------------- */

size_t SGN_PCGraph_get_parents(SGN_PCGraph * this, int v, int * buff){
  size_t id = SGN_PCGraph_first_parent_id(this, v);
  size_t pos = 0;
  while(id != SGN_PCGraph_NULL){
	id = SGN_PCGraph_parent_next(this, id, buff + pos);
	pos++;
  }
  return pos;
}
/* SGN_PCGraph_get_parents() --------------------------------------------------- */

/*!\brief Gets the list of parents in \ref SGN_IntArray.
 *
 *
 *\return the number of parents of \a v.
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
size_t SGN_PCGraph_parents(SGN_PCGraph * this, int v, SGN_IntArray * ar){
  size_t id = SGN_PCGraph_first_parent_id(this, v);
  size_t pos = 0;

  SGN_IntArray_clear(ar);
  
  while(id != SGN_PCGraph_NULL){
	int pa;
	id = SGN_PCGraph_parent_next(this, id, &pa);
	if(pa < 0) continue;
	SGN_IntArray_add(ar, pa);
	pos++;
  }
  return pos;
}
/* SGN_PCGraph_parents() ------------------------------------------------------- */

size_t SGN_PCGraph_get_children(SGN_PCGraph * this, int v, int * buff){
  size_t id = SGN_PCGraph_first_child_id(this, v);
  size_t pos = 0;
  while(id != SGN_PCGraph_NULL){
	id = SGN_PCGraph_child_next(this, id, buff + pos);
	pos++;
  }
  return id;
}
/* SGN_PCGraph_get_children() -------------------------------------------------- */

typedef struct {
  size_t a;
  float b;
} SGN_PCGraph_DAG_t;

int SGN_PCGraph_DAG_cmp(const void * x, const void * y){
  if(((const SGN_PCGraph_DAG_t *) x)->b > ((const SGN_PCGraph_DAG_t *) y)->b){
	return -1;
  } else if(((const SGN_PCGraph_DAG_t *) x)->b < ((const SGN_PCGraph_DAG_t *) y)->b){
	return 1;
  } else {
	return 0;
  }
}

void SGN_PCGraph_DAG_fsc_th(SGN_PCGraph * g, SGN_FloatArray * sc, float th, SGN_PCGraph * dag){
  size_t m = SGN_PCGraph_num_edges(g);
  size_t n = SGN_PCGraph_num_nodes(g);
  SGN_PCGraph_DAG_t ar[m];
  size_t i;
  size_t num = 0;
  OTO_PK * oto = OTO_PK_new(n);

  for(i = 0; i < m; i++){
	float pr = SGN_FloatArray_get(sc, i);
	if(pr >= th){
	  ar[num].a = i;
	  ar[num].b = pr;
	  num++;
	}
  }
  qsort(ar, num, sizeof(SGN_PCGraph_DAG_t), SGN_PCGraph_DAG_cmp);

  /*
  for(i = 0; i < 10; i++){
	int pa, ch;
	SGN_PCGraph_get_edge_index(g, ar[i].a, &pa, &ch);
	fprintf(SIGN_ERR, "%zu: %f %zu  %d->%d\n", i, ar[i].b, ar[i].a,
			pa, ch);
	
  }
  */

  SGN_PCGraph_clear(dag);

  for(i = 0; i < num; i++){
	int pa, ch;
	SGN_PCGraph_get_edge_index(g, ar[i].a, &pa, &ch);

	OTO_PK_add_edge(oto, pa, ch, dag);
  }/* for i */

  OTO_PK_delete(oto);
}
/* SGN_PCGraph_DAG_fsc_th() ---------------------------------------------------- */

/*!\brief Fills the empty slot in the buffer.
 *
 * This re-allocate the buffer so that the users can access
 * all the edges by SGN_PCGraph_get_edge_index().
 *
 * This is required to call SGN_PCGraph_get_edge_index() if
 * SGN_PCGraph_remove_edge() or SGN_PCGraph_remove_edge_id() have been called once.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\memberof SGN_PCGraph */
void SGN_PCGraph_fill(SGN_PCGraph * this){
  int count = 0;
  size_t id = this->last;
  int offset = 0;

  /* Count the number of slots in the removed buffer to
	 obtain the last position of the slot being used in the buffer.
   */
  while(id != SGN_PCGraph_NULL){
	count++;
	id = this->ptr[id].pa_next;
  }

  fprintf(SIGN_ERR, "SGN_PCGraph_fill(): Moving %d slots. (edges=%zu)\n", 
		  count, this->edges);

  /* The number of slots that are used at least once is
	 the number of edges + 'count'. And the last position
	 we are looking for is (num edges + 'count') - 1.

	 We are going to move 'count' slots to empty slots.
  */
  
  while(count != 0){
	size_t mid = this->edges + count + offset - 1 + this->nodes;

	if(this->ptr[mid].pa < 0){
	  offset--; continue;
	}
	
	id = this->last;
	this->last = this->ptr[this->last].pa_next;

	if(id >= this->edges + this->nodes){
	  /* We do not need to fill this slot */
	  count--; offset++; continue;
	}

	/*
	fprintf(SIGN_ERR, "Moving %zu ==> %zu  (pa=%d,ch=%d)\n", 
			mid, id, this->ptr[mid].pa, this->ptr[mid].ch);
	*/

	this->ptr[id].pa = this->ptr[mid].pa;
	this->ptr[id].ch = this->ptr[mid].ch;

	this->ptr[id].pa_prev = this->ptr[mid].pa_prev;
	this->ptr[this->ptr[mid].pa_prev].pa_next = id;
	this->ptr[id].pa_next = this->ptr[mid].pa_next;

	this->ptr[id].ch_prev = this->ptr[mid].ch_prev;
	this->ptr[this->ptr[mid].ch_prev].ch_next = id;
	this->ptr[id].ch_next = this->ptr[mid].ch_next;

	if(this->pa_last[this->ptr[mid].ch] == mid){
	  this->pa_last[this->ptr[mid].ch] = id;
	} else {
	  this->ptr[this->ptr[mid].pa_next].pa_prev = id;
	}

	if(this->ch_last[this->ptr[mid].pa] == mid){
	  this->ch_last[this->ptr[mid].pa] = id;
	} else {
	  this->ptr[this->ptr[mid].ch_next].ch_prev = id;
	}

	this->ptr[mid].pa = -5;
	this->ptr[mid].ch = -5;

	count--;
  }
}
/* SGN_PCGraph_fill() ---------------------------------------------------------- */

/* End of file.  Copyright (C) 2009-2011  SiGN Project Members */
