/*
  math/SGN_PCGraph.h : SGN_PCGraph module for efficient manipulation
  
  Copyright (C) 2009 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.
  
  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#ifndef __SGN_PC_GRAPH_H
#define __SGN_PC_GRAPH_H

#include "../util/SGN_IntArray.h"
#include "../util/SGN_FloatArray.h"

#ifdef DOXY
/*!\brief Value representing NULL.
 *\memberof SGN_PCGraph */
#define SGN_PCGraph_NULL
#else
#define SGN_PCGraph_NULL ((size_t) -1)
#endif /* DOXY */

typedef struct SGN_PCGraph_Edge_t {
  int pa;
  int ch;

  size_t pa_next;
  size_t ch_next;

  size_t pa_prev;
  size_t ch_prev;
} SGN_PCGraph_Edge;

/*!\brief Graph structure for efficient parents and children handling.
 *
 * \ref SGN_PCGraph is similar to \ref SGN_Graph except that 
 * this maintains the list of edges both for parents and
 * children.  \ref SGN_Graph maintains only the list of
 * parents.  Therefore, if users often need to access both
 * children and parents of certain nodes, then this is suitable
 * for that purposes.
 *
 * \see SGN_Graph, SGN_Network
 *
 * \author Yoshinori Tamada &lt;tamada at is.s.u-tokyo.ac.jp&gt;
 */
typedef struct {
#ifdef DOXY
private:
#endif
  size_t nodes;
  size_t edges;
  size_t buff_size;
  SGN_PCGraph_Edge * ptr; /* First 'nodes' pointers are used to
						 maintain parents and children for nodes */

  /* 
	 'pa_last' & 'ch_last' maintain the last edge for each node.
	 ''pa_last[v]' ('ch_last[v]') points to the position in 'ptr'
	 of the last parent (child) node for 'v'.
   */
  size_t * pa_last; 
  size_t * ch_last;

  size_t last;

  /* degrees of nodes. */
  int * in_degree;
  int * out_degree;

} SGN_PCGraph;

SGN_PCGraph * SGN_PCGraph_new(size_t nodes);
void      SGN_PCGraph_resize_buff(SGN_PCGraph * this, size_t size);
void      SGN_PCGraph_delete(void * this);
void      SGN_PCGraph_copy(const SGN_PCGraph * this, SGN_PCGraph * g);
void      SGN_PCGraph_clear(SGN_PCGraph * this);
void      SGN_PCGraph_status(SGN_PCGraph * this, FILE * fp);
void      SGN_PCGraph_fill(SGN_PCGraph * this);

size_t    SGN_PCGraph_num_nodes(const SGN_PCGraph * this);
size_t    SGN_PCGraph_num_edges(const SGN_PCGraph * this);

size_t    SGN_PCGraph_add_edge(SGN_PCGraph * this, int u, int v);
int       SGN_PCGraph_check_edge(SGN_PCGraph * this, int u, int v);
size_t    SGN_PCGraph_check_edge_index(SGN_PCGraph * this, int u, int v);
size_t    SGN_PCGraph_check_edge_id(SGN_PCGraph * this, int u, int v);
void      SGN_PCGraph_remove_edge_id(SGN_PCGraph * this, size_t id);
void      SGN_PCGraph_remove_edge(SGN_PCGraph * this, int u, int v);

size_t    SGN_PCGraph_first_parent_id(SGN_PCGraph * this, int v);
size_t    SGN_PCGraph_first_child_id(SGN_PCGraph * this, int v);
size_t    SGN_PCGraph_next_parent_id(SGN_PCGraph * this, size_t id);
size_t    SGN_PCGraph_next_child_id(SGN_PCGraph * this, size_t id);
size_t    SGN_PCGraph_parent_next(SGN_PCGraph * this, size_t id, int * pa);
size_t    SGN_PCGraph_child_next(SGN_PCGraph * this, size_t id, int * ch);

int       SGN_PCGraph_get_parent(SGN_PCGraph * this, size_t id);
int       SGN_PCGraph_get_child(SGN_PCGraph * this, size_t id);
void      SGN_PCGraph_get_edge(SGN_PCGraph * this, size_t id, int * u, int * v);
void      SGN_PCGraph_get_edge_index(const SGN_PCGraph * this, size_t index, int * u, int * v);

size_t    SGN_PCGraph_first_edge(SGN_PCGraph * this);
size_t    SGN_PCGraph_next_edge(SGN_PCGraph * this, size_t id);

size_t    SGN_PCGraph_parents(SGN_PCGraph * this, int v, SGN_IntArray * ar);


int       SGN_PCGraph_check_cyclic(SGN_PCGraph * this, int u, int v);

size_t    SGN_PCGraph_num_parents(SGN_PCGraph * this, int v);
size_t    SGN_PCGraph_num_children(SGN_PCGraph * this, int v);
size_t    SGN_PCGraph_degree(SGN_PCGraph * this, int v);

void      SGN_PCGraph_DAG_fsc_th(SGN_PCGraph * g, SGN_FloatArray * sc, float th, 
							 SGN_PCGraph * dag);

void      SGN_PCGraph_print_all_parents(SGN_PCGraph * this, FILE * fp);
void      SGN_PCGraph_print_parents(SGN_PCGraph * this, int j, FILE * fp);
void      SGN_PCGraph_print_children(SGN_PCGraph * this, int j, FILE * fp);

#endif /* __SGN_PC_GRAPH_H */

/* End of file.  Copyright (C) 2009-2011  SiGN Project Members. */
