/*
  sign/math/SGN_RNG.c : Wrapper for Random Number Generators

  Copyright (C) 2010-2012  SiGN Project Members.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#include "SGN_RNG.h"

#if defined(USE_FUJITSU_FCC_SSL)
#include <cssl.h>
#include "../math/SGN_Matrix.h"
#elif defined(USE_INTEL_MKL)
#include <mkl_vsl.h>
#else
#include "../math/SGN_Stat.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*!\brief Generates an SGN_RNG instance.
 *
 * Call SGN_RNG_set_seed() to initialize the random number generator.
 *
 * Call SGN_RNG_delete() to release the resources allocated by this routine.
 *
 *\return Newly generated SGN_RNG instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_RNG */
SGN_RNG * SGN_RNG_new(){

  SGN_RNG * rng = malloc(sizeof(SGN_RNG));
  
#if defined(USE_FUJITSU_FCC_SSL)
  rng->ssl_dvrau4_seed = 1;
  rng->ssl_dvrau4_work_size = 45000; /* according to the SSL II Manual */
  rng->ssl_dvrau4_work = malloc(sizeof(double) * rng->ssl_dvrau4_work_size);

  rng->ssl_dvran3_seed = 1;
  rng->ssl_dvran3_work_size = 100000; /* according to the SSL II Manual */
  rng->ssl_dvran3_work = malloc(sizeof(double) * rng->ssl_dvran3_work_size);

#elif defined(USE_INTEL_MKL)
  rng->mkl_vsl_stream = NULL;
#else
  rng->dsfmt = SGN_Random_dSFMT_new();
#endif

  return rng;
}
/* SGN_RNG_new() =============================================================== */

/*!\brief Releases the resources allocated for the SGN_RNG instance.
 *
 *\param rng SGN_RNG instance to be deleted.  
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_RNG */
void SGN_RNG_delete(SGN_RNG * rng){
#if defined(USE_FUJITSU_FCC_SSL)
  free(rng->ssl_dvrau4_work);
  free(rng->ssl_dvran3_work);

#elif defined(USE_INTEL_MKL)
  if(rng->mkl_vsl_stream != NULL){
	vslDeleteStream(&rng->mkl_vsl_stream);
  }
#elif defined(USE_MINGW)
  /* do nothing */
#else
  SGN_Random_dSFMT_delete(rng->dsfmt);
#endif
  free(rng);
}
/* SGN_RNG_delete() ============================================================ */

/*!\brief Initializes the random number generator.
 *
 * The appropriate algorithm is selected based on the environment.
 *
 * - Fujitsu C Compiler + SSL II Math Libarary : c_dvrau4() and c_dvran3().
 * - Others: random() in the standard C library.
 *
 * In Fujitsu SSL II library, 0 cannot be specified as the seed.
 * Thus, if 0 is specified, it is regarded as 1.
 *
 *\param rng   SGN_RNG instance to be initialized.
 *\param seed  Seed of the random number generator.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_RNG */
void SGN_RNG_set_seed(SGN_RNG * rng, int seed){
#if defined(USE_FUJITSU_FCC_SSL)
  if(seed == 0) seed = 1;
  rng->ssl_dvrau4_seed = seed;
  rng->ssl_dvran3_seed = seed;

#elif defined(USE_INTEL_MKL)
  vslNewStream(&rng->mkl_vsl_stream, VSL_BRNG_MCG31, seed);

#elif defined(USE_MINGW)
  srand(seed);
#else
  //srandom(seed);
  SGN_Random_dSFMT_init(rng->dsfmt, seed);
#endif
}
/* SGN_RNG_set_seed() ========================================================== */

/*!\brief Generates the normally distributed random numbers.
 *
 * The appropriate algorithm is selected based on the environment.
 *
 * - Fujitsu C Compiler + SSL II Math Libarary : c_dvran3().
 * - Others: Original routine based on random() in the standard C library.
 *
 *\param rng   Initialized SGN_RNG instance.
 *\param ar    Array in which the generated random numbers are stored.
 *\param n     Number of random numbers to generate.
 *\param mean  Mean of the normal distribution.
 *\param sigma Standard deviation of the normal distribution.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_RNG */
void SGN_RNG_normal(SGN_RNG * rng, double * ar, int n, double mean, double sigma){
#if defined(USE_FUJITSU_FCC_SSL)
  int ret;
  c_dvran3(mean, sigma, &rng->ssl_dvran3_seed, ar, n,
		   rng->ssl_dvran3_work, rng->ssl_dvran3_work_size, &ret);
#elif defined(USE_INTEL_MKL)
  vdRngGaussian(VSL_METHOD_DGAUSSIAN_ICDF, rng->mkl_vsl_stream, n, ar, mean, sigma);
#else
  for(int i = 0; i < n; i++){
	ar[i] = mean + SGN_Stat_rnorm() * sigma;
  }
#endif  
}
/* SGN_RNG_normal() ============================================================ */

/*!\brief Generates the normally distributed random numbers.
 *
 * The appropriate algorithm is selected based on the environment.
 *
 * - Fujitsu C Compiler + SSL II Math Libarary : c_dvran3().
 * - Intel MKL : vdRngUniform().
 * - Others: Original routine based on Random() in the standard C library.
 *
 *\param rng   Initialized SGN_RNG instance.
 *\param ar    Array in which the generated random numbers are stored.
 *\param n     Number of random numbers to generate.
 *\param min   Lower bound of the uniform distribution.
 *\param max   Upper bound of the uniform distribution.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_RNG */
void SGN_RNG_uniform(SGN_RNG * rng, double * ar, int n, double min, double max){
#if defined(USE_FUJITSU_FCC_SSL)
  int ret;
  double v = max - min; /* range */
  c_dvrau4(&rng->ssl_dvrau4_seed, ar, n,
		   rng->ssl_dvrau4_work, rng->ssl_dvrau4_work_size, &ret);
  SGN_Matrix_dscal(&n, &v, ar, &SGN_Matrix_INT_ONE);
  SGN_Matrix_daxpy(&n, &SGN_Matrix_DOUBLE_ONE, &min, &SGN_Matrix_INT_ZERO, ar, &SGN_Matrix_INT_ONE);
#elif defined(USE_INTEL_MKL)
  vdRngUniform(VSL_METHOD_DUNIFORM_STD, rng->mkl_vsl_stream, n, ar, min, max);
#else
  if(n < DSFMT_N64){
	double tmp[DSFMT_N64];
	SGN_Random_dSFMT_close_open_array(rng->dsfmt, tmp, DSFMT_N64);
	memcpy(ar, tmp, sizeof(double) * n);
  } else {
	SGN_Random_dSFMT_close_open_array(rng->dsfmt, ar, n);
  }
  double v = max - min; /* range */
  for(int i = 0; i < n; i++){
	//ar[i] = min + SGN_Stat_runif() * v;
	ar[i] = min + ar[i] * v;
  }
#endif  
}
/* SGN_RNG_normal() ============================================================ */

/*!\brief Generates integer random numbers.
 *
 * This generates integer random numbers ranging [\a min,\a max).
 *
 *\param rng  Intitialized SGN_RNG instance.
 *\param ar   Array in which the generated random numbers are stored.
 *\param n    Numbers of random number to generate.
 *\param min  Lower bound of the random values (inclusive).
 *\param max  Upper bound of the random values (exclusive).
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_RNG */
void SGN_RNG_integer(SGN_RNG * rng, int * ar, int n, int min, int max){
#if defined(USE_FUJITSU_FCC_SSL)
  int ret;
  double tmp[n];
  double v = max - min; /* range */
  c_dvrau4(&rng->ssl_dvrau4_seed, tmp, n,
		   rng->ssl_dvrau4_work, rng->ssl_dvrau4_work_size, &ret);
  SGN_Matrix_dscal(&n, &v, tmp, &SGN_Matrix_INT_ONE);
  for(int i = 0; i < n; i++){
	ar[i] = ((int) tmp[i]) + min;
  }
#elif defined(USE_INTEL_MKL)
  viRngUniform(VSL_METHOD_IUNIFORM_STD, rng->mkl_vsl_stream, n, ar, min, max);
#elif defined(USE_MINGW)
  int v = max - min;
  for(int i = 0; i < n; i++){
	ar[i] = (rand() % v) + min;
  }
#else
  int v = max - min;
  for(int i = 0; i < n; i++){
	ar[i] = (SGN_Random_dSFMT_uint32(rng->dsfmt) % v) + min;
  }
#endif
}
/* SGN_RNG_integer() =========================================================== */

/*!\brief Generates an integer random permuation vector.
 *
 * This generates an random permutation ranging from 0 to \a n - 1.
 *
 *\param rng  Initialized SGN_RNG instance.
 *\param ar   Array in which the generated random permuation vector is stored.
 *\param n    Length of the permutation vector.
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_RNG */
void SGN_RNG_perm(SGN_RNG * rng, int * ar, int n){
  int i;
  for(i = 0; i < n; i++) ar[i] = i;
  for(i = n - 1; i > 0; i--){
	int r;
	SGN_RNG_integer(rng, &r, 1, 0, i + 1);
	int t = ar[i];
	ar[i] = ar[r];
	ar[r] = t;
  }
}
/* SGN_RNG_perm() ============================================================== */

/* End of file.  Copyright (C) 2010-2012 SiGN Project Members */
