/*
  sign/math/SGN_RNG.h : Wrapper for Random Number Generators

  Copyright (C) 2010-2012  SiGN Project Members.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#ifndef __SGN_RNG_H
#define __SGN_RNG_H

#if defined(USE_FUJITSU_FCC_SSL)

#elif defined(USE_INTEL_MKL)
#include <mkl_vsl.h>
#else
#include "SGN_Random.h"
#endif

/*!\brief Structure for encapsulating the random number generator.
 *
 * Depending on the environment, this uses the appropriate implementation /
 * library to generate random numbers.  The following is the list
 * of supported environment (C macro) and its math library.
 *
 * - USE_FUJITSU_FCC_SSL : Fujitsu SSL II
 * - USE_INTEL_MKL : Intel MKL (Math Kernel Library)
 * - Others : dSFTM (double precision SIMD-oriented Fast Mersenne Twister)
 *
 *\author Yoshinori Tamada &gt;tamada at is.s.u-tokyo.ac.jp&lt;
 */
typedef struct {
#if defined(USE_FUJITSU_FCC_SSL)
  /*!\brief [SSL] Seed for uniform random number c_dvrau4(). */
  int ssl_dvrau4_seed;
  /*!\brief [SSL] Workarea for uniform random number c_dvrau4(). */
  double * ssl_dvrau4_work;
  /*!\brief [SSL] Workarea size for uniform random number c_dvrau4(). */
  int ssl_dvrau4_work_size;

  /*!\brief [SSL] Seed for normal random number c_dvran3(). */
  int ssl_dvran3_seed;
  /*!\brief [SSL] Workarea for normal random number c_dvran3(). */
  double * ssl_dvran3_work;
  /*!\brief [SSL] Workarea for normal random number c_dvran3(). */
  int ssl_dvran3_work_size;
#elif defined(USE_INTEL_MKL)
  VSLStreamStatePtr mkl_vsl_stream;
#else
  SGN_Random_dSFMT * dsfmt;
#endif
} SGN_RNG;

SGN_RNG * SGN_RNG_new();
void      SGN_RNG_delete(SGN_RNG * rng);
void      SGN_RNG_set_seed(SGN_RNG * rng, int seed);
void      SGN_RNG_normal(SGN_RNG * rng, double * ar, int n, double mean, double sigma);
void      SGN_RNG_uniform(SGN_RNG * rng, double * ar, int n, double min, double max);
void      SGN_RNG_integer(SGN_RNG * rng, int * ar, int n, int min, int max);
void      SGN_RNG_perm(SGN_RNG * rng, int * ar, int n);

#endif /* __SGN_RNG_H */

/* End of file.  Copyright (C) 2010-2012 SiGN Project Members */
