/*
  sign/math/SGN_Stat.h : Statistical Functions

  Copyright (C) 2008-2011 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

/*!\defgroup SGN_Stat
 *\brief Statistical Functions
 *\author Yoshinori Tamada &lt;tamada at ims.u-tokyo.ac.jp&gt;
 */

#include "../signcom.h"
#include "SGN_Stat.h"
#include "SGN_Math.h"

#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <math.h>

/*!\brief Calculates the mean of values.
 *
 *\param vec  array (or pointer to the first element) of values.
 *\param n    size of the array.
 *\return mean of the given values.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Stat */
double SGN_Stat_mean(double * vec, size_t n){
  double sum = 0.0;
  size_t i;
  for(i = 0; i < n; i++){
	sum += vec[i];
  }
  return sum / ((double) n);
}
/* SGN_Stat_mean() ------------------------------------------------------------- */

/*!\brief Calculates the variance of values using the specified mean.
 *
 *\param vec  array (or pointer to the first element) of values.
 *\param n    size of the array.
 *\param mean  mean of the values
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Stat */
double SGN_Stat_variance2(double * vec, size_t n, double mean){
  size_t i;
  double rss = 0.0;
  for(i = 0; i < n; i++){
	double x = vec[i] - mean;
	rss += (x * x);
  }
  return rss / ((double) n);
}
/* SGN_Stat_variance2() -------------------------------------------------------- */

/*!\brief Generates the uniformly distributed random number.
 *
 * This generates (0,1) - uniformly distributed random number using
 * <TT>random()</TT> function of the standard C library.
 * Note that this does never return 0 or 1.
 *
 *\return Uniform random number.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Stat */
double SGN_Stat_runif(){

#ifdef USE_FUJITSU_FCC_SSL
  const double max = (double) 0x7FFFFFFF;
#else
  const double max = (double) RAND_MAX;
#endif

#ifdef SIGN_DEBUG
  long r = random();
  //fprintf(SIGN_ERR, "SGN_Stat_runif(): r=%ld, RAND_MAX=%f\n", r, max);
  return (((double) r) + 0.5) * (1.0 / (max + 1.0));
#else
  return (((double) random()) + 0.5) * (1.0 / (max + 1.0));
#endif
}
/* SGN_Stat_runif() ------------------------------------------------------------ */



/*!\brief Quantile function for the normal distribution.
 *
 * The algorithm is a C translation of
 * Wichura, M.J. (1998). Algorithm AS 241: The percentage points of the
 * normal distribution. <I>Applied Statisitics</I>, Vol. 37, No. 3,
 * 477-484.
 *
 * The algorithm is the same as one used in R.
 *
 *\param p  value of lower tail area.
 *\return quantile value <I>x</I> that satisfies 
 *               <I>P</I> [<I>X</I> <= <I>x</I>] = \a p.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Stat */
double SGN_Stat_qnorm(double p){
  const double ZERO = 0.0E0;
  const double ONE = 1.0E0;
  const double HALF = ONE / 2.0E0;
  const double SPLIT1 = 0.425E0;
  const double SPLIT2 = 5.0E0;
  const double CONST1 = 0.180625E0;
  const double CONST2 = 1.6E0;

  const double A0 = 3.3871328727963666080E0;
  const double A1 = 1.3314166789178437745E2;
  const double A2 = 1.9715909503065514427E3;
  const double A3 = 1.3731693765509461125E4;
  const double A4 = 4.5921953931549871457E4;
  const double A5 = 6.7265770927008700853E4;
  const double A6 = 3.3430575583588128105E4;
  const double A7 = 2.5090809287301226727E3;
  const double B1 = 4.2313330701600911252E1;
  const double B2 = 6.8718700749205790830E2;
  const double B3 = 5.3941960214247511077E3;
  const double B4 = 2.1213794301586595867E4;
  const double B5 = 3.9307895800092710610E4;
  const double B6 = 2.8729085735721942674E4;
  const double B7 = 5.2264952788528545610E3;

  const double C0 = 1.42343711074968357734E0;
  const double C1 = 4.63033784615654529590E0;
  const double C2 = 5.76949722146069140550E0;
  const double C3 = 3.64784832476320460504E0;
  const double C4 = 1.27045825245236838258E0;
  const double C5 = 2.41780725177450611770E-1;
  const double C6 = 2.27238449892691845833E-2;
  const double C7 = 7.74545014278341407640E-4;
  const double D1 = 2.05319162663775882187E0;
  const double D2 = 1.67638483018380384940E0;
  const double D3 = 6.89767334985100004550E-1;
  const double D4 = 1.48103976427480074590E-1;
  const double D5 = 1.51986665636164571966E-2;
  const double D6 = 5.47593808499534494600E-4;
  const double D7 = 1.05075007164441684324E-9;

  const double E0 = 6.65790464350110377720E0;
  const double E1 = 5.46378491116411436990E0;
  const double E2 = 1.78482653991729133580E0;
  const double E3 = 2.96560571828504891230E-1;
  const double E4 = 2.65321895265761230930E-2;
  const double E5 = 1.24266094738807843860E-3;
  const double E6 = 2.71155556874348757815E-5;
  const double E7 = 2.01033439929228813265E-7;
  const double F1 = 5.99832206555887937690E-1;
  const double F2 = 1.36929880922735805310E-1;
  const double F3 = 1.48753612908506148525E-2;
  const double F4 = 7.86869131145613259100E-4;
  const double F5 = 1.84631831751005468180E-5;
  const double F6 = 1.42151175831644588870E-7;
  const double F7 = 2.04426310338993978564E-15;

  double Q, R, PPND16;

  Q = p - HALF;
  if(fabs(Q) <= SPLIT1){
	R = CONST1 - Q * Q;
	PPND16 = Q * (((((((A7 * R + A6) * R + A5) * R + A4) * R + A3)
      * R + A2) * R + A1) * R + A0) / (((((((B7 * R + B6) * R + B5)
      * R + B4) * R + B3) * R + B2) * R + B1) * R + ONE);
  } else {
	if(Q < 0) {
	  R = p;
	} else {
	  R = ONE - p;
	}
	if(R <= ZERO){
	  return 0;
	}
	R = sqrt(-log(R));
	if(R <= SPLIT2){
	  R = R - CONST2;
	  PPND16 = (((((((C7 * R + C6) * R + C5) * R + C4) * R
         + C3) * R + C2) * R + C1) * R + C0) / (((((((D7 * R
         + D6) * R + D5) * R + D4) * R + D3) * R + D2) * R
		 + D1) * R + ONE);
	} else {
	  R = R - SPLIT2;
	  PPND16 = (((((((E7 * R + E6) * R + E5) * R + E4) * R
		 + E3) * R + E2) * R + E1) * R + E0) / (((((((F7 * R
         + F6) * R + F5) * R + F4) * R + F3) * R + F2) * R 
		 + F1) * R + ONE);
	}
	if(Q < 0) PPND16 = - PPND16;
  }
  return PPND16;
}
/* SGN_Stat_qnorm() ------------------------------------------------------------ */

/*!\brief Vector version of SGN_Stat_qnorm().
 *
 *\param q array to which the quantiles are set.
 *\param p input values.
 *\param n number of elements in the array.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Stat */
void SGN_Stat_vqnorm(double * restrict q, 
				 const double * restrict p, const int n){
  int i;
  for(i = 0; i < n; i++){
	q[i] = SGN_Stat_qnorm(p[i]);
  }
}
/* SGN_Stat_vqnorm() ----------------------------------------------------------- */

/*!\brief Normal distribution function.
 *
 * The algorithm is based on a C tranlation of 
 * Cody, W.D. (1993). ALgorithm 715: SPECFUN - A portable FORTRAN package of
 * special function routines and test drivers. <I>ACM Transductions on 
 * Mathematical Software</I>, Vol 19, 22-32.
 *
 *\param q 
 *\return value of the normal distribution function <I>P</I> [<I>X</I> <= \a q].
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Stat */
double SGN_Stat_pnorm(const double q){
  const double THRSH = 0.67448975; /* same as in R */
  const double EPS = 1.11E-16;
  const double XLOW = -37.519E0;
  const double XUPPR = 8.572E0;
  const double ROOT32 = 5.656854248E0;
  const double SIXTEN = 1.60E1;
  const double SQRPI = 3.9894228040143267794E-1;

  const double A[] = {2.2352520354606839287E0,
					  1.6102823106855587881E2,
					  1.0676894854603709582E3,
					  1.8154981253343561249E4,
					  6.5682337918207449113E-2};
  const double B[] = {4.7202581904688241870E1,
					  9.7609855173777669322E2,
					  1.0260932208618978205E4,
					  4.5507789335026729956E4};
  const double C[] = {3.9894151208813466764E-1,
					  8.8831497943883759412E0,
					  9.3506656132177855979E1,
					  5.9727027639480026226E2,
					  2.4945375852903726711E3,
					  6.8481904505362823326E3,
					  1.1602651437647350124E4,
					  9.8427148383839780218E3,
					  1.0765576773720192317E-8};
  const double D[] =  {2.2266688044328115691E1,
					   2.3538790178262499861E2,
					   1.5193775994075548050E3,
					   6.4855582982667607550E3,
					   1.8615571640885098091E4,
					   3.4900952721145977266E4,
					   3.8912003286093271411E4,
					   1.9685429676859990727E4};
  const double P[] = {2.1589853405795699E-1,
					  1.274011611602473639E-1,
					  2.2235277870649807E-2,
					  1.421619193227893466E-3,
					  2.9112874951168792E-5,
					  2.307344176494017303E-2};
  const double Q[] = {1.28426009614491121E0,
					  4.68238212480865118E-1,
					  6.59881378689285515E-2,
					  3.78239633202758244E-3,
					  7.29751555083966205E-5};

  double X = q;
  double Y = fabs(X);
  double RESULT;
  int i;
  
  if(Y <= THRSH){
	double XSQ = 0.0;
	double XDEN;
	double XNUM;
	if(Y > EPS) XSQ = X * X;
	XNUM = A[4] * XSQ;
	XDEN = XSQ;
	for(i = 0; i < 3; i++){
	  XNUM = (XNUM + A[i]) * XSQ;
	  XDEN = (XDEN + B[i]) * XSQ;
	}

	RESULT = X * (XNUM + A[3]) / (XDEN + B[3]);
	RESULT = 0.5 + RESULT;

  } else if(Y <= ROOT32){
	double XNUM = C[8] * Y;
	double XDEN = Y;
	double XSQ, DEL;
	for(i = 0; i < 7; i++){
	  XNUM = (XNUM + C[i]) * Y;
	  XDEN = (XDEN + D[i]) * Y;
	}

	RESULT = (XNUM + C[7]) / (XDEN + D[7]);
	XSQ = floor(Y * SIXTEN) / SIXTEN;
	DEL = (Y - XSQ) * (Y + XSQ);
	RESULT = exp(-XSQ * XSQ * 0.5) * exp(-DEL * 0.5) * RESULT;

	if(X > 0.0) RESULT = 1.0 - RESULT;

  } else {
	RESULT = 0.0;
	if((X > XLOW) && (X < XUPPR)){
	  double XSQ = 1.0 / (X * X);
	  double XNUM = P[5] * XSQ;
	  double XDEN = XSQ;
	  double DEL;
	  for(i = 0; i < 4; i++){
		XNUM = (XNUM + P[i]) * XSQ;
		XDEN = (XDEN + Q[i]) * XSQ;
	  }

	  RESULT = XSQ * (XNUM + P[4]) / (XDEN + Q[4]);
	  RESULT = (SQRPI -  RESULT) / Y;
	  XSQ = floor(X * SIXTEN) / SIXTEN;
	  DEL = (X - XSQ) * (X + XSQ);
	  RESULT = exp(-XSQ * XSQ * 0.5) * exp(-DEL * 0.5) * RESULT;
	}

	if(X > 0.0) RESULT = 1.0 - RESULT;
  }

  return RESULT;
}
/* SGN_Stat_pnorm() ------------------------------------------------------------ */



/*!\brief Normal random number.
 *
 * The algorithm is based on Algorithm FL<SUB><I>m</I></SUB> (<I>m</I> = 5) in
 * Ahrens,  J.H. and Dieter, U. (1973).
 * Extensions of Forsythe's method for randome sampling from the
 * normal distribution. <I>Mathematics of Computation</I>, Vol. 27, No. 124,
 * 927-937.
 *
 * The algorithm is the same as one implemented in R as "Aherens-Dieter"
 * algorithm (not used by default).
 *
 *\return normally distributed random number.
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Stat */
double SGN_Stat_rnorm(){

  double u, uu, a, w, t, y;
  int i, s;

  double A[] = {
	0.00000000000000, 0.03917608550309, 0.07841241273311,
	0.11776987457909, 0.15731068461017, 0.19709908429430,
	0.23720210932878, 0.27769043982157, 0.31863936396437,
	0.36012989178957, 0.40225006532172, 0.44509652498551,
	0.48877641111466, 0.53340970624127, 0.57913216225555,
	0.62609901234641, 0.67448975019607, 0.72451438349236,
	0.77642176114792, 0.83051087820539, 0.88714655901887,
	0.94678175630104, 1.00999016924958, 1.07751556704027,
	1.15034938037600, 1.22985875921658, 1.31801089730353,
	1.41779713799625, 1.53412054435253, 1.67593972277344,
	1.86273186742164, 2.15387469406144};

  double T[] = {
	0.00076738283767, 0.00230687039764, 0.00386061844387,
	0.00543845406707, 0.00705069876857, 0.00870839582019,
	0.01042356984914, 0.01220953194966, 0.01408124734637,
	0.01605778804548, 0.01815290075142, 0.02039573175398,
	0.02281176732513, 0.02543407332319, 0.02830295595118,
	0.03146822492920, 0.03499233438388, 0.03895482964836,
	0.04345878381672, 0.04864034918076, 0.05468333844273,
	0.06184222395816, 0.07047982761667, 0.08113194985866,
	0.09462443534514, 0.11230007889456, 0.13649799954975,
	0.17168856004707, 0.22762405488269, 0.33049802776911,
	0.58470309390507};
  
  double H[] = {
	0.03920617164634, 0.03932704963665, 0.03950999486086,
	0.03975702679515, 0.04007092772490, 0.04045532602655,
	0.04091480886081, 0.04145507115859, 0.04208311051344,
	0.04280748137995, 0.04363862733472, 0.04458931789605,
	0.04567522779560, 0.04691571371696, 0.04833486978119,
	0.04996298427702, 0.05183858644724, 0.05401138183398,
	0.05654656186515, 0.05953130423884, 0.06308488965373,
	0.06737503494905, 0.07264543556657, 0.07926471414968,
	0.08781922325338, 0.09930398323927, 0.11555994154118,
	0.14043438342816, 0.18361418337460, 0.27900163464163,
	0.70104742502766};

  double D[] = {
	0.67448975019607, 0.47585963017993, 0.38377116397654,
	0.32861132306910, 0.29114282663980, 0.26368432217502,
	0.24250845238097, 0.22556744380930, 0.21163416577204,
	0.19992426749317, 0.18991075842246, 0.18122518100691,
	0.17360140038056, 0.16684190866667, 0.16079672918053,
	0.15534971747692, 0.15040938382813, 0.14590257684509,
	0.14177003276856, 0.13796317369537, 0.13444176150074,
	0.13117215026483, 0.12812596512583, 0.12527909006226,
	0.12261088288608, 0.12010355965651, 0.11774170701949,
	0.11551189226063, 0.11340234879117, 0.11140272044119,
	0.10950385201710};

  /* Step 1 */
  u = SGN_Stat_runif();
  if(u < 0.5){ s = 0; } else { s = 1; }
  u = 2 * u - s;

  /* Step 2 */
  u *= 32.0; /* shift m bits */
  i = (int) u; if(i == 0) goto Step9;

  /* Step 3 */
  uu = u - i;  a = A[i - 1];
  
  /* Step 4 */
 Step4:
  if(uu > T[i - 1]){ w = (uu - T[i - 1]) * H[i - 1]; goto Step17; }

  /* Step 5 */
  u = SGN_Stat_runif(); w = u * (A[i] - a);  t = (w / 2.0 + a) * w;
  
  /* Step 6 */
 Step6:
  if(uu > t) goto Step17;
  
  /* Step 7 */
  u = SGN_Stat_runif(); if(uu < u){ uu = SGN_Stat_runif(); goto Step4; }
  
  /* Step 8 */
  t = u;  uu = SGN_Stat_runif();  goto Step6;

  /* Step 9 */
 Step9:
  i = 6; a = A[31];
  
  /* Step 10 */
 Step10:
  u = 2 * u; if(u >= 1.0) goto Step12;

  /* Step 11 */
  a = a + D[i - 1]; i++; goto Step10;

  /* Step 12 */
 Step12:
  u = u - 1.0;

  /* Step 13 */
 Step13:
  w = u * D[i - 1]; t = (w / 2.0 + a) * w;

  /* Step 14 */
 Step14:
  uu = SGN_Stat_runif(); if(uu > t) goto Step17;

  /* Step 15 */
  u = SGN_Stat_runif(); if(uu < u){ u = SGN_Stat_runif(); goto Step13; }
  
  /* Step 16 */
  t = u; goto Step14;

 /* Step 17 */
 Step17:
  y = a + w; if(s == 0){ return y; } else { return -y; }
}
/* SGN_Stat_rnorm() ------------------------------------------------------------ */

/*!\brief Meta-analysis.
 *
 *\param x list of p-values.
 *\param n number of p-values in \a x.
 *\return <I>p</I>-value for the given <I>p</I>-values.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Stat */
double SGN_Stat_meta_analysis(const double * x, const int n){
  double tmp[n];
  double qnorm[n];
  double sum;
  int i;

  for(i = 0; i < n; i++) tmp[i] = 1.0 - x[i];
  SGN_Stat_vqnorm(qnorm, tmp, n);
  sum = SGN_Math_sum(qnorm, n);
  
  return 1.0 - SGN_Stat_pnorm(sum / sqrt(n));
}
/* SGN_Stat_meta_analysis() ---------------------------------------------------- */

/* End of file.  Copyright (C) 2008-2011 SiGN Project Members */
