/*
  perm_test.c : SiGN-SSM - Gene Network Estimation with State Sapce Model

  Copyright (C) 2010, 2014  SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.
  
  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#include "signssm.h"
#include "SSM.h"
#include "SSMData.h"
#include "SSMResult.h"
#include "signcom.h"
#include <io/TXTS.h>
#include <math/SGN_RNG.h>
#include <math/SGN_Matrix.h>
#include <util/SGN_Network.h>
#include <util/SGN_BS.h>

void perm_gen(int p, int samples, const double * Y, SGN_RNG * rng, double * Z){
  int perm[samples];
  for(int i = 0; i < p; i++){
	SGN_RNG_perm(rng, perm, samples);
	for(int j = 0; j < samples; j++){
	  Z[i + j * p] = Y[i + perm[j] * p];
	}
  }
}
/* perm_gen() ============================================================== */

void perm_test(Args * args, SSMData * Y){
  fprintf(SIGN_ERR, "*************************************\n");
  fprintf(SIGN_ERR, "**   Permutation Test (SGE) Mode   **\n");
  fprintf(SIGN_ERR, "*************************************\n");

  /* Read the original result */
  fprintf(SIGN_ERR, "Reading the estimated model parameters.\n");
  fflush(SIGN_ERR);
  FILE * fp = fopen(args->perm_file, "r");
  if(fp == NULL){
	fprintf(SIGN_ERR, "ERROR: Cannot open file: %s\n", args->perm_file);
	exit(1);
  }
  SSMResult orig = SSMResult_read_fp(fp);
  fclose(fp);
  int dim = orig.dim;

  if(orig.p != Y->p){
	fprintf(SIGN_ERR, "ERROR: Number of objects inconsistent: Data:%d  vs  SSM:%d\n",
			Y->p, orig.p);
	exit(1);
  }

  fprintf(SIGN_ERR, "Preparing permutation test. Y->p=%d, Y->samples=%d, dim=%d\n",
	  Y->p, Y->samples, dim);
  fflush(SIGN_ERR);
  double * Psi = malloc(sizeof(double) * Y->p * Y->p);
  double * Phi = malloc(sizeof(double) * Y->p * Y->p);
  double * D = malloc(sizeof(double) * Y->p * dim);
  double * L = malloc(sizeof(double) * dim);
  double * origY = Y->Y; /* Save the original input data matrix */
  Y->Y = malloc(sizeof(double) * Y->p * Y->samples);

  /* Prepare instances */
  SSM * ssm1 = SSM_new(Y->p, dim);
  SSM * ssm2 = SSM_new(Y->p, dim);
  SSMState * X = SSMState_new(dim, Y->N);
  SSMSuffStat * T = SSMSuffStat_new(Y->p, dim);
  SSMSuffStat * sumT = SSMSuffStat_new(Y->p, dim);
  SSMWork * work = SSMWork_new(Y->p, dim, Y->samples, args->em->constrain_F);
  if(args->em->constrain_F) SSM_EM_set_C(dim, work->C);

  SSMResult_calc_Psi(&orig, Psi, work->Eyx);

  /* File preparation */
  char buff[256];
  snprintf(buff, 256, "%s.%06d", args->output_file, args->id);
  fprintf(SIGN_ERR, "Opening the file for storing test results: %s\n", buff);
  fflush(SIGN_ERR);
  fp = fopen(buff, "w");
  if(fp == NULL){
	fprintf(SIGN_ERR, "ERROR: Cannot open file: %s\n", buff);
	exit(1);
  }

  for(int t = 0; t < args->sets; t++){
	fprintf(SIGN_ERR, "== Starting the permutation test for set %d (/%d) ==\n", 
			t + 1, args->sets);

	fprintf(SIGN_ERR, "Generating permuted data.\n"); fflush(SIGN_ERR);
	perm_gen(Y->p, Y->samples, origY, args->init->rng, Y->Y);

	/* Perform the EM algorithm */
	fprintf(SIGN_ERR, "Starting the SSM estimation.\n"); fflush(SIGN_ERR);
	SSMResult result = SSM_EM_iterate(args->em, args->init, args->iteration,
									  Y, dim, ssm1, ssm2, X, T, sumT, work);
	fprintf(SIGN_ERR, "Finished: dim=%d, likelihood=%f\n", dim, result.likelihood);
	fflush(SIGN_ERR);

	result.L = L;
	result.D = D;
	SSMResult_calc_D_L(&result, work);
	SSMResult_calc_Psi(&result, Phi, work->Eyx);

	int count = 0;
	
	/* Psi: Original one-to-one coefficients,
	   Phi: Permutaed one-to-one coefficients */
	
	if(t != 0) fprintf(fp, "-\n"); /* separator */

	/* Compares the coefficients and output */
	for(int i = 0; i < Y->p; i++){
	  fprintf(fp, "%d", i);
	  for(int j = 0; j < Y->p; j++){
		if(Psi[i + j * Y->p] < 0){ /* original: negative */
		  if(Phi[i + j * Y->p] < Psi[i + j * Y->p]){
			fprintf(fp, "\t%d", j);
			count++;
		  }
		} else { /* original: positive */
		  if(Psi[i + j * Y->p] < Phi[i + j * Y->p]){
			fprintf(fp, "\t%d", j);
			count++;
		  }
		}
	  }
	  fprintf(fp, "\n");
	}/* for i */
	fprintf(SIGN_ERR, "Ouptput: %d edges.\n", count);
  }/* for t < args->sets */
  fclose(fp);

  fprintf(SIGN_ERR, "Finished.\n");
  fflush(SIGN_ERR);
}
/* perm_test() ============================================================= */

void perm_compile(SGN_KeyValues * args, const SSMData * data, const char * output){
  char * prefix = NULL;
  double th = 0.05;
  int bg = 1;
  int ed = 1000;
  char * ssm = NULL;
  int dist = 0;
  char * suffix = "";
  int digits = 6;
  SGN_KeyValueSet argset[] = {
	{"prefix",            SGN_Type_STRING, &prefix},
	{"th",                SGN_Type_DOUBLE, &th},
	{"bg",                SGN_Type_INT,    &bg},
	{"ed",                SGN_Type_INT,    &ed},
	{"ssm",               SGN_Type_STRING, &ssm},
	{"dist",              SGN_Type_ON_OFF, &dist},
	{NULL},
  };
  SGN_Network * network = NULL;
  int r = SGN_KeyValues_set_values_kv(args, argset);
  if(r >= 0){
	fprintf(SIGN_ERR, 
			"ERROR: ssmperm: Unknown argument specified: %s\n",
			SGN_KeyValues_get_key(args, r));
	exit(1);
  }
  if(network != NULL) SGN_Network_delete(network);
  if(prefix == NULL){
	fprintf(SIGN_ERR, "ERROR: \"prefix\" not specified.\n");
	exit(1);
  }
  IOReader reader = TXTS_read_fp; /* Only supports 'TXTS' reader */
  SGN_StrArray * files = SGN_BS_gen_files(prefix, suffix, digits, bg, ed);
  SGN_Network * fnet = SGN_BS_first_network(files, reader);
  SGN_BS * bs = SGN_BS_new(SGN_Network_num_nodes(fnet));
  SGN_BS_set_SSM(bs);
  int total = TXTS_count_edges(files, bs->graph);
  fprintf(SIGN_ERR, "Processed %d networks\n", total);
  network = SGN_BS_compile_dynamic(bs, th, total, 1);

  if(ssm != NULL){
	fprintf(SIGN_ERR, "Calculating edge properties from the estimated results.\n");
	fflush(SIGN_ERR);
	int p = SGN_Network_num_nodes(network);
	FILE * fp = fopen(ssm, "r");
	if(fp == NULL){
	  fprintf(SIGN_ERR, "ERROR: Cannot open file: %s\n", ssm);
	}
	SSMResult result = SSMResult_read_fp(fp);
	fclose(fp);
	if(result.p != p){
	  fprintf(SIGN_ERR, "ERROR: Number of nodes inconsistent.\n");
	  exit(1);
	}
	double * Psi = malloc(sizeof(double) * p * p);
	double * W = malloc(sizeof(double) * p * result.dim);
	SSMResult_calc_Psi(&result, Psi, W);

	fprintf(SIGN_ERR, "  Processing edges...\n"); fflush(SIGN_ERR);

	for(int i = 0; i < p; i++){
	  const SGN_IntArray * parents = SGN_Network_get_parent_indexes(network, i);
	  for(int jj = 0; jj < SGN_IntArray_size(parents); jj++){
		int j = SGN_IntArray_get(parents, jj); /* j -> i */
		SGN_Edge * edge = SGN_Network_get_edge_index(network, j, i);
		double pv = SGN_Edge_get_property_double(edge, "SSM.P", 1.0);
		double c = Psi[i + j * p];
		SGN_Edge_add_property_double(edge, "SSM.coeff", c);
		SGN_Edge_add_property_double(edge, "edgeScore", 1.0 - pv);
		if(c > 0){
		  /* Up */
		  SGN_Edge_add_property_int(edge, "EdgeType", 0);
		  SGN_Edge_add_property_string(edge, "Up/Down", "up");
		  SGN_Edge_set_color(edge, 153, 102, 255, 255);
		} else {
		  SGN_Edge_add_property_int(edge, "EdgeType", 1);
		  SGN_Edge_add_property_string(edge, "Up/Down", "down");
		  SGN_Edge_set_color(edge, 0, 124, 0, 255);
		}
	  }
	}
	free(Psi);
	free(W);
  }/* if(ssm != NULL) */

  /* Set node names */
  for(int i = 0; i < SGN_Network_num_nodes(network); i++){
	SGN_Node * node = SGN_Network_get_node(network, i);
	SGN_Node_set_name(node, data->names[i]);
	SGN_Node_set_alias(node, data->names[i]);
  }

  /* output */
  if(output == NULL){
	fprintf(SIGN_ERR, "WARNING: Output file not specified (-o).\n");
	return;
  }
  FILE * fp = fopen(output, "w");
  if(fp == NULL){
	fprintf(SIGN_ERR, "WARNING: Cannot open file: %s\n", output);
	return;
  }

  for(int i = 0; i < SGN_Network_num_edges(network); i++){
	SGN_Edge * edge = SGN_Network_get_edge(network, i);
	SGN_Node * pa = SGN_Edge_get_parent(edge);
	SGN_Node * ch = SGN_Edge_get_child(edge);

	double pv = SGN_Edge_get_property_double(edge, "SSM.P", NAN);
	if(isnan(pv)){
	  fprintf(SIGN_ERR, "ERROR: Cannot find \"SSM.P\" property.\n");
	  exit(1);
	}
	double co = SGN_Edge_get_property_double(edge, "SSM.coeff", NAN);
	if(isnan(pv)){
	  fprintf(SIGN_ERR, "ERROR: Cannot find \"SSM.coeff\" property.\n");
	  exit(1);
	}
	const char * ud = SGN_Edge_get_property_string(edge, "Up/Down");
	if(ud == NULL){
	  fprintf(SIGN_ERR, "ERROR: Cannot find \"Up/Down\" property.\n");
	  exit(1);
	}

	fprintf(fp, "%s\t%s\t%f\t%f\t%s\n", 
			SGN_Node_name(pa), SGN_Node_name(ch), pv, co, ud);
  }

  fclose(fp);
}
/* perm_compile() ========================================================== */

/* End of file.  Copyright (C) 2010, 2014  SiGN Project Members */
