/*
  pvalues.c

  Copyright (C) 2010  SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  
  Visit <http://sign.hgc.jp/signssm/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#include "SSM.h"
#include "SSMData.h"

#include "signcom.h"
#include <math/SGN_Stat.h>
#include <math/SGN_Matrix.h>

#include <math.h>
#include <float.h>

/*!\brief Calculates p-values
 *
 * This is available only when HGC macro is defined.
 *
 *\param ssm Estimated state space model.
 *\param X   Calculated state variables.
 *\param Y   Input data matrix.
 *\param k   Size of dimensions.
 *\param r   Replicate ID.
 *\param y  [INPUT] (\a p)-vector.  If NULL, calculate in this routine.
 *\param PV [OUTPUT] (\a p, \a n)-matrix where \a n is the number of samples in
 *           the \a r -th replicate. \a PV[\a i + \a j * \a p] corresponds to the p values
 *           of the i -th at the j -th sample where \a p represents the number of objects.
 *\param W [WORK] (\a p, \a k)-matrix.
 *\param y [WORK] (\a p)-vector.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SiGNSSM */
double * calc_pvalues(SSM * ssm, SSMState * X, SSMData * Y, int k, int r,
					  double * PV, double * W, double * y){

  for(int n = 0; n < Y->N; n++){
	int j = Y->table[r][n];
	if(j == -1) continue;
	
	/* y := H x.  */
	SGN_Matrix_dgemv(SGN_Matrix_NO_TRANS, &Y->p, &k, &SGN_Matrix_DOUBLE_ONE, ssm->H, &Y->p,
					 X->predict_exp[n], &SGN_Matrix_INT_ONE, &SGN_Matrix_DOUBLE_ZERO,
					 y, &SGN_Matrix_INT_ONE);

	/* Sigma :=  H     V_n     H'  +  R     */
	/* (p,p)     (p,k) (k,k) (k,p)  (p,p) */

	/* Only diagonal elements of Sigma are required. */
	/* Thus, we don't calculate H V_n H' directly.   */
	/* Instead, because the diagonal elements of H V_n H' are
	   the dot products of rows of (H V) and columns of H'. */
	
	/*    W    := H    V */
	/*   (p,k)   (p,k) (k,k) */
	SGN_Matrix_dgemm(SGN_Matrix_NO_TRANS, SGN_Matrix_NO_TRANS, &Y->p, &k, &k,
				 &SGN_Matrix_DOUBLE_ONE, ssm->H, &Y->p, X->predict_var[n], &k,
				 &SGN_Matrix_DOUBLE_ZERO, W, &Y->p);

	int m = 0;
	for(int i = 0; i < Y->p; i++){
	  if(i == Y->missing[j][m]){
		*PV = NAN;
		m++;
	  } else {
		/*   s  := W_(i) .  H'^(i) + R_i   (".": dot product)*/
		double s = SGN_Matrix_ddot(&k, W + i, &Y->p, ssm->H + i, &Y->p) + ssm->R[i];
		double z = (Y->Y[i + j * Y->p] - W[i]) / sqrt(s); /* z ~ N(0,1) */
		*PV = (1.0 - SGN_Stat_pnorm(fabs(z))) * 2.0; /* Double sided test */
	  }
	  PV++;
	}/* for i < Y->p */
  }/* for n < Y->N */
  return PV;
}
/* calc_pvalues() ========================================================== */

/*!\brief Output the calculated p-values.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SiGNSSM */
void output_pvalues(const char * prefix, int k, int id, const char * postfix,
					int p, int n, double * PV){
  char buff[SSM_BUFF_SIZE];
  int len = snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.P.%s", prefix, k, id, postfix);
  if(SSM_BUFF_SIZE <= len){
	fprintf(SIGN_ERR, "ERROR: Output prefix and/or postfix too long.\n");
	exit(1);
  }
  fprintf(SIGN_ERR, "Outputting the p-values: %s\n", buff);
  FILE * fp = fopen(buff, "w");
  if(fp == NULL){
	fprintf(SIGN_ERR, "WARNING: Cannot open file: %s\n", buff);
	return;
  }
  SGN_Matrix_fprintf(fp, 18, PV, p, n);
  fclose(fp);

  /* Perform Meta Analysis */
  snprintf(buff, SSM_BUFF_SIZE, "%s.D%03d.S%03d.m.%s", prefix, k, id, postfix);
  fprintf(SIGN_ERR, "Outputting the meta-analysis of p-values: %s\n", buff);
  fp = fopen(buff, "w");
  if(fp == NULL){
	fprintf(SIGN_ERR, "WARNING: Cannot open file: %s\n", buff);
	return;
  }
  double pvs[n];
  for(int j = 0; j < p; j++){
	/* Copy the vector of p-values. */
	int nn = 0;
	for(int i = 0; i < n; i++){
	  if(!isnan(PV[j + i * p])) pvs[nn++] = PV[j + i * p];
	}
	double pv = SGN_Stat_meta_analysis(pvs, nn);
	fprintf(fp, "%.18f\n", pv);
  }/* for j < p */
  fclose(fp);
}
/* output_pvalues() ======================================================== */

/* End of file.  Copyright (C) 2010  HGC, IMS, The University of Tokyo       */
