/*
  signssm.c : SiGN-SSM - Gene Network Estimation with State Sapce Model

  Copyright (C) 2010 - 2014
    Yoshinori Tamada, Rui Yamaguchi, Seiya Imoto, Osamu Hirose, Ryo Yoshida,
	Masao Nagasaki and Satoru Miyano
      Laboratory of DNA Information Analysis &
	  Laboratory of Sequence Analysis, Human Genome Center,
      The Institute of Medical Science, The Unversity of Tokyo

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  
  Visit <http://sign.hgc.jp/signssm/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
  Originally developed by Osamu Hirose, Ryo Yoshida, Seiya Imoto, Rui Yamaguchi,
                          Tomoyuki Higuchi, and Satoru Miyano.
*/

#include "signcom.h"
#include "util/SGN_KeyValues.h"
#include <math/SGN_RNG.h>

#include "signssm.h"
#include "dispatch.h"
#include "execute.h"
#include "single.h"
#include "thread.h"
#include "Kalman.h"
#include "SSM_Defaults.h"
#include "SSMInit.h"
#include "perm_test.h"

#ifdef SSM_DEBUG
#include "debug.h"
#endif

#include <string.h>
#ifdef USE_OMP
#include <omp.h>
#endif

#ifdef USE_EDF
#include "read_EDF.h"
#endif

#define SIGNSSM_VERSION "1.2.1"
#define SIGNSSM_DATE    "Fri Dec 19 17:15:38 2014 JST"

/*!\defgroup SiGNSSM SiGNSSM
 *
 */

/* Proto-types for main() */
Args *    Args_new();
SSMData * read_file(const char * input_file);
void      parse_args(int argc, char ** argv, Args * args);
void      apply_Kalman(const char * input, SSMData * Y, int infer,
					   const char * prefix, const char * postfix, int id,
					   int each, int pv);

/* ========================================================================= */
/*  M A I N   R O U T I N E                                                  */
/* ========================================================================= */
/*!\brief Main routine.
 *
 *\param argc  Number of command line arguments.
 *\param argv  Array of command line arguments.
 *\returns 0 if successfully finished.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SiGNSSM */
int main(int argc, char ** argv){
  SGN_Timer * all_timer = SGN_Timer_new();
  SGN_init();
  SGN_Timer_start(all_timer);

#ifdef USE_MPI
  MPI_Init(&argc, &argv);
#endif

  /* Parse command line arguments. */
  Args * args = Args_new();
  parse_args(argc, argv, args);

  /* Outpu the title message */
  fprintf(SIGN_ERR, "SiGN-SSM  version %s (%s)\n", SIGNSSM_VERSION, SIGNSSM_DATE);
  fprintf(SIGN_ERR, "  Copyright (C) 2010-2014  SiGN Project Members.\n\n");
  fprintf(SIGN_ERR, "! ABSOLUTELY NO WARRANTY.  SEE LICENSE FOR DETAILS.  Visit http://sign.hgc.jp/\n\n");
#ifdef USE_EDF
  print_EDF_support(SIGN_ERR);
#endif /* USE_EDF */
  if(args->MPI_size == 1){
	fprintf(SIGN_ERR, "Single process mode.\n");
  } else if(args->MPI_rank == 0){
	fprintf(SIGN_ERR, "MPI mode: This is the root process.\n");
  } else {
	fprintf(SIGN_ERR, "MPI mode: This is an execution process (rank=%d)\n", args->MPI_rank);
  }
  fflush(SIGN_ERR);

  /* Initialize the random number generator. */
  SGN_RNG_set_seed(args->init->rng, args->seed + args->id + args->MPI_rank);

  /* Read the input file. */
  SSMData * Y = NULL;
  if(args->MPI_rank == 0){
#ifdef USE_EDF
	Y = read_EDF(args->input_file, args->input_args);
#else
	Y = read_file(args->input_file);
#endif
	if(args->mean_shift > 0){
	  fprintf(SIGN_ERR, "Applying mean shift for the input data set.\n");
	  SSMData_node_mean_shift(Y, args->mean_shift == 1 ? 1 : 0);
	}
  }

  if(args->MPI_size == 1){
	/* Single execution mode */

	if(args->perm_file != NULL){
	  /* permutation test */
	  perm_test(args, Y);
	} else if(args->ssmperm != NULL){
	  /* compilation of permutation test results */
	  perm_compile(args->ssmperm, Y, args->output_file);
	} else if(args->ssm_file != NULL){
	  /* Read the SSM model parameter file and calculate
		 the state variables. */
		apply_Kalman(args->ssm_file, Y, args->em->update_mu0, 
				   args->output_file, args->output_ext, args->id,
				   args->output_each, args->output_pvalues);
	} else if(args->sge){
	  /* SGE mode */
	  single_set(args, Y);
#ifdef USE_OMP
	} else if(args->threads != 1) {
	  /* Multi-thread mode */
	  thread(args, Y);
#endif /* USE_OMP */
	} else {
	  /* Normal single process (thread) mode. */
	  single_one(args, Y);
	}
  } else {
#ifdef USE_MPI
	/* Parallel execution mode */
	fprintf(SIGN_ERR, "Broadcasting the input data.\n"); fflush(SIGN_ERR);
	SSMData_MPI_Bcast(&Y, 0, MPI_COMM_WORLD);
	fprintf(SIGN_ERR, "  Done.\n");	fflush(SIGN_ERR);
	
	if(args->MPI_rank == 0){
	  dispatch_jobs(args);
	} else {
	  execute_jobs(args, Y);
	}
#else /* USE_MPI */
	fprintf(SIGN_ERR, "ERROR: MPI support not enabled.\n");
	exit(1);
#endif /* USE_MPI */
  }/* if(args->MPI_size ==1) .. else .. */
  
#ifdef USE_MPI
  MPI_Finalize();
#endif

  double time = SGN_Timer_get(all_timer);
  fprintf(SIGN_ERR, "All time: ");
  SGN_Timer_print_DDHHMMSS_sec(SIGN_ERR, time);
  fprintf(SIGN_ERR, "\nX-ALL-TIME:\t%f\n", time);

  return 0;
}
/* main() ================================================================== */

/* ========================================================================= */
/*  S U B   R O U T I N E S                                                  */
/* ========================================================================= */

/*!\brief  Allocates the arugment container and initializes it by default values.
 *
 * Call parse_args() to parse the command line arguments and set
 * them to this.
 *
 *\sa parse_args()
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SiGNSSM */
Args * Args_new(){
  Args * args = malloc(sizeof(Args));

  args->output_file = "result";
  args->output_ext = "dat";
  args->input_file = NULL;
  args->log_mode = 0;
  //args->log_mode = 1;  /* For debug */
  args->sge = 0;
  args->output_each = 0;
  args->output_state = 1;
  args->mean_shift = 1;
  args->ssm_file = NULL;
  args->add_proc_file = 0;
  args->perm_file = NULL;
  args->ssmperm = NULL;
  
  /* Set default values */
  args->id           = 1;
  args->seed         = SSM_DEFAULT_SEED;
  args->dims         = malloc(sizeof(int));
  args->dims[0]      = 4;
  args->len_dims     = 1;
  args->sets         = SSM_DEFAULT_SETS;
  args->iteration    = SSM_DEFAULT_ITERATION;

  args->em = SSM_EM_new();

  args->init = SSMInit_new();

  args->threads = 1;
  args->gsl = 0;
  args->debug = 0;

#ifdef USE_MPI
  MPI_Comm_rank(MPI_COMM_WORLD, &args->MPI_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &args->MPI_size);
#else
  args->MPI_rank = 0;
  args->MPI_size = 1;
#endif /* USE_MPI */

  args->output_pvalues = 1;

#ifdef USE_EDF
  args->input_args = NULL;
#endif

  return args;
}
/* Args_new() ============================================================== */

/*!\brief Prints the simple help message.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SiGNSSM */
void show_help(){
  fprintf(SIGN_ERR, 
		  "SiGN-SSM  version " SIGNSSM_VERSION " (" SIGNSSM_DATE ")\n"
		  "  Copyright (C) 2010-2014 SiGN Project Members.\n\n"
		  "Usage:\n"
		  "  [mpirun [mpi options]] signssm [options] file\n\n"
		  "Options: (default values)\n"
		  "  -d X[-Y:[Z]]  Region of dimensions. X, Y, and Z are integer values. (4)\n"
		  "  -h            Show this help.\n"
		  "  -i N          ID to distinguish concurrent executions. (N=1)\n"
		  "  -l N          Loop limit in the EM algorithm. (N=%d)\n", SSM_DEFAULT_EM_LOOPS);
  fprintf(SIGN_ERR,
		  "  -L {0|1|2}    Log mode. 0: auto, 1: all, 2: stderr. (0)\n"
		  "  -r N          Random seed. (N=%d)\n", SSM_DEFAULT_SEED);
  fprintf(SIGN_ERR,
		  "  --shift {0|1|2}\n"
		  "                Mean shift mode. \n"
		  "                  0: do nothing, 1: shift for replicates, 2: shift for all. (1)\n"
		  "  --sge         SGE mode. (not specified)\n"
		  "  --ssm file\n"
		  "                Read SSM model file to calculate state variables.\n"
		  "  -s N          Number of models (sets) for one size of dimensions. (N=1)\n"
#ifdef USE_OMP
		  "  --threads N   Number of threads. (1)\n"
#endif
		  "  -e STR        Output file name suffix. (STR=\"dat\")\n"
		  "  --each {on|off}\n"
		  "                Output resultant matrices/vectors in separate files. (off)\n"
		  "  -o STR        Output file name prefix. (STR=\"result\")\n"
		  "  --proc-file | -P\n"
		  "                Insert the number of processes to the output file prefix. (not\n"
		  "                 specified)\n"
		  "  --pvalues {on|off}\n"
		  "                Output p-values. (on)\n"
		  "  --state {on|off}\n"
		  "                Output the estimated state variables. (on)\n"
		  "  -n N          Iterations (num. of inital values) for EM. (N=%d)\n", SSM_DEFAULT_ITERATION);
  fprintf(SIGN_ERR,
		  "  -l N | --em-loop N\n"
		  "                Maximum limit of EM loops to converge. (N=20000)\n"
		  "  --constrain-F {on|off} | -F {on|off}\n"
		  "                Constrain F. (on)\n"
		  "  --constrain-Fg X | -g X\n"
		  "                Strength of contraining F. 0 < X < 1. (X=1.0)\n"
		  "  --update-mu {on|off}\n"
		  "                Update mu_0 or not. (on)\n"
		  "  -c X | --converge X\n"
		  "                Condition of convergence. (X=1.0E-5)\n"
		  "  -R {on|off} | --RrI {on|off}\n"
		  "                Assume R=rI or diagonal. R: system noise. (off)\n"
		  "  --retry N     Maximum retry count. -1 for unlimited retry. (-1)\n"
		  "  --abs-converge {on|off}\n"
		  "                Use absolute convergence condition. (on)\n"
		  "  --ridge X     Ridge. (X=0.0)\n"
		  "  --H-mean X    Mean of H (elementwise Gaussian) (X=0.0)\n"
		  "  --H-SD X      S.D. of H (elementwise Gaussian) (X=1.0)\n"
		  "  --F-max X     Upper bound of F (uniform distribution) (X=1.5)\n"
		  "  --F-min X     Lower bound of F (uniform distribution) (X=-1.5)\n"
		  "  --mu X        Mean mu_0 (diagonal-elementwise Gaussian) (X=0.0)\n"
		  "  --SD X        S.D sd_0 (diagonal-elementwise Gaussian) (X=1.0)\n"
		  "  --perm file   Permutation test mode. (not specified)\n"
		  "  --ssmperm args...\n"
		  "                Compiling permutation results.\n"
		  );
  exit(0);
}
/* show_help() ============================================================= */

/*!\brief Parses command line arguments.
 *
 *\param argc  Number of arguments in \a argv, given by the main() function.
 *\param argv  Null terminated array of character strings containing command line
 *          arguments.
 *\param args [OUTPUT] Arguments container in which the parsed arguments are set.
 *
 *\sa Args_new().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SiGNSSM */
void parse_args(int argc, char ** argv, Args * args){
  int * dims = NULL;
  int dims_size = 0;
  
  SGN_KeyValueSet arg_set[] = {
	{SGN_KeyValues_DEFAULT_KEY,
                       SGN_Type_STRING,     &args->input_file},
	{"-i",             SGN_Type_INT,        &args->id},
	{"-d",             SGN_Type_INT_REGION, &dims, &dims_size},
	{"-s",             SGN_Type_INT,        &args->sets},
	{"-n",             SGN_Type_INT,        &args->iteration},
	{"-o",             SGN_Type_STRING,     &args->output_file},
	{"-e",             SGN_Type_STRING,     &args->output_ext},
	{"--each",         SGN_Type_ON_OFF,     &args->output_each},
	{"-E",             SGN_Type_ON_OFF,     &args->output_each},
	{"-L",             SGN_Type_INT,        &args->log_mode},
	{"-r",             SGN_Type_INT,        &args->seed},
	{"--sge",          SGN_Type_ON,         &args->sge},
	{"--gsl",          SGN_Type_ON,         &args->gsl},
	{"--state",        SGN_Type_ON_OFF,     &args->output_state},
	{"--shift",        SGN_Type_INT,        &args->mean_shift},
	{"--threads",      SGN_Type_INT,        &args->threads},
	{"--pvalues",      SGN_Type_ON_OFF,     &args->output_pvalues},
	{"--converge",     SGN_Type_DOUBLE,     &args->em->converge},
	{"-c",             SGN_Type_DOUBLE,     &args->em->converge},
	{"--EM-loop",      SGN_Type_INT,        &args->em->loops},
	{"--em-loop",      SGN_Type_INT,        &args->em->loops},
	{"-l",             SGN_Type_INT,        &args->em->loops},
	{"--RrI",          SGN_Type_ON_OFF,     &args->em->R_rI},
	{"-R",             SGN_Type_ON_OFF,     &args->em->R_rI},
	{"--update-mu",    SGN_Type_ON_OFF,     &args->em->update_mu0},
	{"--abs-converge", SGN_Type_ON_OFF,     &args->em->abs_converge},
	{"--ridge",        SGN_Type_DOUBLE,     &args->em->ridge},
	{"--constrain-F",  SGN_Type_ON_OFF,     &args->em->constrain_F},
	{"-F",             SGN_Type_ON_OFF,     &args->em->constrain_F},
	{"--constrain-Fg", SGN_Type_DOUBLE,     &args->em->g},
	{"-g",             SGN_Type_DOUBLE,     &args->em->g},
	{"--retry",        SGN_Type_INT,        &args->em->retry},
	{"--verbose",      SGN_Type_INT,        &args->em->verbose},
	{"-v",             SGN_Type_INT,        &args->em->verbose},

	{"--H-mean",       SGN_Type_DOUBLE,     &args->init->H_mean},
	{"--H-SD",         SGN_Type_DOUBLE,     &args->init->H_SD},
	{"--F-max",        SGN_Type_DOUBLE,     &args->init->F_max},
	{"--F-min",        SGN_Type_DOUBLE,     &args->init->F_min},
	{"--mu",           SGN_Type_DOUBLE,     &args->init->mu0},
	{"--SD",           SGN_Type_DOUBLE,     &args->init->SD0},

	{"--ssm",          SGN_Type_STRING,     &args->ssm_file},
	{"--proc-file",    SGN_Type_ON,         &args->add_proc_file},
	{"-P",             SGN_Type_ON,         &args->add_proc_file},

	{"--debug",        SGN_Type_INT,        &args->debug},

	{"-h",             SGN_Type_FUNC,       show_help},
	{"--perm",         SGN_Type_STRING,     &args->perm_file},
	{"--ssmperm",      SGN_Type_KEY_VALUES, &args->ssmperm},
#ifdef USE_EDF
	{"-I",             SGN_Type_STRING,     &args->input_args},
#endif
	{NULL},
  };

  int r = SGN_KeyValues_set_args(argc, argv, arg_set);
  if(r > 0){
	fprintf(SIGN_ERR, "ERROR: Unknown argument or error found at #%d \"%s\"\n", r, argv[r]);
	show_help();
  }

  if(dims != NULL){
	free(args->dims);
	args->dims = dims;
	args->len_dims = dims_size;
  }

  if(args->add_proc_file){
	size_t len = strlen(args->output_file) + 10;
	char * file = malloc(sizeof(char) * len);
	snprintf(file, len, "%s.P%04d", args->output_file, args->MPI_size);
	args->output_file = file;
  }

  /* Check the length of the output file name */
  if(strlen(args->output_file) + strlen(args->output_ext) > SSM_BUFF_SIZE - 50){
	fprintf(SIGN_ERR, "ERROR: Output file name too long.\n");
	exit(1);
  }

  /* Check the range of g (constraint parameter). */
  if(args->em->g < 0.0 || 1.0 < args->em->g){
	fprintf(SIGN_ERR, "ERROR: Value for -g option neends to be from 0 to 1.\n");
	exit(1);
  }
  if(args->em->g == 1.0){
	args->em->g -= 1.0E-4;
  }

  /* log file */
  FILE * fp;
  char buff[SSM_BUFF_SIZE];
  if(args->log_mode == 0){
	/* auto mode */
	if(args->sge || args->perm_file != NULL){
	  /* Single SGE/permutation mode */
	  snprintf(buff, SSM_BUFF_SIZE, "%s.log.%06d", args->output_file, args->id);
	  fp = fopen(buff, "w");
	} else if(args->MPI_rank == 0 && args->MPI_size == 1 && !args->sge){
	  /* Single process */
	  snprintf(buff, SSM_BUFF_SIZE, "%s", "<stderr>");
	  fp = stderr;
	} else if(args->log_mode == 0 && args->MPI_rank == 0){
	  /* MPI execution */
	  snprintf(buff, SSM_BUFF_SIZE, "%s.log", args->output_file);
	  fp = fopen(buff, "w");
	} else {
	  snprintf(buff, SSM_BUFF_SIZE, "%s", "/dev/null");
	  fp = fopen(buff, "w");
	}
  } else if(args->log_mode == 1){
	/* Force to output all processes to files */
	if(args->MPI_size == 1 && args->perm_file != NULL){
	  /* Single SGE/permutation mode */
	  snprintf(buff, SSM_BUFF_SIZE, "%s.log.%06d", args->output_file, args->id);
	} else {
	  snprintf(buff, SSM_BUFF_SIZE, "%s.log.%06d", args->output_file, args->MPI_rank);
	}
	fp = fopen(buff, "w");
  } else if(args->log_mode == 2){
	/* Force to output to standard error */
	snprintf(buff, SSM_BUFF_SIZE, "%s", "<stderr>");
	fp = stderr;
  } else {
	/* Do not output */
	snprintf(buff, SSM_BUFF_SIZE, "%s", "/dev/null");
	fp = fopen(buff, "w");
  }
  if(fp == NULL){
	fprintf(SIGN_ERR, "ERROR: Cannot open log file: %s\n", buff);
  }
  SIGN_ERR = fp;
}
/* parse_args() ============================================================ */

/*!\brief Reads the input data from a file.
 *
 * The input file format is a subset version of EDF (Expression Data Format)
 * developed in the author's laboratory.   See other documents for
 * details about this file format.
 *
 *\param input_file  Input file name.  NULL is acceptable.  If so,
 *   reads the input data from the standard input.
 *\return SSMData instance read from the file.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SiGNSSM */
SSMData * read_file(const char * input_file){
  /* Read the input file. */
  SSMData * Y = NULL;
  FILE * fp = stdin;
  if(input_file != NULL){
	fprintf(SIGN_ERR, "Reading input data file: %s\n", input_file);
	fp = fopen(input_file, "r");
	if(fp == NULL){
	  fprintf(SIGN_ERR, "ERROR: Cannot open file: %s\n", input_file);
	  exit(1);
	}
  } else {
	fprintf(SIGN_ERR, "Reading input data from the standard input.\n");
  }
  Y = SSMData_read_fp(fp);
  if(fp != stdin) fclose(fp);
  
  if(Y == NULL){
	fprintf(SIGN_ERR, "ERROR: Failed to read the input file.\n");
	fflush(SIGN_ERR);
	exit(1);
  }
  
  return Y;
}
/* read_file() ============================================================= */

/*!\brief Applies the Kalman filter, prediction and smoother for calculating
 *   filter, prediction and smooth distribution.
 *
 *\param input    Input SSM model parameters file.
 *\param dataset  Input time series dataset.
 *\param infer  If this is non-zero,
 *   the initial values are inferred by the SSM model from the original
 *   SSM::exp and SSM::var using the input data.   Otherwise, uses the
 *   SSM::exp and SSM::var read from file are directly used
 *   as the initial values.
 *\param prefix   Output file prefix.
 *\param postfix  Output file postfix.
 *\param id       ID to be inserted to the output file.
 *\param each     If non-zero, output state variables in each file.
 *\param pv       If non-zero, output p-values.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SiGNSSM */
void apply_Kalman(const char * input, SSMData * Y, int infer,
				  const char * prefix, const char * postfix, int id,
				  int each, int pv){
  
  /* Read SSM model parameters */
  int p, k;

  fprintf(SIGN_ERR, "Reading the SSM model parameters from the file: %s\n", input);
  fflush(SIGN_ERR);
  FILE * fp = fopen(input, "r");
  SSM * ssm = SSM_read_fp(fp, &p, &k);
  fclose(fp);
  fprintf(SIGN_ERR, "  Done. p = %d, dim = %d.\n", p, k); fflush(SIGN_ERR);
  if(p != Y->p){
	fprintf(SIGN_ERR, "ERROR: Number of objects inconsistent.\n");
	fprintf(SIGN_ERR, "     :  input data: %d  vs  input SSM: %d\n", Y->p, p);
	exit(1);
  }

  /* Generate the required instances */
  fprintf(SIGN_ERR, "--- INPUT DATA ---------\n");
  fprintf(SIGN_ERR, " observed samples = %d\n", Y->samples);
  fprintf(SIGN_ERR, " objects = %d\n", Y->p);
  fprintf(SIGN_ERR, " time points = %d\n", Y->N);
  fprintf(SIGN_ERR, " replicates = %d\n", Y->replicates);
  fprintf(SIGN_ERR, "------------------------\n");

  SSMState * X = SSMState_new(k, Y->N);
  SSMWork * work = SSMWork_new(Y->p, k, Y->samples, 0);

  if(infer){
	/* To infer the initial value x0, apply Kalman filter and
	   smoother. */
	fprintf(SIGN_ERR, "Applying Kalman filter and smoother for estimating"
			" the initial state variable.\n"); fflush(SIGN_ERR);
	Kalman_infer_mu0(ssm, X, k, Y, work);
  }
  fprintf(SIGN_ERR, "Calculating state variables and outputting files...\n"); 
  fflush(SIGN_ERR);
  if(each){
	Kalman_apply_output_each(ssm, k, Y, X, work, prefix, postfix, id, pv);
  } else {
	Kalman_apply_output(ssm, k, Y, X, work, prefix, postfix, id, pv);
  }
  /* It is better to free resources allocated for Y, X, and work, here.
	 But the current implementation does not do it. */
}
/* apply_Kalman() ========================================================== */

/*!\brief General routine to output the estimation result.
 *
 *\param result  SSM Estimation result to output.
 *\param work    Working area used to generate \a result.
 *\param X       State variables to output if \a state is non-zero.
 *\param Y       Input data to output if \a state is non-zero.
 *\param prefix  Output file name prefix.
 *\param postfix Output file name postfix (extension).
 *\param id      ID number included in the file name.
 *\param each    If non-zero, outputs matrices/vectors of the result in separate files.
 *\param state   If non-zero, outputs state variables.
 *\param pv      If non-zero, output p-values.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SiGNSSM */
void ssm_output(SSMResult * result, SSMWork * work, SSMState * X, SSMData * Y,
				const char * prefix, const char * postfix, int id, int each, int state, int pv){
  SSMResult_alloc(result, result->p, result->dim);
  SSMResult_calc_D_L(result, work);

  if(state){
	if(each){
	  Kalman_apply_output_each(result->ssm, result->dim, Y, X, work, 
							   prefix, postfix, id, pv);
	} else {
	  Kalman_apply_output(result->ssm, result->dim, Y, X, work, prefix, postfix, id, pv);
	}
  }
  if(each){
	SSMResult_output_each(result, prefix, postfix, id);
  } else {
	SSMResult_output(result, prefix, postfix, id);
  }
  SSMResult_free(result);
}
/* ssm_output() ============================================================ */

/* End of file.  Copyright (C) 2010-2014  HGC, IMS, The University of Tokyo  */
