/*
  single.c : SiGN-SSM - Gene Network Estimation with State Sapce Model

  Copyright (C) 2010
    Yoshinori Tamada, Rui Yamaguchi, Seiya Imoto, Osamu Hirose, Ryo Yoshida,
	Masao Nagasaki and Satoru Miyano
      Laboratory of DNA Information Analysis &
	  Laboratory of Sequence Analysis, Human Genome Center,
      The Institute of Medical Science, The Unversity of Tokyo

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  
  Visit <http://sign.hgc.jp/signssm/> for more information.

  Written by Yoshinori Tamada <tamada at ims.u-tokyo.ac.jp>
  Originally developed by Osamu Hirose, Ryo Yoshida, Seiya Imoto, Rui Yamaguchi,
                          Tomoyuki Higuchi, and Satoru Miyano.
*/

#include "signcom.h"

#include "single.h"
#include "dispatch.h"
#include "execute.h"

#include "SSMData.h"
#include "SSM.h"
#include "Kalman.h"

#ifdef SSM_DEBUG
#include "debug.h"
#endif

void single_one(Args * args, SSMData * Y){
  fprintf(SIGN_ERR, "*******************************\n");
  fprintf(SIGN_ERR, "**   Single Execution Mode   **\n");
  fprintf(SIGN_ERR, "*******************************\n");

#ifdef SSM_DEBUG
  debug_init();
#endif

  fprintf(SIGN_ERR, "--- INPUT DATA ---------\n");
  fprintf(SIGN_ERR, " observed samples = %d\n", Y->samples);
  fprintf(SIGN_ERR, " objects = %d\n", Y->p);
  fprintf(SIGN_ERR, " time points = %d\n", Y->N);
  fprintf(SIGN_ERR, " replicates = %d\n", Y->replicates);
  fprintf(SIGN_ERR, "------------------------\n");
  //SSMData_dump(Y, SIGN_ERR);

  //fprintf(SIGN_ERR, "DEBUG: len_dims=%d\n", args->len_dims);
  //for(int i = 0; i < args->len_dims
  int max_dim = int_ar_max(args->dims, args->len_dims);
  fprintf(SIGN_ERR, "Max dimension = %d\n", max_dim);

  /* Prepare instances */
  SSM * ssm1 = SSM_new(Y->p, max_dim);
  SSM * ssm2 = SSM_new(Y->p, max_dim);
  SSMState * X = SSMState_new(max_dim, Y->N);
  SSMSuffStat * T = SSMSuffStat_new(Y->p, max_dim);
  SSMSuffStat * sumT = SSMSuffStat_new(Y->p, max_dim);
  SSMWork * work = SSMWork_new(Y->p, max_dim, Y->samples, args->em->constrain_F);

  SGN_Timer * timer = SGN_Timer_new();

  /* Visits all the dimensions in "args->dims". */
  for(int i = 0; i < args->len_dims; i++){
	int dim = args->dims[i];
	fprintf(SIGN_ERR, "========================================\n");
	fprintf(SIGN_ERR, "= Start the loop for dim=%d\n", dim); fflush(SIGN_ERR);
	if(args->em->constrain_F) SSM_EM_set_C(dim, work->C);

	for(int j = 0; j < args->sets; j++){
	  fprintf(SIGN_ERR, "Set %d...\n", j + 1); fflush(SIGN_ERR);
	  SGN_Timer_clear(timer);
	  SGN_Timer_start(timer);

	  SSMResult result = SSM_EM_iterate(args->em, args->init, args->iteration,
										  Y, dim, ssm1, ssm2, X, T, sumT, work);

	  SGN_Timer_stop(timer);
	  fprintf(SIGN_ERR, "Finished: dim=%d, likelihood=%f\n", dim, result.likelihood);
	  double sec = SGN_Timer_get(timer);
	  SGN_Timer_print_DDHHMMSS_sec(SIGN_ERR, sec);
	  fprintf(SIGN_ERR, "\nX-SINGLE-TIME:\t%f\n", sec);
	  fflush(SIGN_ERR);
	  
	  /* Output the results.  */
	  if(!isnormal(result.likelihood) || result.monotone == 0){
		fprintf(SIGN_ERR, "Estimation failed for this set.\n");
		fflush(SIGN_ERR);
	  } else {
		ssm_output(&result, work, X, Y,
				   args->output_file, args->output_ext, j + args->id,
				   args->output_each, args->output_state, args->output_pvalues);
	  }
	  //Debug code for checking if the likelihood is consistent between before and
	  //  after the constraint.
	  //double llh = SSM_EM_estimate(result.ssm, X, T, sumT, Y, dim, args->em->ridge, work);
	  //fprintf(SIGN_ERR, "DEBUG: After constraint: LLH=%f\n", llh);
	}/* for j < sets */
  }/* for i < |dims| */

#ifdef SSM_DEBUG
  debug_put_max();
#endif
  fprintf(SIGN_ERR, "Going to exit.\n");
}
/* single_one() ============================================================ */

void single_set(Args * args, SSMData * Y){
  int total_sets = args->len_dims * args->sets;
  fprintf(SIGN_ERR, "***********************************\n");
  fprintf(SIGN_ERR, "**   Single SGE Execution Mode   **\n");
  fprintf(SIGN_ERR, "***********************************\n");
  fprintf(SIGN_ERR, "  Total Sets = %d\n\n", total_sets);
  fprintf(SIGN_ERR, "  ID = %d\n", args->id);
  fflush(SIGN_ERR);
  
  if(total_sets < args->id){
	fprintf(SIGN_ERR, "WARNING: ID exceeded the total sets.\n");
	fprintf(SIGN_ERR, "       : Going to quit.\n");
	return;
  }

  fprintf(SIGN_ERR, "--- INPUT DATA ---------\n");
  fprintf(SIGN_ERR, " observed samples = %d\n", Y->samples);
  fprintf(SIGN_ERR, " objects = %d\n", Y->p);
  fprintf(SIGN_ERR, " time points = %d\n", Y->N);
  fprintf(SIGN_ERR, " replicates = %d\n", Y->replicates);
  fprintf(SIGN_ERR, "------------------------\n");

  int dim_idx;
  int set_id;
  int dim;

  calc_first_dispatch(1, args->sets, args->len_dims,
					  args->id, &dim_idx, &set_id);
  dim = args->dims[dim_idx];

  fprintf(SIGN_ERR, "[JobID:%d] SGE Job dispatched. dim=%d, set ID=%d\n",
		  args->id, dim, set_id);
  fflush(SIGN_ERR);

  /* Prepare instances */
  SSM * ssm1 = SSM_new(Y->p, dim);
  SSM * ssm2 = SSM_new(Y->p, dim);
  SSMState * X = SSMState_new(dim, Y->N);
  SSMSuffStat * T = SSMSuffStat_new(Y->p, dim);
  SSMSuffStat * sumT = SSMSuffStat_new(Y->p, dim);
  SSMWork * work = SSMWork_new(Y->p, dim, Y->samples, args->em->constrain_F);
  if(args->em->constrain_F) SSM_EM_set_C(dim, work->C);

  /* Perform the EM algorithm */
  fprintf(SIGN_ERR, "Starting the SSM estimation.\n"); fflush(SIGN_ERR);
  SSMResult result = SSM_EM_iterate(args->em, args->init, args->iteration,
									Y, dim, ssm1, ssm2, X, T, sumT, work);
  fprintf(SIGN_ERR, "Finished: dim=%d, likelihood=%f\n", dim, result.likelihood);
  fflush(SIGN_ERR);

  /* Output the result */
	  /* Output the results.  */
  if(!isnormal(result.likelihood) || result.monotone == 0){
	fprintf(SIGN_ERR, "Estimation failed for this set.\n");
	fflush(SIGN_ERR);
  } else {
	ssm_output(&result, work, X, Y, args->output_file, args->output_ext,
			   (set_id % args->sets) + 1, args->output_each, args->output_state,
			   args->output_pvalues);
  }
  fprintf(SIGN_ERR, "\n"); fflush(SIGN_ERR);
}
/* single_set() ============================================================ */

/* End of file.  Copyright (C) 2010  HGC, IMS, The University of Tokyo       */
