/*
  thread.c : SiGN-SSM - Gene Network Estimation with State Sapce Model

  Copyright (C) 2010
    Yoshinori Tamada, Rui Yamaguchi, Seiya Imoto, Osamu Hirose, Ryo Yoshida,
	Masao Nagasaki and Satoru Miyano
      Laboratory of DNA Information Analysis &
	  Laboratory of Sequence Analysis, Human Genome Center,
      The Institute of Medical Science, The Unversity of Tokyo

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  
  Visit <http://sign.hgc.jp/signssm/> for more information.

  Written by Yoshinori Tamada <tamada at ims.u-tokyo.ac.jp>
  Originally developed by Osamu Hirose, Ryo Yoshida, Seiya Imoto, Rui Yamaguchi,
                          Tomoyuki Higuchi, and Satoru Miyano.
*/

#include "signcom.h"
#include "signssm.h"
#include "SSMData.h"
#include "SSM.h"
#include "Kalman.h"
#include "dispatch.h"
#include "execute.h"

#include "util/SGN_Array.h"
#include <string.h>
#include <strings.h>

#ifdef USE_OMP
#include <omp.h>
#endif /* USE_OMP */

void thread(Args * args, SSMData * Y){
#ifdef USE_OMP
  fprintf(SIGN_ERR, "*******************************\n");
  fprintf(SIGN_ERR, "**   Thread Execution Mode   **\n");
  fprintf(SIGN_ERR, "*******************************\n");
  fprintf(SIGN_ERR, "Number of threads = %d\n", args->threads);
  omp_set_num_threads(args->threads);

  fprintf(SIGN_ERR, "--- INPUT DATA ---------\n");
  fprintf(SIGN_ERR, " observed samples = %d\n", Y->samples);
  fprintf(SIGN_ERR, " objects = %d\n", Y->p);
  fprintf(SIGN_ERR, " time points = %d\n", Y->N);
  fprintf(SIGN_ERR, " replicates = %d\n", Y->replicates);
  fprintf(SIGN_ERR, "------------------------\n");

  int max_dim = int_ar_max(args->dims, args->len_dims);
  int total = args->iteration * args->sets * args->len_dims;
  fprintf(SIGN_ERR, "Max dimension = %d\n", max_dim);
  fprintf(SIGN_ERR, "Total single tasks = %d\n", total);
  fflush(SIGN_ERR);
  int total_sets = args->sets * args->len_dims;
  double best_scores[total_sets + 1];
  int finished_iter[total_sets + 1];
  SSMResult best_results[total_sets + 1];
  //for(int i = 0; i < total_sets + 1; i++){
  //SSMResult_init(&best_results[i], Y->p, max_dim);
  //}
  SGN_Array * stack = SGN_Array_new();
  for(int i = 1; i <= total_sets; i++){
	best_scores[i] = - DBL_MAX;
	best_results[i].ssm = NULL;
  }
  memset(finished_iter, 0, sizeof(int) * (total_sets + 1));

  int iteration = 0;
  int set = 0;
  int dim_idx = 0;
  int global_set_id = 1; /* 1-origin */
  
#pragma omp parallel default (shared)
  {
	int thread = omp_get_thread_num();
	  
	/* Prepare instances */
	SSM * ssm = SSM_new(Y->p, max_dim);
	SSMState * X = SSMState_new(max_dim, Y->N);
	SSMSuffStat * T = SSMSuffStat_new(Y->p, max_dim);
	SSMSuffStat * sumT = SSMSuffStat_new(Y->p, max_dim);
	SSMWork * work = SSMWork_new(Y->p, max_dim, Y->samples, args->em->constrain_F);
	int prev_dim = -1;

#ifdef USE_MINGW
	/* For MinGW compiler, initialize RNG again using thread ID. */
#pragma omp critical
	RNG_set_seed(args->init->rng, args->seed + args->id + thread);
#endif /* USE_MINGW */

	while(1){
	  /* Job Dispatching */
	  int set_id, dim;
#pragma omp critical
	  {
		set_id = global_set_id;
		dim = args->dims[dim_idx];
		inc_variables(args->iteration, args->sets, args->len_dims,
					  &iteration, &set, &dim_idx, &global_set_id);
#pragma omp flush (iteration, set, dim_idx, global_set_id)
	  }/* end of critical block */

	  if(set_id == 0) break; /* Job finished for this thread */

	  SSMResult result;
	  if(args->em->constrain_F && prev_dim != dim){ 
		SSM_EM_set_C(dim, work->C);
		prev_dim = dim;
	  }

	  int retry = 0;
	  
	  while(1) {
#pragma omp critical
		SSM_init(ssm, args->init, Y->p, dim, Y->Y, Y->samples);

		
		result = SSM_EM_algo(args->em, ssm, X, T, sumT, Y, dim, work);
		if((args->em->retry < 0 || retry < args->em->retry)
		   && ((!isnormal(result.likelihood) || result.monotone == 0))){
		  retry++;
		  fprintf(SIGN_ERR, "Estimation failed.  Retrying (%d)...\n", retry);
		  fflush(SIGN_ERR);
		  continue;
		} else {
		  if(args->em->retry != -1 && retry >= args->em->retry){
			fprintf(SIGN_ERR, "WARNING: Retry count exceeded the limitation.\n");
			fflush(SIGN_ERR);
		  }
		  break;
		}
	  }/* while(1) */
	  
#pragma omp critical
	  {
		finished_iter[set_id]++;
		fprintf(SIGN_ERR, "{%d} set=%d, dim=%d, likelihood=%f, loops=%d, finished=%d\n", 
				thread, set_id, dim, result.likelihood, result.loops,
				finished_iter[set_id]);
		if(isnormal(result.likelihood) && best_scores[set_id] < result.likelihood){
		  best_scores[set_id] = result.likelihood;
		  update(set_id, best_scores, &result, best_results, stack, max_dim);
		}
		fflush(SIGN_ERR);
		if(finished_iter[set_id] == args->iteration){
		  if(best_results[set_id].ssm != NULL){

			ssm_output(best_results + set_id, work, X, Y,
					   args->output_file, args->output_ext,  (set_id % args->sets) + 1,
					   args->output_each, args->output_state, args->output_pvalues);

			SGN_Array_add(stack, best_results[set_id].ssm);
			best_results[set_id].ssm = NULL;
		  } else {
			fprintf(SIGN_ERR, "Estimation failed for this set.\n");
			fflush(SIGN_ERR);
		  }
		}
	  }
	  /* end of critical block */
	}/* while(1) */
  }
  /* End of parallel section */
#else
  fprintf(SIGN_ERR, "ERROR: Multi-thread mode not enabled in this executable.\n");
  exit(1);
#endif /* USE_OMP */
}
/* thread() ================================================================ */

/* End of file.  Copyright (C) 2010  HGC, IMS, The University of Tokyo       */
