/*
  sign/util/SGN_Array.c : Expandable array for general purposes.

  Copyright (C) 2008-2011 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#include "SGN_Array.h"
#ifndef __SIGN_SSM
#include "SGN_Memory.h"
#include "../signcom.h"
#endif /* __SIGN_SSM */

#include <stdio.h>
#include <assert.h>
#include <string.h>

/*!\brief Generates a new \ref SGN_Array instance.
 *
 * This generates a new \ref SGN_Array instance and initializes it.
 * The genereated instance should be deleted by SGN_Array_delete() when
 * the user no longer need to use it.
 *
 *\return pointer to the generated \ref SGN_Array instance, or NULL if
 *   failed to allocate the instance.
 *\sa SGN_Array_delete()
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Array */
SGN_Array * SGN_Array_new(){
  SGN_Array * ar;

  ar = (SGN_Array *) malloc(sizeof(SGN_Array));
  if(ar == NULL) return NULL;

  ar->ptr = NULL;
  ar->buff_size = 0;
  ar->size = 0;

  return ar;
}
/* SGN_Array_new() ------------------------------------------------------------- */

/*!\brief Generates a new \ref SGN_Array instance with the initial buffer
 *  size.
 *
 * Unlike SGN_Array_new(), this can specify the initial buffer size.
 * Note that the buffer size does not mean the length of the array.
 * The initial size of the array equals zero.
 *
 *\param size initial buffer size.
 *\return pointer to the generated \ref SGN_Array instance, or NULL if
 *   failed to allocate the instance.
 *\sa SGN_Array_delete()
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Array */
SGN_Array * SGN_Array_news(size_t size){
  SGN_Array * ar;

  ar = SGN_Array_new();
  if(ar == NULL) return NULL;
  if(size == 0) return ar;

  ar->ptr = (void **) malloc(sizeof(void *) * size);
  if(ar->ptr == NULL){
	SGN_Array_delete(ar);
	return NULL;
  }

  ar->buff_size = size;
  ar->size = 0;

  return ar;
}
/* SGN_Array_news() ========================================================= */

/*!\brief Intializes the SGN_Array instance.
 *
 * <b>THIS IS AN INTERNAL FUNCTION</b>.
 *
 * This initializes the SGN_Array instance allocated on, for example, 
 * stack memory.
 *
 * The resources allocated by this function can be release by
 * SGN_Array_release().
 *
 *\internal
 *\param ar pointer to the SGN_Array instance.
 *\param size  the initial buffer size.
 *\sa SGN_Array_release()
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
void SGN_Array_init(SGN_Array * ar, size_t size){
  ar->ptr = (void **) malloc(sizeof(void *) * size);
  if(ar->ptr == NULL){
	return;
  }
  ar->buff_size = size;
  ar->size = 0;
}
/* SGN_Array_init() ========================================================= */

/*!\brief Releases the resources allocated for the SGN_Array instance.
 *
 * <b>THIS IS AN INTERNAL FUNCTION</b>.
 *
 * This releases resources allocated by SGN_Array_init().
 * This does not release the pointer \a this itself.
 *
 *\internal
 *\param this  pointer to the SGN_Array instance whose resources are to be released.
 *\sa SGN_Array_init()
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
void SGN_Array_release(SGN_Array * this){
  if(this == NULL) return;
  free(this->ptr);
  this->ptr = NULL;
  this->size = 0;
  this->buff_size = 0;
}
/* SGN_Array_release() ========================================================= */

/*!\brief Deletes the \ref SGN_Array instance.
 *
 * This deletes the instance allocated by SGN_Array_new().
 * Note that this does not delete (free) the elements in the array.
 * The user is responsible to manage them.
 * If you want to delete (free) all the elements in the array at the
 * same time, use SGN_Array_delete_all().
 *
 * If NULL is given, this does nothing.
 *
 *\param this pointer to the \ref SGN_Array instance to be deleted.  NULL is
 *          acceptable.
 *\sa SGN_Array_new(), SGN_Array_news()
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
void SGN_Array_delete(void * this){
  if(this == NULL) return;
  
  free(((SGN_Array *) this)->ptr); /* If NULL, no operation is performed. 
									 Thus this is safe. */
  free(this);
}
/* SGN_Array_delete() ======================================================= */

/*!\brief Deletes the \ref SGN_Array instance and its elements.
 *
 * This calls free() for all the elements stored in the array
 * before deleting the array itself.
 *
 * If free() is not applicable to elements in the array,
 * use SGN_Array_delete_allf() in which the user can specify
 * the deletion function for the elements.
 *
 *\param this pointer to the \ref SGN_Array instance to be deleted.  NULL is
 *         acceptable
 *\sa SGN_Array_delete_allf(), SGN_Array_new(), SGN_Array_news()
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
void SGN_Array_delete_all(SGN_Array * this){
  size_t i;
  if(this == NULL) return;
  
  for(i = 0; i < this->size; i++){
#ifdef SIGN_DEBUG
	/*
	fprintf(SIGN_ERR, "SGN_Array_delete_all(): deleting ptr[%zu]=%p...\n", i, this->ptr[i]);
	fflush(SIGN_ERR);
	*/
#endif
	free(this->ptr[i]);
  }
  SGN_Array_delete(this);
}
/* SGN_Array_delete_all() =================================================== */

/*!\brief Deletes the \ref SGN_Array instance and its elements by the specified
 *          deletion function.
 *
 * This is similar to SGN_Array_delete_all() but deletes the elements in the
 * array by the specified deletion function \a del().
 *
 *\param this pointer to the \ref SGN_Array instance to be deleted.  NULL is
 *        acceptable.
 *\param del  pointer to the function to delete elements in the array.
 *\sa SGN_Array_delete_all()
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Array */
void SGN_Array_delete_allf(SGN_Array * this, void (*del)(void *)){
  size_t i;
  if(this == NULL) return;
  for(i = 0; i < this->size; i++){
	del(this->ptr[i]);
  }
  SGN_Array_delete(this);
}
/* SGN_Array_delete_allf() ================================================== */

/*!\brief Generates multiple \ref SGN_Array instances.
 *
 * This returns the pointer to the first address of the array of
 * SGN_Array instances.  All the instances are allocated
 * in the consecutive memory and initialized.  Therefore,
 * The <tt>i</tt>-th instance can be accessed by <tt>p + i</tt>
 * where <tt>p</tt> is the pointer returned by this function.
 *
 * Use SGN_Array_delete_ar() to delete the returned array of the instances.
 *
 *\param ar_size  number of instances.
 *\param init_size default buffer size for each instance.
 *\return pointer to the first element of the generated instances.
 *\sa SGN_Array_delete_ar()
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
SGN_Array * SGN_Array_new_ar(size_t ar_size, size_t init_size){
  int i;
  SGN_Array * ar = (SGN_Array *) malloc(sizeof(SGN_Array) * ar_size);
  for(i = 0; i < ar_size; i++){
	SGN_Array_init(ar + i, init_size);
  }
  return ar;
}
/* SGN_Array_new_ar() ======================================================= */

/*!\brief Deletes the SGN_Array instances allocated by SGN_Array_new_ar().
 *
 * Note that this does not free elements in the arrays.
 * The number of arrays (\a ar_size) needs to be the same
 * as one specified in SGN_Array_new_ar().
 *
 *\param this pointer to the first element of the instances to be deleted.
 *\param ar_size  number of arrays.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
void SGN_Array_delete_ar(SGN_Array * this, size_t ar_size){
  for(size_t i = 0; i < ar_size; i++){
	SGN_Array_release(this + i);
	free(this);
  }
}
/* SGN_Array_delete_ar() ==================================================== */

/*!\brief Expands the size of the buffer.
 *
 * THIS IS AN INTERNAL FUNCTION
 *
 * The new buffere size will be the current size + length
 * specified by \a length argument.
 *
 * If \a length is 0, then it is regarded as 1.
 *
 * Generally this is called internally and not called by
 * the users.
 *
 *\param this pointer to the \ref SGN_Array instance to expand.
 *\param length  buffer size to be newly allocated.
 *\return \a this, or NULL if failed.
 *\internal
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
SGN_Array * SGN_Array_expand(SGN_Array * this, size_t length){
  void * new_ptr;
  size_t new_size;

  if(length == 0) length = 1;
  new_size = this->buff_size + length;
  new_ptr = realloc(this->ptr, sizeof(void *) * new_size);
  if(new_ptr != NULL){
	this->ptr = new_ptr;
	this->buff_size = new_size;
	return this;
  } else {
	return NULL;
  }
}
/* SGN_Array_expand() ======================================================= */

/*!\brief  Specifies the size of the array.
 *
 * If the specified size is bigger than the current size,
 * the values (pointers) in the expanded area are undefined.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Array */
SGN_Array * SGN_Array_set_size(SGN_Array * this, size_t size){
  if(this->buff_size < size){
	SGN_Array_expand(this, size - this->buff_size);
  }
  this->size = size;
  return this;
}
/* SGN_Array_set_size() ===================================================== */

/*!\brief Removes all the elements in the SGN_Array instance.
 *
 * This does not affect the already allocated memory, i.e. this does
 * not free the allocated memory.  Therefore this works very fast.
 * Also this does not free the elements.
 * Use SGN_Array_clear_all() if you want to free the elements when
 * you clear the array.
 *
 *
 *\param this  pointer to the \ref SGN_Array instance to be cleared.
 *\return \a this.
 *\sa SGN_Array_clear_all()
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
SGN_Array * SGN_Array_clear(SGN_Array * this){
  this->size = 0;
  return this;
}
/* SGN_Array_clear() ======================================================== */

/*!\brief Removes all then elements in the SGN_Array instance.
 *
 * This calls free() for all the elemens in the array.
 *
 *\param this  pointer to the \ref SGN_Array instance.
 *\return \a this.
 *\sa SGN_Array_clear()
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
SGN_Array * SGN_Array_clear_all(SGN_Array * this){
  size_t i;
  for(i = 0; i < this->size; i++){
	free(this->ptr[i]);
  }
  this->size = 0;
  return this;
}
/* SGN_Array_clear_all() ==================================================== */

/*!\brief Clears all the elements in the array by the deletion function.
 *
 * This calls \a del() for all the elements before it clears the array.
 *
 *\param this pointer to the SGN_Array instance to be cleared.
 *\param del  pointer to the function to delete the elements in the array.
 *\return \a this.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Array */
SGN_Array * SGN_Array_clear_allf(SGN_Array * this, void (*del)(void *)){
  if(this == NULL) return NULL;
  for(int i = 0; i < this->size; i++){
	del(this->ptr[i]);
  }
  this->size = 0;
  return this;
}
/* SGN_Array_clear_allf() =================================================== */

/*!\brief Adds an element to the SGN_Array instance.
 *
 * The pointer \a obj is added at the end of the array.
 * Note that this does not make a copy (or a clone) of
 * the specified object to be added.
 *
 *\param this  pointer to the \ref SGN_Array instance.
 *\param obj  pointer to the object to be stored in the array.
 *\return \a this, or NULL if failed to expand the buffer memory.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
SGN_Array * SGN_Array_add(SGN_Array * this, void * obj){
  if(this->buff_size == this->size){
	SGN_Array * ar = SGN_Array_expand(this, this->size);
	if(ar == NULL) return NULL;
  }

  this->ptr[this->size++] = obj;
  return this;
}
/* SGN_Array_add() ========================================================== */

/*!\brief Adds all the elements in the other array to the array.
 *
 * This copes all the elements in the second array \a array
 * into the first array \a this.
 *
 *\param this   pointer to the \ref SGN_Array instance to be copied.
 *\param array  pointer to the \ref SGN_Array instance to copy.
 *\return \a this.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
SGN_Array * SGN_Array_add_array(SGN_Array * this, const SGN_Array * array){
  int i;
  for(i = 0; i < SGN_Array_size(array); i++){
	SGN_Array_add(this, SGN_Array_get(array, i));
  }
  return this;
}
/* SGN_Array_add_array() ==================================================== */
  
/*!\brief Returns the element at the specified position in the array.
 *
 * Note: This does not check the array size boundary,
 * that is, the overflow of \a index.
 *
 *\param this  pointer to the \ref SGN_Array instance.
 *\param index  position of the element to return.
 *\return pointer to the object stored at the specified position in the
 *   array.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
void * SGN_Array_get(const SGN_Array * this, size_t index){
#ifdef SIGN_DEBUG
  if(this->size <= index) {
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_Array_get(): index exceeded the size. (%zu >= %zu)\n",
			index, this->size);
	assert(0);
  }
#endif /* SIGN_DEBUG */

  return this->ptr[index];
}
/* SGN_Array_get() ========================================================== */

/*!\brief Returns the element at the end of the array.
 *
 * The returned element is deleted from the array.
 * Therefore the size of array becomes one smaller from the current
 * size. 
 *
 *\param this  pointer to the \ref SGN_Array instance.
 *\return  pointer to the element stored at the end of the array.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
void * SGN_Array_pop(SGN_Array * this){
  return this->ptr[--(this->size)];
}
/* SGN_Array_pop() ========================================================== */

/*!\brief Sets the element at the specified position in the array.
 *
 * The pointer (object) at the specified position is overwritten
 * by the new one.
 *
 *\param this   pointer to the \ref SGN_Array instance.
 *\param index   position at which the object is set in the array.
 *\param object   pointer to the object to be set.
 *\return \a this.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
SGN_Array * SGN_Array_set(SGN_Array * this, size_t index, void * object){
#ifdef SIGN_DEBUG
  if(this->size <= index){
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_Array_set(): index exceeded the size.\n");
	exit(1);
  }
#endif /* SIGN_DEBUG */
  this->ptr[index] = object;
  return this;
}
/* SGN_Array_set() ========================================================== */

/*!\brief Inserts the element at the specified position in the array.
 *
 * The size of the array becomes one bigger than the current size.
 *
 *\param this  pointer to the \ref SGN_Array instance.
 *\param index  position at which the new object is inserted.
 *\param object  pointer to the object to be inserted.
 *\return \a this.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Array */
SGN_Array * SGN_Array_insert(SGN_Array * this, size_t index, void * object){
#ifdef SIGN_DEBUG
  if(this->size <= index){
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_Array_insert(): index exceeded the size.\n");
	exit(1);
  }
#endif
  if(this->buff_size == this->size){
	SGN_Array * ar = SGN_Array_expand(this, this->size);
	if(ar == NULL) return NULL;
  }
  
  memmove(this->ptr + index + 1, this->ptr + index,
		  sizeof(void **) * (this->size - index));
  this->ptr[index] = object;
  this->size++;
  return this;
}
/* SGN_Array_insert() ======================================================= */

/*!\brief Removes the element in the array.
 *
 * Note that this does not free the memory for the element removed from
 * the array by this method.
 *
 *\param this  pointer to the \ref SGN_Array instance.
 *\param index  index of the element to be removed from the array.
 *\return  pointer to the removed element.
 * 
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
void * SGN_Array_remove(SGN_Array * this, size_t index){
#ifdef SIGN_DEBUG
  if(this->size <= index){
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_Array_remove(): index exceeded the size.\n");
	exit(1);
  }
#endif
  void * obj = this->ptr[index];
  memmove(this->ptr + index, this->ptr + index + 1, 
		  sizeof(void **) * (this->size - index - 1));
  this->size--;
  return obj;
}
/* SGN_Array_remove() ======================================================= */

/*!\brief Returns the number of elements stored in the array.
 *
 *\param this pointer to the \ref SGN_Array instance.
 *\return the number of elements in the array.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Array */
size_t SGN_Array_size(const SGN_Array * this){
  return this->size;
}
/* SGN_Array_size() ========================================================= */

/*!\brief Returns the index of the specified object in the array.
 *
 * This only compares the pointer of the object, not
 * the contents of the instances whether or not they are identical.
 * This is an O(n) algorithm where n is the number of elements in the array.
 *
 *\param this  pointer to the \ref SGN_Array instance.
 *\param object   pointer to the object to find.
 *\return  index of the object, or \ref SGN_Array_ERROR_VALUE if not found.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
size_t SGN_Array_find(const SGN_Array * this, const void * object){
  size_t i;
  for(i = 0; i < this->size; i++){
	if(this->ptr[i] == object) return i;
  }
  return SGN_Array_ERROR_VALUE;
}
/* SGN_Array_find() ========================================================= */

/*!\brief Sorts the elements in the array.
 *
 * This sorts the elements using a user defined compare function.
 *
 * A typical compare function is as follows.  
 * Assume that you added the pointers to type T to the array.
 *
 * \code
 * int compare(const void * a, const void * b){
 *   const T * aa = (const T *) *((const void **) a);
 *   const T * bb = (const T *) *((const void **) b);
 *
 *   // Compare aa and bb.  E.g., If T has a member x, and
 *   // if you want to sort according to x ...
 *   if(aa->x > bb->x){
 *     return -1;
 *   } else if(aa->x < bb->x){
 *     return 1;
 *   } else {
 *     return 0;
 *   }
 * }
 * \endcode
 *
 *\param this   pointer to the \ref SGN_Array instance.
 *\param cmp    comparison function.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Array */
void SGN_Array_sort(SGN_Array * this, int (* cmp)(const void *, const void *)){
  qsort(this->ptr, this->size, sizeof(void *), cmp);
}
/* SGN_Array_sort() ========================================================= */

void SGN_Array_fprint_v(const void * v, FILE * fp){
  const SGN_Array * ar = (const SGN_Array *) v;
  fprintf(fp, "(Array elm=%zu)", SGN_Array_size(ar));
}
/* SGN_Array_fprint_v() ==================================================== */


/* ========================================================================= */
/*  MPI Functions                                                            */
/* ========================================================================= */

#ifdef USE_MPI
/*
void SGN_Array_MPI_Bcast(SGN_Array ** ar, int root, MPI_Comm comm,
					 void * (*bcast)(void *)){
  
}
*/
/* SGN_Array_MPI_Bcast() ------------------------------------------------------- */

#endif /* USE_MPI */

/* End of file.  Copyright (C) 2008-2011 SiGN project Members */
