/*
  sign/util/SGN_Array.h : Expandable array for general purposes.

  Copyright (C) 2008-2011 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#ifndef __SGN_ARRAY_H
#define __SGN_ARRAY_H

#include <stdio.h>
#include <stdlib.h>

/*!\brief Expandable data container for general purposes.
 *
 * \ref SGN_Array class provides functions to realize the automatically
 * expandable data container for general purposes.  It can contain void
 * pointers so that a user can store any pointer without managing the size
 * of the container.
 *
 * Do not use the SGN_Array structure directly, for example, allocating SGN_Array
 * structure by malloc().  To use \ref SGN_Array class, use SGN_Array_new() to
 * generate the \ref SGN_Array 'instance', and when you no longer need to use it,
 * call SGN_Array_delete() to release the resources used by the instance.
 *
 * \section example Example
 * \code
 * #include "util/SGN_Array.h"
 * SGN_Array * ar = SGN_Array_new();
 *
 * // Do someting. E.g.  To add an object pointed by 'ptr', do
 * SGN_Array_add(ar, ptr);
 *
 * SGN_Array_delete(ar);   // Delete the array instance.
 * \endcode
 *
 *\author Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
 */
typedef struct {
#ifdef DOXY
private:
#endif
  void ** ptr;
  size_t size;
  size_t buff_size;
} SGN_Array;

/*! Constant used to represent an error.
 *
 *\memberof SGN_Array */
#define SGN_Array_ERROR_VALUE  ((size_t) -1)

SGN_Array * SGN_Array_new(void);
SGN_Array * SGN_Array_news(size_t size);
void       SGN_Array_init(SGN_Array * ar, size_t size);
void       SGN_Array_delete(void * this);
void       SGN_Array_delete_all(SGN_Array * this);
void       SGN_Array_delete_allf(SGN_Array * this, void (*del)(void *));
void       SGN_Array_release(SGN_Array * this);
SGN_Array * SGN_Array_new_ar(size_t ar_size, size_t init_size);
SGN_Array * SGN_Array_expand(SGN_Array * this, size_t length);
SGN_Array * SGN_Array_set_size(SGN_Array * this, size_t size);
SGN_Array * SGN_Array_add(SGN_Array * this, void * object);
size_t     SGN_Array_size(const SGN_Array * this);
void *     SGN_Array_get(const SGN_Array * this, size_t index);
void *     SGN_Array_pop(SGN_Array * this);
SGN_Array * SGN_Array_insert(SGN_Array * this, size_t index, void * object);
void *     SGN_Array_remove(SGN_Array * this, size_t index);
SGN_Array * SGN_Array_set(SGN_Array * this, size_t index, void * object);
size_t     SGN_Array_find(const SGN_Array * this, const void * object);
void       SGN_Array_sort(SGN_Array * this, int (* cmp)(const void *, const void *));
SGN_Array * SGN_Array_clear(SGN_Array * this);
SGN_Array * SGN_Array_clear_all(SGN_Array * this);
SGN_Array * SGN_Array_clear_allf(SGN_Array * this, void (*del)(void *));
void        SGN_Array_fprint_v(const void * v, FILE * fp);

#ifdef USE_MPI
#include <mpi.h>

#endif /* USE_MPI */

#endif /* __SGN_ARRAY_H */

/* End of file.  Copyright (C) 2008-2011 SiGN Project Members */
