/*
  SGN_BS.c : Bootstrap related routines

  Copyright (C) 2010-2011, 2014  SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.
  
  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#include "SGN_BS.h"
#include "SGN_Util.h"

#include <string.h>
#include <strings.h>
#include <math.h>

/*!\brief Sets up a new bootstrap method environment.
 *
 * The returned instance keeps various settings for
 * performing a bootstrap method.
 * It is initialized for the normal bootstrap method.
 *
 * To use it for SSM permuation test,
 * call SGN_BS_set_SSM() to set up the appropriate
 * default settings.
 *
 *\param network  the first network used to set up
 *            a new bootstrap method environment.
 *\return pointer to the SGN_BS instance.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_BS */
SGN_BS * SGN_BS_new(int nodes){
  size_t total = 0, mem;
  SGN_BS * bs = malloc(mem = sizeof(SGN_BS)); total += mem;
  bs->nodes = nodes;
  bs->graph = malloc(mem = sizeof(double *) * bs->nodes); total += mem;
  if(bs->graph == NULL){
	fprintf(SIGN_ERR, "ERROR: SGN_BS: Failed to allocate memory for \'graph\'.\n");
	exit(1);
  }
  bs->gain = malloc(mem = sizeof(double *) * bs->nodes); total += mem;
  if(bs->gain == NULL){
	fprintf(SIGN_ERR, "ERROR: SGN_BS: Failed to allocate memory for \'gain\'.\n");
	exit(1);
  }
  for(int i = 0; i < bs->nodes; i++){
	bs->graph[i] = malloc(mem = sizeof(double) * bs->nodes); total += mem;
	if(bs->graph[i] == NULL){
	  fprintf(SIGN_ERR, "ERROR: SGN_BS: Failed to allocate memory for \'graph[i=%d]\'.\n", i);
	  exit(1);
	}
	memset(bs->graph[i], 0, sizeof(double) * bs->nodes);

	bs->gain[i] = malloc(mem = sizeof(double) * bs->nodes); total += mem;
	if(bs->gain[i] == NULL){
	  fprintf(SIGN_ERR, "ERROR: SGN_BS: Failed to allocate memory for \'gain[i=%d]\'.\n", i);
	  exit(1);
	}
	memset(bs->gain[i], 0, sizeof(double) * bs->nodes);
  }
  bs->ud = malloc(mem = sizeof(double **) * bs->nodes); total += mem;
  if(bs->ud == NULL){
	fprintf(SIGN_ERR, "ERROR: SGN_BS: Failed to allocate memory for \'ud\'.\n");
	exit(1);
  }
  for(int i = 0; i < bs->nodes; i++){
	bs->ud[i] = malloc(mem = sizeof(double*) * bs->nodes); total += mem;
	if(bs->ud[i] == NULL){
	  fprintf(SIGN_ERR, "ERROR: SGN_BS: Failed to allocate memory for \'ud[i=%d]\'.\n", i);
	  exit(1);
	}
	for(int j = 0; j < bs->nodes; j++){
	  bs->ud[i][j] = malloc(mem = sizeof(double) * 4); total += mem;
	  if(bs->ud[i][j] == NULL){
		fprintf(SIGN_ERR, "ERROR: SGN_BS: Failed to allocate memory for \'ud[i=%d][j=%d]\'.\n", 
				i, j);
		exit(1);
	  }
	  memset(bs->ud[i][j], 0, sizeof(double) * 4);
	}/* for j */
  }/* for i */

  bs->label_prob =  "BS.Prob";
  bs->label_gain =  "BS.Gain";
  bs->label_type =  "EdgeType";
  bs->label_ud =    "BS.up/down";
  bs->label_ratio = "BS.EdgeTypeRatio";
  bs->label_dir =   "BS.Direction";
  bs->label_score = "edgeScore";

  bs->label_hubness = "hubness";
  bs->label_hubindex = "BS.HubIndex";
  bs->hubindex = malloc(mem = sizeof(double) * bs->nodes); total += mem;
  if(bs->hubindex == NULL){
	fprintf(SIGN_ERR, "ERROR: SGN_BS: Failed to allocate memory for \'hubindex\'.\n");
	exit(1);
  }
  memset(bs->hubindex, 0, sizeof(double) * bs->nodes);

  bs->one_minus = 0;
  bs->fnet = NULL;
  bs->data = NULL;
  bs->verbose = 0;

  if(bs->verbose >= 0){
	fprintf(SIGN_ERR, "SGN_BS: Memory allocated for the bootstrap: ");
	SGN_Util_print_size_t_comma(SIGN_ERR, total / 1024 / 1024);
	fprintf(SIGN_ERR, " [MiB].\n");
  }
  
  return bs;
}
/* SGN_BS_new() ============================================================ */

/*!\brief Deletes the bootstrap method environment.
 *
 *\param bs  pointer to the SGN_BS instance to be deleted.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_BS */
void SGN_BS_delete(SGN_BS * bs){
}
/* SGN_BS_delete() ========================================================= */

void SGN_BS_set_verbose(SGN_BS * bs, int verbose){
  bs->verbose = verbose;
}
/* SGN_BS_set_verbose() ==================================================== */

/*!\brief Set the parameters for processing SSM permutation test files.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_BS */
void SGN_BS_set_SSM(SGN_BS * bs){
  bs->one_minus       = 0;
  bs->label_prob      = "SSM.P";
  //bs->label_gain    = "SSM.Gain";
  bs->label_gain      = NULL;
  bs->label_type      = NULL;
  //bs->label_ud      = "SSM.up/down";
  bs->label_ud        = NULL;
  //bs->label_ratio   = "SSM.EdgeTypeRatio";
  bs->label_ratio     = NULL;
  bs->label_dir       = NULL;
  bs->label_hubindex  = NULL;
  bs->label_score     = NULL;
}
/* SGN_BS_set_SSM() ============================================================ */


/*!\brief Generates the file names to be processed.
 *
 * This routine simply generates a set of file names to be processed
 * from the file prefix and numbering information such as the number
 * of suffix digits, the initial number and the last number.
 *
 * The generate file names are returned as a string array (\ref SGN_StrArray),
 * and can be used as an input for SGN_BS_first_network() and SGN_BS_process_files().
 *
 *\param prefix
 *\param suffix
 *\param digits
 *\param bg
 *\param ed
 *\return
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_BS */
SGN_StrArray * SGN_BS_gen_files(const char * prefix, const char * suffix, 
								int digits,	int bg, int ed){
  char format[SGN_BS_BUFF_SIZE];
  SGN_StrArray * files = SGN_StrArray_new();
  snprintf(format, SGN_BS_BUFF_SIZE, "%%s.%%0%dd%s", digits, suffix);

  int len = strlen(prefix) + strlen(suffix);
  for(int i = bg; i <= ed; i++){
	char * file = (char *) malloc(sizeof(char) * (len + 10));
	if(file == NULL){
	  fprintf(SIGN_ERR, "ERROR: SGN_BS_gen_files(): Memory allocation failed.\n");
	  exit(1);
	}
	snprintf(file, len + 10, format, prefix, i);
	SGN_StrArray_add(files, file);
  }/* for i */

  return files;
}
/* SGN_BS_gen_files() ========================================================== */

SGN_Network * SGN_BS_first_network(SGN_StrArray * filenames, IOReader reader){
  SGN_Network * network = NULL;
  fprintf(SIGN_ERR, "Try to read the first network to prepare.\n");
  for(int i = 0; i < SGN_StrArray_size(filenames); i++){
	const char * file = SGN_StrArray_get(filenames, i);
	fprintf(SIGN_ERR, "  trying: %s\n", file);
	FILE * fp = fopen(file, "r");
	if(fp == NULL){
	  fprintf(SIGN_ERR, "Cannot open file: %s\n", file);
	  network = NULL;
	} else {
	  network = reader(fp, NULL);
	  fclose(fp);
	}
	if(network != NULL) break;
  }
  if(network == NULL){
	fprintf(SIGN_ERR, "ERROR: No input file could be read.\n");
	exit(1);
  }
  return network;
}
/* SGN_BS_first_network() ====================================================== */

void SGN_BS_set_network(SGN_BS * bs, SGN_Network * network){
  bs->fnet = network;
}

void SGN_BS_set_Data(SGN_BS * bs, SGN_Data * data){
  bs->data = data;
}


int SGN_BS_process(SGN_BS * bs, const SGN_Network * network,
				   const char * freq_key, const char * ud_key, const char * total_key){

  double total = 1.0; /* This is for processing a pre-compiled network. */
  if(total_key != NULL){
	int v = SGN_Network_get_property_int(network, total_key, -99999);
	if(v == -99999){
	  fprintf(SIGN_ERR, "ERROR: SGN_BS: Cannot find network property: %s\n", total_key);
	  exit(1);
	} else {
	  total = (double) v;
	}
  }

  if(SGN_Network_num_nodes(network) != bs->nodes){
	fprintf(SIGN_ERR, "ERROR: inconsistent number of nodes. %d (this) vs %d (global).\n",
			(int) SGN_Network_num_nodes(network), bs->nodes);
	exit(1);
  }

  /* Visit all the edges */
  for(int i = 0; i < bs->nodes; i++){
	const SGN_IntArray * parents = SGN_Network_get_parent_indexes(network, i);
	for(int jj = 0; jj < SGN_IntArray_size(parents); jj++){
	  int j = SGN_IntArray_get(parents, jj);
	  
	  SGN_Edge * edge = SGN_Network_get_edge_index(network, j, i);
	  
	  double coef = 1.0;
	  if(freq_key != NULL){
		double v = SGN_Edge_get_property_double(edge, freq_key, NAN);
		if(isnan(v)){
		  fprintf(SIGN_ERR, 
				  "ERROR: SGN_BS: Cannot find edge property: %s\n", freq_key);
		  exit(1);
		} else {
		  coef = v * total;
		}
	  }
	  
	  bs->graph[j][i] += coef;
	  
	  int * ptr = (int *) SGN_Edge_get_property_with_type(edge, "EdgeType", SGN_Type_INT);
	  
	  double udr = 1.0;
	  if(ud_key != NULL){
		double v = SGN_Edge_get_property_double(edge, ud_key, NAN);
		if(isnan(v)){
		  fprintf(SIGN_ERR, 
				  "ERROR: SGN_BS: Cannot find edge property: %s\n", ud_key);
		  exit(1);
		} else {
		  udr = v * coef;
		}
	  }
	  
	  if(ptr == NULL){
		//fprintf(SIGN_ERR, "WARNING: EdgeType not found.\n");
	  } else {
		bs->ud[j][i][*ptr] += udr;
		bs->ud[i][j][*ptr] += coef - udr;
	  }
	  
	  double * dptr = (double *) SGN_Edge_get_property_with_type(edge, "gain", SGN_Type_DOUBLE);
	  //double * dptr = NULL;  // FOR DEBUG
	  if(dptr == NULL){
		//fprintf(SIGN_ERR, "WARNING: gain not found.\n");
	  } else {
		if(*dptr < 0){
		  fprintf(SIGN_ERR, "WARNING: Minus gain detected: %f.\n", *dptr);
		}
		bs->gain[j][i] += *dptr * total;
	  }
	}
  }/* for i < args->nodes */

  return (int) total;
}
/* SGN_BS_process() ======================================================== */

void SGN_BS_process_serialized(SGN_BS * bs, void * buff, size_t size){

  if(bs->verbose >= 1){
	fprintf(SIGN_ERR, "SGN_BS_process_serialized called. [size = %zu]\n", size);
	fflush(SIGN_ERR);
  }

  size_t pos = 0;
  int i = 0;
  while(1){
	/* Read the number of parents */
	int qj;

	if(size < pos + sizeof(int)){
	  fprintf(SIGN_ERR, "ERROR: SGN_BS: Failed to read the serialized network.\n");
	  exit(1);
	}
	memcpy(&qj, (uint8_t *) buff + pos, sizeof(int));
	pos += sizeof(int);

	/*
	if(bs->verbose){
	  fprintf(SIGN_ERR, "  i=%d, qj=%d\n", i, qj); fflush(SIGN_ERR);
	}
	*/

	for(int jj = 0; jj < qj; jj++){
	  int j;
	  if(size < pos + sizeof(int)){
		fprintf(SIGN_ERR, "ERROR: SGN_BS: Failed to read the serialized network.\n");
		exit(1);
	  }
	  memcpy(&j, (uint8_t *) buff + pos, sizeof(int));
	  pos += sizeof(int);

	  /*
	  if(bs->verbose){
		fprintf(SIGN_ERR, "    j=%d\n", j); fflush(SIGN_ERR);
	  }
	  */

	  bs->graph[j][i] += 1.0;

	  /* ud */
	  char ud;
	  if(size < pos + sizeof(char)){
		fprintf(SIGN_ERR, "ERROR: SGN_BS: Failed to read the serialized network.\n");
		exit(1);
	  }
	  memcpy(&ud, (uint8_t *) buff + pos, sizeof(char));
	  pos += sizeof(char);
	  
	  /*
	  if(bs->verbose){
		fprintf(SIGN_ERR, "    ud=%d\n", ud); fflush(SIGN_ERR);
	  }
	  */

	  bs->ud[j][i][(int)ud] += 1.0;

	  /* gain */
	  double gain;
	  if(size < pos + sizeof(double)){
		fprintf(SIGN_ERR, "ERROR: SGN_BS: Failed to read the serialized network.\n");
		exit(1);
	  }
	  memcpy(&gain, (uint8_t *) buff + pos, sizeof(double));
	  pos += sizeof(double);
	  
	  /*
	  if(bs->verbose){
		fprintf(SIGN_ERR, "    gain=%f\n", gain); fflush(SIGN_ERR);
	  }
	  */

	  bs->gain[j][i] += gain;
	  
	  /*
	  if(bs->verbose){
		fprintf(SIGN_ERR, "    pos=%zu\n", pos); fflush(SIGN_ERR);
	  }
	  */
	}/* for jj < qj */
	i++;

	if(size == pos) break; /* Correctly read */

  }/* while(1) */

  if(i != bs->nodes){
	fprintf(SIGN_ERR, "ERROR: inconsistent number of nodes. %d (this) vs %d (global).\n",
			i, bs->nodes);
	exit(1);
  }

}
/* SGN_BS_process_serialized() ============================================= */

/*
  If \a pv_key and \a total_key is specified, use the frequency information
  stored in the networks is used.
 */
int SGN_BS_process_multi(SGN_BS * bs, IOReader reader, const char * file, 
					 const char * freq_key, const char * ud_key, const char * total_key){
  SGN_Network * network;
  int count = 0;
  int real_count = 0;

  fprintf(SIGN_ERR, "Processing: %s\n", file);
  FILE * fp = fopen(file, "r");

  if(fp == NULL) return 0;

  while(!feof(fp)){
	network = reader(fp, NULL);
	if(network == NULL) continue;

	int total = SGN_BS_process(bs, network, freq_key, ud_key, total_key);

	count++;
	real_count += total;
	SGN_Network_delete(network);
  }
  fclose(fp);
  fprintf(SIGN_ERR, "Processed %d network(s).\n", count);
  if(real_count != count) fprintf(SIGN_ERR, "  Total Processed %d network(s) assumed.\n", 
								  real_count);
  fflush(SIGN_ERR);
	
  return real_count;
}
/* SGN_BS_process_multi() ====================================================== */

int SGN_BS_process_files(SGN_BS * bs, IOReader reader, SGN_StrArray * files){
  int count = 0;
  //fprintf(SIGN_ERR, "DEBUG: SGN_StrArray_size(files) = %zu\n", SGN_StrArray_size(files));
  for(int i = 0; i < SGN_StrArray_size(files); i++){
	count += SGN_BS_process_multi(bs, reader, SGN_StrArray_get(files, i),
								  NULL, NULL, NULL);
  }
  return count;
}
/* SGN_BS_process_files() ====================================================== */

int SGN_BS_process_files_dist(SGN_BS * bs, IOReader reader, SGN_StrArray * files,
						  const char * freq_key, const char * ud_key, const char * total_key){
  int count = 0;
  //fprintf(SIGN_ERR, "DEBUG: SGN_StrArray_size(files) = %zu\n", SGN_StrArray_size(files));
  for(int i = 0; i < SGN_StrArray_size(files); i++){
	count += SGN_BS_process_multi(bs, reader, SGN_StrArray_get(files, i),
							  freq_key, ud_key, total_key);
  }
  return count;
}
/* SGN_BS_process_files() ====================================================== */


int SGN_BS_op_ge(const double a, const double b){
  return a >= b;
}
int SGN_BS_op_le(const double a, const double b){
  return a <= b;
}

/*!\brief Compiles bootstrapped networks into a single network.
 *
 * The following network propeties are added to the returned \ref SGN_Network instance.
 *  - <tt>BS.total</tt> (integer):  Total number of processed networks.
 *
 *\param bs
 *\param th Threshold.
 *\param total Total number of networks to be processed.
 *\param cmp 0: Extracts edges whose frequences are greater than \a th,
 *           1: Extracts edges whose frequences are less than \a th.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_BS */
SGN_Network * SGN_BS_compile_dynamic(SGN_BS * bs, double th, int total, int cmp){
  SGN_Network * network = SGN_Network_new();
  int i, j;
  char buff[SGN_BS_BUFF_SIZE];
  int (*op)(const double, const double);

  fprintf(SIGN_ERR, "Processing for the dynamic networks.\n");
  if(cmp == 0){
	op = SGN_BS_op_ge;
  } else {
	op = SGN_BS_op_le;
  }

  if(bs->fnet != NULL){
	for(i = 0; i < bs->nodes; i++){
	  SGN_Node * node = SGN_Network_get_node(bs->fnet, i);
	  SGN_Network_add_node(network, SGN_Node_clone(node));
	}
  } else if(bs->data != NULL){
	for(i = 0; i < bs->nodes; i++){
	  SGN_Node * node = SGN_Node_new(SGN_Data_get_name(bs->data, i),
									 SGN_Data_get_alias(bs->data, i));
	  SGN_Network_add_node(network, node);
	}
  } else {
	fprintf(SIGN_ERR, "WARNING: SGN_BS: Network or dataset not set."
			" Using default node names.\n");
	fflush(SIGN_ERR);
	char buff[16];
	for(int i = 0; i < bs->nodes; i++){
	  sprintf(buff, "gene%d", i + 1);
	  SGN_Node * node = SGN_Node_new(buff, buff);
	  SGN_Network_add_node(network, node);
	}
  }

  for(i = 0; i < bs->nodes; i++){
	for(j = 0; j < bs->nodes; j++){
	  double p = bs->graph[i][j] / (double) total;
	  double gain = bs->gain[i][j];

	  //fprintf(SIGN_ERR, "p=%f, th=%f\n", p, th); fflush(SIGN_ERR);

	  if(op(p, th)){
		SGN_Node * parent = SGN_Network_get_node(network, i);
		SGN_Node * child = SGN_Network_get_node(network, j);
		SGN_Edge * edge = SGN_Edge_new(parent, child);

		//fprintf(SIGN_ERR, "hoge\n");

		if(bs->one_minus) p = 1.0 - p;
		SGN_Edge_add_property_double(edge, bs->label_prob, p);
		if(bs->label_gain != NULL){
		  SGN_Edge_add_property_double(edge, bs->label_gain, gain / bs->graph[i][j]);
		}
		
		int type;
		if(bs->ud[i][j][3] > bs->ud[i][j][0] && bs->ud[i][j][3] > bs->ud[i][j][1]){
		  /* Unknown */
		  type = 3;
		} else if(bs->ud[i][j][0] > bs->ud[i][j][1]){
		  /* Up */
		  type = 0;
		} else if(bs->ud[i][j][0] < bs->ud[i][j][1]){
		  /* Down */
		  type = 1;
		} else {
		  /* Unknown */
		  type = 3;
		}
		if(bs->label_type != NULL){
		  SGN_Edge_add_property_int(edge, bs->label_type, type);
		}
		if(bs->label_ud != NULL){
		  if(type == 0){
			SGN_Edge_add_property_string(edge, bs->label_ud, "up");
		  } else if(type == 1) {
			SGN_Edge_add_property_string(edge, bs->label_ud, "down");
		  } else {
			SGN_Edge_add_property_string(edge, bs->label_ud, "unknown");
		  }
		}
		int edges = bs->ud[i][j][0] + bs->ud[i][j][1] + bs->ud[i][j][3];
		if(edge == 0){
		  fprintf(SIGN_ERR, "FATAL ERROR: no edges detected.\n");
		  exit(1);
		}
		if(bs->label_ratio != NULL){
		  sprintf(buff, "%.2f/%.2f/%.2f",
				  (double) bs->ud[i][j][0] / (double) edges,
				  (double) bs->ud[i][j][1] / (double) edges,
				  (double) bs->ud[i][j][3] / (double) edges);
		  SGN_Edge_add_property_string(edge, bs->label_ratio, buff);
		}

		SGN_Network_add_edge(network, edge);

		//if(i == j) fprintf(SIGN_ERR, " selfloop detected i=%d\n", i);
	  }
	}/* for j */
  }/* for i */

  SGN_Network_add_property_int(network, "BS.total", total);
  return network;
}
/* SGN_BS_compile_dynamic() ================================================ */

/*!\brief
 *
 *\param cmp  0 for collecting edges whose probabilies are greater than the threshold.
 *              1 for less than the threshold.
 *
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_BS */
SGN_Network * SGN_BS_compile(SGN_BS * bs, double th, int total, int cmp, int dynamic){
  int (*op)(const double, const double);
  char buff[SGN_BS_BUFF_SIZE];

  if(cmp == 0){
	op = SGN_BS_op_ge;
  } else {
	op = SGN_BS_op_le;
  }

  SGN_Network * network = SGN_Network_new();

  /* Add nodes to the network */
  if(bs->fnet != NULL){
	for(int i = 0; i < bs->nodes; i++){
	  SGN_Node * node = SGN_Network_get_node(bs->fnet, i);
	  SGN_Network_add_node(network, SGN_Node_clone(node));
	}
  } else if(bs->data != NULL){
	for(int i = 0; i < bs->nodes; i++){
	  SGN_Node * node = SGN_Node_new(SGN_Data_get_name(bs->data, i),
									 SGN_Data_get_alias(bs->data, i));
	  SGN_Network_add_node(network, node);
	}
  } else {
	fprintf(SIGN_ERR, "WARNING: SGN_BS: Network or dataset not set."
			" Using default node names.\n");
	fflush(SIGN_ERR);
	char buff[16];
	for(int i = 0; i < bs->nodes; i++){
	  sprintf(buff, "gene%d", i + 1);
	  SGN_Node * node = SGN_Node_new(buff, buff);
	  SGN_Network_add_node(network, node);
	}
  }

  for(int i = 0; i < bs->nodes; i++){
	int j_begin = dynamic ? 0 : (i + 1);
	
	for(int j = j_begin; j < bs->nodes; j++){
	  double p;
	  double sum;

	  if(dynamic){
		sum = (double) (bs->graph[i][j]);
		p = sum / (double) total;
	  } else {
		sum = (double) (bs->graph[i][j] + bs->graph[j][i]);
		p = sum / (double) total;
	  }

	  if(op(p, th)){
		int pa = i;
		int ch = j;

		if(!dynamic && bs->graph[i][j] < bs->graph[j][i]){
		  pa = j;
		  ch = i;
		}

		SGN_Node * parent = SGN_Network_get_node(network, pa);
		SGN_Node * child = SGN_Network_get_node(network, ch);
		SGN_Edge * edge = SGN_Edge_new(parent, child);
		int dir = bs->graph[pa][ch];
		double gain = bs->gain[pa][ch];

		if(bs->one_minus) p = 1.0 - p;

		/* BS.Prob */
		SGN_Edge_add_property_double(edge, bs->label_prob, p);

		/* edgeScore */
		if(bs->label_score != NULL){
		  SGN_Edge_add_property_double(edge, bs->label_score, p);
		}

		/* BS.Gain */
		if(bs->label_gain != NULL){
		  SGN_Edge_add_property_double(edge, bs->label_gain, gain / (double) dir);
		}

		/* BS.HubIndex */
		double hubindex;
		if(dynamic){
		  hubindex = p;
		} else {
		  hubindex = (double) dir / (double) sum;
		}
		bs->hubindex[pa] += hubindex;
		if(!dynamic){
		  bs->hubindex[ch] += (1.0 - hubindex);
		}
		
		/* BS.Direction */
		if(!dynamic && bs->label_dir != NULL){
		  SGN_Edge_add_property_double(edge, bs->label_dir, hubindex);
		}

		/* BS.up/down */
		int type;
		if(bs->ud[pa][ch][3] > bs->ud[pa][ch][0] &&
		   bs->ud[pa][ch][3] > bs->ud[pa][ch][1]){
		  /* Unknown */
		  type = 3;
		} else if(bs->ud[pa][ch][0] > bs->ud[pa][ch][1]){
		  /* Up */
		  type = 0;
		} else if(bs->ud[pa][ch][0] < bs->ud[pa][ch][1]){
		  /* Down */
		  type = 1;
		} else {
		  /* Unknown */
		  type = 3;
		}
		if(bs->label_type != NULL){
		  SGN_Edge_add_property_int(edge, bs->label_type, type);
		}
		if(bs->label_ud != NULL){
		  if(type == 0){
			SGN_Edge_add_property_string(edge, bs->label_ud, "up");
		  } else if(type == 1) {
			SGN_Edge_add_property_string(edge, bs->label_ud, "down");
		  } else {
			SGN_Edge_add_property_string(edge, bs->label_ud, "unknown");
		  }
		}
		
		/* BS.EdgeTypeRatio */
		int edges = bs->ud[pa][ch][0] + bs->ud[pa][ch][1] + bs->ud[pa][ch][3];
		if(edge == 0){
		  fprintf(SIGN_ERR, "FATAL ERROR: no edges detected.\n");
		  exit(1);
		}
		if(bs->label_ratio != NULL){
		  sprintf(buff, "%.2f/%.2f/%.2f",
				  (double) bs->ud[pa][ch][0] / (double) edges,
				  (double) bs->ud[pa][ch][1] / (double) edges,
				  (double) bs->ud[pa][ch][3] / (double) edges);
		  SGN_Edge_add_property_string(edge, bs->label_ratio, buff);
		}
		
		SGN_Network_add_edge(network, edge);
	  }/* if(op(p, th)) */
	} /* for j */
  }/* for i */

  for(int i = 0; i < bs->nodes; i++){
	SGN_Node * node = SGN_Network_get_node(network, i);
	CSMLObject * csmlobj = SGN_Node_get_CSMLObject(node);
	if(bs->label_hubness != NULL){
	  CSMLObject_add_property(csmlobj, bs->label_hubness,
							  SGN_Object_new_int(SGN_Network_num_children(network, i)));
	}
	if(bs->label_hubindex != NULL){
	  CSMLObject_add_property(csmlobj, bs->label_hubindex,
							  SGN_Object_new_double(bs->hubindex[i]));
	}
  }/* for i < bs->nodes */

  SGN_Network_add_property_int(network, "BS.total", total);
  return network;
}
/* SGN_BS_compile() ======================================================== */

/* End of file.  Copyright (C) 2010-2011, 2014  SiGN Project Members */
