/*
  util/SGN_Color.c : Color module

  Copyright (C) 2010 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.
  
  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
 */

/*!\defgroup SGN_Color
 *\brief Color module.
 *
 * An \ref SGN_Color instance is to represent an RGB color.
 *
 *\author Yoshinori Taada &lt;tamada at ims.u-tokyo.ac.jp&gt;
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "SGN_Color.h"

/*!\brief Generates a new \ref SGN_Color instance.
 *
 * This generates and returns a new \ref SGN_Color instance.
 * The default color is black.
 *
 * The returning pointer needs to be released by
 * SGN_Color_delete() when it is no longer required.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Color */
SGN_Color * SGN_Color_new(){
  SGN_Color * this = malloc(sizeof(SGN_Color));
  this->r = 0;
  this->g = 0;
  this->b = 0;
  this->a = 255;
  return this;
}
/* SGN_Color_new() ============================================================= */

SGN_Color * SGN_Color_new_RGB(int r, int g, int b, int a){
  SGN_Color * this = SGN_Color_new();
  SGN_Color_set_RGB(this, r, g, b, a);
  return this;
}
/* SGN_Color_new_RGB() ========================================================= */

/*!\brief Deletes the \ref SGN_Color instance.
 *
 * This deletes the \ref SGN_Color instance generated by SGN_Color_new().
 *
 *\param this  pointer to the \ref SGN_Color instance to be deleted.
 *                   NULL acceptable.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Color */
void SGN_Color_delete(void * this){
  free(this);
}
/* SGN_Color_delete() ========================================================== */

/*!\brief Generates the complete copy of the \ref SGN_Color instance.
 *
 *\param this  pointer to the \ref SGN_Color instance to be cloned.
 *                  NULL is acceptable.  If so, NULL will return.
 *\return new \ref SGN_Color instance identical to the given \ref SGN_Color instance.
 *                     Or NULL if NULL is given as an argument.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Color */
SGN_Color * SGN_Color_clone(const SGN_Color * this){
  if(this == NULL) return NULL;
  
  SGN_Color * color = SGN_Color_new();
  *color = *this; /* Shallow copy is enogh here */

  return color;
}
/* SGN_Color_clone() =========================================================== */

/*!\brief Sets the color by the \ref SGN_Color instance.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Color */
void SGN_Color_set(SGN_Color * this, const SGN_Color * color){
  *this = *color; /* Shallow copy */
}
/* SGN_Color_set() ============================================================= */

/*!\brief Sets a color in RGB.
 *
 * All values of red, green and blue are raningin 0 through 255.
 * when a value outside of this ranige is specified, it is converted to fit
 * to this range.
 *
 *\param r   value of red.
 *\param g   value of green.
 *\param b   value of blue.
 *\param a   value of alpha blending.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Color */
void SGN_Color_set_RGB(SGN_Color * this, int r, int g, int b, int a){
  if(r > 255) r = 255;  if(r < 0) r = 0;
  if(g > 255) g = 255;  if(g < 0) g = 0;
  if(b > 255) b = 255;  if(b < 0) b = 0;
  if(a > 255) a = 255;  if(a < 0) a = 0;

  this->r = r;
  this->g = g;
  this->b = b;
  this->a = a;
}
/* SGN_Color_set_RGB() ========================================================= */

/*!\brief Sets a color in HSV.
 *
 *\param h  value of hue.
 *\param s  value of saturation.
 *\param v  value of brightness.
 *\param a  value of alpha blending.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Color */
void SGN_Color_set_HSV(SGN_Color * this, int h, int s, int v, int a){
  SGN_Color_HSV2RGB(h, s, v, &this->r, &this->g, &this->b);
  this->a = a;
}
/* SGN_Color_set_HSV() ========================================================= */

/*!\brief Prints the \ref SGN_Color expression in RGB to the string buffer.
 *
 * This prints the color simply in space-separated format, such as
 * &quot;<TT>255 155 128 255</TT>&quot;.
 * The order of colors is red, green, blue and alpha.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Color */
void SGN_Color_sprint(const SGN_Color * this, char * buff){
  sprintf(buff, "%d %d %d %d", this->r, this->g, this->b, this->a);
}
/* SGN_Color_sprint() ========================================================== */

/*!\brief Prints the \ref SGN_Color expression in RGB to the file stream.
 *
 * This prints the color simply in space-separated format, such as
 * &quot;<TT>255 155 128 255</TT>&quot;.
 * The order of colors is red, green, blue and alpha.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Color */
void SGN_Color_fprint(const SGN_Color * this, FILE * fp){
  fprintf(fp, "%d %d %d %d", this->r, this->g, this->b, this->a);
}
/* SGN_Color_fprint() ========================================================== */

/*!\brief Converts a color in HSV to be in RGB.
 *
 *\param h Hue (0 - 255).  If the value greater than 255 is specified, then
 *          \a h mod 256 is used.
 *\param s Satulation (0 - 255).
 *\param v Value (Brightness) (0 - 255).
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Color */
void SGN_Color_HSV2RGB(int h, int s, int v,
				   int * r, int * g, int * b){
  
  if(h < 0) h = 0;
  if(255 < h) h = h % 256;
  if(s < 0) s = 0;
  if(255 < s) s = 255;
  if(v < 0) v = 0;
  if(255 < v) v = 255;

  if(s == 0){
	*r = v; *g = v; *b = v;
	return;
  }

  if(h == 255) h = 0;
  double hh = ((double) h) * (360.0 / 255.0);
  double ss = ((double) s) / 255.0;
  double vv = ((double) v) / 255.0;

  //fprintf(SIGN_ERR, "hh=%f, ss=%f, vv=%f\n", hh, ss, vv);

  int hi = ((int) floor(hh / 60.0)) % 6;

  /*
  double c = v * s;
  double x = c * (1.0 - fabs(hh / 60 ));
  double m = v - c;
  */

  double f = (hh / 60.0) - (double) hi;
  double p = vv * (1.0 - ss);
  double q = vv * (1.0 - f * ss);
  double t = vv * (1.0 - (1.0 - f) * ss);

  //fprintf(SIGN_ERR, "f=%f\n", f);

  switch(hi){
  case 0:
	*r = v;                  /* c + m = c + (v - c) = v*/
	*g = (int) (t * 255.0);  /* x + m = c(1.0 - ) (v - c) */
	*b = (int) (p * 255.0);  /* m = v - c = v - (v * s) = v (1 - s) */
	break;
  case 1:
	*r = (int) (q * 255.0);
	*g = v;
	*b = (int) (p * 255.0);
	break;
  case 2:
	*r = (int) (p * 255.0);
	*g = v;
	*b = (int) (t * 255.0);
	break;
  case 3:
	*r = (int) (p * 255.0);
	*g = (int) (q * 255.0);
	*b = v;
	break;
  case 4:
	*r = (int) (t * 255.0);
	*g = (int) (p * 255.0);
	*b = v;
	break;
  case 5:
	*r = v;
	*g = (int) (p * 255.0);
	*b = (int) (q * 255.0);
	break;
  }
  //fprintf(SIGN_ERR, "r=%d, g=%d, b=%d\n", *r, *g, *b);
}
/* SGN_Color_HSV2RGB() ================================================== */

void SGN_Color_set_grad(const SGN_Color * c1, const SGN_Color * c2,	double r, SGN_Color * c){
  if(r < 0.0) r = 0.0;
  if(1.0 < r) r = 1.0;

  c->r = c1->r + (c2->r - c1->r) * r;
  c->g = c1->g + (c2->g - c1->g) * r;
  c->b = c1->b + (c2->b - c1->b) * r;
  c->a = c1->a + (c2->a - c1->a) * r;
}
/* void SGN_Color_set_grad() ============================================== */

/* End of file.  Copyright (C) 2010  SiGN Project Members */
