/*
  sign/util/SGN_Data.c : SGN_Data container

  Copyright (C) 2008-2011 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#include <string.h>

#include "../signcom.h"
#include "SGN_Data.h"
#include "SGN_Memory.h"
#include "SGN_IntArray.h"
#include "SGN_Str.h"
#include "SGN_Util.h"
#include "../math/SGN_Math.h"

/*!\brief Generates a new \ref SGN_Data instance.
 *
 * The generated instance has no data matrix values.
 * To generate the instance from the file,
 * use functions in \ref DataMatrix or \ref EDF class,
 * instead using this.
 *
 *\return pointer to a new \ref SGN_Data instance that has no data.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
SGN_Data * SGN_Data_new(){
  SGN_Data * dataset = (SGN_Data *) malloc(sizeof(SGN_Data));
  dataset->n = 0;
  dataset->p = 0;
  dataset->type = SGN_Data_TYPE_NUMERICAL;
  dataset->X = NULL;
  dataset->Y = NULL;
  dataset->categories = NULL;
  dataset->labels = NULL;
  dataset->meta = NULL;
  dataset->attr = NULL;
  dataset->primaryID = NULL;
  dataset->secondaryID = NULL;

  dataset->names = SGN_StrArray_new();
  dataset->aliases = SGN_StrArray_new();
  return dataset;
}
/* SGN_Data_new() ----------------------------------------------------------- */

/*!\brief Generates a clone of the \ref SGN_Data instance.
 *
 *\param \ref SGN_Data instance to be cloned.
 *\return pointer to the new instance that has the same data.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
SGN_Data * SGN_Data_clone(const SGN_Data * data){
  SGN_Data * ds = (SGN_Data *) malloc(sizeof(SGN_Data));

  ds->n = data->n;
  ds->p = data->p;
  ds->type = data->type;

  ds->X = malloc(sizeof(double) * ds->n * ds->p);
  memcpy(ds->X, data->X, sizeof(double) * ds->n * ds->p);
  if(data->X != data->Y){
	ds->Y = malloc(sizeof(double) * ds->n * ds->p);
	memcpy(ds->Y, data->Y, sizeof(double) * ds->n * ds->p);
  } else {
	ds->Y = ds->X;
  }

  ds->names = SGN_StrArray_clone(data->names);
  ds->aliases = SGN_StrArray_clone(data->aliases);

  if(data->categories != NULL){
	ds->categories = SGN_StrArray_clone_ar(data->categories, ds->n);
  }
  if(data->labels != NULL){
	ds->labels = SGN_StrArray_clone(data->labels);
  }
  if(data->primaryID != NULL){
	ds->primaryID = SGN_IntArray_clone(data->primaryID);
  }
  if(data->secondaryID != NULL){
	ds->secondaryID = SGN_IntArray_clone(data->secondaryID);
  }
  return ds;
}
/* SGN_Data_clone() ========================================================= */

/*!\brief Deletes the \ref SGN_Data instance.
 *
 * This releases all resources allocated for the given \ref SGN_Data instance.
 *\param dataset  pointer to the \ref SGN_Data instance to be deleted.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_delete(SGN_Data * data){
  if(data->Y != data->X){
	free(data->Y);
  }
  free(data->X);

#ifdef SIGN_DEBUG
  fprintf(SIGN_ERR, "SGN_Data_delete(): Deleting names.\n");
  fflush(SIGN_ERR);
#endif
  SGN_StrArray_delete_all(data->names);
#ifdef SIGN_DEBUG
  fprintf(SIGN_ERR, "SGN_Data_delete(): Deleting aliases.\n");
  fflush(SIGN_ERR);
#endif
  SGN_StrArray_delete_all(data->aliases);
#ifdef SIGN_DEBUG
  fprintf(SIGN_ERR, "SGN_Data_delete(): Deleting categories.\n");
  fflush(SIGN_ERR);
#endif
  if(data->categories != NULL){
	SGN_StrArray_delete_ar_all(data->categories, data->p);
  }
#ifdef SIGN_DEBUG
  fprintf(SIGN_ERR, "SGN_Data_delete(): Deleting labels.\n");
  fflush(SIGN_ERR);
#endif
  if(data->meta != NULL){
	SGN_KeyValues_delete(data->meta);
  }
  if(data->attr != NULL){
	SGN_KeyValues_delete(data->attr);
  }
  if(data->primaryID != NULL){
	SGN_IntArray_delete(data->primaryID);
  }
  if(data->secondaryID != NULL){
	SGN_IntArray_delete(data->secondaryID);
  }

  SGN_StrArray_delete_all(data->labels);

  free(data);
}
/* SGN_Data_delete() -------------------------------------------------------- */

/*!\brief Returns the number of nodes in the \ref SGN_Data.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
int SGN_Data_nodes(const SGN_Data * data){
  return data->p;
}
/* SGN_Data_nodes() ========================================================= */

/*!\brief Returns the number of samples in the \ref SGN_Data.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
int SGN_Data_samples(const SGN_Data * data){
  return data->n;
}
/* SGN_Data_samples() ======================================================= */

/*!\brief Returns the name of the node.
 *\param data pointer to the \ref SGN_Data instance.
 *\param index  position of the node whose name is returned.
 *\return pointer to the string of the node name, or NULL if the name is not
 *           set.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
const char * SGN_Data_get_name(const SGN_Data * data, size_t index){
  if(data->names == NULL){
	return NULL;
  } if(SGN_StrArray_size(data->names) <= index){
	return NULL;
  } else {
	return SGN_StrArray_get(data->names, index);
  }
}
/* SGN_Data_get_name() ------------------------------------------------------ */

/*!\brief Returns the alias name of the node.
 *\param data pointer to the \ref SGN_Data instance.
 *\param index  position of the node whose name is returned.
 *\return pointer to the string of the node alias name, or NULL if the name is
 *          not set.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
const char * SGN_Data_get_alias(const SGN_Data * data, size_t index){
  if(data->aliases == NULL){
	return NULL;
  } if(SGN_StrArray_size(data->aliases) <= index){
	return NULL;
  } else {
	return SGN_StrArray_get(data->aliases, index);
  }
}
/* SGN_Data_get_alias() ----------------------------------------------------- */

double SGN_Data_get_value(SGN_Data * data, int row, int col){
  return data->X[row + col * data->n];
}
/* SGN_Data_get_value() ----------------------------------------------------- */

double SGN_Data_value(SGN_Data * data, int node, int sample){
  return data->X[node + sample * data->n];
}
/* SGN_Data_value() ========================================================= */

/*!\brief Returns a value in X.
 *
 *
 *\param node  index of node (row)
 *\param sample index of sample (column)
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
double SGN_Data_value_X(SGN_Data * data, int node, int sample){
  return data->X[sample + node * data->n];
}
/* SGN_Data_value_X() ======================================================= */

double SGN_Data_get_value_Y(SGN_Data * data, int row, int col){
  return data->Y[row + col * data->n];
}
/* SGN_Data_get_value_Y() =================================================== */

/*!\brief Returns a value in Y.
 *
 *
 *\param node  index of node (row)
 *\param sample index of sample (column)
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
double SGN_Data_value_Y(SGN_Data * data, int node, int sample){
  return data->Y[sample + node * data->n];
}
/* SGN_Data_value_Y() ======================================================= */

/*!\brief Finds the node which has the specified name.
 *
 *
 *\return index of the node whose name is \a name, or SGN_StrArray_ERROR_VALUE
 *   if not found.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
size_t SGN_Data_find_name(const SGN_Data * data, const char * name){
  if(data->names == NULL) return SGN_StrArray_ERROR_VALUE;

  return SGN_StrArray_find(data->names, name);
}
/* ========================================================================= */

/*!\brief Adds the indexes of nodes given by the file.
 *
 * This reads node names from a tab-separated file,
 * then searches for them from the given SGN_Data instance
 * to acquire indexes of the nodes.
 *
 * The collected indexes are added to SGN_IntArray \a nodes.
 * This does not clear \a nodes before adding the indexes.
 * When adding indexes into \a nodes, duplication is checked.
 *
 * Warning messages can be controled by SGN_set_warning_level().
 * When the warning level is set to 0, this does not output any warning
 * messages.
 *
 *\param data SGN_Data instance.
 *\param nodes SGN_IntArray instance to which the searched indexes are added.
 *\param file  tab-separated file of the node names.
 *\param col column position of the node names in the file. 0-origin.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_read_node_set_file(const SGN_Data * data, SGN_IntArray * nodes,
								 const char * file, int col){
  FILE * fp;
  char * buff = NULL;
  size_t buff_size = 0;
  
  fp = fopen(file, "r");
  if(fp == NULL){
	fprintf(SIGN_ERR, "ERROR: cannot open file: %s\n", file);
  }
  while(SGN_Str_readline(fp, &buff, &buff_size) == 0){
	int id, r;
	SGN_StrArray * tokens = SGN_Str_split(buff, '\t');
	char * node;
	if(col >= SGN_StrArray_size(tokens)){
	  fprintf(SIGN_ERR, "ERROR: SGN_Data: node selection: "
			  "index out of range :%d\n", col);
	  exit(1);
	}
	node = SGN_StrArray_get(tokens, col);
	id  = SGN_StrArray_find(data->names, node);
	if(id == SGN_StrArray_ERROR_VALUE && 1 <= SGN_get_warning_level()){
	  fprintf(SIGN_ERR, "WARNING: not found: %s\n", node);
	  continue;
	}

	r = SGN_IntArray_find(nodes, id);
	if(r == SGN_IntArray_ERROR_VALUE){
	  SGN_IntArray_add(nodes, id);
	  //fprintf(SIGN_ERR, "%s\n", node);
	} else if(1 <= SGN_get_warning_level()){
	  fprintf(SIGN_ERR, "WARNING: duplication detected: %s\n", node);
	}
	SGN_StrArray_delete_all(tokens);
  }
  fclose(fp);
}
/* SGN_Data_read_node_set_file() -------------------------------------------- */

/*!\brief Selects nodes from the multiple files.
 *
 *\param data  pointer to the \ref SGN_Data instance.
 *\param files  array containing the file names.
 *\param col  column  index (position) of each file to read.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_select_nodes_files(SGN_Data * data, SGN_StrArray * files, int col){
  SGN_IntArray * nodes;
  int i;

  nodes = SGN_IntArray_new();
  for(i = 0; i < SGN_StrArray_size(files); i++){
	//fprintf(SIGN_ERR, "select nodes: %s\n", SGN_StrArray_get(files, i));
	SGN_Data_read_node_set_file(data, nodes, SGN_StrArray_get(files, i), col);
  }
  fprintf(SIGN_ERR, "Select nodes: Total %d nodes.\n", (int) SGN_IntArray_size(nodes));

  SGN_Data_select_nodes_index(data, nodes);
  SGN_IntArray_delete(nodes);
}
/* SGN_Data_select_nodes_files() --------------------------------------------- */

/*!\brief Selects nodes from the multiple files.
 *
 * Unlike SGN_Data_select_nodes_file(), this returns the selected
 * indices in the original data.
 *
 *\param data  pointer to the \ref SGN_Data instance.
 *\param files  array containing the file names.
 *\param col  column  index (position) of each file to read.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
SGN_IntArray * SGN_Data_select_nodes_files2(SGN_Data * data, SGN_StrArray * files, int col){
  SGN_IntArray * nodes;
  int i;

  nodes = SGN_IntArray_new();
  for(i = 0; i < SGN_StrArray_size(files); i++){
	//fprintf(SIGN_ERR, "select nodes: %s\n", SGN_StrArray_get(files, i));
	SGN_Data_read_node_set_file(data, nodes, SGN_StrArray_get(files, i), col);
  }
  fprintf(SIGN_ERR, "Select nodes: Total %d nodes.\n", (int) SGN_IntArray_size(nodes));

  SGN_Data_select_nodes_index(data, nodes);

  return nodes;
}
/* SGN_Data_select_nodes_files() --------------------------------------------- */

void SGN_Data_select_nodes_index(SGN_Data * data, SGN_IntArray * nodes){
  double * new_X;
  int new_p = SGN_IntArray_size(nodes);
  SGN_StrArray * new_names = SGN_StrArray_new();
  SGN_StrArray * new_aliases = SGN_StrArray_new();
  int i, j;

  new_X = malloc(sizeof(double) * data->n * new_p);
  for(j = 0; j < SGN_IntArray_size(nodes); j++){
	int jj = SGN_IntArray_get(nodes, j);
	for(i = 0; i < data->n; i++){
	  new_X[i + j * data->n] = data->X[i + jj * data->n];
	}
	SGN_StrArray_add(new_names, SGN_Str_copy(SGN_StrArray_get(data->names, jj)));
	SGN_StrArray_add(new_aliases, SGN_Str_copy(SGN_StrArray_get(data->aliases, jj)));
  }

  free(data->X);
  SGN_StrArray_delete_all(data->names);
  SGN_StrArray_delete_all(data->aliases);

  data->X = new_X;
  data->Y = new_X;
  data->names = new_names;
  data->aliases = new_aliases;
  data->p = new_p;
}
/* SGN_Data_select_nodes_index() -------------------------------------------- */

SGN_Data * SGN_Data_select_nodes_index_new(const SGN_Data * data,
										 SGN_IntArray * nodes){
  SGN_Data * ret = SGN_Data_new();
  size_t i, j;
  int flag = 1;

  ret->p = SGN_IntArray_size(nodes);
  ret->n = data->n;
  ret->X = malloc(sizeof(double) * ret->p * ret->n);

  if(data->X == data->Y){
	ret->Y = ret->X;
	flag = 0;
  }
  
  for(j = 0; j < SGN_IntArray_size(nodes); j++){
	size_t jj = SGN_IntArray_get(nodes, j);
	for(i = 0; i < ret->n; i++){
	  ret->X[i + j * ret->n] = data->X[i + jj * data->n];
	  if(flag){
		ret->Y[i + j * ret->n] = data->Y[i + jj * data->n];
	  }
	}
	SGN_StrArray_add(ret->names, SGN_Str_copy(SGN_StrArray_get(data->names, jj)));
	SGN_StrArray_add(ret->aliases, SGN_Str_copy(SGN_StrArray_get(data->aliases, jj)));
  }

  return ret;
}
/* SGN_Data_select_nodes_index_new() ---------------------------------------- */

/*!\brief Selects the samples (columns) in the data to use.
 *
 * This sets the same pointer to SGN_Data::X and SGN_Data::Y.
 *
 *\param samples   indexes of samples (columns) (1-origin).
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_select_samples(SGN_Data * data, SGN_IntArray * samples){
  double * tmp;
  size_t new_n = SGN_IntArray_size(samples);
  int i, j;

  if(new_n == 0) return;

  tmp = (double *) malloc(sizeof(double) * data->p * new_n);
  for(i = 0; i < new_n; i++){
	int ii = SGN_IntArray_get(samples, i) - 1;
	if(ii < 0){
	  fprintf(SIGN_ERR, "ERROR: SGN_Data: invalid sample index specified: %d", 
			  ii + 1);
	  exit(1);
	}
	for(j = 0; j < data->p; j++){
	  tmp[i + j * new_n] = data->X[ii + j * data->n];
	}/* for j */
  }/* for i */

  if(data->X != data->Y) free(data->Y);
  free(data->X);
  data->n = new_n;
  data->X = tmp;
  data->Y = tmp;
}
/* SGN_Data_select_samples() ------------------------------------------------ */

/*!\brief Selects the samples (columns) at random with replacement.
 *
 * This is to perform random sampling for bootstrap methods.
 *
 * This sets the same pointer to SGN_Data::X and SGN_Data::Y.
 * Do not apply this for dynamic (time course) data.
 *
 * If users want to make a smaller sub set of samples at random,
 * use SGN_Data_random_sampling2() instead of this.
 *
 *\param data  pointer to the \ref SGN_Data instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_random_sampling(SGN_Data * data){
  double * tmp;
  int i, j;

  tmp = (double *) malloc(sizeof(double) * data->p * data->n);
  for(i = 0; i < data->n; i++){
	int ii = random() % data->n;
	for(j = 0; j < data->p; j++){
	  tmp[i + j * data->n] = data->X[ii + j * data->n];
	}/* for j */
  }/* for i */

  if(data->X != data->Y) free(data->Y);
  free(data->X);
  data->X = tmp;
  data->Y = tmp;
}
/* SGN_Data_random_sampling() ----------------------------------------------- */

/*!\brief Selects the samples (columns) at random with replacement and
 *   sets them to the data.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_random_sampling_SGN_Data(SGN_Data * data, const SGN_Data * src){
  for(int i = 0; i < data->n; i++){
	int ii = random() % data->n;
	for(int j = 0; j < data->p; j++){
	  data->X[i + j * data->n] = src->X[ii + j * data->n];
	}/* for j */
  }/* for i */

  data->Y = data->X;
}
/* SGN_Data_random_sampling_SGN_Data() ======================================= */

/*!\brief Selects the samples (columns) at random with replacement and
 *   sets them to the data.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_random_sampling_SGN_Data_RNG(SGN_Data * data, const SGN_Data * src, SGN_RNG * rng){
  int rnd[data->n];
  SGN_RNG_integer(rng, rnd, data->n, 0, data->n);
  for(int i = 0; i < data->n; i++){
	for(int j = 0; j < data->p; j++){
	  data->X[i + j * data->n] = src->X[rnd[i] + j * data->n];
	}/* for j */
  }/* for i */

  data->Y = data->X;
}
/* SGN_Data_random_sampling_SGN_Data() ======================================= */

/*!\brief Select the samples (columns) at random.
 *
 * Unlike SGN_Data_random_sampling(), this extracts a subset of
 * samples at random without replacement.  That is, there are no
 * overlap columns after this random sampling.  The argument
 * \a num must be equal to or smaller than the original number of
 * samples in the \ref SGN_Data.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_random_sampling2(SGN_Data * data, int num){
  int cols[num];

  SGN_Util_random_sample(data->n, num, cols);
  double * tmp = (double *) malloc(sizeof(double) * data->p * num);
  for(int i = 0; i < num; i++){
	int ii = cols[i];
	for(int j = 0; j < data->p; j++){
	  tmp[i + j * num] = data->X[ii + j * data->n];
	}/* for j < p */
  }/* for i < num */

  if(data->labels != NULL){
	SGN_StrArray * new_labels = SGN_StrArray_news(num);
	for(int i = 0; i < num; i++){
	  SGN_StrArray_add_copy(new_labels, 
						SGN_StrArray_get(data->labels, cols[i]));
	}
	SGN_StrArray_delete_all(data->labels);
	data->labels = new_labels;
  }

  if(data->X != data->Y) free(data->Y);
  free(data->X);
  data->X = tmp;
  data->Y = tmp;
  data->n = num;

}
/* SGN_Data_random_sampling2() ============================================== */

/*!\brief Outputs the data into the file stream.
 *
 *\param data pointer to the \ref SGN_Data instance.
 *\param fp file pointer to which the data is output.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_output(SGN_Data * data, FILE * fp){
  int j, i;
  for(j = 0; j < data->p; j++){
	fprintf(fp, "%s\t%s",
			SGN_Data_get_name(data, j),
			SGN_Data_get_alias(data, j));
	for(i = 0; i < data->n; i++){
	  fprintf(fp, "\t%f", data->X[i + j * data->n]);
	}
	fprintf(fp, "\n");
  }
}
/* SGN_Data_output() -------------------------------------------------------- */

/*!\brief Discretizes the values into categorical values.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_discretize(SGN_Data * data, int categories){

  fprintf(SIGN_ERR, "SGN_Data: discretizaing ... (categories = %d)\n", 
		  categories);
  fflush(SIGN_ERR);

  if(data->X != data->Y){
	fprintf(SIGN_ERR, "ERROR: SGN_Data_discretize(): call this before converting"
			" for the dynamic model.\n");
	exit(1);
  }

  data->categories = SGN_StrArray_new_ar(data->p);
  data->type = SGN_Data_TYPE_CATEGORICAL;
  
  for(int j = 0; j < data->p; j++){
	double min, max;

	for(int k = 0; k < categories; k++){
	  char buff[100];
	  sprintf(buff, "Cat_%d", k + 1);
	  SGN_StrArray_add_copy(data->categories + j, buff);
	}/* for k < categories */

	SGN_Math_minmax(data->X + j * data->n, data->n, &min, &max);
	double interval = (max - min) / (double) categories;

	for(int i = 0; i < data->n; i++){
	  double th = min + interval;

	  int flag = 0;
	  //for(int k = 0; k < categories; k++){
	  for(int k = 0; k < categories - 1; k++){
		if(data->X[i + j * data->n] < th){
		  data->X[i + j * data->n] = (double) k;
		  flag = 1;
		  break;
		}
		th += interval;
	  }
	  if(flag == 0) data->X[i + j * data->n] = categories - 1;
	}/* for i < data->n */
  }/* for j < data->p */
  fprintf(SIGN_ERR, "SGN_Data:   finished.\n");
  fflush(SIGN_ERR);
}
/* SGN_Data_discretize() ==================================================== */

/*!\brief Converts the \ref SGN_Data for dynamical models.
 *
 *\param points  list of indexes pointing to the first positions (column indexes)
 *               of the contiguous blocks if there are more than two contiguous blocks
 *               in the data.  Note that the first index begins from
 *               1 (one), not 0 (zero), and you do not need to specify the first
 *               block in \a points.  If NULL is given, all the columns are
 *               regarded as a contiguous block.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_dynamic(SGN_Data * data, SGN_IntArray * points){
#ifdef SIGN_DEBUG
  fprintf(SIGN_ERR, "SGN_Data_dynamic() called.\n");
  fflush(SIGN_ERR);
#endif /* SIGN_DEBUG */

  
  int blocks;
  int n = data->n;
  if(points == NULL){
	blocks = 1;
  } else {
	blocks = SGN_IntArray_size(points) + 1;
  }
  int new_n = data->n - blocks;

  fprintf(SIGN_ERR, "Converting for the dynamical models.\n");
  fprintf(SIGN_ERR, "  blocks = %d, points={", blocks);
  if(points != NULL){
	for(int i = 0; i < SGN_IntArray_size(points); i++){
	  if(i != 0) fprintf(SIGN_ERR, ", ");
	  fprintf(SIGN_ERR, "%d", SGN_IntArray_get(points, i));
	}
  }
  fprintf(SIGN_ERR, "}\n  new n = %d\n", new_n);
  fflush(SIGN_ERR);

  double * new_Y, * new_X;
  new_Y = malloc(sizeof(double) * new_n * data->p);
  new_X = malloc(sizeof(double) * new_n * data->p);
  if(new_Y == NULL || new_X == NULL){
	fprintf(SIGN_ERR, "ERROR: SGN_Data_dynamic(): memory allocation failure.\n");
	exit(1);
  }

  int next = 0;

  for(int k = 0; k < blocks; k++){
	int pos = next;
	if(k < blocks - 1){
	  next = SGN_IntArray_get(points, k) - 1;
	} else {
	  next = data->n;
	}

#ifdef SIGN_DEBUG
	fprintf(SIGN_ERR, "  loop: k = %d, pos = %d, next = %d\n", k, pos, next);
	fflush(SIGN_ERR);
#endif /* SIGN_DEBUG */

	int npos = pos - k;
	size_t size = sizeof(double) * (next - pos - 1);
	for(int j = 0; j < data->p; j++){
	  memcpy(new_X + (npos + j * new_n), data->X + (pos     + j * n), size);
	  memcpy(new_Y + (npos + j * new_n), data->X + (pos + 1 + j * n), size);
	}/* for j < data->p */
  }/* for k < blocks */

  free(data->X);
  if(data->X != data->Y) free(data->Y);

  data->X = new_X;
  data->Y = new_Y;
  data->n = new_n;
}
/* SGN_Data_dynamic() ====================================================== */

void SGN_Data_dynamic_Data(SGN_Data * data, const SGN_IntArray * points,
						   const SGN_Data * src){
  int blocks;
  if(points == NULL){
	blocks = 1;
  } else {
	blocks = SGN_IntArray_size(points) + 1;
  }
  int new_n = src->n - blocks;

  int next = 0;
  for(int k = 0; k < blocks; k++){
	int pos = next;
	if(k < blocks - 1){
	  next = SGN_IntArray_get(points, k) - 1;
	} else {
	  next = src->n;
	}
	int npos = pos - k;
	size_t size = sizeof(double) * (next - pos - 1);
	for(int j = 0; j < src->p; j++){
	  memcpy(data->X + (npos + j * new_n), src->X + (pos     + j * src->n), size);
	  memcpy(data->Y + (npos + j * new_n), src->X + (pos + 1 + j * src->n), size);
	}/* for j < src->p */
  }/* for k < blocks */
  data->n = new_n;
}
/* SGN_Data_dynamic_Data() ================================================= */
							   

/*!\brief Performs resampling with replacement for pseudo bootstrapping.
 *
 *
 * The columns need to be in replicate-first order.
 *
 * The total number of samples used for the network estimation
 * becomes (Num. of time points) x (\a num).
 *
 *\param data   \ref SGN_Data instance to resample.
 *
 *\param replicates The number of replicates for each time point.  If it contains
 *         only one value,  all the time points are considered to have the same
 *         number of replicates.  Otherwise, it needs to have the numbers of
 *         replicates for all the time points separately.
 *
 *\param num        The number of blocks to be resampled.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_pseudo_bootstrap(SGN_Data * data, const SGN_IntArray * replicates, int num){

  SGN_IntArray * reps = SGN_IntArray_new();
  int n, newN;

  if(SGN_IntArray_size(replicates) == 1){
	/*
	  All the time points have the same numbers of replicates.
	 */
	int rep = SGN_IntArray_get(replicates, 0);
	if(data->n % rep != 0){
	  fprintf(SIGN_ERR, "ERROR: SGN_Data_pseudo_bootstrap(): The number of columns inconsistent"
			  " with the specifed number of replicates.\n");
	  exit(1);
	}
	n = data->n / rep;
	newN = num * (n - 1);

	for(int i = 0; i < n; i++) SGN_IntArray_add(reps, rep);
  } else {
	/*
	  Different replicates for time points.
	 */
	int sum = 0;
	SGN_IntArray_copy(replicates, reps);
	n = SGN_IntArray_size(reps);
	newN = num * (n - 1);

	for(int i = 0; i < n; i++) sum += SGN_IntArray_get(reps, i);
	if(data->n != sum){
	  fprintf(SIGN_ERR, "ERROR: SGN_Data_pseudo_bootstrap(): The number of columns inconsistent"
			  " with the specified numbers of replicates.\n");
	  fprintf(SIGN_ERR, "  reps={%d", SGN_IntArray_get(reps, 0));
	  for(int i = 1; i < n; i++) fprintf(SIGN_ERR, ", %d", SGN_IntArray_get(reps, i));
	  fprintf(SIGN_ERR, "}: ");
	  fprintf(SIGN_ERR, "  samples = %d  v.s.  total replicates = %d\n",
			  data->n, sum);
	  exit(1);
	}
  }

  double * newX = malloc(sizeof(double) * newN * data->p);
  double * newY = malloc(sizeof(double) * newN * data->p);
  
  SGN_Data_pseudo_bootstrap2(data->X, data->n, data->p,
							SGN_IntArray_ptr(reps), n, num,	NULL, 0,
							newX, newY);

  data->n = num * (n-1);

  if(data->X != data->Y) free(data->Y);
  free(data->X);
  
  data->X = newX;
  data->Y = newY;

  SGN_IntArray_delete(reps);
}
/* SGN_Data_pseudo_bootstrap() ============================================== */

/*!\brief Performs resampling with replacement for pseudo bootstrapping.
 *
 * \a X needs to be sorted in replicate-first order.
 * The resampled data values are set \a newX and \a newY.
 *
 *\param X   \a n x \a p matrix to be resampled.
 *\param n   number of samples of \a X.
 *\param p   number of nodes in \a X.
 *\param replicates  array of length \a t containing numbers of
 *           of replicates for all the time points.
 *\param t  number of time points (= length of \a replicates).
 *\param num  number of blocks to be resampled.
 *\param nodes list of nodes to be resampled.  if NULL, then all the
 *             \a p  nodes are resampled.
 *\param pp  number of resampling nodes if \a nodes is not NULL.
 *\param newX m x \a p matrix where m = (\a t - 1) x \a num.
 *\param newY m x \a p matrix where m = (\a t - 1) x \a num.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Data */
void SGN_Data_pseudo_bootstrap2(const double * X, int n, int p,
							   const int * replicates, int t, int num,
							   const int * nodes, int pp,
							   double * newX, double * newY){

  int newN = num * (t - 1);
  int pos[t]; /* Position indices of the first replicate for each time point. */
  int inc = 1;
  int j;

  if(nodes == NULL){ pp = p; inc = 0; nodes = &j; }

  /* Calculates the positions of the first replicate of all the time points. */
  pos[0] = 0;
  for(int i = 1; i < t; i++) pos[i] = pos[i - 1] + replicates[i - 1];

  fprintf(SIGN_ERR, "PSEUDO BOOTSTRAP ========\n");
  fprintf(SIGN_ERR, "  t = %d, num = %d, total = %d\n", t, num, newN);
  fprintf(SIGN_ERR, "  reps={%d", replicates[0]);
  for(int i = 1; i < t; i++) fprintf(SIGN_ERR, ", %d", replicates[i]);
  fprintf(SIGN_ERR, "}\n");
  fprintf(SIGN_ERR, "  pos={%d", pos[0]);
  for(int i = 1; i < t; i++) fprintf(SIGN_ERR, ", %d", pos[i]);
  fprintf(SIGN_ERR, "}\n");
  fflush(SIGN_ERR);
  
  int ii = 0;
  for(int i = 0; i < num; i++){
	int prev = random() % replicates[0] + pos[0];
	
	for(int k = 1; k < t; k++){
	  int next = random() % replicates[k] + pos[k];

	  int l = 0;
	  for(j = 0; j < pp; j++){
		/* If nodes == NULL, then 'jj' below becomes aloways equal to 'j' 
		   because '&j' is set to 'nodes' and l == 0 for all the time. */
		int jj = nodes[l];

		newX[ii + j * newN] = X[prev + jj * n];
		newY[ii + j * newN] = X[next + jj * n];

		l += inc;
	  }/* for j < pp */

	  ii++;
	  prev = next;
	}/* for k < t */
  }/* for i < num */
}
/* SGN_Data_pseudo_bootstrap2() ============================================= */

#include <limits.h>
/*
 *\param t   Number of time points (= length of \a replicates)
 *\param num Number of blocks to be resampled. 
 */
void SGN_Data_pseudo_bootstrap_RNG(const double * X, int n, int p,
								   const int * replicates, int t, int num,
								   const int * nodes, int pp,
								   double * newX, double * newY,
								   SGN_RNG * rng){
  int newN = num * (t - 1);
  int pos[t]; /* Position indices of the first replicate for each time point. */
  int inc = 1;
  int j;
  int rnd[num * t];

  if(nodes == NULL){ pp = p; inc = 0; nodes = &j; }

  SGN_RNG_integer(rng, rnd, num * t, 0, INT_MAX);

  /* Calculates the positions of the first replicate of all the time points. */
  pos[0] = 0;
  for(int i = 1; i < t; i++) pos[i] = pos[i - 1] + replicates[i - 1];
  
  int ii = 0;
  int rpos = 0;
  for(int i = 0; i < num; i++){
	int prev = rnd[rpos++] % replicates[0] + pos[0];
	
	for(int k = 1; k < t; k++){
	  int next = rnd[rpos++] % replicates[k] + pos[k];

	  int l = 0;
	  for(j = 0; j < pp; j++){
		/* If nodes == NULL, then 'jj' below becomes aloways equal to 'j' 
		   because '&j' is set to 'nodes' and l == 0 for all the time. */
		int jj = nodes[l];

		newX[ii + j * newN] = X[prev + jj * n];
		newY[ii + j * newN] = X[next + jj * n];

		l += inc;
	  }/* for j < pp */

	  ii++;
	  prev = next;
	}/* for k < t */
  }/* for i < num */
}
/* SGN_Data_pseudo_bootstrap_RNG() ========================================= */

void SGN_Data_pseudo_bootstrap2ushort(const double * X, int n, int p,
									 const int * replicates, int t, int num,
									 const unsigned short * nodes, int pp,
									  double * newX, double * newY,
									  SGN_RNG * rng){

  int newN = num * (t - 1);
  int pos[t]; /* Position indices of the first replicate for each time point. */
  int inc = 1;
  unsigned short j;
  int rnd[num * t];

  if(nodes == NULL){ pp = p; inc = 0; nodes = &j; }
  SGN_RNG_integer(rng, rnd, num * t, 0, INT_MAX);

  /* Calculates the positions of the first replicate of all the time points. */
  pos[0] = 0;
  for(int i = 1; i < t; i++) pos[i] = pos[i - 1] + replicates[i - 1];

  fprintf(SIGN_ERR, "PSEUDO BOOTSTRAP ========\n");
  fprintf(SIGN_ERR, "  t = %d, num = %d, total = %d\n", t, num, newN);
  fprintf(SIGN_ERR, "  reps={%d", replicates[0]);
  for(int i = 1; i < t; i++) fprintf(SIGN_ERR, ", %d", replicates[i]);
  fprintf(SIGN_ERR, "}\n");
  fprintf(SIGN_ERR, "  pos={%d", pos[0]);
  for(int i = 1; i < t; i++) fprintf(SIGN_ERR, ", %d", pos[i]);
  fprintf(SIGN_ERR, "}\n");
  fflush(SIGN_ERR);
  
  int ii = 0;
  int rpos = 0;
  for(int i = 0; i < num; i++){
	int prev = rnd[rpos++] % replicates[0] + pos[0];
	
	for(int k = 1; k < t; k++){
	  int next = rnd[rpos++] % replicates[k] + pos[k];

	  int l = 0;
	  for(j = 0; j < pp; j++){
		/* If nodes == NULL, then 'jj' below becomes aloways equal to 'j' 
		   because '&j' is set to 'nodes' and l == 0 for all the time. */
		int jj = nodes[l];

		newX[ii + j * newN] = X[prev + jj * n];
		newY[ii + j * newN] = X[next + jj * n];

		l += inc;
	  }/* for j < pp */

	  ii++;
	  prev = next;
	}/* for k < t */
  }/* for i < num */
}
/* SGN_Data_pseudo_bootstrap2() ============================================= */

/*!\brief Appends the specified data into the other.
 *
 *
 *\return the number of appended nodes, or -1 if an error occurred.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
int SGN_Data_append(SGN_Data * data1, const SGN_Data * data2){

  /* Check consistency */
  if(SGN_Data_samples(data1) != SGN_Data_samples(data2)){
	fprintf(SIGN_ERR, "WARNING: SGN_Data_append(): Num. of samples not consisitent.\n");
	fprintf(SIGN_ERR, "           %d vs %d.\n",
			SGN_Data_samples(data1), SGN_Data_samples(data2));
	return -1;
  }

  int num = 0;
  SGN_IntArray * indexes = SGN_IntArray_new();

  for(int j = 0; j < SGN_Data_nodes(data2); j++){
	const char * node = SGN_Data_get_name(data2, j);
	size_t idx = SGN_Data_find_name(data1, node);

	if(idx == SGN_StrArray_ERROR_VALUE){
	  SGN_IntArray_add(indexes, j);
	  SGN_StrArray_add_copy(data1->names, node);
	  SGN_StrArray_add_copy(data1->aliases, SGN_Data_get_alias(data2, j));
	  num++;
	}
  }/* for j < */

  int new_p = SGN_Data_nodes(data1) + num;
  int n = SGN_Data_samples(data1);
  double * new_X = malloc(sizeof(double) * n * new_p);
  double * new_Y = new_X;
  if(data1->X != data1->Y){
	new_Y = malloc(sizeof(double) * n * new_p);
  }

  int offset = n * data1->p;
  memcpy(new_X, data1->X, sizeof(double) * offset);
  for(int j = 0; j < num; j++){
	int idx = SGN_IntArray_get(indexes, j);
	memcpy(new_X + offset + n * j, data2->X + idx * n, sizeof(double) * n);
  }
  if(data1->X != data1->Y){
	memcpy(new_Y, data1->Y, offset);
	for(int j = 0; j < num; j++){
	  int idx = SGN_IntArray_get(indexes, j);
	  memcpy(new_Y + offset + n * j, data2->Y + idx * n, sizeof(double) * n);
	}
  }
  data1->p = new_p;
  free(data1->X);
  if(data1->X != data1->Y) free(data1->Y);
  data1->X = new_X;
  data1->Y = new_Y;
  
  SGN_IntArray_delete(indexes);
  return num;
}
/* SGN_Data_append() ======================================================== */

/*!\brief Adds the meta data.
 *
 * Meta data is a collection of pairs of keys and their values.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_add_meta_data(SGN_Data * data, const char * key, const char * value){
  if(data->meta == NULL)	data->meta = SGN_KeyValues_new();

  SGN_KeyValues_add_string(data->meta, key, value);
}
/* SGN_Data_add_meta_data() ================================================= */

void SGN_Data_set_meta_data_kv(SGN_Data * data, const SGN_KeyValues * kv){
  if(data->meta == NULL) data->meta = SGN_KeyValues_new();
  SGN_KeyValues_add_KeyValues(data->meta, kv);
}
/* SGN_Data_set_meta_data_kv() ============================================== */

/*!\brief Adds the attribute data.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_add_attributes(SGN_Data * data, const char * key, const SGN_StrArray * values){
  if(data->attr == NULL) data->attr = SGN_KeyValues_new();

  SGN_KeyValues_add(data->attr, key, SGN_Type_STR_ARRAY, SGN_StrArray_clone(values));
}
/* SGN_Data_add_attributes() ================================================ */

//const SGN_StrArray * SGN_Data_get_attributes(const SGN_Data * data, const char * key){
//  if(data->attr == NULL) return NULL;
//  return (SGN_StrArray *) SGN_KeyValues_get_with_type(data->attr,
//											  
//}
/* SGN_Data_get_attributes() ================================================ */


typedef struct {
  size_t index;
  SGN_IntArray * primary;
  SGN_IntArray * secondary;
} SGN_Data_sort_t;

int SGN_Data_sort_t_cmp(const void * a, const void * b){
  const SGN_Data_sort_t * aa = a;
  const SGN_Data_sort_t * bb = b;

  int r = SGN_IntArray_get(aa->primary, aa->index) - SGN_IntArray_get(bb->primary, bb->index);

  if(r != 0){
	return r;
  } else {
	return SGN_IntArray_get(aa->secondary, aa->index) - SGN_IntArray_get(bb->secondary, bb->index);
  }
}
/* SGN_Data_sort_t_cmp() ==================================================== */

/*!\brief Sort columns (samples) according to the primary key IDs.
 *
 * This does nothing if the primary key IDs are not available.
 * The primary Key ID is generally used to represent time points in EDF.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_sort_by_primary(SGN_Data * data){
  if(data->primaryID == NULL){
	fprintf(SIGN_ERR, "WARNING: primary key ID is not set.\n");
	return;
  }
  if(data->secondaryID == NULL){
	fprintf(SIGN_ERR, "WARNING: secondary key ID is not set.\n");
	return;
  }

  const int n = data->n;
  SGN_Data_sort_t ar[n];
  for(int i = 0; i < n; i++){
	ar[i].index = i;
	ar[i].primary = data->primaryID;
	ar[i].secondary = data->secondaryID;
  }
  qsort(ar, data->n, sizeof(SGN_Data_sort_t), SGN_Data_sort_t_cmp);

  double * new_X = malloc(sizeof(double) * n * data->p);
  SGN_IntArray * new_primaryID = SGN_IntArray_new();
  SGN_IntArray * new_secondaryID = SGN_IntArray_new();
  for(int i = 0; i < n; i++){
	int ii = ar[i].index;
#ifdef SIGN_DEBUG
	fprintf(SIGN_ERR, "SGN_Data: ii=%d, i=%d\n", ii, i);
#endif
	for(int j = 0; j < data->p; j++){
	  new_X[i + j * n] = data->X[ii + j * n];
	}
	SGN_IntArray_add(new_primaryID, SGN_IntArray_get(data->primaryID, ar[i].index));
	SGN_IntArray_add(new_secondaryID, SGN_IntArray_get(data->secondaryID, ar[i].index));
  }

  free(data->X);
  if(data->X != data->Y) free(data->Y);
  SGN_IntArray_delete(data->primaryID);
  SGN_IntArray_delete(data->secondaryID);

  data->X = new_X;
  data->Y = new_X;
  data->primaryID = new_primaryID;
  data->secondaryID = new_secondaryID;
}
/* SGN_Data_sort_by_primary() =============================================== */

/*!\brief Sort columns (samples) according to the secondary key IDs.
 *
 * This does nothing if the secondary key IDs are not available.
 * The primary Key ID is generally used to represent time points in EDF,
 * and the scondary Key ID represents replicate ID.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_sort_by_secondary(SGN_Data * data){
  if(data->primaryID == NULL){
	fprintf(SIGN_ERR, "WARNING: primary key ID is not set.\n");
	return;
  }
  if(data->secondaryID == NULL){
	fprintf(SIGN_ERR, "WARNING: secondary key ID is not set.\n");
	return;
  }

  const int n = data->n;
  SGN_Data_sort_t ar[n];
  for(int i = 0; i < n; i++){
	ar[i].index = i;
	ar[i].primary = data->secondaryID;
	ar[i].secondary = data->primaryID;
  }
  qsort(ar, data->n, sizeof(SGN_Data_sort_t), SGN_Data_sort_t_cmp);

  double * new_X = malloc(sizeof(double) * n * data->p);
  SGN_IntArray * new_primaryID = SGN_IntArray_new();
  SGN_IntArray * new_secondaryID = SGN_IntArray_new();
  for(int i = 0; i < n; i++){
	int ii = ar[i].index;
#ifdef SIGN_DEBUG
	fprintf(SIGN_ERR, "SGN_Data: ii=%d, i=%d\n", ii, i);
#endif
	for(int j = 0; j < data->p; j++){
	  new_X[i + j * n] = data->X[ii + j * n];
	}
	SGN_IntArray_add(new_primaryID, SGN_IntArray_get(data->primaryID, ar[i].index));
	SGN_IntArray_add(new_secondaryID, SGN_IntArray_get(data->secondaryID, ar[i].index));
  }

  free(data->X);
  if(data->X != data->Y) free(data->Y);
  SGN_IntArray_delete(data->primaryID);
  SGN_IntArray_delete(data->secondaryID);

  data->X = new_X;
  data->Y = new_X;
  data->primaryID = new_primaryID;
  data->secondaryID = new_secondaryID;
}
/* SGN_Data_sort_by_secondary() ============================================= */

/*!\brief Returns the array containing the numbers of replicates
 *  (secondary IDs) for all the time points (primary IDs).
 *
 * The maximum primary ID is assumed as the number of time points.
 * Thus, the size of the returned array is equal to the maximum primary
 * ID.  It is possible that the returned array has 0 as its element.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
SGN_IntArray * SGN_Data_get_replicates(const SGN_Data * data){
  if(data->primaryID == NULL || data->secondaryID == NULL){
	return NULL;
  }
  SGN_IntArray * reps = SGN_IntArray_new();
  for(int i = 0; i < data->n; i++){
	int id = SGN_IntArray_get(data->primaryID, i);
#ifdef SIGN_DEBUG
	fprintf(SIGN_ERR, "SGN_Data_get_replicates: i=%d, id=%d\n", i, id);
#endif
	if(id == 0) {
	  fprintf(SIGN_ERR, "ERROR: SGN_Data_get_replicates(): "
			  "Primary ID contains 0 at the %d-th value column.\n",
			  i + 1);
	  exit(1);
	}
	while(SGN_IntArray_size(reps) < id) SGN_IntArray_add(reps, 0);
	SGN_IntArray_inc(reps, id - 1);
  }
  return reps;
}
/* SGN_Data_get_replicates() ================================================ */

/*!\brief Returns the largest scondary (replicate) ID.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
int SGN_Data_max_replicates(const SGN_Data * data){
  if(data->secondaryID == NULL) return 1;

  int ret = -1;
  for(int i = 0; i < SGN_IntArray_size(data->secondaryID); i++){
	int v = SGN_IntArray_get(data->secondaryID, i);
	if(v > ret) ret = v;
  }
  return ret;
}
/* SGN_Data_max_replicates() ================================================ */

/*
  This generates the vector generally specified by '-Y' option from secondaryID.
 */
SGN_IntArray * SGN_Data_get_dynamic_points(const SGN_Data * data){
  if(data->primaryID == NULL || data->secondaryID == NULL){
	return NULL;
  }
  SGN_IntArray * points = SGN_IntArray_new();
  for(int i = 0; i < data->n; i++){
	int id = SGN_IntArray_get(data->secondaryID, i);
#ifdef SIGN_DEBUG
	fprintf(SIGN_ERR, "SGN_Data_get_replicates: i=%d, id=%d\n", i, id);
#endif
	while(SGN_IntArray_size(points) < id) SGN_IntArray_add(points, 0);
	//SGN_IntArray_set(points, id - 1, SGN_IntArray_get(points, id - 1) + 1);
	SGN_IntArray_inc(points, id - 1);
  }
  SGN_IntArray_remove(points, SGN_IntArray_size(points) - 1);
  SGN_IntArray_set(points, 0, SGN_IntArray_get(points, 0) + 1);
  int prev = SGN_IntArray_get(points, 0);
  for(int i = 1; i < SGN_IntArray_size(points); i++){
	int v = SGN_IntArray_get(points, i) + prev;
	SGN_IntArray_set(points, i, v);
	prev = v;
  }
	
  return points;
}
/* SGN_Data_get_dynamic_points() ============================================ */

/*!\brief Generates an array containing time points.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
SGN_IntArray * SGN_Data_get_timepoints(const SGN_Data * data){
  if(data->primaryID == NULL || data->secondaryID == NULL){
	return NULL;
  }
  SGN_IntArray * points = SGN_IntArray_new();
  for(int i = 0; i < data->n; i++){
	int id = SGN_IntArray_get(data->secondaryID, i);
	while(SGN_IntArray_size(points) < id) SGN_IntArray_add(points, 0);
	SGN_IntArray_inc(points, id - 1);
  }
  return points;
}
/* SGN_Data_get_timepoints() ================================================ */

/*!\brief Returns the list of unique secondary IDs.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
SGN_IntArray * SGN_Data_unique_secondaryID(const SGN_Data * data){
  if(data->primaryID == NULL || data->secondaryID == NULL){
	return NULL;
  }
  SGN_IntArray * IDs = SGN_IntArray_new();
  for(int i = 0; i < SGN_IntArray_size(data->secondaryID); i++){
	SGN_IntArray_add_nr(IDs, SGN_IntArray_get(data->secondaryID, i));
  }
  return IDs;
}
/* SGN_Data_unique_secondaryID() ============================================ */

int SGN_Data_max_primaryID(const SGN_Data * data){
  if(data->primaryID == NULL) return 1;
  int ret = -1;
  for(int i = 0; i < SGN_IntArray_size(data->primaryID); i++){
	int v = SGN_IntArray_get(data->primaryID, i);
	if(v > ret) ret = v;
  }
  return ret;
}
/* SGN_Data_max_primaryID() ================================================= */

int SGN_Data_max_secondaryID(const SGN_Data * data){
  if(data->secondaryID == NULL) return 1;
  int ret = -1;
  for(int i = 0; i < SGN_IntArray_size(data->secondaryID); i++){
	int v = SGN_IntArray_get(data->secondaryID, i);
	if(v > ret) ret = v;
  }
  return ret;
}
/* SGN_Data_max_secondaryID() =============================================== */

/*!\brief Performs mean shift for each node.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_node_mean_shift(SGN_Data * data){
  for(int j = 0; j < data->p; j++){
	/* Calculate the mean for a node */
	double v = 0.0;
	for(int i = 0; i < data->n; i++){
	  v += data->X[i + j * data->n];
	}
	v /= (double) data->n;

	/* Perform mean shift. */
	for(int i = 0; i < data->n; i++){
	  data->X[i + j * data->n] -= v;
	}
  }
}
/* SGN_Data_node_mean_shift() =============================================== */

#ifdef USE_MPI
/*!\brief Broadcasts the \ref SGN_Data instance by MPI.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Data */
void SGN_Data_MPI_Bcast(SGN_Data ** pds, int root, MPI_Comm comm){
  int rank;
  int flag = 0;
  SGN_Data * data;

  //fprintf(SIGN_ERR, "SGN_Data_MPI_Bcast(): called [root=%d].\n", root); fflush(SIGN_ERR);

  MPI_Comm_rank(comm, &rank);
  if(rank != root) *pds = SGN_Data_new();
  data = *pds;

  //fprintf(SIGN_ERR, "SGN_Data_MPI_Bcast(): rank = %d.\n", rank);

  /* Broadcast n, p, and type. */
  MPI_Bcast(&data->n, 1, MPI_INT, root, comm);
  MPI_Bcast(&data->p, 1, MPI_INT, root, comm);
  MPI_Bcast(&data->type, 1, MPI_INT, root, comm);

  //fprintf(SIGN_ERR, "SGN_Data_MPI_Bcast(): exchanged n, p, & type: n=%d, p=%d\n",
  //	  data->n, data->p);
  //fflush(SIGN_ERR);

  /* Broadcast data values in 'X'. */
  if(rank != root){
	//fprintf(SIGN_ERR, "SGN_Data_MPI_Bcast(): allocating X.\n");
	data->X = (double *) malloc(sizeof(double) * data->n * data->p);
  }
  MPI_Bcast(data->X, data->n * data->p, MPI_DOUBLE, root, comm);
  //fprintf(SIGN_ERR, "SGN_Data_MPI_Bcast(): X exchanged.\n");
  //fflush(SIGN_ERR);

  /* Check whether or not broadcast data values in 'Y'. */
  if(rank == root && data->X != data->Y) flag = 1;

  /* Broadcast whether or not 'Y' is different from 'X' ('flag') */
  MPI_Bcast(&flag, 1, MPI_INT, root, comm);
  if(rank != root){
	if(flag){
	  data->Y = (double *) malloc(sizeof(double) * data->n * data->p);
	} else {
	  data->Y = data->X;
	}
  }
  /* If 'flag' is true, broadcast data values in 'Y'. */
  if(flag){
	MPI_Bcast(&data->Y, data->n * data->p, MPI_DOUBLE, root, comm);
  }
  
  /* ------------------------------ */
  /* Broadcast gene names & aliases */
  /* ------------------------------ */
  SGN_StrArray_MPI_Bcast(&data->names, root, comm);
  SGN_StrArray_MPI_Bcast(&data->aliases, root, comm);
  //fprintf(SIGN_ERR, "SGN_Data_MPI_Bcast(): Names & Aliases exchanged.\n");
  //fflush(SIGN_ERR);

  /* -------------------------------------------- */
  /* Broadcast category names if type is discrete */
  /* -------------------------------------------- */
  if(data->type == SGN_Data_TYPE_CATEGORICAL){
	size_t size = (size_t) data->p;
	//fprintf(SIGN_ERR, "SGN_Data_MPI_Bcast(): categorical data\n"); fflush(SIGN_ERR);
	SGN_StrArray_ar_MPI_Bcast(&data->categories, &size, root, comm);
  }

  char msg;

  /* Broadcast primary ID */
  msg = (rank == root && data->primaryID != NULL) ? 1 : 0;
  MPI_Bcast(&msg, 1, MPI_CHAR, root, comm);
  if(msg){
	SGN_IntArray_MPI_Bcast(&data->primaryID, root, comm);
  }

  /* Broadcast secondary ID */
  msg = (rank == root && data->secondaryID != NULL) ? 1 : 0;
  MPI_Bcast(&msg, 1, MPI_CHAR, root, comm);
  if(msg){
	SGN_IntArray_MPI_Bcast(&data->secondaryID, root, comm);
  }
}
/* SGN_Data_MPI_Bcast() ----------------------------------------------------- */

#endif /* USE_MPI */

/* End of file.  Copyright (C) 2008-2011 SiGN Project Members */
