/*
  sign/util/SGN_Data.h : Data matrix container

  Copyright (C) 2008-2011 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#ifndef __SGN_DATA_H
#define __SGN_DATA_H

#include "SGN_StrArray.h"
#include "SGN_IntArray.h"
#include "SGN_KeyValues.h"
#include "../math/SGN_RNG.h"

#ifndef DOXY
typedef enum {
  SGN_Data_TYPE_NUMERICAL,
  SGN_Data_TYPE_CATEGORICAL,
} SGN_Data_TYPE_t;
#endif

/*!\brief Constant that represents categorical (discrete) values.
 *\memberof SGN_Data */
//#define SGN_Data_TYPE_CATEGORICAL  1

/*!\brief Data matrix container.
 *
 * This class is designed to handle a numerical data matrix.
 * Several operations on the data matrix are defined in this class.
 *
 * To construct a \ref SGN_Data instance from a file,
 * use \ref DataMatrix module or \ref EDF class.
 *
 * By default, the primary ID corresponds to time points, and
 * the secondary ID, to replicate IDs.
 *
 *\sa \ref DataMatrix, \ref EDF
 *
 *\author Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
 */
typedef struct {

  /*!\brief The number of samples. */
  int n;

  /*!\brief The number of nodes */
  int p;

  /*!\brief \ref n x \ref p input data matrix. */
  double * X; 

  /*!\brief \ref n x \ref p matrix containing the shifted (predicted) data. */
  double * Y; 

  SGN_StrArray * names;
  SGN_StrArray * aliases;

  /*!\brief \ref SGN_Data_TYPE_NUMERICAL or \ref SGN_Data_TYPE_CATEGORICAL. */
  int type;

  /*!\brief Categories for the samples (columns) if the value of
	SGN_Data::type is \ref SGN_Data_TYPE_CATEGORICAL.
	NULL can be set. This is the array of \ref SGN_StrArray. */
  SGN_StrArray * categories;

  SGN_StrArray * labels;

  SGN_KeyValues * meta;
  SGN_KeyValues * attr;

  SGN_IntArray * primaryID;
  SGN_IntArray * secondaryID;

#ifdef DOXY
  enum SGN_Data_TYPE_t {
	/*! Numerical data */
	SGN_Data_TYPE_NUMERICAL,
	/*! Categorical data */
	SGN_Data_TYPE_CATEGORICAL,
  };
#endif
  
} SGN_Data;

SGN_Data *    SGN_Data_new();
void         SGN_Data_delete(SGN_Data * dataset);
SGN_Data *    SGN_Data_clone(const SGN_Data * dataset);
int          SGN_Data_nodes(const SGN_Data * dataset);
int          SGN_Data_samples(const SGN_Data * dataset);
const char * SGN_Data_get_name(const SGN_Data * dataset, size_t index);
const char * SGN_Data_get_alias(const SGN_Data * dataset, size_t index);
double       SGN_Data_get_value(SGN_Data * dataset, int row, int col);
double       SGN_Data_get_value_Y(SGN_Data * dataset, int row, int col);
double       SGN_Data_value(SGN_Data * dataset, int node, int sample);
double       SGN_Data_value_X(SGN_Data * dataset, int node, int sample);
double       SGN_Data_value_Y(SGN_Data * dataset, int node, int sample);
size_t       SGN_Data_find_name(const SGN_Data * dataset, const char * name);
void         SGN_Data_select_nodes_files(SGN_Data * dataset, SGN_StrArray * files, int col);
SGN_IntArray *   SGN_Data_select_nodes_files2(SGN_Data * dataset, SGN_StrArray * files, int col);

void         SGN_Data_select_nodes_index(SGN_Data * dataset, SGN_IntArray * nodes);
SGN_Data *    SGN_Data_select_nodes_index_new(const SGN_Data * dataset,
											SGN_IntArray * nodes);
void         SGN_Data_select_samples(SGN_Data * dataset, SGN_IntArray * samples);
void         SGN_Data_random_sampling(SGN_Data * dataset);
void         SGN_Data_random_sampling_SGN_Data(SGN_Data * dataset, const SGN_Data * src);
void         SGN_Data_random_sampling_SGN_Data_RNG(SGN_Data * data, const SGN_Data * src, SGN_RNG * rng);
void         SGN_Data_random_sampling2(SGN_Data * dataset, int num);
void         SGN_Data_output(SGN_Data * dataset, FILE * fp);
void         SGN_Data_discretize(SGN_Data * dataset, int categories);
void         SGN_Data_dynamic(SGN_Data * dataset, SGN_IntArray * points);
void         SGN_Data_pseudo_bootstrap(SGN_Data * dataset, const SGN_IntArray * replicates, int num);
void         SGN_Data_pseudo_bootstrap2(const double * X, int n, int p, const int * replicates, int t, int num, const int * nodes, int pp, double * newX, double * newY);
void         SGN_Data_pseudo_bootstrap2ushort(const double * X, int n, int p, const int * replicates, int t, int num, const unsigned short * nodes, int pp, double * newX, double * newY, SGN_RNG * rng);
void         SGN_Data_pseudo_bootstrap_RNG(const double * X, int n, int p, const int * replicates, int t, int num, const int * nodes, int pp, double * newX, double * newY, SGN_RNG * rng);
int          SGN_Data_append(SGN_Data * dataset1, const SGN_Data * dataset2);
void         SGN_Data_add_meta_data(SGN_Data * dataset, const char * key, const char * value);
void         SGN_Data_set_meta_data_kv(SGN_Data * dataset, const SGN_KeyValues * kv);
void         SGN_Data_add_attributes(SGN_Data * dataset, const char * key, const SGN_StrArray * values);

void         SGN_Data_sort_by_primary(SGN_Data * dataset);
void         SGN_Data_sort_by_secondary(SGN_Data * dataset);
SGN_IntArray *   SGN_Data_get_replicates(const SGN_Data * dataset);
int          SGN_Data_max_replicates(const SGN_Data * dataset);
SGN_IntArray *   SGN_Data_get_dynamic_points(const SGN_Data * dataset);
SGN_IntArray *   SGN_Data_get_timepoints(const SGN_Data * dataset);
int          SGN_Data_max_primaryID(const SGN_Data * dataset);
int          SGN_Data_max_secondaryID(const SGN_Data * dataset);

SGN_IntArray *   SGN_Data_unique_secondaryID(const SGN_Data * dataset);

void SGN_Data_node_mean_shift(SGN_Data * dataset);
void SGN_Data_read_node_set_file(const SGN_Data * dataset, SGN_IntArray * nodes,
								const char * file, int col);


#ifdef USE_MPI

#include <mpi.h>
void SGN_Data_MPI_Bcast(SGN_Data ** pds, int root, MPI_Comm comm);
#endif /* USE_MPI */

#endif /* __SGN_DATA_H */

/* End of file.  Copyright (C) 2008-2011 SiGN Project Members */
