/*
  sign/util/SGN_DoubleArray.c : Expandable array for double values.

  Copyright (C) 2008 - 2010
    Yoshinori Tamada
      Laboratory of DNA Information Analysis, Human Genome Center
      The Institute of Medical Science, The Unversity of Tokyo

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at ims.u-tokyo.ac.jp>
*/

#ifndef __SIGN_SSM
#include "SGN_Memory.h" /* Do not include for SiGN-SSM */
#endif /* __SIGN_SSM */
#include "SGN_DoubleArray.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <stdint.h>

/*!\brief Generates a new SGN_DoubleArray instance.
 *
 * This generates the new SGN_DoubleArray instance and initialize it.
 * The generated instance should be deleted (freed) by calling
 * SGN_DoubleArray_delete() when it is no longer needed.
 *
 *\return Pointer to the generated SGN_DoubleArray instance, or NULL
 *   if failed to allocate the enough memory.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_DoubleArray */
SGN_DoubleArray * SGN_DoubleArray_new(){
  SGN_DoubleArray * ar = malloc(sizeof(SGN_DoubleArray));
  if(ar == NULL) return NULL;

  ar->ptr = NULL;
  ar->buff_size = 0;
  ar->size = 0;

  return ar;
}
/* SGN_DoubleArray_new() ======================================================= */

/*!\brief Generates a new SGN_DoubleArray instance.
 *
 * This generates the new SGN_DoubleArray instance and initialize it.
 * The generated instance should be deleted (freed) by calling
 * SGN_DoubleArray_delete() when it is no longer needed.
 *
 *\param size  Initial buffer size.
 *\return  Pointer to the newly generated SGN_DoubleArray instance, or NULL
 *   if failed to allocate the enough memory.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_DoubleArray */
SGN_DoubleArray * SGN_DoubleArray_news(size_t size){
  SGN_DoubleArray * this = SGN_DoubleArray_new();
  if(this == NULL) return NULL;
  if(size != 0){
	SGN_DoubleArray_expand(this, size);
  }
  return this;
}
/* SGN_DoubleArray_news() ====================================================== */

void SGN_DoubleArray_init(SGN_DoubleArray * ar, size_t size){
  ar->ptr = (double *) malloc(sizeof(double) * size);
  if(ar->ptr == NULL){
	free(ar);
	return;
  }

  ar->buff_size = size;
  ar->size = 0;
}
/* SGN_DoubleArray_init() ------------------------------------------------------ */

/*!\brief Deletes the \ref SGN_DoubleArray instance.
 *
 *\param this  Pointer to the \ref SGN_DoubleArray instance to be deleted.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_DoubleArray */
void SGN_DoubleArray_delete(SGN_DoubleArray * this){
  free(this->ptr);
  free(this);
}
/* ========================================================================= */

void SGN_DoubleArray_delete_v(void * this){
  free(((SGN_DoubleArray *) this)->ptr);
  free(this);
}
/* ========================================================================= */

/*!\brief Copies the \ref SGN_DoubleArray instance.
 *\param this   \ref SGN_DoubleArray instance
 *\return Duplicated \ref SGN_DoubleArray instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_DoubleArray */
SGN_DoubleArray * SGN_DoubleArray_copy(const SGN_DoubleArray * this){
  SGN_DoubleArray * da = SGN_DoubleArray_news(this->size);
  
  memcpy(da->ptr, this->ptr, sizeof(double) * this->size);
  da->size = this->size;
  return da;
}
/* SGN_DoubleArray_copy() ------------------------------------------------------ */

double * SGN_DoubleArray_ptr(SGN_DoubleArray * this){
  return this->ptr;
}
/* SGN_DoubleArray_ptr() ------------------------------------------------------- */

/*!\brief Generates a new \ref SGN_DoubleArray instance from a string.
 *
 * This generates a new \ref SGN_DoubleArray instance from a string expression.
 * The delimiter to split each element is a colon ``<TT>:</TT>''.
 *
 * Empty elements are not allowed.  If it is found, then NULL is returned.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_DoubleArray */
SGN_DoubleArray * SGN_DoubleArray_new_str(const char * str){
  int count = 0;
  const char * pos = str;
  SGN_DoubleArray * ret;
  char * r = NULL;
  double v;

  /* Count the number of elements first */
  while(1){
	pos = strchr(pos, ':');
	if(pos != NULL){
	  count++;
	  pos++;
	} else {
	  break;
	}
  }
  count++;
  ret = SGN_DoubleArray_news(count);

  pos = str;
  while(1){
	char * npos;

	npos = strchr(pos, ':');
	if(npos == NULL) break;

	v = strtod(pos, &r);

	if(r == pos){ /* error */
	  SGN_DoubleArray_delete(ret);
	  return NULL;
	}
	
	SGN_DoubleArray_add(ret, v);
	pos = npos + 1;
  }

  /* Add the last element */
  v = strtod(pos, &r);
  if(r == pos){
	SGN_DoubleArray_delete(ret);
	return NULL;
  }
  SGN_DoubleArray_add(ret, v);

  return ret;
}
/* SGN_DoubleArray_new_str() --------------------------------------------------- */

/*!\brief Generates multiple \ref SGN_DoubleArray instances.
 *
 *\return  pointer to the first \ref SGN_DoubleArray instances.
 *\memberof SGN_DoubleArray */
SGN_DoubleArray * SGN_DoubleArray_new_ar(size_t ar_size, size_t init_size){
  int i;
  SGN_DoubleArray * iar = (SGN_DoubleArray *) malloc(sizeof(SGN_DoubleArray) * ar_size);
  
  for(i = 0; i < ar_size; i++){
	SGN_DoubleArray_init(iar + i, init_size);
  }
  
  return iar;
}
/* SGN_DoubleArray_new_ar() ---------------------------------------------------- */

/*!\brief Expands the size of the buffer memory.
 *
 * The new buffer size will be the current size + length secified
 * by \a length argument.
 *
 * Generally this is called internally, and users do not
 * need to use this.
 *
 *\param this  pointer to the \ref SGN_DoubleArray instance whose buffer size
 *    is to be expanded.
 *\param length  buffer size to be newly allocated.
 *\return \a this, or NULL if failed.
 *\memberof SGN_DoubleArray */
SGN_DoubleArray * SGN_DoubleArray_expand(SGN_DoubleArray * this, size_t length){
  double * new_ptr;
  size_t new_size;
  if(length == 0) length = 1;
  new_size = this->buff_size + length;
  new_ptr = (double *) realloc(this->ptr, sizeof(double) * new_size);
  if(new_ptr != NULL){
	this->ptr = new_ptr;
	this->buff_size = new_size;
	return this;
  } else {
	return NULL;
  }
}
/* SGN_DoubleArray_expand() ---------------------------------------------------- */

/*!\brief Removes all elements in the array.
 *
 * This does not free allocated memory.
 *
 *\param this  pointer to the \ref DoulbeArray instance.
 *\return \a this
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_DoubleArray */
SGN_DoubleArray * SGN_DoubleArray_clear(SGN_DoubleArray * this){
  this->size = 0;
  return this;
}
/* SGN_DoubleArray_clear() ----------------------------------------------------- */

/*!\brief Adds an integer value to the array.
 *
 *\param this  pointer to the \ref SGN_DoubleArray instance.
 *\param value  integer value to be added in the array.
 *\return \a this, or NULL if failed to expand the buffer memory.
 *\memberof Intarray */
SGN_DoubleArray * SGN_DoubleArray_add(SGN_DoubleArray * this, double value){
  if(this->buff_size == this->size){
	SGN_DoubleArray * ar = SGN_DoubleArray_expand(this, this->size);
	if(ar == NULL) return NULL;
  }

  this->ptr[this->size++] = value;
  return this;
}
/* SGN_DoubleArray_add() ------------------------------------------------------- */

/*!\brief Sets the value into the specified position of the array.
 *
 *\param this  pointer to the \ref SGN_DoubleArray instance.
 *\param index  position at which the value is set.
 *\param value   value to be set in the array.
 *\return  \a this itself.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_DoubleArray */
SGN_DoubleArray * SGN_DoubleArray_set(SGN_DoubleArray * this, size_t index, double value){

  while(this->buff_size <= index){
	SGN_DoubleArray * ar = SGN_DoubleArray_expand(this, this->buff_size);
	if(ar == NULL) return NULL;
  }

  if(this->size <= index) this->size = index + 1;

  this->ptr[index] = value;
  return this;
}
/* SGN_IntArray_set() ---------------------------------------------------------- */

/*!\brief Returns the number of values stored in the array.
 *\param this  pointer to the \ref SGN_DoubleArray instance.
 *\return the number of integer values stored in the array.
 *\memberof SGN_DoubleArray */
size_t SGN_DoubleArray_size(const SGN_DoubleArray * this){
  return this->size;
}
/* SGN_DoubleArray_size() ------------------------------------------------------ */

/*!\brief Returns the double value stored at the specified
 *   position in the array.
 *
 *\param this  pointer to the \ref SGN_DoubleArray instance.
 *\param index  position of the value to return.
 *\return double value stored at the specified position.
 *\memberof SGN_DoubleArray */
double SGN_DoubleArray_get(const SGN_DoubleArray * this, size_t index){
  return this->ptr[index];
}
/* SGN_DoubleArray_get() ------------------------------------------------------- */

/*!\brief Sets values by a string.
 *
 * This sets values represented by a string expression.
 * The delimiter to split each value is a colon ``<TT>:</TT>''.
 *
 * Empty elements are not allowed.  If it is found, then \r this
 * is cleared. 
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_DoubleArray */
void SGN_DoubleArray_set_str(SGN_DoubleArray * this, const char * str){
  const char * pos = str;
  char * r = NULL;
  double v;

  SGN_DoubleArray_clear(this);

  while(1){
	char * npos;

	npos = strchr(pos, ':');
	if(npos == NULL) break;

	v = strtod(pos, &r);

	if(r == pos){ /* error */
	  SGN_DoubleArray_clear(this);
	  return;
	}
	
	SGN_DoubleArray_add(this, v);
	pos = npos + 1;
  }

  /* Add the last element */
  v = strtod(pos, &r);
  if(r == pos){
	SGN_DoubleArray_clear(this);
	return;
  }
  SGN_DoubleArray_add(this, v);
}
/* SGN_DoubleArray_set_str() --------------------------------------------------- */

/*!\brief Returns the index at which the given value is stored.
 *
 * This requires at most O(n) time where n is the number of values
 * stored in the array.
 *
 *\param this  pointer to the \ref SGN_DoubleArray instance.
 *\param value  double value to find.
 *\return index at which the given value is stored, or INT_ARRAY_ERROR_VALUE
 *   if failed to find.
 *\memberof SGN_DoubleArray */
size_t SGN_DoubleArray_find(const SGN_DoubleArray * this, double value){
  size_t i;
  for(i = 0; i < this->size; i++){
	if(this->ptr[i] == value) return i;
  }
  return SGN_DoubleArray_ERROR_VALUE;
}
/* SGN_DoubleArray_find() ------------------------------------------------------ */

/*!\brief Returns the minimum value in the array.
 *
 *\param this  pointer to the \ref SGN_DoubleArray instance.
 *\return minimum value in the array, or 0 if the array is empty.
 *  
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_DoubleArray */
double SGN_DoubleArray_get_min(const SGN_DoubleArray * this){
  /* This can be implemented using optimized libraries */

  size_t i;
  double v = 0.0;

  if(this->size != 0){
	v = this->ptr[0];
	for(i = 1; i < this->size; i++){
	  if(this->ptr[i] < v) v = this->ptr[i];
	}
  }
  
  return v;
}
/* SGN_DoubleArray_get_min() --------------------------------------------------- */

/*!\brief Reads a double value in the array and copy it to the specified
 *   position of memory.
 *
 * This simply copies a value at the specified position of the array
 * to the specified position of the memory space.
 *
 *\param this SGN_DoubleArray instance.
 *\param index Position of the value to read.
 *\param mem Pointer to which the value is written.
 *\return Pointer pointing to the next position of \a mem after writing the value.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_DoubleArray */
void * SGN_DoubleArray_read(const SGN_DoubleArray * this, size_t index, void * mem){
  memcpy(mem, this->ptr + index, sizeof(double));
  return (uint8_t *) mem + sizeof(mem);
}
/* SGN_DoubleArray_read() ====================================================== */

/*!\brief Prints elements in the array.
 *
 *\param da  \SGN_DoubleArray instance.
 *\param fp  file pointer to output.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_DoubleArray */
void SGN_DoubleArray_fprint(const SGN_DoubleArray * da, FILE * fp){
  const SGN_DoubleArray * this = (const SGN_DoubleArray *) da;
  for(size_t i = 0; i < SGN_DoubleArray_size(this); i++){
	if(i != 0) fprintf(fp, ", ");
	fprintf(fp, "%f", SGN_DoubleArray_get(this, i));
  }
}
/* SGN_DoubleArray_fprint() ================================================ */

void SGN_DoubleArray_fprint_v(const void * da, FILE * fp){
  const SGN_DoubleArray * this = (const SGN_DoubleArray *) da;
  for(size_t i = 0; i < SGN_DoubleArray_size(this); i++){
	if(i != 0) fprintf(fp, ":");
	fprintf(fp, "%f", SGN_DoubleArray_get(this, i));
  }
}

/* End of file.  Copyright (C) 2008  HGC, IMS, The University of Tokyo       */
