/*
  util/SGN_Edge.c : Module for representing an edge in a network.

  Copyright (C) 2008-2009 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.
  
  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/


#include "SGN_Edge.h"
#include "SGN_Memory.h"
#include "../lang/SGN_Obj.h"

/*!\brief Generates the new \ref SGN_Edge instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Edge */
SGN_Edge * SGN_Edge_new(SGN_Node * parent, SGN_Node * child){
  SGN_Edge * this = (SGN_Edge *) malloc(sizeof(SGN_Edge));

  this->parent = parent;
  this->child = child;
  //this->attr = SGN_KeyValues_new();
  this->csmlobj = CSMLObject_new();
  CSMLObject_set_type(this->csmlobj, CSMLObject_TYPE_EDGE);
  CSMLObject_set_line_width(this->csmlobj, SGN_Edge_DEFAULT_LINE_WIDTH);

  return this;
}
/* SGN_Edge_new() ============================================================== */

/*!\brief Deletes the \ref SGN_Edge instance.
 *
 * Note that this does not delete the nodes linked to this edge.
 * This just releases the resources occupied by this instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Edge */
void SGN_Edge_delete(SGN_Edge * this){
  //SGN_KeyValues_delete(this->attr);
  CSMLObject_delete(this->csmlobj);
  free(this);
}
/* SGN_Edge_delete() ----------------------------------------------------------- */

/*!\brief Returns the clone of the given \ref SGN_Edge instance.
 *
 * Note that this does not clone \ref SGN_Node instances linked to this
 * edge as a parent and a child.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Edge */
SGN_Edge * SGN_Edge_clone(const SGN_Edge * this){
  SGN_Edge * edge = SGN_Edge_new(this->parent, this->child);

  //edge->attr = SGN_KeyValues_clone(this->attr);
  edge->csmlobj = CSMLObject_clone(this->csmlobj);

  return edge;
}
/* SGN_Edge_clone() ============================================================ */

/*!
 *\memberof SGN_Edge */
SGN_Node * SGN_Edge_get_parent(const SGN_Edge * this){
  return this->parent;
}

/*!
 *
 *\param this
 *\param parent
 *\memberof SGN_Edge */
void SGN_Edge_set_parent(SGN_Edge * this, SGN_Node * parent){
  this->parent = parent;
}

void SGN_Edge_set_child(SGN_Edge * this, SGN_Node * child){
  this->child = child;
}

/*!\brief Sets parent and child nodes to this edge.
 *
 *\param this
 *\param parent
 *\param child
 *\memberof SGN_Edge */
void SGN_Edge_set(SGN_Edge * this, SGN_Node * parent, SGN_Node * child){
  this->parent = parent;
  this->child = child;
}
/* SGN_Edge_set() ========================================================== */

/*!\ Returns
 *
 * Same as SGN_Edge_get_parent().
 *\ingropu SGN_Edge */
SGN_Node * SGN_Edge_parent(const SGN_Edge * this){
  return this->parent;
}

/*!\brief Returns the child node of the edge.
 *
 *\memberof SGN_Edge */
SGN_Node * SGN_Edge_get_child(const SGN_Edge * this){
  return this->child;
}

/*!\brief Returns the child node of the edge.
 *
 * Same as SGN_Edge_get_child().
 *\memberof SGN_Edge */
SGN_Node * SGN_Edge_child(const SGN_Edge * this){
  return this->child;
}

/*!\brief Returns the name of the parent node.
 *\memberof SGN_Edge */
const char * SGN_Edge_parent_name(const SGN_Edge * this){
  return SGN_Node_name(this->parent);
}

/*!\brief Returns the name of the child node.
 *\memberof SGN_Edge */
const char * SGN_Edge_child_name(const SGN_Edge * this){
  return SGN_Node_name(this->child);
}

/*!\brief Adds an edge property (attribute).
 *
 *
 *
 *\memberof SGN_Edge */
void SGN_Edge_add_property(SGN_Edge * this, const char * key, SGN_Type type, void * value){
  SGN_Object * obj = SGN_Object_new();
  SGN_Object_set(obj, value, type);

  CSMLObject_add_property(this->csmlobj, key, obj);
  
  //SGN_KeyValues_add(this->attr, key, type, value);
  //CSMLObject_add
  //fprintf(SIGN_ERR, "Fatal Error: SGN_Edge_add_property() not implemented.\n");
  //fflush(SIGN_ERR);
}
/* SGN_Edge_add_property() ----------------------------------------------------- */

/*!\brief Add an edge property.
 *
 *\memberof SGN_Edge */
void SGN_Edge_add_property_Object(SGN_Edge * this, const char * key, SGN_Object * obj){
  CSMLObject_add_property(this->csmlobj, key, obj);
}
/* SGN_Edge_add_property_Object() ============================================== */

/*!\brief
 *\memberof SGN_Edge */
void SGN_Edge_add_property_int(SGN_Edge * this, const char * key, int value){
  //SGN_KeyValues_put_int(this->attr, key, value);
  CSMLObject_add_property_int(this->csmlobj, key, value);
}
/* SGN_Edge_add_property_int() ------------------------------------------------- */

/*!\brief
 *\memberof SGN_Edge */
void SGN_Edge_add_property_double(SGN_Edge * this, const char * key, double value){
  //SGN_KeyValues_put_double(this->attr, key, value);
  CSMLObject_add_property_double(this->csmlobj, key, value);
}
/* SGN_Edge_add_property_double() ---------------------------------------------- */

void SGN_Edge_set_property_double(SGN_Edge * this, const char * key, double value){
  CSMLObject_set_property_double(this->csmlobj, key, value);
}
/* SGN_Edge_set_property_double() ========================================== */

/*!\brief Adds a string edge property.
 *
 * The specified string is copied by this function.
 *
 *\param this
 *\param key
 *\param value  String property added to this edge.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Edge */
void SGN_Edge_add_property_string(SGN_Edge * this, const char * key, const char * value){
  //SGN_KeyValues_add_string(this->attr, key, value);
  CSMLObject_add_property_string(this->csmlobj, key, value);
}
/* SGN_Edge_add_property_string() ============================================== */

/*!\brief
 *\memberof SGN_Edge */
void SGN_Edge_add_properties(SGN_Edge * this, const SGN_KeyValues * properties){
  CSMLObject_append_properties_KeyValues(this->csmlobj, properties);
  //SGN_KeyValues_add_KeyValues(this->attr, properties);
}
/* SGN_Edge_add_properties() --------------------------------------------------- */

/*!\brief Returns the value of the edge property for the specified key.
 *\memberof SGN_Edge */
/*
void * SGN_Edge_get_property(const SGN_Edge * this, const char * key){
  if(this == NULL || key == NULL) return NULL;

  return SGN_KeyValues_get(this->attr, key);
}
*/
/* SGN_Edge_get_property() ----------------------------------------------------- */

/*!
 *\memberof SGN_Edge */
void * SGN_Edge_get_property_value(const SGN_Edge * this, size_t index){
  SGN_Object * obj = CSMLObject_get_property_value(this->csmlobj, index);
  if(obj == NULL){
	return NULL;
  } else {
	return SGN_Object_ptr(obj);
  }
}
/* SGN_Edge_get_property_value() ----------------------------------------------- */

/*!
 *\memberof SGN_Edge */
void * SGN_Edge_get_property_with_type(const SGN_Edge * this, 
								   const char * key, SGN_Type type){
  const SGN_Object * obj = CSMLObject_get_property(this->csmlobj, key);
  if(obj != NULL && SGN_Object_type(obj) == type){
	return SGN_Object_ptr(obj);
  } else {
	return NULL;
  }
  //return SGN_KeyValues_get_with_type(this->attr, key, type);
}
/* SGN_Edge_get_property_type() ------------------------------------------------ */

/*!\brief Returns the number of propeties (attributes) in the edge.
 *
 *\param this  pointer to the \ref SGN_Edge instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Edge */
size_t SGN_Edge_get_property_size(const SGN_Edge * this){
  //return SGN_KeyValues_size(this->attr);
  return CSMLObject_num_properties(this->csmlobj);
}
/* SGN_Edge_get_property_size() ------------------------------------------------ */

/*!\brief Returns the key of the properties at the specified position.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Edge */
const char * SGN_Edge_get_property_key(const SGN_Edge * this, size_t index){
  return CSMLObject_get_property_key(this->csmlobj, index);
  //return SGN_KeyValues_get_key(this->attr, index);
}
/* SGN_Edge_get_property_key() ------------------------------------------------- */

/*!\brief Return the type of the properties at the specified position.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Edge */
SGN_Type SGN_Edge_get_property_type(const SGN_Edge * this, size_t index){
  return CSMLObject_get_property_type(this->csmlobj, index);
}
/* SGN_Edge_get_property_type() ------------------------------------------------ */

/*!
 *\memberof SGN_Edge */
double SGN_Edge_get_property_double(const SGN_Edge * this, const char * key, double def_value){
  const SGN_Object * obj = CSMLObject_get_property(this->csmlobj, key);
  
  if(SGN_Object_type(obj) == SGN_Type_DOUBLE){
	return SGN_Object_double_value(obj, 0.0);
  } else {
	return def_value;
  }
}
/* SGN_Edge_get_property_double() ---------------------------------------------- */

/*!
 *\memberof SGN_Edge */
const char * SGN_Edge_get_property_string(const SGN_Edge * this, const char * key){
  const SGN_Object * obj = CSMLObject_get_property(this->csmlobj, key);
  if(SGN_Object_type(obj) == SGN_Type_STRING){
	return SGN_Object_string_value(obj, NULL);
  } else {
	return NULL;
  }
}
/* SGN_Edge_get_property_string() ========================================== */

/*
int SGN_Edge_get_property_int(const SGN_Edge * this, const char * key){
  return SGN_KeyValues_get_int(this->attr, key);
}
*/
/* SGN_Edge_get_property_int() ---------------------------------------------- */

/*!\brief
 *\return index corresponding to the key, or \t SGN_Edge_ERROR_VALUE if not 
 *    found.
 *\memberof SGN_Edge */
size_t SGN_Edge_find_property_key(const SGN_Edge * this, const char * key){
  size_t index = CSMLObject_find_property(this->csmlobj, key);
  if(index == CSMLObject_ERROR_VALUE) return SGN_Edge_ERROR_VALUE;
  return index;
}
/* SGN_Edge_find_property_key() ------------------------------------------------ */

size_t SGN_Edge_find_property(const SGN_Edge * this, const char * key){
  return CSMLObject_find_property(this->csmlobj, key);
}
/* SGN_Edge_find_property() ==================================================== */

int SGN_Edge_get_property_int_index(const SGN_Edge * this, const size_t index){
  /*
  void * obj;
  if(SGN_KeyValues_get_type(this->attr, index) != SGN_Type_INT){
	return 0;
  }
  obj = SGN_KeyValues_get_value(this->attr, index);
  if(obj == NULL) return 0;
  return *((int *) obj);
  */
  return CSMLObject_get_property_int_at(this->csmlobj, index, -1);
}
/* SGN_Edge_get_property_int_index() ------------------------------------------- */

/*!\brief Removes a property at the specified position.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Edge */
void SGN_Edge_remove_property_at(SGN_Edge * this, size_t index){
  CSMLObject_remove_property_at(this->csmlobj, index);
}
/* SGN_Edge_remove_property_at() =============================================== */

/*!\brief Removes all the properties of the edge.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Edge */
void SGN_Edge_clear_properties(SGN_Edge * this){
  //SGN_KeyValues_clear(this->attr);
  CSMLObject_clear_properties(this->csmlobj);
}
/* SGN_Edge_clear_properties() ------------------------------------------------- */

void SGN_Edge_print_properties(SGN_Edge * this, FILE * fp){
  //SGN_KeyValues_print(this->attr, fp);
  CSMLObject_fprint_properties(this->csmlobj, fp);
}
/* SGN_Edge_print_properties() ================================================= */

/*!\brief Prints the attribute value bound to the given key to the stream.
 *
 *\param this   pointer to the \ref SGN_Edge instance.
 *\param key    key of the value.
 *\param fp     output stream.
 *\return 1 if succeeded, or 0 if failed.
 *\memberof SGN_Edge */
int SGN_Edge_print_property_value(SGN_Edge * this, const char * key, FILE * fp){

  return CSMLObject_fprint_property_value(this->csmlobj, key, fp);

  /*
  size_t pos = SGN_KeyValues_find_key(this->attr, key);
  if(pos != SGN_KeyValues_ERROR_VALUE){
	void * value = SGN_KeyValues_get_value(this->attr, pos);
	int type = SGN_KeyValues_get_type(this->attr, pos);
	SGN_Types_fprint_value(type, value, fp);
	return 1;
  } else {
	return 0;
  }
  */
}
/* SGN_Edge_print_property_value() --------------------------------------------- */

/*!\brief Returns the \ref CSMLObject instance used in this \ref SGN_Edge
 * instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Edge */
CSMLObject * SGN_Edge_get_CSMLObject(const SGN_Edge * this){
  return this->csmlobj;
}
/* SGN_Edge_get_CSMLObject() =================================================== */


/*!\brief Sets the edge color.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Edge */
void SGN_Edge_set_color(SGN_Edge * this, int r, int g, int b, int a){
  CSMLObject_set_line_color_RGB(this->csmlobj, r, g, b, a);
}
/* SGN_Edge_set_color() ======================================================== */

void SGN_Edge_get_color_expr(const SGN_Edge * this, char * buff){
  CSMLObject_get_line_color_expr(this->csmlobj, buff);
}
/* SGN_Node_get_color_expr() --------------------------------------------------- */

/*!\brief Sets the line width.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Edge */
void SGN_Edge_set_line_width(SGN_Edge * this, double width){
  CSMLObject_set_line_width(this->csmlobj, width);
}
/* SGN_Node_set_line_width() =================================================== */

/*!\breif Returns the line width.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Edge */
double SGN_Edge_get_line_width(const SGN_Edge * this){
  return CSMLObject_get_line_width(this->csmlobj);
}
/* SGN_Edge_get_line_width() =================================================== */

/*!\brief Sets the depth of the \ref SGN_Edge.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Edge */
void SGN_Edge_set_depth(SGN_Edge * this, int depth){
  CSMLObject_set_depth(this->csmlobj, depth);
}
/* SGN_Edge_set_depth() ======================================================== */

/*!\brief Returns the depth of the \ref SGN_Edge.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Edge */
int SGN_Edge_get_depth(const SGN_Edge * this){
	return CSMLObject_get_depth(this->csmlobj);
}
/* SGN_Edge_get_depth() ======================================================== */

/* End of file.  Copyright (C) 2008-2009 SiGN Project Members */
