/*
  util/SGN_Edge.h : Module for representing an edge in a network.

  Copyright (C) 2008-2011  SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#ifndef __SGN_EDGE_H
#define __SGN_EDGE_H

#include "SGN_KeyValues.h"
#include "SGN_Node.h"
#include "../csml/CSMLObject.h"

#define SGN_Edge_ERROR_VALUE (-1)

/*!\brief Edge connecting nodes in a network.
 *
 * <H2>List of edge property keys</H2>
 *
 *
 * <DL>
 * <DT><TT>gain</TT> (double)</DT>
 * <DD>Edge gain.</DD>
 *
 * <DT><TT>up/down</TT> (string)</DT>
 * <DD>This takes one of &quot;<TT>up</TT>&auot;, &quot;<TT>down</TT>&quot;, 
 *     and &quot;<TT>unknown</TT>&quot;,
 *     and represents the estimated effect of edges whether it is up-regulation
 *     or down-regulation.<BR>
 *     Available for \ref BNRC score function.
 *  </DD>
 *
 * <DT><TT>EdgeType</TT> (int)</DT>
 * <DD>This is the integer version of &quot;<TT>up/down</TT>&quot; property.
 *     0, 1, and 2 are assigned to &quot;<TT>up</TT>&quot;, &quot;<TT>down</TT>&quot;,
 *     and &quot;<TT>unknown</TT>&quot;, respectively.</DD>
 *
 * <DT><TT>bspline</TT> (double array)</DT>
 * <DD><I>B</I>-spline coefficients (gamma) of the edge.</DD>
 *
 * <DT><TT>BS.Prob</TT> (double)</DT>
 * <DD></DD>
 *
 * <DT><TT>BS.Direction</TT> (double)</DT>
 * <DD></DD>
 *
 * </DL>
 *
 *\see SGN_Network, SGN_Node
 *\author Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
 */
typedef struct {
#ifdef DOXY
private:
#endif
  SGN_Node * parent;
  SGN_Node * child;

  CSMLObject * csmlobj;

} SGN_Edge;

#define SGN_Edge_DEFAULT_LINE_WIDTH (1.0)

SGN_Edge * SGN_Edge_new(SGN_Node * parent, SGN_Node * child);
void   SGN_Edge_delete(SGN_Edge * this);
SGN_Edge * SGN_Edge_clone(const SGN_Edge * this);
SGN_Node * SGN_Edge_get_parent(const SGN_Edge * this);
SGN_Node * SGN_Edge_parent(const SGN_Edge * this);
SGN_Node * SGN_Edge_get_child(const SGN_Edge * this);
SGN_Node * SGN_Edge_child(const SGN_Edge * this);
void   SGN_Edge_set_parent(SGN_Edge * this, SGN_Node * parent);
void   SGN_Edge_set_child(SGN_Edge * this, SGN_Node * child);
const char * SGN_Edge_parent_name(const SGN_Edge * this);
const char * SGN_Edge_child_name(const SGN_Edge * this);
void SGN_Edge_set(SGN_Edge * this, SGN_Node * parent, SGN_Node * child);

size_t SGN_Edge_find_property(const SGN_Edge * this, const char * key);
size_t SGN_Edge_find_property_key(const SGN_Edge * this, const char * key);
void   SGN_Edge_add_property(SGN_Edge * this, const char * key, SGN_Type type, void * value);
void   SGN_Edge_add_property_Object(SGN_Edge * this, const char * key, SGN_Object * obj);
void * SGN_Edge_get_property_with_type(const SGN_Edge * this, const char * key, SGN_Type type);
void   SGN_Edge_add_property_int(SGN_Edge * this, const char * key, int value);
void   SGN_Edge_add_property_double(SGN_Edge * this, const char * key, double value);
void   SGN_Edge_set_property_double(SGN_Edge * this, const char * key, double value);
void   SGN_Edge_add_property_string(SGN_Edge * this, const char * key, const char * value);
void   SGN_Edge_add_properties(SGN_Edge * this, const SGN_KeyValues * properties);
void * SGN_Edge_get_property_value(const SGN_Edge * this, size_t index);
const char * SGN_Edge_get_property_key(const SGN_Edge * this, size_t index);
size_t SGN_Edge_get_property_size(const SGN_Edge * this);
SGN_Type SGN_Edge_get_property_type(const SGN_Edge * this, size_t index);
double       SGN_Edge_get_property_double(const SGN_Edge * this, const char * key, double def_value);
int          SGN_Edge_get_property_int_index(const SGN_Edge * this, const size_t index);
const char * SGN_Edge_get_property_string(const SGN_Edge * this, const char * key);

void   SGN_Edge_clear_properties(SGN_Edge * this);
void   SGN_Edge_print_properties(SGN_Edge * this, FILE * fp);
int SGN_Edge_print_property_value(SGN_Edge * this, const char * key, FILE * fp);
void SGN_Edge_remove_property_at(SGN_Edge * this, size_t index);

CSMLObject * SGN_Edge_get_CSMLObject(const SGN_Edge * this);
void         SGN_Edge_set_color(SGN_Edge * this, int r, int g, int b, int a);
void         SGN_Edge_get_color_expr(const SGN_Edge * this, char * buff);
void         SGN_Edge_set_line_width(SGN_Edge * this, double width);
double       SGN_Edge_get_line_width(const SGN_Edge * this);
void         SGN_Edge_set_depth(SGN_Edge * this, int depth);
int          SGN_Edge_get_depth(const SGN_Edge * this);

#endif /* __SGN_EDGE_H */

/* End of file.  Copyright (C) 2008-2011  HGC, IMS, The University of Tokyo       */
