/*
  util/SGN_FloatArray.c : Expandable array for float values.

  Copyright (C) 2009 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.
  
  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

/*!\defgroup SGN_FloatArray
 *
 *\sa \ref DoubleArray, \ref IntArray
 *\author Yoshinori Tamada &lt;tamada at ims.u-tokyo.ac.jp&gt;
 */

#include "../signcom.h"
#include "SGN_FloatArray.h"

#include <stdlib.h>
#include <string.h>

/*!\brief Generates a new \ref SGN_FloatArray instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_FloatArray */
SGN_FloatArray * SGN_FloatArray_new(){
  SGN_FloatArray * ar = malloc(sizeof(SGN_FloatArray));
  ar->ptr = NULL;
  ar->size = 0;
  ar->buff_size = 0;

  return ar;
}
/* SGN_FloatArray_new() -------------------------------------------------------- */

void SGN_FloatArray_delete(void * this){
  free(((SGN_FloatArray *) this)->ptr);
  free(this);
}
/* SGN_FloatArray_delete() ----------------------------------------------------- */

SGN_FloatArray * SGN_FloatArray_expand(SGN_FloatArray * this, size_t length){
  void * new_ptr;
  size_t new_size = this->buff_size + length;
  //fprintf(SIGN_ERR, "SGN_FloatArray_expand(length=%zu)\n", length); fflush(SIGN_ERR);
  //fprintf(SIGN_ERR, "SGN_FloatArray_expand() buff_size==%zu\n", this->buff_size); fflush(SIGN_ERR);
  if(new_size == 0) new_size = 5;
  new_ptr = realloc(this->ptr, sizeof(float) * new_size);
  if(new_ptr != NULL){
	this->ptr = new_ptr;
	this->buff_size = new_size;
	return this;
  } else {
	return NULL;
  }
}
/* SGN_FloatArray_expand() ----------------------------------------------------- */

SGN_FloatArray * SGN_FloatArray_add(SGN_FloatArray * this, float value){
  if(this->buff_size == this->size){
	SGN_FloatArray * ar = SGN_FloatArray_expand(this, this->size);
	if(ar == NULL) return NULL;
  }

  this->ptr[this->size++] = value;
  return this;
}
/* SGN_FloatArray_add() -------------------------------------------------------- */
		  
float SGN_FloatArray_get(const SGN_FloatArray * this, size_t index){
  return this->ptr[index];
}
/* SGN_FloatArray_get() -------------------------------------------------------- */

SGN_FloatArray * SGN_FloatArray_set(SGN_FloatArray * this, size_t index, float value){
  if(this->buff_size <= index){
	size_t new_size = this->buff_size * 2;
	SGN_FloatArray * ar;
	if(new_size == 0) new_size = 1;
	while(new_size <= index) new_size += new_size;
	ar = SGN_FloatArray_expand(this, new_size - this->buff_size);
	if(ar == NULL) return NULL;
  }

  if(this->size <= index) this->size = index + 1;
  this->ptr[index] = value;

  return this;
}
/* SGN_FloatArray_set() -------------------------------------------------------- */

SGN_FloatArray * SGN_FloatArray_clear(SGN_FloatArray * this){
  this->size = 0;
  return this;
}
/* SGN_FloatArray_clear() ------------------------------------------------------ */

SGN_FloatArray * SGN_FloatArray_set_size(SGN_FloatArray * this, size_t size){
  if(this->buff_size < size){
	SGN_FloatArray_expand(this, size - this->buff_size);
  }
  this->size = size;
  return this;
}
/* SGN_FloatArray_set_size() --------------------------------------------------- */

float * SGN_FloatArray_ptr(SGN_FloatArray * this){
  return this->ptr;
}
/* SGN_FloatArray_ptr() -------------------------------------------------------- */

SGN_FloatArray * SGN_FloatArray_set_ar(SGN_FloatArray * this, size_t pos,
							   const SGN_FloatArray * ar, size_t size){
  if(this->size < pos + size){
	SGN_FloatArray_set_size(this, pos + size);
  }
  memcpy(this->ptr + pos, ar->ptr, sizeof(float) * size);
  return this;
}
/* SGN_FloatArray_set_ar() ----------------------------------------------------- */

/* End of file.  Copyright (C) 2009  SiGN Project Members */
