/*
  SGN_IntArray.c : Expandable array for integer values.

  Copyright (C) 2008-2011 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#ifndef __SIGN_SSM
#include "SGN_Memory.h"   /* Do not include for SiGN-SSM */
#endif /* __SIGN_SSM */
#include "SGN_IntArray.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <limits.h>

/*!\brief Generates the new SGN_IntArray instance.
 *
 * This generates the new SGN_IntArray instance and initialize it.
 * The generated instance should be deleted (freed) by calling
 * SGN_IntArray_delete() when it is no longer needed.
 *
 *\return  Pointer to the generated \ref SGN_IntArray instance, or NULL
 *   if failed to allocate the enough memory.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_new(){
  SGN_IntArray * ar = malloc(sizeof(SGN_IntArray));
  if(ar == NULL) return NULL;

  ar->ptr = NULL;
  ar->buff_size = 0;
  ar->size = 0;

  return ar;
}
/* SGN_IntArray_new() ========================================================== */

/*!\brief Generates the new \ref SGN_IntArray instance.
 *
 * This generates the new \ref SGN_IntArray instance and initialize it.
 * The generated instance should be deleted (freed) by calling
 * SGN_IntArray_delete() when the user no longer needed to use it.
 *
 *\param size  Initial buffer size.
 *\return  Pointer to the newly generated SGN_IntArray instance, or NULL
 *   if failed to allocate the enough memory.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_news(size_t size){
  SGN_IntArray * this = SGN_IntArray_new();
  if(this == NULL) return NULL;
  if(size != 0){
	SGN_IntArray_expand(this, size);
  }
  return this;
}
/* SGN_IntArray_news() ========================================================= */

void SGN_IntArray_init(SGN_IntArray * ar, size_t size){
  ar->ptr = (int *) malloc(sizeof(int) * size);
  if(ar->ptr == NULL){
	free(ar);
	return;
  }

  ar->buff_size = size;
  ar->size = 0;
}
/* SGN_IntArray_init() ========================================================= */

/*!\brief Deletes the \ref SGN_IntArray instance.
 *
 * This releases all the resources of the instance which is
 * allocated by SGN_IntArray_new().
 * The given pointer can no longer be used.
 *
 * This function can be specified as the second argument
 * of Array_delete_allf().
 *
 *\param this pointer to the \ref SGN_IntArray instance to be deleted.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
void SGN_IntArray_delete(void * this){
  free(((SGN_IntArray *) this)->ptr);
  free(this);
}
/* SGN_IntArray_delete() ------------------------------------------------------- */

/*!\brief Generates multiple \ref SGN_IntArray instances.
 *
 *\param ar_size   the number of \ref SGN_IntArray instances to generate.
 *\param init_size   the initial buffer size for each array.
 *\return  pointer to the first \ref SGN_IntArray instances.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_new_ar(size_t ar_size, size_t init_size){
  int i;
  SGN_IntArray * iar = (SGN_IntArray *) malloc(sizeof(SGN_IntArray) * ar_size);
  
  for(i = 0; i < ar_size; i++){
	SGN_IntArray_init(iar + i, init_size);
  }
  
  return iar;
}
/* SGN_IntArray_new_ar() ------------------------------------------------------- */

/*!\brief Delets the multiple \ref SGN_IntArray instances allocated by
 *    SGN_IntArray_new_ar().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
void SGN_IntArray_delete_ar(SGN_IntArray * this, size_t ar_size){
  int i;
  for(i = 0; i < ar_size; i++){
	free(this[i].ptr);
  }
  free(this);
}
/* SGN_IntArray_delete_ar() ---------------------------------------------------- */

/*!\brief Generates a clone of the \ref SGN_IntArray instance.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_clone(const SGN_IntArray * this){
  SGN_IntArray * ar = SGN_IntArray_news(this->size);
  memcpy(ar->ptr, this->ptr, sizeof(int) * this->size);
  ar->size = this->size;
  return ar;
}
/* SGN_IntArray_clone() ======================================================== */

/*!\brief Expands the size of the buffer memory.
 *
 * The new buffer size will be the current size + length secified
 * by \a length argument.
 *
 * Generally this is called internally, and users do not
 * need to use this.
 *
 *\param this  pointer to the \ref SGN_IntArray instance whose buffer size
 *    is to be expanded.
 *\param length  buffer size to be newly allocated.
 *\return \a this, or NULL if failed.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_expand(SGN_IntArray * this, size_t length){
  int * new_ptr;
  size_t new_size;
  if(length == 0) length = 1;
  new_size = this->buff_size + length;
  new_ptr = realloc(this->ptr, sizeof(int) * new_size);
  if(new_ptr != NULL){
	this->ptr = new_ptr;
	this->buff_size = new_size;
	return this;
  } else {
	return NULL;
  }
}
/* SGN_IntArray_expand() ------------------------------------------------------- */

/*!\brief Adds an integer value to the array.
 *
 *\param this  pointer to the \ref SGN_IntArray instance.
 *\param value  integer value to be added in the array.
 *\return \a this, or NULL if failed to expand the buffer memory.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof Intarray */
SGN_IntArray * SGN_IntArray_add(SGN_IntArray * this, int value){
  if(this->buff_size == this->size){
	SGN_IntArray * ar = SGN_IntArray_expand(this, this->size);
	if(ar == NULL) return NULL;
  }

  this->ptr[this->size++] = value;
  return this;
}
/* SGN_IntArray_add() ---------------------------------------------------------- */

/*!\brief Adds an integer value to the \ref SGN_IntArray.
 *
 * This is idential to SGN_IntArray_add().
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof Intarray */
SGN_IntArray * SGN_IntArray_push(SGN_IntArray * this, int value){
  if(this->buff_size == this->size){
	SGN_IntArray * ar = SGN_IntArray_expand(this, this->size);
	if(ar == NULL) return NULL;
  }

  this->ptr[this->size++] = value;
  return this;
}
/* SGN_IntArray_push() ========================================================= */

/*!\brief Adds multiple values in the integer array.
 *\param this   pointer to the \ref SGN_IntArray instance.
 *\param ar    integer array containing values to be added.
 *\param length   length of \a ar.
 *\return \a this, or NULL if faild to allocate the memory. 
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_add_ar(SGN_IntArray * this, int * ar, size_t length){
  if(this->buff_size < this->size + length){
	SGN_IntArray * tmp = SGN_IntArray_expand(this, 
											 (this->size + length) - this->buff_size);
	if(tmp == NULL) return NULL;
  }
  memcpy(this->ptr + this->size, ar, sizeof(int) * length);
  this->size += length;
  return this;
}
/* SGN_IntArray_add_ar() ------------------------------------------------------- */

/*!\brief Adds an integer value only when it is not already in the array.
 *
 *\param this  pointer to the \ref SGN_IntArray instance.
 *\param value  integer value to be added in the array.
 *\return \a this, or NULL if failed to expand the buffer memory.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray  */
SGN_IntArray * SGN_IntArray_add_nr(SGN_IntArray * this, int value){
  size_t i = SGN_IntArray_find(this, value);
  SGN_IntArray * ret = this;
  if(i == SGN_IntArray_ERROR_VALUE) ret = SGN_IntArray_add(this, value);
  return ret;
}
/* SGN_IntArray_add_nr() ------------------------------------------------------- */

/*!\brief Adds integer values in the second array to the first array.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_add_IntArray(SGN_IntArray * this, const SGN_IntArray * values){
  return SGN_IntArray_add_ar(this, values->ptr, values->size);
}
/* SGN_IntArray_add_IntArray_nr() ---------------------------------------------- */

/*!\brief Adds integer values in the array with checking redundancy.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_add_IntArray_nr(SGN_IntArray * this, const SGN_IntArray * values){
  int i;
  for(i = 0; i < SGN_IntArray_size(values); i++){
	SGN_IntArray_add_nr(this, SGN_IntArray_get(values, i));
  }
  return this;
}
/* SGN_IntArray_add_IntArray_nr() ---------------------------------------------- */

/*!\brief Returns the number of values stored in the array.
 *\param this  pointer to the \ref SGN_IntArray instance.
 *\return the number of integer values stored in the array.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
size_t SGN_IntArray_size(const SGN_IntArray * this){
  return this->size;
}
/* SGN_IntArray_size() --------------------------------------------------------- */

/*!\brief Returns the integer value stored at the specified
 *   position in the array.
 *
 *\param this  pointer to the \ref SGN_IntArray instance.
 *\param index  position of the value to return.
 *\return integer value stored at the specified position.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
int SGN_IntArray_get(const SGN_IntArray * this, size_t index){
  return this->ptr[index];
}
/* SGN_IntArray_get() ---------------------------------------------------------- */

/*!\brief Returns an integer value at the last position of the \ref SGN_IntArray.
 *
 * The size of the array will be automatically decremented.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
int SGN_IntArray_pop(SGN_IntArray * this){
  return this->ptr[--(this->size)];
}
/* SGN_IntArray_pop() ========================================================== */

/*!\brief Returns the index at which the given value is stored.
 *
 * This requires at most O(n) time where n is the number of values
 * stored in the array.
 *
 *\param this  pointer to the \ref SGN_IntArray instance.
 *\param value  integer value to find.
 *\return index at which the given value is stored, or SGN_IntArray_ERROR_VALUE
 *   if failed to find.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
size_t SGN_IntArray_find(const SGN_IntArray * this, int value){
  size_t i;
  for(i = 0; i < this->size; i++){
	if(this->ptr[i] == value) return i;
  }
  return SGN_IntArray_ERROR_VALUE;
}
/* SGN_IntArray_find() --------------------------------------------------------- */

/*!\brief Sets the value into the specified position of the array.
 *
 *\param this  pointer to the \ref SGN_IntArray instance.
 *\param index  position at which the value is set.
 *\param value   value to be set in the array.
 *\return  \a this itself.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_set(SGN_IntArray * this, size_t index, int value){

  while(this->buff_size <= index){
	SGN_IntArray * ar = SGN_IntArray_expand(this, this->buff_size);
	if(ar == NULL) return NULL;
  }

  if(this->size <= index) this->size = index + 1;

  this->ptr[index] = value;
  return this;
}
/* SGN_IntArray_set() ---------------------------------------------------------- */

/*!\brief Increments a value in the array.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_inc(SGN_IntArray * this, size_t index){
  this->ptr[index]++;
  return this;
}
/* SGN_IntArray_inc() ========================================================== */

/*!\brief Returns the maximum value in the array.
 *
 * If the size of the array is zero, then returns <tt>-INT_MAX</tt>.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
int SGN_IntArray_max(const SGN_IntArray * this){
  int max = - INT_MAX;
  for(size_t i = 0; i < this->size; i++){
	if(max < this->ptr[i]) max = this->ptr[i];
  }
  return max;
}
/* SGN_IntArray_max() ========================================================== */

/*!\brief Inserts a value at the specified position in the array.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_insert(SGN_IntArray * this, size_t index, int value){
  if(this->buff_size == this->size){
	SGN_IntArray * ar = SGN_IntArray_expand(this, this->size);
	if(ar == NULL) return NULL;
  }

  memmove(this->ptr + index + 1, this->ptr + index,
		  sizeof(int) * (this->size - index));
  this->ptr[index] = value;
  this->size++;

  return this;
}
/* SGN_IntArray_insert() ======================================================= */

/*!\brief Removes a value from the array.
 *
 *\param this  pointer to the \ref SGN_IntArray instance.
 *\param index position of the element to be removed from the array.
 *\return \a this itself.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_remove(SGN_IntArray * this, size_t index){
  memmove(this->ptr + index, this->ptr + index + 1, 
		  sizeof(int) * (this->size - index - 1));
  this->size--;
  return this;
}
/* SGN_IntArray_remove() ------------------------------------------------------- */

/*!\brief Removes a value in the array.
 *
 * If \a value is not found in the array, then this does nothing.
 *
 *\param this  pointer to the \ref IntArra instance.
 *\param value  value to be removed from the array.
 *\return \a this itself
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_remove_value(SGN_IntArray * this, int value){
  size_t r = SGN_IntArray_find(this, value);
  if(r != SGN_IntArray_ERROR_VALUE){
	SGN_IntArray_remove(this, r);
  }
  return this;
}
/* SGN_IntArray_remove_value() ------------------------------------------------- */

/*!\brief Allocates the buffer memory for the specified size.
 *
 * This allocates the buffer memory for at least the specified
 * number of elements.  If the buffer is bigger than the
 * specified size, this does nothing.
 *
 * Note: This does not affect the number of elements in the array.
 * This expands only the buffer if required.
 *
 *\param this  pointer to the \ref SGN_IntArray instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_reserve(SGN_IntArray * this, size_t size){
  if(this->buff_size < size){
	SGN_IntArray * ar = SGN_IntArray_expand(this, size - this->buff_size);
	if(ar == NULL) return NULL;
  }
  return this;
}
/* SGN_IntArray_reserve() ------------------------------------------------------ */

/*!\brief  Clears the all elements in the array.
 *
 * This does not release memory allocated by the array.
 * Instead, this simply sets the size as 0, therefore this works
 * very fast.
 *
 *\param this   pointer to the \ref SGN_IntArray instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_clear(SGN_IntArray * this){
  this->size = 0;
  return this;
}
/* SGN_IntArray_clear() -------------------------------------------------------- */

/*!\brief  Specifies the size of the array.
 *
 * If the specified size is bigger than the current size,
 * the values in the expanded area are undefined.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
SGN_IntArray * SGN_IntArray_set_size(SGN_IntArray * this, size_t size){
  if(this->buff_size < size){
	SGN_IntArray_expand(this, size - this->buff_size);
  }
  this->size = size;
  return this;
}
/* SGN_IntArray_set_size() ----------------------------------------------------- */

/*!\brief Copies the contents of the array to another array.
 *
 * Values in the array are over-written.
 * If you do not want to delete existing values and want to add values
 * from the second array, use SGN_IntArray_add_SGN_IntArray().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
void SGN_IntArray_copy(const SGN_IntArray * this, SGN_IntArray * dst){
  if(dst->buff_size < this->size){
	SGN_IntArray_expand(dst, this->size - dst->buff_size);
  }
  memcpy(dst->ptr, this->ptr, sizeof(int) * this->size);
  dst->size = this->size;
}
/* SGN_IntArray_copy() --------------------------------------------------------- */

/*!\brief Copies the contents of the array to another array.
 *
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
void SGN_IntArray_copy_add(const SGN_IntArray * this, SGN_IntArray * dst, int value){
  if(dst->buff_size < this->size + 1){
	SGN_IntArray_expand(dst, this->size + 1 - dst->buff_size);
  }
  memcpy(dst->ptr, this->ptr, sizeof(int) * this->size);
  dst->ptr[this->size] = value;
  dst->size = this->size + 1;
}
/* SGN_IntArray_copy_add() ----------------------------------------------------- */

int * SGN_IntArray_ptr(const SGN_IntArray * this){
  return this->ptr;
}
/* SGN_IntArray_ptr() ---------------------------------------------------------- */

void SGN_IntArray_copy_remove(const SGN_IntArray * this, SGN_IntArray * dst, int index){
  if(dst->buff_size < this->size - 1){
	SGN_IntArray_expand(dst, this->size - dst->buff_size - 1);
  }
  if(index > 0){
	memcpy(dst->ptr, this->ptr, sizeof(int) * index);
  }
  memcpy(dst->ptr + index, this->ptr + index + 1, 
		 sizeof(int) * (this->size - index - 1));
  dst->size = this->size - 1;
}
/* SGN_IntArray_copy_remove() -------------------------------------------------- */

/*!\brief Swaps the contents of two arrays.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
void SGN_IntArray_swap(SGN_IntArray * array1, SGN_IntArray * array2){
  int tmp_size = array1->size;
  int * tmp_ptr = array1->ptr;

  array1->size = array2->size;
  array1->ptr = array2->ptr;
  
  array2->size = tmp_size;
  array2->ptr = tmp_ptr;
}
/* SGN_IntArray_swap() --------------------------------------------------------- */

/*!\brief Fills the array with zero.
 *
 * This sets zero for all the elements.  Note that this does not
 * set zero for unused buffer area.
 *
 *\param this  pointer to the \ref SGN_IntArray instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
void SGN_IntArray_fill_zero(SGN_IntArray * this){
  memset(this->ptr, 0, sizeof(int) * this->size);
}
/* SGN_IntArray_fill_zero() ---------------------------------------------------- */

int SGN_IntArray_cmp(const void * a, const void * b){
  return *((const int *) a) - *((const int *) b);
}

void SGN_IntArray_sort(SGN_IntArray * this){
  qsort(this->ptr, this->size, sizeof(int), SGN_IntArray_cmp);
}
/* SGN_IntArray_sort() --------------------------------------------------------- */

/*!\brief Tests if two \ref SGN_IntArray instances are identical.
 *
 *
 *\return 1 if two arrays are identical, or 0 otherwise.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
int SGN_IntArray_equal(const SGN_IntArray * ar1, const SGN_IntArray * ar2){
  if(ar1->size != ar2->size) return 0;
  for(int i = 0; i < ar1->size; i++){
	if(ar1->ptr[i] != ar2->ptr[i]) return 0;
  }
  return 1;
}
/* SGN_IntArray_equal() ======================================================== */

/*!\brief Tests if the \ref SGN_IntArray instance and the int array are identical.
 *
 *
 *\return 1 if two arrays are identical, or 0 otherwise.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
int SGN_IntArray_equal_ar(const SGN_IntArray * ar1, const int * ar2, size_t len){
  if(ar1->size != len) return 0;
  for(int i = 0; i < ar1->size; i++){
	if(ar1->ptr[i] != ar2[i]) return 0;
  }
  return 1;
}
/* SGN_IntArray_equal_ar() ================================================= */

/*!\brief Prints the contents of the array to the file stream.
 *
 * a comma is used as a delimiter of printing values.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
void SGN_IntArray_fprint(const SGN_IntArray * ar, FILE * fp){
  if(SGN_IntArray_size(ar) == 0) return;
  fprintf(fp, "%d", SGN_IntArray_get(ar, 0));
  for(int i = 1; i < SGN_IntArray_size(ar); i++){
	fprintf(fp, ", %d", SGN_IntArray_get(ar, i));
  }
}
/* SGN_IntArray_fprintf() ====================================================== */

/*!\brief Generates a C-style array containing the values in the array.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
int * SGN_IntArray_gen_array(const SGN_IntArray * ar){
  int * ret = malloc(sizeof(int) * ar->size);
  memcpy(ret, ar->ptr, sizeof(int) * ar->size);
  return ret;
}
/* SGN_IntArray_gen_array() ==================================================== */

#ifdef USE_MPI
/*!\brief [MPI] Broadcasts the SGN_IntArray instance.
 *
 * This is available only for MPI-enabled SiGN library.
 *
 *\param par address of the pointer to the SGN_IntArray instance to be broadcasted.
 *            The instance is automatically allocated for the non-root processes.
 *\param root  root process ID.
 *\param comm  MPI communicator.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_IntArray */
void SGN_IntArray_MPI_Bcast(SGN_IntArray ** par, int root, MPI_Comm comm){
  int rank;
  unsigned long size;

  //fprintf(SIGN_ERR, "SGN_IntArray_MPI_Bcast(root=%d): called.\n",  root);
  //fflush(SIGN_ERR);

  MPI_Comm_rank(comm, &rank);
  //fprintf(SIGN_ERR, "                              rank=%d.\n", rank);
  //fflush(SIGN_ERR);
  if(rank == root) size = (*par)->size;
  MPI_Bcast(&size, 1, MPI_UNSIGNED_LONG, root, comm);
  //fprintf(SIGN_ERR, "                              size=%zu.\n", size);
  //fflush(SIGN_ERR);

  if(rank != root) *par = SGN_IntArray_news(size);
  (*par)->size = size;

  MPI_Bcast((*par)->ptr, size, MPI_INT, root, comm);
}
/* SGN_IntArray_Bcast() -------------------------------------------------------- */

void SGN_IntArray_MPI_Bcast_ar(SGN_IntArray ** par, size_t * num, int root, MPI_Comm comm){
  int rank;
  unsigned long numm = * num;
  unsigned long i;
  SGN_IntArray * iar;

  MPI_Comm_rank(comm, &rank);
  MPI_Bcast(&numm, 1, MPI_UNSIGNED_LONG, root, comm);

  if(root != rank){
	*num = numm;
	*par = (SGN_IntArray *) malloc(sizeof(SGN_IntArray) * numm);
  }
  iar = *par;

  for(i = 0; i < numm; i++){
	unsigned long size = iar[i].size;
	MPI_Bcast(&size, 1, MPI_UNSIGNED_LONG, root, comm);
	if(root != rank){
	  iar[i].size = size;
	  iar[i].buff_size = size;
	  iar[i].ptr = (int *) malloc(sizeof(int) * size);
	}
	MPI_Bcast(iar[i].ptr, size, MPI_INT, root, comm);
  }
}
/* SGN_IntArray_Bcast_ar() ----------------------------------------------------- */

void SGN_IntArray_MPI_Bcast_copy(SGN_IntArray * array, int root, MPI_Comm comm){
  int rank;
  unsigned long size = array->size;
  
  MPI_Comm_rank(comm, &rank);
  MPI_Bcast(&size, 1, MPI_UNSIGNED_LONG, root, comm);

  if(array->buff_size < size){
	SGN_IntArray_expand(array, size - array->buff_size);
  }

  MPI_Bcast(array->ptr, size, MPI_INT, root, comm);
  array->size = size;
}
/* SGN_IntArray_Bcast_copy() --------------------------------------------------- */

/*!\brief [MPI] Sends the contents to the specified process.
 *
 * This is available only with USE_MPI compile option.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *
 *\memberof SGN_IntArray */
void SGN_IntArray_MPI_Send(SGN_IntArray * ar, int dst, MPI_Comm comm){
  unsigned long size = (unsigned long) ar->size;
  MPI_Send(&size, 1, MPI_UNSIGNED_LONG, dst, 0, comm);
  MPI_Send(ar->ptr, size, MPI_INT, dst, 1, comm);
}
/* SGN_IntArray_Send() --------------------------------------------------------- */

/*!\brief [MPI] Receives the contents of the array from the specified process.
 *
 *\param ar
 *\param src
 *\param comm
 *\sa SGN_IntArray_MPI_Send()
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
void SGN_IntArray_MPI_Recv(SGN_IntArray * ar, int src, MPI_Comm comm){
  unsigned long size;
  MPI_Status status;
  MPI_Recv(&size, 1, MPI_UNSIGNED_LONG, src, 0, comm, &status);
  if(ar->buff_size < size){
	SGN_IntArray_expand(ar, size - ar->buff_size);
  }

  MPI_Recv(ar->ptr, size, MPI_INT, src, 1, comm, &status);
  ar->size = (size_t) size;
}
/* SGN_IntArray_MPI_Recv() ----------------------------------------------------- */

/*!\brief [MPI]
 *
 *\sa SGN_IntArray_MPI_Send(), SGN_IntArray_MPI_Recv()
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntArray */
void SGN_IntArray_MPI_Recv_add(SGN_IntArray * ar, int src, MPI_Comm comm){
  unsigned long size;
  MPI_Status status;
  MPI_Recv(&size, 1, MPI_UNSIGNED_LONG, src, 0, comm, &status);
  if(ar->buff_size - ar->size < size){
	SGN_IntArray_expand(ar, size);
  }
  MPI_Recv(ar->ptr + ar->size, size, MPI_INT, src, 1, comm, &status);
  ar->size += size;
}
/* SGN_IntArray_MPI_Recv_add() ------------------------------------------------- */

#endif /* USE_MPI */

/* End of file.  Copyright (C) 2008-2011 SiGN Project Members */
