/*
  sign/util/SGN_IntArray.h : Expandable array for integer values.

  Copyright (C) 2008-2011 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#ifndef __SGN_INT_ARRAY_H
#define __SGN_INT_ARRAY_H

#include <stdio.h>

/*!\brief Expandable array for integer values.
 *
 * SGN_IntArray module provides functions to realize the automatically
 * expandable data container for integer values.
 *
 * Do not deal with values in the fields of SGN_IntArray structure directly.  
 * To use SGN_IntArray module, use SGN_IntArray_new() to generate an SGN_IntArray 
 * instance, and use SGN_IntArray_delete() to release the memory used by the 
 * instance when it is no longer needed.
 *
 *\author Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
 */
typedef struct {
#ifdef DOXY
private:
#endif
  int * ptr;
  size_t size;
  size_t buff_size;
} SGN_IntArray;

#define SGN_IntArray_ERROR_VALUE ((size_t) -1)

SGN_IntArray * SGN_IntArray_new(void);
SGN_IntArray * SGN_IntArray_news(size_t size);
SGN_IntArray * SGN_IntArray_new_ar(size_t ar_size, size_t init_size);
void           SGN_IntArray_delete(void * this);
void           SGN_IntArray_delete_ar(SGN_IntArray * this, size_t ar_size);
SGN_IntArray * SGN_IntArray_clone(const SGN_IntArray * this);
SGN_IntArray * SGN_IntArray_expand(SGN_IntArray * this, size_t length);
size_t         SGN_IntArray_size(const SGN_IntArray * this);
SGN_IntArray * SGN_IntArray_add(SGN_IntArray * this, int value);
SGN_IntArray * SGN_IntArray_push(SGN_IntArray * this, int value);
SGN_IntArray * SGN_IntArray_add_ar(SGN_IntArray * this, int * ar, size_t length);
SGN_IntArray * SGN_IntArray_add_nr(SGN_IntArray * this, int value);
SGN_IntArray * SGN_IntArray_add_IntArray(SGN_IntArray * this, const SGN_IntArray * values);
SGN_IntArray * SGN_IntArray_add_IntArray_nr(SGN_IntArray * this, const SGN_IntArray * values);
int            SGN_IntArray_get(const SGN_IntArray * this, size_t index);
int            SGN_IntArray_pop(SGN_IntArray * this);
size_t         SGN_IntArray_find(const SGN_IntArray * this, int value);
SGN_IntArray * SGN_IntArray_set(SGN_IntArray * this, size_t index, int value);
SGN_IntArray * SGN_IntArray_inc(SGN_IntArray * this, size_t index);
int            SGN_IntArray_max(const SGN_IntArray * this);
SGN_IntArray * SGN_IntArray_insert(SGN_IntArray * this, size_t index, int value);
SGN_IntArray * SGN_IntArray_remove(SGN_IntArray * this, size_t index);
SGN_IntArray * SGN_IntArray_remove_value(SGN_IntArray * this, int value);
SGN_IntArray * SGN_IntArray_clear(SGN_IntArray * this);
SGN_IntArray * SGN_IntArray_set_size(SGN_IntArray * this, size_t size);
void           SGN_IntArray_copy(const SGN_IntArray * this, SGN_IntArray * dst);
void           SGN_IntArray_copy_add(const SGN_IntArray * this, SGN_IntArray * dst, int value);
int *          SGN_IntArray_ptr(const SGN_IntArray * this);
void           SGN_IntArray_copy_remove(const SGN_IntArray * this, SGN_IntArray * dst, int index);
void           SGN_IntArray_swap(SGN_IntArray * array1, SGN_IntArray * array2);
SGN_IntArray * SGN_IntArray_reserve(SGN_IntArray * this, size_t size);
void           SGN_IntArray_fill_zero(SGN_IntArray * this);
void           SGN_IntArray_sort(SGN_IntArray * this);
int            SGN_IntArray_equal(const SGN_IntArray * ar1, const SGN_IntArray * ar2);
int            SGN_IntArray_equal_ar(const SGN_IntArray * ar1, const int * ar2, size_t len);
int *          SGN_IntArray_gen_array(const SGN_IntArray * ar);
void           SGN_IntArray_fprint(const SGN_IntArray * ar, FILE * fp);

#ifdef USE_MPI
#include <mpi.h>
void SGN_IntArray_MPI_Bcast(SGN_IntArray ** par, int root, MPI_Comm comm);
void SGN_IntArray_MPI_Bcast_ar(SGN_IntArray ** par, size_t * num, int root, MPI_Comm comm);
void SGN_IntArray_MPI_Bcast_copy(SGN_IntArray * array, int root, MPI_Comm comm);
void SGN_IntArray_MPI_Send(SGN_IntArray * ar, int dst, MPI_Comm comm);
void SGN_IntArray_MPI_Recv(SGN_IntArray * ar, int src, MPI_Comm comm);
void SGN_IntArray_MPI_Recv_add(SGN_IntArray * ar, int src, MPI_Comm comm);
#endif /* USE_MPI */

#endif /* __SGN_INT_ARRAY_H */

/* End of file.  Copyright (C) 2008-2011  SiGN Project Members */
