/*
  SGN_IntQueue.c : Expandable Integer Queue

  Copyright (C) 2008 - 2010
    Yoshinori Tamada
      Laboratory of DNA Information Analysis, Human Genome Center
      The Institute of Medical Science, The Unversity of Tokyo

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at ims.u-tokyo.ac.jp>
*/

#ifndef __SIGN_SSM
#include "../signcom.h"
#endif /* __SIGN_SSM */
#include "SGN_IntQueue.h"
#include <string.h>

void SGN_IntQueue_init(SGN_IntQueue * this){
  this->ptr = NULL;
  this->buff_size = 0;
  this->pos = 0;
  this->size = 0;
}

/*!\brief Generates a new \ref SGN_IntQueue instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntQueue */
SGN_IntQueue * SGN_IntQueue_new(){
  SGN_IntQueue * this = malloc(sizeof(SGN_IntQueue));
  SGN_IntQueue_init(this);
  return this;
}
/* SGN_IntQueue_new() ========================================================== */

/*!\brief Generates an array of \ref SGN_IntQueue instances.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntQueue */
SGN_IntQueue * SGN_IntQueue_new_ar(size_t n){
  SGN_IntQueue * this = malloc(sizeof(SGN_IntQueue) * n);
  SGN_IntQueue * ret = this;
  for(size_t i = 0; i < n; i++){
	SGN_IntQueue_init(this++);
  }
  return ret;
}
/* SGN_IntQueue_new_ar() ======================================================= */

/*!\brief Deletes the \ref SGN_IntQueue instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntQueue */
void SGN_IntQueue_delete(void * this){
  free(((SGN_IntQueue *) this)->ptr);
  free(this);
}
/* SGN_IntQueue_delete() ======================================================= */

/*!\brief Deletes the array of the \ref SGN_IntQueue instances.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntQueue */
void SGN_IntQueue_delete_ar(void * this, size_t n){
  SGN_IntQueue * queue = (SGN_IntQueue *) this;
  for(size_t i = 0; i < n; i++){
	free(queue->ptr);
	queue++;
  }
  free(this);
}
/* SGN_IntQueue_delete_ar() ==================================================== */

/*!\brief Clears the \ref SGN_IntQueue instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntQueue */
void SGN_IntQueue_clear(SGN_IntQueue * this){
  this->pos = 0;
  this->size = 0;
}
/* SGN_IntQueue_clear() ======================================================== */

/*!\brief Returns the number of elements in the \ref SGN_IntQueue instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntQueue */
size_t SGN_IntQueue_size(const SGN_IntQueue * this){
  return this->size;
}
/* SGN_IntQueue_size() ========================================================= */

/*!\brief Checks if the \ref SGN_IntQueue is empty.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntQueue */
int SGN_IntQueue_is_empty(const SGN_IntQueue * this){
  if(SGN_IntQueue_size(this) == 0){
	return 1;
  } else {
	return 0;
  }
}
/* SGN_IntQueue_is_empty() ==================================================== */

/*!\brief Expands the buffer size of the \ref SGN_IntQueue instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntQueue */
SGN_IntQueue * SGN_IntQueue_expand(SGN_IntQueue * this, size_t size){
  if(size == 0) size = 1;
  size_t new_size = this->buff_size + size;
  
  int * newptr = realloc(this->ptr, sizeof(int) * new_size);
  if(newptr == NULL) return NULL;
  this->ptr = newptr;

  if(this->pos + this->size > this->buff_size){
	/* The queue locates over the boundary. */
	/* Thus, needs to move memory. */
	memmove(this->ptr + this->pos + size, this->ptr + this->pos,
			sizeof(int) * (this->buff_size - this->pos));
	this->pos += size;
  }
  this->buff_size = new_size;
  return this;
}
/* SGN_IntQueue_expand() ======================================================= */

/*!\brief Enqueues an integer value into the \ref SGN_IntQueue.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntQueue */
void SGN_IntQueue_enqueue(SGN_IntQueue * this, int value){
  if(this->size == this->buff_size){
	SGN_IntQueue_expand(this, this->buff_size);
  }

  size_t rpos = this->pos + this->size;
  if(rpos >= this->buff_size) rpos -= this->buff_size;
  this->ptr[rpos] = value;
  this->size++;
}
/* SGN_IntQueue_enqueue() ====================================================== */

/*!\brief Dequeues an integer value from the \ref SGN_IntQueue.
 *
 * The dequeued value is set to the passed pointer.
 *
 *\return 0 if successful, or 1 if failure because no value is queued.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_IntQueue */
int SGN_IntQueue_dequeue(SGN_IntQueue * this, int * v){

  if(this->size == 0) return 1; /* error */

  *v = this->ptr[this->pos++];
  if(this->pos == this->buff_size) this->pos = 0;
  this->size--;
  return 0;
}
/* SGN_IntQueue_dequeue() ====================================================== */

int SGN_IntQueue_get(const SGN_IntQueue * this, size_t index){
  return this->ptr[(this->pos + index) % this->buff_size];
}
/* SGN_IntQueue_get() ========================================================== */

void SGN_IntQueue_fprint(const SGN_IntQueue * this, FILE * fp){

  for(size_t i = 0; i < this->size; i++){
	if(i != 0) fprintf(fp, ", ");
	fprintf(fp, "%d", this->ptr[(this->pos + i) % this->buff_size]);
  }
}
/* SGN_IntQueue_fprintf() ====================================================== */

/* End of file.  Copyright (C) 2010  HGC, IMS, The University of Tokyo       */
