/*
  util/SGN_KeyObjects.c : Key and its associated objects.
  
  Copyright (C) 2010 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.
  
  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
 */

/*!\defgroup SGN_KeyObjects
 *\brief Key and its associated objects.
 *
 *\author Yoshinori Tamada &lt;tamada at is.s.u-tokyo.ac.jp&gt;
 */

#include "SGN_KeyObjects.h"
#include "SGN_Str.h"
#include "../lang/SGN_Obj.h"

/* FOR REFERENCE
typedef struct {
  SGN_StrArray * keys;
  SGN_ObjArray * objects;
} SGN_KeyObjects;
*/

/*!\brief Generates a new \ref SGN_KeyObjects instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyObjects */ 
SGN_KeyObjects * SGN_KeyObjects_new(){
  SGN_KeyObjects * this = malloc(sizeof(SGN_KeyObjects));
  this->keys = SGN_StrArray_new();
  this->objects = SGN_ObjArray_new();
  return this;
}
/* SGN_KeyObjects_new() ======================================================== */

SGN_KeyObjects * SGN_KeyObjects_new_KeyValues(const SGN_KeyValues * kv){
  SGN_KeyObjects * ko = SGN_KeyObjects_new();
  
  for(int i = 0; i < SGN_KeyValues_size(kv); i++){
	const char * key = SGN_KeyValues_get_key(kv, i);
	SGN_Type type = SGN_KeyValues_get_type(kv, i);
	const void * value = SGN_KeyValues_get_value(kv, i);

	SGN_Object * obj = SGN_Object_new();
	SGN_Object_set(obj, SGN_Type_duplicate(value, type), type);

#ifdef SIGN_DEBUG
	if((int) SGN_Object_type(obj) > (int) SGN_Type_ERROR){
	  fprintf(SIGN_ERR, "ERROR: SGN_KeyObjects_new_KeyValues(): Illigal type number detected: %d\n",
			  (int) SGN_Object_type(obj));
	  abort();
	}
#endif

	SGN_KeyObjects_add(ko, key, obj); 
  }
  return ko;
}

/*!\brief Deletes the \ref SGN_KeyObjects instance.
 *
 *\param this pointer to the \ref SGN_KeyObjects instance to be deleted.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyObjects */ 
void SGN_KeyObjects_delete(SGN_KeyObjects * this){
  SGN_StrArray_delete_all(this->keys);
  SGN_ObjArray_delete_all(this->objects);
  free(this);
}
/* SGN_KeyObjects_delete() ===================================================== */

void SGN_KeyObjects_delete_v(void * this){
  SGN_KeyObjects_delete((SGN_KeyObjects *) this);
}
/* SGN_KeyObjects_delete_v() =================================================== */

/*!\brief Generates a clone of this \ref SGN_KeyObjects instance.
 *
 * NULL is acceptable.
 *
 *\param this pointer to the \ref SGN_KeyObjects instance to clone.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyObjects */ 
SGN_KeyObjects * SGN_KeyObjects_clone(const SGN_KeyObjects * this){
  if(this == NULL) return NULL;
  SGN_KeyObjects * ko = malloc(sizeof(SGN_KeyObjects));
  ko->keys = SGN_StrArray_clone(this->keys);
  ko->objects = SGN_ObjArray_clone(this->objects);

  return ko;
}
/* SGN_KeyObjects_clone() ====================================================== */

/*!\brief Clears the objects in this \ref SGN_KeyObjects instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyObjects */ 
SGN_KeyObjects * SGN_KeyObjects_clear(SGN_KeyObjects * this){
  SGN_StrArray_clear_all(this->keys);
  SGN_ObjArray_clear_all(this->objects);
  return this;
}
/* SGN_KeyObjects_clear() ====================================================== */

/*!\brief Returns the number of objects in this \ref SGN_KeyObjects instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyObjects */ 
size_t SGN_KeyObjects_size(const SGN_KeyObjects * this){
  return SGN_StrArray_size(this->keys);
}
/* SGN_KeyObjects_size() ======================================================= */

/*!\brief
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyObjects */ 
SGN_Object * SGN_KeyObjects_get(SGN_KeyObjects * this, const char * key){
#ifdef SIGN_DEBUG
  //fprintf(SIGN_ERR, "SGN_KeyObjects_get(key=%s): called.\n", key);
#endif
  size_t pos = SGN_StrArray_find(this->keys, key);
  if(pos == SGN_StrArray_ERROR_VALUE){
	return NULL;
  } else {
	//fprintf(SIGN_ERR, "  found at %zu\n", pos);
#ifdef SIGN_DEBUG
	SGN_Object * obj = SGN_ObjArray_get(this->objects, pos);
	//Object * obj = SGN_KeyObjects_get_value(this, pos);
	//fprintf(SIGN_ERR, "  obj->type=%d\n", (int) obj->type);
	return obj;
#else
	return SGN_ObjArray_get(this->objects, pos);
#endif
  }
}
/* SGN_KeyObjects_get() ======================================================== */

SGN_Object * SGN_KeyObjects_get_const(const SGN_KeyObjects * this, const char * key){
#ifdef SIGN_DEBUG
  //fprintf(SIGN_ERR, "SGN_KeyObjects_get(key=%s): called.\n", key);
#endif
  size_t pos = SGN_StrArray_find(this->keys, key);
  if(pos == SGN_StrArray_ERROR_VALUE){
	return NULL;
  } else {
	//fprintf(SIGN_ERR, "  found at %zu\n", pos);
#ifdef SIGN_DEBUG
	SGN_Object * obj = SGN_ObjArray_get(this->objects, pos);
	//Object * obj = SGN_KeyObjects_get_value(this, pos);
	//fprintf(SIGN_ERR, "  obj->type=%d\n", (int) obj->type);
	return obj;
#else
	return SGN_ObjArray_get(this->objects, pos);
#endif
  }
}

SGN_Object * SGN_KeyObjects_get_value(SGN_KeyObjects * this, size_t index){
  return SGN_ObjArray_get(this->objects, index);
}
/* SGN_KeyObjects_get_value() ================================================== */

const SGN_Object * SGN_KeyObjects_get_const_value(const SGN_KeyObjects * this, size_t index){
  return SGN_ObjArray_get(this->objects, index);
}
/* KeyObject_get_const_value() ============================================= */

SGN_Type SGN_KeyObjects_get_type(const SGN_KeyObjects * this, size_t index){
  const SGN_Object * obj = SGN_KeyObjects_get_const_value(this, index);
  return SGN_Object_type(obj);
}
/* SGN_KeyObjects_get_type() =================================================== */

size_t SGN_KeyObjects_find(const SGN_KeyObjects * this, const char * key){
  return SGN_StrArray_find(this->keys, key);
}
/* SGN_KeyObjects_find() ======================================================= */

const char * SGN_KeyObjects_get_key(const SGN_KeyObjects * this, size_t index){
  return SGN_StrArray_get(this->keys, index);
}
/* SGN_KeyObjects_get_key() ==================================================== */

SGN_KeyObjects * SGN_KeyObjects_add(SGN_KeyObjects * this, const char * key, SGN_Object * object){
  SGN_StrArray_add(this->keys, SGN_Str_copy(key));
  SGN_ObjArray_add(this->objects, object);
  return this;
}
/* SGN_KeyObjects_add() ======================================================== */

SGN_KeyObjects * SGN_KeyObjects_add_type(SGN_KeyObjects * this, const char * key,
								 void * pointer, SGN_Type type){
  SGN_StrArray_add(this->keys, SGN_Str_copy(key));
  SGN_ObjArray_add_type(this->objects, pointer, type);
  return this;
}
/* keyObjects_add_type() =================================================== */

SGN_KeyObjects * SGN_KeyObjects_add_int(SGN_KeyObjects * this, const char * key, int value){
  SGN_Object * object = SGN_Object_new_int(value);
  SGN_KeyObjects_add(this, key, object);
  return this;
}
/* SGN_KeyObjects_add_int() ==================================================== */

SGN_KeyObjects * SGN_KeyObjects_set_int(SGN_KeyObjects * this, const char * key, int value){
  
  SGN_Object * object = SGN_KeyObjects_get(this, key);
  if(object == NULL){
	SGN_KeyObjects_add_int(this, key, value);
  } else {
	*((int *) SGN_Object_ptr(object)) = value;
  }
  return this;
}
/* SGN_KeyObjects_set_int() ==================================================== */

int SGN_KeyObjects_get_int(const SGN_KeyObjects * this, const char * key, int def_value){
  const SGN_Object * object = SGN_KeyObjects_get_const(this, key);
  if(object == NULL){
	return def_value;
  } else {
	return SGN_Object_int_value(object, def_value);
  }
}
/* SGN_KeyObjects_get_int() ==================================================== */

int SGN_KeyObjects_get_int_at(const SGN_KeyObjects * this, size_t index, int def_value){
  const SGN_Object * object = SGN_KeyObjects_get_const_value(this, index);
  if(object == NULL){
	return def_value;
  } else {
	return SGN_Object_int_value(object, def_value);
  }
}
/* SGN_KeyObjects_get_int_at() ================================================= */

SGN_KeyObjects * SGN_KeyObjects_add_double(SGN_KeyObjects * this, const char * key, double value){
  SGN_Object * object = SGN_Object_new_double(value);
  SGN_KeyObjects_add(this, key, object);
  return this;
}
/* SGN_KeyObjects_add_double() ================================================= */

SGN_KeyObjects * SGN_KeyObjects_set_double(SGN_KeyObjects * this, const char * key, 
								   double value){
  SGN_Object * object = SGN_KeyObjects_get(this, key);
  if(object == NULL){
	SGN_KeyObjects_add_double(this, key, value);
  } else {
	*((double *) SGN_Object_ptr(object)) = value;
  }

  return this;
}
/* SGN_KeyObjects_set_double() ================================================= */

SGN_KeyObjects * SGN_KeyObjects_add_string(SGN_KeyObjects * this, const char * key,
								   const char * value){
  SGN_Object * object = SGN_Object_new_string(value);
  SGN_KeyObjects_add(this, key, object);
  return this;
}
/* SGN_KeyObjects_add_string() ================================================= */

SGN_KeyObjects * SGN_KeyObjects_append_KeyObjects(SGN_KeyObjects * this, const SGN_KeyObjects * ko){
  for(size_t i = 0; i < SGN_StrArray_size(ko->keys); i++){
	const char * key = SGN_KeyObjects_get_key(ko, i);
	const SGN_Object * object = SGN_KeyObjects_get_const_value(ko, i);
	SGN_KeyObjects_add(this, key, SGN_Object_clone(object));
  }
  return this;
}
/* SGN_KeyObjects_add_SGN_KeyObjects() ============================================= */

SGN_KeyObjects * SGN_KeyObjects_append_KeyValues(SGN_KeyObjects * this, const SGN_KeyValues * kv){
  for(int i = 0; i < SGN_KeyValues_size(kv); i++){
	const char * key = SGN_KeyValues_get_key(kv, i);
	SGN_Type type = SGN_KeyValues_get_type(kv, i);
	const void * value = SGN_KeyValues_get_value(kv, i);

	SGN_Object * obj = SGN_Object_new();
	SGN_Object_set(obj, SGN_Type_duplicate(value, type), type);

	SGN_KeyObjects_add(this, key, obj);
  }
  return this;
}
/* SGN_KeyObjects_append_KeyValues() =========================================== */

SGN_KeyObjects * SGN_KeyObjects_add_by_string(SGN_KeyObjects * this, const char * key,
									  const char * value, SGN_Type type){
  void * v = SGN_Type_new_str(value, type);

  return SGN_KeyObjects_add_type(this, key, v, type);
}
/* SGN_KeyObjects_add_by_string() ========================================== */

/*!\brief Removes a pair of a key and its value at the specified position.
 *
 *\memberof SGN_KeyObjects */
SGN_KeyObjects * SGN_KeyObjects_remove_at(SGN_KeyObjects * this, size_t index){
  if(index < SGN_KeyObjects_size(this)){
	SGN_StrArray_remove_elm(this->keys, index);
	SGN_ObjArray_remove_obj(this->objects, index);
  }
  return this;
}
/* SGN_KeyObjects_remove_at() ============================================== */

/*
char * SGN_KeyObjects_get_string(const SGN_KeyObjects * this, const char * key){
  SGN_Object * object = SGN_KeyObjects_get(this, key);
  if(object == NULL){
	return NULL;
  } else {
  }
}
*/

void SGN_KeyObjects_fprint(const SGN_KeyObjects * this, FILE * fp){
  for(int i = 0; i < SGN_KeyObjects_size(this); i++){
	if(i != 0) fprintf(fp, ", ");
	const char * key = SGN_KeyObjects_get_key(this, i);
	const SGN_Object * object = SGN_KeyObjects_get_const_value(this, i);
	fprintf(fp, "%s=", key);
	SGN_Object_fprint(object, fp);
  }
}
/* SGN_KeyObjects_fprint() ===================================================== */

void SGN_KeyObjects_check_validity(const SGN_KeyObjects * this){
  for(int i = 0; i < SGN_KeyObjects_size(this); i++){
	const char * key = SGN_KeyObjects_get_key(this, i);
	const SGN_Object * object = SGN_KeyObjects_get_const_value(this, i);

	fprintf(SIGN_ERR, "[%d] key=%s\n", i, key);

	if(SGN_Object_type(object) > SGN_Type_ERROR){
	  fprintf(SIGN_ERR, "FATAL ERROR: SGN_KeyObjects_check_validity():"
			  " Check Failed.\n");
	  abort();
	} else {
	  fprintf(SIGN_ERR, "[%d] type=%d\n", i, (int) SGN_Object_type(object));
	}
  }
}
/* SGN_KeyObjects_check_validity() ============================================= */

/* End of file.  Copyright (C) 2010  SiGN Project Members */
