/*
  sign/util/SGN_KeyValues.c : Functions for dealing with pairs of keys and their values.
  
  Copyright (C) 2008-2011 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

/*
  Memo:

  --ab b,c,d
  a=b:c:d
  
 */

#include "../signcom.h"
#include "../lang/SGN_Type.h"
#include "SGN_Str.h"
#include "SGN_KeyValues.h"
#include "SGN_DoubleArray.h"
#include "SGN_StrBuff.h"
#include "SGN_Memory.h"

#include <string.h>
#include <strings.h>

char * SGN_KeyValues_DEFAULT_VALUE = "on";
char * SGN_KeyValues_DEFAULT_KEY = "_default_";
char * SGN_KeyValues_ERROR_KEY = "_error_";
int SGN_KeyValues_DEFAULT_VALUE_ON_V = 1;
int SGN_KeyValues_DEFAULT_VALUE_OFF_V = 0;
int * SGN_KeyValues_DEFAULT_VALUE_ON = &SGN_KeyValues_DEFAULT_VALUE_ON_V;
int * SGN_KeyValues_DEFAULT_VALUE_OFF = &SGN_KeyValues_DEFAULT_VALUE_OFF_V;

/*!\brief Generates the new \ref SGN_KeyValues instance.
 *
 *\return pointer to the newly generated \ref SGN_KeyValues instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
SGN_KeyValues * SGN_KeyValues_new(){
  SGN_KeyValues * kv;
  kv = (SGN_KeyValues *) malloc(sizeof(SGN_KeyValues));
  kv->keys = SGN_StrArray_new();
  kv->values = SGN_Array_new();
  kv->types = SGN_IntArray_new();
  return kv;
}
/* SGN_KeyValues_new() --------------------------------------------------------- */

/*\brief Generates the copy of the \ref SGN_KeyValues instance.
 *
 * This makes a deep copy of the given \ref SGN_KeyValues instance.
 * All the keys and values are duplicated to the newly allocated
 * memory.
 *
 *\param this  pointer to the \ref keyValues instance to be copied.
 *\return newly generated \ref SGN_KeyValues instance that has the same
 *   contents as \a kv.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
//SGN_KeyValues * SGN_KeyValues_copy(const SGN_KeyValues * this){
//  SGN_KeyValues * kv = SGN_KeyValues_new();
//  SGN_KeyValues_add_KeyValues(kv, this);
//  return kv;
//}
/* SGN_KeyValues_copy() -------------------------------------------------------- */

/*!\brief Generates the copy of the \ref SGN_KeyValues instance.
 *
 * This makes a deep copy of the given \ref SGN_KeyValues instance.
 * All the keys and values are duplicated to the newly allocated
 * memory.
 *
 *\param this  pointer to the \ref keyValues instance to be copied.
 *                NULL acceptable.
 *
 *\return newly generated \ref SGN_KeyValues instance that has the same
 *   contents as \a kv, or NULL if NULL is given.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
SGN_KeyValues * SGN_KeyValues_clone(const SGN_KeyValues * this){
  if(this == NULL) return NULL;
  
  SGN_KeyValues * kv = SGN_KeyValues_new();
  SGN_KeyValues_add_KeyValues(kv, this);
  return kv;
}
/* SGN_KeyValues_clone() ------------------------------------------------------- */

/*
 *\param bg  start position
 *\param spos  split position, i.e., position of '='.
 *\param ed  end position (exclusive)
 */
void SGN_KeyValues_parse_sub(SGN_KeyValues * kv,
						 const char * str, int bg, int spos, int ed){
  char * key;
  void * value;
  int type;

  if(ed <= bg) return;
  if(spos == -1){ /* Key only. E.g. "key" */
	key = SGN_Str_substr(str + bg, ed - bg);
	value = SGN_KeyValues_DEFAULT_VALUE;
	type = SGN_Type_STRING;
  } else {
	int keylen = spos - bg;
	int valuelen = ed - spos - 1;
	if(keylen <= 0) return; /* Do nothing, if e.g. "=value" (no key).*/
	key = SGN_Str_substr(str + bg, spos - bg);
	if(valuelen <= 0) { /* Key only. E.g. "key=" */
	  value = SGN_KeyValues_DEFAULT_VALUE;
	  type = SGN_Type_STRING;
	} else {
	  if(valuelen >= 2 && str[spos + 1] == '{' && str[ed - 1] == '}'){
		/* Value is a 'SGN_KeyValue' */
		value = SGN_KeyValues_parse_sl(str + spos + 2, valuelen - 2);
		type = SGN_Type_KEY_VALUES;
	  } else {
		value = SGN_Str_substr(str + spos + 1, valuelen);
		type = SGN_Type_STRING;
	  }
	}
  }
  SGN_StrArray_add(kv->keys, key);
  SGN_Array_add(kv->values, value);
  SGN_IntArray_add(kv->types, type);
}
/* SGN_KeyValues_parse_sub() =================================================== */

/*!\brief Parses a SGN_KeyValue format string and store keys and values
 *     in the generated SGN_KeyValues instance.
 *
 *\param str pointer to the string to be parsed.  NULL acceptable.  If so,
 *     this returns the empty SGN_KeyValues instance.
 *\return pointer to the \ref SGN_KeyValues instance containing keys and 
 *    values, or NULL if failed.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
SGN_KeyValues * SGN_KeyValues_parse(const char * str){
  if(str == NULL){
	return SGN_KeyValues_new();
  } else {
	return SGN_KeyValues_parse_sl(str, strlen(str));
  }
}
/* SGN_KeyValues_parse() =================================================== */

/*!\brief Parses a SGN_KeyValue format string and store keys and values
 *     in the generated SGN_KeyValues instance.
 *
 *\param str pointer to the string to be parsed.
 *\param length  length of the string to be parsed.  Characters
 *   behind \a length of \a str are ignored.
 *\return pointer to the \ref SGN_KeyValues instance containing keys and 
 *    values, or NULL if failed.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
SGN_KeyValues * SGN_KeyValues_parse_sl(const char * str, size_t length){
  SGN_KeyValues * kv;
  int i, len, spos = -1, level = 0, bg = 0;

  /* 
	 'level' keeps track of how deep the nested structure goes.
	 Parsing happens only when it meets ',' and level == 0.
   */

  kv = SGN_KeyValues_new();
  
  if(str == NULL) return kv;
  len = strlen(str);
  if(len < length) length = len;

  for(i = 0; i < length; i++){
	switch(str[i]){
	case ',':
	  if(level == 0){
		SGN_KeyValues_parse_sub(kv, str, bg, spos, i);
		bg = i + 1;
		spos = -1;
	  }
	  break;
	case '=': if(level == 0) spos = i; break;
	case '{': level++; break;
	case '}': level--; break;
	}
  }

  /* process the lsat element */
  if(bg != i) SGN_KeyValues_parse_sub(kv, str, bg, spos, i);

  return kv;
}
/* SGN_KeyValues_parse_sl() ---------------------------------------------------- */

int SGN_KeyValues_release_at(void * obj, SGN_Type type){
  int ret = 0;
  switch(type){
  case SGN_Type_STRING:
	if((char *) obj != SGN_KeyValues_DEFAULT_VALUE &&
	   (char *) obj != SGN_KeyValues_DEFAULT_KEY &&
	   (char *) obj != SGN_KeyValues_ERROR_KEY){
	  free(obj);
	}
	break;

  case SGN_Type_INT:
	if((int *) obj != SGN_KeyValues_DEFAULT_VALUE_ON &&
	   (int *) obj != SGN_KeyValues_DEFAULT_VALUE_OFF){
	  free(obj);
	}
	break;

  case SGN_Type_KEY_VALUES:
  case SGN_Type_DOUBLE:
  case SGN_Type_DOUBLE_ARRAY:
  case SGN_Type_STR_ARRAY:
  case SGN_Type_INT_ARRAY:
#ifdef SIGN_DEBUG
	if(type == SGN_Type_DOUBLE){
	  fprintf(SIGN_ERR, "KeyValues: Deleting a double value.\n");
	}
#endif
	SGN_Type_delete(obj, type);
	break;


  default:
	fprintf(SIGN_ERR, "FATAL ERROR: SGN_KeyValues_release_at(): "
			"Unsupported type specified: %s\n", SGN_Type_name(type));
	exit(1);
	ret = 1;
	break;
  }
  return ret;
}
/* SGN_KeyValues_release_at() ================================================== */

/*!\brief Deletes the \ref SGN_KeyValues instance.
 *\param this pointer to the \ref SGN_KeyValues instance to be deleted.  NULL is
 *    acceptable.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_delete(SGN_KeyValues * this){
  SGN_KeyValues_delete_v(this);
}
/* ========================================================================= */

void SGN_KeyValues_delete_v(void * this){
  if(this == NULL) return;

  SGN_KeyValues * kv = (SGN_KeyValues *) this;
  int i;
  for(i = 0; i < SGN_IntArray_size(kv->types); i++){
	int r = SGN_KeyValues_release_at(SGN_Array_get(kv->values, i),
								 (SGN_Type) SGN_IntArray_get(kv->types, i));
	if(r != 0){
	  fprintf(SIGN_ERR, "FATAL ERROR: SGN_KeyValues_delete(): "
			  "unsupported value type: %s (%d)\n",
			  SGN_Type_name(SGN_IntArray_get(kv->types, i)),
			  SGN_IntArray_get(kv->types, i));
	  exit(1);
	}

	r = SGN_KeyValues_release_at(SGN_StrArray_get(kv->keys, i), SGN_Type_STRING);
	if(r != 0){
	  fprintf(SIGN_ERR, "FATAL ERROR: SGN_KeyValues_delete(): "
			  "unsupported value type: %s (%d)\n",
			  SGN_Type_name(SGN_IntArray_get(kv->types, i)),
			  SGN_IntArray_get(kv->types, i));
	  exit(1);
	}
  }/* for i < |this->types| */
  
  SGN_StrArray_delete(kv->keys);
  SGN_Array_delete(kv->values);
  SGN_IntArray_delete(kv->types);

  free(this);
}
/* SGN_KeyValues_delete() ------------------------------------------------------ */

/*!\brief Clears all the keys and their values.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_clear(SGN_KeyValues * this){
  int i;
  for(i = 0; i < SGN_IntArray_size(this->types); i++){
	int r = SGN_KeyValues_release_at(SGN_Array_get(this->values, i),
								 (SGN_Type) SGN_IntArray_get(this->types, i));
	if(r != 0){
	  fprintf(SIGN_ERR, "FATAL ERROR: SGN_KeyValues_clear(): "
			  "unsupported value type: %s (%d)\n",
			  SGN_Type_name(SGN_IntArray_get(this->types, i)),
			  SGN_IntArray_get(this->types, i));
	  exit(1);
	}

	r = SGN_KeyValues_release_at(SGN_StrArray_get(this->keys, i), SGN_Type_STRING);
	if(r != 0){
	  fprintf(SIGN_ERR, "FATAL ERROR: SGN_KeyValues_clear(): "
			  "unsupported value type: %s (%d)\n",
			  SGN_Type_name(SGN_IntArray_get(this->types, i)),
			  SGN_IntArray_get(this->types, i));
	  exit(1);
	}
  }/* for i < |this->types| */

  SGN_StrArray_clear(this->keys);
  SGN_Array_clear(this->values);
  SGN_IntArray_clear(this->types);
}
/* SGN_KeyValues_clear() ------------------------------------------------------- */

/*!\brief Returns the number of keys in the \ref SGN_KeyValues instance.
 *
 * Note that this does not count the keys in the nested \ref SGN_KeyValues
 * instances in values.
 *
 *\param this  pointer to the \ref SGN_KeyValue instance.
 *\return number of keys in the \ref SGN_KeyValue instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
size_t SGN_KeyValues_size(const SGN_KeyValues * this){
  return SGN_StrArray_size(this->keys);
}

/*!\brief Returns the key stored at the specified position.
 *
 *\param this  pointer to the \ref SGN_KeyValue instance.
 *\param index  position of the value type to return.
 *\return integer value representing the type of the value stored at the
 *    specified position.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
const char * SGN_KeyValues_get_key(const SGN_KeyValues * this, size_t index){
  return SGN_StrArray_get(this->keys, index);
}

/*!\brief Returns the value type stored at the specified position.
 *
 *\param this  pointer to the \ref SGN_KeyValues instance.
 *\param index  position of the value of the type to return.
 *\return  \ref SGN_Type of the value in \ref SGN_KeyValueSGN_Type.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
SGN_Type SGN_KeyValues_get_type(const SGN_KeyValues * this, size_t index){
  return (SGN_Type) SGN_IntArray_get(this->types, index);
}

/*!\brief Returns the value at the specified position.
 *\param this  pointer to the \ref SGN_KeyValues instance.
 *\param index  position of the value to return.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void * SGN_KeyValues_get_value(const SGN_KeyValues * this, size_t index){
  return SGN_Array_get(this->values, index);
}

/*!\brief Returns the position of the specified key.
 *\param this   pointer to the \ref SGN_KeyValues instance.  NULL is acceptable.
 *\param key   key string.
 *\return position (index) of the specified key, or SGN_KeyValues_ERROR_VALUE if
 *   failed to find or NULL is given for \a this.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
size_t SGN_KeyValues_find_key(const SGN_KeyValues * this, const char * key){
  size_t index;

  if(this == NULL) return SGN_KeyValues_ERROR_VALUE;

  index = SGN_StrArray_find(this->keys, key);
  if(index == SGN_StrArray_ERROR_VALUE) return SGN_KeyValues_ERROR_VALUE;

  return index;
}
/* SGN_KeyValues_find_key() ---------------------------------------------------- */

/*!\brief
 *\return pointer bind to the specified key, or NULL if not found.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void * SGN_KeyValues_get(const SGN_KeyValues * this, const char * key){
  size_t i = SGN_StrArray_find(this->keys, key);
  if(i == SGN_StrArray_ERROR_VALUE) return NULL;

  return SGN_Array_get(this->values, i);
}
/* SGN_KeyValues_get() --------------------------------------------------------- */

/*!\brief Obtains the pointer to the value associated to the given key.
 *
 * This version is not case-sensitive.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void * SGN_KeyValues_get_case(const SGN_KeyValues * this, const char * key){
  size_t i = SGN_StrArray_find_case(this->keys, key);
  if(i == SGN_StrArray_ERROR_VALUE) return NULL;

  return SGN_Array_get(this->values, i);
}
/* SGN_KeyValues_get_case() ==================================================== */

/*!\brief Returns the integer value associated with the specified key.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
int SGN_KeyValues_get_int(const SGN_KeyValues * this, const char * key, int def){
  void * obj = SGN_KeyValues_get_with_type(this, key, SGN_Type_INT);
  if(obj == NULL){
	return def;
  } else {
	return *((int *) obj);
  }
}
/* SGN_KeyValues_get_int() ----------------------------------------------------- */

int SGN_KeyValues_get_int_def(const SGN_KeyValues * this, const char * key, int def){
  void * obj = SGN_KeyValues_get_with_type(this, key, SGN_Type_INT);
  if(obj == NULL){
	return def;
  } else {
	return *((int *) obj);
  }
}
/* SGN_KeyValues_get_int_def() ================================================= */

int SGN_KeyValues_get_int_index(const SGN_KeyValues * this, size_t index){
  return *((int *) SGN_KeyValues_get_value(this, index));
}
/* SGN_KeyValues_get_int_index() =============================================== */

/*!\brief Returns the string value associated with the specified key.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
char * SGN_KeyValues_get_string(const SGN_KeyValues * this, const char * key){
  void * obj = SGN_KeyValues_get_with_type(this, key, SGN_Type_STRING);
  if(obj == NULL){
	return NULL;
  } else {
	return (char *) obj;
  }
}
/* SGN_KeyValues_get_string() ================================================== */

double SGN_KeyValues_get_double(const SGN_KeyValues * this, const char * key, double def){
  void * obj = SGN_KeyValues_get_with_type(this, key, SGN_Type_DOUBLE);
  if(obj == NULL){
	return def;
  } else {
	return *((double *) obj);
  }
}
/* SGN_KeyValues_get_double() ================================================== */

/*!\brief Returns the value if the specified type is identical to one
 *    associated with the key.
 *
 *\return pointer to the value or, NULL if the key is not found or the type is 
 *  different.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void * SGN_KeyValues_get_with_type(const SGN_KeyValues * this, 
								   const char * key, SGN_Type type){
  size_t index;

  if(this == NULL || key == NULL) return NULL;
  index = SGN_StrArray_find(this->keys, key);
  if(index == SGN_StrArray_ERROR_VALUE) return NULL;
  if((SGN_Type) SGN_IntArray_get(this->types, index) != type) return NULL;
  
  return SGN_Array_get(this->values, index);
}/* SGN_KeyValues_get_with_type() ---------------------------------------------- */

/*\brief Returns the value if the specified type is the same as the value type.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_KeyValues */
void * SGN_KeyValues_get_with_type_index(const SGN_KeyValues * this,
										 size_t index, SGN_Type type){
  if((SGN_Type) SGN_IntArray_get(this->types, index) != type) return NULL;
  return SGN_Array_get(this->values, index);
}
/* SGN_KeyValues_get_with_type_index() ===================================== */

/*!\brief Adds a value into the \ref SGN_KeyValues intance.
 *
 *
 * The added pointer is managed by the \ref SGN_KeyValues module.
 * Therefore it may be deleted by the deletion function
 * authomatically.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_add(SGN_KeyValues * this,
				   const char * key, SGN_Type type, void * value){
  SGN_StrArray_add(this->keys, SGN_Str_copy(key));
  SGN_IntArray_add(this->types, (int) type);
  SGN_Array_add(this->values, value);
}
/* SGN_KeyValues_add() --------------------------------------------------------- */

/*!\brief Sets a value at the specified position of the \ref SGN_KeyValues instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_set_at(SGN_KeyValues * this, size_t index,
					  SGN_Type type, void * value){
  SGN_IntArray_set(this->types, index, (int) type);
  SGN_Array_set(this->values, index, value);
}
/* SGN_KeyValues_set_at() ====================================================== */

/*!\brief Puts an integer value into the \ref SGN_KeyValues instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_put_int(SGN_KeyValues * this,
					   const char * key, const int value){
  int * v = (int *) malloc(sizeof(int));
  *v = value;
  SGN_StrArray_add(this->keys, SGN_Str_copy(key));
  SGN_IntArray_add(this->types, SGN_Type_INT);
  SGN_Array_add(this->values, v);
}
/* SGN_KeyValues_put_int() ----------------------------------------------------- */

/*!\brief Sets an integer value into the \ref SGN_KeyValues instance.
 *
 * If there already exists the same key, the value will be overwritten.
 * When overwriting, the previous value is not freed.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_set_int(SGN_KeyValues * this,
					   const char * key, const int value){
  size_t idx = SGN_KeyValues_find_key(this, key);
  if(idx != SGN_KeyValues_ERROR_VALUE){
	if(SGN_KeyValues_get_type(this, idx) == SGN_Type_INT){
	  int * v = (int *) SGN_KeyValues_get_value(this, idx);
	  *v = value;
	} else {
	  int * v = malloc(sizeof(int));
	  *v = value;
	  SGN_KeyValues_set_at(this, idx, SGN_Type_INT, v);
	}
  } else {
	SGN_KeyValues_put_int(this, key, value);
  }
}
/* SGN_KeyValues_set_int() ===================================================== */

/*!\brief Increment the integer value.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_inc_int(SGN_KeyValues * this, const char * key){
  size_t idx = SGN_KeyValues_find_key(this, key);
  if(idx == SGN_KeyValues_ERROR_VALUE){
	SGN_KeyValues_put_int(this, key, 1);
  } else {
	int * v = SGN_KeyValues_get_with_type_index(this, SGN_Type_INT, idx);
	if(v != NULL) (*v)++;
  }
}
/* SGN_KeyValues_inc_int() ================================================= */

/*!\brief Puts a double value into the \ref SGN_KeyValues instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_put_double(SGN_KeyValues * this,
						  const char * key, const double value){
  double * v = (double *) malloc(sizeof(double));
  *v = value;
  SGN_StrArray_add(this->keys, SGN_Str_copy(key));
  SGN_IntArray_add(this->types, SGN_Type_DOUBLE);
  SGN_Array_add(this->values, v);
}
/* SGN_KeyValues_put_double() -------------------------------------------------- */

/*!\brief Sets a double precision float value into the \ref SGN_KeyValues instance.
 *
 * If there already exists the same key, the value will be overwritten.
 * When overwriting, the previous value is not freed.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_set_double(SGN_KeyValues * this,
						  const char * key, const double value){
  size_t idx = SGN_KeyValues_find_key(this, key);
  if(idx != SGN_KeyValues_ERROR_VALUE){
	if(SGN_KeyValues_get_type(this, idx) == SGN_Type_DOUBLE){
	  double * v = (double *) SGN_KeyValues_get_value(this, idx);
	  *v = value;
	} else {
	  double * v = malloc(sizeof(double));
	  *v = value;
	  SGN_KeyValues_set_at(this, idx, SGN_Type_DOUBLE, v);
	}
  } else {
	SGN_KeyValues_put_double(this, key, value);
  }
}
/* SGN_KeyValues_set_double() ================================================== */

/*!\brief Puts a character string into the \ref SGN_KeyValues instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_add_string(SGN_KeyValues * this,
							  const char * key, const char * value){
  SGN_StrArray_add(this->keys, SGN_Str_copy(key));
  SGN_IntArray_add(this->types, SGN_Type_STRING);
  SGN_Array_add(this->values, SGN_Str_copy(value));
}
/* SGN_KeyValues_add_string() -------------------------------------------------- */

/*!\brief Puts a \ref SGN_DoubleArray instance the \ref SGN_KeyValues instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_put_DoubleArray(SGN_KeyValues * this,
							   const char * key, SGN_DoubleArray * ar){
  SGN_StrArray_add(this->keys, SGN_Str_copy(key));
  SGN_IntArray_add(this->types, SGN_Type_DOUBLE_ARRAY);
  SGN_Array_add(this->values, ar);
}
/* SGN_KeyValues_put_DoubleArray() --------------------------------------------- */

/*!\brief Writes a value into the \ref SGN_KeyValues instance.
 *
 * Unlike SGN_KeyValues_add(), this overwrites a value if the key
 * already exists in this \ref SGN_KeyValues instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_write(SGN_KeyValues * this,
					 const char * key, int type, void * value){
  size_t i = SGN_KeyValues_find_key(this, key);
  if(i != SGN_KeyValues_ERROR_VALUE){
	int r = SGN_KeyValues_release_at(SGN_Array_get(this->values, i),
								 (SGN_Type) SGN_IntArray_get(this->types, i));
	if(r != 0){
	}
	
	SGN_IntArray_set(this->types, i, type);
	SGN_Array_set(this->values, i, value);
  } else {
	SGN_KeyValues_add(this, key, type, value);
  }
}
/* SGN_KeyValues_write() ======================================================= */

/*!\brief Adds all the keys and the values in the specified 
 *    \ref SGN_KeyValues instance.
 *
 * All the values will be duplicated, then added into \a this.
 *
 *\param this  pointer to the \ref SGN_KeyValue instance.
 *\param kv    \ref SGN_KeyValue instance whose keys and values are added
 *               to \a this.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_add_KeyValues(SGN_KeyValues * this,
								 const SGN_KeyValues * kv){
  size_t i;
  for(i = 0; i < SGN_KeyValues_size(kv); i++){
	SGN_Type type = SGN_KeyValues_get_type(kv, i);
	SGN_StrArray_add(this->keys, SGN_Str_copy(SGN_KeyValues_get_key(kv, i)));
	SGN_IntArray_add(this->types, type);
	SGN_Array_add(this->values, 
				  SGN_Type_duplicate(SGN_KeyValues_get_value(kv, i), type));
  }
}
/* SGN_KeyValues_add_SGN_KeyValues() ----------------------------------------------- */

/*!\brief Adds all the keys and the values in the specified 
 *    \ref SGN_KeyValues instance.
 *
 * All the values will be duplicated, then added into \a this.
 * Unlike SGN_KeyValues_add_SGN_KeyValues(), this overwrites values for
 * the existing keys.
 *
 *\param this  pointer to the \ref SGN_KeyValue instance.
 *\param kv    \ref SGN_KeyValue instance whose keys and values are added
 *               to \a this.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_add_KeyValues_nr(SGN_KeyValues * this,
									const SGN_KeyValues * kv){
  for(size_t i = 0; i < SGN_KeyValues_size(kv); i++){
	SGN_Type type = SGN_KeyValues_get_type(kv, i);
	const char * key = SGN_KeyValues_get_key(kv, i);
	SGN_KeyValues_write(this, key, type,
					SGN_Type_duplicate(SGN_KeyValues_get_value(kv, i), type));
  }
}
/* SGN_KeyValues_add_SGN_KeyValues_nr() ============================================ */

void SGN_KeyValues_put_as_string(SGN_KeyValues * this, 
							 const char * key, int type, const char * value){
  void * v = SGN_Type_from_str(type, value, NULL);
  SGN_KeyValues_add(this, key, type, v);
}
/* SGN_KeyValues_put_as_string() =============================================== */

/*!\brief Removes a key-value pair by specifing the key.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_remove(SGN_KeyValues * this, const char * key){
  size_t idx = SGN_KeyValues_find_key(this, key);
  if(idx != SGN_KeyValues_ERROR_VALUE){

	int r = SGN_KeyValues_release_at(SGN_Array_get(this->values, idx),
								 (SGN_Type) SGN_IntArray_get(this->types, idx));
	if(r != 0){
	  fprintf(SIGN_ERR, "FATAL ERROR: SGN_KeyValues_remove(): "
			  "unsupported value type: %s (%d)\n",
			  SGN_Type_name(SGN_IntArray_get(this->types, idx)),
			  SGN_IntArray_get(this->types, idx));
	  exit(1);
	}

	r = SGN_KeyValues_release_at(SGN_StrArray_get(this->keys, idx), SGN_Type_STRING);

	SGN_StrArray_remove(this->keys, idx);
	SGN_Array_remove(this->values, idx);
	SGN_IntArray_remove(this->types, idx);
  }
}
/* SGN_KeyValues_remove() ====================================================== */

/*!\brief Sets the values in the \ref SGN_KeyValues instance into the variables.
 *
 * The types of all the values in the \ref SGN_KeyValues instance need to be
 * SGN_KeyValue_SGN_Type_STRING (string values), SGN_KeyValue_SGN_Type_KEY_VALUE 
 * (KeValues instance) type, or one specified by argument \a set.
 *
 * The values are converted to the appropriate types and then set to
 * the variables specified by the \ref SGN_KeyValueSet arrays.
 *
 * NULL is acceptable for \a this.  If so, this does nothing and return
 * with -1.
 *
 *
 *\section special_types Special SGN_Types
 *
 * The following types in \ref SGN_Type have special functions and perform
 * its own conversion when the routine sets a value into the variable.
 *
 * - \ref SGN_Type_STRING_ADD <br>
 *    The variable needs to be \ref SGN_StrArray.
 *    The value (string) is copied and added to the array specified by the variable.
 *    If the variable has NULL, the new instance is allocated.
 * - <b>Other types</b> <br>
 *    Basically, if an atomic typessuch as <tt>int</tt>, <tt>double</tt>, etc..
 *    is specified, then the value is converted its instance from the string
 *    as noted above.
 *
 *\param this  NULL is acceptable.
 *\param set
 *\return -1 if succeeded, or index of key whose value 
 *   caused the error.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
int SGN_KeyValues_set_values_kv(const SGN_KeyValues * this, SGN_KeyValueSet * set){
  size_t size;
  int i, j, flag;

  if(this == NULL) return -1;

  size = SGN_KeyValues_size(this);
  for(i = 0; i < size; i++){
	const char * key = SGN_KeyValues_get_key(this, i);
	const void * value;
	void * arg1;
	void * arg2;
	int type;
	int error_key = -1;

	j = 0; flag = -1;
	/* Search for 'key' in 'set[j]' */
	while(set[j].key != NULL){
	  if(strcmp(set[j].key, key) == 0){
		flag = j;
		break;
	  } else if(strcmp(set[j].key, SGN_KeyValues_ERROR_KEY) == 0){
		error_key = j;
	  }
	  j++;
	}
	
	if(flag == -1){ /* not found */
	  if(error_key != -1){ /* not found but ERROR_KEY is acceptable */
		flag = error_key;
		/* If the ERROR_KEY is acceptable, then 'key' is regarded as 'value' */	
		value = key;
	  } else {
		return i; /* Error */
	  }
	} else {
	  value = SGN_KeyValues_get_value(this, i);	  
	}
	
	type = SGN_KeyValues_get_type(this, i);

	//fprintf(stderr, "DEBUG: SGN_KeyValues_set_values_kv(): key=%s, value=%s\n", key, (char *) value);

	/*
	if(type != SGN_Type_STRING && type != SGN_Type_KEY_VALUES){
	  return i;
	}
	*/

	arg1 = set[flag].arg1;
	arg2 = set[flag].arg2;

	switch(set[flag].type){
	case SGN_Type_INT:
	  if(type == SGN_Type_STRING){
		if(value == SGN_KeyValues_DEFAULT_VALUE){
		  return - i - 2; /* argument error */
		} else {
		  *((int *) arg1) = atoi((char *) value);
		}
	  } else if(type == SGN_Type_INT){
		*((int *) arg1) = *((int *) value);
	  } else {
		return i; /* error */
	  }
	  break;

	case SGN_Type_INT_ADD:
	  if(type == SGN_Type_STRING){
		*((int *) arg1) += atoi((char *) value);
	  } else if(type == SGN_Type_INT){
		*((int *) arg1) += *((int *) value);
	  } else {
		return i; /* error */
	  }
	  break;

	case SGN_Type_INT_OPT:
	  if(type == SGN_Type_STRING){
		if(value == SGN_KeyValues_DEFAULT_VALUE){
		  *((int *) arg1) = *((int *) set[flag].arg2);
		} else {
		  *((int *) arg1) = atoi((char *) value);
		}
	  } else if(type == SGN_Type_INT){
		*((int *) arg1) = *((int *) value);
	  } else {
		return i; /* error */
	  }
	  break;

	case SGN_Type_UINT:
	  if(type == SGN_Type_STRING){
		*((unsigned int *) arg1) = strtoul((char *) value, (char **) NULL, 10);
	  } else if(type == SGN_Type_UINT){
		*((unsigned int *) arg1) = *((unsigned int *) value);
	  } else {
		return i;
	  }
	  break;

	case SGN_Type_CHAR:
	  if(type == SGN_Type_STRING){
		*((char *) arg1) = ((char *) value)[0];
	  } else {
		*((char *) arg1) = atoi((char *) value);
	  }
	  break;

	case SGN_Type_ON_OFF:
	  if(type == SGN_Type_STRING){
		if(strcasecmp((char *) value, "on") == 0 ||
		   strcasecmp((char *) value, "t") == 0 ||
		   strcasecmp((char *) value, "1") == 0){
		  *((int *) arg1) = 1;
		} else {
		  *((int *) arg1) = 0;
		}
	  } else if(type == SGN_Type_INT){
		if(*((int *) value) == 0){
		  /* false */
		  *((int *) arg1) = 0;
		} else {
		  /* true */
		  *((int *) arg1) = 1;
		}
	  }
	  break;

	case SGN_Type_ON:
	  *((int *) arg1) = 1;
	  break;

	case SGN_Type_OFF:
	  *((int *) arg1) = 0;
	  break;

	case SGN_Type_STRING:
	  *((char **) arg1) = SGN_Str_copy((char *) value);
	  break;

	case SGN_Type_STRING_OPT:
	  if(value == SGN_KeyValues_DEFAULT_VALUE){
		*((char **) arg1) = *((char **) set[flag].arg2);
	  } else {
		*((char **) arg1) = SGN_Str_copy((char *) value);
	  }
	  break;

	case SGN_Type_STRING_ADD:
	  {
		SGN_StrArray * sa = *((SGN_StrArray **) arg1);
		if(sa == NULL){
		  sa = SGN_StrArray_news(1);
		  *((SGN_StrArray **) arg1) = sa;
		}
		SGN_StrArray_add(sa, SGN_Str_copy((char *) value));
	  }
	  break;

	case SGN_Type_DOUBLE:
	  if(type == SGN_Type_STRING){
		*((double *) arg1) = atof((char *) value);
	  } else if(type == SGN_Type_DOUBLE){
		*((double *) arg1) = *((double *) value);
	  } else {
		fprintf(SIGN_ERR, "FATAL_ERROR: Unexpected type for key \"%s\": %s",
				key, SGN_Type_name(type));
		exit(1);
	  }
	  break;

	case SGN_Type_KEY_VALUES:
	  if(type == SGN_Type_STRING) {
		*((SGN_KeyValues **) arg1) = SGN_KeyValues_parse((char *) value);
	  } else if(type == SGN_Type_KEY_VALUES){
		//SGN_KeyValues_print((SGN_KeyValues *) value, SIGN_ERR);
		*((SGN_KeyValues **) arg1) = SGN_KeyValues_clone((SGN_KeyValues *) value);
	  } else {
		fprintf(SIGN_ERR, "FATAL_ERROR: Unexpected type for key \"%s\": %s",
				key, SGN_Type_name(type));
		exit(1);
	  }
	  break;

	case SGN_Type_KEY_VALUES_ADD:
	  {
		SGN_KeyValues * kv = *((SGN_KeyValues **) arg1);
		if(kv == NULL){
		  kv = SGN_KeyValues_new();
		  *((SGN_KeyValues **) arg1) = kv;
		}
		SGN_KeyValues * nkv;
		if(type == SGN_Type_STRING){
		  nkv = SGN_KeyValues_parse((char *) value);
		} else {
		  nkv = (SGN_KeyValues *) value;
		}

		SGN_KeyValues_add_KeyValues(kv, nkv);

		if(type == SGN_Type_STRING) SGN_KeyValues_delete(nkv);
	  }
	  break;

	case SGN_Type_ARRAY:
	  {
		if(type == SGN_Type_ARRAY){
		  *((SGN_Array **) arg1) = (SGN_Array *) value;
		} else {
		  fprintf(SIGN_ERR, "FATAL ERROR: SGN_KeyValues_set_value_kv().\n");
		  exit(1);
		}
	  }
	  break;

	case SGN_Type_STR_ARRAY:
	  {
		if(type == SGN_Type_STR_ARRAY){
		  *((SGN_StrArray **) arg1) = (SGN_StrArray *) value;
		} else {
		  fprintf(SIGN_ERR, "FATAL ERROR: SGN_KeyValues_set_value_kv().\n");
		  exit(1);
		}
	  }
	  break;

	case SGN_Type_STR_ARRAY_ADD:
	  {
		SGN_StrArray * sa = *((SGN_StrArray **) arg1);
		SGN_StrArray * tokens;
		if(sa == NULL){
		  sa = SGN_StrArray_news(1);
		  *((SGN_StrArray **) arg1) = sa;
		}
		tokens = SGN_Str_split((char *) value, ':');
		for(j = 0; j < SGN_StrArray_size(tokens); j++){
		  SGN_StrArray_add(sa, SGN_StrArray_get(tokens, j));
		}
		SGN_StrArray_delete(tokens);
	  }
	  break;

	case SGN_Type_STR_ARRAY_ADD2:
	  {
		SGN_StrArray * sa = *((SGN_StrArray **) arg1);
		SGN_StrArray * tokens;
		if(sa == NULL){
		  sa = SGN_StrArray_news(1);
		  *((SGN_StrArray **) arg1) = sa;
		}
		tokens = SGN_Str_split((char *) value, ',');
		for(j = 0; j < SGN_StrArray_size(tokens); j++){
		  SGN_StrArray_add(sa, SGN_StrArray_get(tokens, j));
		}
		SGN_StrArray_delete(tokens);
	  }
	  break;

	case SGN_Type_INT_ARRAY_ADD:
	  {
		SGN_IntArray * ia = *((SGN_IntArray **) arg1);
		SGN_StrArray * tokens;
		if(ia == NULL){
		  ia = SGN_IntArray_new();
		  *((SGN_IntArray **) arg1) = ia;
		}
		tokens = SGN_Str_split((char *) value, ':');
		for(j = 0; j < SGN_StrArray_size(tokens); j++){
		  SGN_IntArray_add(ia, atoi(SGN_StrArray_get(tokens, j)));
		}
		SGN_StrArray_delete(tokens);
	  }
	  break;

	case SGN_Type_INT_ARRAY_ADD2:
	  {
		SGN_IntArray * ia = *((SGN_IntArray **) arg1);
		SGN_StrArray * tokens;
		if(ia == NULL){
		  ia = SGN_IntArray_new();
		  *((SGN_IntArray **) arg1) = ia;
		}
		tokens = SGN_Str_split((char *) value, ',');
		for(j = 0; j < SGN_StrArray_size(tokens); j++){
		  SGN_IntArray_add(ia, atoi(SGN_StrArray_get(tokens, j)));
		}
		SGN_StrArray_delete(tokens);
	  }
	  break;

	case SGN_Type_DOUBLE_ARRAY_ADD:
	  {
		SGN_DoubleArray * da = *((SGN_DoubleArray **) arg1);
		SGN_StrArray * tokens;
		if(da == NULL){
		  da = SGN_DoubleArray_new();
		  *((SGN_DoubleArray **) arg1) = da;
		}
		tokens = SGN_Str_split((char *) value, ':');
		for(j = 0; j < SGN_StrArray_size(tokens); j++){
		  SGN_DoubleArray_add(da, atof(SGN_StrArray_get(tokens, j)));
		}
		SGN_StrArray_delete_all(tokens);
	  }
	  break;

	case SGN_Type_DOUBLE_ARRAY_ADD2:
	  {
		SGN_DoubleArray * da = *((SGN_DoubleArray **) arg1);
		SGN_StrArray * tokens;
		if(da == NULL){
		  da = SGN_DoubleArray_new();
		  *((SGN_DoubleArray **) arg1) = da;
		}
		tokens = SGN_Str_split((char *) value, ',');
		for(j = 0; j < SGN_StrArray_size(tokens); j++){
		  SGN_DoubleArray_add(da, atof(SGN_StrArray_get(tokens, j)));
		}
		SGN_StrArray_delete_all(tokens);
	  }
	  break;

	case SGN_Type_FUNC:
	  {
		void (* func)() = (void (*)() ) arg1;
		func();
	  }
	  break;

	case SGN_Type_FUNC2:
	  {
		void (* func)(const char *, const char *) 
		  = (void (*)(const char *, const char *)) arg1;
		func(key, (const char *) value);
	  }
	  break;

	case SGN_Type_FUNC3:
	  {
		void (* func)(const char *, const char *, void *)
		  = (void (*)(const char *, const char *, void *)) arg1;
		func(key, (const char *) value, set[flag].arg2);
	  }
	  break;

	case SGN_Type_INT_REGION:
	  {
#ifdef SIGN_DEBUG
		fprintf(SIGN_ERR, "KeyValues: SGN_Type_INT_REGION.\n");
#endif
		//fprintf(SIGN_ERR, "DEBUG: KeyValues: '%s'\n", (char *) value);
		SGN_StrArray * tokens = SGN_Str_split((char *) value, ':');
		if(SGN_StrArray_size(tokens) >= 3){
		  /* ERROR */
		  SGN_StrArray_delete_all(tokens);
		  return i;
		}
		int inc = 1;
		if(SGN_StrArray_size(tokens) == 2){
		  inc = atoi(SGN_StrArray_get(tokens, 1));
		}
#ifdef SIGN_DEBUG
		fprintf(SIGN_ERR, "DEBUG: KeyValues:   inc=%d.\n", inc);
#endif

		SGN_StrArray * region = SGN_Str_split(SGN_StrArray_get(tokens, 0), '-');
		if(SGN_StrArray_size(region) == 0){
		  /* ERROR */
		  SGN_StrArray_delete_all(region);
		  SGN_StrArray_delete_all(tokens);
		  return i;
		}
		int bg = atoi(SGN_StrArray_get(region, 0));
#ifdef SIGN_DEBUG
		fprintf(SIGN_ERR, "KeyValues:   bg=%d.\n", bg);
#endif
		int ed = bg;
		if(SGN_StrArray_size(region) == 2){
		  ed = atoi(SGN_StrArray_get(region, 1));
		}
#ifdef SIGN_DEBUG
		fprintf(SIGN_ERR, "KeyValues:   ed=%d.\n", ed);
#endif
		if(ed < bg){
		  /* ERROR */
		  SGN_StrArray_delete_all(region);
		  SGN_StrArray_delete_all(tokens);
		  return i;
		}
		int * ar = malloc(sizeof(int) * (ed - bg + 1));
		//SGN_IntArray * ar = SGN_IntArray_new();
		int p = 0;
		for(int i = bg; i <= ed; i += inc){
		  //SGN_IntArray_add(ar, i);
		  ar[p++] = i;
		}
#ifdef SIGN_DEBUG
		fprintf(SIGN_ERR, "KeyValues:   |ar|=%zu.\n", SGN_IntArray_size(ar));
#endif
		//*((SGN_IntArray **) arg1) = ar;
		*((int **) arg1) = ar;
		*((int *) arg2) = (ed - bg + 1);
		SGN_StrArray_delete_all(region);
		SGN_StrArray_delete_all(tokens);
	  }
	  break;

	case SGN_Type_NONE:
	  /* Perform nothing */
	  break;
	  
	default:
	  fprintf(stderr, "FATAL ERROR:%s:%d: unknown key type detected:%d\n",
			  __FILE__, __LINE__, set[flag].type);
	  exit(1);
	}
  }
  return -1;
}
/* SGN_KeyValues_set_values_kv() =============================================== */

/*!\brief Generates the \ref SGN_KeyValues instance from the command line
 *   arguments.
 *
 * This is a funtion to parse the command line arguments based on
 * the user defined argument table by the array of the \ref SGN_KeyValueSet
 * structure.
 *
 * All the strings beginning with '-' followed by alphabets are regarded
 * as keys and the following strings as their values. E.g. &quot;-key value&quot;.
 * If the string after
 * the key is also a key, the formar key is deemed to have the default value
 * \ref SGN_KeyValues_DEFAULT_VALUE.
 *
 * All the values without keys (or preceding option switches beginning 
 * with '-') are regarded as the values of the special key
 *  SGN_KeyValues_DEFAULT_KEY.
 *
 * Values ending with camma represents that the following strings are also
 * values of the former value.
 *
 *\param argc  integer value that the main() function receives.
 *\param argv  pointer to the string array that the main() function receives.
 *\param set   NULL is acceptable.
 *\return pointer to the generated \ref SGN_KeyValues instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
SGN_KeyValues * SGN_KeyValues_parse_args(int argc, char ** argv, SGN_KeyValueSet * set){
  int i = 0, j;
  char * key = NULL;
  SGN_KeyValues * kv = SGN_KeyValues_new();
  int flag = 0; /* If 1, all the tokens are values of the default key */
  SGN_StrBuff * value = SGN_StrBuff_new();
  
  /* Visit all the arguments */
  for(i = 1; i < argc; i++){
	//fprintf(stderr, "DEBUG: KeyValues: <<%s>>\n", argv[i]);

	if(flag){
	  /* argv[i] is a value without a key */
	  SGN_KeyValues_add(kv, SGN_KeyValues_DEFAULT_KEY, 
					SGN_Type_STRING, SGN_Str_copy(argv[i]));

	} else if(strcmp(argv[i], "-") == 0){
	  /* All the latter tokens are values without keys */
	  flag = 1;
	  
	} else if(argv[i][0] == '-' && !SGN_Str_is_num(argv[i][1])){
	  /* argv[i] is a key */
	  if(key != NULL){
		/* The previous is also a key.  Set the default value for the key. */
		SGN_KeyValues_add(kv, key, SGN_Type_STRING, SGN_KeyValues_DEFAULT_VALUE);
	  }
	  key = argv[i];

	  if(set != NULL){
		for(j = 0; set[j].key != NULL; j++){
		  if(strcmp(key, set[j].key) == 0 && 
			 (set[j].type == SGN_Type_ON || 
			  set[j].type == SGN_Type_OFF)){
			/* 'key' does not take its value */
			SGN_KeyValues_add(kv, key, SGN_Type_STRING, 
						  SGN_KeyValues_DEFAULT_VALUE);
			key = NULL;
			break;
		  }
		}
	  }/* if set != NULL */
	} else if(key != NULL){
	  /* argv[i] is a value */
	  //fprintf(stderr, "DEBUG: KeyValues: key=%s,value=%s\n", key, argv[i]);
	  SGN_StrBuff_add_str(value, argv[i]);
	  //fprintf(stderr, "DEBUG: KeyValues:  <<%s>>\n", StrBuff_str(value));
	  if(argv[i][strlen(argv[i]) - 1] != ','){
		SGN_KeyValues_add(kv, key, SGN_Type_STRING, SGN_StrBuff_gen(value));
		key = NULL;
		SGN_StrBuff_clear(value);
	  }
	} else {
	  /* argv[i] is a value without a key */
	  SGN_KeyValues_add(kv, SGN_KeyValues_DEFAULT_KEY, 
						SGN_Type_STRING, SGN_Str_copy(argv[i]));
	}
  }
  if(key != NULL){
	SGN_KeyValues_add(kv, key, SGN_Type_STRING, SGN_KeyValues_DEFAULT_VALUE);
  }
  SGN_StrBuff_delete(value);
  return kv;
}
/* SGN_KeyValues_parse_args() -------------------------------------------------- */

/*!\brief Parses the command line arguments and set the values into
 *    the variables according to the \ref SGN_KeyValueSet settings.
 *
 * See SGN_KeyValues_parse_args() for the details.
 *
 *\param argc  integer value that the main() function receives.
 *\param argv  pointer to the string array that the main() function receives.
 *\param set
 *\return 0 if succeeded, or integer value representing the index of \a argv
 *   that caused the error.  If the argument takes its argument and it is not specified
 *   or there is a syntax error there, the return value becomes the minus of the 
 *    index of \a argv.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
int SGN_KeyValues_set_args(int argc, char ** argv, SGN_KeyValueSet * set){
  SGN_KeyValues * kv = SGN_KeyValues_parse_args(argc, argv, set);
  //SGN_KeyValues_print(kv, SIGN_ERR); fprintf(SIGN_ERR, "\n");
  int ret = SGN_KeyValues_set_values_kv(kv, set);

  /* find the error key in argv */
  if(ret != -1){
	int i;
	int r = ret;
	if(ret < 0) ret = (ret * -1) - 2;
	const char * key = SGN_KeyValues_get_key(kv, ret);
	if(strcmp(key, SGN_KeyValues_DEFAULT_KEY) == 0){
	  key = (const char *) SGN_KeyValues_get_value(kv, ret);
	}
	for(i = 1; i < argc; i++){
	  if(strcmp(argv[i], key) == 0){
		if(r > 0){
		  ret = i;
		} else {
		  ret = -i;
		}
		break;
	  }		
	}
  } else {
	ret = 0;
  }
  SGN_KeyValues_delete(kv);
  return ret;
}
/* SGN_KeyValues_set_args() ---------------------------------------------------- */

/*!\brief Prints the contents of the \ref SGN_KeyValues instance.
 *
 *\param this pointer to the \ref SGN_KeyValues instance.
 *\param fp file stream.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_KeyValues */
void SGN_KeyValues_print(const SGN_KeyValues * this, FILE * fp){
  int i;
  for(i = 0; i < SGN_KeyValues_size(this); i++){
	const char * key = SGN_KeyValues_get_key(this, i);
	const void * value = SGN_KeyValues_get_value(this, i);
	int type = SGN_KeyValues_get_type(this, i);

	if(i != 0) fprintf(fp, ",");
	
	fprintf(fp, "%s=", key);
	if(type == SGN_Type_KEY_VALUES) fprintf(fp, "{");
	SGN_Type_fprint_value(value, type, fp);
	if(type == SGN_Type_KEY_VALUES) fprintf(fp, "}");
  }
}
/* SGN_KeyValues_print() ------------------------------------------------------- */

void SGN_KeyValues_print_v(const void * this, FILE * fp){
  SGN_KeyValues_print((const SGN_KeyValues *) this, fp);
}

/* End of file.  Copyright (C) 2008-2011 SiGN Project Members */
