/*
  sign/util/SGN_KeyValues.h : Functions for dealing with pairs of keys and their values.
  
  Copyright (C) 2008-2011 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#ifndef __SGN_KEY_VALUES_H
#define __SGN_KEY_VALUES_H

#include "SGN_StrArray.h"
#include "SGN_IntArray.h"
#include "SGN_DoubleArray.h"
#include "../lang/SGN_Type.h"
#include "../signcom.h"

//#define SGN_KeyValues_DEFAULT_VALUE "on"

/*!\brief Value used as a default value for no argument keys.
 *\ingroup SGN_KeyValues */
extern char * SGN_KeyValues_DEFAULT_VALUE;
extern char * SGN_KeyValues_DEFAULT_KEY;
extern char * SGN_KeyValues_ERROR_KEY;

extern int * SGN_KeyValues_DEFAULT_VALUE_ON;
extern int * SGN_KeyValues_DEFAULT_VALUE_OFF;

#define SGN_KeyValues_ERROR_VALUE SIGN_SIZE_T_ERROR_VALUE

/*!\brief Functions for dealing with pairs of keys and their associated values.
 *
 * This class is to deal with a set of keys and their associated values.
 * You can store any types of values with its key.  The key must be a string.
 * You can load (read) a values by specifying the key.  When you store a value,
 * you need to specify the type of the value, defined in \ref SGN_Type class.
 * SGN_keyValues class automatically deletes (or releases) instances or memory
 * spaces when needed by using the type information.
 *
 * \section basics Basic Usage
 *
 * SGN_KeyValues_add() and SGN_KeyValues_get() are the basic functions
 * to store and load key-value pairs.
 *
 * \section parse Parsing String Expressions of Key and Values.
 *
 * This class also provides functions to deal with the string expression
 * of key-value pairs.  The format is '<i>key</i><tt>=</tt><i>value</i>'.
 * For example,  '<TT>length=10</TT>' means 
 * the key '<TT>length</TT>' and its associated value '<TT>10</TT>.'.
 * The parsing of these expressions can be realized by SGN_KeyValues_parse().
 *
 * Multiple pairs of keys and values can be processed at once.
 * For this, use ',' (comma) to concatenate keys and values, like
 * '<TT>length=10,height=30</TT>'.
 *
 * Values can be structured (or nested). i.e. a key can have a set
 *  of pairs of keys and values. To realize this, use '<TT>{</TT>' and
 *  '<TT>}</TT>' to enclose pairs of keys and values to bind it to a
 *  key. For example, '<TT>body={height=10,weight=20}</TT>'.
 *
 * Note that the types of all the values parsed by SGN_KeyValues_parse() 
 * are strings.
 *
 * \section x Parsing Command Line Arguments
 *
 * SGN_KeyValues class also provides functions to deal with command line
 * arguments by SGN_KeyValues_set_args().  See the documents of the
 * function for details.  Here is the example of parsing arguments.
 * This accepts "-i" option taking an integer.  Arguments without
 * option key (such as "-i") can also be set to the variable automatically.
 *
 *\code
 int main(int argc, char * argv[]){
   int i = 0;
   char * file = NULL;
   SGN_KeyValueSet argset[] = {
     {"-i",                      SGN_Type_INT,    &i},
     {SGN_KeyValues_DEFAULT_KEY, SGN_Type_STRING, &file},
     {NULL}, // Terminator
   };

   int r = SGN_KeyValue_set_args(argc, argv, argset);

   printf("i=%d\n", i);
   printf("file=%s\n", file);
   
   return 0;
 }
 \endcode
 *
 *
 *\sa \ref SGN_Type
 *
 *\author Yoshinori Tamada &lt;tamada at is.s.u-tokyo.ac.jp&gt;
 */
typedef struct {
#ifdef DOXY
private:
#endif
  SGN_StrArray * keys;
  SGN_Array * values;
  SGN_IntArray * types;
} SGN_KeyValues;

/*!\brief Structure for specifying keys, types and variables set by
 *    KeyValue_set().
 *
 *\ingroup SGN_KeyValues */
typedef struct {
  char * key;
  SGN_Type type;
  void * arg1;
  void * arg2;
} SGN_KeyValueSet;

SGN_KeyValues *   SGN_KeyValues_new();
SGN_KeyValues *   SGN_KeyValues_clone(const SGN_KeyValues * this);
void          SGN_KeyValues_clear(SGN_KeyValues * this);
SGN_KeyValues *   SGN_KeyValues_parse(const char * str);
SGN_KeyValues *   SGN_KeyValues_parse_sl(const char * str, size_t length);
void          SGN_KeyValues_delete(SGN_KeyValues * this);
void          SGN_KeyValues_delete_v(void * this);
size_t        SGN_KeyValues_size(const SGN_KeyValues * this);
const char *  SGN_KeyValues_get_key(const SGN_KeyValues * this, size_t index);
SGN_Type      SGN_KeyValues_get_type(const SGN_KeyValues * this, size_t index);
void *        SGN_KeyValues_get_value(const SGN_KeyValues * this, size_t index);
int           SGN_KeyValues_set_values_kv(const SGN_KeyValues * this, 
										  SGN_KeyValueSet * set);
size_t       SGN_KeyValues_find_key(const SGN_KeyValues * this, const char * key);
void *       SGN_KeyValues_get(const SGN_KeyValues * this, const char * key);
void *       SGN_KeyValues_get_case(const SGN_KeyValues * this, const char * key);
void         SGN_KeyValues_add(SGN_KeyValues * this,
						   const char * key, SGN_Type type, void * value);

int          SGN_KeyValues_get_int(const SGN_KeyValues * this, const char * key, int def);
int          SGN_KeyValues_get_int_def(const SGN_KeyValues * this, const char * key, int def);
int          SGN_KeyValues_get_int_index(const SGN_KeyValues * this, size_t index);

void         SGN_KeyValues_set_int(SGN_KeyValues * this, const char * key, const int value);

double       SGN_KeyValues_get_double(const SGN_KeyValues * this, const char * key, double def);
void         SGN_KeyValues_set_double(SGN_KeyValues * this,
								  const char * key, const double value);

char *       SGN_KeyValues_get_string(const SGN_KeyValues * this, const char * key);
void *       SGN_KeyValues_get_with_type(const SGN_KeyValues * this, const char * key, SGN_Type type);
void *       SGN_KeyValues_get_with_type_index(const SGN_KeyValues * this, size_t index, SGN_Type type);
SGN_KeyValues *  SGN_KeyValues_parse_args(int argc, char ** argv, SGN_KeyValueSet * set);
int          SGN_KeyValues_set_args(int argc, char ** argv, SGN_KeyValueSet * set);
void         SGN_KeyValues_put_int(SGN_KeyValues * this,
							   const char * key, const int value);
void         SGN_KeyValues_inc_int(SGN_KeyValues * this, const char * key);
void         SGN_KeyValues_put_double(SGN_KeyValues * this,
								  const char * key, const double value);
void         SGN_KeyValues_add_string(SGN_KeyValues * this,
								  const char * key, const char * value);
void         SGN_KeyValues_put_DoubleArray(SGN_KeyValues * this,
										   const char * key, SGN_DoubleArray * ar);
void         SGN_KeyValues_add_KeyValues(SGN_KeyValues * this, const SGN_KeyValues * kv);
void         SGN_KeyValues_add_KeyValues_nr(SGN_KeyValues * this, const SGN_KeyValues * kv);

void         SGN_KeyValues_put_as_string(SGN_KeyValues * this, 
									 const char * key, int type, const char * value);

void         SGN_KeyValues_remove(SGN_KeyValues * this, const char * key);

void         SGN_KeyValues_fprint_value(const int type, const void * value, 
									FILE * fp);
void         SGN_KeyValues_print(const SGN_KeyValues * this, FILE * fp);
void         SGN_KeyValues_print_v(const void * this, FILE * fp);

#endif /* __SGN_KEY_VALUES_H */

/* End of file.  Copyright (C) 2008-2011 SiGN Project Members */
