/*
  sign/util/SGN_Memory.c : Memory utilities

  Copyright (C) 2008-2012 SiGN Project Members.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

/*! \defgroup SGN_Memory
 *\brief Memory handling module
 *
 * Currently, this module is designed to manage
 * memory only for unit testing.
 *
 * When SIGN_DEBUG is defined, memory allocation functions such as malloc()
 * are replaced so that all memory allocation is registered and can be
 * traceable.  The following standard C functions are replaced.
 *
 * - malloc();
 * - realloc();
 * - free();
 *
 *\author Yoshinori Tamada <tamada@ims.u-tokyo.ac.jp>
 */

#include "SGN_Memory.h"
#include "../signlib.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

typedef struct {
  size_t size;
  const char * file;
  int line;
  void * ptr;
} SGN_Memory_field;

static int SGN_Memory_table_pos = 0;
static int SGN_Memory_table_size = 0;
static SGN_Memory_field * SGN_Memory_table = NULL;

#undef malloc
#undef realloc
#undef free

void SGN_Memory_extend_table(){
  size_t new_size = SGN_Memory_table_size * 2;
  if(new_size == 0) new_size = 10;
  SGN_Memory_table = realloc(SGN_Memory_table, sizeof(SGN_Memory_field) * new_size);
  if(SGN_Memory_table == NULL){
	fprintf(stderr, 
			"FATAL ERROR: SGN_Memory_allocate(); Memory allocation failed.\n");
	exit(1);
  }
  SGN_Memory_table_size = new_size;
}
/* SGN_Memory_extend_table() --------------------------------------------------- */

/*!\brief Replacement of malloc().
 *
 * This is a replacement of malloc() when SIGN_DEBUG is defined.
 * Generally, this is not directly called by users.
 *
 *\param size  size of allocating memory in bytes.
 *\param file  name of source code file.  Generally this is automatically
 *   set by using the relaced malloc() function.
 *\param line  line number in the source code.  Generally this is 
 *   automatically set by using the relaced malloc() function.
 *\return pointer to the allocated memory.
 *\ingroup SGN_Memory  */
void * SGN_Memory_allocate(size_t size, const char * file, int line){
  void * ptr;

  ptr = malloc(size);

  /*  
  fprintf(SIGN_ERR, "SGN_Memory_allocate(size=%zu,file=\"%s\",line=%d) = %p; called.\n",
		  size, file, line, ptr);
  fflush(SIGN_ERR);
  */

  if(ptr != NULL){
	/* 
	   Add the allocated memory to the table.  If the table is full, 
	   reallocate the memory for the table.
	 */
#pragma omp critical 
	{
	  if(SGN_Memory_table_pos == SGN_Memory_table_size) SGN_Memory_extend_table();
	
	  SGN_Memory_table[SGN_Memory_table_pos].ptr = ptr;
	  SGN_Memory_table[SGN_Memory_table_pos].size = size;
	  SGN_Memory_table[SGN_Memory_table_pos].file = file;
	  SGN_Memory_table[SGN_Memory_table_pos].line = line;
	  
	  SGN_Memory_table_pos++;
	}
  }

  return ptr;
}
/* SGN_Memory_allocate() ======================================================= */

/*!\brief Replacement of realloc().
 *
 * This is a replacement of realloc() when SIGN_DEBUG is defined.
 * Generally, this is not directly called by users.
 *
 *\param ptr  pointer to be passed as the first parameter of realloc().
 *\param size  size of memory in bytes.
 *\param file   file name.
 *\param line   line number in the file.
 *\return pointer to the allocated memory.
 *\ingroup SGN_Memory */
void * SGN_Memory_reallocate(void * ptr, size_t size, const char * file, int line){
  int i;
  void * new_ptr;
  int table_pos = -1;

  if(ptr == NULL) return SGN_Memory_allocate(size, file, line);
  if(size == 0){
	SGN_Memory_free(ptr, file, line);
	return NULL;
  }

  /* At first, find [ptr] from the table. */
  for(i = 0; i < SGN_Memory_table_pos; i++){
	if(SGN_Memory_table[i].ptr == ptr){
	  /* found */
	  table_pos = i;
	  break;
	}
  }
  if(table_pos == -1){
	/* not found */
	fprintf(stderr, "FATAL ERROR: %s:%d: "
			"Already freed memory specified: [%p]\n",
			file, line, ptr);
	assert(0);
	exit(1);
  }

  new_ptr = realloc(ptr, size);

  /*
  fprintf(SIGN_ERR, "SGN_Memory_reallocate(): ptr=%p\n", new_ptr);
  fflush(SIGN_ERR);
  */

  if(new_ptr != NULL){
	/* Update the table entry */
	SGN_Memory_table[table_pos].ptr = new_ptr;
	SGN_Memory_table[table_pos].size = size;
	SGN_Memory_table[table_pos].file = file;
	SGN_Memory_table[table_pos].line = line;
  }
  return new_ptr;
}

/*!\brief Replacement of free().
 *
 * This is a repacement of free().
 * Generally, this is not directly called by users.
 *
 *\param ptr pointer to be freed.
 *\param file file name.
 *\param line line number in the file.
 *\ingroup SGN_Memory */
void SGN_Memory_free(void * ptr, const char * file, int line){
  int i;
  int flag = 1;

  if(ptr == NULL){
	//fprintf(stderr, "SGN_Memory_free(): NULL given\n");
	return;
  }
#ifdef SIGN_DEBUG
  /*
  fprintf(SIGN_ERR, "SGN_Memory_free(%p): %s:%d\n", ptr, file, line);
  fflush(SIGN_ERR);
  */
#endif

  /* Search [ptr]. */
  for(i = 0; i < SGN_Memory_table_pos; i++){
	if(SGN_Memory_table[i].ptr == ptr){
	  /* found */
	  flag = 0;

	  memmove(SGN_Memory_table + i,
			  SGN_Memory_table + (i + 1),
			  sizeof(SGN_Memory_field) * (SGN_Memory_table_pos - i - 1));

	  SGN_Memory_table_pos--;
	  break;
	}
  }
  if(flag){
	fprintf(SIGN_ERR, "FATAL ERROR: %s:%d: Freed already freed memory.\n",
			file, line);
	fprintf(SIGN_ERR, "             ptr=%p.\n", ptr);
	fflush(SIGN_ERR);
	assert(0);
	exit(1);
  }
}

/*!\brief Show the memory allocation status.
 *\ingroup SGN_Memory */
void SGN_Memory_show(FILE * fp){
  int i;
  size_t total = 0;
  fprintf(fp, "====================================================\n");
  fprintf(fp, "Table Size: %d\n", SGN_Memory_table_size);
  fprintf(fp, "# of allocation: %d\n", SGN_Memory_table_pos);
  fprintf(fp, "----------------------------------------------------\n");
  for(i = 0; i < SGN_Memory_table_pos; i++){
	fprintf(fp, "#%3d %10zu  %p  %s:%d \n", 
			i + 1, 
			SGN_Memory_table[i].size,
			SGN_Memory_table[i].ptr,
			SGN_Memory_table[i].file,
			SGN_Memory_table[i].line);
	total += SGN_Memory_table[i].size;
  }
  fprintf(fp, "----------------------------------------------------\n");
  fprintf(fp, "Total size: %u\n", (unsigned int) total);
  fprintf(fp, "====================================================\n");
}

/*!\brief Returns the number of memory allocation.
 *
 *\return the number of memory allocation.
 *\ingroup SGN_Memory */
int SGN_Memory_num(){
  return SGN_Memory_table_pos;
}

/* End of file.  Copyright (C) 2008-2012 SiGN Project Members */
