/*
  util/SGN_Network.c : Network module
  
  Copyright (C) 2009 - 2011
    Yoshinori Tamada  and
    Laboratory of DNA Information Analysis, Human Genome Center,
	The Institute of Medical Science, The Unversity of Tokyo

  DO NOT COPY, DISTRIBUTE, OR MODIFY WITHOUT PERMISSION.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
 */

#include "../signcom.h"
#include "SGN_Network.h"
#include "SGN_Memory.h"
#include "SGN_Str.h"
#include "SGN_StrMatrix.h"
#include "SGN_IntQueue.h"
#include "SGN_ObjArray.h"
#include "../math/SGN_Graph.h"
#include "../math/SGN_PCGraph.h"
#include "../math/SGN_Random.h"
#include "../lang/SGN_Obj.h"

#include <string.h>
#include <strings.h>
#include <math.h>
#include <float.h>
#include <limits.h>

/*!\brief Generates a new \ref SGN_Network instance.
 *\return pointer to the newly allocated \ref SGN_Network instance.
 *
 *\sa SGN_Network_delete()
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Network * SGN_Network_new(){
  SGN_Network * network;

  network = (SGN_Network *) malloc(sizeof(SGN_Network));
  network->nodes = SGN_Array_new();
  network->parents = SGN_Array_new();
  network->edges = SGN_Array_new();
  network->primary_score = NULL;
  network->prop = NULL;
  network->objects = NULL;

  return network;
}
/* SGN_Network_new() ----------------------------------------------------------- */

/*!\brief Deletes the \ref SGN_Network instance.
 *
 * Note that all objects held by the given \ref SGN_Network instance
 * are deleted.
 *
 *\param this  pointer to the \ref SGN_Network instance to be deleted.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_delete(SGN_Network * this){
  int i;
  
  /* Deletes all nodes */
  SGN_Array_delete_allf(this->nodes, SGN_Node_delete);

  for(i = 0; i < SGN_Array_size(this->parents); i++){
	SGN_IntArray * iar = (SGN_IntArray *) SGN_Array_get(this->parents, i);
	SGN_IntArray_delete(iar);
  }
  SGN_Array_delete(this->parents);

  for(i = 0; i < SGN_Array_size(this->edges); i++){
	SGN_Edge * edge = (SGN_Edge *) SGN_Array_get(this->edges, i);
	SGN_Edge_delete(edge);
  }
  SGN_Array_delete(this->edges);

  free(this);
}
/* SGN_Network_delete() -------------------------------------------------------- */

/*!\brief Generates the complete copy of the given \ref SGN_Network instance.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Network * SGN_Network_clone(const SGN_Network * this){
  SGN_Network * network = SGN_Network_new();

  /* Add all the nodes. */
  for(int i = 0; i < SGN_Network_num_nodes(this); i++){
	SGN_Node * node = SGN_Network_get_node(this, i);
	SGN_Network_add_node(network, SGN_Node_clone(node));
  }/* for i < nodes */

  /* Add all the edges. */
  for(int i = 0; i < SGN_Network_num_edges(this); i++){
	SGN_Edge * edge = SGN_Edge_clone(SGN_Network_get_edge(this, i));

	SGN_Node * parent = SGN_Network_find_node(network, SGN_Edge_parent_name(edge));
	SGN_Node * child = SGN_Network_find_node(network, SGN_Edge_child_name(edge));
	SGN_Edge_set_parent(edge, parent);
	SGN_Edge_set_child(edge, child);
	SGN_Network_add_edge(network, edge);
  }/* for i < edges */

  /* CSML Objects */
  if(this->objects != NULL){
	network->objects = SGN_Array_new();
	for(int i = 0; i < SGN_Array_size(this->objects); i++){
	  CSMLObject * object = CSMLObject_clone(SGN_Array_get(this->objects, i));
	  SGN_Network_add_CSMLObject(network, object);
	}
  }

  return network;
}
/* SGN_Network_copy() ========================================================== */

/*!\brief Returns the number of nodes in the \ref SGN_Network.
 *\param this  pointer to the \ref SGN_Network instance.
 *\return the number of nodes in the network.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
size_t SGN_Network_num_nodes(const SGN_Network * this){
  return SGN_Array_size(this->nodes);
}

/*!\brief Returns the number of edges in the \ref SGN_Network.
 *\param this   pointer to the \ref SGN_Network instance.
 *\return number of edges in the network.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
size_t SGN_Network_num_edges(const SGN_Network * this){
  return SGN_Array_size(this->edges);
}

/*!\brief Adds a \ref SGN_Node to the \ref SGN_Network.
 *\param this pointer to the \ref SGN_Network instance.
 *\param node pointer to the \ref SGN_Node instance to be added to the network.
 *\return \a this itself.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Network * SGN_Network_add_node(SGN_Network * this, SGN_Node * node){
  int index = (int) SGN_Array_size(this->nodes);
  SGN_Array_add(this->nodes, node);
  SGN_Array_add(this->parents, SGN_IntArray_new());
  SGN_Node_set_index(node, index);
  return this;
}

/*!\brief Returns the \ref SGN_Node instance specified by the index.
 *
 * To obtain the index of the node by its name, use 
 * SGN_Network_find_node_index().
 *
 * Do not free (delete) the returned \ref SGN_Node instance.
 *
 *\param this   pointer to the \ref SGN_Network instance.
 *\param v   index of the node to return.
 *\return pointer to the \ref SGN_Node instance, or NULL if
 *  there is no such node in the network.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Node * SGN_Network_get_node(const SGN_Network * this, int v){
  if(v < 0 || v >= SGN_Array_size(this->nodes)){
	return NULL;
  } else {
	return (SGN_Node *) SGN_Array_get(this->nodes, v);
  }
}
/* SGN_Network_get_node() ------------------------------------------------------ */

/*!\brief Finds the node in the \ref SGN_Network and returns the index
 *   of it.
 *
 * This requires O(n) time where n is the number of nodes in
 * the network.
 *
 *\param this  pointer to the \ref SGN_Network instance.
 *\param name  name of the node to find.
 *\return index of the node, or SGN_Network_ERROR_VALUE if not found.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
int SGN_Network_find_node_index(const SGN_Network * this, const char * name){
  size_t i = 0;
  for(i = 0; i < SGN_Array_size(this->nodes); i++){
	SGN_Node * node = (SGN_Node *) SGN_Array_get(this->nodes, i);
	if(strcmp(SGN_Node_name(node), name) == 0){
	  return i;
	}
  }
  return SGN_Network_ERROR_VALUE;
}
/* SGN_Network_find_node_index() ----------------------------------------------- */

/*!\brief  Returns the index of the node in the \ref SGN_Network.
 *
 *\param this   pointer to the \ref SGN_Network instance.
 *\param node   poiner to the \ref node to find.
 *\return  index of the node, or SGN_Network_ERROR_VALUE if not found.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
int SGN_Network_find_node_index_node(const SGN_Network * this,
									 const SGN_Node * node){
  size_t r = SGN_Array_find(this->nodes, node);
  if(r == SGN_Array_ERROR_VALUE) return SGN_Network_ERROR_VALUE;
  return r;
}

/*!\brief Finds the \ref SGN_Node in the \ref SGN_Network and returns it.
 *
 *
 *\param this   pointer to the \ref SGN_Network instance.
 *\param name   name of the node to find.
 *\return   pointer to the \ref SGN_Node instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Node * SGN_Network_find_node(const SGN_Network * this, const char * name){
  int i = 0;
  for(i = 0; i < SGN_Array_size(this->nodes); i++){
	SGN_Node * node = (SGN_Node *) SGN_Array_get(this->nodes, i);
	if(strcmp(SGN_Node_name(node), name) == 0){
	  return node;
	}
  }
  return NULL;
}

/*!\brief Returns the name of the node.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
const char * SGN_Network_node_name(const SGN_Network * this, int u){
  SGN_Node * node = SGN_Network_get_node(this, u);
  return SGN_Node_name(node);
}
/* SGN_Network_node_name() ----------------------------------------------------- */

/*!\brief Returns the alias name of the node.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
const char * SGN_Network_node_alias(const SGN_Network * this, int u){
  SGN_Node * node = SGN_Network_get_node(this, u);
  return SGN_Node_alias(node);
}
/* SGN_Network_node_alias() ---------------------------------------------------- */

/*!\brief Removes a node in the network.
 *
 * This also removes edges connecting the the removing node.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_remove_node_index(SGN_Network * this, int v){
  SGN_Node * node = SGN_Network_get_node(this, v);
  SGN_IntArray * targets = SGN_IntArray_new();

  /* Collect edges to be removed */
  for(int i = 0; i < SGN_Network_num_edges(this); i++){
	SGN_Edge * edge = SGN_Network_get_edge(this, i);
	if(SGN_Edge_get_parent(edge) == node || SGN_Edge_get_child(edge) == node){
	  SGN_IntArray_add(targets, i);
	}
  }

  /* Remove edges */
  for(int i = 0; i < SGN_IntArray_size(targets); i++){
	SGN_Network_remove_edge_index(this,
							  SGN_IntArray_get(targets, 
										   SGN_IntArray_size(targets) - i - 1));
  }

  SGN_Array_remove(this->nodes, v);
  SGN_IntArray * parents = SGN_Array_remove(this->parents, v);

  SGN_Node_delete(node);
  SGN_IntArray_delete(parents);
}
/* SGN_Network_remove_node_index() ============================================= */

/*! Addes an \ref SGN_Edge into the \ref SGN_Network.
 *
 *\param this  pointer to the \ref SGN_Network instance.
 *\param edge  pointer to the \ref SGN_Edge instance to be added to the network.
 *\return this, or NULL if failed.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Network * SGN_Network_add_edge(SGN_Network * this, SGN_Edge * edge){

  size_t u, v, r;
  SGN_IntArray * parents;

  u = SGN_Network_find_node_index_node(this, edge->parent);
  if(u == SGN_Network_ERROR_VALUE) return NULL;

  v = SGN_Network_find_node_index_node(this, edge->child);
  if(v == SGN_Network_ERROR_VALUE) return NULL;

  parents = (SGN_IntArray *) SGN_Array_get(this->parents, v);
  r = SGN_IntArray_find(parents, u);
  if(r == SGN_IntArray_ERROR_VALUE){
	SGN_IntArray_add(parents, u);
	SGN_Array_add(this->edges, edge);
  }
  return this;
}
/* SGN_Network_add_edge() ------------------------------------------------------ */

/*!\brief Adds an edge specified by the parent and the child indexes.
 *
 * If the edge (u, v) is already in the network, this does nothing.
 *
 *\param this  pointer to the \ref SGN_Network instance.
 *\param u  index of the parent of the edge.
 *\param v  index of the child of the edge.
 *\return added or existing \ref SGN_Edge instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Edge * SGN_Network_add_edge_index(SGN_Network * this, int u, int v){
  SGN_IntArray * parents = (SGN_IntArray *) SGN_Array_get(this->parents, v);
  size_t r = SGN_IntArray_find(parents, u);
  if(r == SGN_IntArray_ERROR_VALUE){
	SGN_Edge * e = SGN_Edge_new(SGN_Network_get_node(this, u),
								SGN_Network_get_node(this, v));
	SGN_IntArray_add(parents, u);
	SGN_Array_add(this->edges, e);
	return e;
  } else {
	return SGN_Network_get_edge_index(this, SGN_IntArray_get(parents, r), v);
  }
}
/* SGN_Network_add_edge_index() ------------------------------------------------ */

/*!\brief Removes the specified edge.
 *
 * SGN_Edge instance \a edge will be deleted by SGN_Edge_delete().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_remove_edge(SGN_Network * this, SGN_Edge * edge){
  size_t index = SGN_Network_find_edge_index(this, edge);
  SGN_Array_remove(this->edges, index);

  /* Remove from the parent list */
  SGN_Node * pa_node = SGN_Edge_parent(edge);
  SGN_Node * ch_node = SGN_Edge_child(edge);
  size_t pa = SGN_Network_find_node_index_node(this, pa_node);
  size_t ch = SGN_Network_find_node_index_node(this, ch_node);
  SGN_IntArray * parents = (SGN_IntArray *) SGN_Array_get(this->parents, ch);
  SGN_IntArray_remove_value(parents, pa);

  /* Free the SGN_Edge instance */
  SGN_Edge_delete(edge);
}
/* SGN_Network_remove_edge() =============================================== */

/*!\brief Removes the specified edge from the \ref SGN_Network.
 *
 *\param this  pointer to the \ref SGN_Network instance.
 *\param index  index of the edge to be removed.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_remove_edge_index(SGN_Network * this, int index){
  /* Remove from the edge list */
  SGN_Edge * e = (SGN_Edge *) SGN_Array_remove(this->edges, index);
  
  /* Remove from the parent list */
  SGN_Node * pa_node = SGN_Edge_parent(e);
  SGN_Node * ch_node = SGN_Edge_child(e);
  size_t pa = SGN_Network_find_node_index_node(this, pa_node);
  size_t ch = SGN_Network_find_node_index_node(this, ch_node);
  SGN_IntArray * parents = (SGN_IntArray *) SGN_Array_get(this->parents, ch);
  SGN_IntArray_remove_value(parents, pa);

  /* Free the SGN_Edge instance */
  SGN_Edge_delete(e);
}
/* SGN_Network_remove_edge_index() --------------------------------------------- */

/*!\brief Returns the \ref SGN_Edge specified by the index.
 *
 *\param this  pointer to the \ref SGN_Network instance.
 *\param index   index of the edge to return.
 *\return  poiner to the \ref SGN_Edge instance at the specified position.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Edge * SGN_Network_get_edge(const SGN_Network * this, size_t index){
  return (SGN_Edge *) SGN_Array_get(this->edges, index);
}
/* SGN_Network_get_edge() ------------------------------------------------------ */

/*!\brief Returns the \ref SGN_Edge instance specified by the parent's
 *    and child's indexes.
 *
 *\param this   pointer to the \ref SGN_Network instance.
 *\param u   index of the parent of the edge.
 *\param v   index of the child of the edge.
 *\return   pointer to the edge, or NULL if failed to find.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Edge * SGN_Network_get_edge_index(const SGN_Network * this, int u, int v){
  const SGN_Node * parent = SGN_Network_get_node(this, u);
  const SGN_Node * child = SGN_Network_get_node(this, v);
  return SGN_Network_find_edge(this, parent, child);
}
/* SGN_Network_get_edge() ------------------------------------------------------ */

/*!\brief Checks whether there exsits the specified edge.
 *\param this
 *\param u
 *\param v
 *\return 1 if the edge exists, or 0 otherwise.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
int SGN_Network_check_edge_index(const SGN_Network * this, int u, int v){
  SGN_IntArray * parents = (SGN_IntArray *) SGN_Array_get(this->parents, v);
  int r = SGN_IntArray_find(parents, u);
  if(r == SGN_IntArray_ERROR_VALUE){
	return 0;
  } else {
	return 1;
  }
}
/* SGN_Network_check_edge_index() ---------------------------------------------- */

/*!\brief Finds the edge connecting from and to the specified nodes.
 *
 * This does not compare node names to find the specified nodes.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Edge * SGN_Network_find_edge(const SGN_Network * this,
						 const SGN_Node * parent, const SGN_Node * child){
  size_t i;
  for(i = 0; i < SGN_Array_size(this->edges); i++){
	SGN_Edge * e = (SGN_Edge *) SGN_Array_get(this->edges, i);
	if(e->parent == parent && e->child == child){
	  return e;
	}
  }
  return NULL;
}
/* SGN_Network_find_edge() ----------------------------------------------------- */

/*!\brief Finds the edge connecting from and to the specified nodes.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Edge * SGN_Network_find_edge_name(const SGN_Network * this,
							  const char * parent, const char * child){
  int pa = SGN_Network_find_node_index(this, parent);
  int ch = SGN_Network_find_node_index(this, child);

  return SGN_Network_get_edge_index(this, pa, ch);
}
/* SGN_Network_find_edge_name() ================================================ */

/*!\brief Finds the edge by the SGN_Edge instance and returns its index.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
size_t SGN_Network_find_edge_index(const SGN_Network * this, const SGN_Edge * edge){
  return SGN_Array_find(this->edges, edge);
}
/* SGN_Network_find_edge_index() =============================================== */

/*!\brief Returns the set of parent indexes.
 *
 * DO NOT change, free or delete the returning \ref SGN_IntArray instance.
 * This returns directly the internal parent list for efficiency.
 *
 *\param this   pointer to the \ref SGN_Network instance.
 *\param u   index of the node.
 *\return pointer to the \ref SGN_IntArray instance containing the indexes of
 *    the parents of node \a u.  Do not change, free or delete this
 *    instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
const SGN_IntArray * SGN_Network_get_parent_indexes(const SGN_Network * this, int u){
  const SGN_IntArray * parent = (const SGN_IntArray *) SGN_Array_get(this->parents, u);
  return parent;
}
/* SGN_Network_get_parent_indexes() -------------------------------------------- */

/*!\brief Returns the index of the \a k th parent of \a u.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
int SGN_Network_get_parent_index(const SGN_Network * this, int u, int k){
  return SGN_IntArray_get((const SGN_IntArray *) SGN_Array_get(this->parents, u), k);
}
/* SGN_Network_get_parent_index() ---------------------------------------------- */

/*!\brief Returns the set of child indexes.
 *
 * Unlike SGN_Network_get_parent_indexes(), the returning array can be
 * changed.  Users have responsible to delete the instance.
 * 
 *\param this  pointer to the \ref SGN_Network instance.
 *\param u     index of the node.
 *\return   pointer to the \ref SGN_IntArray instance containing the indexes of
 *   the children of node \a u.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_IntArray * SGN_Network_get_child_indexes(const SGN_Network * this, int u){
  int j;
  SGN_IntArray * ar = SGN_IntArray_new();
  for(j = 0; j < SGN_Array_size(this->parents); j++){
	SGN_IntArray * parents = (SGN_IntArray *) SGN_Array_get(this->parents, j);
	int r = SGN_IntArray_find(parents, u);
	if(r != SGN_IntArray_ERROR_VALUE) SGN_IntArray_add(ar, j);
  }
  return ar;
}
/* SGN_Network_get_child_indexes() --------------------------------------------- */

int SGN_Network_get_child_index(const SGN_Network * this, int u, int k){
  int j;
  int pos = 0;
  for(j = 0; j < SGN_Array_size(this->parents); j++){
	SGN_IntArray * parents = (SGN_IntArray *) SGN_Array_get(this->parents, j);
	int r = SGN_IntArray_find(parents, u);
	if(r != SGN_IntArray_ERROR_VALUE){
	  if(k == pos){
		return j;
	  } else {
		pos++;
	  }
	}
  }/* for j */
  return -1;
}
/* SGN_Network_get_child_index() ----------------------------------------------- */

/*!\brief Copies the indexes of children of the node to the array.
 *
 *\param this  pointer to the \ref SGN_Network instance.
 *\param u  index of the node
 *\param ar  pointer to the \ref SGN_IntArray instance to which
 *    the children of node \a u is copied.
 *\return \a this itself.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_IntArray * SGN_Network_copy_child_indexes(const SGN_Network * this, 
									  int u, SGN_IntArray * ar){
  int j;
  SGN_IntArray_clear(ar);
  for(j = 0; j < SGN_Array_size(this->parents); j++){
	SGN_IntArray * parents = (SGN_IntArray *) SGN_Array_get(this->parents, j);
	int r = SGN_IntArray_find(parents, u);
	if(r != SGN_IntArray_ERROR_VALUE) SGN_IntArray_add(ar, j);
  }
  return ar;
}
/* SGN_Network_get_child_indexes() --------------------------------------------- */

/*!\brief Copies the idexes of neighbor nodes into the \ref SGN_IntArray.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_get_neighbor(const SGN_Network * this, int u, SGN_IntArray * ar){
  SGN_IntArray_copy(SGN_Network_get_parent_indexes(this, u), ar);
  for(int j = 0; j < SGN_Array_size(this->parents); j++){
	SGN_IntArray * parents = (SGN_IntArray *) SGN_Array_get(this->parents, j);
	int r = SGN_IntArray_find(parents, u);
	if(r != SGN_IntArray_ERROR_VALUE) SGN_IntArray_add_nr(ar, j);
  }
}
/* SGN_Network_get_neighbor() ================================================= */

/*!\brief Returns the number of parents of the specified node.
 *
 *\param this pointer to the \ref SGN_Network instance.
 *\param u    index of the node.
 *\return the number of parents of the specified node.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
int SGN_Network_num_parents(const SGN_Network * this, int u){
  const SGN_IntArray * parent;

  parent = (const SGN_IntArray *) SGN_Array_get(this->parents, u);
  return (int) SGN_IntArray_size(parent);
}
/* SGN_Network_num_parents() --------------------------------------------------- */

/*!\brief Returns the number of children of the specified node.
 *
 *\param this pointer to the \ref SGN_Network instance.
 *\param u    index of the node.
 *\return the number of chilren of the specified node.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
int SGN_Network_num_children(const SGN_Network * this, int u){
  int j;
  int children = 0;
  for(j = 0; j < SGN_Array_size(this->parents); j++){
	SGN_IntArray * parents = (SGN_IntArray *) SGN_Array_get(this->parents, j);
	int r = SGN_IntArray_find(parents, u);
	if(r != SGN_IntArray_ERROR_VALUE) children++;
  }
  return children;
}
/* SGN_Network_num_children() -------------------------------------------------- */

/*!\brief Returns the \ref SGN_IntArray containing children indexes for all
 *   the nodes.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_IntArray * SGN_Network_get_all_children(const SGN_Network * this){
  int j;
  int p = SGN_Array_size(this->parents);
  SGN_IntArray * children = SGN_IntArray_new_ar(p, 1);
  for(j = 0; j < p; j++){
	SGN_IntArray * parents = (SGN_IntArray *) SGN_Array_get(this->parents, j);
	int k;
	for(k = 0; k < SGN_IntArray_size(parents); k++){
	  int jk = SGN_IntArray_get(parents, k);
	  SGN_IntArray_add(children + jk, j);
	}
  }
  return children;
}
/* SGN_Network_get_all_children() ---------------------------------------------- */

/*!\brief Return the degree of a node.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
int SGN_Network_node_degree(const SGN_Network * this, int u){
  return SGN_Network_num_parents(this, u) + SGN_Network_num_parents(this, u);
}
/* SGN_Network_node_degree() --------------------------------------------------- */

/*!\brief Generates the \ref Graph instance from the \ref SGN_Network.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Graph * SGN_Network_graph(const SGN_Network * this){
  int u, v, vv;
  SGN_Graph * g = SGN_Graph_new(SGN_Network_num_nodes(this));
  for(u = 0; u < SGN_Array_size(this->parents); u++){
	SGN_IntArray * parents = (SGN_IntArray *) SGN_Array_get(this->parents, u);
	for(vv = 0; vv < SGN_IntArray_size(parents); vv++){
	  v = SGN_IntArray_get(parents, vv);
	  SGN_Graph_add_edge_nr(g, v, u);
	}
  }
  return g;
}
/* SGN_Network_graph() --------------------------------------------------------- */

/*!\brief Generates the \ref Graph object based on the dataset.
 *
 * This generates the \ref Graph object from the \ref SGN_Network object.
 * The indexes of nodes in the graph will be the same as in the given
 * dataset.
 *
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Graph * SGN_Network_graph_dataset(const SGN_Network * this, const SGN_Data * dataset){
  SGN_Graph * g = SGN_Graph_new(dataset->p);
  int np = SGN_Network_num_nodes(this);
  size_t table[np];

  /* Constructs the conversion table */
  for(int j = 0; j < np; j++){
	table[j] = SGN_Data_find_name(dataset, SGN_Network_node_name(this, j));
	if(table[j] == SGN_StrArray_ERROR_VALUE){
	  fprintf(SIGN_ERR,
			  "WARNING: not found: %s\n", SGN_Network_node_name(this,j));
	}
  }/* for j < np */

  for(int u = 0; u < np; u++){
	if(table[u] == SGN_StrArray_ERROR_VALUE) continue;
	
	SGN_IntArray * parents = (SGN_IntArray *) SGN_Array_get(this->parents, u);
	for(int k = 0; k < SGN_IntArray_size(parents); k++){
	  int v = SGN_IntArray_get(parents, k);
	  if(table[v] == SGN_StrArray_ERROR_VALUE) continue;
	  SGN_Graph_add_edge_nr(g, table[v], table[u]);
	}/* k < |parents| */
  }/* u < np */
  
  return g;
}
/* SGN_Network_graph_dataset() ================================================= */


/*!\brief Generates the \ref AdjGraph based on the dataset from the
 *   given \ref SGN_Network instance.
 *
 * This generates the \ref AdjGraph instance from the \ref SGN_Network instance.
 * The indexes of nodes in the graph will be the same as in the given
 * dataset.
 *
 * Warning messages can be controled by SGN_set_warning_level().
 * When the warning level is set to 0, this does not output any warning
 * messages.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_AdjGraph * SGN_Network_AdjGraph_dataset(const SGN_Network * this, const SGN_Data * dataset){
  SGN_AdjGraph * g = SGN_AdjGraph_news(dataset->p);
  int np = SGN_Network_num_nodes(this);
  size_t table[np];

  /* Constructs the conversion table */
  for(int j = 0; j < np; j++){
	table[j] = SGN_Data_find_name(dataset, SGN_Network_node_name(this, j));
	if(table[j] == SGN_StrArray_ERROR_VALUE && 1 <= SGN_get_warning_level()){
	  fprintf(SIGN_ERR,
			  "WARNING: not found: %s\n", SGN_Network_node_name(this,j));
	}
  }/* for j < np */

  for(int u = 0; u < np; u++){
	if(table[u] == SGN_StrArray_ERROR_VALUE) continue;
	
	SGN_IntArray * parents = (SGN_IntArray *) SGN_Array_get(this->parents, u);
	for(int k = 0; k < SGN_IntArray_size(parents); k++){
	  int v = SGN_IntArray_get(parents, k);
	  if(table[v] == SGN_StrArray_ERROR_VALUE) continue;
	  SGN_AdjGraph_add_edge(SGN_AdjGraph_to_GraphI(g), table[v], table[u]);
	}/* k < |parents| */
  }/* u < np */
  
  return g;
}
/* SGN_Network_graph_dataset() ================================================= */

/*!\brief Generate the \ref SGN_PCGraph instance from the \ref SGN_Network.
 *
 * The order of edges is not the same as in the original network.
 * If you want to keep the order of edges, use
 * SGN_Network_pcgraph_ord() instead of this.
 *
 *\param network   pointer to the \ref SGN_Network instance.
 *\return \ref SGN_PCGraph instance generated from \a network.
 *\sa SGN_Network_pcgraph_ord().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_PCGraph * SGN_Network_pcgraph(const SGN_Network * network){
  size_t j, k;
  SGN_PCGraph * g = SGN_PCGraph_new(SGN_Network_num_nodes(network));
  for(j = 0; j < SGN_Network_num_nodes(network); j++){
	const SGN_IntArray * parents = SGN_Network_get_parent_indexes(network, j);
	for(k = 0; k < SGN_IntArray_size(parents); k++){
	  SGN_PCGraph_add_edge(g, SGN_IntArray_get(parents, k), j);
	}
  }
  return g;
}
/* SGN_Network_pcgraph() ------------------------------------------------------- */

/*!\brief Generates the \ref SGN_PCGraph instance from the \ref SGN_Network.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_PCGraph * SGN_Network_pcgraph_ord(const SGN_Network * network){
  size_t i;
  SGN_PCGraph * g = SGN_PCGraph_new(SGN_Network_num_nodes(network));
  
  for(i = 0; i < SGN_Network_num_edges(network); i++){
	SGN_Edge * edge = SGN_Network_get_edge(network, i);
	int pa, ch;

	pa = SGN_Network_find_node_index_node(network, SGN_Edge_get_parent(edge));
	ch = SGN_Network_find_node_index_node(network, SGN_Edge_get_child(edge));

	SGN_PCGraph_add_edge(g, pa, ch);
  }
  return g;
}
/* SGN_Network_pcgraph_ord() -------------------------------------------------- */

/*!\brief Constructs the \ref SGN_Network from a \ref Graph.
 *
 *
 * \a network is used to specify node names and so on.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Network * SGN_Network_from_graph_network(const SGN_Graph * graph, 
											 const SGN_Network * network){
  size_t u;
  SGN_Network * ret = SGN_Network_new();
  
  for(u = 0; u < SGN_Network_num_nodes(network); u++){
	SGN_IntArray * parents = SGN_Graph_parents(graph, u);
	int j;

	SGN_Node * uu = NULL;
	
	for(j = 0; j < SGN_IntArray_size(parents); j++){
	  int v = SGN_IntArray_get(parents, j);
	  SGN_Node * vv = SGN_Network_find_node(ret, SGN_Network_node_name(network, v));
	  if(vv == NULL){
		vv = SGN_Node_new(SGN_Network_node_name(network, v),
					  SGN_Network_node_alias(network, v));
		SGN_Network_add_node(ret, vv);
	  }

	  if(uu == NULL){
		uu = SGN_Network_find_node(ret, SGN_Network_node_name(network, u));
		if(uu == NULL){
		  uu = SGN_Node_new(SGN_Network_node_name(network, u),
						SGN_Network_node_alias(network, u));
		  SGN_Network_add_node(ret, uu);
		}
	  }

	  /*
	  fprintf(stderr, "%s -> %s (%s -> %s)\n",
			  SGN_Node_alias(vv), SGN_Node_alias(uu),
			  SGN_Network_node_alias(network, v), SGN_Network_node_alias(network, u));
	  */

	  SGN_Network_add_edge(ret, SGN_Edge_new(vv, uu));
	}
  }
  return ret;
}
/* SGN_Network_from_graph_network() -------------------------------------------- */

/*!\brief Constructs the \ref SGN_Network from a \ref Graph.
 *
 * This is similar to SGN_Network_from_graph_network() but this uses
 * a \ref SGN_Data instance to extract node names.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Network * SGN_Network_from_graph(const SGN_Graph * graph, const SGN_Data * dataset){
  SGN_Network * net = SGN_Network_new();
  int j, k;
  const int p = dataset->p;

  /* Add nodes */
  for(j = 0; j < p; j++){
	SGN_Node * node = SGN_Node_new(SGN_Data_get_name(dataset, j),
						   SGN_Data_get_alias(dataset, j));
  fflush(SIGN_ERR);
	SGN_Network_add_node(net, node);
  }

  /* Add edges */
  for(j = 0; j < p; j++){
	SGN_IntArray * parents = SGN_Graph_parents(graph, j);
	SGN_IntArray_sort(parents);
	for(k = 0; k < SGN_IntArray_size(parents); k++){
	  int kj = SGN_IntArray_get(parents, k);
	  SGN_Network_add_edge_index(net, kj, j);
	}
  }

  return net;
}
/* SGN_Network_from_graph() ==================================================== */

/*!\brief Generates a new \ref SGN_Network instance from a \ref SGN_PCGraph.
 *
 * SGN_Node names will be <TT>g_1</TT> ... <TT>g_</TT><I>n</I>
 * where <I>n</I> is the number of nodes in the graph.
 *
 *\param g  pointer to the \ref SGN_PCGraph instance.
 *\return newly generated \ref SGN_Network instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Network * SGN_Network_from_pcgraph(const SGN_PCGraph * g, SGN_Data * dataset){
  SGN_Network * network = SGN_Network_new();
  size_t n = SGN_PCGraph_num_nodes(g);
  size_t i;
  char buff[256];

  for(i = 0; i < n; i++){
	SGN_Node * node;
	if(dataset == NULL){
	  sprintf(buff, "g_%d", (int)(i + 1));
	  node = SGN_Node_new(buff, buff);
	} else {
	  node = SGN_Node_new(SGN_Data_get_name(dataset, i),
					  SGN_Data_get_alias(dataset, i));
	}
	SGN_Network_add_node(network, node);
  }

  for(i = 0; i < SGN_PCGraph_num_edges(g); i++){
	int u, v;
	SGN_PCGraph_get_edge_index(g, i, &u, &v);
	SGN_Network_add_edge_index(network, u, v);
  }
  
  return network;
}
/* SGN_Network_from_pcgraph() -------------------------------------------------- */

/*!\brief Sets the primary score name to the network.
 *
 * The specified string will be copied (duplicated).
 *
 *\param this  pointer to the \ref SGN_Network instance.
 *\param score_name
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_set_primary_score(SGN_Network * this, const char * score_name){
  this->primary_score = SGN_Str_copy(score_name);
}
/* SGN_Network_set_primary_score() --------------------------------------------- */

/*!\brief Returns the primary score attribute name.
 *\param this  pointer to the \ref SGN_Network instance.
 *\return edge attribute name of the primary score.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
const char * SGN_Network_get_primary_score(const SGN_Network * this){
  return this->primary_score;
}
/* SGN_Network_get_primary_score() --------------------------------------------- */

/*!\brief Replace node names according to the name table.
 *
 * The name table is give as a \SGN_StrMatrix table where the first column
 * is the name to be replaced, and the second one to replace, i.e. the new 
 * name.
 *
 * The alias names are not affected by this function.
 *
 *\param network   pointer to the \ref SGN_Network instance whose node
 *   names will be replaced.
 *\param table   pointer to the \ref SGN_StrMatrix instance which is
 *   the name table for the name replacement.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_replace_node_names(SGN_Network * network, const SGN_StrMatrix * table){
  int i, j;
  for(i = 0; i < SGN_Network_num_nodes(network); i++){
	SGN_Node * node = SGN_Network_get_node(network, i);
	const char * new_name = NULL;
	/* Find the node name. */
	for(j = 0; j < SGN_StrMatrix_lines(table); j++){
	  SGN_StrArray * line = SGN_StrMatrix_get_line(table, j);
	  if(strcmp(SGN_StrArray_get(line, 0), SGN_Node_name(node)) == 0){
		/* found */
		new_name = SGN_StrArray_get(line, 1);
		break;
	  }
	}/* for j < |table| */

	if(new_name != NULL) SGN_Node_set_name(node, new_name);
  }/* for i < # of nodes */
}
/* SGN_Network_replace_node_names() -------------------------------------------- */

/*!\brief Replace node alias names according to the name table.
 *
 * The name table is give as a \SGN_StrMatrix table where the first column
 * is the alias name to be replaced, and the second one to replace, i.e. the new 
 * alias name.
 *
 * The (primary) node names are not affected by this function.
 *
 * If you want to use arbitrary columns to replace aliases, use
 * SGN_Network_replace_node_aliases_cols() instead of this.
 * This is a wrapper function of SGN_Network_replace_node_aliases_cols().
 *
 *\param network   pointer to the \ref SGN_Network instance whose node
 *   alise names will be replaced.
 *\param table   pointer to the \ref SGN_StrMatrix instance which is
 *   the name table for the name replacement.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_replace_node_aliases(SGN_Network * network, const SGN_StrMatrix * table){
  SGN_Network_replace_node_aliases_cols(network, table, 0, 1);
}
/* SGN_Network_replace_node_aliases() ------------------------------------------ */

/*!\brief Replace node alias names according to the name table.
 *
 * The name table is give as a \SGN_StrMatrix table where the column
 * specified by \a oldcol argument
 * is the alias name to be replaced, and the column
 * by \a newcol to replace, i.e. the new alias name.
 *
 * The (primary) node names are not affected by this function.
 *
 *
 * Note that the column position starts from zero (0).
 *
 *\param network   pointer to the \ref SGN_Network instance whose node
 *   alise names will be replaced.
 *\param table   pointer to the \ref SGN_StrMatrix instance which is
 *   the name table for the name replacement.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_replace_node_aliases_cols(SGN_Network * network, 
									   const SGN_StrMatrix * table,
									   const int oldcol, const int newcol){
  int i, j;
  for(i = 0; i < SGN_Network_num_nodes(network); i++){
	SGN_Node * node = SGN_Network_get_node(network, i);
	const char * new_name = NULL;
	/* Find the node name. */
	for(j = 0; j < SGN_StrMatrix_lines(table); j++){
	  SGN_StrArray * line = SGN_StrMatrix_get_line(table, j);
	  if(strcmp(SGN_StrArray_get(line, oldcol), SGN_Node_alias(node)) == 0){
		/* found */
		new_name = SGN_StrArray_get(line, newcol);
		if(strlen(new_name) == 0) new_name = NULL;
		break;
	  }
	}/* for j < |table| */

	if(new_name != NULL) SGN_Node_set_alias(node, new_name);
  }/* for i < # of nodes */
}
/* SGN_Network_replace_node_aliases_cols() ------------------------------------- */

/*!\brief Constructs a subnetwork from the list of node indexes.
 *
 *
 *\param this   pointer to the \ref SGN_Network instance.
 *\param indexes  list of node indexes.
 *\return pointer to the \ref SGN_Network insance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Network * SGN_Network_subnetwork_index(const SGN_Network * this, SGN_IntArray * indexes){
  SGN_Network * ret = SGN_Network_new();
  int jj;
  int table[SGN_Network_num_nodes(this)];

  for(jj = 0; jj < SGN_Network_num_nodes(this); jj++){
	table[jj] = -1;
  }

  /* Add given nodes to the new network. */
  for(jj = 0; jj < SGN_IntArray_size(indexes); jj++){
	int j = SGN_IntArray_get(indexes, jj);
	SGN_Node * n = SGN_Network_get_node(this, j);
	SGN_Network_add_node(ret, SGN_Node_clone(n));
	table[j] = jj;
  }

  /* add edges */
  for(jj = 0; jj < SGN_IntArray_size(indexes); jj++){
	int j = SGN_IntArray_get(indexes, jj);
	const SGN_IntArray * parents = SGN_Network_get_parent_indexes(this, j);
	int k;
	SGN_Node * st = SGN_Network_get_node(ret, table[j]);
	for(k = 0; k < SGN_IntArray_size(parents); k++){
	  int kj = SGN_IntArray_get(parents, k);

	  if(table[kj] != -1){
		SGN_Node * ed = SGN_Network_get_node(ret, table[kj]);
		SGN_Edge * edge = SGN_Network_get_edge_index(this, kj, j);
		SGN_Edge * e = SGN_Edge_clone(edge);
		SGN_Edge_set(e, st, ed);
		SGN_Network_add_edge(ret, e);
	  }
	}
  }/* for j < |index| */

  return ret;
}
/* SGN_Network_subnetwork() ---------------------------------------------------- */

SGN_Network * SGN_Network_subnetwork_index_dist(const SGN_Network * this, 
										const SGN_IntArray * indexes, size_t dist){
  int p = SGN_Network_num_nodes(this);
  int table[p];
  SGN_Network * ret = SGN_Network_new();

  for(int j = 0; j < p; j++) table[j] = -1;

  /* Add nodes */
  for(int jj = 0; jj < SGN_IntArray_size(indexes); jj++){
	int j = SGN_IntArray_get(indexes, jj);
	SGN_Node * n = SGN_Network_get_node(this, j);
	SGN_Network_add_node(ret, SGN_Node_clone(n));
	table[j] = jj;
  }
  
  /* Visit all the ordered pairs */
  for(int jj = 0; jj < SGN_IntArray_size(indexes); jj++){
	for(int ii = 0; ii < SGN_IntArray_size(indexes); ii++){
	  SGN_Array * pathways;
	  int k;
	  /* Find pathways from ii to jj. */
	  pathways = SGN_Network_collect_paths_index(this, ii, jj, dist, 0);

	  /* Adds the collected pathways */
	  for(k = 0; k < SGN_Array_size(pathways); k++){
		SGN_IntArray * path = (SGN_IntArray *) SGN_Array_get(pathways, k);
		int l;
		int u = SGN_IntArray_get(path, 0);
		/* u is definitely already added in the network
		   because it is a start node. */
		for(l = 1; l < SGN_IntArray_size(path); l++){
		  int v = SGN_IntArray_get(path, l);
		  if(table[v] == -1){
			SGN_Network_add_node(ret, SGN_Node_clone(SGN_Network_get_node(this, v)));
			table[v] = SGN_Network_num_nodes(ret) - 1;
		  }
		  SGN_Network_add_edge_index(ret, table[u], table[v]);
		  u = v;
		}/* for l */
	  }/* for k < |pathways| */
	  SGN_Array_delete_allf(pathways, SGN_IntArray_delete);
	}/* for ii */
  }/* for jj */

  return ret;
}
/* SGN_Network_subnetwork_index_dist() ----------------------------------------- */

SGN_Network * SGN_Network_subnetwork_index_collect(const SGN_Network * this,
												   const SGN_IntArray * indexes,
												   int dist){
  int p = SGN_Network_num_nodes(this);
  int table[p];
  int visited[p];
  SGN_Network * ret = SGN_Network_new();
  int np = 0;
  SGN_IntArray * new_nodes = SGN_IntArray_new();
  SGN_IntArray * added_nodes = SGN_IntArray_new();
  SGN_IntArray * children = SGN_IntArray_new();

  if(dist < 0){
	fprintf(SIGN_ERR, "ERROR: SGN_Network_subnetwork_index_collect():"
			" Invalid distance specified (dist = %d)\n", dist);
	exit(1);
  }
  
  for(int j = 0; j < p; j++) { table[j] = -1; visited[j] = 0; }

  SGN_IntArray_copy(indexes, new_nodes);

  for(int d = 0; d < dist; d++){
	SGN_IntArray * tmp = new_nodes;
	new_nodes = added_nodes;
	added_nodes = tmp;
	SGN_IntArray_clear(new_nodes);

	//fprintf(SIGN_ERR, "DEBUG: d=%d, |added_nodes|=%zu\n", 
	//d, SGN_IntArray_size(added_nodes));

	for(int jj = 0; jj < SGN_IntArray_size(added_nodes); jj++){
	  int j = SGN_IntArray_get(added_nodes, jj);
	  if(visited[j]) continue;
	  visited[j] = 1;

	  if(table[j] == -1){
		SGN_Node * n = SGN_Network_get_node(this, j);
		SGN_Network_add_node(ret, SGN_Node_clone(n));
		table[j] = np++;
	  }

	  SGN_Node * node = SGN_Network_get_node(ret, table[j]);

	  /* Visit parents */
	  const SGN_IntArray * parents = SGN_Network_get_parent_indexes(this, j);
	  for(int k = 0; k < SGN_IntArray_size(parents); k++){
		int kj = SGN_IntArray_get(parents, k);
		if(visited[kj]) continue;
		if(table[kj] == -1){
		  SGN_Node * n = SGN_Network_get_node(this, kj);
		  SGN_Network_add_node(ret, SGN_Node_clone(n));
		  table[kj] = np++;
		}
		SGN_IntArray_add(new_nodes, kj);
		/* Add edge */
		SGN_Node * st = SGN_Network_get_node(ret, table[kj]);
		SGN_Edge * e = SGN_Edge_clone(SGN_Network_get_edge_index(this, kj, j));
		SGN_Edge_set(e, st, node);
		SGN_Network_add_edge(ret, e);
	  }
	  /* Visit children */
	  SGN_Network_copy_child_indexes(this, j, children);
	  for(int k = 0; k < SGN_IntArray_size(children); k++){
		int kj = SGN_IntArray_get(children, k);
		if(visited[kj]) continue;
		if(table[kj] == -1){
		  SGN_Node * n = SGN_Network_get_node(this, kj);
		  SGN_Network_add_node(ret, SGN_Node_clone(n));
		  table[kj] = np++;
		}
		SGN_IntArray_add(new_nodes, kj);
		/* Add edge */
		SGN_Node * ed = SGN_Network_get_node(ret, table[kj]);
		SGN_Edge * e = SGN_Edge_clone(SGN_Network_get_edge_index(this, j, kj));
		SGN_Edge_set(e, node, ed);
		SGN_Network_add_edge(ret, e);
	  }/* for k < |children| */
	}
  } /* for d < dist */
  return ret;
}
/* SGN_Network_subnetwork_index_collect() ================================== */

/*!\brief Collects pathways from \a x to \a y within distance \a dist.
 *
 * This collects all possible pathways within the specified distance.
 * If \a exact is set to 1, this collects pathways whose distances are
 * exactly \a dist.
 *
 *\return the \ref SGN_Array instance where each element is an \ref SGN_IntArray
 *   instance that represents a collected pathway.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_Array * SGN_Network_collect_paths_index(const SGN_Network * this,
									size_t x, size_t y, size_t dist, int exact){
  int p = SGN_Network_num_nodes(this);
  int flag[p]; /* index */
  int back[p]; /* where from */
  int j;
  int depth = 0;
  SGN_Array * paths = SGN_Array_new();

  for(j = 0; j < p; j++) flag[j] = -1;
  back[y] = -1;
 
  while(y != -1){
	flag[y]++;

	if(flag[y] == 0 && y == x && (! exact || (exact && depth == dist))){
	  SGN_IntArray * pw = SGN_IntArray_news(dist);
	  int xx = y;
	  while(xx != -1){
		SGN_IntArray_add(pw, xx);
		xx = back[xx];
	  }
	  SGN_Array_add(paths, pw);
	}

	if(depth == dist){
	  flag[y] = -1;
	  y = back[y];
	  depth--;
	  continue;
	}

	if(flag[y] < SGN_Network_num_parents(this, y)){
	  /* Go to the next parent */
	  int ny = SGN_Network_get_parent_index(this, y, flag[y]);
	  if(flag[ny] < 0){
		/* not visited. Go there */
		back[ny] = y;
		y = ny;
		depth++;
	  }
	} else {
	  /* Already visited all the parents for y */
	  /* Go back */
	  flag[y] = -1;
	  y = back[y];
	  depth--;
	}
  }/* while() */
  return paths;
}
/* SGN_Network_collect_paths_index() ------------------------------------------- */

/*!\brief Collects undirected pathways.
 *
 *
 *\param dir  non-zero - takes direction into account,  or 0 - does not.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
SGN_IntArray * SGN_Network_collect_udpaths_index(const SGN_Network * this, 
												 const SGN_IntArray * children,
												 int x, int y, int dist, 
												 int exact, int direc){
  int p = SGN_Network_num_nodes(this);
  int flag[p];
  int back[p];
  int dir[p];
  int j;
  int depth = 0;
  SGN_IntArray * ret = SGN_IntArray_new();

  for(j = 0; j < p; j++) flag[j] = -1;
  back[y] = -1;

  while(y != -1){
	flag[y]++;

	if(flag[y] == 0 && y == x && (! exact || (exact && depth == dist))){
	  /* found a path */
	  int xx = y;
	  int yy = back[xx];
	  while(yy != -1){
		if(dir[xx] == 1){
		  SGN_IntArray_add(ret, xx);
		  SGN_IntArray_add(ret, yy);
		} else {
		  SGN_IntArray_add(ret, yy);
		  SGN_IntArray_add(ret, xx);
		}
		xx = yy;
		yy = back[xx];
	  }
	}

	if(depth == dist){
	  flag[y] = -1;
	  y = back[y];
	  depth--;
	  continue;
	}

	if(flag[y] < SGN_Network_num_parents(this, y)){
	  /* Go to the next parent */
	  int ny = SGN_Network_get_parent_index(this, y, flag[y]);
	  if(flag[ny] < 0){
		/* not visited. Go ther */
		back[ny] = y;
		dir[ny] = 1;
		y = ny;
		depth++;
	  }
	} else if(direc == 0 &&
			  flag[y] - SGN_Network_num_parents(this, y) < SGN_IntArray_size(children + y)){
	  /* Go to the next children */
	  int ny = SGN_IntArray_get(children + y, 
							flag[y] - SGN_Network_num_parents(this, y));
	  if(flag[ny] < 0){
		/* not visited. Go there */
		back[ny] = y;
		dir[ny] = -1;
		y = ny;
		depth++;
	  }
	} else {
	  /* Visited all the parents and children for y */
	  flag[y] = -1;
	  y = back[y];
	  depth--;
	}
  }/* while() */

  return ret;
}
/* SGN_Network_collet_udpaths_index() ------------------------------------------ */

/*!\brief Sets the color of nodes.
 *
 *\param indexes  \ref SGN_IntArray containing the indexes of nodes to be
 *               colored.  If NULL is given, all the nodes in the
 *            network are colored.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_set_node_color(SGN_Network * this, const SGN_IntArray * indexes, 
							int r, int g, int b, int a){
  
  if(indexes != NULL){
	for(int i = 0; i < SGN_IntArray_size(indexes); i++){
	  SGN_Node * node = SGN_Network_get_node(this, SGN_IntArray_get(indexes, i));
	  SGN_Node_set_color(node, r, g, b, a);
	}
  } else {
	for(int i = 0; i < SGN_Network_num_nodes(this); i++){
	  SGN_Node * node = SGN_Network_get_node(this, i);
	  SGN_Node_set_color(node, r, g, b, a);
	}
  }
}
/* SGN_Network_set_color() ----------------------------------------------------- */

void SGN_Network_set_node_color_by_names(SGN_Network * this, SGN_StrArray * names,
									 int r, int g, int b, int a){
  int i;
  for(i = 0; i < SGN_StrArray_size(names); i++){
	SGN_Node * node = SGN_Network_find_node(this, SGN_StrArray_get(names, i));
	if(node != NULL) SGN_Node_set_color(node, r, g, b, a);
  }
}
/* SGN_Network_set_node_color_by_names() --------------------------------------- */

/*!\brief Sets the size of nodes in the \ref SGN_Network.
 *
 *\param indexes  \ref SGN_IntArray containing the indexes of nodes whose
 *                size will be set.  If NULL is given, the size of 
 *                all the nodes in the network is changed.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_set_node_size(SGN_Network * this, const SGN_IntArray * indexes,
						   int size){
  if(indexes != NULL){
	for(int i = 0; i < SGN_IntArray_size(indexes); i++){
	  SGN_Node * node = SGN_Network_get_node(this, SGN_IntArray_get(indexes, i));
	  double s = (SGN_Node_get_size(node) - size) / 2.0;
	  SGN_Node_set_size(node, size);
	  SGN_Node_move_position(node, s, s);
	}
  } else {
	for(int i = 0; i < SGN_Network_num_nodes(this); i++){
	  SGN_Node * node = SGN_Network_get_node(this, i);
	  double s = (SGN_Node_get_size(node) - size) / 2.0;
	  SGN_Node_set_size(node, size);
	  SGN_Node_move_position(node, s, s);
	}
  }
}
/* SGN_Network_set_node_size() ================================================= */

/*!\brief Sets the outline width of nodes in the \ref SGN_Network.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_set_node_line_width(SGN_Network * this, const SGN_IntArray * indexes,
								 double width){
  if(indexes != NULL){
	for(int i = 0; i < SGN_IntArray_size(indexes); i++){
	  SGN_Node * node = SGN_Network_get_node(this, SGN_IntArray_get(indexes, i));
	  SGN_Node_set_line_width(node, width);
	}
  } else {
	for(int i = 0; i < SGN_Network_num_nodes(this); i++){
	  SGN_Node * node = SGN_Network_get_node(this, i);
	  SGN_Node_set_line_width(node, width);
	}
  }
}
/* SGN_Network_set_node_line_width() =========================================== */

/*!\brief Sets the color of edges.
 *
 *\param indexes  \ref SGN_IntArray containing the indexes of edges to be
 *     colored.  If NULL is given, all the edges in the network are
 *     colored.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_set_edge_color(SGN_Network * this, const SGN_IntArray * indexes,
							int r, int g, int b, int a){
  if(indexes != NULL){
	for(int i = 0; i < SGN_IntArray_size(indexes); i++){
	  SGN_Edge * edge = SGN_Network_get_edge(this, SGN_IntArray_get(indexes, i));
	  SGN_Edge_set_color(edge, r, g, b, a);
	}
  } else {
	for(int i = 0; i < SGN_Network_num_edges(this); i++){
	  SGN_Edge * edge = SGN_Network_get_edge(this, i);
	  SGN_Edge_set_color(edge,r, g, b, a);
	}
  }
}
/* SGN_Network_set_edge_color() ================================================ */

/*!\brief Sets the visibility of the \ref CSMLObjects in the \ref SGN_Network.
 *
 *\param indexes \ref SGN_IntArray containing indexes of the total \ref CSMLObject's,
 *           or NULL to specify all the objects as targets.
 *
 *\sa SGN_Network_total_CSMLObjects().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_set_visible(SGN_Network * this, const SGN_IntArray * indexes, int visible){
  if(indexes != NULL){
	for(int i = 0; i < SGN_IntArray_size(indexes); i++){
	  CSMLObject * csmlobj = SGN_Network_get_total_CSMLObject(this,
														  SGN_IntArray_get(indexes, i));
	  CSMLObject_set_visible(csmlobj, visible);
	}
  } else {
	for(int i = 0; i < SGN_Network_total_CSMLObjects(this); i++){
	  CSMLObject * csmlobj = SGN_Network_get_total_CSMLObject(this, i);
	  CSMLObject_set_visible(csmlobj, visible);
	}
  }
}
/* SGN_Network_set_visible(); ================================================== */

SGN_Array * SGN_Network_connected_components(SGN_Network * this){
  int p = SGN_Network_num_nodes(this);
  SGN_IntArray * ar_a;
  const SGN_IntArray * parents;
  SGN_IntArray * children = SGN_IntArray_new();
  int added[p];
  int total = 0;
  SGN_Array * comps = SGN_Array_new();
  int ar_a_pos = 0, v, k;

  bzero(added, sizeof(int) * p);

  ar_a = SGN_IntArray_news(1);
  SGN_IntArray_add(ar_a, 0);
  added[0] = 1;
  total++;

  while(1){
	v = SGN_IntArray_get(ar_a, ar_a_pos++);

	parents = SGN_Network_get_parent_indexes(this, v);
	for(k = 0; k < SGN_IntArray_size(parents); k++){
	  int u = SGN_IntArray_get(parents, k);
	  if(added[u]) continue;
	  SGN_IntArray_add(ar_a, u);
	  added[u] = 1;
	  total++;
	}
	SGN_Network_copy_child_indexes(this, v, children);
	for(k = 0; k < SGN_IntArray_size(children); k++){
	  int u = SGN_IntArray_get(children, k);
	  if(added[u]) continue;
	  SGN_IntArray_add(ar_a, u);
	  added[u] = 1;
	  total++;
	}

	if(ar_a_pos == SGN_IntArray_size(ar_a)){
	  /* Start a new component if available */
	  SGN_Array_add(comps, ar_a);

	  if(total == p) break;

	  for(k = 0; k < p; k++) if(added[k] == 0) break;

	  ar_a = SGN_IntArray_news(1);
	  ar_a_pos = 0;
	  SGN_IntArray_add(ar_a, k);
	  added[k] = 1;
	  total++;
	}
  }/* while(1); */
  
  SGN_IntArray_delete(children);
  return comps;
}
/* SGN_Network_connected_components() ------------------------------------------ */

/*!\brief Filters edges by a condition.
 *
 * This leaves edges whose edge properties satisfy the specified condition.
 *
 *\param op  type specific operator to compare values.  Do not specify
 *    general operators.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_filter(SGN_Network * network, 
					   const char * key, SGN_Type type, 
					   SGN_Type_Operator op, void * value){
  for(size_t i = 0; i < SGN_Network_num_edges(network); i++){
	SGN_Edge * edge = SGN_Network_get_edge(network, i);
	size_t idx = SGN_Edge_find_property_key(edge, key);
	SGN_Type ty;
	void * val; /* Edge property value */
	
	if(idx == SGN_Edge_ERROR_VALUE){
	  SGN_Network_remove_edge_index(network, i);
	  i--;
	  continue;
	}

	ty = SGN_Edge_get_property_type(edge, idx);
	if(ty != type){
	  SGN_Network_remove_edge_index(network, i);
	  i--;
	  continue;
	}
	
	val = SGN_Edge_get_property_value(edge, idx);
	if(!SGN_Type_compare_op(ty, op, val, value)){
	  SGN_Network_remove_edge_index(network, i);
	  i--;
	  continue;
	}
  }/* for i < |network| */
}
/* SGN_Network_filter() -------------------------------------------------------- */

typedef struct {
  int index;
  int value;
  int value2;
} SGN_Network_int_t;

int SGN_Network_int_t_cmp(const void * x, const void * y){
  return ((const SGN_Network_int_t *) y)->value - ((const SGN_Network_int_t *) x)->value;
}

/*!\brief Sorts nodes in the network with repsect to the number of
 *     children.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_get_hub_list(SGN_Network * this, int * list){
  int p = SGN_Network_num_nodes(this);
  SGN_Network_int_t iar[p];
  for(int j = 0; j < p; j++){
	iar[j].index = j;
	iar[j].value = SGN_Network_num_children(this, j);
	iar[j].value2 = SGN_Network_num_parents(this, j);
  }
  qsort(iar, p, sizeof(SGN_Network_int_t), SGN_Network_int_t_cmp);

  for(int j = 0; j < p; j++){
	list[j] = iar[j].index;
  }
}
/* SGN_Network_get_hub_list() ================================================== */

/*!\brief Calculates the hub ranking and adds it as node properties.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_gen_prop_hub_rank(SGN_Network * this){
  int p = SGN_Network_num_nodes(this);

  SGN_Network_int_t iar[p];
  for(int j = 0; j < p; j++){
	iar[j].index = j;
	iar[j].value = SGN_Network_num_children(this, j);
	iar[j].value2 = SGN_Network_num_parents(this, j);
  }
  qsort(iar, p, sizeof(SGN_Network_int_t), SGN_Network_int_t_cmp);

  int prev = -1;
  int rank = -1;
  for(int j = 0; j < p; j++){
	int jj = iar[j].index;
	//SGN_Node * node = SGN_Network_get_node(this, jj);
	CSMLObject * csmlobj = SGN_Node_get_CSMLObject(SGN_Network_get_node(this, jj));

	if(iar[j].value != prev){
	  rank = j + 1;
	  prev = iar[j].value;
	}

	//SGN_Node_set_property_int(node, "hub.rank", pos);
	CSMLObject_set_property_int(csmlobj, "hubness", iar[j].value);
	CSMLObject_set_property_int(csmlobj, "hubness.rank", rank);
  }
}
/* SGN_Network_gen_prop_hub_rank() ============================================= */

void SGN_Network_gen_prop_hubness(SGN_Network * this){
  SGN_Network_gen_prop_hub_rank(this);
}

/*!\brief Applies the circle layout.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_layout_circle(SGN_Network * this){
  double cx = 500.0;
  double cy = 400.0;
  double theta = 3.14159265 / 2.0;
  double r = 350.0;
  double w = 1.0;
  double n = SGN_Network_num_nodes(this);
  int j;
  int max_deg = 0;
  int p = SGN_Network_num_nodes(this);
  int deg[p];
  int max_size = 40;
  int min_size = 10;

  for(j = 0; j < SGN_Network_num_nodes(this); j++){
	SGN_Node * node = SGN_Network_get_node(this, j);
	double ntheta = theta + ((2.0 * 3.1415926535 / n) * (double) j);
	double x = w * r * cos(ntheta) + cx;
	double y = r * sin(ntheta) + cy;

	SGN_Node_set_position(node, x, y);
	//deg[j] = SGN_Network_node_degree(this, j);
	deg[j] = SGN_Network_num_children(this, j);
	if(deg[j] > max_deg) max_deg = deg[j];
  }

  for(j = 0; j < p; j++){
	SGN_Node * node = SGN_Network_get_node(this,j);

	SGN_Node_set_size(node, 
				  (deg[j] / (double) max_deg) * (max_size - min_size) +
				  min_size);
  }
}
/* SGN_Network_layout_circle() ------------------------------------------------- */

/*!\brief Layouts the network randomly
 *
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_layout_random(SGN_Network * network, 
						   const SGN_Array * nodes, const SGN_KeyValues * args){

  const SGN_Array * targets;
  double x1 = 10.0, y1 = 10.0, x2 = 640.0, y2 = 480.0;
  double width, height;
  SGN_Random_dSFMT dsfmt;
  unsigned int seed = 0;
  SGN_KeyValueSet argset[] = {
	{"x1",   SGN_Type_DOUBLE, &x1},
	{"x2",   SGN_Type_DOUBLE, &x2},
	{"y1",   SGN_Type_DOUBLE, &y1},
	{"y2",   SGN_Type_DOUBLE, &y2},
	{"seed", SGN_Type_UINT, &seed},
	{NULL},
  };
  int r = SGN_KeyValues_set_values_kv(args, argset);
  if(r >= 0){
	fprintf(SIGN_ERR,
			"ERROR: SGN_Network_layout_random(): unknown argument specified: %s\n",
			SGN_KeyValues_get_key(args, r));
	exit(1);
  }

  SGN_Random_dSFMT_init(&dsfmt, seed);
  
  if(x2 < x1){
	double tmp = x2; 
	x2 = x1; x1 = tmp;
  }
  if(y2 < y1){
	double tmp = y2;
	y2 = y1; y1 = tmp;
  }
  width = x2 - x1;
  height = y2 - y1;
  
  if(nodes == NULL){
	targets = network->nodes;
  } else {
	targets = nodes;
  }

  for(int i = 0; i < SGN_Array_size(targets); i++){
	double x, y;
	x = SGN_Random_dSFMT_close_open(&dsfmt) * width + x1;
	y = SGN_Random_dSFMT_close_open(&dsfmt) * height + y1;
	SGN_Node * node = (SGN_Node *) SGN_Array_get(targets, i);
	SGN_Node_set_position(node, x, y);
  }
}
/* SGN_Network_layout_random() ================================================= */

void SGN_Network_layout_fit(SGN_Network * network, const SGN_KeyValues * args){
  double x1 = 50.0, y1 = 50.0, x2 = 900.0, y2 = 600.0, xoffset = 0, yoffset = 0;
  SGN_KeyValueSet argset[] = {
	{"x1", SGN_Type_DOUBLE, &x1},
	{"x2", SGN_Type_DOUBLE, &x2},
	{"y1", SGN_Type_DOUBLE, &y1},
	{"y2", SGN_Type_DOUBLE, &y2},
	{"x",  SGN_Type_DOUBLE, &xoffset},
	{"y",  SGN_Type_DOUBLE, &yoffset},
	{NULL},
  };
  int r = SGN_KeyValues_set_values_kv(args, argset);
  if(r >= 0){
	fprintf(SIGN_ERR,
			"ERROR: SGN_Network_layout_fit(): unknown argument specified: %s\n",
			SGN_KeyValues_get_key(args, r));
	exit(1);
  }

  if(x2 < x1){
	double tmp = x2; 
	x2 = x1; x1 = tmp;
  }
  if(y2 < y1){
	double tmp = y2;
	y2 = y1; y1 = tmp;
  }

  double x_min = DBL_MAX;
  double x_max = - DBL_MAX;
  double y_min = DBL_MAX;
  double y_max = - DBL_MAX;

  for(int i = 0; i < SGN_Network_num_nodes(network); i++){
	SGN_Node * node = SGN_Network_get_node(network, i);

	double x = SGN_Node_get_position_x(node);
	double y = SGN_Node_get_position_y(node);

	if(x < x_min) x_min = x;
	if(x_max < x) x_max = x;
	if(y < y_min) y_min = y;
	if(y_max < y) y_max = y;
  }

  double rx = (x2 - x1) / (x_max - x_min);
  double ry = (y2 - y1) / (y_max - y_min);

  for(int i = 0; i < SGN_Network_num_nodes(network); i++){
	SGN_Node * node = SGN_Network_get_node(network, i);
	
	double x = SGN_Node_get_position_x(node);
	double y = SGN_Node_get_position_y(node);

	SGN_Node_set_position(node,
						  (x - x_min) * rx + x1 + xoffset,
						  (y - y_min) * ry + y1 + yoffset);
  }
}
/* SGN_Network_layout_fit() ==================================================== */

/*!\brief Changes the node size according to the hubness.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_layout_hub_size(SGN_Network * this, const SGN_KeyValues * args){
  int p = SGN_Network_num_nodes(this);
  int deg[p];
  int max_deg = 0;
  int max_size = 40;
  int min_size = 10;
  SGN_Network_int_t iar[p];

  for(int j = 0; j < p; j++){
	deg[j] = SGN_Network_num_children(this, j);
	if(deg[j] > max_deg) max_deg = deg[j];

	iar[j].index = j;
	iar[j].value = SGN_Network_num_children(this, j);
	iar[j].value2 = SGN_Network_num_parents(this, j);
  }
  qsort(iar, p, sizeof(SGN_Network_int_t), SGN_Network_int_t_cmp);

  for(int j = 0; j < p; j++){
	int jj = iar[j].index;
	SGN_Node * node = SGN_Network_get_node(this, jj);
	SGN_Node_set_size(node,
				  (deg[jj] / (double) max_deg) * (max_size - min_size) +
				  min_size);
	if(j + 1 > 1000){
	  SGN_Node_set_depth(node, 1000);
	} else {
	  SGN_Node_set_depth(node, j + 1);
	}
  }

  for(int i = 0; i < SGN_Network_num_edges(this); i++){
	int depth = i + p + 1;
	SGN_Edge * edge = SGN_Network_get_edge(this, i);

	if(depth > 1000) depth = 1000;
	SGN_Edge_set_depth(edge, depth);
  }
}
/* SGN_Network_layout_hub_size() =============================================== */

typedef struct {
  int index;
  double value;
} SGN_Network_double_t;

int SGN_Network_double_t_cmp(const void * x, const void * y){
  return ((const SGN_Network_double_t *) y)->value - ((const SGN_Network_double_t *) x)->value;
}

void SGN_Network_layout_prop_node_size(SGN_Network * this, const char * prop,
									   const SGN_KeyValues * args){
  double min_size = 10.0;
  double max_size = 40.0;
  double max_value = 0.0;
  double min_value = DBL_MAX;
  SGN_Network_double_t dar[SGN_Network_num_nodes(this)];

  for(int j = 0; j < SGN_Network_num_nodes(this); j++){
	SGN_Node * node = SGN_Network_get_node(this, j);
	SGN_Object * obj = CSMLObject_get_property(SGN_Node_get_CSMLObject(node), prop);
	double value;

	if(obj == NULL){
	  /* not found */
	  value = - DBL_MAX;
	} else {
	  switch(SGN_Object_type(obj)){
	  case SGN_Type_INT: value = SGN_Object_int_value(obj, 0); break;
	  case SGN_Type_DOUBLE: value = SGN_Object_double_value(obj, 0.0); break;
	  default:
		fprintf(SIGN_ERR, "FATAL ERROR: SGN_Network_layout_prop_node_size(): "
				"Unsupported type: %s\n", SGN_Object_type_name(obj));
		exit(1);
	  }

	  if(value > max_value) max_value = value;
	  if(min_value < value) min_value = value;
	}
	dar[j].value = value;
	dar[j].index = j;
  }
  qsort(dar, SGN_Network_num_nodes(this), 
		sizeof(SGN_Network_double_t), SGN_Network_double_t_cmp);
  
  for(int j = 0; j < SGN_Network_num_nodes(this); j++){
	int jj = dar[j].index;
	SGN_Node * node = SGN_Network_get_node(this, jj);
	if(dar[j].value == - DBL_MAX){
	  SGN_Node_set_size(node, min_size);
	} else {
	  SGN_Node_set_size(node, (dar[j].value / max_value) * (max_size - min_size) + min_size);
	}

	SGN_Node_set_depth(node, j + 1);
  }

  for(int i = 0; i < SGN_Network_num_edges(this); i++){
	int depth = i + SGN_Network_num_nodes(this) + 1;
	SGN_Edge * edge = SGN_Network_get_edge(this, i);

	if(depth > 1000) depth = 1000;
	SGN_Edge_set_depth(edge, depth);
  }
}
/* SGN_Network_layout_prop_node_size() ========================================= */


/*!\brief Removes all the edge properties.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_clear_all_edge_properties(SGN_Network * this){
  int i;
  for(i = 0; i < SGN_Array_size(this->edges); i++){
	SGN_Edge * edge = (SGN_Edge *) SGN_Array_get(this->edges, i);
	SGN_Edge_clear_properties(edge);
  }
}
/* SGN_Network_clear_all_edge_properties() ------------------------------------- */

/*!\brief Adds the network property.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_add_property(SGN_Network * this, 
						  const char * key, SGN_Type type, void * value){
  if(this->prop == NULL) this->prop = SGN_KeyValues_new();
  SGN_KeyValues_add(this->prop, key, type, value);
}
/* SGN_Network_add_property() ================================================== */

/*!\brief Adds the network property in string.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_add_property_int(SGN_Network * this, const char * key, int value){
  if(this->prop == NULL) this->prop = SGN_KeyValues_new();
  SGN_KeyValues_put_int(this->prop, key, value);
}
/* SGN_Network_add_property_int() ============================================== */

int SGN_Network_get_property_int(const SGN_Network * this, const char * key, int def){
  if(this->prop == NULL) return def;
  return SGN_KeyValues_get_int_def(this->prop, key, def);
}
/* SGN_Network_get_property_int() ============================================== */

double SGN_Network_get_property_double(const SGN_Network * this, const char * key,
								   double def){
  if(this->prop == NULL) return def;
  return SGN_KeyValues_get_double(this->prop, key, def);
}
/* SGN_Network_get_property_double() =========================================== */

/*!\brief Adds the properties in a \ref SGN_KeyValues instance to this
 * \ref SGN_Network.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_add_properties_KeyValues(SGN_Network * this, const SGN_KeyValues * kv){
  if(this->prop == NULL) this->prop = SGN_KeyValues_new();
  SGN_KeyValues_add_KeyValues_nr(this->prop, kv);
}
/* SGN_Network_add_properties_KeyValues() ====================================== */

/*!\brief Compares two \ref SGN_Network instances and return 1 if their
 *    structures are identical.
 *
 * This compares only the structures not node names etc..
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
int SGN_Network_compare_structure(const SGN_Network * net1, const SGN_Network * net2){
  if(SGN_Array_size(net1->nodes) != SGN_Array_size(net2->nodes)) return 0;

  if(SGN_Array_size(net1->edges) != SGN_Array_size(net2->edges)) return 0;

  SGN_IntArray * parents1 = SGN_IntArray_new();
  SGN_IntArray * parents2 = SGN_IntArray_new();

  int ret = 1;
  
  for(int j = 0; j < SGN_Array_size(net1->nodes); j++){
	SGN_IntArray_copy(SGN_Network_get_parent_indexes(net1, j), parents1);
	SGN_IntArray_copy(SGN_Network_get_parent_indexes(net2, j), parents2);

	SGN_IntArray_sort(parents1);
	SGN_IntArray_sort(parents2);

	if(SGN_IntArray_equal(parents1, parents2) == 0){
	  ret = 0;
	  break;
	}
  }/* for i < nodes */

  SGN_IntArray_delete(parents1);
  SGN_IntArray_delete(parents2);
  
  return ret;
}
/* SGN_Network_compare_structure() ============================================= */

/*!\brief Returns the \ref SGN_Edge property value in double.
 *
 *\return - DBL_MAX if (\a i, \a j) edge does not exist in the network.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
double SGN_Network_get_edge_property_double(SGN_Network * network,
										int i, int j, char * key){
  SGN_Edge * edge = SGN_Network_get_edge_index(network, i, j);
  if(edge == NULL) return - DBL_MAX;

  double v = SGN_Edge_get_property_double(edge, key, 0.0);

  return v;
}
/* SGN_Network_get_edge_property_double() ====================================== */

SGN_Network * SGN_Network_merge_union(SGN_Network ** networks, int num){
  SGN_Network * network = SGN_Network_new();
  for(int i = 0; i < num; i++){
	SGN_Network * net = networks[i];
	int c = 0;

	/* SGN_Node */
	for(int j = 0; j < SGN_Network_num_nodes(net); j++){
	  SGN_Node * node = SGN_Network_get_node(net, j);
	  
	  SGN_Node * node2 = SGN_Network_find_node(network, SGN_Node_name(node));
	  if(node2 == NULL){
		SGN_Node * new_node = SGN_Node_new(SGN_Node_name(node),
								   SGN_Node_alias(node));
		SGN_Network_add_node(network, new_node);
		SGN_Node_set_position(new_node,
						  SGN_Node_get_position_x(node),
						  SGN_Node_get_position_y(node));
		SGN_Node_set_size(new_node, SGN_Node_get_size(node));
		//fprintf(SIGN_ERR, "size: %f\n", SGN_Node_get_size(node));
		c++;
	  }
	}
	fprintf(SIGN_ERR, "SGN_Network:  added %d nodes.\n", c);

	/* SGN_Edge */
	for(int j = 0; j < SGN_Network_num_edges(net); j++){
	  SGN_Edge * edge = SGN_Network_get_edge(net, j);
	  SGN_Node * pa_node = SGN_Edge_get_parent(edge);
	  SGN_Node * ch_node = SGN_Edge_get_child(edge);

	  SGN_Edge * e = SGN_Network_find_edge_name(network,
										SGN_Node_name(pa_node),
										SGN_Node_name(ch_node));
	  if(e == NULL){
		SGN_Node * pa2 = SGN_Network_find_node(network, SGN_Node_name(pa_node));
		SGN_Node * ch2 = SGN_Network_find_node(network, SGN_Node_name(ch_node));
		SGN_Edge * new_edge = SGN_Edge_new(pa2, ch2);
		SGN_Network_add_edge(network, new_edge);
	  }
	}/* for j < Num. edges */
	  
  }/* for i < num */

  return network;
}
/* SGN_Network_merge_union() =================================================== */

/*!\brief Calculates the betweenness for all the nodes in the \ref SGN_Network.
 *
 * This routine calculates the betweenness of all the nodes
 * by Brandes algorithm.
 *
 * Brandes U. (2001). <I>J. Mathematical Sociology</I> <B>25</B>(2), 163-177.
 *
 *\param this  pointer to the SGN_Network instance of whose betweenness is
 *  calculated.
 *\param cb  pointer to the double array whose size is equal to the number of
 *  the nodes in the network.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_calc_betweenness(const SGN_Network * this, double * cb){
  int p = SGN_Network_num_nodes(this);
  bzero(cb, sizeof(double) * p);

  SGN_IntQueue * Q = SGN_IntQueue_new();
  SGN_IntArray * S = SGN_IntArray_new();
  SGN_IntArray * neighbor = SGN_IntArray_new();
  SGN_IntArray * P = SGN_IntArray_new_ar(p, 0);

  for(int s = 0; s < p; s++){
	//fprintf(SIGN_ERR, "==== for s=%d ====\n", s);
	SGN_IntArray_clear(S);  /* empty stack */
	for(int i = 0; i < p; i++) SGN_IntArray_clear(P + i); /* empty list */
	
	int sigma[p];
	bzero(sigma, sizeof(int) * p);
	sigma[s] = 1;

	int d[p];
	for(int i = 0; i < p; i++) d[i] = -1;
	d[s] = 0;

	SGN_IntQueue_clear(Q);      /* empty queue. */
	SGN_IntQueue_enqueue(Q, s); /* enqueue s -> Q */
	//fprintf(SIGN_ERR, "Enqueue s=%d; Q={", s);
	//SGN_IntQueue_fprint(Q, SIGN_ERR);
	//fprintf(SIGN_ERR, "}\n");

	while(!SGN_IntQueue_is_empty(Q)){
	  int v;
	  SGN_IntQueue_dequeue(Q, &v);
	  SGN_IntArray_push(S, v);
	  //fprintf(SIGN_ERR, "  Dequeue v=%d; Q={", v);
	  //SGN_IntQueue_fprint(Q, SIGN_ERR);
	  //fprintf(SIGN_ERR, "}\n");

	  /* foreach neighbor w of v do */
	  //SGN_Network_get_neighbor(this, v, neighbor);
	  SGN_Network_copy_child_indexes(this, v, neighbor);
	  //fprintf(SIGN_ERR, "Neighbor of v=%d: ", v);
	  //SGN_IntArray_fprint(neighbor, SIGN_ERR);
	  //fprintf(SIGN_ERR, "\n");
	  for(int i = 0; i < SGN_IntArray_size(neighbor); i++){
		int w = SGN_IntArray_get(neighbor, i);

		if(d[w] < 0){
		  SGN_IntQueue_enqueue(Q, w);
		  d[w] = d[v] + 1;
		  //fprintf(SIGN_ERR, "  Enqueue w=%d; Q={", w);
		  //SGN_IntQueue_fprint(Q, SIGN_ERR);
		  //fprintf(SIGN_ERR, "}\n");
		}

		if(d[w] == d[v] + 1){
		  sigma[w] += sigma[v];
		  /* append v -> P[w] */
		  SGN_IntArray_add(P + w, v);
		}
	  }/* foreach neighbor w of v */
	}/* while queue is not empty. */

	double delta[p];
	bzero(delta, sizeof(double) * p);
	
	while(SGN_IntArray_size(S) != 0){
	  int w = SGN_IntArray_pop(S);
	  for(int i = 0; i < SGN_IntArray_size(P + w); i++){
		int v = SGN_IntArray_get(P + w, i);
		delta[v] += (((double) sigma[v]) / ((double) sigma[w])) * (1.0 + delta[w]);
	  }
	  if(w != s) cb[w] += delta[w];
	}/* while S is not empty */
  }/* for s < p */

  SGN_IntArray_delete_ar(P, p);
  SGN_IntArray_delete(neighbor);
  SGN_IntArray_delete(S);
  SGN_IntQueue_delete(Q);
}
/* SGN_Network_calc_betweenness() ============================================== */

/*!\brief Calculates and sets betweeness of nodes in the \ref SGN_Network.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_gen_prop_betweenness(SGN_Network * this){
  int p = SGN_Network_num_nodes(this);
  double cb[p];
  SGN_Network_double_t dar[SGN_Network_num_nodes(this)];

  SGN_Network_calc_betweenness(this, cb);
  for(int i = 0; i < p; i++){
	//SGN_Node * node = SGN_Network_get_node(this, i);
	CSMLObject * csmlobj = SGN_Node_get_CSMLObject(SGN_Network_get_node(this, i));
	CSMLObject_set_property_double(csmlobj, "betweenness", cb[i]);
	dar[i].index = i;
	dar[i].value = cb[i];
  }

  qsort(dar, SGN_Network_num_nodes(this), 
		sizeof(SGN_Network_double_t), SGN_Network_double_t_cmp);

  double prev = - DBL_MAX;
  int rank = -1;

  for(int i = 0; i < p; i++){
	int ii = dar[i].index;
	//SGN_Node * node = SGN_Network_get_node(this, ii);
	CSMLObject * csmlobj = SGN_Node_get_CSMLObject(SGN_Network_get_node(this, ii));
	if(dar[i].value != prev){
	  rank = i + 1;
	  prev = dar[i].value;
	}
	CSMLObject_set_property_int(csmlobj, "betweenness.rank", rank);

	//if(i < 10){
	//	  fprintf(SIGN_ERR, "SGN_Network: bt=%f\n", dar[i].value);
	  //CSMLObject_fprint_properties(csmlobj, SIGN_ERR);
	  //fprintf(SIGN_ERR, "\n");
	//}
  }
}
/* SGN_Network_gen_prop_betweenness() ========================================== */

int SGN_Network_move_front_cmp_old(const void * a, const void * b){
  const SGN_Object * aa = (const SGN_Object *) *((const void **) a);
  const SGN_Object * bb = (const SGN_Object *) *((const void **) b);
  int av = INT_MAX;
  int bv = INT_MAX;

  if(SGN_Object_type(aa) == SGN_Type_NODE){
	av = SGN_Node_get_depth((const SGN_Node *) SGN_Object_pointer(aa));
  } else if(SGN_Object_type(aa) == SGN_Type_EDGE){
	av = SGN_Edge_get_depth((const SGN_Edge *) SGN_Object_pointer(aa));
  }
  if(SGN_Object_type(bb) == SGN_Type_NODE){
	bv = SGN_Node_get_depth((const SGN_Node *) SGN_Object_pointer(bb));
  } else if(SGN_Object_type(aa) == SGN_Type_EDGE){
	bv = SGN_Edge_get_depth((const SGN_Edge *) SGN_Object_pointer(bb));
  }
  return av - bv;
}
/* SGN_Network_move_front_cmp() ================================================ */

//DEBUG
//#include "math/SGN_Matrix.h"

/** Calculates the closeness centrality of all the nodes.
 *
 * This calculates the closeness centrality of all the nodes
 * by the Floyd-Warshall algorithm.
 *
 *
 * This takes direction of edges into account.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_calc_closeness(const SGN_Network * network, double * cn){
  int n = SGN_Network_num_nodes(network);
  double * d = malloc(sizeof(double) * n * n);
  double * p = malloc(sizeof(double) * n * n);
  for(int i = 0; i < n; i++){
	for(int j = 0; j < n; j++){
	  if(i == j){
		d[i + j * n] = 0;
	  } else if(SGN_Network_check_edge_index(network, i, j)){
		d[i + j * n] = 1;
	  } else {
		d[i + j * n] = INFINITY;
	  }
	}
  }
  for(int k = 0; k < n; k++){
	for(int i = 0; i < n; i++){
	  for(int j = 0; j < n; j++){
		if(i == j) continue;

		if(d[i + j * n] > d[i + k * n] + d[k + j * n]){
		  p[i + j * n] = d[i + k * n] + d[k + j * n];
		} else {
		  p[i + j * n] = d[i + j * n];
		}
	  }
	}
	double * tmp = p;
	p = d;
	d = tmp;
	//fprintf(SIGN_ERR, "DEBUG: k=%d\n", k);
	//SGN_Matrix_print(SIGN_ERR, d, n, n); // DEBUG
  }
  for(int i = 0; i < n; i++){
	double sum = 0.0;
	for(int j = 0; j < n; j++){
	  if(i == j){
		continue;
	  } else if(!isinf(d[i + j * n])){
		sum += 1.0 / d[i + j * n];
	  }
	}
	cn[i] = sum / (double) n;
  }
  free(d);
  free(p);
}
/* SGN_Network_calc_closeness() ================================================ */

/*****************************************************************************/
/*  CSMLObject related routines                                              */
/*****************************************************************************/

/*!\brief Adds an \ref CSMLObject instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_add_CSMLObject(SGN_Network * this, CSMLObject * obj){
  if(this->objects == NULL) this->objects = SGN_Array_new();

  SGN_Array_add(this->objects, obj);
}
/* SGN_Network_add_CSMLObject() =============================================== */

/*!\brief Returns the number of \ref CSMLObject instances.
 *
 * This returns the number of \ref CSMLObject instances
 *  in this \ref SGN_Network except for the \ref SGN_Node and \ref SGN_Edge instances.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
size_t  SGN_Network_num_CSMLObjects(const SGN_Network * this){
  if(this->objects == NULL){
	return 0;
  } else {
	return SGN_Array_size(this->objects);
  }
}
/* SGN_Network_num_CSMLObjects() ============================================== */

CSMLObject * SGN_Network_get_CSMLObject(const SGN_Network * this, int index){
  if(this->objects == NULL){
	return NULL;
  } else {
	return (CSMLObject *) SGN_Array_get(this->objects, index);
  }
}
/* SGN_Network_get_CSMLObject() ================================================ */

void SGN_Network_clear_CSMLObject(SGN_Network * network){
  if(network->objects != NULL){
	for(int i = 0; i < SGN_Array_size(network->objects); i++){
	  CSMLObject * obj = (CSMLObject *) SGN_Array_get(network->objects, i);
	  CSMLObject_delete(obj);
	}
	SGN_Array_clear(network->objects);
  }
}
/* SGN_Network_clear_CSMLObject() ============================================== */

/*!\brief Returns the total number of the \ref CSMLObject instances in this
 *  \ref SGN_Network instance.
 *
 * This returns the total number of objects that can be handled as a
 * \ref CSMLObject instance.
 *
 * If you want to obtain the number of the \ref CSMLObject instances
 * other than \ref SGN_Node and \ref SGN_Edge instances, use SGN_Network_num_CSMLObjects().
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
size_t SGN_Network_total_CSMLObjects(const SGN_Network * this){
  return SGN_Network_num_nodes(this) 
	+ SGN_Network_num_edges(this) + SGN_Network_num_CSMLObjects(this);
}
/* SGN_Network_total_CSMLObjects() ============================================= */

/*!\brief Returns the \ref CSMLObject from this \ref SGN_Network.
 *
 * This does not distinguish whether it is \ref SGN_Node, \ref SGN_Edge or other
 * \ref CSMLObject instances.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
CSMLObject * SGN_Network_get_total_CSMLObject(const SGN_Network * this, size_t index){
  if(index < SGN_Network_num_nodes(this)){
	SGN_Node * node = SGN_Network_get_node(this, index);
	return SGN_Node_get_CSMLObject(node);
  } else {
	index -= SGN_Network_num_nodes(this);
	if(index  < SGN_Network_num_edges(this)){
	  SGN_Edge * edge = SGN_Network_get_edge(this, index);
	  return SGN_Edge_get_CSMLObject(edge);
	} else {
	  index -= SGN_Network_num_edges(this);
	  return SGN_Network_get_CSMLObject(this, index);
	}
  }
}
/* SGN_Network_get_total_CSMLObject() ========================================== */


int SGN_Network_move_front_cmp(const void * a, const void * b){
  const SGN_Object * aa = (const SGN_Object *) *((const void **) a);
  const SGN_Object * bb = (const SGN_Object *) *((const void **) b);

  int av = INT_MAX;
  int bv = INT_MAX;
  if(SGN_Object_type(aa) == SGN_Type_CSMLObject){
	av = CSMLObject_get_depth((CSMLObject *) SGN_Object_ptr(aa));
  }
  if(SGN_Object_type(bb) == SGN_Type_CSMLObject){
	bv = CSMLObject_get_depth((CSMLObject *) SGN_Object_ptr(bb));
  }
  return av - bv;
}
/* SGN_Network_move_front_cmp() ================================================ */

/*!\brief Moves \CSMLObject instances in the \ref SGN_Network to the front.
 *
 *\param objects  \ref SGN_ObjArray containing \ref CSMLObject instances.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_move_front(SGN_Network * network, SGN_ObjArray * objects){
  int size = (int) SGN_ObjArray_size(objects);

  /* Sort SGN_ObjArray */
  SGN_ObjArray_sort(objects, SGN_Network_move_front_cmp);

  /* Move others back. */
  for(size_t i = 0; i < SGN_Network_total_CSMLObjects(network); i++){
	CSMLObject * obj = SGN_Network_get_total_CSMLObject(network, i);
	int depth = CSMLObject_get_depth(obj);
	depth += size;
	if(depth > SIGN_MAX_DEPTH) depth = SIGN_MAX_DEPTH;
	CSMLObject_set_depth(obj, depth);
  }

  /* Move objects front */
  for(size_t i = 0; i < SGN_ObjArray_size(objects); i++){
	SGN_Object * obj = SGN_ObjArray_get(objects, i);
	if(SGN_Object_type(obj) == SGN_Type_CSMLObject){
	  CSMLObject_set_depth((CSMLObject *) SGN_Object_ptr(obj), i + 1);
	}
  }
}
/* SGN_Network_move_front() ==================================================== */

void SGN_Network_check_validity(const SGN_Network * this){
  for(int i = 0; i < SGN_Network_num_nodes(this); i++){
	SGN_Node * node = SGN_Network_get_node(this, i);
	CSMLObject * csmlobj = SGN_Node_get_CSMLObject(node);
	CSMLObject_check_validity(csmlobj);
  }
  fprintf(SIGN_ERR, "SGN_Network_check_validity(): Validated.\n");
}
/* SGN_Network_check_validity() ================================================ */

/* ========================================================================= */
/*  MPI Functions                                                            */
/* ========================================================================= */

#ifdef USE_MPI
/*\brief
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Network */
void SGN_Network_MPI_Bcast(SGN_Network ** network, int root, MPI_Comm comm){
  int rank;

  MPI_Comm_rank(comm, &rank);
  if(rank != root) *network = SGN_Network_new();
}
/* SGN_Network_MPI_Bcast() ---------------------------------------------------- */
#endif /* USE_MPI */

/* End of file.  Copyright (C) 2008  HGC, IMS, The University of Tokyo       */
