/*
  SGN_Network.h : SGN_Network class
  
  Copyright (C) 2008-2011  SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
 */

#ifndef __SGN_NETWORK_H
#define __SGN_NETWORK_H

#include "SGN_StrArray.h"
#include "SGN_IntArray.h"
#include "SGN_StrMatrix.h"
#include "SGN_Node.h"
#include "SGN_Edge.h"
#include "../math/SGN_Graph.h"
#include "../math/SGN_PCGraph.h"
#include "../math/SGN_AdjGraph.h"
#include "../csml/CSMLObject.h"
#include "SGN_Data.h"

/*!\brief Class for handling networks.
 *
 * This class is to represent a network consisting of
 * nodes (vertices) and directed edges.
 *
 * SGN_Network represents a network structure by
 * \ref SGN_Node and \ref SGN_Edge instances.
 *
 * This class is not intended for the fast processing and/or handling
 * of the graph structure.  Therefore it is not appropriate to develop
 * the some graph algorithms working on the \ref SGN_Network instance.
 *
 * \ref IO module provides functions for writing and reading an SGN_Network
 * instance to and from files.
 *
 *\see \ref SGN_Edge, \ref SGN_Node, \ref Graph, \ref PCGraph
 *\author Yoshinori Tamada &lt;tamada at is.s.u-tokyo.ac.jp&gt;
 */
typedef struct {
#ifdef DOXY
private:
#endif
  SGN_Array * nodes; /* List of nodes */
  SGN_Array * edges;
  SGN_Array * parents;
  char * primary_score;

  SGN_KeyValues * prop;
  SGN_Array * objects;
} SGN_Network;

struct SGN_Score_t;

#define SGN_Network_ERROR_VALUE  (-1)

SGN_Network * SGN_Network_new();
void      SGN_Network_delete(SGN_Network * this);
SGN_Network * SGN_Network_clone(const SGN_Network * this);
SGN_Network * SGN_Network_add_node(SGN_Network * this, SGN_Node * node);
SGN_Node *    SGN_Network_get_node(const SGN_Network * this, int v);
size_t    SGN_Network_num_nodes(const SGN_Network * this);
size_t    SGN_Network_num_edges(const SGN_Network * this);
int       SGN_Network_find_node_index(const SGN_Network * this, const char * name);
SGN_Node *    SGN_Network_find_node(const SGN_Network * this, const char * name);
int       SGN_Network_find_node_index_node(const SGN_Network * this,
									   const SGN_Node * node);
const char * SGN_Network_node_name(const SGN_Network * this, int u);
const char * SGN_Network_node_alias(const SGN_Network * this, int u);
void         SGN_Network_remove_node_index(SGN_Network * this, int v);

/* SGN_Edge Operations */
SGN_Edge *       SGN_Network_add_edge_index(SGN_Network * this, int u, int v);
SGN_Network *    SGN_Network_add_edge(SGN_Network * this, SGN_Edge * edge);
SGN_Edge *       SGN_Network_get_edge(const SGN_Network * this, size_t index);
SGN_Edge *       SGN_Network_get_edge_index(const SGN_Network * this, int u, int v);
int          SGN_Network_check_edge_index(const SGN_Network * this, int u, int v);
void         SGN_Network_remove_edge(SGN_Network * this, SGN_Edge * edge);
void         SGN_Network_remove_edge_index(SGN_Network * this, int index);



const SGN_IntArray * 
             SGN_Network_get_parent_indexes(const SGN_Network * this, int u);
SGN_IntArray *    SGN_Network_get_child_indexes(const SGN_Network * this, int u);
SGN_IntArray *   SGN_Network_copy_child_indexes(const SGN_Network * this,
										int u, SGN_IntArray * ar);
int          SGN_Network_get_parent_index(const SGN_Network * this, int u, int k);
SGN_IntArray *   SGN_Network_get_all_children(const SGN_Network * this);
int          SGN_Network_num_parents(const SGN_Network * this, int u);
int          SGN_Network_num_children(const SGN_Network * this, int u);
void         SGN_Network_get_neighbor(const SGN_Network * this, int u, SGN_IntArray * ar);

int          SGN_Network_node_degree(const SGN_Network * this, int u);

SGN_Edge *   SGN_Network_find_edge(const SGN_Network * this, const SGN_Node * parent, const SGN_Node * child);
SGN_Edge *   SGN_Network_find_edge_name(const SGN_Network * this, const char * parent, const char * child);
size_t       SGN_Network_find_edge_index(const SGN_Network * this, const SGN_Edge * edge);

SGN_Graph *    SGN_Network_graph(const SGN_Network * this);
SGN_Graph *    SGN_Network_graph_dataset(const SGN_Network * this, const SGN_Data * data);
SGN_AdjGraph * SGN_Network_AdjGraph_dataset(const SGN_Network * this, const SGN_Data * data);
SGN_PCGraph *  SGN_Network_pcgraph(const SGN_Network * network);
SGN_PCGraph *  SGN_Network_pcgraph_ord(const SGN_Network * network);
SGN_Network *  SGN_Network_from_graph_network(const SGN_Graph * graph, const SGN_Network * network);
SGN_Network *    SGN_Network_from_graph(const SGN_Graph * graph, const SGN_Data * data);
SGN_Network *    SGN_Network_from_pcgraph(const SGN_PCGraph * g, SGN_Data * data);

void         SGN_Network_set_primary_score(SGN_Network * this, const char * score_name);
const char * SGN_Network_get_primary_score(const SGN_Network * this);
void         SGN_Network_replace_node_names(SGN_Network * network, 
											const SGN_StrMatrix * table);
void         SGN_Network_replace_node_aliases(SGN_Network * this, 
											  const SGN_StrMatrix * table);
void         SGN_Network_replace_node_aliases_cols(SGN_Network * network, 
												   const SGN_StrMatrix * table,
												   const int oldcol, 
												   const int newcol);

void SGN_Network_remove_unconnected_nodes(SGN_Network * this);
SGN_Network * SGN_Network_subnetwork_index(const SGN_Network * this, SGN_IntArray * indexes);
SGN_Network * SGN_Network_subnetwork_index_dist(const SGN_Network * this, 
										const SGN_IntArray * indexes, size_t dist);
SGN_Network * SGN_Network_subnetwork_index_collect(const SGN_Network * this,
												   const SGN_IntArray * indexes,
												   int dist);
SGN_Array *   SGN_Network_collect_paths_index(const SGN_Network * this, size_t x, size_t y, 
									  size_t dist, int exact);
SGN_IntArray * SGN_Network_collect_udpaths_index(const SGN_Network * this,
										 const SGN_IntArray * children,
										 int x, int y, int dist, int exact,
										 int dir);

void SGN_Network_set_node_color(SGN_Network * this, const SGN_IntArray * indexes, 
							int r, int g, int b, int a);
void SGN_Network_set_node_color_by_names(SGN_Network * this, SGN_StrArray * names,
									 int r, int g, int b, int a);
void SGN_Network_set_node_line_width(SGN_Network * this, const SGN_IntArray * indexes, double width);
void SGN_Network_set_node_size(SGN_Network * this, const SGN_IntArray * indexes, int size);
void SGN_Network_set_edge_color(SGN_Network * this, const SGN_IntArray * indexes,
							int r, int g, int b, int a);
void SGN_Network_set_visible(SGN_Network * this, const SGN_IntArray * indexes, int visible);


SGN_Array * SGN_Network_connected_components(SGN_Network * this);

void SGN_Network_filter(SGN_Network * network, 
					   const char * key, SGN_Type type, SGN_Type_Operator op, void * value);

/* Layout routines */
void SGN_Network_layout_circle(SGN_Network * this);
void SGN_Network_layout_random(SGN_Network * network, 
						   const SGN_Array * nodes, const SGN_KeyValues * args);
void SGN_Network_layout_fit(SGN_Network * network, const SGN_KeyValues * args);
void SGN_Network_layout_hub_size(SGN_Network * this, const SGN_KeyValues * args);
void SGN_Network_layout_prop_node_size(SGN_Network * this, const char * prop, const SGN_KeyValues * args);
void SGN_Network_move_front(SGN_Network * network, SGN_ObjArray * objects);


void SGN_Network_clear_all_edge_properties(SGN_Network * this);

void SGN_Network_add_property(SGN_Network * this, const char * key, SGN_Type type, void * value);
int  SGN_Network_get_property_int(const SGN_Network * this, const char * key, int def);
void SGN_Network_add_property_int(SGN_Network * this, const char * key, int value);
void SGN_Network_add_properties_KeyValues(SGN_Network * this, const SGN_KeyValues * kv);
double SGN_Network_get_property_double(const SGN_Network * this, const char * key, double def);

void SGN_Network_get_hub_list(SGN_Network * this, int * list);
void SGN_Network_gen_prop_hub_rank(SGN_Network * this);
void SGN_Network_gen_prop_hubness(SGN_Network * this);

int SGN_Network_compare_structure(const SGN_Network * net1, const SGN_Network * net2);

double SGN_Network_get_edge_property_double(SGN_Network * network, int i, int j, char * key);

double SGN_Network_calc_score(SGN_Network * network, struct SGN_Score_t * score);

SGN_Network * SGN_Network_merge_union(SGN_Network ** this, int num);

void      SGN_Network_calc_betweenness(const SGN_Network * this, double * cb);
void      SGN_Network_gen_prop_betweenness(SGN_Network * this);
void      SGN_Network_calc_closeness(const SGN_Network * network, double * cn);

void         SGN_Network_add_CSMLObject(SGN_Network * this, CSMLObject * obj);
size_t       SGN_Network_num_CSMLObjects(const SGN_Network * this);
CSMLObject * SGN_Network_get_CSMLObject(const SGN_Network * this, int index);
void         SGN_Network_clear_CSMLObject(SGN_Network * this);
size_t       SGN_Network_total_CSMLObjects(const SGN_Network * this);
CSMLObject * SGN_Network_get_total_CSMLObject(const SGN_Network * this, size_t index);

void SGN_Network_check_validity(const SGN_Network * this);

#endif /* __SGN_NETWORK_H */

/* End of file.  Copyright (C) 2008-2011 SiGN Project Members */
