/*
  SGN_Node.h : Node in a network.

  Copyright (C) 2008-2011  SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#include "SGN_Node.h"
#include "SGN_Memory.h"

#include "SGN_Str.h"

/*!\brief Generates a new SGN_Node instance.
 *
 * Both names are copied (duplicated) into the new memory allocated
 * by this function.
 *
 *\param name  pointer to the string representing the name of the node.
 *\param alias  pointer to the string representing the alias name of the node.
 *\return newly generated SGN_Node instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
SGN_Node * SGN_Node_new(const char * name, const char * alias){
  SGN_Node * this = (SGN_Node *) malloc(sizeof(SGN_Node));
  this->name = SGN_Str_copy(name);
  this->alias = SGN_Str_copy(alias);

  this->csmlobj = CSMLObject_new();
  CSMLObject_set_type(this->csmlobj, CSMLObject_TYPE_NODE);
  CSMLObject_set_CSO_type(this->csmlobj, CSOType_MRNA);
  CSMLObject_set_line_width(this->csmlobj, SGN_Node_DEFAULT_LINE_WIDTH);
  CSMLObject_set_height(this->csmlobj, SGN_Node_DEFAULT_SIZE);
  CSMLObject_set_width(this->csmlobj, SGN_Node_DEFAULT_SIZE);

  this->index = -1;

  return this;
}
/* SGN_Node_new() -------------------------------------------------------------- */

/*!\brief Deletes the \ref SGN_Node instance.
 *\param this  pointer to the \ref SGN_Node instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
void SGN_Node_delete(void * this){
  SGN_Node * node = (SGN_Node *) this;
  
  free(node->name);
  free(node->alias);
  CSMLObject_delete(node->csmlobj);
  
  free(node);
}
/* SGN_Node_delete() ----------------------------------------------------------- */

/*!\brief Copies the \ref SGN_Node instance.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
//SGN_Node * SGN_Node_copy(const SGN_Node * this){
//  return SGN_Node_new(this->name, this->alias);
//}

/*!\brief Generates the complete copy of the \ref SGN_Node instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
SGN_Node * SGN_Node_clone(const SGN_Node * this){
  SGN_Node * node = SGN_Node_new(this->name, this->alias);

  //node->attr = SGN_KeyValues_clone(this->attr);
  node->csmlobj = CSMLObject_clone(this->csmlobj);

  return node;
}
/* SGN_Node_clone() ============================================================ */


/*!\brief Returns the \ref CSMLObject instance used in this \ref SGN_Node
 * instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
CSMLObject * SGN_Node_get_CSMLObject(const SGN_Node * this){
  return this->csmlobj;
}
/* SGN_Node_get_CSMLObject() =================================================== */

//const CSMLObject * SGN_Node_get_const_CSMLObject(const SGN_Node * this){
  //if(this->csmlobj == NULL) SGN_Node_prepare_CSMLObject(this);
  //return this->csmlobj;
//}

/*!\brief Returns the name of the node.
 *
 *\param this  pointer to the \ref SGN_Node instance.
 *\return pointer to the string representing the name of the node.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
const char * SGN_Node_name(const SGN_Node * this){
  return this->name;
}
/* SGN_Node_name() ------------------------------------------------------------- */

/*!\brief Returns the alias name of the node.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
const char * SGN_Node_alias(const SGN_Node * this){
  return this->alias;
}
/* SGN_Node_alias() ------------------------------------------------------------ */

/*!\brief Sets the new name for the node.
 *
 * The old string will be freed and the new name \a name will be copied and
 * set to the node as the new node name.
 *
 *\param this  pointer to \ref SGN_Node instance.
 *\param name  string of the node name to replace.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
void SGN_Node_set_name(SGN_Node * this, const char * name){
  free(this->name);
  this->name = SGN_Str_copy(name);
}
/* SGN_Node_set_name() --------------------------------------------------------- */

/*!\brief Sets the new alias name for the node.
 *
 * The old string will be freed and the new alias name \a alias will be copied and
 * set to the node as the new node alias name.
 *
 *\param this  pointer to \ref SGN_Node instance.
 *\param name  string of the node alias name to replace.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
void SGN_Node_set_alias(SGN_Node * this, const char * alias){
  free(this->alias);
  this->alias = SGN_Str_copy(alias);
}
/* SGN_Node_set_alias() ---------------------------------------------------- */

/*!\brief Sets the index for the node.
 *
 * The node index is used to effciently refer to the index in a network.
 * This is intended to be called after this node is set to a network.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Node */
void SGN_Node_set_index(SGN_Node * this, int index){
  this->index = index;
}
/* SGN_Node_set_index() ==================================================== */

/*\brief Returns the index of the node.
 *
 *
 *\return index in a network, or -1 if not set to a network.
 *\sa SGN_Node_set_index().
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Node */
int SGN_Node_get_index(const SGN_Node * this){
  return this->index;
}
/* SGN_Node_get_index() ==================================================== */

/*!\breif Sets the position.
 *
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
void SGN_Node_set_position(SGN_Node * this, double x, double y){
  CSMLObject_set_position(this->csmlobj, x, y);
}
/* SGN_Node_set_position() ----------------------------------------------------- */

/*!\brief Sets the relative position to the current position.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
void SGN_Node_move_position(SGN_Node * this, double x, double y){
  CSMLObject_move_position(this->csmlobj, x, y);
}
/* SGN_Node_move_position() ==================================================== */

/*!\breif Returns the position.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
double SGN_Node_get_position_x(const SGN_Node * this){
  return CSMLObject_get_position_x(this->csmlobj);
}
/* SGN_Node_get_position_x() =================================================== */

/*!\breif Returns the position.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
double SGN_Node_get_position_y(const SGN_Node * this){
  return CSMLObject_get_position_y(this->csmlobj);
}
/* SGN_Node_get_position_y() =================================================== */

/*!\brief fill color
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
void SGN_Node_set_fill_color(SGN_Node * this, const SGN_Color * color){
  CSMLObject_set_fill_color(this->csmlobj, color);
}
/* SGN_Node_set_fill_color() =================================================== */

/*!\brief line color
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
void SGN_Node_set_line_color(SGN_Node * this, const SGN_Color * color){
  CSMLObject_set_line_color(this->csmlobj, color);
}
/* SGN_Node_set_line_color() =================================================== */

/*!\brief Sets the color for this node.
 *
 * This sets both the fill color and line color at once.
 *
 *\sa SGN_Node_set_fill_color(), SGN_Node_set_line_color().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
void SGN_Node_set_color(SGN_Node * this, int r, int g, int b, int a){
  CSMLObject_set_fill_color_RGB(this->csmlobj, r, g, b, a);
  CSMLObject_set_line_color_RGB(this->csmlobj, r, g, b, a);
}
/* SGN_Node_set_color() -------------------------------------------------------- */

void SGN_Node_get_color_expr(const SGN_Node * this, char * buff){
  CSMLObject_get_fill_color_expr(this->csmlobj, buff);
}
/* SGN_Node_get_color_expr() --------------------------------------------------- */

void SGN_Node_get_line_color_expr(const SGN_Node * this, char * buff){
  CSMLObject_get_line_color_expr(this->csmlobj, buff);
}
/* SGN_Node_get_line_color_expr() ============================================== */

/*!\brief Sets the size of this node.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
void SGN_Node_set_size(SGN_Node * this, double s){
  CSMLObject_set_width(this->csmlobj, s);
  CSMLObject_set_height(this->csmlobj, s);
}
/* SGN_Node_set_size() --------------------------------------------------------- */

/*!\brief Sets the size of this node with keeping the center position.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
void SGN_Node_set_size_move(SGN_Node * this, double size){
  double x = (CSMLObject_get_width(this->csmlobj) - size) / 2.0;
  double y = (CSMLObject_get_height(this->csmlobj) - size) / 2.0;

  SGN_Node_set_size(this, size);
  SGN_Node_move_position(this, x, y);
}
/* SGN_Node_set_size_move() ==================================================== */

/*!\brief Returns the size of this node.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
double SGN_Node_get_size(const SGN_Node * this){
  return CSMLObject_get_width(this->csmlobj);
}
/* SGN_Node_get_size() --------------------------------------------------------- */

/*!\brief Sets the line width.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
void SGN_Node_set_line_width(SGN_Node * this, double width){
  CSMLObject_set_line_width(this->csmlobj, width);
}
/* SGN_Node_set_line_width() =================================================== */

/*!\brief Returns the line width.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
double SGN_Node_get_line_width(const SGN_Node * this){
  return CSMLObject_get_line_width(this->csmlobj);
}
/* SGN_Node_get_line_width() =================================================== */

/*!\brief Sets the depth of the \ref SGN_Node.
 *
 * This specifies the depth (z-order) of the \ref SGN_Node instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
void SGN_Node_set_depth(SGN_Node * this, int depth){
  CSMLObject_set_depth(this->csmlobj, depth);
}
/* SGN_Node_set_depth() ======================================================== */

/*!\brief Returns the depth of the \ref SGN_Node.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
int SGN_Node_get_depth(const SGN_Node * this){
  return CSMLObject_get_depth(this->csmlobj);
}
/* SGN_Node_get_depth() ======================================================== */

/*!\brief Returns the property associated with the specified key.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
SGN_Object * SGN_Node_get_property(const SGN_Node * this, const char * key){
  return CSMLObject_get_property(this->csmlobj, key);
}
/* SGN_Node_get_property() ===================================================== */

/*!\brief Finds the property value associated with the specified key and
 *          returns its position index.
 *
 *
 *\return index of the key, or \ref SGN_Node_ERROR_VALUE if not found.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
//size_t SGN_Node_find_property(const SGN_Node * this, const char * key){
//  return SGN_KeyValues_find_key(this->attr, key);
//}
/* SGN_Node_find_property() ==================================================== */

/*
const char * SGN_Node_get_property_key(const SGN_Node * this, size_t index){
  return SGN_KeyValues_get_key(this->attr, index);
}
*/
/* SGN_Node_get_property_key() ================================================= */

/*!\brief Returns the property value at the specified position.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
/*
void * SGN_Node_get_property_value(const SGN_Node * this, size_t index){
  return SGN_KeyValues_get_value(this->attr, index);
}
*/
/* SGN_Node_get_property_value() =============================================== */

/*!\brief Returns the value type of the property key.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Node */
/*
Type SGN_Node_get_property_type(const SGN_Node * this, size_t index){
  return SGN_KeyValues_get_type(this->attr, index);
}
*/
/* SGN_Node_get_property_type() ================================================ */


/*
void SGN_Node_add_property_string(SGN_Node * this, 
							  const char * key, const char * value){
  SGN_KeyValues_add_string(this->attr, key, value);
}
*/
/* SGN_Node_add_property_string() ---------------------------------------------- */

/*
void SGN_Node_set_property_int(SGN_Node * this,
						   const char * key, const int value){
  SGN_KeyValues_set_int(this->attr, key, value);
}
*/
/* SGN_Node_set_proprety_int() ================================================= */

/*
void SGN_Node_set_property_double(SGN_Node * this,
							  const char * key, const double value){
  SGN_KeyValues_set_double(this->attr, key, value);
}
*/
/* SGN_Node_set_property_double() ============================================== */

/*
const char * SGN_Node_get_property_string(const SGN_Node * this, const char * key){
  size_t index;
  const void * obj;
  if(key == NULL) {
	return "";
  }
  index = SGN_KeyValues_find_key(this->attr, key);
  if(index == SGN_KeyValues_ERROR_VALUE) return "";
  if(SGN_KeyValues_get_type(this->attr, index) != TYPE_STRING){
	return "";
  }
  obj = SGN_KeyValues_get_value(this->attr, index);
  if(obj == NULL) return "";
  return (const char *) obj;
}
*/
/* SGN_Node_get_property_string() ---------------------------------------------- */

/*
int SGN_Node_get_property_int(const SGN_Node * this, const char * key, int def){
  return SGN_KeyValues_get_int_def(this->attr, key, def);
}
*/
/* SGN_Node_get_property_int() ================================================= */

#include "SGN_KeyValues.h"
void SGN_Node_set_properties(SGN_Node * this, SGN_KeyValues * kv){
  //if(this->attr != NULL) SGN_KeyValues_delete(this->attr);
  //this->attr = kv;

  CSMLObject_set_properties(this->csmlobj, SGN_KeyObjects_new_KeyValues(kv));
  SGN_KeyValues_delete(kv);
}
/* SGN_Node_set_properties() =================================================== */

/*!\brief Add an integer property value to the node.
 *
 *\memberof SGN_Node */
void SGN_Node_add_property_int(SGN_Node * node, const char * key, int value){
  CSMLObject_add_property_int(node->csmlobj, key, value);
}
/* SGN_Node_add_property_int() ================================================= */

void SGN_Node_copy_properties(SGN_Node * this, const SGN_Node * node){
  /*
  if(this->attr != NULL){
	SGN_KeyValues_clear(this->attr);
  } else {
	this->attr = SGN_KeyValues_new();
  }
  SGN_KeyValues_add_KeyValues(this->attr, node->attr);
  */
  /*
  if(this->csmlobj == NULL){
	if(node->csmlobj == NULL){
	  return;
	} else {
	  SGN_Node_prepare_CSMLObject(this);
	}
  }
  */
  CSMLObject_copy_properties(this->csmlobj, node->csmlobj);
}
/* SGN_Node_copy_properties() ================================================== */

//SGN_KeyValues * SGN_Node_get_properties(const SGN_Node * this){
//CSMLObject
//CSMLObject_get_properties(this->
  //return this->attr;
//}

//size_t SGN_Node_property_size(const SGN_Node * this){
//  return SGN_KeyValues_size(this->attr);
//}

//void SGN_Node_clear_properties(SGN_Node * this){
  //if(this->attr != NULL) SGN_KeyValues_clear(this->attr);
//  if(this->csmlobj != NULL) CSMLObject_clear(this->attr);
//}
/* SGN_Node_clear_properties() ================================================= */

/* End of file.  Copyright (C) 2008-2011 Yoshinori Tamada & Univ. Tokyo          */
