/*
  SGN_Node.h : Node in a network.
  
  Copyright (C) 2008-2011  SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

#ifndef __SGN_NODE_H
#define __SGN_NODE_H

#include "../signcom.h"
#include "SGN_KeyValues.h"
#include "../csml/CSMLObject.h"

/*!\defgroup Node
 *
 *\section List of Node Properties
 *
 * <DL>
 * <DT><TT>hub.rank</TT> (integer)</DT>
 * <DD>Hub ranking begining from 1.</DD>
 *
 * <DT><TT>betweenness</TT> (double)</DT>
 * <DD>The betweenness, a kind of the centerilty of a node.</DD>
 *
 * <DT><TT>betweenness.rank</TT> (integer)</DT>
 * <DD>Betweenness ranking begining from 1.</DD>
 *
 * </DL>
 *
 * \sa \ref Network, \ref Edge
 * \author Yoshinori Tamada <tamada@ims.u-tokyo.ac.jp>
 */
typedef struct {
#ifdef DOXY
private:
#endif
  char * name;
  char * alias;

  CSMLObject * csmlobj;
  int index; /* The index in a network. */
} SGN_Node;

/*!\brief Represents an error value for index values.
 *\ingroup SGN_Node */
#ifdef DOXY
#  define SGN_Node_ERROR_VALUE
#else
#  define SGN_Node_ERROR_VALUE SIGN_SIZE_T_ERROR_VALUE
#endif

#define SGN_Node_DEFAULT_LINE_WIDTH (2.0)
#define SGN_Node_DEFAULT_SIZE (10.0)
#define SGN_Node_DEFAULT_COLOR_EXPR "0 0 255 255"


SGN_Node *   SGN_Node_new(const char *name, const char * alias);
void         SGN_Node_delete(void * this);
const char * SGN_Node_name(const SGN_Node * this);
const char * SGN_Node_alias(const SGN_Node * this);
//SGN_Node * SGN_Node_copy(const SGN_Node * this);
SGN_Node *   SGN_Node_clone(const SGN_Node * this);
void         SGN_Node_set_name(SGN_Node * this, const char * name);
void         SGN_Node_set_alias(SGN_Node * this, const char * alias);
CSMLObject * SGN_Node_get_CSMLObject(const SGN_Node * this);
void         SGN_Node_set_index(SGN_Node * this, int index);
int          SGN_Node_get_index(const SGN_Node * this);
//const 
//CSMLObject * SGN_Node_get_const_CSMLObject(const SGN_Node * this);
void         SGN_Node_set_position(SGN_Node * this, double x, double y);
double       SGN_Node_get_position_x(const SGN_Node * this);
double       SGN_Node_get_position_y(const SGN_Node * this);
void         SGN_Node_move_position(SGN_Node * this, double x, double y);
void         SGN_Node_set_color(SGN_Node * this, int r, int g, int b, int a);
void         SGN_Node_set_fill_color(SGN_Node * this, const SGN_Color * color);
void         SGN_Node_set_line_color(SGN_Node * this, const SGN_Color * color);
void         SGN_Node_get_color_expr(const SGN_Node * this, char * buff);
void         SGN_Node_get_line_color_expr(const SGN_Node * this, char * buff);
void         SGN_Node_set_size(SGN_Node * this, double s);
void         SGN_Node_set_size_move(SGN_Node * this, double size);
double       SGN_Node_get_size(const SGN_Node * this);
void         SGN_Node_set_line_width(SGN_Node * this, double width);
double       SGN_Node_get_line_width(const SGN_Node * this);
void         SGN_Node_set_depth(SGN_Node * this, int depth);
int          SGN_Node_get_depth(const SGN_Node * this);

SGN_Object * SGN_Node_get_property(const SGN_Node * this, const char * key);
/*
size_t       SGN_Node_find_property(const SGN_Node * this, const char * key);
const char * SGN_Node_get_property_key(const SGN_Node * this, size_t index);
void *       SGN_Node_get_property_value(const SGN_Node * this, size_t index);
Type         SGN_Node_get_property_type(const SGN_Node * this, size_t index);
int          SGN_Node_get_property_int(const SGN_Node * this, const char * key, int def);
void         SGN_Node_set_property_int(SGN_Node * this, const char * key, const int value);
void         SGN_Node_set_property_double(SGN_Node * this,
									  const char * key, const double value);
void         SGN_Node_add_property_string(SGN_Node * this, const char * key, const char * value);
const char * SGN_Node_get_property_string(const SGN_Node * this, const char * key);
*/
//const SGN_KeyValues * SGN_Node_get_properties(const SGN_Node * this);
void         SGN_Node_set_properties(SGN_Node * this, SGN_KeyValues * kv);
void         SGN_Node_copy_properties(SGN_Node * this, const SGN_Node * node);
void         SGN_Node_add_property_int(SGN_Node * node, const char * key, int value);
/*
void         SGN_Node_copy_properties(SGN_Node * this, const SGN_Node * node);
void         SGN_Node_clear_properties(SGN_Node * this);
size_t       SGN_Node_property_size(const SGN_Node * this);
*/

#endif /*  __SGN_NODE_H */

/* End of file.  Copyright (C) 2008-2011 SiGN Project Members */
