/*
  util/SGN_ObjArray.c : Array for Object module

  Copyright (C) 2010 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
 */

#include "SGN_ObjArray.h"

/*!\brief Generates a new \ref SGN_ObjArray instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_ObjArray */
SGN_ObjArray * SGN_ObjArray_new(){
  return SGN_Array_new();
}
/* SGN_ObjArray_new() ========================================================== */

/*!\brief Generates a new \ref SGN_ObjArray instance with the default buffer size.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_ObjArray */
SGN_ObjArray * SGN_ObjArray_news(size_t buff_size){
  return SGN_Array_news(buff_size);
}
/* SGN_ObjArray_news() ========================================================= */

void SGN_ObjArray_delete(void * this){
  SGN_Array_delete(this);
}
/* SGN_ObjArray_delete() ======================================================= */

void SGN_ObjArray_delete_all(void * this){
  SGN_Array_delete_allf((SGN_Array *) this, SGN_Object_delete_v);
}
/* SGN_ObjArray_delete_all() =================================================== */

void SGN_ObjArray_delete_ref(void * this){
  for(int i = 0; i < SGN_Array_size((SGN_Array *) this); i++){
	SGN_Object * obj = SGN_ObjArray_get((SGN_ObjArray *) this, i);
	SGN_Type_delete(SGN_Object_pointer(obj), SGN_Object_type(obj));
  }
}
/* SGN_ObjArray_delete_ref() =================================================== */

/*!\brief Generates a complete copy of this \ref SGN_ObjArray instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_ObjArray */
SGN_ObjArray * SGN_ObjArray_clone(const SGN_ObjArray * this){
  SGN_ObjArray * array = SGN_ObjArray_news(this->size);
  for(int i = 0; i < this->size; i++){
	SGN_ObjArray_add_copy(array, SGN_ObjArray_get(this, i));
  }
  return array;
}
/* SGN_ObjArray_clone() ======================================================== */

void SGN_ObjArray_clear(SGN_ObjArray * this){
  SGN_Array_clear(this);
}
/* SGN_ObjArray_clear() ======================================================= */

/*!\brief Clears the \ref SGN_ObjArray instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_ObjArray */
void SGN_ObjArray_clear_all(SGN_ObjArray * this){
  SGN_Array_clear_allf(this, SGN_Object_delete_v);
}
/* SGN_ObjArray_clear_all() ==================================================== */

size_t SGN_ObjArray_size(const SGN_ObjArray * this){
  return SGN_Array_size(this);
}
/* SGN_ObjArray_size() ========================================================= */

SGN_ObjArray * SGN_ObjArray_add(SGN_ObjArray * this, SGN_Object * object){
  return (SGN_ObjArray *) SGN_Array_add(this, object);
}
/* SGN_ObjArray_add() ========================================================== */

SGN_ObjArray * SGN_ObjArray_add_copy(SGN_ObjArray * this, const SGN_Object * object){
  return (SGN_ObjArray *) SGN_Array_add(this, SGN_Object_clone(object));
}
/* SGN_ObjArray_add_copy() ===================================================== */
 
/*!\brief Adds an objects as a void pointer with its type.
 *
 * This creates a new \ref SGN_Object and sets the given pointer to it.
 * Then, adds it to the array.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_ObjArray */
SGN_ObjArray * SGN_ObjArray_add_type(SGN_ObjArray * this, void * pointer, SGN_Type type){
  SGN_Object * object = SGN_Object_new();
  SGN_Object_set(object, pointer, type);

  return SGN_Array_add(this, object);
}
/* SGN_ObjArray_add_type() ===================================================== */

SGN_Object * SGN_ObjArray_get(const SGN_ObjArray * this, size_t index){
  return (SGN_Object *) SGN_Array_get(this, index);
}
/* SGN_ObjArray_get() ====================================================== */

/*\brief Removes the object at the specified position in the array.
 *
 * This does not delete (free) the object itself.
 *
 *\param this  pointer to the \ref SGN_ObjArray instance.
 *\param index  position index of the object to delete in the array.
 *\return pointer to the removed object.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_ObjArray */
SGN_Object * SGN_ObjArray_remove(SGN_ObjArray * this, size_t index){
  return (SGN_Object *) SGN_Array_remove(this, index);
}
/* SGN_ObjArray_remove() =================================================== */

/*\brief Removes the object at the specified position in the array.
 *
 * This also deletes (frees the resource of) the object by SGN_Object_delete().
 *
 *\param this  pointer to the \ref SGN_ObjArray instance.
 *\param index  position index of the object to delete in the array.
 *\return pointer to the removed object.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_ObjArray */
void SGN_ObjArray_remove_obj(SGN_ObjArray * this, size_t index){
  SGN_Object * obj = SGN_ObjArray_remove(this, index);
  SGN_Object_delete(obj);
}
/* SGN_ObjeArray_remove_obj() ============================================== */

//size_t SGN_ObjArray_find(const SGN_ObjArray * this, const SGN_Object * 


void SGN_ObjArray_sort(SGN_ObjArray * this, int (*cmp)(const void *, const void *)){
  qsort(this->ptr, this->size, sizeof(void *), cmp);
}

/* End of file.  Copyright (C) 2010  SiGN Project Members */
