/*
  sign/util/SGN_Str.c : Utility funtions for strings

  Copyright (C) 2008-2011 SiGN Project Members.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

/*!\defgroup SGN_Str
 * \brief Utility functions for strings.
 * \author Yoshinori Tamada <tamada@ims.u-tokyo.ac.jp>
 */

#include "SGN_Str.h"
#include "SGN_Memory.h"
#include "../signcom.h"
#include "../lang/SGN_Type.h"

#include <string.h>
#include <strings.h>
#include <ctype.h>

#ifndef SIGN_DEBUG
#define STR_BUFF_SIZE 1024
#else
#define STR_BUFF_SIZE 5
#endif

/*!\brief Returns the substring.
 *
 * If the string is shorter than the specified length,
 * the returning string will be the same string.
 *
 * This allocates the new memory for copying the substring,
 * therefore user is responsible to manage it.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Str */
char * SGN_Str_substr(const char * str, size_t length){
  char * ret = malloc(sizeof(char) * (length + 1));
  strncpy(ret, str, length);
  ret[length] = '\0';
  return ret;
}

/*!\brief Splits the string into tokens by the delimiter.
 *
 * The last element will be added even if it is an empty string.
 * For example, <TT>a;b;c;</TT> will be split into 4 tokens
 * with delimiter '<TT>;</TT>', and the last token will be the
 * empty (length = 0) string.
 *
 *\param str  pointer to the string to be spilt.
 *\param delim  delimiter character.
 *\return pointer to the \ref SGN_StrArray instance whose elements are
 *    split tokens.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Str */
SGN_StrArray * SGN_Str_split(const char * str, const char delim){
  size_t len = strlen(str);
  SGN_StrArray * ar;
  int i, tokens = 0, pos = 0;

  /* If the empty string is given, returns the empty array. */
  if(len == 0)	return SGN_StrArray_new();

  /* Count the number of tokens to allocate the minimum size of the memory */
  for(i = 0; i < len; i++){
	if(str[i] == delim) tokens++;
  }
  tokens++; /* The number of tokens is alwasy the number of delims + 1. */

  ar = SGN_StrArray_news(tokens);
  for(i = 0; i < len; i++){
	if(str[i] == delim){
	  char * token = SGN_Str_substr(str + pos, i - pos);
	  SGN_StrArray_add(ar, token);
	  pos = i + 1;
	}
  }
  /* Add the last token */
  SGN_StrArray_add(ar, SGN_Str_substr(str + pos, i - pos));
  return ar;
}
/* SGN_Str_split() ============================================================= */

/*!\brief Splits the string into tokens by the delimiter.
 *
 * Unlike SGN_Str_split(), this removes white spaces around tokens.
 * 
 *\param str  pointer to the string to be spilt.
 *\param delim  delimiter character.
 *\return pointer to the \ref SGN_StrArray instance whose elements are
 *    split tokens.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Str */
SGN_StrArray * SGN_Str_split_trim(const char * str, const char delim){
  size_t len = strlen(str);
  SGN_StrArray * ar;
  int i, tokens = 0, pos = 0;
  size_t bg, ed;

  /* If the empty string is given, returns the empty array. */
  if(len == 0)	return SGN_StrArray_new();

  /* Count the number of tokens to allocate the minimum size of the memory */
  for(i = 0; i < len; i++){
	if(str[i] == delim) tokens++;
  }
  tokens++; /* The number of tokens is alwasy the number of delims + 1. */

  ar = SGN_StrArray_news(tokens);
  for(i = 0; i < len; i++){
	if(str[i] == delim){
	  bg = pos; ed = i;
	  SGN_Str_trim2(str, &bg, &ed);
	  char * token = SGN_Str_substr(str + bg, ed - bg);
#ifdef SIGN_DEBUG
	  fprintf(SIGN_ERR, "SGN_Str_split_trim(): bg=%zu, ed=%zu, str=\"%s\"\n", bg, ed, token);
#endif
	  SGN_StrArray_add(ar, token);
	  pos = i + 1;
	}
  }

  /* Add the last token */
  bg = pos; ed = i;
  SGN_Str_trim2(str, &bg, &ed);
  SGN_StrArray_add(ar, SGN_Str_substr(str + bg, ed - bg));
#ifdef SIGN_DEBUG
  fprintf(SIGN_ERR, "SGN_Str_split_trim(): bg=%zu, ed=%zu, str=\"%s\"\n", 
		  bg, ed, SGN_StrArray_get(ar, SGN_StrArray_size(ar) - 1));
#endif
  return ar;
}
/* SGN_Str_split_trim() ======================================================== */

/*!\brief Splits the string into tokens by the delimiter.
 *
 * Unlike SGN_Str_split(), this regards consecutive delimiters as a single
 * delimiter.  Delimiters at the end of the string will be ignored.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Str */
SGN_StrArray * SGN_Str_split2(const char * str, const char delim){
  size_t len = strlen(str);
  SGN_StrArray * ar;

  if(len == 0) return SGN_StrArray_new();

  /* Count the number of tokens to allocate the minimum size of the memory */
  int prev = 0; /* 0: delimiter, 1: token */
  int tokens = 0;
  for(int i = 0; i < len; i++){
	if(str[i] != delim){
	  if(prev == 0){
		tokens++;
	  }
	  prev = 1;
	} else {
	  prev = 0;
	}
  }

  //fprintf(stderr, "(%d tokens)\n", tokens);

  ar = SGN_StrArray_news(tokens);
  prev = 0;
  int pos = 0;
  for(int i = 0; i < len; i++){
	if(str[i] != delim){
	  if(prev == 0){
		pos = i;
	  }
	  prev = 1;
	} else {
	  if(prev == 1){
		char * token = SGN_Str_substr(str + pos, i - pos);
		SGN_StrArray_add(ar, token);
	  }
	  prev = 0;
	}
  }
  if(prev == 1){
	char * token = SGN_Str_substr(str + pos, len - pos);
	SGN_StrArray_add(ar, token);
  }
  return ar;
}
/* SGN_Str_split2() ============================================================ */

/*!\brief Splits the string into tokens by the delimiter.
 *
 * Unlike SGN_Str_split(), this uses a string as a delimiter, instead of
 * a single character.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Str */
SGN_StrArray * SGN_Str_split_str(const char * str, const char * delim){
  size_t len = strlen(str);
  size_t delimlen = strlen(delim);
  SGN_StrArray * ar;
  
  if(len == 0) return SGN_StrArray_new();
  if(delimlen == 0){
	ar = SGN_StrArray_new();
	SGN_StrArray_add_copy(ar, str);
	return ar;
  }

  int tokens = 0;
  
  /* Count the number of tokens */
  for(int i = 0; i < len - delimlen + 1;){
	if(strncmp(str + i, delim, delimlen) == 0){
	  tokens++;
	  i += delimlen;
	} else {
	  i++;
	}
  }/* for i */
  tokens++;

  ar = SGN_StrArray_news(tokens);
  int pos = 0;
  for(int i = 0; i < len - delimlen + 1;){
	if(strncmp(str + i, delim, delimlen) == 0){
	  char * token = SGN_Str_substr(str + pos, i - pos);
	  SGN_StrArray_add(ar, token);
	  i += delimlen;
	  pos = i;
	} else {
	  i++;
	}
  }
  SGN_StrArray_add(ar, SGN_Str_substr(str + pos, len - pos));
  return ar;
}
/* SGN_Str_split_str() ========================================================= */

/*!\brief Counts the number of tokens in the string split by the delimiter.
 *
 * The last element will be included even if it is an empty string.
 * Therefore this returns 1 if the empty string is given.
 *
 *\param str  pointer to the string.
 *\param delim  delimiter character.
 *\return number of tokens in the string.
 *\ingroup SGN_Str */
int SGN_Str_num_tokens(const char * str, const char delim){
  int count = 0;
  const char * pos = str;
  while((pos = index(pos, delim)) != NULL) {count++; pos++;}
  count++;
  return count;
}
/* SGN_Str_num_tokens() -------------------------------------------------------- */

/*!\brief Copies the string into the newly allocated memory.
 *
 * This allocates the new memory and copies the given string
 * to it.  The user is reponsible to manage the returning pointer.
 *
 *\param str pointer to the string to be copied.
 *\return pointer to the newly allocated string that has the same
 *   contents of the given string.
 *\ingroup SGN_Str */
char * SGN_Str_copy(const char * str){
  size_t size;
  char * ret;

  if(str == NULL) return NULL;
  
  size = (strlen(str) + 1) * sizeof(char);
  ret = (char *) malloc(size);
  memcpy(ret, str, size);
  return ret;
}
/* SGN_Str_copy() -------------------------------------------------------------- */

/*!\brief Reads one line from the given file stream.
 *
 * This automatically extends the size of the buffer required to
 * read the entire line.  To do so, user needs to pass the buffer and
 * the size of it by pointers.
 * The buffer can be empty (NULL).  If so, set 0 to the variable pointed
 * by \a size.
 *
 * The newline at the end of line is removed.
 *
 * This regards all CR, CR+LF, and LF as a return character.
 *
 *\param fp  input stream to be read.
 *\param buff pointer to the pointer to the buffer.  The buffer
 *  needs to be the pointer returned by malloc and so on.
 *\param size pointer to the buffer size.
 *\return 0 if succeeded, or non-zero if failed or reached to the end of
 *  the file.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Str */
int SGN_Str_readline(FILE * fp, char ** buff, size_t * size){
  char work[STR_BUFF_SIZE];
  size_t c = 0;
  size_t prev = 0;
  int ch;

  while((ch = getc(fp)) != EOF){
	if(ch == 10){ /* LF */
	  break;
	} else if(ch == 13) { /* CR */
	  /* Try to read one more character. */
	  ch = getc(fp);
	  if(ch != 10) ungetc(ch, fp);
	  break;
	} else {
	  if(c >= STR_BUFF_SIZE - 1){
		/* working buffer full.  move them to the user buffer. */
		
		if(*size < c + prev){
		  char * new_ptr;
		  new_ptr = realloc(*buff, sizeof(char) * (c + prev));
		  *buff = new_ptr;
		  *size = c + prev;
		}
		memcpy(*buff + prev, work, sizeof(char) * c);
		prev += c;
		c = 0;
	  }
	  work[c++] = (char) ch;
	}
  }/* while() */
  
  if(feof(fp) && (c + prev) == 0){
	return 1;
  }

  /* append \0 at the end of the string. */
  work[c++] = '\0';
  
  if(*size < c + prev){
	char * new_ptr;
	new_ptr = realloc(*buff, sizeof(char) * (c + prev));
	*buff = new_ptr;
	*size = c + prev;
  }
  memcpy(*buff + prev, work, sizeof(char) * c);
  return 0;
}
/* SGN_Str_readline() ========================================================== */

/*!\brief Reads one line from the given stream.
 *
 * This automatically extends the size of the buffer required to
 * read the entire line.  To do so, user needs to pass the buffer and
 * the size of it by pointers.
 * The buffer can be empty (NULL).  If so, set 0 to the variable pointed
 * by \a size.
 *
 * The newline at the end of line is removed.
 *
 *\param fp  input stream to be read.
 *\param buff pointer to the pointer to the buffer.  The buffer
 *  needs to be the pointer returned by malloc and so on.
 *\param size pointer to the buffer size.
 *\return 0 if succeeded, or non-zero if failed or reached to the end of
 *  the file.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Str */
int SGN_Str_readline_buff(FILE * fp, char ** buff, size_t * size){
  char work[STR_BUFF_SIZE];
  char * r;
  size_t length;

  if(*size != 0){
	r = fgets(*buff, *size, fp);
	if(r == NULL) return 1; /* Error */
	length = strlen(r);
  } else {
	length = 0;
  }

  if(*size == 0 || (*buff)[length - 1] != '\n'){
	while(1){
	  size_t len;
	  char * new_ptr;
	  r = fgets(work, STR_BUFF_SIZE, fp);
	  if(r == NULL) return 1; /* Error */

	  len = strlen(work);
	  new_ptr = (char *) realloc(*buff, sizeof(char) * (length + len));
	  if(work[len - 1] == '\n'){
		memcpy(new_ptr + length, work, sizeof(char) * (len - 1));
		length += len;
		*size = length;
		*buff = new_ptr;
		break;
	  } else {
		memcpy(new_ptr + length, work, sizeof(char) * (len));
		length += len;
		*buff = new_ptr;
	  }
	}/* while(1) */
  }

  (*buff)[length - 1] = '\0';
  return 0;
}
/* SGN_Str_readline_buff() ----------------------------------------------------- */

/*!\brief Removes white spaces from the beginning and the ending.
 *
 * A white space here is one of the follwings: space, tab, and newline.
 *
 * This simply returns the first position of the given string
 * that are not one of white spaces listed above, and
 * modifies the string by putting '0' to removes white spaces
 * in the end of string.
 * Thus, do not free the returned pointer directly.
 *
 *
 *\param str pointer to the string in which white spaces in the both ends
 *    are removed.
 *\return pointer to the first position of the given string, that is not
 *    a white space.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Str */
char * SGN_Str_trim(char * str){
  size_t pos, lastpos;

  /* Find the first non-white-space character */
  for(pos = 0; str[pos] == ' '  || 
		       str[pos] == '\t' ||
		       str[pos] == '\n'; pos++);

  /* Find the last non-white-space character */
  lastpos = strlen(str);
  if(lastpos == 0) return str + pos;
  for(lastpos-- ; str[lastpos] == ' '  ||
		          str[lastpos] == '\t' ||
		          str[lastpos] == '\n'; lastpos--);

  str[lastpos + 1] = '\0';
  return str + pos;
}
/* SGN_Str_trim() -------------------------------------------------------------- */

/*!\brief Removes spaces at the beginning and the end.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Str */
char * SGN_Str_trim_sp(char * str){
  size_t pos, lastpos;

  /* Find the first non-white-space character */
  for(pos = 0; str[pos] == ' '; pos++);

  /* Find the last non-white-space character */
  lastpos = strlen(str);
  if(lastpos == 0) return str + pos;
  for(lastpos-- ; str[lastpos] == ' '; lastpos--);

  str[lastpos + 1] = '\0';
  return str + pos;
}
/* SGN_Str_trim_sp() =========================================================== */

/*!\brief Returns positions of the trimmed string.
 *
 * This is similar to SGN_Str_trim() but returns just the start and last 
 * positions of the white space-trimmed string.
 *
 * Parameters \a bg and \a ed needs to be set at first.
 * This enables users to use this for non-NULL terminated strings.
 *
 * The last positions points to the terminal charactor or
 * the first white space character appearing the end part of the string.
 * Thus ( \a ed - \a bg) represents the length of the trimmed string.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Str */
void SGN_Str_trim2(const char * str, size_t * bg, size_t * ed){
  /* Find the first non-white-space character */
  for(; str[*bg] == ' '  || 
		str[*bg] == '\t' ||
		str[*bg] == '\n'; (*bg)++);

  /* Find the last non-white-space character */
  for((*ed)-- ; str[*ed] == ' '  ||
		          str[*ed] == '\t' ||
                  str[*ed] == '\n'; (*ed)--);
  (*ed)++;
  if(*bg > *ed) *bg = *ed;
}
/* SGN_Str_trim2() ============================================================= */

/*!\brief
 *\ingroup SGN_Str */
int SGN_Str_is_num(char ch){
  int ret = 0;
  switch(ch){
  case '0': case '1': case '2': case '3': case '4':
  case '5': case '6': case '7': case '8': case '9':
	ret = 1;
	break;
  }
  return ret;
}
/* SGN_Str_is_num() ------------------------------------------------------------ */

/*!\brief Converts the latin characters to their lower cases.
 *
 *\ingroup SGN_Str */
char * SGN_Str_lower(char * str){
  char * pos = str;
  while(*pos != '\0'){
	*pos = (char) tolower(*pos);
	pos++;
  }
  return str;
}
/* SGN_Str_lower() ============================================================= */

/*\brief Converts the latin characters to their lower cases.
 *
 */
void SGN_Str_lower2(const char * str, char * buff){
  while(*str != '\0'){
	*buff = (char) tolower(*str);
	str++;
	buff++;
  }
  *buff = '\0';
}
/* SGN_Str_lower2() ======================================================== */

/*!\brief Detects the type of string value expression.
 *
 * This simply determines the given string can be
 * parsed as a double value or a string value.
 *
 *\param str string to be detected.
 *\return SGN_Type_STRING if \a str is string,
 * SGN_Type_DOUBLE if \a str can be parsed as a double value,
 * or SGN_Type_NULL if \a str is NULL.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\ingroup SGN_Str */
SGN_Type SGN_Str_detect_string_double(const char * str){

  int q = 0;
  int i = 0;
  SGN_Type t = SGN_Type_STRING;
  int flag = 1;

  if(str == NULL) return SGN_Type_NULL;
  
  while(flag){
	switch(q){
	case 0:  /* q0 */
	  switch(str[i++]){
	  case ' ': case '\t': case '\n':  break; /* do nothing */
	  case '0': case '1': case '2': case '3': case '4':
	  case '5': case '6': case '7': case '8': case '9':
		q = 1;
		break;
	  case '+': case '-': q = 4; break;
	  case '.': q = 5; break;
		
	  case '\0':
	  default:
		q = -1;
	  }
	  break;
	  
	case 1: /* q1 */
	  switch(str[i++]){
	  case '0': case '1': case '2': case '3': case '4':
	  case '5': case '6': case '7': case '8': case '9':
		/* do nothing (keep q=1) */
		break;
	  case '\0': q = -2; break; /* real */	  
	  case '.': q = 2; break;
	  default: q = -1; break;
	  }
	  break;
	  
	case 2: /* q2 */
	  switch(str[i++]){
	  case '0': case '1': case '2': case '3': case '4':
	  case '5': case '6': case '7': case '8': case '9':
		break;
	  case ' ': case '\t': case '\n':  q = 3; break;
	  case '\0': q = -2; break;
	  default: q = -1; break;
	  }
	  break;
	  
	case 3: /* q3 */
	  switch(str[i++]){
	  case ' ': case '\t': case '\n': break;
	  case '\0': q = -2; break;
	  default: q = -1; break;
	  }
	  break;
	  
	case 4: /* q4 */
	  switch(str[i++]){
	  case '0': case '1': case '2': case '3': case '4':
	  case '5': case '6': case '7': case '8': case '9':
		q = 1; break;
	  case '.': q = 5; break;
	  default: q = -1; break;
	  }
	  break;

	case 5: /* q5 */
	  switch(str[i++]){
	  case '0': case '1': case '2': case '3': case '4':
	  case '5': case '6': case '7': case '8': case '9':
		q = 2; break;

	  default: q = -1; break;
	  }
	  break;
	  
	case -1: /* terminal 1 (string) */
	  flag = 0; break;
	  
	case -2: /* terminal 2 (real) */
	  flag = 0;
	  t = SGN_Type_DOUBLE;
	  break;
	  
	default: /* error */
	  fprintf(SIGN_ERR, "FATAL ERROR: SGN_Str_detect_string_double()\n");
	  exit(1);
	}
  }/* while(flag) */

  return t;
}
/* SGN_Type SGN_Str_detect_string_double() ================================= */

/* End of file.  Copyright (C) 2008-2011 SiGN Project Members */

