/*
  util/SGN_StrArray.c : Expandable array for strings.

  Copyright (C) 2008  
    Laboratory of DNA Information Analysis, Human Genome Center,
	The Institute of Medical Science, The University of Tokyo

  DO NOT COPY, DISTRIBUTE, OR MODIFY WITHOUT PERMISSION.

  Written by Yoshinori Tamada <tamada at ims.u-tokyo.ac.jp>
 */


#include "SGN_StrArray.h"
#include "SGN_Str.h"
#include "SGN_Memory.h"
#include <string.h>
#include <strings.h>

/*!\brief Generates the new \ref SGN_StrArray instance.
 *
 * This generates the new \ref SGN_StrArray instance and initialize it.
 * The generated instance should be deleted (freed) by calling
 * SGN_StrArray_delete() when the user no longer need to use it.
 *
 *\param size  initial buffer size.
 *\return pointer to the generated \ref SGN_StrArray instance, or NULL
 *   if failed to allocate the enough memory.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
SGN_StrArray * SGN_StrArray_new(){
  return (SGN_StrArray *) SGN_Array_new();
}

/*!\brief Generates the new \ref SGN_StrArray instance.
 *
 * This generates the new \ref SGN_StrArray instance and initialize it.
 * The generated instance should be deleted (freed) by calling
 * SGN_StrArray_delete() when the user no longer need to use it.
 *
 *\param size  initial buffer size.
 *\return pointer to the generated \ref SGN_StrArray instance, or NULL
 *   if failed to allocate the enough memory.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
SGN_StrArray * SGN_StrArray_news(size_t size){
  return (SGN_StrArray *) SGN_Array_news(size);
}

/*!\brief Deletes the \ref SGN_StrArray instance.
 *
 * Note that this does not delete (free) the strings in the array.
 * The user is responsible to manage them.
 * If you want to delete all the strings in the array at the same
 * time, use SGN_StrArray_delete_all().
 *
 *\param this  pointer to the \ref SGN_StrArray instace to be deleted.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
void SGN_StrArray_delete(SGN_StrArray * this){
  SGN_Array_delete(this);
}

/*!\brief Deletes the \ref SGN_StrArray instance and its elements.
 *
 * This calls free() for all the strings (char pointer) stored
 * in the array before deleting the array itself.
 *
 *\param this  pointer to the \ref SGN_StrArray instance to be deleted.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
void SGN_StrArray_delete_all(SGN_StrArray * this){
  SGN_Array_delete_all(this);
}
/* SGN_StrArray_delete_all() =================================================== */

/*!\brief Removes all the elements in the array.
 *
 * This DOES NOT call free() for all the strings stored in the array.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
void SGN_StrArray_clear(SGN_StrArray * this){
  SGN_Array_clear(this);
}
/* SGN_StrArray_clear() ======================================================== */

/*!\brief Removes all the elements in the array.
 *
 * This calls free() for all the strings stored in the
 * array.  The size of array will become zero after calling this
 * function.   The instance itself is not deleted.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
void SGN_StrArray_clear_all(SGN_StrArray * this){
  SGN_Array_clear_all(this);
}
/* SGN_StrArray_clear_all() ---------------------------------------------------- */

/*!\brief Generates multiple \ref SGN_StrArray instances.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
SGN_StrArray * SGN_StrArray_new_ar(size_t ar_size){
  SGN_StrArray * sar = (SGN_StrArray *) malloc(sizeof(SGN_StrArray) * ar_size);

  for(size_t i = 0; i < ar_size; i++){
	SGN_Array_init((SGN_Array *) sar + i, 1);
  }/* for i < ar_size */

  return sar;
}
/* SGN_StrArray_new_ar() ======================================================= */

/*!\brief Deletes the instances allocated by \ref SGN_StrArray_new_ar().
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
void SGN_StrArray_delete_ar(SGN_StrArray * this, size_t ar_size){
  for(size_t i = 0; i < ar_size; i++){
	SGN_Array_release((SGN_Array *) this + i);
  }
  free(this);
}
/* SGN_StrArray_delete_ar() ==================================================== */

void SGN_StrArray_delete_ar_all(SGN_StrArray * this, size_t ar_size){
  for(size_t i = 0; i < ar_size; i++){
	for(size_t j = 0; j < SGN_Array_size(this + i); j++){
	  free(SGN_Array_get(this + i, j));
	}
	SGN_Array_release((SGN_Array *) this + i);
  }
  free(this);
}

void SGN_StrArray_delete_v(void * this){
  SGN_Array_delete_all((SGN_Array *) this);
}

 /*!\brief Generates a complete copy of \ref SGN_StrArray instance.
 *
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
SGN_StrArray * SGN_StrArray_clone(const SGN_StrArray * this){
  SGN_StrArray * sar = SGN_StrArray_news(this->size);
  for(size_t i = 0; i < this->size; i++){
	SGN_StrArray_add(sar, SGN_Str_copy(SGN_StrArray_get(this, i)));
  }

  return sar;
}
/* SGN_StrArray * SGN_StrArray_clone() ============================================= */

/*!\brief Generates a clone of the array of \ref SGN_StrArray instances.
 *
 *\param size  Size of the \ref SGN_StrArray array.
 *
 *\memberof SGN_StrArray */
SGN_StrArray * SGN_StrArray_clone_ar(const SGN_StrArray * this, size_t size){
  SGN_StrArray * sar = SGN_StrArray_new_ar(size);
  for(size_t i = 0; i < size; i++){
	for(size_t j = 0; j < SGN_StrArray_size(this + i); j++){
	  SGN_StrArray_add(sar + i, SGN_Str_copy(SGN_StrArray_get(this + i, j)));
	}
  }
  return sar;
}
/* SGN_StrArray_clone_ar() ===================================================== */

/*!\brief Adds a string to the array.
 *
 *\param this  pointer to the \ref SGN_StrArray instance.
 *\param str   pointer to the string to be added in the array.
 *\return \a this, or NULL if failed to expand the buffer memory.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
SGN_StrArray * SGN_StrArray_add(SGN_StrArray * this, char * str){
  return (SGN_StrArray *) SGN_Array_add(this, str);
}

/*!\brief Adds a string to the array non-redundantly.
 *
 * If the same string already exists in the array,
 * then this does do nothing.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
SGN_StrArray * SGN_StrArray_add_nr(SGN_StrArray * this, char * str){
  size_t idx = SGN_StrArray_find(this, str);
  if(idx == SGN_StrArray_ERROR_VALUE) SGN_StrArray_add(this, str);

  return this;
}
/* SGN_StrArray_add_nr() ======================================================= */

/*!\brief Adds a copy of a string to the array.
 *
 *\param this  pointer to the \ref SGN_StrArray instance.
 *\param str   pointer to the string to be added in the array.
 *\return \a this, or NULL if failed to expand the buffer memory.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
SGN_StrArray * SGN_StrArray_add_copy(SGN_StrArray * this, const char * str){
  return (SGN_StrArray *) SGN_Array_add(this, SGN_Str_copy(str));
}


/*!\brief Returns the string at the specified position in the array.
 *
 *\param this  pointer to the \ref SGN_StrArray instance.
 *\param index  position of the string to return.
 *\return pointer to the string stored at the specified position in the
 *   array.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
char * SGN_StrArray_get(const SGN_StrArray * this, size_t index){
  return (char *) SGN_Array_get(this, index);
}

/*!\brief Sets the string at the specified position of the array.
 *
 *\param this
 *\param index
 *\param char
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
SGN_StrArray * SGN_StrArray_set(SGN_StrArray * this, size_t index, char * str){
  return (SGN_StrArray *) SGN_Array_set(this, index, str);
}
/* SGN_StrArray_set() ---------------------------------------------------------- */

/*!\brief Removes a string in the \ref SGN_StrArray.
 *
 * This does NOT free the string (char pointer).
 *
 *\param this  pointer to the \ref SGN_Array instance.
 *\param index  index of the element to be removed from the array.
 * 
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_StrArray */
void SGN_StrArray_remove(SGN_StrArray * this, size_t index){
  SGN_Array_remove(this, index);
}

/*!\brief Removes a string in the \ref SGN_StrArray.
 *
 * This also frees the string (char pointer).
 *
 *\param this  pointer to the \ref SGN_Array instance.
 *\param index  index of the element to be removed from the array.
 * 
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_StrArray */
void SGN_StrArray_remove_elm(SGN_StrArray * this, size_t index){
  free(this->ptr[index]);
  SGN_Array_remove(this, index);
}
/* SGN_StrArray_remove() ------------------------------------------------------- */

SGN_StrArray * SGN_StrArray_set_size(SGN_StrArray * this, size_t size){
  return SGN_Array_set_size(this, size);
}
/* SGN_StrArray_set_size() ================================================= */

/*!\brief Returns the number of strings stored in the array.
 *\param this  pointer to the \ref SGN_StrArray instance.
 *\return the number of strings stored in the array.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
size_t SGN_StrArray_size(const SGN_StrArray * this){
  return SGN_Array_size(this);
}
/* SGN_StrArray_size() --------------------------------------------------------- */

void SGN_StrArray_copy_all(const SGN_StrArray * this, SGN_StrArray * dst){
  size_t i;
  
  if(dst->buff_size < this->size){
	SGN_Array_expand(dst, this->size - dst->buff_size);
  }
  SGN_Array_clear_all(dst);

  for(i = 0; i < this->size; i++){
	SGN_StrArray_add(dst, SGN_Str_copy((char *) this->ptr[i]));
  }
  dst->size = this->size;
}
/* SGN_StrArray_copy_all() ----------------------------------------------------- */

/*!\brief Finds the given string in the array.
 *
 * This requires at most <I>O</I>(<I>n</I>) time where <I>n</I> 
 * represents the number of strings in the array.
 *
 *\param this  pointer to the \ref SGN_StrArray instance.
 *\return index of the string in the array, or \ref SGN_StrArray_ERROR_VALUE if
 *   failed to find.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
size_t SGN_StrArray_find(const SGN_StrArray * this, const char * str){
  size_t i;
  for(i = 0; i < this->size; i++){
	if(strcmp(SGN_StrArray_get(this, i), str) == 0){
	  return i;
	}
  }
  return SGN_StrArray_ERROR_VALUE;
}
/* SGN_StrArray_find() --------------------------------------------------------- */

/*!\brief Finds the given string in the array.
 *
 * This requires at most <I>O</I>(<I>n</I>) time where <I>n</I> 
 * represents the number of strings in the array.
 *
 * Unlike SGN_StrArray_find(), this does not care letter cases (case-insensitive).
 *
 *\param this  pointer to the \ref SGN_StrArray instance.
 *\return index of the string in the array, or \ref SGN_StrArray_ERROR_VALUE if
 *   failed to find.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
size_t SGN_StrArray_find_case(const SGN_StrArray * this, const char * str){
  size_t i;
  for(i = 0; i < this->size; i++){
	if(strcasecmp(SGN_StrArray_get(this, i), str) == 0){
	  return i;
	}
  }
  return SGN_StrArray_ERROR_VALUE;
}
/* SGN_StrArray_find_case() ==================================================== */

/*!\brief Finds or adds the given string in the array.
 *
 * This requires at most <I>O</I>(<I>n</I>) time where <I>n</I> 
 * represents the number of strings in the array.
 *
 * If the given string \a str is not found, this routine adds it to the array.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_StrArray */
size_t SGN_StrArray_find_add(SGN_StrArray * this, const char * str){
  size_t p = SGN_StrArray_find(this, str);
  if(p == SGN_StrArray_ERROR_VALUE){
	p = SGN_StrArray_size(this);
	SGN_StrArray_add_copy(this, str);
  }
  return p;
}
/* SGN_StrArray_find_add() ================================================= */

SGN_IntArray * SGN_StrArray_collect_indexes(const SGN_StrArray * this, const char * str){
  SGN_IntArray * ret = SGN_IntArray_new();
  size_t i;
  for(i = 0; i < this->size; i++){
	if(strcmp(SGN_StrArray_get(this, i), str) == 0){
	  SGN_IntArray_add(ret, (int) i);
	}
  }
  return ret;
}
/* SGN_StrArray_collect_indexes() ---------------------------------------------- */

void SGN_StrArray_fprint_v(const void * v, FILE * fp){
  const SGN_StrArray * sa = (const SGN_StrArray *) v;
  if(SGN_StrArray_size(sa) >= 1){
	fprintf(fp, "%s", SGN_StrArray_get(sa, 0));
	for(int i = 1; i < SGN_StrArray_size(sa); i++){
	  fprintf(fp, ":%s", SGN_StrArray_get(sa, i));
	}
  }
}
/* SGN_StrArray_fprint_v() ================================================= */


#ifdef USE_MPI
#include "SGN_Str.h"

/*!\brief Broadcasts the array using MPI.
 *
 * This can be used only when \c USE_MPI is defined.
 *
 * This allocates working memory for sending/receving the data.
 *
 *\param this  pointer to the pointer to the \ref SGN_StrArray instance (or 
 *       variable) of the sending process (or receiving process).
 *\param root   index of the root process to broadcast.
 *\param comm   communicator of the MPI.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
void SGN_StrArray_MPI_Bcast(SGN_StrArray ** this, int root, MPI_Comm comm){
  int rank;
  unsigned long size = (*this)->size;
  unsigned long total = 0;
  size_t i;
  char * buff;

  MPI_Comm_rank(comm, &rank);
  MPI_Bcast(&size, 1, MPI_UNSIGNED_LONG, root, comm);

  if(rank != root) *this = SGN_StrArray_news(size);

  if(rank == root){
	for(i = 0; i < (*this)->size; i++){
	  total += strlen((char *) (*this)->ptr[i]) + 1;
	}
  }
  MPI_Bcast(&total, 1, MPI_UNSIGNED_LONG, root, comm);
  buff = (char *) malloc(sizeof(char) * total);
  
  if(rank == root){
	total = 0;
	for(i = 0; i < (*this)->size; i++){
	  size_t len = strlen((char *) (*this)->ptr[i]) + 1;
	  memcpy(buff + total, (*this)->ptr[i], sizeof(char) * len);
	  total += len;
	}
  }
  MPI_Bcast(buff, total, MPI_CHAR, root, comm);
  if(rank != root){
	total = 0;
	for(i = 0; i < size; i++){
	  size_t len = strlen(buff + total) + 1;
	  SGN_StrArray_add(*this, SGN_Str_copy(buff + total));
	  total += len;
	}
  }
  free(buff);
}
/* SGN_StrArray_MPI_Bcast() ---------------------------------------------------- */

/*!\brief Broadcasts the arrays using MPI.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrArray */
void SGN_StrArray_ar_MPI_Bcast(SGN_StrArray ** this, size_t * size, 
						   int root, MPI_Comm comm){
  int rank;
  unsigned long size2 = *size;
  unsigned long total;
  char * buff = NULL;
  size_t buff_size = 0;
  
  MPI_Comm_rank(comm, &rank);
  MPI_Bcast(&size2, 1, MPI_UNSIGNED_LONG, root, comm);

  if(rank != root){
	*this = SGN_StrArray_new_ar(size2);
	*size = (size_t) size2;
  }

  for(int j = 0; j < size2; j++){
	unsigned long elm_size = (unsigned long) (*this + j)->size;
	MPI_Bcast(&elm_size, 1, MPI_UNSIGNED_LONG, root, comm);
	total = 0;
	if(rank == root){
	  for(int i = 0; i < (*this + j)->size; i++){
		total += strlen((char *) (*this + j)->ptr[i]) + 1;
	  }
	}

	MPI_Bcast(&total, 1, MPI_UNSIGNED_LONG, root, comm);
	if(buff_size < total){
	  buff_size = total;
	  buff = realloc(buff, sizeof(char) * buff_size);
	}

	if(rank == root){
	  total = 0;
	  for(int i = 0; i < (*this + j)->size; i++){
		size_t len = strlen((char *) (*this + j)->ptr[i]) + 1;
		memcpy(buff + total, (*this + j)->ptr[i], sizeof(char) * len);
		total += len;
	  }
	}
	MPI_Bcast(buff, total, MPI_CHAR, root, comm);
	if(rank != root){
	  total = 0;
	  for(int i = 0; i < elm_size; i++){
		size_t len = strlen(buff + total) + 1;
		SGN_StrArray_add(*this + j, SGN_Str_copy(buff + total));
		total += len;
	  }
	}
	
  }/* for j < size2 */
  free(buff);
}
/* SGN_StrArray_ar_MPI_Bcast() ================================================= */
						   
#endif /* USE_MPI */
/* End of file.  Copyright (C) 2008  HGC, IMS, The University of Tokyo       */
