/*
  util/SGN_StrBuff.c : Expandable string buffer.

  Copyright (C) 2008-2011 SiGN Project Members.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include "SGN_StrBuff.h"
#include "SGN_Str.h"
#include "SGN_Memory.h"

/*!\brief Generates the new \ref SGN_StrBuff instance.
 *
 *
 *\return pointer to the new \ref SGN_StrBuff instance, or NULL if failed
 *  to allocate.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrBuff */
SGN_StrBuff * SGN_StrBuff_new(){
  SGN_StrBuff * this = malloc(sizeof(SGN_StrBuff));
  if(this == NULL) return NULL;
  
  this->ptr = NULL;
  this->size = 0;
  this->buff_size = 0;

  return this;
}
/* SGN_StrBuff_new() =========================================================== */

/*!\brief Deletes the \ref SGN_StrBuff instance.
 *
 *\param this  pointer to the \ref SGN_StrBuff to be deleted.  The pointer
 *               has to be one returned by SGN_StrBuff_new().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrBuff */
void SGN_StrBuff_delete(void * this){
  SGN_StrBuff * sb = (SGN_StrBuff *) this;
  free(sb->ptr);
  free(this);
}
/* SGN_StrBuff_delete() ======================================================== */

/*!\brief Expands the size of the allocated buffer memory.
 *
 * The new buffer size will be the current size + \a length.
 *
 * Generally this is used internally, and users do not need to use
 * this directly.
 *
 *\param this  pointer to the \ref SGN_StrBuff instance.
 *\param length  buffer size to be newly allocated.
 *\return \a this, or NULL if failed.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrBuff */
SGN_StrBuff * SGN_StrBuff_expand(SGN_StrBuff * this, size_t length){
  char * new_ptr;
  size_t new_size = this->buff_size + length;

  new_ptr = realloc(this->ptr, sizeof(char) * new_size);
  if(new_ptr != NULL){
	this->ptr = new_ptr;

	/* Fills it with '\0' */
	bzero(this->ptr + this->buff_size, sizeof(char) * length);
	
	this->buff_size = new_size;
	return this;
  } else {
	return NULL;
  }
}
/* SGN_StrBuff_expand() ======================================================== */

/*!\brief Adds a character to the end of the buffer.
 *
 *\param this  pointer to the \ref SGN_StrBuff instance.
 *\param c  character letter added to the \ref SGN_StrBuff.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrBuff */
SGN_StrBuff * SGN_StrBuff_add(SGN_StrBuff * this, char c){
  if(this->buff_size <= this->size + 1){
	size_t size = this->buff_size == 0 ? 10 : this->buff_size;
	SGN_StrBuff * sb = SGN_StrBuff_expand(this, size);
	if(sb == NULL) return NULL;
  }

  this->ptr[this->size++] = c;
	
  return this;
}
/* SGN_StrBuff_add() =========================================================== */

/*!\brief Adds a string to the end of the buffer.
 *
 *\param this  pointer to the \ref SGN_StrBuff instance.
 *\param str   string to be added to the \ref SGN_StrBuff.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrBuff */
SGN_StrBuff * SGN_StrBuff_add_str(SGN_StrBuff * this, const char * str){
  size_t len = strlen(str);
  size_t lenp = len + 1;
  if(this->buff_size < this->size + lenp){
	size_t size = (this->size + lenp) - this->buff_size;
	if(size < this->buff_size) size = this->buff_size;
	SGN_StrBuff * sb = SGN_StrBuff_expand(this, size);
	if(sb == NULL) return NULL;
  }

  memcpy(this->ptr + this->size, str, sizeof(char) * len);
  this->size += len;
  return this;
}
/* SGN_StrBuff_add_str() ======================================================= */

/*!\brief Returns the length of the string in the buffer.
 *
 * This returns the length of the string, excluding the terminal NULL letter.
 *
 *\param this  pointer to the \ref SGN_StrBuff instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrBuff */
size_t SGN_StrBuff_length(const SGN_StrBuff * this){
  return this->size;
}
/* SGN_StrBuff_length() ======================================================== */

/*!\brief Returns the length of the string in the buffer.
 *
 * This returns the length of the string, excluding the terminal NULL letter.
 *
 * This is the same as SGN_StrBuff_length().
 *
 *\param this  pointer to the \ref SGN_StrBuff instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrBuff */
size_t SGN_StrBuff_len(const SGN_StrBuff * this){
  return this->size;
}
/* SGN_StrBuff_len() =========================================================== */

/*!\brief Returns the pointer pointing to the current string.
 *
 * If the buffer contains nothing, in other words, if no memory space
 * is allocated for the buffer,  then this returns the constant pointer
 * to the empty string.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrBuff */
const char * SGN_StrBuff_str(const SGN_StrBuff * this){
  if(this->buff_size == 0){
	//SGN_StrBuff_expand(this, 1);
	return "";
  } else {
	return this->ptr;
  }
}
/* SGN_StrBuff_str() =========================================================== */

/*!\brief Clears the buffer.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrBuff */
void SGN_StrBuff_clear(SGN_StrBuff * this){
  if(this->size != 0){
	bzero(this->ptr, sizeof(char) * this->size);
	this->size = 0;
  }
}
/* SGN_StrBuff_clear() ========================================================= */

/*!\brief Generates a new string identical to the buffer.
 *
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_StrBuff */
char * SGN_StrBuff_gen(const SGN_StrBuff * this){
  return SGN_Str_copy(SGN_StrBuff_str(this));
}
/* SGN_StrBuff_gen() =========================================================== */

/* End of file.  Copyright (C) 2008-2011 SiGN Project Members */
