/*
  sign/util/SGN_StrMatrix.c : Matrices module for containing strings.

  Copyright (C) 2008-2012 SiGN Project Members.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

/*!\defgroup SGN_StrMatrix
 * \brief Matrices for containing strings.
 *
 * \author Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
 */

#include "SGN_StrMatrix.h"
#include "SGN_Str.h"
#include <string.h>

/*!\brief Generates an empty \ref SGN_StrMatrix instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_StrMatrix */
SGN_StrMatrix * SGN_StrMatrix_new(){
  SGN_Array * array = SGN_Array_new();
  return array;
}
/* SGN_StrMatrix_new() --------------------------------------------------------- */

/*!\brief Generates an empty \ref SGN_StrMatrix instance.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_StrMatrix */
void SGN_StrMatrix_delete(SGN_StrMatrix * this){
  int i;
  for(i = 0; i < SGN_StrMatrix_lines(this); i++){
	SGN_StrArray * sar = SGN_StrMatrix_get_line(this, i);
	SGN_StrArray_delete_all(sar);
  }
  SGN_Array_delete(this);
}
/* SGN_StrMatrix_delete() ------------------------------------------------------ */

/*!\brief Reads the string matrix from the file pointer.
 *
 * This reads the tab separated text file and stores each
 * element into the string matrix.
 *
 *\param fp  file pointer to be read.
 *\return \ref SGN_StrMatrix instance containing the contents of the file.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_StrMatrix */
SGN_StrMatrix * SGN_StrMatrix_read_fp(FILE * fp){
  char * buff = NULL;
  size_t buff_size = 0;
  SGN_Array * array = SGN_Array_new();

  /* fprintf(stderr, "SGN_StrMatrix_read_fp() called.\n"); */

  while(SGN_Str_readline(fp, &buff, &buff_size) == 0){
	SGN_StrArray * line = SGN_Str_split(buff, '\t');
	SGN_Array_add(array, line);
  }
  free(buff);

  return array;
}
/* SGN_StrMatrix_read_fp() ----------------------------------------------------- */

/*!\brief Reads the string matrix from the file pointer using a specific
 *          delimiter.
 *
 * This reads the text file in which each cell is separated by the specified
 * delimieter into a string matrix.
 *
 *\param fp  file pointer to be read.
 *\param delim delimiter
 *\return \ref SGN_StrMatrix instance containing the contents of the file.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_StrMatrix */
SGN_StrMatrix * SGN_StrMatrix_read_delim_fp(FILE * fp, char delim){
  char * buff = NULL;
  size_t buff_size = 0;
  SGN_Array * array = SGN_Array_new();

  while(SGN_Str_readline(fp, &buff, &buff_size) == 0){
	SGN_StrArray * line = SGN_Str_split(buff, delim);
	SGN_Array_add(array, line);
  }
  free(buff);

  return array;
}
/* SGN_StrMatrix_read_delim_fp() =============================================== */

/*!\brief Reads the string matrix from the file.
 *
 * This reads the tab separated text file and stores each
 * element into the string matrix.
 *
 *\param file file name to read.
 *\return \ref SGN_StrMatrix instance containing the contents of the file, or
 *   NULL if failed to read the file.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_StrMatrix */
SGN_StrMatrix * SGN_StrMatrix_read_file(const char * file){
  SGN_StrMatrix * mat;
  FILE * fp = fopen(file, "r");
  if(fp == NULL) return NULL;
  mat = SGN_StrMatrix_read_fp(fp);
  fclose(fp);
  return mat;
}
/* SGN_StrMatrix_read_file() --------------------------------------------------- */

/*!\brief Writes the matrix into the file stream.
 *\ingroup SGN_StrMatrix */
void SGN_StrMatrix_write_fp(FILE * fp, const SGN_StrMatrix * this){
  int i, j;
  for(i = 0; i < SGN_StrMatrix_lines(this); i++){
	SGN_StrArray * line = SGN_StrMatrix_get_line(this, i);
	for(j = 0; j < SGN_StrArray_size(line); j++){
	  if(j != 0) fprintf(fp, "\t");
	  fprintf(fp, "%s", SGN_StrArray_get(line, j));
	}
	fprintf(fp, "\n");
  }
}
/* SGN_StrMatrix_write_fp() ---------------------------------------------------- */

/*! Returns strings in a line
 *
 *\param this
 *\param line
 *\return \ref SGN_StrArray
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_StrMatrix */
SGN_StrArray * SGN_StrMatrix_get_line(const SGN_StrMatrix * this, size_t line){
  return (SGN_StrArray *) SGN_Array_get(this, line);
}
/* SGN_StrMatrix_get_line() ---------------------------------------------------- */

/*!\brief Returns the element at the specifid line and column.
 *
 *\param this pointer to the \ref SGN_StrMatrix instance.
 *\param line  index of the line.
 *\param col   index of the column in the line.
 *\return string at the specified position, or NULL if the specified
 *   line does not have an element at the specified column.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_StrMatrix */
char * SGN_StrMatrix_get(const SGN_StrMatrix * this, size_t line, size_t col){
  const SGN_StrArray * l = SGN_StrMatrix_get_line(this, line);
  if(col < SGN_StrArray_size(l)){
	return SGN_StrArray_get(l, col);
  } else {
	return NULL;
  }
}
/* SGN_StrMatrix_get() --------------------------------------------------------- */

/*!\brief Returns the number of rows (lines) in the matrix.
 *\param this   pointer to the \ref SGN_StrMatrix.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_StrMatrix */
size_t SGN_StrMatrix_lines(const SGN_StrMatrix * this){
  return SGN_Array_size(this);
}
/* SGN_StrMatrix_lines() ------------------------------------------------------- */

/*!\brief Returns the number of rows (lines) in the matrix.
 *
 * This simply calls SGN_StrMatrix_lines().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_StrMatrix */
size_t SGN_StrMatrix_rows(const SGN_StrMatrix * this){
  return SGN_StrMatrix_lines(this);
}
/* SGN_StrMatrix_rows() ======================================================== */

/*!\brief Adds a line.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_StrMatrix */
void SGN_StrMatrix_add_line(SGN_StrMatrix * this, SGN_StrArray * line){
  SGN_Array_add(this, line);
}
/* SGN_StrMatrix_add_line() ---------------------------------------------------- */

/*!\brief Returns a new \ref SGN_StrArray containing the all strings at the
 *    specified column position.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_StrMatrix */
SGN_StrArray * SGN_StrMatrix_get_column(const SGN_StrMatrix * this, size_t col){
  SGN_StrArray * ret = SGN_StrArray_new();
  int i;

  for(i = 0; i < SGN_StrMatrix_lines(this); i++){
	const SGN_StrArray * line = SGN_StrMatrix_get_line(this, i);
	if(SGN_StrArray_size(line) <= col){
	  SGN_StrArray_add(ret, SGN_Str_copy(""));
	} else {
	  SGN_StrArray_add(ret, SGN_Str_copy(SGN_StrArray_get(line, col)));
	}
  }
  return ret;
}
/* SGN_StrMatrix_get_column() -------------------------------------------------- */

/*!\brief Returns the number of columns at the specified row of the matrix.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_StrMatrix */
size_t SGN_StrMatrix_cols(const SGN_StrMatrix * this, size_t row){
  return SGN_StrArray_size((const SGN_StrArray *) SGN_Array_get(this, row));
}
/* SGN_StrMatrix_cols() ======================================================== */

/*!\brief Finds the line containing the specified string.
 *
 *\param this pointer to the \ref SGN_StrMatrix instance.
 *\param str  string to find in the matrix.
 *\return  index of the line containing the given string, or
 *   \ref SGN_StrArray_ERROR_VALUE if not found.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_StrMatrix */
size_t SGN_StrMatrix_find_line(const SGN_StrMatrix * this, const char * str){
  int i;
  size_t ret = SGN_StrArray_ERROR_VALUE;
  for(i = 0; i < SGN_StrMatrix_lines(this); i++){
	const SGN_StrArray * line = SGN_StrMatrix_get_line(this, i);
	size_t r = SGN_StrArray_find(line, str);

	if(r != SGN_StrArray_ERROR_VALUE) return i;
  }/* for i */
  return ret;
}
/* SGN_StrMatrix_find_line() --------------------------------------------------- */

/*!\brief Finds the line containing the specified string at the specified
 *    column.
 *
 *\retun index of the line, or \ref SGN_StrArray_ERROR_VALUE if not found.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_StrMatrix */
size_t SGN_StrMatrix_find_line_col(const SGN_StrMatrix * this, const char * str,
							   size_t col){
  size_t i;
  for(i = 0; i < SGN_StrMatrix_lines(this); i++){
	const char * cell = SGN_StrMatrix_get(this, i, col);
	if(cell != NULL && strcmp(str, cell) == 0) return i;
  }
  /* not found */
  return SGN_StrArray_ERROR_VALUE;
}
/* SGN_StrMatrix_find_line_col() ----------------------------------------------- */

/* End of file.  Copyright (C) 2008-2012 SiGN Project Members */
