/*
  sign/util/SGN_Timer.c : Timer functions

  Copyright (C) 2008-2012 SiGN Project Members

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada AT is.s.u-tokyo.ac.jp>
*/

#include "SGN_Timer.h"
#include "../signcom.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

struct SGN_Timer_t {
  /*!\brief Start time. */
  struct timeval start;

  /*!\brief Start time in CPU clock. */
  clock_t clock_start;

  /*!\brief Accumulated time in second. */
  double sec;

  /*!\brief Whether or not currently timing. */
  int timing;

  /*!\brief Timer type */
  SGN_Timer_type_t type;

  /* The number of times of timing */
  int laps;
};

/*!\brief Structure for Timer functions.
 *
 * This module is to provide timer functions for measuring
 * various running times between any points in the program.
 * Users can start and stop the timer any times and can obtain
 * the current elapsed time of the timer in various formats.
 *
 * This modules uses gettimeofday() function to measure
 * times by default.
 *
 *\author Yoshinori Tamada &lt;tamada at is.s.u-tokyo.ac.jp&gt;
 *\struct SGN_Timer
 */


/*!\brief Constructs the \ref SGN_Timer instance.
 *
 * The time is measured by gettimeofday() function by default.
 * To change it, use SGN_Timer_type().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Timer */
SGN_Timer * SGN_Timer_new(){
  SGN_Timer * timer = malloc(sizeof(SGN_Timer));
  timer->sec = 0.0;
  timer->timing = 0;
  timer->type = SGN_Timer_NORMAL;
  timer->laps = 0;

  return timer;
}
/* SGN_Timer_new() ============================================================= */

/*!\brief Construct multiple \ref SGN_Timer instances.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Timer */
SGN_Timer * SGN_Timer_new_ar(size_t size){
  SGN_Timer * timers = malloc(sizeof(SGN_Timer) * size);
  for(size_t i = 0; i < size; i++){
	timers[i].sec = 0.0;
	timers[i].timing = 0;
	timers[i].type = SGN_Timer_NORMAL;
	timers[i].laps = 0;
  }
  return timers;
}
/* SGN_Timer_new_ar() ====================================================== */

/*!\brief Deletes the \ref SGN_Timer instance.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Timer */
void SGN_Timer_delete(SGN_Timer * timer){
  free(timer);
}
/* SGN_Timer_delete() ========================================================== */

/*!\brief Sets the timer type.
 *
 * The available types are:
 * - \ref SGN_Timer_NORMAL - default type measured by gettimeofday() function.
 * - \ref SGN_Timer_CLOCK - measured by clock() function.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Timer */
void SGN_Timer_set_type(SGN_Timer * timer, SGN_Timer_type_t type){
  if(type != timer->type){
	SGN_Timer_stop(timer);
	timer->type = type;
	SGN_Timer_start(timer);
  } else {
	timer->type = type;
  }
}
/* SGN_Timer_set_type() ==================================================== */

/*!\brief Returns the timer type.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Timer */
SGN_Timer_type_t SGN_Timer_get_type(const SGN_Timer * timer){
  return timer->type;
}
/* SGN_Timer_get_type() ==================================================== */

/*!\brief Starts to measure the time.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Timer */
void SGN_Timer_start(SGN_Timer * timer){
  switch(timer->type){
  case SGN_Timer_NORMAL: gettimeofday(&timer->start, NULL); break;
  case SGN_Timer_CLOCK: timer->clock_start = clock(); break;
  }
  timer->timing = 1;
}
/* SGN_Timer_start() ======================================================= */

/*!\brief Start to measure the time with a timer type.
 *
 * This is simply an alias of calling Time_set_type() and then SGN_Timer_start().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Timer */
void SGN_Timer_start_type(SGN_Timer * timer, SGN_Timer_type_t type){
  SGN_Timer_set_type(timer, type);
  SGN_Timer_start(timer);
}
/* SGN_Timer_start_type() ================================================== */

double SGN_Timer_timeval_span(const struct timeval * start_time,
						 const struct timeval * end_time){
  double st = 
    (double) start_time->tv_sec + ((double) start_time->tv_usec / 1000000.0);
  double ed = 
    (double) end_time->tv_sec + ((double) end_time->tv_usec / 1000000.0);

  return ed - st;
}
/* SGN_Timer_timeval_span() ================================================ */

/*!\brief Stops to measure the time.
 *
 * This increases the number of laps.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Timer */
void SGN_Timer_stop(SGN_Timer * timer){
  switch(timer->type){
  case SGN_Timer_NORMAL:
	{
	  struct timeval end;
	  gettimeofday(&end, NULL);
	  /* If the timer is not working, just ignore */
	  if(timer->timing){
		timer->timing = 0;
		timer->sec += SGN_Timer_timeval_span(&timer->start, &end);
	  }
	}
	break;
  case SGN_Timer_CLOCK:
	{
	  clock_t clock_end = clock();
	  if(timer->timing){
		timer->timing = 0;
		timer->sec += (double) (clock_end - timer->clock_start) / CLOCKS_PER_SEC;
	  }
	}
	break;
  default:
	fprintf(SIGN_ERR, "ERROR: SGN_Timer: illegal timer type.\n");
	exit(1);
  }
  timer->laps++;
}
/* SGN_Timer_stop() ======================================================== */

/*!\brief Clears the timer.
 *
 * This clears the timer.  If the timer is working, then the
 * the timer stops and is cleared.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Timer */
void SGN_Timer_clear(SGN_Timer * timer){
  timer->timing = 0;
  timer->sec = 0.0;
}
/* SGN_Timer_clear() ======================================================= */

/*!\brief Clears and then starts the timer.
 *
 * This is idential to call SGN_Timer_clear() and then SGN_Timer_start().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Timer */
void SGN_Timer_clear_start(SGN_Timer * timer){
  SGN_Timer_clear(timer);
  SGN_Timer_start(timer);
}
/* SGN_Timer_clear_start() ================================================= */
  
/*!\brief Returns the measured time in second.
 *
 * NOTE: This does not stop the timer.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Timer */
double SGN_Timer_get(const SGN_Timer * timer){
  switch(timer->type){
  case SGN_Timer_NORMAL:
	{
	  struct timeval end;
	  gettimeofday(&end, NULL);
	  if(timer->timing){
		return timer->sec + SGN_Timer_timeval_span(&timer->start, &end);
	  } else {
		return timer->sec;
	  }
	}
	break;
  case SGN_Timer_CLOCK:
	{
	  clock_t clock_end = clock();
	  if(timer->timing){
		return timer->sec + (double) (clock_end - timer->clock_start) / CLOCKS_PER_SEC;
	  } else {
		return timer->sec;
	  }
	}
	break;
  default:
	fprintf(SIGN_ERR, "ERROR: SGN_Timer: illegal timer type.\n");
	exit(1);
  }
}
/* SGN_Timer_get() ========================================================= */

void SGN_Timer_add(SGN_Timer * timer, const SGN_Timer * timer2){
}
/* SGN_Timer_add() ========================================================= */

/*!\brief Returns the average time of the laps.
 *
 * Use this after calling SGN_Timer_stop().
 * This returns the average time of the laps recorded until
 * the latest calling of SGN_Timer_stop().
 * Therefore, if the timer is currently measuring time, then
 * this does not include it for calculating the average time.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\memberof SGN_Timer  */
double SGN_Timer_get_avg_lap(const SGN_Timer * timer){
  fflush(SIGN_ERR);
  return timer->sec / (double) timer->laps;
}
/* SGN_Timer_get_avg_lap() ================================================= */

void SGN_Timer_print_DDHHMMSS_sec(FILE * fp, double seconds){
  int time = (int) (seconds + 0.5);
  int sec, min, hour;

  sec = time % 60;
  time /= 60;

  min = time % 60;
  time /= 60;

  hour = time % 24;
  time /= 24;

  fprintf(fp, "%d day %02d hour %02d min %02d sec",
		  time, hour, min, sec);
}
/* SGN_Timer_print_DDHHMMSS() ============================================== */

/*!\brief Prints the current timer in day, hour, min, sec format.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof SGN_Timer */
void SGN_Timer_print_DDHHMMSS(FILE * fp, SGN_Timer * timer){
  SGN_Timer_print_DDHHMMSS_sec(fp, SGN_Timer_get(timer));
}
/* SGN_Timer_print_DDHHMMSS() ============================================== */

/* End of file.  Copyright (C) 2008-2012  SiGN Project Members. */
