/*
  util/SGN_Util.c : Utility Functions

  Copyright (C) 2008-2011 SiGN Project Members.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a part of SiGN Library.
  Visit <http://sign.hgc.jp/> for more information.

  Written by Yoshinori Tamada <tamada at is.s.u-tokyo.ac.jp>
*/

/*!\defgroup SGN_Util
 *\brief Utility functions
 *\author Yoshinori Tamada <tamada at ims.u-tokyo.ac.jp>
 */

#include "SGN_Util.h"
#include "SGN_Str.h"

#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <sys/time.h>

/*!\brief Prints a <TT>size_t</TT> value with delimiters for each 3-digit.
 *
 *\param fp  file stream.
 *\param value  value to be printed.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Util */
void SGN_Util_print_size_t_comma(FILE * fp, size_t value){

  /* 
	 The maximum 64 bit value equals to 18,446,744,073,709,551,615.
	 Thus we need only 7 3-digit blocks to keep all digits.
   */

  int block[7];
  int i;

  for(i = 0; i < 7; i++){
	block[i] = value % 1000;
	value /= 1000;
	if(value == 0) break;
  }
  fprintf(fp, "%d", block[i]);
  for(i = i - 1; i >= 0; i--){
	fprintf(fp, ",%03d", block[i]);
  }
}
/* SGN_Util_print_size_t_comma() ----------------------------------------------- */

/*!\brief Prints a <TT>int</TT> value with delimiters for each 3-digit.
 *
 *\param fp  file stream.
 *\param value  value to be printed.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Util */
void SGN_Util_print_int_comma(FILE * fp, int value){

  /* 
	 The maximum 64 bit value equals to 18,446,744,073,709,551,615.
	 Thus we need only 7 3-digit blocks to keep all digits.
   */

  int block[7];
  int i;
  int sign = 0;
  
  if(value < 0){
	sign = 1;
	value = abs(value);
  }

  for(i = 0; i < 7; i++){
	block[i] = value % 1000;
	value /= 1000;
	if(value == 0) break;
  }
  if(sign) fprintf(fp, "-");
  fprintf(fp, "%d", block[i]);
  for(i = i - 1; i >= 0; i--){
	fprintf(fp, ",%03d", block[i]);
  }
}
/* SGN_Util_print_int_comma() -------------------------------------------------- */

/*!\brief Calculates the difference between two <tt>timeval</tt> times.
 *
 * To obtain the current time in a <tt>timeval</tt> structure,
 * use gettimeofday() in <tt>sys/time.h</tt>.
 *
 *\param start_time start time to calculate the difference
 *\param end_time end time to calculate the difference
 *\return time difference in second.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Util */
double SGN_Util_timeval_span(const struct timeval * start_time,
						 const struct timeval * end_time){
  double st = 
    (double) start_time->tv_sec + ((double) start_time->tv_usec / 1000000.0);
  double ed = 
    (double) end_time->tv_sec + ((double) end_time->tv_usec / 1000000.0);

  return ed - st;
}
/* SGN_Util_timeval_span() ----------------------------------------------------- */

void SGN_Util_print_timeval_span(FILE * fp,
							 const struct timeval * start_time,
							 const struct timeval * end_time){
  double total = SGN_Util_timeval_span(start_time, end_time);
  int diff = (int) total;
  int sec, min, hour;

  sec = diff % 60;
  diff = diff / 60;

  min = diff % 60;
  diff = diff / 60;
  
  hour = diff % 24;
  diff = diff / 24;

  fprintf(fp, "%d day %02d hour %02d min %02d sec\n", diff, hour, min, sec);
  fprintf(fp, "%f\tsec in total.\n", total);

}
/* SGN_Util_print_timeval_span() ----------------------------------------------- */

void SGN_Util_sec_to_day_hour_min_sec(double in_sec,
								  int * day, int * hour, int * min, int * sec){
  int diff = (int) (in_sec + 0.5);
  *sec = diff % 60;
  diff = diff / 60;

  *min = diff % 60;
  diff = diff / 60;
  
  *hour = diff % 24;
  *day = diff / 24;
}
/* SGN_Util_sec_to_day_hour_min_sec() ========================================== */

void SGN_Util_print_sec_in_day_hour_min_sec(FILE * fp, double in_sec){
  int day, hour, min, sec;
  SGN_Util_sec_to_day_hour_min_sec(in_sec, &day, &hour, &min, &sec);
  fprintf(fp, "%d day %02d hour %02d min %02d sec\n", day, hour, min, sec);
}
/* SGN_Util_print_Sec_in_day_hour_min_sec() ==================================== */

char SGN_Util_buff[256];
const char * SGN_Util_get_svn_rev(const char * svnrev){
  char * pos;
  strcpy(SGN_Util_buff, svnrev);
  pos = index(SGN_Util_buff, ':');
  if(pos == NULL){
	pos = SGN_Util_buff;
	SGN_Util_buff[0] = '-';
	SGN_Util_buff[1] = '\0';
  } else {
	int len = strlen(svnrev);
	SGN_Util_buff[len - 2] = '\0';
	pos += 2;
  }
  return pos;
}
/* SGN_Util_get_svn_rev() ------------------------------------------------------ */

/*!\brief Splits jobs evenly and returns the start index and the size for the
 *   process.
 *
 *  This routine calculates the start index (position) and the size of the 
 *  jobs for the process indicated by argument \a rank if \a total jobs are 
 *  given.  The jobs are split into equal sizes as much as possible for \a 
 *  procs processes.
 *
 *  More exactly, this splits \a total jobs into <I>x</I> + 1 jobs to
 *  <I>y</I> processes,  and <I>x</I> jobs to \a total - <I>y</I> processes.
 *  Here <I>x</I> and <I>y</I> are returned as a value of <TT>div_t</TT>.
 *  Member <TT>quot</TT> (<TT>rem</TT>) has <I>x</I> (<I>y</I>).
 *
 *  The calculated start index and the size are set to arguments 
 *  \a start and \a size.
 *
 * This can be also used for cyclic separation of data.
 * In such a case, however, the returned value in \a start has no meaning.
 * \a start can be NULL for such a case.
 *
 *\param total
 *\param procs
 *\param rank
 *\param start  NULL acceptable.
 *\param size
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Util */
div_t SGN_Util_job_split(int total, int procs, int rank, int * start, int * size){
  div_t v;

  if(procs == 0){
	if(start != NULL) *start = -1; 
	*size = -1;
	v.quot = 0;
	v.rem = 0;
	return v;
  }

  v = div(total, procs);

  if(v.rem == 0){
	*size = v.quot;
	if(start != NULL) *start = v.quot * rank;
  } else {
	if(rank < v.rem){
	  *size = v.quot + 1;
	  if(start != NULL) *start = *size * rank;
	} else {
	  *size = v.quot;
	  if(start != NULL) *start = v.quot * rank + v.rem;
	}
  }
  return v;
}
/* SGN_Util_job_split() -------------------------------------------------------- */

/*!\brief Calculates the process assigned to the given job.
 *
 *
 *\param quot  <TT>quot</TT> member of <TT>div_t</TT> instance returned by
 *                        SGN_Util_job_split().
 *\param rem  <TT>rem</TT> member of <TT>div_t</TT> instance returned by
 *                        SGN_Util_job_split().
 *\param job  Job index.
 *\return Process ID (rank) that is assigned to job \a job.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Util */
int SGN_Util_job_rank(int quot, int rem, int job){

  div_t v;
  
  v = div(job, quot + 1);
  if(v.quot < rem){
	return v.quot;
  } else {
	return (job - rem) / quot;
  }
}
/* SGN_Util_job_rank() --------------------------------------------------------- */

/*!\brief Calculates the start position of the job assigned to the given rank.
 *
 *\param quot <tt>quot</tt> member of the <tt>div_t</tt> value returned by
 *             SGN_Util_job_split().
 *\param rem <tt>rem</tt> member of the <tt>div_t</tt> value returnd by
 *             SGN_Util_job_split().
 *\param rank  Rank to calculate the start position.
 *
 *<div align="right"><a href="#top">Go to Top</a></div>
 *\ingroup SGN_Util. */
int SGN_Util_job_start(int quot, int rem, int rank){
  if(rem == 0){
	return quot * rank;
  } else if(rank < rem){
	return (quot + 1) * rank;
  } else {
	return quot * rank + rem;
  }
}
/* SGN_Util_job_start() ======================================================== */

/*!\brief Generates unoverlapped integer random values.
 *
 * This generates \a num random values ranging from 0 to \a max - 1.
 * The generated values are not overlapped.
 *
 * This uses <TT>random()</TT> in the standard C library, therefore
 * it is required to initialize the random generator by <TT>srandom()</TT>
 * by yourself before using this function.
 *
 * Note: he current implementation is very slow. It requires <I>O</I>(\a num)
 * calculations to check overlappings.
 *
 * If \a max is smaller than \a num, then this does nothing.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Util */
void SGN_Util_random_sample(int max, int num, int * buff){
  int i, j;

  if(max < num) return;

  for(i = 0; i < num; i++){
	int r, flag;
	do {
	  flag = 0;
	  r = random() % max;

	  /* Check overlapping */
	  for(j = 0; j < i; j++){
		if(buff[j] == r){
		  flag = 1; break;
		}
	  }
	} while(flag);
	
	buff[i] = r;
  }/* for i < num */
}
/* SGN_Util_random_sample() ---------------------------------------------------- */

void SGN_Util_random_sample_ushort(unsigned short max, int num,
							   unsigned short * buff){
  int i, j;

  if(max < num) return;

  for(i = 0; i < num; i++){
	unsigned short r;
	int flag;
	do {
	  flag = 0;
	  r = random() % max;

	  /* Check overlapping */
	  for(j = 0; j < i; j++){
		if(buff[j] == r){
		  flag = 1; break;
		}
	  }
	} while(flag);
	
	buff[i] = r;
  }/* for i < num */
}
/* SGN_Util_random_sample_ushort() --------------------------------------------- */

void SGN_Util_random_sample_ushort_RNG(unsigned short max, int num,
									   unsigned short * buff, SGN_RNG * rng){
  int i, j;

  if(max < num) return;

  for(i = 0; i < num; i++){
	unsigned short r;
	int flag;
	do {
	  flag = 0;
	  int v;
	  SGN_RNG_integer(rng, &v, 1, 0, (int) max);
	  //r = random() % max;
	  r = (unsigned short) v;

	  /* Check overlapping */
	  for(j = 0; j < i; j++){
		if(buff[j] == r){
		  flag = 1; break;
		}
	  }
	} while(flag);
	
	buff[i] = r;
  }/* for i < num */
}
/* SGN_Util_random_sample_ushort_RNG() ===================================== */

/*!\brief Generates unoverlapped integer random values with some fixed values.
 *
 * This is similar to SGN_Util_random_sample() but users can specify
 * already sampled values.  Users need to set \a fix fixed values ranging
 * from 0 to \a max - 1 into the buffer pointed by \a buff.
 *
 * If argument \a fix equals to 0, the result is idential to
 * SGN_Util_random_sample().
 *
 *\param max
 *\param num
 *\param buff
 *\param fix
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Util */
void SGN_Util_random_sample_fix(int max, int num, int * buff, int fix){
  int i, j;
  if(max < num) return;

  for(i = fix; i < num; i++){
	int r, flag;
	do {
	  flag = 0;
	  r = random() % max;

	  /* Check overlapping */
	  for(j = 0; j < i; j++){
		if(buff[j] == r){
		  flag = 1; break;
		}
	  }
	} while(flag);
	
	buff[i] = r;
  }/* for i < num */
}
/* SGN_Util_random_sample_fix() ------------------------------------------------ */

/*!\brief Generates unoverlapped integer random values with some fixed values.
 *
 * Unsigned short version of SGN_Util_random_sample_fix().
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Util */
void SGN_Util_random_sample_fix_ushort(unsigned short max, int num,
								   unsigned short * buff, int fix){
  int i, j;
  if(max < num) return;

  for(i = fix; i < num; i++){
	unsigned short r;
	int flag;
	do {
	  flag = 0;
	  r = random() % max;

	  /* Check overlapping */
	  for(j = 0; j < i; j++){
		if(buff[j] == r){
		  flag = 1; break;
		}
	  }
	} while(flag);
	
	buff[i] = r;
  }/* for i < num */
}
/* SGN_Util_random_sample_fix_ushort() ----------------------------------------- */

void SGN_Util_random_sample_fix_ushort_RNG(unsigned short max, int num,
										   unsigned short * buff, int fix,
										   SGN_RNG * rng){
  int i, j;
  if(max < num) return;
  for(i = fix; i < num; i++){
	unsigned short r;
	int flag;
	do {
	  flag = 0;
	  int v;
	  SGN_RNG_integer(rng, &v, 1, 0, (int) max);
	  //r = random() % max;
	  r = (unsigned short) v;
	  /* Check overlapping */
	  for(j = 0; j < i; j++){
		if(buff[j] == r){
		  flag = 1; break;
		}
	  }
	} while(flag);
	
	buff[i] = r;
  }/* for i < num */
}
/* SGN_Util_random_sample_fix_ushort_RNG() ================================= */

/*!\brief Generates unoverlapped integer random values from the array.
 *
 *
 *\param set   array containing values to be sampled.
 *\param size  size of \a set.
 *\param num   number of values to be sampled.
 *\param buff  array where the sampled values are stored.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Util */
void SGN_Util_random_sample_set(int * set, int size, int num, int * buff){
  int i;
  if(size < num) return;

  SGN_Util_random_sample(size, num, buff);
  
  for(i = 0; i < num; i++){
	buff[i] = set[buff[i]];
  }
}
/* SGN_Util_random_sample_set() ------------------------------------------------ */

/*!\brief Generates a random permutation.
 *
 * This generates a random permutation for values from 0 to \a num - 1.
 *
 *\param buff integer array of \a num size.
 *\param num size of the array.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Util */
void SGN_Util_random_perm(int * buff, int num){
  int i;
  for(i = 0; i < num; i++) buff[i] = i;
  for(i = num - 1; i > 0; i--){
	int r = random() % (i + 1);
	int t = buff[i];
	buff[i] = buff[r];
	buff[r] = t;
  }
}
/* SGN_Util_random_perm() ------------------------------------------------------ */

int SGN_Util_int_array_sort_asc_cmp(const void * a, const void * b){
  return *((const int *) a) - *((const int *) b);
}
/* SGN_Util_int_array_sort_dec_cmp() =========================================== */

/*!\brief Sorts values of the array in ascending order .
 *
 * This sorts values of the array in ascending by the qsort function in the
 * C standard library.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Util */
void SGN_Util_int_array_sort_asc(int * ar, size_t size){
  qsort(ar, size, sizeof(int), SGN_Util_int_array_sort_asc_cmp);
}
/* SGN_Util_int_array_sort_asc() =============================================== */

int SGN_Util_int_array_sort_dec_cmp(const void * a, const void * b){
  return *((const int *) b) - *((const int *) a);
}
/* SGN_Util_int_array_sort_dec_cmp() =========================================== */

/*!\brief Sorts values of the array in descending order .
 *
 * This sorts values of the array in descending by the qsort function in the
 * C standard library.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Util */
void SGN_Util_int_array_sort_dsc(int * ar, size_t size){
  qsort(ar, size, sizeof(int), SGN_Util_int_array_sort_dec_cmp);
}
/* SGN_Util_int_array_sort_dsc() =============================================== */

uint32_t SGN_Util_badrnd32_prev = 0;

/*!\brief  Generates a bad random number.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Util */
uint32_t SGN_Util_badrnd32(){
  uint32_t a = 69069;
  uint32_t c = 1234567;

  SGN_Util_badrnd32_prev = a * SGN_Util_badrnd32_prev + c;

  return SGN_Util_badrnd32_prev;
  
}
/* SGN_Util_badrnd32() ========================================================= */

/*!\brief Sets the initial seed number of the bad random generator.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\ingroup SGN_Util */
void SGN_Util_badrnd32_seed(uint32_t seed){
  SGN_Util_badrnd32_prev = seed;
}

/* End of file.  Copyright (C) 2008-2011 SiGN Project Members */
